Package: tinytable
Type: Package
Title: Simple and Configurable Tables in 'HTML', 'LaTeX', 'Markdown',
        'Word', 'PNG', 'PDF', and 'Typst' Formats
Description: Create highly customized tables with this simple and dependency-free package. Data frames can be converted to 'HTML', 'LaTeX', 'Markdown', 'Word', 'PNG', 'PDF', or 'Typst' tables. The user interface is minimalist and easy to learn. The syntax is concise. 'HTML' tables can be customized using the flexible 'Bootstrap' framework, and 'LaTeX' code with the 'tabularray' package.
Version: 0.18.0
Imports: methods
Depends: R (>= 4.1.0)
Enhances: knitr
Suggests: base64enc, broom, data.table (>= 1.15.2), estimatr, ggplot2,
        gh, glue, htmltools, litedown (>= 0.6), magrittr,
        marginaleffects, modelsummary, pandoc, pkgdown, png, quarto,
        Rdatasets, rmarkdown, rstudioapi, scales, stringi, tibble,
        tinysnapshot (>= 0.2.0), tinytest, tinytex, xfun (>= 0.51),
        webshot2
URL: https://vincentarelbundock.github.io/tinytable/
BugReports: https://github.com/vincentarelbundock/tinytable/issues
Authors@R: c(
  person("Vincent", "Arel-Bundock",
  email = "vincent.arel-bundock@umontreal.ca",
  role = c("aut", "cre"),
  comment = c(ORCID = "0000-0003-2042-7063")))
License: GPL (>= 3)
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-08-19 20:33:12 UTC; vincent
Author: Vincent Arel-Bundock [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2042-7063>)
Maintainer: Vincent Arel-Bundock <vincent.arel-bundock@umontreal.ca>
Repository: CRAN
Date/Publication: 2026-08-20 05:20:21 UTC

# Generated by roxygen2: do not edit by hand

S3method(knitr::knit_print, tinytable)
S3method(print,tinytable)
S3method(subset,tinytable)
S3method(tt,data.frame)
S3method(tt,data.table)
S3method(tt,default)
S3method(tt,tbl_df)
S3method(xfun::record_print,tinytable)
export(format_tt)
export(format_vector)
export(group_tt)
export(knit_print.tinytable)
export(pkgdown_print.tinytable)
export(plot_tt)
export(plot_vector)
export(save_tt)
export(style_tt)
export(style_vector)
export(theme_default)
export(theme_empty)
export(theme_grid)
export(theme_html)
export(theme_latex)
export(theme_markdown)
export(theme_revealjs)
export(theme_rotate)
export(theme_striped)
export(theme_tinytable)
export(theme_tt)
export(theme_typst)
export(tt)
export(tt_format)
export(tt_group)
export(tt_plot)
export(tt_save)
export(tt_style)
exportClasses(tinytable)
exportMethods("colnames<-")
exportMethods("names<-")
exportMethods(colnames)
exportMethods(names)
exportMethods(rbind2)
importFrom(methods,rbind2)

<p align="center">

<img src="man/figures/gallery/tinytable_gallery.gif" height = "250" class = "center">
<br> <!-- badges: start -->
<a href = "https://github.com/vincentarelbundock/tinytable/blob/main/LICENSE.md" target = "_blank"><img src="https://img.shields.io/badge/license-GPLv3-blue"></a>
<a href = "https://vincentarelbundock.github.io/tinytable/" target = "_blank"><img src="https://img.shields.io/static/v1?label=Website&message=Visit&color=blue"></a>
<a href = "https://cran.r-project.org/package=tinytable" target = "_blank"><img src="https://cranlogs.r-pkg.org/badges/tinytable"></a>
<!-- badges: end -->
</p>

## What?

`tinytable` is a small but powerful `R` package to draw beautiful tables
in a variety of formats: HTML, LaTeX, Word[^1], PDF, PNG, Markdown, and
Typst. The user interface is minimalist and easy to learn, while giving
users access to powerful frameworks to create endlessly customizable
tables.

<https://vincentarelbundock.github.io/tinytable/>

> [!NOTE]
>
> The documentation on this website uses the latest development version
> of `tinytable`. This version can normally be installed from
> R-Universe:
>
> ``` r
> install.packages('tinytable', repos = c('https://vincentarelbundock.r-universe.dev', 'https://cloud.r-project.org'))
> ```
>
> or from Github:
>
> ``` r
> remotes::install_github("vincentarelbundock/tinytable")
> ```
>
> Remember to restart `R` completely after installation, to ensure that
> the latest version is loaded.
>
> The documentation for the current version of `tinytable` is available
> on the [CRAN package
> page.](https://CRAN.R-project.org/package=tinytable)

## Why?

There are already many excellent table-drawing packages in the `R`
ecosystem. Why release a new one? As [the maintainer of
`modelsummary`](https://modelsummary.com), I needed a table-drawing
package which was:

- *Simple*: Streamlined, consistent, and uncluttered user interface,
  with few functions to learn.
- *Flexible*: Expressive frameworks to customize tables in HTML and
  LaTeX formats.[^2]
- *Zero-dependency*: Avoid importing any other `R` package.[^3]
- *Concise*: Draw beautiful tables without typing a lot of code.
- *Safe*: User inputs are checked thoroughly, and informative errors are
  returned early.
- *Maintainable*: A small code base which does not rely on too many
  complex regular expressions.
- *Readable*: HTML and LaTeX code should be human-readable and editable.
- *Free*: This package will always be free. Tiny tables for a tiny
  price!

To achieve these goals, the design philosophy of `tinytable` rests on
three pillars:

1)  *Data is separate from style.* The code that this package creates
    keeps the content of a table separate from the style sheet that
    applies to its cells. This is in contrast to other `R` packages that
    modify the actual text in each cell to style it. Keeping data and
    style separate allows `tinytable` to create human-readable files
    which are easy to edit, debug, and extend. It also enables
    developers to keep a simpler code base, with minimal use of messy
    regular expressions.

2)  *Flexibility.* Users’ needs are extremely varied, and a
    table-drawing package must be flexible enough to accomodate
    different ideas. To achieve this, `tinytable` builds on
    battle-tested and versatile frameworks like `Bootstrap` for HTML and
    `tabularray` for LaTeX.

3)  Lightweight. Some of the most popular table-drawing packages in the
    `R` ecosystem are very heavy: A single `library()` call can
    sometimes load upwards of 65 `R` packages. In contrast, `tinytable`
    imports zero 3rd party `R` package by default.

## First steps

The best feature of `tinytable` is its simplicity. To draw a table,
simply call the `tt()` function with your data frame as the first
argument:

``` r
library(tinytable)

x <- mtcars[1:5, 1:5]

tt(x)
```

<p align="center">

<img src="man/figures/readme_01.png" height = "180" class = "center">
<br>

More complex tables can be created by calling arguments and chaining
functions together. In the next example, we add a caption, footnote,
colors, styles, and spanning column headers:

``` r
cap <- "A simple \\texttt{tinytable} example."
not <- "Nullam odio est, ullamcorper scelerisque lectus a, eleifend luctus nisl. Etiam ullamcorper, nibh vel interdum auctor, odio nulla mollis tortor, vel fringilla ante quam quis est."

tt(x, 
   caption = cap, 
   notes = not, 
   width = .5) |> 
 style_tt(
   i = 1:3,
   j = 1:2,
   background = "teal",
   color = "white",
   bold = TRUE) |>
 group_tt(
   j = list("Halloumi" = 1:2, "Tofu" = 4:5))
```

<p align="center">

<img src="man/figures/readme_02.png" height = "350" class = "center">
<br>

## Tutorial

The `tinytable` 0.17.0.10002 tutorial will take you much further. It is
available in HTML and PDF formats at:
<https://vincentarelbundock.github.io/tinytable/>

[^1]: Styling options in Word are somewhat limited. See the FAQ page and
    the `style_tt()` documentation for details.

[^2]: Other formats like Markdown and Typst are also available, but less
    flexible.

[^3]: Some extra packages can be imported to access specific
    functionality, such as integration with Quarto, inserting `ggplot2`
    objects as inline plots, and saving tables to PNG images or PDF
    documents.

<p align="center">

<img src="man/figures/gallery/tinytable_gallery.gif" height = "250" class = "center">
<br> <!-- badges: start -->
<a href = "https://github.com/vincentarelbundock/tinytable/blob/main/LICENSE.md" target = "_blank"><img src="https://img.shields.io/badge/license-GPLv3-blue"></a>
<a href = "https://vincentarelbundock.github.io/tinytable/" target = "_blank"><img src="https://img.shields.io/static/v1?label=Website&message=Visit&color=blue"></a>
<a href = "https://cran.r-project.org/package=tinytable" target = "_blank"><img src="https://cranlogs.r-pkg.org/badges/tinytable"></a>
<!-- badges: end -->
</p>

## What?

`tinytable` is a small but powerful `R` package to draw beautiful tables
in a variety of formats: HTML, LaTeX, Word[^1], PDF, PNG, Markdown, and
Typst. The user interface is minimalist and easy to learn, while giving
users access to powerful frameworks to create endlessly customizable
tables.

<https://vincentarelbundock.github.io/tinytable/>

> [!NOTE]
>
> The documentation on this website uses the latest development version
> of `tinytable`. This version can normally be installed from
> R-Universe:
>
> ``` r
> install.packages('tinytable', repos = c('https://vincentarelbundock.r-universe.dev', 'https://cloud.r-project.org'))
> ```
>
> or from Github:
>
> ``` r
> remotes::install_github("vincentarelbundock/tinytable")
> ```
>
> Remember to restart `R` completely after installation, to ensure that
> the latest version is loaded.
>
> The documentation for the current version of `tinytable` is available
> on the [CRAN package
> page.](https://CRAN.R-project.org/package=tinytable)

## Why?

There are already many excellent table-drawing packages in the `R`
ecosystem. Why release a new one? As [the maintainer of
`modelsummary`](https://modelsummary.com), I needed a table-drawing
package which was:

- *Simple*: Streamlined, consistent, and uncluttered user interface,
  with few functions to learn.
- *Flexible*: Expressive frameworks to customize tables in HTML and
  LaTeX formats.[^2]
- *Zero-dependency*: Avoid importing any other `R` package.[^3]
- *Concise*: Draw beautiful tables without typing a lot of code.
- *Safe*: User inputs are checked thoroughly, and informative errors are
  returned early.
- *Maintainable*: A small code base which does not rely on too many
  complex regular expressions.
- *Readable*: HTML and LaTeX code should be human-readable and editable.
- *Free*: This package will always be free. Tiny tables for a tiny
  price!

To achieve these goals, the design philosophy of `tinytable` rests on
three pillars:

1)  *Data is separate from style.* The code that this package creates
    keeps the content of a table separate from the style sheet that
    applies to its cells. This is in contrast to other `R` packages that
    modify the actual text in each cell to style it. Keeping data and
    style separate allows `tinytable` to create human-readable files
    which are easy to edit, debug, and extend. It also enables
    developers to keep a simpler code base, with minimal use of messy
    regular expressions.

2)  *Flexibility.* Users’ needs are extremely varied, and a
    table-drawing package must be flexible enough to accomodate
    different ideas. To achieve this, `tinytable` builds on
    battle-tested and versatile frameworks like `Bootstrap` for HTML and
    `tabularray` for LaTeX.

3)  Lightweight. Some of the most popular table-drawing packages in the
    `R` ecosystem are very heavy: A single `library()` call can
    sometimes load upwards of 65 `R` packages. In contrast, `tinytable`
    imports zero 3rd party `R` package by default.

## First steps

The best feature of `tinytable` is its simplicity. To draw a table,
simply call the `tt()` function with your data frame as the first
argument:

``` r
library(tinytable)

x <- mtcars[1:5, 1:5]

tt(x)
```

<p align="center">

<img src="man/figures/readme_01.png" height = "180" class = "center">
<br>

More complex tables can be created by calling arguments and chaining
functions together. In the next example, we add a caption, footnote,
colors, styles, and spanning column headers:

``` r
cap <- "A simple \\texttt{tinytable} example."
not <- "Nullam odio est, ullamcorper scelerisque lectus a, eleifend luctus nisl. Etiam ullamcorper, nibh vel interdum auctor, odio nulla mollis tortor, vel fringilla ante quam quis est."

tt(x, 
   caption = cap, 
   notes = not, 
   width = .5) |> 
 style_tt(
   i = 1:3,
   j = 1:2,
   background = "teal",
   color = "white",
   bold = TRUE) |>
 group_tt(
   j = list("Halloumi" = 1:2, "Tofu" = 4:5))
```

<p align="center">

<img src="man/figures/readme_02.png" height = "350" class = "center">
<br>

## Tutorial

The `tinytable` 0.15.2.5 tutorial will take you much further. It is
available in HTML and PDF formats at:
<https://vincentarelbundock.github.io/tinytable/>

[^1]: Styling options in Word are somewhat limited. See the FAQ page and
    the `style_tt()` documentation for details.

[^2]: Other formats like Markdown and Typst are also available, but less
    flexible.

[^3]: Some extra packages can be imported to access specific
    functionality, such as integration with Quarto, inserting `ggplot2`
    objects as inline plots, and saving tables to PNG images or PDF
    documents.

source("helpers.R")

# date
dates <- as.Date(c("2023-01-01", "2023-12-31"))
formatted_dates <- format_vector(dates, date = "%Y")
expect_equivalent(formatted_dates, c("2023", "2023"))
formatted_dates_full <- format_vector(dates, date = "%B %d, %Y")
expect_equivalent(formatted_dates_full, c("January 01, 2023", "December 31, 2023"))
formatted_vector <- format_vector(dates, date = "%Y")
expect_equivalent(formatted_vector, c("2023", "2023"))
d <- data.frame(dates = dates)
tab <- tt(d) |> format_tt(j = 1, date = "%Y")
formatted_table <- save_tt(tab, "dataframe")
expect_equivalent(formatted_table[[1]], c("dates", "2023", "2023"))


# markdown
input <- c(
  "This is **bold** text",
  "This is *italic* text",
  "This is `code` text")
output <- c(
  "This is <strong>bold</strong> text",
  "This is <em>italic</em> text",
  "This is <code>code</code> text")
expect_equivalent(format_vector(input, markdown = TRUE), output)

source("helpers.R")
using("tinysnapshot")

dat <- data.frame(
  "LaTeX_Special" = c(
    "Dollars $",
    "Percent %",
    "Underscore _",
    "Backslash \\",
    "Hash #",
    "Ampersand &"
  ),
  "HTML_Special" = c(
    "<br>",
    "<sup>4</sup>",
    "<emph>blah</emph>",
    "&amp;",
    "&lt;",
    "&gt;"
  ),
  "Mixed_Content" = c(
    "$100 & <b>bold</b>",
    "50% & <i>italic</i>",
    "user_name & <code>code</code>"
  )
)

set.seed(1024) # reproducibility of html unique IDs
tab <- tt(dat) |> format_tt(escape = TRUE)
t <- expect_table(tab, formats = c("html"))
expect_snapshot_print(t[["html"]], "escape-latex.html")
tab <- tt(dat, escape = TRUE)
t <- expect_table(tab, formats = c("html"))
expect_snapshot_print(t[["html"]], "escape-latex.html")

tab <- tt(dat, escape = TRUE)
t <- expect_table(tab, formats = c("all", "html", "typst"))
expect_snapshot_print(t[["latex"]], "escape-simple.tex")
expect_snapshot_print(t[["html"]], "escape-simple.html")
expect_snapshot_print(t[["typst"]], "escape-simple.typ")

# escape columns
dat <- data.frame(
  "blah_blah_underscore" = 1:2,
  "dollar$sign" = 3:4,
  "percent%sign" = 5:6
)
tab <- tt(dat) |> format_tt(escape = TRUE)
t <- expect_table(tab, formats = c("latex", "html", "typst"))
expect_snapshot_print(t[["latex"]], "escape-latex_colnames.tex")
expect_snapshot_print(t[["html"]], "escape-latex_colnames.html")
expect_snapshot_print(t[["typst"]], "escape-latex_colnames.typ")

# Issue #150: avoid double escaping captions
tab <- tt(
  head(iris),
  caption = "Blah blah $100 & <b>bold</b> \\label{tab:blah-blah}"
)
t <- expect_table(tab, formats = c("latex", "html", "typst"))
expect_snapshot_print(t[["latex"]], "escape-issue150_caption.tex")
expect_snapshot_print(t[["html"]], "escape-issue150_caption.html")
# backslash should be escpaed in typst, so this breaks compile
# expect_snapshot_print(t[["typst"]], "escape-issue150_caption.typ")

tab <- mtcars[1:3, 1:4] |>
  setNames(c(
    "blah_blah_underscore",
    "dollar$sign",
    "percent%sign",
    "ampersand&sign"
  )) |>
  tt() |>
  format_tt(escape = TRUE) |>
  group_tt(j = list("foo_bar" = 1:2, "banana_fish" = 3:4))
t <- expect_table(tab, formats = c("latex", "html", "typst"))
expect_snapshot_print(t[["latex"]], "escape-issue150_caption_02.tex")
expect_snapshot_print(t[["html"]], "escape-issue150_caption_02.html")
expect_snapshot_print(t[["typst"]], "escape-issue150_caption_02.typ")

# Escape caption and notes
tab <- mtcars[1:3, 1:4] |>
  setNames(c(
    "blah_blah_underscore",
    "dollar$sign",
    "percent%sign",
    "ampersand&sign"
  )) |>
  tt(
    caption = "banana_fish $100 & <b>bold</b>",
    notes = list(
      "banana_fish $100 & <b>bold</b>",
      "b" = list(i = 1, j = 1:2, text = "banana_fish $100 & <b>bold</b>")
    )
  ) |>
  format_tt(escape = TRUE) |>
  group_tt(j = list("foo_bar" = 1:2, "banana_fish" = 3:4))
t <- expect_table(tab, formats = c("latex", "html", "typst"))
expect_snapshot_print(t[["latex"]], "escape-issue150_caption_03.tex")
expect_snapshot_print(t[["html"]], "escape-issue150_caption_03.html")
expect_snapshot_print(t[["typst"]], "escape-issue150_caption_03.typ")

source("helpers.R")
using("tinysnapshot")

x <- data.frame(
  short = c("H1", "H2", "H3"),
  hypothesis = c(
    "Lorem ipsum dolor sit amet, consectetur adipisicing elit",
    "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua",
    "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi"
  )
)

# tt(x, width = c(.2, .4)) |>
#     style_tt(i = 0:4, line = "tblr") |>
#     print("latex")

source("helpers.R")
using("tinysnapshot")

# Examples from @examples section in rbind2.R
x <- tt(mtcars[1:3, 1:2], caption = "Combine two tiny tables.")
y <- tt(mtcars[4:5, 8:10])

# basic combination using rbind()
tab1 <- rbind(x, y)
expect_snapshot_print(tab1, label = "rbind2-basic_rbind.md")

# basic combination with format_tt
tab2 <- rbind(x, y) |> format_tt(replace = "")
expect_snapshot_print(tab2, label = "rbind2-basic_rbind_format.md")

# rbind2() with headers = FALSE (omit y header)
tab3 <- rbind2(x, y, headers = FALSE)
expect_snapshot_print(tab3, label = "rbind2-no_headers.md")

# rbind2() with use_names = FALSE (bind by position rather than column names)
tab4 <- rbind2(x, y, use_names = FALSE)
expect_snapshot_print(tab4, label = "rbind2-position_bind.md")

# Examples from vignettes (using a, b variable names like in vignette)
a <- tt(mtcars[1:3, 1:2], caption = "Combine two tiny tables.")
b <- tt(mtcars[4:5, 8:10])

# vignette rbind() example
tab5 <- rbind(a, b)
expect_snapshot_print(tab5, label = "rbind2-vignette_basic.md")

# vignette rbind() with format_tt
tab6 <- rbind(a, b) |> format_tt(replace = "")
expect_snapshot_print(tab6, label = "rbind2-vignette_format.md")

# vignette rbind2() with headers = FALSE
tab7 <- rbind2(a, b, headers = FALSE)
expect_snapshot_print(tab7, label = "rbind2-vignette_no_headers.md")

# vignette rbind2() with use_names = FALSE
tab8 <- rbind2(a, b, use_names = FALSE)
expect_snapshot_print(tab8, label = "rbind2-vignette_position.md")

# Test with identical column names
c <- tt(mtcars[1:2, 1:3])
d <- tt(mtcars[6:7, 1:3])
tab9 <- rbind2(c, d)
expect_snapshot_print(tab9, label = "rbind2-identical_columns.md")

# Test with notes carrying over
e <- tt(mtcars[1:2, 1:2], notes = "First table note")
f <- tt(mtcars[3:4, 1:2], notes = "Second table note")
tab10 <- rbind2(e, f)
expect_snapshot_print(tab10, label = "rbind2-notes_carryover.md")

source("helpers.R")
using("tinysnapshot")
options(tinytable_print_output = "latex")

# Create a LaTeX table with $ signs in each element type
# Testing format_tt(escape = TRUE) with different target elements

# Base data with $ signs in all elements
dat <- data.frame(
  "column$one" = c("data$1", "value$2"),
  "column$two" = c("cost$100", "price$50"),
  check.names = FALSE
)

# Create base table with all elements containing $ signs
tab <- tt(dat,
          caption = "Budget $100 analysis",
          notes = "Cost data $50-$100") |>
  group_tt(i = list("Section$A" = 1:2)) |>
  group_tt(j = list("Price$Data" = 1:2))

# No escaping (baseline)
t1 <- expect_table(tab, formats = "latex")
expect_snapshot_print(t1[["latex"]], "escape_i-no_escape.tex")

# Escape all content (default behavior)
t2 <- expect_table(tab |> format_tt(escape = TRUE), formats = "latex")
expect_snapshot_print(t2[["latex"]], "escape_i-escape_all.tex")

# Escape only column names
t3 <- expect_table(tab |> format_tt("colnames", escape = TRUE), formats = "latex")
expect_snapshot_print(t3[["latex"]], "escape_i-escape_colnames.tex")

# Escape row groups
t4 <- expect_table(tab |> format_tt("groupi", escape = TRUE), formats = "latex")
expect_snapshot_print(t4[["latex"]], "escape_i-escape_groupi.tex")

# Escape column groups
t5 <- expect_table(tab |> format_tt("groupj", escape = TRUE), formats = "latex")
expect_snapshot_print(t5[["latex"]], "escape_i-escape_groupj.tex")

# Escape caption
t6 <- expect_table(tab |> format_tt("caption", escape = TRUE), formats = "latex")
expect_snapshot_print(t6[["latex"]], "escape_i-escape_caption.tex")

# Escape notes
t7 <- expect_table(tab |> format_tt("notes", escape = TRUE), formats = "latex")
expect_snapshot_print(t7[["latex"]], "escape_i-escape_notes.tex")

# Selective escaping - multiple elements
t8 <- expect_table(tab |> 
                   format_tt("colnames", escape = TRUE) |>
                   format_tt("caption", escape = TRUE), 
                   formats = "latex")
expect_snapshot_print(t8[["latex"]], "escape_i-selective_escape.tex")

# All elements escaped
t9 <- expect_table(tab |> 
                   format_tt(c("colnames", "groupi", "groupj", "caption", "notes"), escape = TRUE),
                   formats = "latex")
expect_snapshot_print(t9[["latex"]], "escape_i-escape_all_elements.tex")

source("helpers.R")
using("tinysnapshot")

dat <- do.call(rbind, by(iris, ~Species, head, n = 2))
tab <- tt(dat) |>
  group_tt(i = Species) |>
  subset(select = -Species)
expect_snapshot_print(tab, label = "methods-groupi-subset.md")


tab <- tt(mtcars[1:4, ]) |>
  subset(select = c(am, mpg, cyl))
expect_snapshot_print(tab, label = "methods-subset_01.md")

source("helpers.R")
using("tinysnapshot")
options(tinytable_print_output = "html")

x <- mtcars[1:4, 1:5]

tab <- tt(x, theme = "striped")
expect_snapshot_print(print_html(tab), "html-striped.html")

tab <- tt(x, theme = "striped") |>
  style_tt(color = "orange")
expect_snapshot_print(print_html(tab), "html-striped_orange.html")

# Issue #92: header alignment
k <- structure(
  list(
    Column1 = c("Some text", "123"),
    Column2 = c(
      "Some text",
      "456"
    )
  ),
  row.names = c(NA, -2L),
  class = "data.frame"
)
tab <- tt(k, width = 1) |> style_tt(j = 2, align = "r")
expect_snapshot_print(print_html(tab), "html-issue92.html")

# tutorial.qmd: vectorized settings
tab <- tt(x) |>
  style_tt(
    j = 2:3,
    color = c("orange", "green"),
    background = "black"
  )
expect_snapshot_print(print_html(tab), "html-vectorized_color_j.html")

# Issue #58
tab <- tt(iris[1:10, ]) |>
  style_tt(align = "c") |>
  group_tt(j = list("Sepal" = 1:2, "Petal" = 3:4))
expect_snapshot_print(print_html(tab), "html-issue58.html")

# Issue #88: indent
tab <- tt(iris[1:10, ]) |>
  style_tt(i = 2:4, indent = 3)
expect_snapshot_print(print_html(tab), "html-issue88.html")

# tutorial.qmd: heatmap
k <- data.frame(matrix(1:20, ncol = 5))
colnames(k) <- NULL
bg <- hcl.colors(20, "Inferno")
fg <- ifelse(as.matrix(k) < 17, tail(bg, 1), head(bg, 1))
fs <- 1:20
tab <- tt(k, width = .5, theme = "empty") |>
  style_tt(
    i = 1:4,
    j = 1:5,
    color = fg,
    background = bg,
    fontsize = fs
  )
expect_snapshot_print(print_html(tab), "html-heatmap.html")

# Caption
tab <- tt(mtcars[1:3, 1:3], caption = "Blah blah")
expect_snapshot_print(print_html(tab), "html-caption.html")

# Footnote
tab <- tt(mtcars[1:3, 1:3], notes = list(a = "Blah.", b = "Blah blah."))
expect_snapshot_print(print_html(tab), "html-footnote.html")

# Style individual cells
tab <- tt(mtcars[1:4, 1:4]) |>
  style_tt(
    i = 2:3,
    j = c(1, 3, 4),
    italic = TRUE,
    background = "pink",
    color = "orange"
  )
expect_snapshot_print(print_html(tab), "html-individual_cells.html")

# Issue #432: S4 refactor broke bootstrap_css argument
x <- mtcars[1:4, 1:5]
tab <- tt(x) |>
  style_tt(j = 1, bootstrap_css = "font-weight: bold; color: red;")
expect_snapshot_print(print_html(tab), "html-bootstrap_css.html")

# Line break
d <- data.frame(
  "{Sed ut \\\\ perspiciatis unde}",
  "dicta sunt<br> explicabo. Nemo"
) |>
  setNames(c("LaTeX line break", "HTML line break"))
d <- tt(d)
expect_snapshot_print(print_html(d), "html-line_break.html")

# Formatting
dat <- data.frame(
  w = c(143002.2092, 201399.181, 100188.3883),
  x = c(1.43402, 201.399, 0.134588),
  y = as.Date(c(999, 675, 3), origin = "1970-01-01"),
  z = c(TRUE, TRUE, FALSE)
)
dat <- tt(dat, digits = 2)
expect_snapshot_print(print_html(dat), "html-formatting.html")

# Missing value replacement
tab <- data.frame(a = c(NA, 1, 2), b = c(3, NA, 5))
tab <- tt(tab)
tab <- format_tt(tab, replace = "-")
expect_snapshot_print(print_html(tab), "html-missing_value.html")

# Alignment
dat <- data.frame(
  a = c("a", "aa", "aaa"),
  b = c("b", "bb", "bbb"),
  c = c("c", "cc", "ccc")
)
dat <- tt(dat) |> style_tt(j = 1:3, align = "lcr")
expect_snapshot_print(print_html(dat), "html-alignment.html")

# Font size
x <- mtcars[1:4, 1:5]
x <- tt(x) |> style_tt(j = "mpg|hp|qsec", fontsize = 1.5)
expect_snapshot_print(print_html(x), "html-font_size.html")

# Merge cells
x <- (mtcars[1:4, 1:5])
x <- tt(x) |>
  style_tt(
    i = 2,
    j = 2,
    colspan = 3,
    rowspan = 2,
    align = "c",
    alignv = "m",
    color = "white",
    background = "black",
    bold = TRUE
  )
expect_snapshot_print(print_html(x), "html-merge_cells.html")

# Spanning cells
tab <- aggregate(mpg ~ cyl + am, FUN = mean, data = mtcars)
tab <- tab[order(tab$cyl, tab$am), ]
tab <- tt(tab, digits = 2) |>
  style_tt(i = c(1, 3, 5), j = 1, rowspan = 2, alignv = "t")
expect_snapshot_print(print_html(tab), "html-spanning_cells.html")

# Omit headers
k <- (mtcars[1:4, 1:5])
colnames(k) <- NULL
k <- tt(k)
expect_snapshot_print(print_html(k), "html-omit_headers.html")

# Conditional styling
k <- mtcars[1:10, c("mpg", "am", "vs")]
k <- tt(k) |>
  style_tt(
    i = which(k$am == k$vs),
    background = "teal",
    color = "white"
  )
expect_snapshot_print(print_html(k), "html-conditional_styling.html")

# Borders
x <- tt(mtcars[1:4, 1:5], theme = "empty") |>
  style_tt(
    i = 0:3,
    j = 1:3,
    line = "tblr",
    line_width = 0.4,
    line_color = "orange"
  )
expect_snapshot_print(print_html(x), "html-borders.html")

# Images
if (Sys.info()["sysname"] == "Darwin") {
  dat <- data.frame(
    Species = c("Spider", "Squirrel"),
    Image = ""
  )
  img <- c(
    "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcSeLPSPrPtVgPg6BLCiN6lBYy8l1xNy0T5yttVjkIk0L3Rva8Zl",
    "https://encrypted-tbn2.gstatic.com/images?q=tbn:ANd9GcQdBlFVajljNz5qMbO622ihkIU2r6yA5whM9b8MbRGKOfJ8_UmZ"
  )
  dat <- tt(dat) |>
    plot_tt(j = 2, images = img, height = 3)
  expect_snapshot_print(print_html(dat), "html-images.html")
}

# Issue #297: group_tt() breaks alignment
if (Sys.info()["sysname"] == "Darwin") {
  tab <- data.frame(
    Person = c("Alice", "Bob", "Charlemagne"),
    Fruit = c("Apple", "Banana", "Cantaloupe"),
    Count = c(4, 238432, 32)
  ) |>
    tt() |>
    group_tt(i = list("Thing" = 1, "Thing again" = 2)) |>
    style_tt(i = c(1, 3), align = "l") |>
    style_tt(j = 1:3, align = "l")
  expect_snapshot_print(print_html(dat), "html-issue297.html")
}


# # Issue #355a: rowspan breaks indexing
tab <- tt(mtcars[1:6, 1:5]) |>
  group_tt(j = list("Hello" = 1:2, "World" = 3:5)) |>
  group_tt(j = list("Foo" = 2:3, "Bar" = 4:5)) |>
  style_tt(i = c(1, 3, 5), j = 1:2, bold = TRUE, color = "red", rowspan = 2) |>
  style_tt(i = 1 + c(1, 3, 5), j = 3, bold = TRUE, color = "green") |>
  style_tt(i = c(1, 3, 5), j = 3, bold = TRUE, color = "orange") |>
  style_tt(i = 0, background = "black", color = "white") |>
  style_tt(i = -1, color = "pink") |>
  style_tt(i = -2, color = "blue") |>
  style_tt(i = 6, background = "pink")
expect_snapshot_print(print_html(tab), "html-issue355a.html")

tab <- tt(mtcars[1:9, 1:8]) |>
  group_tt(
    j = list("Foo" = 1:3, "Bar" = 4:8),
    i = list(
      "I like (fake) hamburgers" = 3,
      "She prefers halloumi" = 4,
      "They love tofu" = 7
    )
  ) |>
  style_tt(
    i = c(3, 5, 9),
    align = "c",
    color = "white",
    background = "gray",
    bold = TRUE
  ) |>
  style_tt(4, background = "pink")
expect_snapshot_print(print_html(tab), "html-issue355b.html")



# Issue #575
tab <- head(iris) |>
  tt() |>
  theme_empty() |>
  group_tt(j = list("Hello" = 1:2, "Cruel" = 3, "World" = 4:5)) |>
  group_tt(j = list("Foo" = 1, "Bar" = 2:5)) |>
  style_tt(align = "c") |>
  style_tt(i = 2, j = 2, colspan = 3) |>
  style_tt(i = 2, j = 2, line = "b", line_color = "green") |>
  style_tt(i = 2, j = 5, line = "b", line_color = "orange") |>
  style_tt(i = -1, j = 1:2, line = "b", line_color = "orange") |>
  style_tt(i = -2, j = 3, line = "b", line_color = "green") |>
  style_tt(i = -1, j = 4:5, line = "b", line_color = "blue") |>
  style_tt(i = 0, line = "b", line_color = "pink") |>
  style_tt(i = 6, line = "b", line_color = "pink")
expect_snapshot_print(print_html(tab), "html-issue575.html")




## TODO: reinstate portable test, but there's a snapshot challenge
# # Issue #340: plot_tt should be able to create self-contained HTML
# if (Sys.info()["user"] == "vincent") {
#     dat <- data.frame(
#         Name = c("bar", "line"),
#         Image = "") |>
#         tt() |>
#         plot_tt(i = 1, j = 2, fun = "bar", data = list(2)) |>
#         plot_tt(i = 2, j = 2, fun = "line", data = list(data.frame(x = 1:3, y = 3:1)))
#     expect_snapshot_print(print_html(dat, "html_portable"), "html-images-portable.html")
#     op = options("tinytable_html_portable" = TRUE)
#     expect_snapshot_print(print_html(dat, "html"), "html-images-portable.html")
#     options(op)
# }

# # Built-in plots
# # cannot be tested because the names of plots are random and set seed doesn't work
# set.seed(1024)
# plot_data <- list(mtcars$mpg, mtcars$hp, mtcars$qsec)
# dat <- data.frame(
#   Variables = c("mpg", "hp", "qsec"),
#   Histogram = "",
#   Density = "",
#   Bar = "",
#   Line = ""
# )
# lines <- lapply(1:3, \(x) data.frame(x = 1:10, y = rnorm(10)))
# dat<-tt(dat) |>
#   plot_tt(j = 2, fun = "histogram", data = plot_data) |>
#   plot_tt(j = 3, fun = "density", data = plot_data, color = "darkgreen") |>
#   plot_tt(j = 4, fun = "bar", data = list(2, 3, 6), color = "orange") |>
#   plot_tt(j = 5, fun = "line", data = lines, color = "blue") |>
#   style_tt(j = 2:5, align = "c")
# expect_snapshot_print(print_html(dat), "html-built_in_plots")

options(tinytable_print_output = NULL)

source("helpers.R")

input <- style_vector(1:2, color = c("orange", "teal"))
output <- c(
  "<span style='color:#FFA500'>1</span>",
  "<span style='color:#008080'>2</span>")
expect_equivalent(input, output)

source("helpers.R")
using("tinysnapshot")

# latex table with cell marker table
options(tinytable_print_output = "latex")
x <- mtcars[1:3, 1:3]
n <- list("Blah blah", "*" = list(i = 0:1, j = 2, text = "foo bar"))
tab <- tt(x, notes = n)
expect_snapshot_print(tab, label = "notes-latex_cell_markers.tex")
options(tinytable_print_output = NULL)

# s4 refactor bug along the way
x <- mtcars[1:3, 1:3]
tab <- save_tt(tt(x, notes = "Hello World!"), "markdown")
expect_inherits(tab, "character")

source("helpers.R")
using("tinysnapshot")
requiet("tibble")
requiet("pillar")
exit_file("fail on github actions")

# Issue #426
x <- tibble(a = c(pi, 2.3, 1), b = letters[1:3])
x <- format_tt(x, digits = 1)
expect_inherits(x, "tbl_df")
expect_equivalent(x[[1]], c("3", "2", "1"))
expect_equivalent(x[[2]], c("a", "b", "c"))

# Issue #141
k2 <- tibble(
  x3 = 9:11 * 100 + 0.5,
  x4 = 9:11 * 100 + 0.5,
  x5 = 9:11 * 100 + 0.5,
)
tab <- tt(k2, digits = 1)

options("tinytable_print_output" = "markdown")
expect_snapshot_print(tab, "tibble-tibble_issue141")


options("tinytable_print_output" = NULL)
x <- tibble(
  x3 = pillar::num(9:11 * 100 + 0.5, sigfig = 3),
  x4 = pillar::num(9:11 * 100 + 0.5, sigfig = 4),
  x5 = pillar::num(9:11 * 100 + 0.5, sigfig = 5),
)
tab <- tt(x)
t <- expect_table(tab)
expect_snapshot_print(t[["typst"]], "tibble-pillar_num.typ")
expect_snapshot_print(t[["markdown"]], "tibble-pillar_num.md")
expect_snapshot_print(t[["latex"]], "tibble-pillar_num.tex")

exit_file("fail on github actions")
expect_snapshot_print(t[["html"]], "tibble-pillar_num.html")


# markdown
exit_file("ANSI tinysnapshot issue")
options("tinytable_print_output" = "markdown")
expect_snapshot_print(tab, "tibble-pillar_num")
options("tinytable_print_output" = NULL)

source("helpers.R")
using("tinysnapshot")
options(tinytable_print_output = "markdown")

# issue #98: Grid alignment
x <- mtcars[1:4, 1:5]
tab <- tt(x) |> group_tt(j = list(a = 2:3, b = 4:5))
expect_snapshot_print(tab, label = "docx-issue98_01")

x[1, 1] <- "$\\sigma$"
expect_snapshot_print(tt(x), label = "docx-issue98_02")

colnames(x)[1] <- "$\\sigma$"
expect_snapshot_print(tt(x), label = "docx-issue98_03")

x[1, 1] <- "1"
expect_snapshot_print(tt(x), label = "docx-issue98_04")

options(tinytable_print_output = NULL)

options("tinysnapshot_device" = "svglite")
options("tinysnapshot_tol" = 300)
options("tinysnapshot_mode" = "sidebyside")
options("width" = 200) # for diff printout readability
options("tinytable_print_output" = NULL)
options("tinytable_html_engine" = NULL)

# Stabilize collation-dependent ordering in snapshot tests.
try(Sys.setlocale("LC_COLLATE", "C"), silent = TRUE)

# common formatting options
options(tinytable_format_bool = function(x) tools::toTitleCase(tolower(x)))
options(tinytable_format_replace = "")

strip_random_sequential <- function(x, stem = "tinytable_css_") {
  pattern <- paste0(stem, "\\w+\\b")
  match_list <- lapply(x, function(s) {
    regmatches(s, gregexpr(pattern, s, perl = TRUE))[[1]]
  })
  all_matches <- unlist(match_list)
  unique_matches <- unique(all_matches)
  new_ids <- sprintf("tinytable_css_%02d", seq_along(unique_matches))
  replaced <- x
  for (i in seq_along(unique_matches)) {
    replaced <- gsub(unique_matches[i], new_ids[i], replaced, perl = TRUE)
  }
  replaced
}
strip_random <- function(x) {
  for (stem in c(
    "insertSpanRow",
    "spanCell_",
    "styleCell_",
    "styleHeaderCell_",
    "tableFns_",
    "tinytable_(?!css)",
    "tinytable_css_",
    "tinytable/"
  )) {
    x <- strip_random_sequential(x, stem)
  }
  x
}
options(tinysnapshot_fn_current = strip_random)
options(tinysnapshot_fn_target = strip_random)

is_local <- isTRUE(grepl("(?i)vincent|mbp|mandelbrot", Sys.info()["user"]))

# libraries
requiet <- function(package) {
  void <- capture.output(
    pkg_available <- tryCatch(
      suppressPackageStartupMessages(
        suppressWarnings(
          suppressMessages(
            tryCatch(
              isTRUE(
                require(package, warn.conflicts = FALSE, character.only = TRUE)
              ),
              error = function(e) FALSE
            )
          )
        )
      )
    )
  )
  return(pkg_available)
}

requiet("tinytable")
requiet("tinytest")
requiet("tinysnapshot")

print.custom_html_string <- function(x, ...) {
  cat(x, "\n", sep = "", ...)
  invisible(x)
}

print_html <- function(x, output = c("html")) {
  output <- match.arg(output)
  x <- save_tt(x, output)
  class(x) <- c("custom_html_string", class(x))
  x
}

format_extensions <- list(
  markdown = "md",
  latex = "tex",
  typst = "typ",
  html = "html"
)

# Helper function to prepare tables for testing across multiple formats
expect_table <- function(tab, formats = names(format_extensions)) {
  result <- list()

  for (format in formats) {
    # Create a copy and set the output format
    table_copy <- tab
    table_copy@output <- format

    # Create the table output
    if (format == "html") {
      table_output <- print_html(table_copy)
    } else {
      table_output <- table_copy
    }

    # Store the table output
    result[[format]] <- table_output
  }

  return(result)
}

source("helpers.R")
using("tinysnapshot")

requiet("Rdatasets")

# bug discovered in modelsummary
z <- factor(c("a", "b", NA))
a <- format_tt(z, replace = "-")
b <- c("a", "b", "-")
expect_equivalent(a, b)

# numeric vector input
a <- c(98938272783457, 7288839482, 29111727, 93945)
b <- format_tt(a, num_suffix = TRUE, digits = 2)
expect_equivalent(b, c("99T", "7.3B", "29M", "94K"))

# num_suffix (vignette)
options(tinytable_print_output = "markdown")

dat <- data.frame(
  a = c("Burger", "Halloumi", "Tofu", "Beans"),
  b = c(1.43202, 201.399, 0.146188, 0.0031),
  c = c(98938272783457, 7288839482, 29111727, 93945)
)
tab <- tt(dat) |>
  format_tt(j = "a", sprintf = "Food: %s") |>
  format_tt(j = 2, digits = 1) |>
  format_tt(j = "c", digits = 2, num_suffix = TRUE)
expect_snapshot_print(tab, label = "format_tt-num_suffix_vignette")

set.seed(1024)
dat <- data.frame(
  w = c(143002.2092, 201399.181, 100188.3883),
  x = c(1.43402, 201.399, 0.134588),
  y = as.Date(sample(1:1000, 3), origin = "1970-01-01"),
  z = c(TRUE, TRUE, FALSE)
)
tab <- tt(dat, digits = 2)
expect_snapshot_print(tab, label = "format_tt-vignette_digits")

tab <- tt(dat) |>
  format_tt(
    j = 2:4,
    digits = 1,
    date = "%B %d %Y"
  ) |>
  format_tt(
    j = 1,
    digits = 2,
    num_mark_big = " ",
    num_mark_dec = ",",
    num_fmt = "decimal"
  )
expect_snapshot_print(tab, label = "format_tt-vignette_misc")

expect_snapshot_print(
  format_tt(dat, digits = 1, num_suffix = TRUE),
  label = "format_tt-dataframe"
)

# custom formatting
x <- mtcars[1:3, 1:3]
tab <- tt(x) |> format_tt(fn = function(x) paste("Ya", x))
expect_snapshot_print(tab, "format_tt-fn")

# Issue #142
k <- data.frame(x = c(0.000123456789, 12.4356789))
tab <- tt(k, digits = 2)
expect_snapshot_print(tab, "format_tt-issue142_01")
tab <- tt(k) |> format_tt(digits = 2, num_fmt = "significant_cell")
expect_snapshot_print(tab, "format_tt-issue142_02")

# Issue #147: format_tt(escape = TRUE) zaps previous formatting
options(tinytable_print_output = "latex")
x <- data.frame(num = c(pi, pi), char = c("10$", "blah_blah"))
tab <- tt(x) |>
  format_tt(i = 1, j = 1, digits = 2) |>
  format_tt(i = 1, j = 1, digits = 3) # overwrite
expect_snapshot_print(tab, "format_tt-issue147_01.tex")
tab <- tt(x) |>
  format_tt(i = 1, j = 1, digits = 2) |>
  format_tt(i = 2, j = 1, digits = 3) # different cell
expect_snapshot_print(tab, "format_tt-issue147_02.tex")
tab <- tt(x) |>
  format_tt(i = 1, j = 1, digits = 2) |>
  format_tt(i = 2, j = 1, digits = 3) |>
  format_tt(escape = TRUE) # do not zap
expect_snapshot_print(tab, "format_tt-issue147_03.tex")
options(tinytable_print_output = NULL)

# Issue #149: num_mark_big requires digits
x <- data.frame(x = c(pi * 1e6, pi * 1e9))
tab <- format_tt(tt(x), digits = 1, num_mark_big = " ")
expect_snapshot_print(tab, "format_tt-issue149")
# expect_error(format_tt(tt(x), num_mark_big = " "))

x <- data.frame(x = 1:5, y = c(pi, NA, NaN, -Inf, Inf))
dict <- list("-" = c(NA, NaN), "Tiny" = -Inf, "Huge" = Inf)
tab <- tt(x) |>
  format_tt(replace = dict) |>
  save_tt("dataframe")
expect_equivalent(tab[[2]], c("y", "3.141593", "-", "-", "Tiny", "Huge"))

# Issue #256: big mark for integers
tab <- data.frame(x = c(1332037, 1299128, 805058, 206840, 698511)) |>
  tt() |>
  format_tt(num_mark_big = " ", digits = 0, num_fmt = "decimal") |>
  save_tt("dataframe")
expect_equivalent(
  tab[[1]],
  c("x", "1 332 037", "1 299 128", "805 058", "206 840", "698 511")
)

x <- data.frame(x = pi, y = NA)
options(tinytable_tt_digits = 2)
tab <- tt(x) |> save_tt("dataframe")
expect_equivalent(tab[[2]], c("y", ""))

options(tinytable_format_replace = "zzz")
tab <- tt(x) |> save_tt("dataframe")
expect_equivalent(tab[[2]], c("y", "zzz"))

options(tinytable_format_replace = FALSE)
tab <- tt(x) |> save_tt("dataframe")
expect_equivalent(tab[[2]], c("y", "NA"))

options(tinytable_print_output = NULL)
options(tinytable_tt_digits = NULL)
options(tinytable_format_replace = NULL)

# Issue #263: NA processing with Quarto doesn't work
x <- data.frame(
  Points = c(40, 0, 10, 10, 10, 10, 10, 20, 110),
  Assignment = c(
    "Analytical Assignments",
    "Team Project Part 0 - County Selection",
    "Team Project Part 1 - Summary Background",
    "Team Project Part 2 - SWOT Analysis",
    "Team Project Part 3 - Strategic Direction/Action Plan",
    "Team Project Part 4 - Evaluation Framework",
    "Team Project Part 5 - Plan Presentation",
    "Class participation",
    "Total"
  ),
  Percent = c("36%", "0%", "9%", "9%", "9%", "9%", "9%", "18%", NA)
)
x <- tt(x) |>
  format_tt(replace = "!", quarto = TRUE) |>
  save_tt("latex")
expect_true(grepl("IQ==", x))

# Website scaling example
thumbdrives <- data.frame(
  date_lookup = as.Date(
    c("2024-01-15", "2024-01-18", "2024-01-14", "2024-01-16")
  ),
  price = c(18.49, 19.99, 24.99, 24.99),
  price_rank = c(1, 2, 3, 3),
  memory = c(16e9, 12e9, 10e9, 8e9),
  speed_benchmark = c(0.6, 0.73, 0.82, 0.99)
)
tab <- tt(thumbdrives) |>
  format_tt(j = 1, fn = scales::date_format("%B %d %Y")) |>
  format_tt(j = 2, fn = scales::label_currency()) |>
  format_tt(j = 3, fn = scales::label_ordinal()) |>
  format_tt(j = 4, fn = scales::label_bytes()) |>
  format_tt(j = 5, fn = scales::label_percent()) |>
  format_tt(escape = TRUE) |>
  save_tt("dataframe")
expect_true("January 15 2024" %in% tab[[1]])
expect_false("2024-01-15" %in% tab[[1]])
expect_true("$18.49" %in% tab[[2]])
expect_true("16 GB" %in% tab[[4]])
expect_true("99%" %in% tab[[5]])

# Issue #409: both NA and NaN should be replaced
options(tinytable_format_replace = NULL)
tab <- data.frame(x = c(1, NA, NaN, Inf))
tab0 <- tt(tab) |> save_tt("dataframe")
tab1 <- tt(tab) |>
  format_tt() |>
  save_tt("dataframe")
tab2 <- tt(tab) |>
  format_tt(replace = TRUE) |>
  save_tt("dataframe")
expect_equivalent(tab0[[1]], c("x", "1", "NA", "NaN", "Inf"))
expect_equivalent(tab1[[1]], c("x", "1", "NA", "NaN", "Inf"))
expect_equivalent(tab2[[1]], c("x", "1", "", "", "Inf"))

# bug: duplicated columns with markdown html
dat <- data.frame(
  markdown = c(
    "This is _italic_ text.",
    "This sentence ends with a superscript.^2^"
  )
)
tab <- tt(dat) |>
  format_tt(j = 1, markdown = TRUE) |>
  style_tt(j = 1, align = "c")
expect_snapshot_print(print_html(tab), "format_tt-vignette_html_markdown.html")

# boolean formatting
tab <- tt(data.frame(a = c(TRUE, FALSE, TRUE, NA)))
tab1 <- format_tt(tab, bool = tolower) |> save_tt("dataframe")
tab2 <- format_tt(tab, j = 1, bool = tolower) |> save_tt("dataframe")
expect_equivalent(tab1$a, tab2$a)


# Issue #392
pen <- rddata("penguins", "palmerpenguins")[1:3, c("species", "body_mass_g")]
expect_error(format_tt(tt(pen), num_mark_big = ","), "requires.*digits")
expect_error(format_tt(1234, num_mark_big = ","), "requires.*digits")


# Component-specific formatting Issue #418
dat <- data.frame(x = 1:2, y = 3:4)
tab <- tt(dat, caption = "Test Caption", notes = "Test note")

tab1 <- format_tt(tab, i = "colnames", fn = function(x) paste0("Col_", x))
expect_true(grepl("Col_x", save_tt(tab1, "markdown")))
expect_false(grepl("Col_2", save_tt(tab1, "markdown")))

tab1 <- format_tt(tab, i = "caption", fn = function(x) paste0("Col_", x))
expect_true(grepl("Col_Test", save_tt(tab1, "markdown")))
expect_false(grepl("Col_2", save_tt(tab1, "markdown")))

tab1 <- format_tt(tab, i = "notes", fn = function(x) paste0("Col_", x))
expect_true(grepl("| Col_Test", save_tt(tab1, "markdown"), fixed = TRUE))
expect_false(grepl("Col_2", save_tt(tab1, "markdown")))

# Test formatting only notes
tab3 <- format_tt(tab, i = "notes", fn = function(x) paste0("Note: ", x))
expect_true(grepl("| Note: |", save_tt(tab3, "markdown"), fixed = TRUE))
expect_true(grepl("| Test  |", save_tt(tab3, "markdown"), fixed = TRUE))

# Test multiple components at once
tab4 <- format_tt(tab, i = c("colnames", "caption"), fn = function(x) {
  paste0("Prefix_", x)
})
expect_true(grepl("Prefix_x", save_tt(tab4, "markdown"), fixed = TRUE))
expect_true(grepl(
  "Prefix_Test Caption",
  save_tt(tab4, "markdown"),
  fixed = TRUE
))


# groupi vs. ~groupi
tab = head(iris) |>
  tt() |>
  group_tt(i = list("Hello" = 1, "World" = 3)) |>
  format_tt("~groupi", j = 1, sprintf = "--%s--") |>
  save_tt("dataframe")
expect_true("Hello" %in% tab[[1]])
expect_true("--5.1--" %in% tab[[1]])
expect_false("--Hello--" %in% tab[[1]])
expect_false("5.1" %in% tab[[1]])

tab = head(iris) |>
  tt() |>
  group_tt(i = list("Hello" = 1, "World" = 3)) |>
  format_tt("groupi", j = 1, sprintf = "--%s--") |>
  save_tt("dataframe")
expect_true("--Hello--" %in% tab[[1]])
expect_true("5.1" %in% tab[[1]])
expect_false("Hello" %in% tab[[1]])
expect_false("--5.1--" %in% tab[[1]])


# Vignette with multiple components
options(tinytable_print_output = "latex")
set.seed(48103)
tab <- data.frame(
  "A_B" = rnorm(5),
  "B_C" = rnorm(5),
  "C_D" = rnorm(5)
)
tab <- tt(tab, digits = 2, notes = "_Source_: Simulated data.") |>
  group_tt(i = list("Down" = 1, "Up" = 3)) |>
  format_tt("colnames", fn = \(x) sub("_", " / ", x)) |>
  format_tt("notes", markdown = TRUE) |>
  format_tt("groupi", replace = list("↓" = "Down", "↑" = "Up"))

expect_snapshot_print(
  tab,
  label = "format_tt-format_components_vignette_01.tex"
)
options(tinytable_print_output = NULL)


# format group_tt(i)
dat <- data.frame(x = 1:4, y = letters[1:4])
tab <- tt(dat) |>
  group_tt(i = list("Group A" = 2, "Group B" = 4)) |>
  format_tt(i = c(2, 5), j = 1, sprintf = "_%s_") |>
  format_tt(i = 1:3, j = 1, sprintf = "*%s*")
expect_snapshot_print(tab, label = "format_tt-group_tt_01.md")


dat1 <- data.frame(
  values = c(123.456, 789.012, 345.678, 901.234),
  labels = c("A", "B", "C", "D")
)
tab1 <- tt(dat1) |>
  group_tt(i = list("First Group" = 2, "Second Group" = 4)) |>
  format_tt(j = 1, sprintf = "((%s))")
expect_snapshot_print(tab1, label = "format_tt-group_tt_02.md")


# Line breaks using linebreak argument
d <- data.frame(Text = "First line<br>Second line")
tab <- tt(d) |> format_tt(linebreak = "<br>")
t <- expect_table(tab, formats = c("html", "latex", "typst"))
expect_true(grepl("First line<br>Second line", t[["html"]]))
expect_true(grepl("First line\\\\Second line", t[["latex"]], fixed = TRUE))
expect_true(grepl("First line \\ Second line", t[["typst"]], fixed = TRUE))

# Date formatting tests
dates <- as.Date(c("2023-01-01", "2023-12-31"))
formatted_dates <- format_vector(dates, date = "%Y")
expect_equivalent(formatted_dates, c("2023", "2023"))
formatted_dates_full <- format_vector(dates, date = "%B %d, %Y")
expect_equivalent(formatted_dates_full, c("January 01, 2023", "December 31, 2023"))
formatted_vector <- format_vector(dates, date = "%Y")
expect_equivalent(formatted_vector, c("2023", "2023"))
d <- data.frame(dates = dates)
tab <- tt(d) |> format_tt(j = 1, date = "%Y")
formatted_table <- save_tt(tab, "dataframe")
expect_equivalent(formatted_table[[1]], c("dates", "2023", "2023"))


# Conditional formatting
dat <- data.frame(a = c("*Hello*", "**World**"))
cap <- "HTML render 1st row. LaTeX render 2nd row."
tab <- tt(dat, caption = cap) |>
  format_tt(i = 1, j = 1, markdown = TRUE, output = "html") |>
  format_tt(i = 2, j = 1, markdown = TRUE, output = "latex")
t <- expect_table(tab, formats = c("html", "latex"))
expect_snapshot_print(t[["html"]], "format_tt-conditional_output.html")
expect_snapshot_print(t[["latex"]], "format_tt-conditional_output.tex")

source("helpers.R")

if (isTRUE(requiet("marginaleffects"))) {
  mod <- lm(mpg ~ hp * qsec, mtcars)
  s <- avg_slopes(mod)
  k <- tt(s) |> save_tt("dataframe")
  expect_equal(nrow(k), 3)
  expect_true(all(c("term", "estimate", "std.error") %in% k[1, ]))
}

# Issue #264: Row names global option
options(tinytable_tt_rownames = TRUE)
x <- mtcars[1:4, 1:4]
tab <- tt(x)
expect_equal(nrow(tab), 4)
expect_equal(ncol(tab), 5)
expect_equal(colnames(tab)[1], "rowname")
options(tinytable_tt_rownames = NULL)

source("helpers.R")
using("tinysnapshot")


# Issue #514: vignette white lines with blue background
dat <- data.frame(1:2, 3:4, 5:6, 7:8)
tab <- tt(dat, theme = "empty", colnames = FALSE) |>
    style_tt(
        line = "tblr",
        line_color = "white",
        line_width = 0.5,
        background = "blue",
        color = "white"
    )
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "style-issue514_white_blue.md")
expect_snapshot_print(t[["latex"]], "style-issue514_white_blue.tex")
expect_snapshot_print(t[["typst"]], "style-issue514_white_blue.typ")
expect_snapshot_print(t[["html"]], "style-issue514_white_blue.html")


# Issue #507: Test group row styling functionality
tab <- tt(mtcars[1:8, 1:3]) |>
    group_tt(i = list("Hello" = 3, "World" = 6)) |>
    group_tt(j = list("Hello" = 1, "World" = 2:3)) |>
    style_tt(i = 2:4, italic = TRUE) |>
    style_tt(i = 6, strikeout = TRUE) |>
    style_tt(i = 8, bold = TRUE)
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "style-issue507_markdown_styles.md")
expect_snapshot_print(t[["latex"]], "style-issue507_markdown_styles.tex")
expect_snapshot_print(t[["typst"]], "style-issue507_markdown_styles.typ")
expect_snapshot_print(t[["html"]], "style-issue507_markdown_styles.html")


# Issue #512: Many tables with identical output
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    style_tt("colnames", italic = TRUE) |>
    style_tt(-1, bold = TRUE)
expect_snapshot_print(tab, "style-groupj_colnames.md")
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    style_tt(0, italic = TRUE) |>
    style_tt(-1, bold = TRUE)
expect_snapshot_print(tab, "style-groupj_colnames.md")
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    format_tt(0, sprintf = "_%s_") |>
    format_tt(-1, sprintf = "**%s**")
expect_snapshot_print(tab, "style-groupj_colnames.md")
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    format_tt("colnames", sprintf = "_%s_") |>
    format_tt(-1, sprintf = "**%s**")
expect_snapshot_print(tab, "style-groupj_colnames.md")

tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    style_tt("colnames", italic = TRUE) |>
    style_tt("groupj", bold = TRUE)
expect_snapshot_print(tab, "style-groupj_colnames.md")


# Issue #564: Empty which() index should proceed silently
dat <- data.frame(feed = c("fish", "meat", "soy"), weight = c(10, 20, 30))
tab1 <- tt(dat)
tab2 <- tt(dat)

# These should not error when no rows match the condition
expect_silent(style_tt(tab1, i = {
    feed == "cool"
}, bold = TRUE))
expect_silent(style_tt(tab2, i = which(feed == "cool"), bold = TRUE))


# smallcap styling functionality
tab <- tt(head(iris),
    caption = "Motor Trend Car Road Tests",
    notes = "Source: Henderson and Velleman (1981)") |>
    style_tt(i = "colnames", smallcap = TRUE) |>
    style_tt(i = "caption", smallcap = TRUE) |>
    style_tt(i = "notes", smallcap = TRUE) |>
    style_tt(i = 1:3, j = 5, smallcap = TRUE)
t <- expect_table(tab)
expect_snapshot_print(t[["html"]], "style-smallcap.html")
expect_snapshot_print(t[["latex"]], "style-smallcap.tex")
expect_snapshot_print(t[["typst"]], "style-smallcap.typ")
expect_snapshot_print(t[["markdown"]], "style-smallcap.md")


# partial align
tab <- tt(mtcars[1:5, 1:6]) |> style_tt(j = c(2, 4), align = "cr")
t <- expect_table(tab)
expect_snapshot_print(t[["html"]], "style-align_partial.html")
expect_snapshot_print(t[["latex"]], "style-align_partial.tex")
expect_snapshot_print(t[["typst"]], "style-align_partial.typ")
expect_snapshot_print(t[["markdown"]], "style-align_partial.md")


# logical matrix with all FALSE values should not error
testdata <- data.frame(
  names = c('a', 'b', 'c', 'd', 'e'),
  values1 = c(1, 2, 3, 4, 5),
  values2 = c(6, 7, 8, 9, 10)
)
testdata_override <- matrix(FALSE, nrow = nrow(testdata), ncol = ncol(testdata))
result <- testdata |>
  tt() |>
  style_tt(i = testdata_override, background = "red") |>
  save_tt("html")
expect_true(is.character(result))
expect_true(nchar(result) > 0)

source("helpers.R")
using("tinysnapshot")


# Issue #514: vignette white lines with blue background
dat <- data.frame(1:2, 3:4, 5:6, 7:8)
tab <- tt(dat, theme = "empty", colnames = FALSE) |>
    style_tt(
        line = "tblr",
        line_color = "white",
        line_width = 0.5,
        background = "blue",
        color = "white"
    )
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "style-issue514_white_blue.md")
expect_snapshot_print(t[["latex"]], "style-issue514_white_blue.tex")
expect_snapshot_print(t[["typst"]], "style-issue514_white_blue.typ")
expect_snapshot_print(t[["html"]], "style-issue514_white_blue.html")


# Issue #507: Test group row styling functionality
tab <- tt(mtcars[1:8, 1:3]) |>
    group_tt(i = list("Hello" = 3, "World" = 6)) |>
    group_tt(j = list("Hello" = 1, "World" = 2:3)) |>
    style_tt(i = 2:4, italic = TRUE) |>
    style_tt(i = 6, strikeout = TRUE) |>
    style_tt(i = 8, bold = TRUE)
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "style-issue507_markdown_styles.md")
expect_snapshot_print(t[["latex"]], "style-issue507_markdown_styles.tex")
expect_snapshot_print(t[["typst"]], "style-issue507_markdown_styles.typ")
expect_snapshot_print(t[["html"]], "style-issue507_markdown_styles.html")


# Issue #512: Many tables with identical output
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    style_tt("colnames", italic = TRUE) |>
    style_tt(-1, bold = TRUE)
expect_snapshot_print(tab, "style-groupj_colnames.md")
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    style_tt(0, italic = TRUE) |>
    style_tt(-1, bold = TRUE)
expect_snapshot_print(tab, "style-groupj_colnames.md")
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    format_tt(0, sprintf = "_%s_") |>
    format_tt(-1, sprintf = "**%s**")
expect_snapshot_print(tab, "style-groupj_colnames.md")
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    format_tt("colnames", sprintf = "_%s_") |>
    format_tt(-1, sprintf = "**%s**")
expect_snapshot_print(tab, "style-groupj_colnames.md")

tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    style_tt("colnames", italic = TRUE) |>
    style_tt("groupj", bold = TRUE)
expect_snapshot_print(tab, "style-groupj_colnames.md")


# Issue #564: Empty which() index should proceed silently
dat <- data.frame(feed = c("fish", "meat", "soy"), weight = c(10, 20, 30))
tab1 <- tt(dat)
tab2 <- tt(dat)

# These should not error when no rows match the condition
expect_silent(style_tt(tab1, i = {
    feed == "cool"
}, bold = TRUE))
expect_silent(style_tt(tab2, i = which(feed == "cool"), bold = TRUE))


# smallcap styling functionality
tab <- tt(head(iris),
    caption = "Motor Trend Car Road Tests",
    notes = "Source: Henderson and Velleman (1981)") |>
    style_tt(i = "colnames", smallcap = TRUE) |>
    style_tt(i = "caption", smallcap = TRUE) |>
    style_tt(i = "notes", smallcap = TRUE) |>
    style_tt(i = 1:3, j = 5, smallcap = TRUE)
t <- expect_table(tab)
expect_snapshot_print(t[["html"]], "style-smallcap.html")
expect_snapshot_print(t[["latex"]], "style-smallcap.tex")
expect_snapshot_print(t[["typst"]], "style-smallcap.typ")
expect_snapshot_print(t[["markdown"]], "style-smallcap.md")


# partial align
tab <- tt(mtcars[1:5, 1:6]) |> style_tt(j = c(2, 4), align = "cr")
t <- expect_table(tab)
expect_snapshot_print(t[["html"]], "style-align_partial.html")
expect_snapshot_print(t[["latex"]], "style-align_partial.tex")
expect_snapshot_print(t[["typst"]], "style-align_partial.typ")
expect_snapshot_print(t[["markdown"]], "style-align_partial.md")


# logical matrix with all FALSE values should not error
testdata <- data.frame(
  names = c('a', 'b', 'c', 'd', 'e'),
  values1 = c(1, 2, 3, 4, 5),
  values2 = c(6, 7, 8, 9, 10)
)
testdata_override <- matrix(FALSE, nrow = nrow(testdata), ncol = ncol(testdata))
result <- testdata |>
  tt() |>
  style_tt(i = testdata_override, background = "red") |>
  save_tt("html")
expect_true(is.character(result))
expect_true(nchar(result) > 0)


# Issue #647: Conditional styling with output argument
tab <- data.frame(A = 1:2, B = 3:4) |>
  tt() |>
  style_tt(bold = TRUE, output = "typst") |>
  style_tt(italic = TRUE, output = "html")
t <- expect_table(tab, formats = c("html", "typst", "latex"))
expect_snapshot_print(t[["html"]], "style-output_conditional.html")
expect_snapshot_print(t[["typst"]], "style-output_conditional.typ")
expect_snapshot_print(t[["latex"]], "style-output_conditional.tex")

source("helpers.R")
using("tinysnapshot")
options(tinytable_print_output = "typst")

# semi complicated
tab <- tt(mtcars[1:4, 1:5], caption = "Hello World") |>
  group_tt(j = list("Group 1" = 4:5, "Group 2" = 2:3)) |>
  style_tt(j = 1:5, align = "lcccr") |>
  style_tt(
    i = 2,
    j = 1:3,
    strikeout = TRUE,
    bold = TRUE,
    background = "black",
    color = "white"
  ) |>
  style_tt(j = 1, color = "red", italic = TRUE)
expect_snapshot_print(tab, label = "typst-complicated.typ")

# issue #139
tab <- tt(mtcars[1:10, 1:5]) |>
  group_tt(i = list("Feta" = 2, "Brie" = 6)) |>
  group_tt(j = list("First" = 2:3, "Second" = 4:5)) |>
  style_tt(1:5, align = "c", background = "blue", color = "white")
expect_snapshot_print(
  tab,
  label = "typst-issue-139_misaligned_rule_with_group_tt.typ"
)

# Theme striped
x <- mtcars[1:4, 1:5]
tab <- tt(x, theme = "striped")
expect_snapshot_print(tab, label = "typst-striped.typ")

# Theme grid
tab <- tt(x, theme = "grid")
expect_snapshot_print(tab, label = "typst-grid.typ")

# Formatting
dat <- data.frame(
  w = c(143002.2092, 201399.181, 100188.3883),
  x = c(1.43402, 201.399, 0.134588),
  y = as.Date(c(897, 232, 198), origin = "1970-01-01"),
  z = c(TRUE, TRUE, FALSE)
)
dat <- tt(dat, digits = 2)
expect_snapshot_print(dat, label = "typst-formatting.typ")

# More formatting
dat <- data.frame(
  a = c("Burger", "Halloumi", "Tofu", "Beans"),
  b = c(1.43202, 201.399, 0.146188, 0.0031),
  c = c(98938272783457, 7288839482, 29111727, 93945)
)
tab <- tt(dat) |>
  format_tt(j = "a", sprintf = "Food: %s") |>
  format_tt(j = 2, digits = 1) |>
  format_tt(j = "c", digits = 2, num_suffix = TRUE)
expect_snapshot_print(tab, label = "typst-more_formatting.typ")

# Significant cell
dat <- data.frame(x = c(0.000123456789, 12.4356789))
tab <- tt(dat) |> format_tt(digits = 2, num_fmt = "significant_cell")
expect_snapshot_print(tab, label = "typst-significant_cell.typ")

# Missing value replacement
tab <- tt(data.frame(a = c(NA, 1, 2), b = c(3, NA, 5)))
tab <- format_tt(tab, replace = "-")
expect_snapshot_print(tab, label = "typst-missing_value_replacement.typ")

# Italic markdown
dat <- data.frame(markdown = c("This is _italic_ text."))
tab <- tt(dat) |>
  format_tt(j = 1, markdown = TRUE) |>
  style_tt(j = 1, align = "c")
expect_snapshot_print(tab, label = "typst-italic_markdown.typ")

# Font size
dat <- tt(x) |> style_tt(j = "mpg|hp|qsec", fontsize = 1.5)
expect_snapshot_print(dat, label = "typst-font_size.typ")

# No headers
k <- x
colnames(k) <- NULL
k <- tt(k)
expect_snapshot_print(k, label = "typst-no_headers.typ")

# Portable images
if (requiet("base64enc")) {
  img <- tempfile(fileext = ".png")
  grDevices::png(img, width = 20, height = 20)
  graphics::par(mar = c(0, 0, 0, 0))
  graphics::plot.new()
  graphics::rect(0, 0, 1, 1, col = "black", border = NA)
  grDevices::dev.off()

  tab <- tt(data.frame(x = "")) |>
    plot_tt(j = 1, images = img) |>
    theme_typst(portable = TRUE) |>
    save_tt("typst")

  expect_true(grepl("#image(bytes(", tab, fixed = TRUE))
  expect_true(grepl("data:image/png;base64,", tab, fixed = TRUE))
  expect_false(grepl(img, tab, fixed = TRUE))

  unlink(img)
}

# Group rows
dat <- mtcars[1:9, 1:8]
dat <- tt(dat) |>
  group_tt(
    i = list(
      "I like (fake) hamburgers" = 3,
      "She prefers halloumi" = 4,
      "They love tofu" = 7
    )
  )
expect_snapshot_print(dat, label = "typst-group_rows.typ")

# Group columns
dat <- mtcars[1:9, 1:8]
tab <- tt(dat) |>
  group_tt(
    j = list(
      "Hamburgers" = 1:3,
      "Halloumi" = 4:5,
      "Tofu" = 7
    )
  )
expect_snapshot_print(tab, label = "typst-group_columns.typ")

# issue #323
dat <- mtcars[1:9, 1:8]
tab <- tt(dat) |>
  group_tt(
    i = list(
      "I like (fake) hamburgers" = 3,
      "She prefers halloumi" = 4,
      "They love tofu" = 7
    ),
    j = list("Hamburgers" = 1:3, "Halloumi" = 4:5, "Tofu" = 7)
  ) |>
  style_tt(
    i = c(3, 5, 9),
    align = "c",
    background = "black",
    color = "orange"
  ) |>
  style_tt(i = -1, color = "orange")
tab@output <- "typst"
expect_snapshot_print(tab, label = "typst-issue323_group_tt_style_tt.typ")

# Frame
tab <- tt(mtcars[1:5, 1:5]) |>
  style_tt(2:3, 2:3, line_color = "red", line = "tblr", line_width = .05)
tab@output <- "typst"
expect_snapshot_print(tab, label = "typst-tblr.typ")

# Issue #357
tab <- tt(head(iris), notes = "blah") |> save_tt("typst")
expect_true(grepl("blah", tab))

# Bug: Footnotes with cell coordinates #456
tab <- tt(
  mtcars[1:4, 1:5],
  notes = list(
    a = list(i = c(0:1), j = 1, text = "Blah."),
    b = "Blah blah."
  )
)
expect_snapshot_print(tab, label = "typst-issue456.typ")

# Single row colspan bug fix
df <- data.frame(
  column_a = c("No data"),
  column_b = c(NA),
  column_c = c(NA)
)
tab <- df |>
  tt() |>
  style_tt(i = 1, j = 1, colspan = ncol(df), align = "c") |>
  save_tt("typst")
expect_true(is.character(tab) && nchar(tab) > 0)


# Figure alignment is opt-in for Typst output
tab <- tt(mtcars[1:4, 1:5]) |>
  save_tt("typst")
expect_false(grepl("#align(", tab, fixed = TRUE))

tab <- tt(mtcars[1:4, 1:5]) |>
  theme_typst(align_figure = "c") |>
  save_tt("typst")
expect_true(grepl("#align(center,", tab, fixed = TRUE))

tab <- tt(mtcars[1:4, 1:5]) |>
  theme_typst(align_figure = "r") |>
  save_tt("typst")
expect_true(grepl("#align(right,", tab, fixed = TRUE))


# Typst resizing follows theme_latex() API
tab <- tt(mtcars[1:4, 1:5]) |>
  theme_typst(resize_width = 0.8, resize_direction = "down") |>
  save_tt("typst")
expect_true(grepl("#layout(size => {", tab, fixed = TRUE))
expect_true(grepl("let target-width = size.width * 0.8", tab, fixed = TRUE))
expect_true(grepl("if body-size.width > target-width", tab, fixed = TRUE))
expect_true(grepl("scale(x: factor, y: factor, reflow: true, body)", tab, fixed = TRUE))

tab <- tt(mtcars[1:4, 1:5]) |>
  theme_typst(resize_height = 0.6, resize_direction = "up") |>
  save_tt("typst")
expect_true(grepl("let target-height = size.height * 0.6", tab, fixed = TRUE))
expect_true(grepl("if body-size.height < target-height", tab, fixed = TRUE))

expect_error(
  theme_typst(tt(mtcars[1:4, 1:5]), resize_direction = "sideways"),
  "resize_direction"
)


# Issue #592
table <- data.frame(
  stringsAsFactors = FALSE,
  check.names = FALSE,
  metric = c(
    "Adoption",
    "Adoption",
    "Adoption",
    "Adoption",
    "Adoption",
    "NPS",
    "NPS",
    "NPS",
    "NPS",
    "NPS",
    "Overall Satisfaction",
    "Overall Satisfaction",
    "Overall Satisfaction",
    "Overall Satisfaction",
    "Overall Satisfaction"),
  tier = c(
    "1",
    "2 ", "3 ", "3 ", "3 ", "1",
    "2 ", "3 ", "3 ", "3 ", "1",
    "2 ", "3 ", "3 ", "3"),
  role = c(
    "",
    "x", "x", "y", "z", "",
    "x", "x", "y", "z", "",
    "x", "x", "y", "z"),
  Overall = c(
    "90.9%", "87.9%", "90.6%", "90.6%", "100%", "19",
    "39.9", "47.2", "14.6", "100", "80.7%",
    "90.4%", "80.9%", "91.8%", "100%"),
  state1 = c(
    "90.5%", "91%", "91%", "92.3%", "100%", "24.2",
    "40.3", "51.4", "11.5", "100", "82.4%",
    "89.4%", "78.4%", "91.6%", "100%"),
  state2 = c(
    "95.7%", "75%", NA, NA, NA, "11.5", "45.7", NA,
    NA, NA, "79.7%", "94.4%", NA, NA, NA),
  state3 = c(
    "91.5%", "85.9%", "95.4%", "100%", NA, "23.8",
    "44.7", "70.8", "22.2", NA, "82.4%", "88.1%",
    "55.6%", "98.4%", NA)
)
tab <- tt(table) |>
  theme_empty() |>
  style_tt(i = c(3, 8, 13), j = 2, rowspan = 3, alignv = "m") |>
  style_tt(i = 0, line = "tblr") |>
  style_tt(j = 1, line = "l") |>
  style_tt(j = 2, i = 1:nrow(table), alignv = "m", line = "lt") |>
  style_tt(j = 3:length(table), i = 1:nrow(table), line = "r") |>
  style_tt(j = 4:length(table), i = 1:nrow(table), line = "t") |>
  style_tt(j = 3, i = c(1, 2, 3, 6, 7, 8, 11, 12, 13, 16), line = "t") |>
  style_tt(
    i = seq(1, nrow(table), by = 5),
    j = 1,
    rowspan = 5,
    align = "c",
    alignv = "m",
    line = "lt"
  ) |>
  style_tt(i = nrow(table), j = 1:length(table), line = "b") |>
  format_tt(linebreak = "\n", replace = "", output = "typst")
expect_snapshot_print(tab, label = "typst-issue592.typ")



options(tinytable_print_output = NULL)

source("helpers.R")
using("tinysnapshot")
requiet("data.table")

dat1 <- structure(list(EU = c("Austria", "Belgium", "Bulgaria", "Croatia", "Cyprus", "Czech Republic", "Denmark", "Estonia", "Finland", "France", "Germany", "Greece", "Hungary", "Ireland", "Italy", "Latvia", "Lithuania", "Luxembourg", "Malta", "Netherlands", "Poland", "Portugal", "Romania", "Slovakia", "Slovenia", "Spain", "Sweden")), class = "data.frame", row.names = c(NA, -27L)) # noqa
DT1 <- as.data.table(dat1)
colnames(DT1) <- NULL
tab <- tt(DT1)
expect_inherits(tab, "tinytable")


# Bug group_tt(i) accepts factor variables
dat <- do.call(rbind, by(iris, iris$Species, head, n = 3))
tab <- tt(dat) |>
  group_tt(dat$Species) |>
  save_tt("markdown")
tab <- unlist(strsplit(tab, "\\n"))
expect_equal(sum(grepl("^\\| virginica", tab)), 1)

source("helpers.R")
using("tinysnapshot")
requiet("data.table")

dat1 <- structure(list(EU = c("Austria", "Belgium", "Bulgaria", "Croatia", "Cyprus", "Czech Republic", "Denmark", "Estonia", "Finland", "France", "Germany", "Greece", "Hungary", "Ireland", "Italy", "Latvia", "Lithuania", "Luxembourg", "Malta", "Netherlands", "Poland", "Portugal", "Romania", "Slovakia", "Slovenia", "Spain", "Sweden")), class = "data.frame", row.names = c(NA, -27L)) # noqa
DT1 <- as.data.table(dat1)
colnames(DT1) <- NULL
tab <- tt(DT1)
expect_inherits(tab, "tinytable")


# Bug group_tt(i) accepts factor variables
dat <- do.call(rbind, by(iris, iris$Species, head, n = 3))
tab <- tt(dat) |>
  group_tt(dat$Species) |>
  save_tt("markdown")
tab <- unlist(strsplit(tab, "\\n"))
expect_equal(sum(grepl("^\\| virginica", tab)), 1)


# Bug: OutDec should not affect structural markup numbers
local({
  old <- options(OutDec = ",")
  on.exit(options(old), add = TRUE)

  has_bad_markup_decimal <- function(x) {
    grepl("[0-9],[0-9]+(em|%|deg)", x, perl = TRUE)
  }

  dat <- data.frame(a = 1.23, b = 4.56)
  tab <- tt(dat, height = 0.25, width = 0.333) |>
    format_tt(digits = 2, num_fmt = "decimal") |>
    style_tt(
      i = 1,
      j = 1,
      fontsize = 0.8,
      indent = 0.5,
      line = "tblr",
      line_width = 0.05
    ) |>
    theme_rotate(angle = 12.5, i = 1, j = 1)

  html <- save_tt(tab, "html")
  latex <- save_tt(tab, "latex")
  typst <- save_tt(tab, "typst")
  tabulator <- tt(dat, height = 0.25) |>
    format_tt(digits = 2, num_fmt = "decimal") |>
    theme_html(engine = "tabulator") |>
    save_tt("html")

  expect_false(has_bad_markup_decimal(html))
  expect_false(has_bad_markup_decimal(latex))
  expect_false(has_bad_markup_decimal(typst))
  expect_false(has_bad_markup_decimal(tabulator))

  expect_true(grepl("1,23", html, fixed = TRUE))
  expect_true(grepl("1,23", latex, fixed = TRUE))
  expect_true(grepl("1,23", typst, fixed = TRUE))
  expect_true(grepl("1,23", tabulator, fixed = TRUE))
})

source("helpers.R")
using("tinysnapshot")

# Basic matrix insertion - single position, single row
rowmat <- matrix(c("Inserted", "Row", "Data", "Here", "Extra"), nrow = 1)
tab <- tt(head(iris, 5)) |> group_tt(i = 3, j = rowmat)
t <- expect_table(tab)
expect_snapshot_print(
  t[["markdown"]],
  "group_matrix-single_position_single_row.md"
)
expect_snapshot_print(
  t[["latex"]],
  "group_matrix-single_position_single_row.tex"
)
expect_snapshot_print(
  t[["typst"]],
  "group_matrix-single_position_single_row.typ"
)
expect_snapshot_print(
  t[["html"]],
  "group_matrix-single_position_single_row.html"
)

# Matrix insertion - multiple positions and rows (covers both multiple positions and multiple rows)
rowmat <- matrix(
  c(
    "Row1",
    "At",
    "Pos1",
    "Data",
    "A",
    "Row2",
    "At",
    "Pos1",
    "Data",
    "B",
    "Row3",
    "At",
    "Pos2",
    "Data",
    "C"
  ),
  nrow = 3,
  byrow = TRUE
)
tab <- tt(head(iris, 7)) |> group_tt(i = c(2, 2, 5), j = rowmat)
t <- expect_table(tab)
expect_snapshot_print(
  t[["markdown"]],
  "group_matrix-multiple_positions_multiple_rows.md"
)
expect_snapshot_print(
  t[["latex"]],
  "group_matrix-multiple_positions_multiple_rows.tex"
)
expect_snapshot_print(
  t[["typst"]],
  "group_matrix-multiple_positions_multiple_rows.typ"
)
expect_snapshot_print(
  t[["html"]],
  "group_matrix-multiple_positions_multiple_rows.html"
)

# Matrix insertion at edge positions (combines position 1 and last position)
expect_error(
  tt(head(iris, 4)) |>
    group_tt(
      i = 1,
      j = matrix(c("Header", "Row", "At", "Top", "Position"), nrow = 1)
    ) |>
    group_tt(
      i = 6,
      j = matrix(c("Footer", "Row", "At", "Bottom", "Position"), nrow = 1)
    ),
  pattern = "Only one group row"
)

# Matrix with single column reshape and styling combined
rowmat <- matrix(c("A", "B", "C", "D", "E"))
tab <- tt(head(iris, 4)) |>
  group_tt(i = 2, j = rowmat) |>
  style_tt(i = "groupi", background = "lightblue")
t <- expect_table(tab)
expect_snapshot_print(
  t[["markdown"]],
  "group_matrix-single_column_with_styling.md"
)
expect_snapshot_print(
  t[["latex"]],
  "group_matrix-single_column_with_styling.tex"
)
expect_snapshot_print(
  t[["typst"]],
  "group_matrix-single_column_with_styling.typ"
)
expect_snapshot_print(
  t[["html"]],
  "group_matrix-single_column_with_styling.html"
)

# Matrix row duplication - single row matrix with multiple positions
rowmat <- matrix(colnames(iris))
tab <- tt(head(iris, 7)) |> group_tt(i = c(2, 5), j = rowmat)
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "group_matrix-row_duplication.md")
expect_snapshot_print(t[["latex"]], "group_matrix-row_duplication.tex")
expect_snapshot_print(t[["typst"]], "group_matrix-row_duplication.typ")
expect_snapshot_print(t[["html"]], "group_matrix-row_duplication.html")

source("helpers.R")
using("tinysnapshot")

if (!is_local) {
  exit_file("Run on Vincent's machine")
}

# vector row labels
set.seed(48103)
dat <- data.frame(
  label = c("a", "a", "a", "b", "b", "c", "a", "a"),
  x1 = rnorm(8),
  x2 = rnorm(8)
)
tab <- tt(dat[, 2:3]) |> group_tt(i = dat$label)
tab@output <- "markdown"
expect_snapshot_print(tab, label = "group-vector_row_labels.md")

# 3 level tests across formats
if (Sys.info()[["sysname"]] == "Darwin") {
  x <- mtcars[1:3, 1:5]
  tab <- tt(x) |>
    group_tt(j = list("a" = 2:3, "b" = 4:5)) |>
    group_tt(j = list("c" = 1:2, "d" = 3:5)) |>
    group_tt(j = list("e" = 1:3, "f" = 4))
  t <- expect_table(tab, formats = c("markdown", "latex", "typst"))
  expect_snapshot_print(t[["markdown"]], "group-3level.md")
  expect_snapshot_print(t[["latex"]], "group-3level.tex")
  expect_snapshot_print(t[["typst"]], "group-3level.typ")
}

tab <- tt(mtcars[1:10, 1:5]) |>
  group_tt(
    i = list("Hello" = 3, "World" = 8),
    j = list("Foo" = 2:3, "Bar" = 4:5)
  )
expect_snapshot_print(print_html(tab), "group-html_tutorial_01.html")


# Issue #258: group_tt indices with 0 and duplicates
tab1 <- tt(head(iris)) |> group_tt(i = list("hello" = 1))
t1 <- expect_table(tab1)
expect_snapshot_print(t1[["latex"]], "group-issue258_01_latex.tex")
expect_snapshot_print(t1[["typst"]], "group-issue258_01_typst.typ")
expect_snapshot_print(t1[["markdown"]], "group-issue258_01_markdown.md")
expect_snapshot_print(t1[["html"]], "group-issue258_01_html.html")

tab2 <- tt(head(iris)) |> group_tt(i = list("hello" = 2, "world" = 2))
t2 <- expect_table(tab2)
expect_snapshot_print(t2[["latex"]], "group-issue258_02_latex.tex")
expect_snapshot_print(t2[["typst"]], "group-issue258_02_typst.typ")
expect_snapshot_print(t2[["markdown"]], "group-issue258_02_markdown.md")
expect_snapshot_print(t2[["html"]], "group-issue258_02_html.html")

# Issue #362: group_tt with duplicate column group labels in latex
tab <- data.frame(
  sub1 = 1,
  sub2 = 2,
  sub1 = 3,
  sub2 = 4,
  check.names = FALSE
) |>
  tt() |>
  group_tt(j = list(a = 1:2, a = 3:4))
tab@output <- "latex"
expect_snapshot_print(tab, "group-issue362_duplicate_colum_labels.tex")

# Issue #413: Automatic row groups fail if column is a factor
df <- mtcars |> head(10)
df <- df[order(df$am), ]
df$am <- factor(df$am)
tab <- tt(df) |> group_tt(i = df$am)
expect_inherits(tab, "tinytable")
expect_inherits(save_tt(tab, "markdown"), "character")


# delimiters in column names
# No delimiter
tab <- tt(mtcars |> head())
tab@output <- "markdown"
expect_snapshot_print(tab, "group-delim-no-delim.md")
tab <- tt(data.frame(
  A_id = 1,
  A_a1 = 2,
  A_a2 = "3",
  B_b1 = 4,
  B_b2 = 5,
  B_C = 6
)) |>
  group_tt(j = "_")
tab@output <- "markdown"
expect_snapshot_print(tab, "group-delim-all-delim.md")
tab <- tt(data.frame(
  id = 1,
  A_a1 = 2,
  A_a2 = "3",
  B_b1 = 4,
  B_b2 = 5,
  C = 6
)) |>
  group_tt(j = "_")
tab@output <- "markdown"
expect_snapshot_print(tab, "group-delim-some-delim.md")
tab <- tt(data.frame(
  id = 1,
  Axa1 = 2,
  Axa2 = "3",
  Bxb1 = 4,
  Bxb2 = 5,
  C = 6
)) |>
  group_tt(j = "x")
tab@output <- "markdown"
expect_snapshot_print(tab, "group-delim-x-delim.md")



# bug: same number of delimiters
dat <- data.frame(
  "A__D" = rnorm(3),
  "A_B_D" = rnorm(3),
  "A_B_" = rnorm(3),
  "_C_E" = rnorm(3),
  check.names = FALSE
)
tt(dat) |> group_tt(j = "_")


# Issue #466: group_tt without delimiter returns a valid table
tab <- tt(data.frame(Aid = 1, Aa1 = 2, Aa2 = "3", Bb1 = 4, Bb2 = 5, BC = 5)) |>
  group_tt(j = "_")
expect_inherits(tab, "tinytable")


# Issue #471: `style_tt()` group
set.seed(48103)
dat <- data.frame(
  label = c("a", "a", "a", "b", "b", "c", "a", "a"),
  x1 = rnorm(8),
  x2 = rnorm(8)
)
tab <- tt(dat[, 2:3]) |>
  group_tt(i = dat$label) |>
  style_tt(i = "groupi", color = "white", background = "black", align = "c")
tab@output <- "latex"
expect_snapshot_print(tab, label = "group_style_tt_group.tex")


# Issue #165: Column labels not centered, last row not styled because of row group offset
k = mtcars[1:5, 1:5]
k[1, 1] <- pi
tab <- tt(k) |>
  format_tt(num_fmt = "decimal", num_zero = TRUE, digits = 3) |>
  style_tt(i = 1:3, j = 1, bold = TRUE) |>
  style_tt(j = 2, italic = TRUE) |>
  group_tt(i = list("hello world" = 2), j = list("foo" = 1:2, "bar" = 3:4))
expect_snapshot_print(
  print_html(tab),
  "group-issue165_html_centering_style.html"
)


# informative errors
expect_error(
  tt(head(iris)) |> style_tt("groupj", color = "orange") |> print(),
  pattern = "No column grouping"
)

# informative errors
x <- head(iris)
colnames(x) <- NULL
expect_error(
  tt(x) |> style_tt("colnames", color = "orange") |> print(),
  pattern = "No column names found"
)
expect_error(
  tt(head(iris)) |> style_tt("groupj", color = "orange") |> print(),
  pattern = "No column grouping"
)


# i = "colnames" vs. "groupj"
tab <- tt(head(iris)) |>
  group_tt(j = list("Foo" = 2:3, "Bar" = 5)) |>
  group_tt(j = list("Hello" = 1:2, "World" = 4:5)) |>
  style_tt("groupj", color = "orange") |>
  style_tt("colnames", color = "teal")
tab <- expect_table(tab, "latex")[["latex"]]
expect_snapshot_print(tab, "group-style_tt_i_colnames_vs_groupj.tex")


# Issue #165: group_tt insert extra row at the bottom
k <- mtcars[1:4, 1:4]
tab <- tt(k) |>
  format_tt(num_fmt = "decimal", num_zero = TRUE, digits = 3) |>
  style_tt(i = 1:3, j = 1, bold = TRUE) |>
  style_tt(j = 2, italic = TRUE) |>
  group_tt(
    i = list("hello world" = 2),
    j = list("foo" = 1:2, "bar" = 3:4)
  )
t <- expect_table(tab)
expect_snapshot_print(t[["latex"]], "group-issue165_extra_row.tex")
expect_snapshot_print(t[["typst"]], "group-issue165_extra_row.typ")
expect_snapshot_print(t[["html"]], "group-issue165_extra_row.html")
expect_snapshot_print(t[["markdown"]], "group-issue165_extra_row.md")


# Issue #518: Multi-level delimiter grouping
# Test basic multi-level structure
tab <- data.frame(
  hi_hi_hi = 1,
  hi_ho_hu = 2
) |>
  tt() |>
  group_tt(j = "_")
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "group-multilevel-basic.md")
expect_snapshot_print(t[["latex"]], "group-multilevel-basic.tex")
expect_snapshot_print(t[["html"]], "group-multilevel-basic.html")
expect_snapshot_print(t[["typst"]], "group-multilevel-basic.typ")

# Mixed-level delimiter structure should error
tab_mixed_levels <- data.frame(
  A_X_1 = 1,
  A_X_2 = 2,
  A_Y_1 = 3,
  B_Z_1 = 4,
  B_Z_2 = 5,
  C_W = 6
)
expect_error(
  tab_mixed_levels |> tt() |> group_tt(j = "_"),
  "Each column name must have the same number of delimiters.",
  fixed = TRUE
)

# Test multi-level structure with empty fields.
# The trailing space in "A_B_ " preserves the delimiter count.
tab <- data.frame(
  "A__C" = 1,
  "A_B_C" = 2,
  "A_B_ " = 3,
  "_D_E" = 4,
  check.names = FALSE
) |>
  tt() |>
  group_tt(j = "_")
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "group-multilevel-empty.md")
expect_snapshot_print(t[["latex"]], "group-multilevel-empty.tex")
expect_snapshot_print(t[["html"]], "group-multilevel-empty.html")
expect_snapshot_print(t[["typst"]], "group-multilevel-empty.typ")


# Issue #543: Multi-level delimiter with shared level names spanning all columns
dat <- data.frame(
  "Strata group\nSex\nStrata level\nMale" = NA,
  "Strata group\nSex\nStrata level\nFemale" = NA,
  "Strata group\nAge\nStrata level\n<18" = NA,
  "Strata group\nAge\nStrata level\n>=18" = NA,
  check.names = FALSE
)
tab <- tt(dat) |> group_tt(j = "\n")
t <- expect_table(tab, "markdown")
expect_snapshot_print(t[["markdown"]], "group-shared-level-spanning.md")

# Indent markdown
dat <- do.call(rbind, by(iris, iris$Species, head, 3))
tab <- tt(dat) |>
  group_tt(i = dat$Species) |>
  style_tt(i = "~groupi", j = 1, indent = 3)
expect_snapshot_print(tab, "group-indent.md")

# nse
t <- mtcars |>
  sort_by(~am) |>
  tt() |>
  group_tt(i = am)
t <- expect_table(t)
expect_snapshot_print(t[["markdown"]], "group-nse.md")
expect_snapshot_print(t[["latex"]], "group-nse.tex")
expect_snapshot_print(t[["html"]], "group-nse.html")
expect_snapshot_print(t[["typst"]], "group-nse.typ")

# Issue #638: group_tt(j) after subset() placed column spans at wrong positions
dat <- mtcars[1:5, ] |>
  setNames(c("A", "B", "C", "D", "E", "F", "G", "H", "Total", "X", "Y"))
tab <- dat |>
  tt() |>
  group_tt(i = list("grp" = 3)) |>
  subset(select = -A) |>
  group_tt(j = list("AAA" = 3:4, "BBB" = 5:6))
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "group-subset_j-638.md")
expect_snapshot_print(t[["latex"]], "group-subset_j-638.tex")
expect_snapshot_print(t[["html"]], "group-subset_j-638.html")
expect_snapshot_print(t[["typst"]], "group-subset_j-638.typ")

source("helpers.R")
using("tinysnapshot")

# Bug: \begin{table}[H][H]
options(tinytable_latex_placement = "H")
x <- mtcars[1:4, 1:4]
tab <- tt(x) |> theme_latex(resize_width = .9, resize_direction = "down")
tab@output <- "latex"
expect_snapshot_print(tab, label = "theme-placement_options_no_doubling")
options(tinytable_latex_placement = "H")

# Issue #206: resize with footnote
k <- data.frame(X = 1) |>
  tt(note = "abc") |>
  theme_latex(resize_width = 0.9, resize_direction = "down") |>
  save_tt("latex")
expect_inherits(k, "character")

# Issue ##439: function theme breaks HTML
theme_mitex <- function(x, ...) {
  fn <- function(table) {
    if (isTRUE(table@output == "typst")) {
      table@table_string <- gsub(
        "\\$(.*?)\\$",
        "#mitex(`\\1`)",
        table@table_string
      )
    }
    return(table)
  }
  x <- style_tt(x, finalize = fn)
  return(x)
}
tab <- data.frame(Math = c("$\\alpha$", "$a_{it}$", "$e^{i\\pi} + 1 = 0$")) |>
  tt(theme = theme_mitex) |>
  save_tt("html")
expect_inherits(tab, "character")

# HTML rotation
tab <- tt(head(mtcars)) |>
  theme_rotate(angle = 45) |>
  save_tt("html")
expect_true(grepl("rotate\\(45deg\\)", tab))


# Issue #460: rowhead is not inserted in LaTeX
tmp <- rbind(mtcars, mtcars)[, 1:6]
cap <- "A long 80\\% width table with repeating headers."
tab <- tt(tmp, width = .8, caption = cap) |>
  theme_latex(multipage = TRUE, rowhead = 1) |>
  save_tt("latex")
expect_true(grepl("rowhead=1", tab))


# theme_tt() deprecation warning
expect_warning(theme_tt(tt(head(iris)), "striped"), pattern = "deprecated")


# theme_default() deprecation warning
expect_warning(theme_default(tt(head(iris))), pattern = "deprecated")


# Issue #531: style_tt() overrides triped theme
iris_dt <- do.call(rbind, by(iris, ~Species, head, 2))
cap <- "Stripes override purple and teal, but not yellow."
t = tab <- tt(iris_dt, theme = "empty", caption = cap) |>
  style_tt(i = 1:2, j = 1, background = "#4B0055") |>
  style_tt(i = 3:4, j = 1, background = "#009B95") |>
  theme_striped() |>
  style_tt(i = 5:6, j = 1, background = "#FDE333")
expect_snapshot_print(
  print_html(tab),
  "theme-issue531_style_colors_override_stripes.html"
)

source("helpers.R")
using("tinysnapshot")

# Basic matrix insertion - single position, single row
rowmat <- matrix(c("Inserted", "Row", "Data", "Here", "Extra"), nrow = 1)
tab <- tt(head(iris, 5)) |> group_tt(i = 3, j = rowmat)
t <- expect_table(tab)
expect_snapshot_print(
  t[["markdown"]],
  "group_matrix-single_position_single_row.md"
)
expect_snapshot_print(
  t[["latex"]],
  "group_matrix-single_position_single_row.tex"
)
expect_snapshot_print(
  t[["typst"]],
  "group_matrix-single_position_single_row.typ"
)
expect_snapshot_print(
  t[["html"]],
  "group_matrix-single_position_single_row.html"
)

# Matrix insertion - multiple positions and rows (covers both multiple positions and multiple rows)
rowmat <- matrix(
  c(
    "Row1",
    "At",
    "Pos1",
    "Data",
    "A",
    "Row2",
    "At",
    "Pos1",
    "Data",
    "B",
    "Row3",
    "At",
    "Pos2",
    "Data",
    "C"
  ),
  nrow = 3,
  byrow = TRUE
)
tab <- tt(head(iris, 7)) |> group_tt(i = c(2, 2, 5), j = rowmat)
t <- expect_table(tab)
expect_snapshot_print(
  t[["markdown"]],
  "group_matrix-multiple_positions_multiple_rows.md"
)
expect_snapshot_print(
  t[["latex"]],
  "group_matrix-multiple_positions_multiple_rows.tex"
)
expect_snapshot_print(
  t[["typst"]],
  "group_matrix-multiple_positions_multiple_rows.typ"
)
expect_snapshot_print(
  t[["html"]],
  "group_matrix-multiple_positions_multiple_rows.html"
)

# Stacked matrix insertions at edge positions (position 1 and last position)
# used to error with "Only one group row insertion is allowed at a time"
tab <- tt(head(iris, 4)) |>
  group_tt(
    i = 1,
    j = matrix(c("Header", "Row", "At", "Top", "Position"), nrow = 1)
  ) |>
  group_tt(
    i = 6,
    j = matrix(c("Footer", "Row", "At", "Bottom", "Position"), nrow = 1)
  )
expect_inherits(tab, "tinytable")
expect_equal(tab@group_index_i, c(1, 6))
md <- strsplit(save_tt(tab, "markdown"), "\n")[[1]]
expect_true(any(grepl("Header", md)))
expect_true(any(grepl("Footer", md)))
expect_true(min(grep("Header", md)) < min(grep("5.1", md, fixed = TRUE)))
expect_true(max(grep("Footer", md)) > max(grep("4.6", md, fixed = TRUE)))

# Matrix with single column reshape and styling combined
rowmat <- matrix(c("A", "B", "C", "D", "E"))
tab <- tt(head(iris, 4)) |>
  group_tt(i = 2, j = rowmat) |>
  style_tt(i = "groupi", background = "lightblue")
t <- expect_table(tab)
expect_snapshot_print(
  t[["markdown"]],
  "group_matrix-single_column_with_styling.md"
)
expect_snapshot_print(
  t[["latex"]],
  "group_matrix-single_column_with_styling.tex"
)
expect_snapshot_print(
  t[["typst"]],
  "group_matrix-single_column_with_styling.typ"
)
expect_snapshot_print(
  t[["html"]],
  "group_matrix-single_column_with_styling.html"
)

# Matrix row duplication - single row matrix with multiple positions
rowmat <- matrix(colnames(iris))
tab <- tt(head(iris, 7)) |> group_tt(i = c(2, 5), j = rowmat)
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "group_matrix-row_duplication.md")
expect_snapshot_print(t[["latex"]], "group_matrix-row_duplication.tex")
expect_snapshot_print(t[["typst"]], "group_matrix-row_duplication.typ")
expect_snapshot_print(t[["html"]], "group_matrix-row_duplication.html")

source("helpers.R")
using("tinysnapshot")


# Issue #514: vignette white lines with blue background
dat <- data.frame(1:2, 3:4, 5:6, 7:8)
tab <- tt(dat, theme = "empty", colnames = FALSE) |>
    style_tt(
        line = "tblr",
        line_color = "white",
        line_width = 0.5,
        background = "blue",
        color = "white"
    )
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "style-issue514_white_blue.md")
expect_snapshot_print(t[["latex"]], "style-issue514_white_blue.tex")
expect_snapshot_print(t[["typst"]], "style-issue514_white_blue.typ")
expect_snapshot_print(t[["html"]], "style-issue514_white_blue.html")


# Issue #507: Test group row styling functionality
tab <- tt(mtcars[1:8, 1:3]) |>
    group_tt(i = list("Hello" = 3, "World" = 6)) |>
    group_tt(j = list("Hello" = 1, "World" = 2:3)) |>
    style_tt(i = 2:4, italic = TRUE) |>
    style_tt(i = 6, strikeout = TRUE) |>
    style_tt(i = 8, bold = TRUE)
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "style-issue507_markdown_styles.md")
expect_snapshot_print(t[["latex"]], "style-issue507_markdown_styles.tex")
expect_snapshot_print(t[["typst"]], "style-issue507_markdown_styles.typ")
expect_snapshot_print(t[["html"]], "style-issue507_markdown_styles.html")


# Issue #512: Many tables with identical output
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    style_tt("colnames", italic = TRUE) |>
    style_tt(-1, bold = TRUE)
expect_snapshot_print(tab, "style-groupj_colnames.md")
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    style_tt(0, italic = TRUE) |>
    style_tt(-1, bold = TRUE)
expect_snapshot_print(tab, "style-groupj_colnames.md")
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    format_tt(0, sprintf = "_%s_") |>
    format_tt(-1, sprintf = "**%s**")
expect_snapshot_print(tab, "style-groupj_colnames.md")
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    format_tt("colnames", sprintf = "_%s_") |>
    format_tt(-1, sprintf = "**%s**")
expect_snapshot_print(tab, "style-groupj_colnames.md")

tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    style_tt("colnames", italic = TRUE) |>
    style_tt("groupj", bold = TRUE)
expect_snapshot_print(tab, "style-groupj_colnames.md")


# Issue #564: Empty which() index should proceed silently
dat <- data.frame(feed = c("fish", "meat", "soy"), weight = c(10, 20, 30))
tab1 <- tt(dat)
tab2 <- tt(dat)

# These should not error when no rows match the condition
expect_silent(style_tt(tab1, i = {
    feed == "cool"
}, bold = TRUE))
expect_silent(style_tt(tab2, i = which(feed == "cool"), bold = TRUE))


# smallcap styling functionality
tab <- tt(head(iris),
    caption = "Motor Trend Car Road Tests",
    notes = "Source: Henderson and Velleman (1981)") |>
    style_tt(i = "colnames", smallcap = TRUE) |>
    style_tt(i = "caption", smallcap = TRUE) |>
    style_tt(i = "notes", smallcap = TRUE) |>
    style_tt(i = 1:3, j = 5, smallcap = TRUE)
t <- expect_table(tab)
expect_snapshot_print(t[["html"]], "style-smallcap.html")
expect_snapshot_print(t[["latex"]], "style-smallcap.tex")
expect_snapshot_print(t[["typst"]], "style-smallcap.typ")
expect_snapshot_print(t[["markdown"]], "style-smallcap.md")


# Caption and notes styles must not duplicate the preceding cell style frame
tab <- tt(
    data.frame(a = 1:2),
    caption = "Caption",
    notes = "Note") |>
    style_tt(i = 1, line = "b", line_color = "red", line_width = 0.123) |>
    style_tt(i = "caption", bold = TRUE) |>
    style_tt(i = "notes", italic = TRUE)
tab <- tinytable:::build_tt(tab, output = "typst")
user_lines <- tab@style_lines[
    tab@style_lines$line_color == "red" & tab@style_lines$line_width == 0.123,
    ,
    drop = FALSE]
expect_equal(nrow(user_lines), 1L)
line_matches <- gregexpr(
    'stroke: 0.123em + rgb("#FF0000")',
    tab@table_string,
    fixed = TRUE)[[1]]
expect_equal(sum(line_matches > 0L), 1L)


# partial align
tab <- tt(mtcars[1:5, 1:6]) |> style_tt(j = c(2, 4), align = "cr")
t <- expect_table(tab)
expect_snapshot_print(t[["html"]], "style-align_partial.html")
expect_snapshot_print(t[["latex"]], "style-align_partial.tex")
expect_snapshot_print(t[["typst"]], "style-align_partial.typ")
expect_snapshot_print(t[["markdown"]], "style-align_partial.md")


# logical matrix with all FALSE values should not error
testdata <- data.frame(
  names = c('a', 'b', 'c', 'd', 'e'),
  values1 = c(1, 2, 3, 4, 5),
  values2 = c(6, 7, 8, 9, 10)
)
testdata_override <- matrix(FALSE, nrow = nrow(testdata), ncol = ncol(testdata))
result <- testdata |>
  tt() |>
  style_tt(i = testdata_override, background = "red") |>
  save_tt("html")
expect_true(is.character(result))
expect_true(nchar(result) > 0)


# Issue #647: Conditional styling with output argument
tab <- data.frame(A = 1:2, B = 3:4) |>
  tt() |>
  style_tt(bold = TRUE, output = "typst") |>
  style_tt(italic = TRUE, output = "html")
t <- expect_table(tab, formats = c("html", "typst", "latex"))
expect_snapshot_print(t[["html"]], "style-output_conditional.html")
expect_snapshot_print(t[["typst"]], "style-output_conditional.typ")
expect_snapshot_print(t[["latex"]], "style-output_conditional.tex")


# Deprecated shims must warn AND take effect (tabularray_inner/outer,
# html_class, html_css_rule). Regression: `theme_latex()` was called without
# `x`, and html shim results were discarded.
tab <- style_tt(tt(mtcars[1:3, 1:3]), tabularray_inner = "rowsep=8pt")
expect_warning(build_tt(tab, "latex"), pattern = "tabularray_inner")
b <- suppressWarnings(build_tt(tab, "latex"))
expect_true(grepl("rowsep=8pt", b@table_string, fixed = TRUE))

tab <- style_tt(tt(mtcars[1:3, 1:3]), tabularray_outer = "baseline=T")
expect_warning(build_tt(tab, "latex"), pattern = "tabularray_outer")
b <- suppressWarnings(build_tt(tab, "latex"))
expect_true(grepl("baseline=T", b@table_string, fixed = TRUE))

tab <- style_tt(tt(mtcars[1:3, 1:3]), html_class = "table-dark-x")
expect_warning(build_tt(tab, "html"), pattern = "html_class")
b <- suppressWarnings(build_tt(tab, "html"))
expect_true(grepl("table-dark-x", b@table_string, fixed = TRUE))

tab <- style_tt(tt(mtcars[1:3, 1:3]), html_css_rule = ".mystyle { color: pink; }")
expect_warning(build_tt(tab, "html"), pattern = "html_css_rule")
b <- suppressWarnings(build_tt(tab, "html"))
expect_true(grepl(".mystyle { color: pink; }", b@table_string, fixed = TRUE))


# Empty row selection with align must be a silent no-op, not a build error
# Regression: "replacement has 1 row, data has 0"
tab <- style_tt(tt(mtcars[1:3, 1:3]), i = integer(0), align = "c")
b <- build_tt(tab, "html")
expect_inherits(b, "tinytable")


# Logical-matrix `i` with per-cell style vectors: values recycle over TRUE
# cells in column-major order (which(i, arr.ind = TRUE))
d <- data.frame(a = 1:3, b = 4:6)
m <- matrix(FALSE, 3, 2)
m[2, 1] <- TRUE
m[1, 2] <- TRUE
b <- build_tt(style_tt(tt(d), i = m, color = c("red", "blue")), "html")
sty <- b@style[!is.na(b@style$color), ]
expect_equal(nrow(sty), 2)
expect_equal(sty$color[sty$i == 2 & sty$j == 1], "red")
expect_equal(sty$color[sty$i == 1 & sty$j == 2], "blue")

# length-1 values still recycle over all TRUE cells
b <- build_tt(style_tt(tt(d), i = m, color = "green"), "html")
sty <- b@style[!is.na(b@style$color), ]
expect_equal(sty$color[sty$i == 2 & sty$j == 1], "green")
expect_equal(sty$color[sty$i == 1 & sty$j == 2], "green")

# wrong vector length errors informatively (valid lengths: 1 or #TRUE cells)
expect_error(
  build_tt(style_tt(tt(d), i = m, color = c("a", "b", "c")), "html"),
  pattern = "one of these lengths: 1, 2"
)

# logical matrix with wrong dimensions must error, not degrade silently
m2 <- matrix(TRUE, 2, 2)
expect_error(
  build_tt(style_tt(tt(d), i = m2, bold = TRUE), "html"),
  pattern = "dimensions must match"
)

# align is ignored (with a warning) when `i` is a logical matrix
expect_warning(
  build_tt(style_tt(tt(d), i = m, align = "r"), "html"),
  pattern = "logical matrix"
)


# Value-to-column mapping honors USER-specified `j` order, consistent with `i`
# Regression: j = c(3, 1) used to sort and put the first value on column 1
d3 <- data.frame(a = 1:2, b = 3:4, c = 5:6)
b <- build_tt(style_tt(tt(d3), j = c(3, 1), background = c("red", "blue")), "html")
sty <- b@style[!is.na(b@style$background), ]
expect_true(all(sty$background[sty$j == 3] == "red"))
expect_true(all(sty$background[sty$j == 1] == "blue"))

# character `j` also honors user order
b <- build_tt(style_tt(tt(d3), j = c("c", "a"), background = c("red", "blue")), "html")
sty <- b@style[!is.na(b@style$background), ]
expect_true(all(sty$background[sty$j == 3] == "red"))
expect_true(all(sty$background[sty$j == 1] == "blue"))

# `i` user-order behavior is unchanged
b <- build_tt(style_tt(tt(d3), i = c(2, 1), j = 1, background = c("red", "blue")), "html")
sty <- b@style[!is.na(b@style$background), ]
expect_true(all(sty$background[sty$i == 2] == "red"))
expect_true(all(sty$background[sty$i == 1] == "blue"))

# Compound line directions do not duplicate non-line properties in @style
b <- build_tt(style_tt(tt(d3, theme = "empty"), i = 1, j = 1, line = "tb", background = "red"), "html")
expect_equal(sum(!is.na(b@style$background)), 1)
expect_equal(sum(!is.na(b@style$line)), 2)

source("helpers.R")
using("tinysnapshot")

# Bug: \begin{table}[H][H]
options(tinytable_latex_placement = "H")
x <- mtcars[1:4, 1:4]
tab <- tt(x) |> theme_latex(resize_width = .9, resize_direction = "down")
tab@output <- "latex"
expect_snapshot_print(tab, label = "theme-placement_options_no_doubling")
options(tinytable_latex_placement = "H")

# Issue #206: resize with footnote
k <- data.frame(X = 1) |>
  tt(note = "abc") |>
  theme_latex(resize_width = 0.9, resize_direction = "down") |>
  save_tt("latex")
expect_inherits(k, "character")

# Issue ##439: function theme breaks HTML
theme_mitex <- function(x, ...) {
  fn <- function(table) {
    if (isTRUE(table@output == "typst")) {
      table@table_string <- gsub(
        "\\$(.*?)\\$",
        "#mitex(`\\1`)",
        table@table_string
      )
    }
    return(table)
  }
  x <- style_tt(x, finalize = fn)
  return(x)
}
tab <- data.frame(Math = c("$\\alpha$", "$a_{it}$", "$e^{i\\pi} + 1 = 0$")) |>
  tt(theme = theme_mitex) |>
  save_tt("html")
expect_inherits(tab, "character")

# HTML: always cell-level rotation (no whole-table rotation)
tab <- tt(head(mtcars)) |>
  theme_rotate(angle = 45) |>
  save_tt("html")
expect_true(grepl("inline-block.*rotate\\(45deg\\)", tab))

# HTML: specific cells
tab <- tt(head(mtcars)) |>
  theme_rotate(i = 0, j = 1, angle = 45) |>
  save_tt("html")
expect_true(grepl("inline-block.*rotate\\(45deg\\)", tab))

# LaTeX: whole-table rotation when no i/j
tab <- tt(head(mtcars)) |>
  theme_rotate(angle = 45) |>
  save_tt("latex")
expect_true(grepl("\\\\rotatebox\\{45\\}", tab))

# LaTeX: cell-level rotation with i/j
tab <- tt(head(mtcars)) |>
  theme_rotate(i = 1, j = 1, angle = 45) |>
  save_tt("latex")
expect_true(grepl("\\\\rotatebox\\{45\\}", tab))

# Typst: whole-table rotation when no i/j
tab <- tt(head(mtcars)) |>
  theme_rotate(angle = 45) |>
  save_tt("typst")
expect_true(grepl("#rotate\\(-45deg, reflow: true", tab))

# Typst: cell-level rotation with i/j
tab <- tt(head(mtcars)) |>
  theme_rotate(i = 1, j = 1, angle = 45) |>
  save_tt("typst")
expect_true(grepl("#rotate\\(-45deg", tab))


# Issue #460: rowhead is not inserted in LaTeX
tmp <- rbind(mtcars, mtcars)[, 1:6]
cap <- "A long 80\\% width table with repeating headers."
tab <- tt(tmp, width = .8, caption = cap) |>
  theme_latex(multipage = TRUE, rowhead = 1) |>
  save_tt("latex")
expect_true(grepl("rowhead=1", tab))


# theme_tt() deprecation warning
expect_warning(theme_tt(tt(head(iris)), "striped"), pattern = "deprecated")


# theme_default() deprecation warning
expect_warning(theme_default(tt(head(iris))), pattern = "deprecated")


# Issue #531: style_tt() overrides triped theme
iris_dt <- do.call(rbind, by(iris, ~Species, head, 2))
cap <- "Stripes override purple and teal, but not yellow."
t = tab <- tt(iris_dt, theme = "empty", caption = cap) |>
  style_tt(i = 1:2, j = 1, background = "#4B0055") |>
  style_tt(i = 3:4, j = 1, background = "#009B95") |>
  theme_striped() |>
  style_tt(i = 5:6, j = 1, background = "#FDE333")
expect_snapshot_print(
  print_html(tab),
  "theme-issue531_style_colors_override_stripes.html"
)

source("helpers.R")
using("tinysnapshot")
options(tinytable_print_output = "typst")

# Use Typst's native table kind so captions inherit the document language and
# share counters and selectors with other tables.
tab <- tt(head(iris), caption = "A caption") |> save_tt("typst")
expect_true(grepl("kind: table", tab, fixed = TRUE))
expect_false(grepl("supplement:", tab, fixed = TRUE))

# semi complicated
tab <- tt(mtcars[1:4, 1:5], caption = "Hello World") |>
  group_tt(j = list("Group 1" = 4:5, "Group 2" = 2:3)) |>
  style_tt(j = 1:5, align = "lcccr") |>
  style_tt(
    i = 2,
    j = 1:3,
    strikeout = TRUE,
    bold = TRUE,
    background = "black",
    color = "white"
  ) |>
  style_tt(j = 1, color = "red", italic = TRUE)
expect_snapshot_print(tab, label = "typst-complicated.typ")

# issue #139
tab <- tt(mtcars[1:10, 1:5]) |>
  group_tt(i = list("Feta" = 2, "Brie" = 6)) |>
  group_tt(j = list("First" = 2:3, "Second" = 4:5)) |>
  style_tt(1:5, align = "c", background = "blue", color = "white")
expect_snapshot_print(
  tab,
  label = "typst-issue-139_misaligned_rule_with_group_tt.typ"
)

# Theme striped
x <- mtcars[1:4, 1:5]
tab <- tt(x, theme = "striped")
expect_snapshot_print(tab, label = "typst-striped.typ")

# Theme grid
tab <- tt(x, theme = "grid")
expect_snapshot_print(tab, label = "typst-grid.typ")

# Formatting
dat <- data.frame(
  w = c(143002.2092, 201399.181, 100188.3883),
  x = c(1.43402, 201.399, 0.134588),
  y = as.Date(c(897, 232, 198), origin = "1970-01-01"),
  z = c(TRUE, TRUE, FALSE)
)
dat <- tt(dat, digits = 2)
expect_snapshot_print(dat, label = "typst-formatting.typ")

# More formatting
dat <- data.frame(
  a = c("Burger", "Halloumi", "Tofu", "Beans"),
  b = c(1.43202, 201.399, 0.146188, 0.0031),
  c = c(98938272783457, 7288839482, 29111727, 93945)
)
tab <- tt(dat) |>
  format_tt(j = "a", sprintf = "Food: %s") |>
  format_tt(j = 2, digits = 1) |>
  format_tt(j = "c", digits = 2, num_suffix = TRUE)
expect_snapshot_print(tab, label = "typst-more_formatting.typ")

# Significant cell
dat <- data.frame(x = c(0.000123456789, 12.4356789))
tab <- tt(dat) |> format_tt(digits = 2, num_fmt = "significant_cell")
expect_snapshot_print(tab, label = "typst-significant_cell.typ")

# Missing value replacement
tab <- tt(data.frame(a = c(NA, 1, 2), b = c(3, NA, 5)))
tab <- format_tt(tab, replace = "-")
expect_snapshot_print(tab, label = "typst-missing_value_replacement.typ")

# Italic markdown
dat <- data.frame(markdown = c("This is _italic_ text."))
tab <- tt(dat) |>
  format_tt(j = 1, markdown = TRUE) |>
  style_tt(j = 1, align = "c")
expect_snapshot_print(tab, label = "typst-italic_markdown.typ")

# Font size
dat <- tt(x) |> style_tt(j = "mpg|hp|qsec", fontsize = 1.5)
expect_snapshot_print(dat, label = "typst-font_size.typ")

# No headers
k <- x
colnames(k) <- NULL
k <- tt(k)
expect_snapshot_print(k, label = "typst-no_headers.typ")

# Portable images
if (requiet("base64enc")) {
  img <- tempfile(fileext = ".png")
  grDevices::png(img, width = 20, height = 20)
  graphics::par(mar = c(0, 0, 0, 0))
  graphics::plot.new()
  graphics::rect(0, 0, 1, 1, col = "black", border = NA)
  grDevices::dev.off()

  tab <- tt(data.frame(x = "")) |>
    plot_tt(j = 1, images = img) |>
    theme_typst(portable = TRUE) |>
    save_tt("typst")

  expect_true(grepl("#image(bytes(", tab, fixed = TRUE))
  expect_true(grepl("data:image/png;base64,", tab, fixed = TRUE))
  expect_false(grepl(img, tab, fixed = TRUE))

  unlink(img)
}

# Group rows
dat <- mtcars[1:9, 1:8]
dat <- tt(dat) |>
  group_tt(
    i = list(
      "I like (fake) hamburgers" = 3,
      "She prefers halloumi" = 4,
      "They love tofu" = 7
    )
  )
expect_snapshot_print(dat, label = "typst-group_rows.typ")

# Group columns
dat <- mtcars[1:9, 1:8]
tab <- tt(dat) |>
  group_tt(
    j = list(
      "Hamburgers" = 1:3,
      "Halloumi" = 4:5,
      "Tofu" = 7
    )
  )
expect_snapshot_print(tab, label = "typst-group_columns.typ")

# issue #323
dat <- mtcars[1:9, 1:8]
tab <- tt(dat) |>
  group_tt(
    i = list(
      "I like (fake) hamburgers" = 3,
      "She prefers halloumi" = 4,
      "They love tofu" = 7
    ),
    j = list("Hamburgers" = 1:3, "Halloumi" = 4:5, "Tofu" = 7)
  ) |>
  style_tt(
    i = c(3, 5, 9),
    align = "c",
    background = "black",
    color = "orange"
  ) |>
  style_tt(i = -1, color = "orange")
tab@output <- "typst"
expect_snapshot_print(tab, label = "typst-issue323_group_tt_style_tt.typ")

# Frame
tab <- tt(mtcars[1:5, 1:5]) |>
  style_tt(2:3, 2:3, line_color = "red", line = "tblr", line_width = .05)
tab@output <- "typst"
expect_snapshot_print(tab, label = "typst-tblr.typ")

# Issue #357
tab <- tt(head(iris), notes = "blah") |> save_tt("typst")
expect_true(grepl("blah", tab))

# Bug: Footnotes with cell coordinates #456
tab <- tt(
  mtcars[1:4, 1:5],
  notes = list(
    a = list(i = c(0:1), j = 1, text = "Blah."),
    b = "Blah blah."
  )
)
expect_snapshot_print(tab, label = "typst-issue456.typ")

# Single row colspan bug fix
df <- data.frame(
  column_a = c("No data"),
  column_b = c(NA),
  column_c = c(NA)
)
tab <- df |>
  tt() |>
  style_tt(i = 1, j = 1, colspan = ncol(df), align = "c") |>
  save_tt("typst")
expect_true(is.character(tab) && nchar(tab) > 0)


# Figure alignment is opt-in for Typst output
tab <- tt(mtcars[1:4, 1:5]) |>
  save_tt("typst")
expect_false(grepl("#align(", tab, fixed = TRUE))

tab <- tt(mtcars[1:4, 1:5]) |>
  theme_typst(align_figure = "c") |>
  save_tt("typst")
expect_true(grepl("#align(center,", tab, fixed = TRUE))

tab <- tt(mtcars[1:4, 1:5]) |>
  theme_typst(align_figure = "r") |>
  save_tt("typst")
expect_true(grepl("#align(right,", tab, fixed = TRUE))


# theme_typst(figure = FALSE) drops only the figure wrapper: the #block[]
# preamble defining get-style() must survive or the document cannot compile
tab <- tt(mtcars[1:4, 1:5]) |>
  theme_typst(figure = FALSE)
expect_snapshot_print(tab, label = "typst-figure_false.typ")


# Typst resizing follows theme_latex() API
tab <- tt(mtcars[1:4, 1:5]) |>
  theme_typst(resize_width = 0.8, resize_direction = "down") |>
  save_tt("typst")
expect_true(grepl("#layout(size => {", tab, fixed = TRUE))
expect_true(grepl("let target-width = size.width * 0.8", tab, fixed = TRUE))
expect_true(grepl("if body-size.width > target-width", tab, fixed = TRUE))
expect_true(grepl("scale(x: factor, y: factor, reflow: true, body)", tab, fixed = TRUE))

tab <- tt(mtcars[1:4, 1:5]) |>
  theme_typst(resize_height = 0.6, resize_direction = "up") |>
  save_tt("typst")
expect_true(grepl("let target-height = size.height * 0.6", tab, fixed = TRUE))
expect_true(grepl("if body-size.height < target-height", tab, fixed = TRUE))

expect_error(
  theme_typst(tt(mtcars[1:4, 1:5]), resize_direction = "sideways"),
  "resize_direction"
)


# Issue #592
table <- data.frame(
  stringsAsFactors = FALSE,
  check.names = FALSE,
  metric = c(
    "Adoption",
    "Adoption",
    "Adoption",
    "Adoption",
    "Adoption",
    "NPS",
    "NPS",
    "NPS",
    "NPS",
    "NPS",
    "Overall Satisfaction",
    "Overall Satisfaction",
    "Overall Satisfaction",
    "Overall Satisfaction",
    "Overall Satisfaction"),
  tier = c(
    "1",
    "2 ", "3 ", "3 ", "3 ", "1",
    "2 ", "3 ", "3 ", "3 ", "1",
    "2 ", "3 ", "3 ", "3"),
  role = c(
    "",
    "x", "x", "y", "z", "",
    "x", "x", "y", "z", "",
    "x", "x", "y", "z"),
  Overall = c(
    "90.9%", "87.9%", "90.6%", "90.6%", "100%", "19",
    "39.9", "47.2", "14.6", "100", "80.7%",
    "90.4%", "80.9%", "91.8%", "100%"),
  state1 = c(
    "90.5%", "91%", "91%", "92.3%", "100%", "24.2",
    "40.3", "51.4", "11.5", "100", "82.4%",
    "89.4%", "78.4%", "91.6%", "100%"),
  state2 = c(
    "95.7%", "75%", NA, NA, NA, "11.5", "45.7", NA,
    NA, NA, "79.7%", "94.4%", NA, NA, NA),
  state3 = c(
    "91.5%", "85.9%", "95.4%", "100%", NA, "23.8",
    "44.7", "70.8", "22.2", NA, "82.4%", "88.1%",
    "55.6%", "98.4%", NA)
)
tab <- tt(table) |>
  theme_empty() |>
  style_tt(i = c(3, 8, 13), j = 2, rowspan = 3, alignv = "m") |>
  style_tt(i = 0, line = "tblr") |>
  style_tt(j = 1, line = "l") |>
  style_tt(j = 2, i = 1:nrow(table), alignv = "m", line = "lt") |>
  style_tt(j = 3:length(table), i = 1:nrow(table), line = "r") |>
  style_tt(j = 4:length(table), i = 1:nrow(table), line = "t") |>
  style_tt(j = 3, i = c(1, 2, 3, 6, 7, 8, 11, 12, 13, 16), line = "t") |>
  style_tt(
    i = seq(1, nrow(table), by = 5),
    j = 1,
    rowspan = 5,
    align = "c",
    alignv = "m",
    line = "lt"
  ) |>
  style_tt(i = nrow(table), j = 1:length(table), line = "b") |>
  format_tt(linebreak = "\n", replace = "", output = "typst")
expect_snapshot_print(tab, label = "typst-issue592.typ")



options(tinytable_print_output = NULL)


# Column-group underlines must not merge into a single rule when two groups
# are adjacent. Typst `table.hline()` snaps to column boundaries and has no
# trim option, so the rules are drawn inside the group header cells with
# `place()`, which stops short of the column edges.
tab <- tt(head(mtcars, 4)) |>
  group_tt(j = list("Cyl" = 1:3, "Disp" = 4:6)) |>
  style_tt(i = 1, background = "pink")
out <- build_tt(tab, "typst")@table_string
expect_true(grepl(
  "table.cell(colspan: 3, align: center)[Cyl #place(bottom, dy: 0.4em, line(length: 100%, stroke: 0.03em))]",
  out,
  fixed = TRUE
))
expect_true(grepl(
  "table.cell(colspan: 3, align: center)[Disp #place(bottom, dy: 0.4em, line(length: 100%, stroke: 0.03em))]",
  out,
  fixed = TRUE
))
# the replaced hline is not also drawn at the table level
expect_false(grepl("table.hline(y: 1,", out, fixed = TRUE))
# other rules are untouched
expect_true(grepl("table.hline(y: 2,", out, fixed = TRUE))

# Single-column groups get the same treatment
tab <- tt(head(mtcars, 4)) |> group_tt(j = list("A" = 1, "B" = 2))
out <- build_tt(tab, "typst")@table_string
expect_true(grepl(
  "[A #place(bottom, dy: 0.4em, line(length: 100%, stroke: 0.03em))]",
  out,
  fixed = TRUE
))

# Issue #679: an unspecified `line_color` must not be hard-coded to black.
# Typst stroke folding then lets the document's own styling supply the paint.
tab <- tt(head(mtcars, 4)) |>
  style_tt(i = c(1, 3), line = "b", line_width = 0.05)
out <- build_tt(tab, "typst")@table_string
expect_false(grepl("black", out, fixed = TRUE))
expect_true(grepl("table.hline(y: 2, start: 0, end: 11, stroke: 0.05em),", out, fixed = TRUE))
# dashed rules omit `paint:` rather than the whole dictionary
tab <- tt(head(mtcars, 4)) |> style_tt(i = 1, line = "b", line_type = "dashed")
out <- build_tt(tab, "typst")@table_string
expect_true(grepl('stroke: (thickness: 0.1em, dash: "dashed")', out, fixed = TRUE))
# an explicit color is still honored
tab <- tt(head(mtcars, 4)) |> style_tt(i = 1, line = "b", line_color = "teal")
out <- build_tt(tab, "typst")@table_string
expect_true(grepl('stroke: 0.1em + rgb("#008080")', out, fixed = TRUE))

# Issue #669: notes must not stretch the table.
dat669 <- data.frame(
  Grp = c("I", "II"),
  ColA = c("2,000", "5,000"),
  ColB = c("10,000", "20,000"),
  Ratio = c("20%", "25%")
)
note669 <- "This is a table note that is moderately long and causes the table to stretch."
tab <- tt(dat669, notes = note669)
out <- build_tt(tab, "typst")@table_string
# proportional fr columns computed from measured natural widths
expect_true(grepl("tinytable-coldata", out, fixed = TRUE))
expect_true(grepl("columns: tinytable-naturals.map(w => w.pt() * 1fr),", out, fixed = TRUE))
# notes are capped at the table's natural width
expect_true(grepl("calc.min(tinytable-naturals.sum(), size.width)", out, fixed = TRUE))
# scalar width keeps equal fixed columns, consistent with other formats
tab <- tt(dat669, width = 0.8, notes = note669)
out <- build_tt(tab, "typst")@table_string
expect_true(grepl("columns: (20.00%, 20.00%, 20.00%, 20.00%),", out, fixed = TRUE))
expect_false(grepl("tinytable-coldata", out, fixed = TRUE))
# no notes and no width: plain auto columns, no measurement machinery
tab <- tt(dat669)
out <- build_tt(tab, "typst")@table_string
expect_true(grepl("columns: (auto, auto, auto, auto),", out, fixed = TRUE))
expect_false(grepl("tinytable-coldata", out, fixed = TRUE))
# per-column widths keep fixed percentages
tab <- tt(dat669, width = c(0.1, 0.2, 0.2, 0.3))
out <- build_tt(tab, "typst")@table_string
expect_true(grepl("columns: (10.00%, 20.00%, 20.00%, 30.00%),", out, fixed = TRUE))

source("helpers.R")
using("tinysnapshot")

if (!is_local) {
  exit_file("Run on Vincent's machine")
}

# vector row labels
set.seed(48103)
dat <- data.frame(
  label = c("a", "a", "a", "b", "b", "c", "a", "a"),
  x1 = rnorm(8),
  x2 = rnorm(8)
)
tab <- tt(dat[, 2:3]) |> group_tt(i = dat$label)
tab@output <- "markdown"
expect_snapshot_print(tab, label = "group-vector_row_labels.md")

# 3 level tests across formats
x <- mtcars[1:3, 1:5]
tab <- tt(x) |>
  group_tt(j = list("a" = 2:3, "b" = 4:5)) |>
  group_tt(j = list("c" = 1:2, "d" = 3:5)) |>
  group_tt(j = list("e" = 1:3, "f" = 4))
t <- expect_table(tab, formats = c("markdown", "latex", "typst"))
expect_snapshot_print(t[["markdown"]], "group-3level.md")
expect_snapshot_print(t[["latex"]], "group-3level.tex")
expect_snapshot_print(t[["typst"]], "group-3level.typ")

tab <- tt(mtcars[1:10, 1:5]) |>
  group_tt(
    i = list("Hello" = 3, "World" = 8),
    j = list("Foo" = 2:3, "Bar" = 4:5)
  )
expect_snapshot_print(print_html(tab), "group-html_tutorial_01.html")


# Issue #258: group_tt indices with 0 and duplicates
tab1 <- tt(head(iris)) |> group_tt(i = list("hello" = 1))
t1 <- expect_table(tab1)
expect_snapshot_print(t1[["latex"]], "group-issue258_01_latex.tex")
expect_snapshot_print(t1[["typst"]], "group-issue258_01_typst.typ")
expect_snapshot_print(t1[["markdown"]], "group-issue258_01_markdown.md")
expect_snapshot_print(t1[["html"]], "group-issue258_01_html.html")

tab2 <- tt(head(iris)) |> group_tt(i = list("hello" = 2, "world" = 2))
t2 <- expect_table(tab2)
expect_snapshot_print(t2[["latex"]], "group-issue258_02_latex.tex")
expect_snapshot_print(t2[["typst"]], "group-issue258_02_typst.typ")
expect_snapshot_print(t2[["markdown"]], "group-issue258_02_markdown.md")
expect_snapshot_print(t2[["html"]], "group-issue258_02_html.html")

# Issue #362: group_tt with duplicate column group labels in latex
tab <- data.frame(
  sub1 = 1,
  sub2 = 2,
  sub1 = 3,
  sub2 = 4,
  check.names = FALSE
) |>
  tt() |>
  group_tt(j = list(a = 1:2, a = 3:4))
tab@output <- "latex"
expect_snapshot_print(tab, "group-issue362_duplicate_colum_labels.tex")

# Issue #413: Automatic row groups fail if column is a factor
df <- mtcars |> head(10)
df <- df[order(df$am), ]
df$am <- factor(df$am)
tab <- tt(df) |> group_tt(i = df$am)
expect_inherits(tab, "tinytable")
expect_inherits(save_tt(tab, "markdown"), "character")


# delimiters in column names
# No delimiter
tab <- tt(mtcars |> head())
tab@output <- "markdown"
expect_snapshot_print(tab, "group-delim-no-delim.md")
tab <- tt(data.frame(
  A_id = 1,
  A_a1 = 2,
  A_a2 = "3",
  B_b1 = 4,
  B_b2 = 5,
  B_C = 6
)) |>
  group_tt(j = "_")
tab@output <- "markdown"
expect_snapshot_print(tab, "group-delim-all-delim.md")
tab <- tt(data.frame(
  id = 1,
  A_a1 = 2,
  A_a2 = "3",
  B_b1 = 4,
  B_b2 = 5,
  C = 6
)) |>
  group_tt(j = "_")
tab@output <- "markdown"
expect_snapshot_print(tab, "group-delim-some-delim.md")
tab <- tt(data.frame(
  id = 1,
  Axa1 = 2,
  Axa2 = "3",
  Bxb1 = 4,
  Bxb2 = 5,
  C = 6
)) |>
  group_tt(j = "x")
tab@output <- "markdown"
expect_snapshot_print(tab, "group-delim-x-delim.md")



# bug: same number of delimiters
dat <- data.frame(
  "A__D" = rnorm(3),
  "A_B_D" = rnorm(3),
  "A_B_" = rnorm(3),
  "_C_E" = rnorm(3),
  check.names = FALSE
)
tt(dat) |> group_tt(j = "_")


# Issue #466: group_tt without delimiter returns a valid table
tab <- tt(data.frame(Aid = 1, Aa1 = 2, Aa2 = "3", Bb1 = 4, Bb2 = 5, BC = 5)) |>
  group_tt(j = "_")
expect_inherits(tab, "tinytable")


# Issue #471: `style_tt()` group
set.seed(48103)
dat <- data.frame(
  label = c("a", "a", "a", "b", "b", "c", "a", "a"),
  x1 = rnorm(8),
  x2 = rnorm(8)
)
tab <- tt(dat[, 2:3]) |>
  group_tt(i = dat$label) |>
  style_tt(i = "groupi", color = "white", background = "black", align = "c")
tab@output <- "latex"
expect_snapshot_print(tab, label = "group_style_tt_group.tex")


# Issue #165: Column labels not centered, last row not styled because of row group offset
k = mtcars[1:5, 1:5]
k[1, 1] <- pi
tab <- tt(k) |>
  format_tt(num_fmt = "decimal", num_zero = TRUE, digits = 3) |>
  style_tt(i = 1:3, j = 1, bold = TRUE) |>
  style_tt(j = 2, italic = TRUE) |>
  group_tt(i = list("hello world" = 2), j = list("foo" = 1:2, "bar" = 3:4))
expect_snapshot_print(
  print_html(tab),
  "group-issue165_html_centering_style.html"
)


# informative errors
expect_error(
  tt(head(iris)) |> style_tt("groupj", color = "orange") |> print(),
  pattern = "No column grouping"
)

# informative errors
x <- head(iris)
colnames(x) <- NULL
expect_error(
  tt(x) |> style_tt("colnames", color = "orange") |> print(),
  pattern = "No column names found"
)
expect_error(
  tt(head(iris)) |> style_tt("groupj", color = "orange") |> print(),
  pattern = "No column grouping"
)


# i = "colnames" vs. "groupj"
tab <- tt(head(iris)) |>
  group_tt(j = list("Foo" = 2:3, "Bar" = 5)) |>
  group_tt(j = list("Hello" = 1:2, "World" = 4:5)) |>
  style_tt("groupj", color = "orange") |>
  style_tt("colnames", color = "teal")
tab <- expect_table(tab, "latex")[["latex"]]
expect_snapshot_print(tab, "group-style_tt_i_colnames_vs_groupj.tex")


# Issue #165: group_tt insert extra row at the bottom
k <- mtcars[1:4, 1:4]
tab <- tt(k) |>
  format_tt(num_fmt = "decimal", num_zero = TRUE, digits = 3) |>
  style_tt(i = 1:3, j = 1, bold = TRUE) |>
  style_tt(j = 2, italic = TRUE) |>
  group_tt(
    i = list("hello world" = 2),
    j = list("foo" = 1:2, "bar" = 3:4)
  )
t <- expect_table(tab)
expect_snapshot_print(t[["latex"]], "group-issue165_extra_row.tex")
expect_snapshot_print(t[["typst"]], "group-issue165_extra_row.typ")
expect_snapshot_print(t[["html"]], "group-issue165_extra_row.html")
expect_snapshot_print(t[["markdown"]], "group-issue165_extra_row.md")


# Issue #518: Multi-level delimiter grouping
# Test basic multi-level structure
tab <- data.frame(
  hi_hi_hi = 1,
  hi_ho_hu = 2
) |>
  tt() |>
  group_tt(j = "_")
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "group-multilevel-basic.md")
expect_snapshot_print(t[["latex"]], "group-multilevel-basic.tex")
expect_snapshot_print(t[["html"]], "group-multilevel-basic.html")
expect_snapshot_print(t[["typst"]], "group-multilevel-basic.typ")

# Mixed-level delimiter structure should error
tab_mixed_levels <- data.frame(
  A_X_1 = 1,
  A_X_2 = 2,
  A_Y_1 = 3,
  B_Z_1 = 4,
  B_Z_2 = 5,
  C_W = 6
)
expect_error(
  tab_mixed_levels |> tt() |> group_tt(j = "_"),
  "Each column name must have the same number of delimiters.",
  fixed = TRUE
)

# Test multi-level structure with empty fields.
# The trailing space in "A_B_ " preserves the delimiter count.
tab <- data.frame(
  "A__C" = 1,
  "A_B_C" = 2,
  "A_B_ " = 3,
  "_D_E" = 4,
  check.names = FALSE
) |>
  tt() |>
  group_tt(j = "_")
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "group-multilevel-empty.md")
expect_snapshot_print(t[["latex"]], "group-multilevel-empty.tex")
expect_snapshot_print(t[["html"]], "group-multilevel-empty.html")
expect_snapshot_print(t[["typst"]], "group-multilevel-empty.typ")


# Issue #543: Multi-level delimiter with shared level names spanning all columns
dat <- data.frame(
  "Strata group\nSex\nStrata level\nMale" = NA,
  "Strata group\nSex\nStrata level\nFemale" = NA,
  "Strata group\nAge\nStrata level\n<18" = NA,
  "Strata group\nAge\nStrata level\n>=18" = NA,
  check.names = FALSE
)
tab <- tt(dat) |> group_tt(j = "\n")
t <- expect_table(tab, "markdown")
expect_snapshot_print(t[["markdown"]], "group-shared-level-spanning.md")

# Indent markdown
dat <- do.call(rbind, by(iris, iris$Species, head, 3))
tab <- tt(dat) |>
  group_tt(i = dat$Species) |>
  style_tt(i = "~groupi", j = 1, indent = 3)
expect_snapshot_print(tab, "group-indent.md")

# nse
t <- mtcars |>
  sort_by(~am) |>
  tt() |>
  group_tt(i = am)
t <- expect_table(t)
expect_snapshot_print(t[["markdown"]], "group-nse.md")
expect_snapshot_print(t[["latex"]], "group-nse.tex")
expect_snapshot_print(t[["html"]], "group-nse.html")
expect_snapshot_print(t[["typst"]], "group-nse.typ")

# Issue #638: group_tt(j) after subset() placed column spans at wrong positions
dat <- mtcars[1:5, ] |>
  setNames(c("A", "B", "C", "D", "E", "F", "G", "H", "Total", "X", "Y"))
tab <- dat |>
  tt() |>
  group_tt(i = list("grp" = 3)) |>
  subset(select = -A) |>
  group_tt(j = list("AAA" = 3:4, "BBB" = 5:6))
t <- expect_table(tab)
expect_snapshot_print(t[["markdown"]], "group-subset_j-638.md")
expect_snapshot_print(t[["latex"]], "group-subset_j-638.tex")
expect_snapshot_print(t[["html"]], "group-subset_j-638.html")
expect_snapshot_print(t[["typst"]], "group-subset_j-638.typ")


# Unsorted list positions place labels identically to sorted input
x1 <- tt(mtcars[1:8, 1:3]) |> group_tt(i = list("B" = 5, "A" = 2))
x2 <- tt(mtcars[1:8, 1:3]) |> group_tt(i = list("A" = 2, "B" = 5))
expect_equal(unname(x1@group_index_i), c(2, 6))
expect_identical(save_tt(x1, "markdown"), save_tt(x2, "markdown"))
md <- strsplit(save_tt(x1, "markdown"), "\n")[[1]]
expect_true(min(grep("^\\| A", md)) < min(grep("^\\| B", md)))

# Character vector `i` combined with `j` keeps both row groups and column span
dat <- data.frame(x1 = 1:4, x2 = 5:8)
tab <- tt(dat) |> group_tt(i = c("g1", "g1", "g2", "g2"), j = list("span" = 1:2))
expect_equal(nrow(tab@group_data_i), 2)
expect_equal(nrow(tab@group_data_j), 1)
md <- save_tt(tab, "markdown")
expect_true(grepl("g1", md) && grepl("g2", md) && grepl("span", md))

# Documented i = list("label" = 0) inserts the label in the first row (synonym for 1)
t0 <- tt(head(iris, 3)) |> group_tt(i = list("first" = 0))
t1 <- tt(head(iris, 3)) |> group_tt(i = list("first" = 1))
expect_equal(unname(t0@group_index_i), 1)
expect_identical(save_tt(t0, "markdown"), save_tt(t1, "markdown"))
expect_false(grepl("NA", save_tt(t0, "markdown")))

# Stacked pure-i group_tt() calls work and match the i+j path
t1 <- tt(mtcars[1:8, 1:3]) |>
  group_tt(i = list("A" = 2)) |>
  group_tt(i = list("B" = 5))
t2 <- tt(mtcars[1:8, 1:3]) |>
  group_tt(i = list("A" = 2), j = list("g" = 1:2)) |>
  group_tt(i = list("B" = 5), j = list("h" = 1:2))
expect_equal(unname(t1@group_index_i), c(2, 5))
expect_equal(unname(t1@group_index_i), unname(t2@group_index_i))
expect_identical(t1@group_data_i, t2@group_data_i)

# Bare numeric `i` without a matrix `j` errors informatively
expect_error(
  tt(head(iris)) |> group_tt(i = c(2, 4)),
  pattern = "named list"
)

# Length-1 character `i` creates a single group covering the whole table
tab <- tt(head(iris, 3)) |> group_tt(i = "only")
expect_equal(nrow(tab@group_data_i), 1)
expect_equal(unname(tab@group_index_i), 1)
expect_true(grepl("only", save_tt(tab, "markdown")))

# NA labels in vector `i` produce a clear error message
expect_error(
  tt(head(iris, 4)) |> group_tt(i = c("a", "a", NA, "b")),
  pattern = "missing values"
)

# Non-contiguous column spans warn instead of silently swallowing columns
expect_warning(
  tt(head(iris)) |> group_tt(j = list("A" = c(1, 3))),
  pattern = "not contiguous"
)

# Duplicate column-group labels (documented example in ?group_tt)
tab <- tt(mtcars[1:3, 1:6]) |>
  group_tt(j = list("Hello" = 1:2, "World" = 3:4, "Hello" = 5:6))
md <- save_tt(tab, "markdown")
expect_equal(unname(lengths(gregexpr("Hello", md, fixed = TRUE))), 2)
expect_true(grepl("World", md, fixed = TRUE))
htm <- strip_random(save_tt(tab, "html"))
expect_equal(unname(lengths(gregexpr("Hello", htm, fixed = TRUE))), 2)

source("helpers.R")
using("tinysnapshot")

tab <- tt(mtcars[1:4, 1:5])

options("tinytable_print_output" = "markdown")
void <- capture.output(x <- print(tab))
expect_equivalent(x@output, "markdown")

options("tinytable_print_output" = "latex")
void <- capture.output(x <- print(tab))
expect_equivalent(x@output, "latex")

options("tinytable_print_output" = "typst")
void <- capture.output(x <- print(tab))
expect_equivalent(x@output, "typst")

# avoid launching a viewer
# options("tinytable_print_output" = "html")
# void <- capture.output(x <- print(tab))
# expect_equivalent(x@output, "html")

options("tinytable_print_output" = NULL)

exit_file("skipping webshot2 screenshot tests: chromote fails to locate visible elements when rendering the styled tinytable (pre-existing flake, unrelated to scrutin)")

# print(output = "knitr"): returns knitr::include_graphics() with a valid PNG
if (is_local && requiet("webshot2") && requiet("knitr")) {
  tab_styled <- tt(mtcars[1:4, 1:5]) |> style_tt(background = "lightblue", i = 1:2)
  out <- print(tab_styled, output = "knitr")
  expect_inherits(out, "knit_asis")
  expect_true(file.exists(out[1]))
  expect_true(grepl("\\.png$", out[1]))
  expect_true(file.info(out[1])$size > 0)
}

# print(output = "raster"): draws on graphics device
if (is_local && requiet("webshot2") && requiet("png")) {
  tab_styled <- tt(mtcars[1:4, 1:5]) |> style_tt(background = "lightblue", i = 1:2)
  tmp_plot <- tempfile(fileext = ".png")
  grDevices::png(tmp_plot, width = 800, height = 400)
  out <- print(tab_styled, output = "raster")
  grDevices::dev.off()
  expect_true(file.exists(tmp_plot))
  expect_true(file.info(tmp_plot)$size > 0)
}

source("helpers.R")
using("tinysnapshot")

requiet("Rdatasets")

# bug discovered in modelsummary
z <- factor(c("a", "b", NA))
a <- format_tt(z, replace = "-")
b <- c("a", "b", "-")
expect_equivalent(a, b)

# numeric vector input
a <- c(98938272783457, 7288839482, 29111727, 93945)
b <- format_tt(a, num_suffix = TRUE, digits = 2)
expect_equivalent(b, c("99T", "7.3B", "29M", "94K"))

# num_suffix (vignette)
options(tinytable_print_output = "markdown")

dat <- data.frame(
  a = c("Burger", "Halloumi", "Tofu", "Beans"),
  b = c(1.43202, 201.399, 0.146188, 0.0031),
  c = c(98938272783457, 7288839482, 29111727, 93945)
)
tab <- tt(dat) |>
  format_tt(j = "a", sprintf = "Food: %s") |>
  format_tt(j = 2, digits = 1) |>
  format_tt(j = "c", digits = 2, num_suffix = TRUE)
expect_snapshot_print(tab, label = "format_tt-num_suffix_vignette")

set.seed(1024)
dat <- data.frame(
  w = c(143002.2092, 201399.181, 100188.3883),
  x = c(1.43402, 201.399, 0.134588),
  y = as.Date(sample(1:1000, 3), origin = "1970-01-01"),
  z = c(TRUE, TRUE, FALSE)
)
tab <- tt(dat, digits = 2)
expect_snapshot_print(tab, label = "format_tt-vignette_digits")

tab <- tt(dat) |>
  format_tt(
    j = 2:4,
    digits = 1,
    date = "%B %d %Y"
  ) |>
  format_tt(
    j = 1,
    digits = 2,
    num_mark_big = " ",
    num_mark_dec = ",",
    num_fmt = "decimal"
  )
expect_snapshot_print(tab, label = "format_tt-vignette_misc")

expect_snapshot_print(
  format_tt(dat, digits = 1, num_suffix = TRUE),
  label = "format_tt-dataframe"
)

# custom formatting
x <- mtcars[1:3, 1:3]
tab <- tt(x) |> format_tt(fn = function(x) paste("Ya", x))
expect_snapshot_print(tab, "format_tt-fn")

# Issue #142
k <- data.frame(x = c(0.000123456789, 12.4356789))
tab <- tt(k, digits = 2)
expect_snapshot_print(tab, "format_tt-issue142_01")
tab <- tt(k) |> format_tt(digits = 2, num_fmt = "significant_cell")
expect_snapshot_print(tab, "format_tt-issue142_02")

# Issue #147: format_tt(escape = TRUE) zaps previous formatting
options(tinytable_print_output = "latex")
x <- data.frame(num = c(pi, pi), char = c("10$", "blah_blah"))
tab <- tt(x) |>
  format_tt(i = 1, j = 1, digits = 2) |>
  format_tt(i = 1, j = 1, digits = 3) # overwrite
expect_snapshot_print(tab, "format_tt-issue147_01.tex")
tab <- tt(x) |>
  format_tt(i = 1, j = 1, digits = 2) |>
  format_tt(i = 2, j = 1, digits = 3) # different cell
expect_snapshot_print(tab, "format_tt-issue147_02.tex")
tab <- tt(x) |>
  format_tt(i = 1, j = 1, digits = 2) |>
  format_tt(i = 2, j = 1, digits = 3) |>
  format_tt(escape = TRUE) # do not zap
expect_snapshot_print(tab, "format_tt-issue147_03.tex")
options(tinytable_print_output = NULL)

# Issue #149: num_mark_big requires digits
x <- data.frame(x = c(pi * 1e6, pi * 1e9))
tab <- format_tt(tt(x), digits = 1, num_mark_big = " ")
expect_snapshot_print(tab, "format_tt-issue149")
# expect_error(format_tt(tt(x), num_mark_big = " "))

x <- data.frame(x = 1:5, y = c(pi, NA, NaN, -Inf, Inf))
dict <- list("-" = c(NA, NaN), "Tiny" = -Inf, "Huge" = Inf)
tab <- tt(x) |>
  format_tt(replace = dict) |>
  save_tt("dataframe")
expect_equivalent(tab[[2]], c("y", "3.141593", "-", "-", "Tiny", "Huge"))

# Issue #256: big mark for integers
tab <- data.frame(x = c(1332037, 1299128, 805058, 206840, 698511)) |>
  tt() |>
  format_tt(num_mark_big = " ", digits = 0, num_fmt = "decimal") |>
  save_tt("dataframe")
expect_equivalent(
  tab[[1]],
  c("x", "1 332 037", "1 299 128", "805 058", "206 840", "698 511")
)

x <- data.frame(x = pi, y = NA)
options(tinytable_tt_digits = 2)
tab <- tt(x) |> save_tt("dataframe")
expect_equivalent(tab[[2]], c("y", ""))

options(tinytable_format_replace = "zzz")
tab <- tt(x) |> save_tt("dataframe")
expect_equivalent(tab[[2]], c("y", "zzz"))

options(tinytable_format_replace = FALSE)
tab <- tt(x) |> save_tt("dataframe")
expect_equivalent(tab[[2]], c("y", "NA"))

options(tinytable_print_output = NULL)
options(tinytable_tt_digits = NULL)
options(tinytable_format_replace = NULL)

# Issue #263: NA processing with Quarto doesn't work
x <- data.frame(
  Points = c(40, 0, 10, 10, 10, 10, 10, 20, 110),
  Assignment = c(
    "Analytical Assignments",
    "Team Project Part 0 - County Selection",
    "Team Project Part 1 - Summary Background",
    "Team Project Part 2 - SWOT Analysis",
    "Team Project Part 3 - Strategic Direction/Action Plan",
    "Team Project Part 4 - Evaluation Framework",
    "Team Project Part 5 - Plan Presentation",
    "Class participation",
    "Total"
  ),
  Percent = c("36%", "0%", "9%", "9%", "9%", "9%", "9%", "18%", NA)
)
x <- tt(x) |>
  format_tt(replace = "!", quarto = TRUE) |>
  save_tt("latex")
expect_true(grepl("IQ==", x))

# Website scaling example
thumbdrives <- data.frame(
  date_lookup = as.Date(
    c("2024-01-15", "2024-01-18", "2024-01-14", "2024-01-16")
  ),
  price = c(18.49, 19.99, 24.99, 24.99),
  price_rank = c(1, 2, 3, 3),
  memory = c(16e9, 12e9, 10e9, 8e9),
  speed_benchmark = c(0.6, 0.73, 0.82, 0.99)
)
tab <- tt(thumbdrives) |>
  format_tt(j = 1, fn = scales::date_format("%B %d %Y")) |>
  format_tt(j = 2, fn = scales::label_currency()) |>
  format_tt(j = 3, fn = scales::label_ordinal()) |>
  format_tt(j = 4, fn = scales::label_bytes()) |>
  format_tt(j = 5, fn = scales::label_percent()) |>
  format_tt(escape = TRUE) |>
  save_tt("dataframe")
expect_true("January 15 2024" %in% tab[[1]])
expect_false("2024-01-15" %in% tab[[1]])
expect_true("$18.49" %in% tab[[2]])
expect_true("16 GB" %in% tab[[4]])
expect_true("99%" %in% tab[[5]])

# Issue #409: both NA and NaN should be replaced
options(tinytable_format_replace = NULL)
tab <- data.frame(x = c(1, NA, NaN, Inf))
tab0 <- tt(tab) |> save_tt("dataframe")
tab1 <- tt(tab) |>
  format_tt() |>
  save_tt("dataframe")
tab2 <- tt(tab) |>
  format_tt(replace = TRUE) |>
  save_tt("dataframe")
expect_equivalent(tab0[[1]], c("x", "1", "NA", "NaN", "Inf"))
expect_equivalent(tab1[[1]], c("x", "1", "NA", "NaN", "Inf"))
expect_equivalent(tab2[[1]], c("x", "1", "", "", "Inf"))

# bug: duplicated columns with markdown html
dat <- data.frame(
  markdown = c(
    "This is _italic_ text.",
    "This sentence ends with a superscript.^2^"
  )
)
tab <- tt(dat) |>
  format_tt(j = 1, markdown = TRUE) |>
  style_tt(j = 1, align = "c")
expect_snapshot_print(print_html(tab), "format_tt-vignette_html_markdown.html")

# boolean formatting
tab <- tt(data.frame(a = c(TRUE, FALSE, TRUE, NA)))
tab1 <- format_tt(tab, bool = tolower) |> save_tt("dataframe")
tab2 <- format_tt(tab, j = 1, bool = tolower) |> save_tt("dataframe")
expect_equivalent(tab1$a, tab2$a)


# Issue #392
pen <- rddata("penguins", "palmerpenguins")[1:3, c("species", "body_mass_g")]
expect_error(format_tt(tt(pen), num_mark_big = ","), "requires.*digits")
expect_error(format_tt(1234, num_mark_big = ","), "requires.*digits")


# Component-specific formatting Issue #418
dat <- data.frame(x = 1:2, y = 3:4)
tab <- tt(dat, caption = "Test Caption", notes = "Test note")

tab1 <- format_tt(tab, i = "colnames", fn = function(x) paste0("Col_", x))
expect_true(grepl("Col_x", save_tt(tab1, "markdown")))
expect_false(grepl("Col_2", save_tt(tab1, "markdown")))

tab1 <- format_tt(tab, i = "caption", fn = function(x) paste0("Col_", x))
expect_true(grepl("Col_Test", save_tt(tab1, "markdown")))
expect_false(grepl("Col_2", save_tt(tab1, "markdown")))

tab1 <- format_tt(tab, i = "notes", fn = function(x) paste0("Col_", x))
expect_true(grepl("| Col_Test", save_tt(tab1, "markdown"), fixed = TRUE))
expect_false(grepl("Col_2", save_tt(tab1, "markdown")))

# Test formatting only notes
tab3 <- format_tt(tab, i = "notes", fn = function(x) paste0("Note: ", x))
expect_true(grepl("| Note: |", save_tt(tab3, "markdown"), fixed = TRUE))
expect_true(grepl("| Test  |", save_tt(tab3, "markdown"), fixed = TRUE))

# Test multiple components at once
tab4 <- format_tt(tab, i = c("colnames", "caption"), fn = function(x) {
  paste0("Prefix_", x)
})
expect_true(grepl("Prefix_x", save_tt(tab4, "markdown"), fixed = TRUE))
expect_true(grepl(
  "Prefix_Test Caption",
  save_tt(tab4, "markdown"),
  fixed = TRUE
))


# groupi vs. ~groupi
tab = head(iris) |>
  tt() |>
  group_tt(i = list("Hello" = 1, "World" = 3)) |>
  format_tt("~groupi", j = 1, sprintf = "--%s--") |>
  save_tt("dataframe")
expect_true("Hello" %in% tab[[1]])
expect_true("--5.1--" %in% tab[[1]])
expect_false("--Hello--" %in% tab[[1]])
expect_false("5.1" %in% tab[[1]])

tab = head(iris) |>
  tt() |>
  group_tt(i = list("Hello" = 1, "World" = 3)) |>
  format_tt("groupi", j = 1, sprintf = "--%s--") |>
  save_tt("dataframe")
expect_true("--Hello--" %in% tab[[1]])
expect_true("5.1" %in% tab[[1]])
expect_false("Hello" %in% tab[[1]])
expect_false("--5.1--" %in% tab[[1]])


# Vignette with multiple components
options(tinytable_print_output = "latex")
set.seed(48103)
tab <- data.frame(
  "A_B" = rnorm(5),
  "B_C" = rnorm(5),
  "C_D" = rnorm(5)
)
tab <- tt(tab, digits = 2, notes = "_Source_: Simulated data.") |>
  group_tt(i = list("Down" = 1, "Up" = 3)) |>
  format_tt("colnames", fn = \(x) sub("_", " / ", x)) |>
  format_tt("notes", markdown = TRUE) |>
  format_tt("groupi", replace = list("↓" = "Down", "↑" = "Up"))

expect_snapshot_print(
  tab,
  label = "format_tt-format_components_vignette_01.tex"
)
options(tinytable_print_output = NULL)


# format group_tt(i)
dat <- data.frame(x = 1:4, y = letters[1:4])
tab <- tt(dat) |>
  group_tt(i = list("Group A" = 2, "Group B" = 4)) |>
  format_tt(i = c(2, 5), j = 1, sprintf = "_%s_") |>
  format_tt(i = 1:3, j = 1, sprintf = "*%s*")
expect_snapshot_print(tab, label = "format_tt-group_tt_01.md")


dat1 <- data.frame(
  values = c(123.456, 789.012, 345.678, 901.234),
  labels = c("A", "B", "C", "D")
)
tab1 <- tt(dat1) |>
  group_tt(i = list("First Group" = 2, "Second Group" = 4)) |>
  format_tt(j = 1, sprintf = "((%s))")
expect_snapshot_print(tab1, label = "format_tt-group_tt_02.md")


# Line breaks using linebreak argument
d <- data.frame(Text = "First line<br>Second line")
tab <- tt(d) |> format_tt(linebreak = "<br>")
t <- expect_table(tab, formats = c("html", "latex", "typst"))
expect_true(grepl("First line<br>Second line", t[["html"]]))
expect_true(grepl("First line\\\\Second line", t[["latex"]], fixed = TRUE))
expect_true(grepl("First line \\ Second line", t[["typst"]], fixed = TRUE))

# Date formatting tests
dates <- as.Date(c("2023-01-01", "2023-12-31"))
formatted_dates <- format_vector(dates, date = "%Y")
expect_equivalent(formatted_dates, c("2023", "2023"))
formatted_dates_full <- format_vector(dates, date = "%B %d, %Y")
expect_equivalent(formatted_dates_full, c("January 01, 2023", "December 31, 2023"))
formatted_vector <- format_vector(dates, date = "%Y")
expect_equivalent(formatted_vector, c("2023", "2023"))
d <- data.frame(dates = dates)
tab <- tt(d) |> format_tt(j = 1, date = "%Y")
formatted_table <- save_tt(tab, "dataframe")
expect_equivalent(formatted_table[[1]], c("dates", "2023", "2023"))


# Conditional formatting
dat <- data.frame(a = c("*Hello*", "**World**"))
cap <- "HTML render 1st row. LaTeX render 2nd row."
tab <- tt(dat, caption = cap) |>
  format_tt(i = 1, j = 1, markdown = TRUE, output = "html") |>
  format_tt(i = 2, j = 1, markdown = TRUE, output = "latex")
t <- expect_table(tab, formats = c("html", "latex"))
expect_snapshot_print(t[["html"]], "format_tt-conditional_output.html")
expect_snapshot_print(t[["latex"]], "format_tt-conditional_output.tex")


# num_suffix: NA must not render "NANA", Inf must not render "InfT", exact
# powers get the right tier (1e6 -> "1M", not "1000K"), and negatives are
# abbreviated via abs()
x <- format_tt(c(NA, -2e6, 1e3, 1500, 2e6, Inf, 0.5), digits = 2, num_suffix = TRUE)
expect_equivalent(x, c("NA", "-2M", "1K", "1.5K", "2M", "Inf", "0.5"))
x <- format_tt(c(1e3, 1e6, 1e9, 1e12), digits = 2, num_suffix = TRUE)
expect_equivalent(x, c("1K", "1M", "1B", "1T"))

# markdown/quarto flags must not crash on data frame or vector input
expect_equivalent(format_tt(c("_a_"), markdown = TRUE), "_a_")
expect_equivalent(format_tt(data.frame(x = "a"), quarto = TRUE)$x, "a")

# component-only i (e.g. "caption") must not reformat body cells
tab <- tt(data.frame(x = c(1.23456, 2.34567)), caption = "cap") |>
  format_tt(i = "caption", digits = 1) |>
  save_tt("dataframe")
expect_equivalent(tab[[1]], c("x", "1.23456", "2.34567"))

# replace matches typed original values only: a cell that merely formats to
# the same string (1.4 -> "1" with digits = 0) must not be clobbered
x <- format_tt(c(1.4, 1, 2.6), digits = 0, num_fmt = "decimal", replace = list("one" = 1))
expect_equivalent(x, c("1", "one", "3"))
tab <- tt(data.frame(x = c(1.4, 1, 2.6))) |>
  format_tt(digits = 0, num_fmt = "decimal", replace = list("one" = 1)) |>
  save_tt("dataframe")
expect_equivalent(tab[[1]], c("x", "1", "one", "3"))

# replace = TRUE blanks real NA values on both the vector and tinytable paths,
# but leaves literal "NA" strings alone
expect_equivalent(format_tt(c(NA, 1.5), digits = 1, replace = TRUE), c(" ", "2"))
tab <- tt(data.frame(x = c(NA, 1.5))) |>
  format_tt(digits = 1, replace = TRUE) |>
  save_tt("dataframe")
expect_equivalent(tab[[1]], c("x", "", "2"))
expect_equivalent(format_tt(c("NA", "FR"), replace = TRUE), c("NA", "FR"))

# decimal formatting must not emit signed zero
expect_equivalent(
  format_tt(c(-0.0001, 0.0001), digits = 2, num_fmt = "decimal"),
  c("0", "0")
)
expect_equivalent(
  format_tt(-0.0001, digits = 2, num_fmt = "decimal", num_zero = TRUE),
  "0.00"
)

source("helpers.R")
using("tinysnapshot")
requiet("data.table")

dat1 <- structure(list(EU = c("Austria", "Belgium", "Bulgaria", "Croatia", "Cyprus", "Czech Republic", "Denmark", "Estonia", "Finland", "France", "Germany", "Greece", "Hungary", "Ireland", "Italy", "Latvia", "Lithuania", "Luxembourg", "Malta", "Netherlands", "Poland", "Portugal", "Romania", "Slovakia", "Slovenia", "Spain", "Sweden")), class = "data.frame", row.names = c(NA, -27L)) # noqa
DT1 <- as.data.table(dat1)
colnames(DT1) <- NULL
tab <- tt(DT1)
expect_inherits(tab, "tinytable")


# Bug group_tt(i) accepts factor variables
dat <- do.call(rbind, by(iris, iris$Species, head, n = 3))
tab <- tt(dat) |>
  group_tt(dat$Species) |>
  save_tt("markdown")
tab <- unlist(strsplit(tab, "\\n"))
expect_equal(sum(grepl("^\\| virginica", tab)), 1)


# Bug: OutDec should not affect structural markup numbers
local({
  old <- options(OutDec = ",")
  on.exit(options(old), add = TRUE)

  has_bad_markup_decimal <- function(x) {
    grepl("[0-9],[0-9]+(em|%|deg)", x, perl = TRUE)
  }

  dat <- data.frame(a = 1.23, b = 4.56)
  tab <- tt(dat, height = 0.25, width = 0.333) |>
    format_tt(digits = 2, num_fmt = "decimal") |>
    style_tt(
      i = 1,
      j = 1,
      fontsize = 0.8,
      indent = 0.5,
      line = "tblr",
      line_width = 0.05
    ) |>
    theme_rotate(angle = 12.5, i = 1, j = 1)

  html <- save_tt(tab, "html")
  latex <- save_tt(tab, "latex")
  typst <- save_tt(tab, "typst")
  tabulator <- tt(dat, height = 0.25) |>
    format_tt(digits = 2, num_fmt = "decimal") |>
    theme_html(engine = "tabulator") |>
    save_tt("html")

  expect_false(has_bad_markup_decimal(html))
  expect_false(has_bad_markup_decimal(latex))
  expect_false(has_bad_markup_decimal(typst))
  expect_false(has_bad_markup_decimal(tabulator))

  expect_true(grepl("1,23", html, fixed = TRUE))
  expect_true(grepl("1,23", latex, fixed = TRUE))
  expect_true(grepl("1,23", typst, fixed = TRUE))
  expect_true(grepl("1,23", tabulator, fixed = TRUE))
})


# Regression test: typst_hlines/typst_vlines used to emit fragmented chains of
# table.hline()/table.vline() segments covering the same span whenever
# theme_tinytable() and the user declared the same line (or any line was
# declared twice with identical properties). This produced redundant, ugly
# output and slowed down rendering. The fix deduplicates cell-level entries
# inside the typst line processors.

# Case 1: user re-declares the theme's middle "below colnames" line.
# Before the fix this produced 6 fragmented hline segments at y=1.
tab <- tt(mtcars[1:5, 1:5]) |>
  style_tt(i = 0, line = "b", line_color = "black", line_width = 0.05)
out <- save_tt(tab, "typst")
hlines <- strsplit(out, "\n", fixed = TRUE)[[1]]
hlines <- hlines[grepl("table.hline", hlines, fixed = TRUE)]
middle <- hlines[grepl("y: 1,", hlines, fixed = TRUE)]
expect_true(length(middle) >= 1)
# The whole row should contain exactly one table.hline(y: 1, ...) segment
n_middle_segments <- lengths(regmatches(middle, gregexpr("table[.]hline", middle)))
expect_true(n_middle_segments == 1)

# Case 2: user adds the same line twice -> should still produce a single segment
tab <- tt(mtcars[1:5, 1:5]) |>
  style_tt(i = 2, line = "b", line_color = "pink") |>
  style_tt(i = 2, line = "b", line_color = "pink")
out <- save_tt(tab, "typst")
hlines <- strsplit(out, "\n", fixed = TRUE)[[1]]
hlines <- hlines[grepl("table.hline", hlines, fixed = TRUE)]
dup <- hlines[grepl("y: 3,", hlines, fixed = TRUE)]
expect_true(length(dup) >= 1)
n_dup_segments <- lengths(regmatches(dup, gregexpr("table[.]hline", dup)))
expect_true(n_dup_segments == 1)

source("helpers.R")
using("tinysnapshot")

dat <- head(iris[, 1:3], 3)

# Render to a single string so grepl() checks below are scalar
render <- function(x, format) {
  paste(build_tt(x, output = format)@table_string, collapse = "\n")
}

# Validation: only the three portable types are accepted
expect_error(
  render(style_tt(tt(dat), i = 1, line = "b", line_type = "dotdash"), "html"),
  pattern = "solid, dashed, dotted"
)

# Default is unchanged: no line_type means solid everywhere
tab <- tt(dat) |> style_tt(i = 1, line = "b")
expect_true(grepl("--line-type-bottom: solid", render(tab, "html"), fixed = TRUE))
expect_true(grepl("solid, black", render(tab, "latex"), fixed = TRUE))
expect_true(grepl("stroke: 0.1em)", render(tab, "typst"), fixed = TRUE))

# HTML: the type reaches the per-cell custom property
tab <- tt(dat) |> style_tt(i = 1, line = "b", line_type = "dashed")
expect_true(grepl("--line-type-bottom: dashed", render(tab, "html"), fixed = TRUE))

# LaTeX: tabularray hline spec
expect_true(grepl("hline{3}={1-3}{dashed, black", render(tab, "latex"), fixed = TRUE))

# Typst: dictionary stroke form, only used for non-solid rules
expect_true(grepl('dash: "dashed"', render(tab, "typst"), fixed = TRUE))

# Dotted
tab <- tt(dat) |> style_tt(i = 1, line = "b", line_type = "dotted")
expect_true(grepl("--line-type-bottom: dotted", render(tab, "html"), fixed = TRUE))
expect_true(grepl("{dotted, black", render(tab, "latex"), fixed = TRUE))
expect_true(grepl('dash: "dotted"', render(tab, "typst"), fixed = TRUE))

# Compound line specs apply the type to every requested side
tab <- tt(dat) |> style_tt(i = 2, j = 2, line = "tblr", line_type = "dashed")
h <- render(tab, "html")
for (side in c("top", "bottom", "left", "right")) {
  expect_true(grepl(sprintf("--line-type-%s: dashed", side), h, fixed = TRUE))
}

# Vertical rules in Typst also carry the dash
expect_true(grepl('table\\.vline\\(.*dash: "dashed"', render(tab, "typst")))

# Types are not merged across rules: distinct rules keep distinct types
tab <- tt(dat) |>
  style_tt(i = 1, line = "b", line_type = "dashed") |>
  style_tt(i = 2, line = "b", line_type = "dotted")
h <- render(tab, "html")
expect_true(grepl("--line-type-bottom: dashed", h, fixed = TRUE))
expect_true(grepl("--line-type-bottom: dotted", h, fixed = TRUE))

# HTML draws borders on pseudo-elements. The stylesheet is often loaded from a
# CDN, where an older copy always draws solid rules, so the type-aware rules
# travel with the table itself -- but only when a non-solid type is used.
expect_true(grepl("var(--line-type-bottom, solid)", h, fixed = TRUE))
expect_false(grepl(
  "var(--line-type-bottom, solid)",
  render(tt(dat) |> style_tt(i = 1, line = "b"), "html"),
  fixed = TRUE
))

# Markdown ignores line types rather than erroring
expect_inherits(render(tab, "markdown"), "character")

# Snapshots
tab <- tt(dat) |>
  style_tt(i = 1, line = "b", line_type = "dashed") |>
  style_tt(i = 2, line = "b", line_type = "dotted", line_color = "red") |>
  style_tt(i = 3, j = 2, line = "tblr", line_type = "dashed")
t <- expect_table(tab, c("html", "latex", "typst"))
expect_snapshot_print(t[["html"]], "line_type-mixed.html")
expect_snapshot_print(t[["latex"]], "line_type-mixed.tex")
expect_snapshot_print(t[["typst"]], "line_type-mixed.typ")

source("helpers.R")
using("tinysnapshot")
options(tinytable_print_output = "markdown")

# markdown without labels
k = iris[1:10, ]
colnames(k) <- NULL
tt(k)
expect_snapshot_print(
  tt(k),
  label = "markdown-nocolnames"
)

# no colnames
tab <- tt(head(iris))
colnames(tab) <- NULL
tab
expect_snapshot_print(tab, label = "markdown-nocolnames_02")

# group rows and columns
tab <- tt(mtcars[1:10, 1:5]) |>
  group_tt(
    i = list(
      "Hello" = 3,
      "World" = 8
    ),
    j = list(
      "Foo" = 2:3,
      "Bar" = 4:5
    )
  )
expect_snapshot_print(tab, label = "markdown-group_tt")

# simple caption
tab <- tt(mtcars[1:3, 1:3], caption = "Blah blah blah")
expect_snapshot_print(tab, label = "markdown-caption")

# Issue #105
x <- mtcars[1:15, ]
i <- list("blah" = 1, "blah" = 3, "blah" = 10)
tab <- tt(x) |> group_tt(i)
expect_snapshot_print(tab, label = "markdown-issue105")

# Issue 127: Large column
test <- data.frame(a = 1, b = 2, c = 3)
tab <- tt(test) |>
  group_tt(j = list("foobar" = 1:2, "hello world" = 3)) |>
  group_tt(j = list("foobar and hello world" = 2:3))
expect_snapshot_print(tab, label = "markdown-group_j_wider_1")

test <- data.frame(a = 1, b = 2, c = 3)
tab <- tt(test) |>
  group_tt(j = list("foobar" = 1:2, "hello world" = 3))
expect_snapshot_print(tab, label = "markdown-group_j_wider_2")

# Test bold columns
tab <- tt(mtcars[1:8, 1:5]) |>
  style_tt(j = c(2, 4), bold = TRUE)
expect_snapshot_print(tab, label = "markdown-bold_columns")

# Test replace missing value
tab <- mtcars[1:4, 1:3]
tab[2, 3] = NA
tab <- tt(tab) |>
  format_tt(replace = "-")
expect_snapshot_print(tab, label = "markdown-missing_value")

# Test footnote
x <- mtcars[1:3, 1:3]
n <- list("Blah blah", "*" = list(i = 0:1, j = 2, text = "foo bar"))
tab <- tt(x, notes = n)

expect_snapshot_print(tab, label = "markdown-footnote")

# GFM
if (!is_local) {
  exit_file("Run on Vincent's machine")
}
options(tinytable_print_output = "markdown")
expect_snapshot_print(tt(head(iris)) |> theme_markdown(style = "gfm"), label = "markdown-gfm")

# Long group labels - row groups
options(tinytable_print_output = "markdown")
x <- mtcars[1:5, 1:2]
tab <- tt(x) |> group_tt(i = list("AAAAAAAAAAAAAAAAAAAAAAAAAAAA" = 1))
expect_snapshot_print(tab, label = "markdown-long_row_group")

# Long group labels - column groups
x <- mtcars[1:3, 1:4]
tab <- tt(x) |>
  group_tt(j = list("Very Long Column Group Name Here" = 1:2, "Short" = 3:4))
expect_snapshot_print(tab, label = "markdown-long_column_group")

# Long group labels - mixed (both row and column)
x <- mtcars[1:3, 1:4]
tab <- tt(x) |>
  group_tt(
    i = list("SUPER LONG ROW GROUP NAME THAT SHOULD EXPAND TABLE" = 1)
  ) |>
  group_tt(
    j = list("Long Column Group" = 1:2, "Another Long Column Group" = 3:4)
  )
expect_snapshot_print(tab, label = "markdown-long_mixed_groups")

# Multiple long row groups
x <- mtcars[1:6, 1:3]
tab <- tt(x) |>
  group_tt(
    i = list(
      "First Very Long Row Group Name Here" = 1,
      "Second Even Longer Row Group Name That Should Also Expand" = 3,
      "Short" = 5
    )
  )
expect_snapshot_print(tab, label = "markdown-multiple_long_row_groups")

# colspan
x <- head(iris)[1:4, ]
tab <- tt(x) |>
  style_tt(1, 1, italic = TRUE) |>
  style_tt(2, 1, colspan = 3)
expect_snapshot_print(tab, label = "markdown-colspan_basic")

x <- head(iris)[1:4, ]
x[2, 1] <- "Very long content that spans multiple columns"
tab <- tt(x) |>
  style_tt(2, 1, colspan = 3)
expect_snapshot_print(tab, label = "markdown-colspan_long_content")

x <- head(iris)[1:5, ]
tab <- tt(x) |>
  style_tt(2, 1, colspan = 3) |>
  style_tt(4, 2, colspan = 2)
expect_snapshot_print(tab, label = "markdown-colspan_multiple")

# Test theme_empty() followed by theme_markdown() - should restore grid lines
options(tinytable_print_output = "markdown")
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_empty() |>
  theme_markdown(vline = TRUE, hline = TRUE)
expect_snapshot_print(tab, label = "markdown-theme_empty_restore_grid")

# Test theme_empty() alone - should have no grid lines
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_empty()
expect_snapshot_print(tab, label = "markdown-theme_empty_no_grid")

# Test theme_markdown grid options - hline only
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_markdown(hline = TRUE, vline = FALSE)
expect_snapshot_print(tab, label = "markdown-hline_only")

# Test theme_markdown grid options - vline only
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_markdown(hline = FALSE, vline = TRUE)
expect_snapshot_print(tab, label = "markdown-vline_only")

# Test theme_markdown with header line disabled
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_markdown(hline_header = FALSE)
expect_snapshot_print(tab, label = "markdown-no_header_line")

# Test theme_empty() then selective restoration
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_empty() |>
  theme_markdown(hline = TRUE, vline = FALSE)
expect_snapshot_print(tab, label = "markdown-theme_empty_hline_only")


# Issue #605
cap <- "Issue #605 lm1 and lm2 span 2 and p is in twice in colnames."
out <- data.frame(
  Parameter = c(
    "grp (2)", "grp (3)", "Days × grp (2)", "Days × grp (3)",
    "Days"
  ),
  `Coefficient (CI)` = c(
    "-4.31 (-15.95,  7.32)", "-1.31 (-13.47, 10.84)", "",
    "", "10.44 (  8.84, 12.03)"
  ),
  p = c(" 0.465", " 0.831", "", "", "<0.001"),
  `Coefficient (CI)` = c(
    " 0.32 (-22.56, 23.20)", " 3.77 (-19.72, 27.26)",
    "-1.01 ( -5.35,  3.32)", "-1.11 ( -5.53,  3.31)",
    "11.23 (  7.87, 14.60)"
  ),
  p = c(" 0.978", " 0.752", " 0.645", " 0.621", "<0.001"),
  check.names = FALSE
)
final <- tt(out, caption = cap)
final <- group_tt(
  final,
  i = list(Groups = 1L, Interactions = 3L, Controls = 5L),
  j = list(lm1 = 2:3, lm2 = 4:5)
)
final <- style_tt(final, i = "~groupi", j = 1, indent = 2)
expect_snapshot_print(tab, label = "markdown-issue605")



# restore
options(tinytable_print_output = NULL)

# colspan test from test-colrowspan.R
options(tinytable_print_output = "markdown")
tab <- aggregate(mpg ~ cyl + am, FUN = mean, data = mtcars)
tab <- tab[order(tab$cyl, tab$am), ]
tab <- tt(tab, digits = 2) |>
  style_tt(i = c(1, 3, 5), j = 1, colspan = 2)
expect_snapshot_print(tab, "colrowspan-markdown_multiple")
options(tinytable_print_output = NULL)

source("helpers.R")
using("tinysnapshot")

options(tinytable_print_output = "markdown")
data <- data.frame(
  Name = c("Alice", "Bob", "Charlie"),
  Age = c(25, 30, 35),
  Score = c(95.5, 87.2, 92.8)
)
tab <- tt(data, caption = "Three friends.") |>
  style_tt(i = c(0, 3), color = "orange") |>
  style_tt(i = 1, background = "teal", color = "black", bold = TRUE) |>
  style_tt(i = 2, j = 2, underline = TRUE, color = "red") |>
  style_tt(i = 3, strikeout = TRUE) |>
  group_tt(j = list("Characteristics" = 2:3)) |>
  style_tt(i = "caption", bold = TRUE, color = "red") |>
  style_tt(i = "notes", bold = TRUE, color = "red") |>
  theme_markdown(ansi = TRUE, hline = FALSE)
expect_snapshot_print(tab, label = "ansi-styles_01.md")

options(tinytable_print_output = "dataframe")
expect_snapshot_print(tab, label = "ansi-styles_01.txt")


# bugfix i = NULL applies to colnames
options(tinytable_print_output = "markdown")
tab <- tt(head(iris)) |>
  style_tt(color = "orange") |>
  theme_markdown(ansi = TRUE)
expect_snapshot_print(tab, label = "ansi-style_colnames_01.md")
tab <- tt(head(iris)) |>
  style_tt(i = 0:3, j = 1:2, color = "orange") |>
  theme_markdown(ansi = TRUE)
expect_snapshot_print(tab, label = "ansi-style_colnames_02.md")

# groupj
tab <- tt(head(iris)) |>
  group_tt(j = list("Group1" = 1:2, "Group2" = 3:5)) |>
  style_tt(i = -1, color = "orange") |>
  theme_markdown(ansi = TRUE)
expect_snapshot_print(tab, label = "ansi-style_groupj_01.md")
tab <- tt(head(iris)) |>
  group_tt(j = list("Group1" = 1:2, "Group2" = 3:5)) |>
  style_tt(i = "groupj", color = "orange") |>
  theme_markdown(ansi = TRUE)
expect_snapshot_print(tab, label = "ansi-style_groupj_01.md")


exit_file("broken")
data <- data.frame(
  Name = c("Alice", "Bob", "Charlie"),
  Age = c(25, 30, 35),
  Score = c(95.5, 87.2, 92.8)
)
tt(data, caption = "Three friends.", notes = "Three enemies") |>
  style_tt(i = c(0, 3), color = "orange") |>
  style_tt(i = 1, background = "teal", color = "black", bold = TRUE) |>
  style_tt(i = 2, j = 2, underline = TRUE, color = "red") |>
  style_tt(i = 3, strikeout = TRUE) |>
  group_tt(j = list("Characteristics" = 2:3)) |>
  style_tt(i = "caption", bold = TRUE, color = "red") |>
  style_tt(i = "notes", bold = TRUE, color = "red") |>
  theme_markdown(ansi = TRUE, hline = FALSE)

source("helpers.R")
requiet("modelsummary")

# modelsummary Issue #711: stars and d-columns
training <- 'https://vincentarelbundock.github.io/Rdatasets/csv/Ecdat/Treatment.csv'
training <- read.csv(training, na.strings = "")
training <- transform(
  training,
  `Earnings Before` = re75 / 1000,
  `Earnings After` = re78 / 1000,
  Treatment = ifelse(treat == TRUE, 'Treatment', 'Control'),
  Married = ifelse(married == TRUE, 'Yes', 'No')
)
cols <- c(
  "Earnings Before",
  "Earnings After",
  "Treatment",
  "ethn",
  "age",
  "educ",
  "Married"
)
names <- c(
  "Earnings Before",
  "Earnings After",
  "Treatment",
  "Ethnicity",
  "Age",
  "Education",
  "Married"
)
training <- training[, cols]
training <- setNames(training, names)
tab <- datasummary_balance(
  ~Treatment,
  training,
  fmt = 3,
  stars = TRUE,
  output = "data.frame"
)
p <- tt(tab) |> style_tt(align = "lllllldl") |> save_tt("latex")
expect_inherits(p, "character")

source("helpers.R")
using("tinysnapshot")

tab <- tt(mtcars[1:4, 1:5])

options("tinytable_print_output" = "markdown")
void <- capture.output(x <- print(tab))
expect_equivalent(x@output, "markdown")

options("tinytable_print_output" = "latex")
void <- capture.output(x <- print(tab))
expect_equivalent(x@output, "latex")

options("tinytable_print_output" = "typst")
void <- capture.output(x <- print(tab))
expect_equivalent(x@output, "typst")

# avoid launching a viewer
# options("tinytable_print_output" = "html")
# void <- capture.output(x <- print(tab))
# expect_equivalent(x@output, "html")

# =============================================================================
# TABULATOR STYLESHEET HANDLING
# =============================================================================

# CDN URLs and the default CSS block are defined once in R/tabulator_build.R

# Tabulator theme mapping
tabulator_themes <- list(
    tinytable = tinytable_tabulator_theme_cdn,
    tabulator = "tabulator.min.css",
    simple = "tabulator_simple.min.css",
    midnight = "tabulator_midnight.min.css",
    modern = "tabulator_modern.min.css",
    site = "tabulator_site.min.css",
    site_dark = "tabulator_site_dark.min.css",
    bootstrap3 = "tabulator_bootstrap.min.css",
    bootstrap4 = "tabulator_bootstrap4.min.css",
    bootstrap5 = "tabulator_bootstrap5.min.css",
    semanticui = "tabulator_semanticui.min.css",
    bulma = "tabulator_bulma.min.css",
    materialize = "tabulator_materialize.min.css"
)

# Default tabulator theme
tabulator_default_theme <- tabulator_themes[["tinytable"]]

#' Helper function to handle stylesheet theme selection
#' @param x tinytable object
#' @param stylesheet Theme name or custom URL
#' @return Modified tinytable object
#' @keywords internal
#' @noRd
tabulator_stylesheet <- function(x, stylesheet) {
    if (identical(stylesheet, "tinytable")) {
        stylesheet <- tabulator_themes[["tinytable"]]
    }

    # Check if it's a custom URL
    if (startsWith(stylesheet, "http")) {
        css_link <- sprintf(
            '<link href="%s" rel="stylesheet">\n    <link href="%s" rel="stylesheet">\n    <link href="%s" rel="stylesheet">',
            tabulator_css_cdn,
            stylesheet,
            fontawesome_css_cdn
        )
    } else {
        # Validate theme choice
        valid_themes <- names(tabulator_themes)
        valid_themes <- c("tinytable", sort(setdiff(valid_themes, "tinytable")))

        if (!stylesheet %in% valid_themes) {
            warning(
                "Invalid theme '",
                stylesheet,
                "'. Valid themes are: ",
                paste(valid_themes, collapse = ", "),
                ". Or provide a custom CDN URL starting with 'http'. Using the default `tinytable` theme.",
                call. = FALSE
            )
            css_link <- tabulator_default_css_block
        } else {
            css_file <- tabulator_themes[[stylesheet]]
            css_link <- sprintf(
                '<link href="%s" rel="stylesheet">\n    <link href="%s%s" rel="stylesheet">\n    <link href="%s" rel="stylesheet">',
                tabulator_css_cdn,
                tabulator_cdn_base,
                css_file,
                fontawesome_css_cdn
            )
        }
    }

    # Replace the CSS links in the table string (both base Tabulator CSS and tinytable theme)
    x@table_string <- gsub(
        tabulator_default_css_block,
        css_link,
        x@table_string,
        fixed = TRUE
    )

    return(x)
}

style_string_html <- function(n, styles) {
  # Collect CSS properties that can be combined into a single span
  css_props <- c()

  if (isTRUE(styles[["smallcap"]])) {
    css_props <- c(css_props, "font-variant: small-caps")
  }
  if (!is.null(styles[["color"]])) {
    css_props <- c(css_props, paste0("color:", styles[["color"]]))
  }
  if (!is.null(styles[["fontsize"]])) {
    css_props <- c(css_props, paste0("font-size:", format_markup_unit(styles[["fontsize"]], "em")))
  }

  # Apply CSS properties as a single span if any exist
  if (length(css_props) > 0) {
    css_style <- paste(css_props, collapse = "; ")
    n <- sprintf("<span style='%s'>%s</span>", css_style, n)
  }

  # Apply HTML tags that cannot be combined with CSS
  if (isTRUE(styles[["italic"]])) {
    n <- sprintf("<i>%s</i>", n)
  }
  if (isTRUE(styles[["strikeout"]])) {
    n <- sprintf("<s>%s</s>", n)
  }
  if (isTRUE(styles[["underline"]])) {
    n <- sprintf("<u>%s</u>", n)
  }
  if (isTRUE(styles[["bold"]])) {
    n <- sprintf("<b>%s</b>", n)
  }
  if (isTRUE(styles[["monospace"]])) {
    n <- sprintf("<code>%s</code>", n)
  }
  n
}

style_string_latex <- function(n, styles) {
  if (isTRUE(styles[["italic"]])) {
    n <- sprintf("\\textit{%s}", n)
  }
  if (isTRUE(styles[["strikeout"]])) {
    n <- sprintf("\\sout{%s}", n)
  }
  if (isTRUE(styles[["underline"]])) {
    n <- sprintf("\\underline{%s}", n)
  }
  if (isTRUE(styles[["bold"]])) {
    n <- sprintf("\\textbf{%s}", n)
  }
  if (isTRUE(styles[["monospace"]])) {
    n <- sprintf("\\texttt{%s}", n)
  }
  if (isTRUE(styles[["smallcap"]])) {
    n <- sprintf("\\textsc{%s}", n)
  }
  if (!is.null(styles[["color"]])) {
    n <- sprintf("\\textcolor{%s}{%s}", styles[["color"]], n)
  }
  if (!is.null(styles[["fontsize"]])) {
    fs <- format_markup_num(styles[["fontsize"]])
    n <- sprintf(
      "{\\fontsize{%sem}{%sem}\\selectfont %s}",
      fs,
      fs,
      n
    )
  }
  n
}

style_string_typst <- function(n, styles) {
  sty <- NULL
  if (isTRUE(styles[["italic"]])) {
    sty <- c(sty, 'style: "italic"')
  }
  if (isTRUE(styles[["bold"]])) {
    sty <- c(sty, 'weight: "bold"')
  }
  if (isTRUE(styles[["strikeout"]])) {
    # not sure how to do this
  }
  if (isTRUE(styles[["underline"]])) {
    # not sure how to do this
  }
  if (!is.null(styles[["fontsize"]])) {
    fs <- sprintf("size: %s", format_markup_unit(styles[["fontsize"]], "em"))
    sty <- c(sty, fs)
  }
  if (!is.null(styles[["color"]])) {
    col <- styles[["color"]]
    if (grepl("^#", col)) {
      col <- sprintf('rgb("%s")', col)
    }
    col <- sprintf("fill: %s", col)
    sty <- c(sty, col)
  }
  if (isTRUE(styles[["smallcap"]])) {
    if (length(sty) > 0) {
      template <- paste0("smallcaps(text(", paste(sty, collapse = ", "), ", [%s]))")
    } else {
      template <- "smallcaps([%s])"
    }
    out <- sprintf(template, n)
    return(out)
  }
  template <- paste0("text(", paste(sty, collapse = ", "), ", [%s])")
  out <- sprintf(template, n)
  out <- sub("text(, ", "text(", out, fixed = TRUE)
  return(out)
}

style_string_markdown <- function(n, styles) {
  out <- n
  if (isTRUE(styles[["bold"]])) {
    out <- sprintf("**%s**", out)
  }
  if (isTRUE(styles[["italic"]])) {
    out <- sprintf("_%s_", out)
  }
  if (isTRUE(styles[["underline"]])) {
    out <- sprintf("<u>%s</u>", out)
  }
  if (isTRUE(styles[["strikeout"]])) {
    out <- sprintf("~~%s~~", out)
  }
  if (isTRUE(styles[["smallcap"]])) {
    out <- toupper(out)
  }
  if (!is.null(styles[["indent"]]) && !is.na(styles[["indent"]])) {
    out <- sprintf("%s%s", strrep(" ", styles[["indent"]]), out)
  }
  return(out)
}

style_string_ansi <- function(n, styles) {
  out <- n
  if (isTRUE(styles[["bold"]])) {
    out <- sprintf("\033[1m%s\033[22m", out)
  }
  if (isTRUE(styles[["italic"]])) {
    out <- sprintf("\033[3m%s\033[23m", out)
  }
  if (isTRUE(styles[["underline"]])) {
    out <- sprintf("\033[4m%s\033[24m", out)
  }
  if (isTRUE(styles[["strikeout"]])) {
    out <- sprintf("\033[9m%s\033[29m", out)
  }
  if (isTRUE(styles[["smallcap"]])) {
    out <- toupper(out)
  }
  if (!is.null(styles[["color"]]) && !is.na(styles[["color"]])) {
    ansi_color_code <- standardize_colors(styles[["color"]], format = "ansi")
    if (!is.na(ansi_color_code) && ansi_color_code != styles[["color"]]) {
      out <- sprintf("\033[%sm%s\033[39m", ansi_color_code, out)
    }
  }
  if (!is.null(styles[["background"]]) && !is.na(styles[["background"]])) {
    ansi_bg_code <- standardize_colors(styles[["background"]], format = "ansi")
    if (!is.na(ansi_bg_code) && ansi_bg_code != styles[["background"]]) {
      # Convert foreground ANSI code to background by replacing 38 with 48
      if (grepl("^38;2;", ansi_bg_code)) {
        ansi_bg_code <- sub("^38;2;", "48;2;", ansi_bg_code)
      } else if (grepl("^3[0-7]$", ansi_bg_code)) {
        # Convert standard foreground colors (30-37) to background (40-47)
        fg_code <- as.numeric(ansi_bg_code)
        ansi_bg_code <- as.character(fg_code + 10)
      } else if (grepl("^9[0-7]$", ansi_bg_code)) {
        # Convert bright foreground colors (90-97) to bright background (100-107)
        fg_code <- as.numeric(ansi_bg_code)
        ansi_bg_code <- as.character(fg_code + 10)
      }
      out <- sprintf("\033[%sm%s\033[49m", ansi_bg_code, out)
    }
  }
  if (!is.null(styles[["indent"]]) && !is.na(styles[["indent"]])) {
    out <- sprintf("%s%s", strrep(" ", styles[["indent"]]), out)
  }
  return(out)
}

style_notes <- function(x) {
  fun <- switch(x@output,
    "typst" = style_string_typst,
    "html" = style_string_html,
    "bootstrap" = style_string_html,
    "tabulator" = style_string_html,
    "latex" = style_string_latex,
    "tabularray" = style_string_latex,
    "markdown" = style_string_markdown,
    function(k, ...) identity(k)
  )
  # Check if we should use ANSI styling for grid output
  if (x@output == "markdown" && isTRUE(x@ansi)) {
    fun <- style_string_ansi
  }
  for (i in seq_along(x@notes)) {
    if (length(x@notes[[i]]) == 3 && "text" %in% names(x@notes[[i]])) {
      x@notes[[i]][["text"]] <- fun(x@notes[[i]][["text"]], x@style_notes)
    } else {
      x@notes[[i]] <- fun(x@notes[[i]], x@style_notes)
    }
  }
  return(x)
}

style_caption <- function(x) {
  fun <- switch(x@output,
    "typst" = style_string_typst,
    "html" = style_string_html,
    "bootstrap" = style_string_html,
    "tabulator" = style_string_html,
    "latex" = style_string_latex,
    "tabularray" = style_string_latex,
    "markdown" = style_string_markdown,
    function(k, ...) identity(k)
  )
  # Check if we should use ANSI styling for grid output
  if (x@output == "markdown" && isTRUE(x@ansi)) {
    fun <- style_string_ansi
  }
  if (length(x@caption) > 0) {
    x@caption <- fun(x@caption, x@style_caption)
  }
  return(x)
}

#' Clean Typst style string
#' @keywords internal
#' @noRd
typst_clean_css <- function(css) {
  css <- gsub(" +", " ", trimws(css))
  css <- sub("^,", "", trimws(css))
  css <- gsub(",+", ",", trimws(css))
  return(css)
}

#' Vectorized CSS builder for the typst style-array
#'
#' Replaces the per-cell for-loop that called typst_insert_field() (which does
#' regex sub() on a growing string per property) with a single pass that builds
#' each cell's "key: value, key: value, ..." string directly. Output is
#' byte-identical to the concatenation of sequential typst_insert_field()
#' calls followed by typst_clean_css().
#'
#' @param other Filtered @style_other data frame (only rows with at least one style)
#' @param color_map Named character vector mapping original color names to typst rgb() strings
#' @return Character vector of css strings (one per row of `other`), un-trimmed
#' @keywords internal
#' @noRd
.typst_build_css_vectorized <- function(other, color_map) {
  n <- nrow(other)
  if (n == 0) return(character(0))

  bold        <- !is.na(other$bold)       & other$bold
  italic      <- !is.na(other$italic)     & other$italic
  underline   <- !is.na(other$underline)  & other$underline
  strikeout   <- !is.na(other$strikeout)  & other$strikeout
  monospace   <- !is.na(other$monospace)  & other$monospace
  smallcap    <- !is.na(other$smallcap)   & other$smallcap
  color_set   <- !is.na(other$color)
  bg_set      <- !is.na(other$background)
  fs_set      <- !is.na(other$fontsize)
  indent_set  <- !is.na(other$indent)     & other$indent > 0
  align_h_set <- !is.na(other$align)
  align_v_set <- !is.na(other$alignv)

  css <- character(n)

  # scalar emission: same string for every matching cell
  add_scalar <- function(mask, key, val) {
    if (!any(mask)) return()
    idx <- which(mask)
    fragment <- sprintf("%s: %s", key, val)
    sep <- ifelse(nzchar(css[idx]), ", ", "")
    css[idx] <<- paste0(css[idx], sep, fragment)
  }
  # vector emission: val is parallel to `other`; only used where mask is TRUE
  add_vec <- function(mask, key, val) {
    if (!any(mask)) return()
    idx <- which(mask)
    fragment <- sprintf("%s: %s", key, val[idx])
    sep <- ifelse(nzchar(css[idx]), ", ", "")
    css[idx] <<- paste0(css[idx], sep, fragment)
  }

  add_scalar(bold, "bold", "true")
  add_scalar(italic, "italic", "true")
  add_scalar(underline, "underline", "true")
  add_scalar(strikeout, "strikeout", "true")
  add_scalar(monospace, "mono", "true")
  add_scalar(smallcap, "smallcaps", "true")

  if (any(color_set)) {
    cv <- character(n)
    cv[color_set] <- color_map[other$color[color_set]]
    add_vec(color_set, "color", cv)
  }
  if (any(bg_set)) {
    bv <- character(n)
    bv[bg_set] <- color_map[other$background[bg_set]]
    add_vec(bg_set, "background", bv)
  }
  if (any(fs_set)) {
    fv <- character(n)
    fv[fs_set] <- vapply(
      other$fontsize[fs_set],
      format_markup_unit,
      character(1),
      "em"
    )
    add_vec(fs_set, "fontsize", fv)
  }

  if (any(align_h_set) || any(align_v_set)) {
    ah <- ifelse(is.na(other$align),  "", other$align)
    av <- ifelse(is.na(other$alignv), "", other$alignv)
    combined <- ifelse(nchar(ah) > 0 & nchar(av) > 0,
                       paste(ah, av, sep = " + "),
                       paste0(ah, av))
    mask <- align_h_set | align_v_set
    add_vec(mask, "align", combined)
  }

  if (any(indent_set)) {
    iv <- character(n)
    iv[indent_set] <- vapply(
      other$indent[indent_set],
      format_markup_unit,
      character(1),
      "em"
    )
    add_vec(indent_set, "indent", iv)
  }

  # Each cell becomes a comma-separated list with a trailing comma, matching
  # the original typst_insert_field() + typst_clean_css() output format.
  ifelse(nzchar(css), paste0(css, ","), css)
}

#' Apply Typst styles and generate style-dict/array
#' @keywords internal
#' @noRd
typst_apply_styles <- function(x, rec) {
  # Filter out empty styles
  idx <- rec$css != ""
  rec <- rec[idx, , drop = FALSE]
  if (nrow(rec) == 0) {
    return(x)
  }

  # Get unique styles. split(rec, rec$css) sorted by CSS value; preserve that
  # order without copying one data frame per style group.
  style_keys <- sort(unique(rec$css))
  style_idx <- match(rec$css, style_keys)

  # Create style-array (unique styles)
  style_array_entries <- sprintf("(%s),", style_keys)

  # Create style-dict (cell positions -> style array indices) using vectorized
  # sprintf instead of growing a vector inside a for loop. Each unique style
  # group can produce many cell-position keys, so we pre-allocate a list and
  # unlist() once.
  style_dict_list <- vector("list", length(style_keys))
  for (idx in seq_along(style_keys)) {
    rows <- which(style_idx == idx)
    style_dict_list[[idx]] <- sprintf('"%s_%s": %s', rec$i[rows], rec$j[rows], idx - 1)
  }
  style_dict_entries <- unlist(style_dict_list, use.names = FALSE)

  # Insert style-dict entries as single line
  if (length(style_dict_entries) > 0) {
    combined_dict <- paste0(
      "    ",
      paste(style_dict_entries, collapse = ", ")
    )
    x@table_string <- lines_insert(
      x@table_string,
      combined_dict,
      "tinytable style-dict after",
      "after"
    )
  }

  # Insert all style-array entries in one pass over the table string. The old
  # loop inserted rev(style_array_entries) after the same marker, which
  # produced style_array_entries order in the final output.
  if (length(style_array_entries) > 0) {
    x@table_string <- lines_insert(
      x@table_string,
      paste0("    ", style_array_entries, collapse = "\n"),
      "tinytable cell style after",
      "after"
    )
  }

  return(x)
}




#' Convert tinytable indexing to Typst 0-based indexing
#'
#' tinytable: 0=colnames, -1=group1, -2=group2, ..., 1,2,3=data
#' Typst: group headers come first (0,1,...), then colnames, then data
#' Need to account for the fact that group headers are inserted at the top
#' @param df Data frame with `i` (row) and `j` (column) tinytable indices
#' @param nhead Number of header rows (colnames + column group rows)
#' @keywords internal
#' @noRd
typst_shift_indices <- function(df, nhead) {
  if (nhead > 0) {
    # Case with headers/colnames: normal conversion
    df$i <- ifelse(df$i < 0,
                   nhead + df$i - 1,  # Headers: -1 becomes nhead-2, -2 becomes nhead-3
                   df$i + nhead - 1)  # Column names (0) and data: 0 becomes nhead-1, 1 becomes nhead
  } else {
    # Case with no headers (nhead = 0): filter out i=0 (non-existent colnames), then convert
    df <- df[df$i > 0, , drop = FALSE]  # Remove i=0 entries
    df$i <- df$i - 1  # Data rows: 1 becomes 0, 2 becomes 1, etc.
  }
  df$j <- df$j - 1
  return(df)
}

typst_split_chunks <- function(x) {
  x <- sort(x)
  breaks <- c(0, which(diff(x) != 1), length(x))
  result <- list()
  for (i in seq_along(breaks)[-length(breaks)]) {
    chunk <- x[(breaks[i] + 1):breaks[i + 1]]
    result[[i]] <- c(min = min(chunk), max = max(chunk))
  }
  out <- data.frame(do.call(rbind, result))
  out$max <- out$max + 1
  return(out)
}

#' Build a Typst stroke specification for a rule
#'
#' Solid rules keep the compact `<width> + <color>` form so that existing
#' output is unchanged. Dashed and dotted rules require the dictionary form,
#' which is the only way to pass `dash:`.
#' @keywords internal
#' @noRd
typst_line_type <- function(lin) {
  out <- if ("line_type" %in% names(lin)) lin$line_type else NA_character_
  ifelse(is.na(out), "solid", out)
}

# A `color` of NA means the user did not request a specific paint. We then omit
# it from the stroke entirely: Typst stroke folding fills the missing field from
# the surrounding `#set` rules, so the document's global styling wins instead of
# being overridden by a hard-coded black. See Issue #679.
typst_stroke <- function(width, color, type) {
  if (is.na(type) || identical(type, "solid")) {
    if (is.na(color)) width else sprintf("%s + %s", width, color)
  } else if (is.na(color)) {
    sprintf('(thickness: %s, dash: "%s")', width, type)
  } else {
    sprintf('(thickness: %s, paint: %s, dash: "%s")', width, color, type)
  }
}

# split() drops NA levels, which would silently discard every line whose paint
# is left to the document. Map NA to a sentinel before using it as a split key.
typst_color_key <- function(color) {
  ifelse(is.na(color), "__inherit__", color)
}

# When the very same rule is declared twice -- once with an inherited paint (by
# theme_tinytable() or group_tt()) and once with an explicit color (by the user)
# -- the explicit declaration wins. Without this, the two entries land in
# different split() groups and two overlapping table.hline() calls are emitted
# for one visual rule. See test-bugfix.R.
typst_drop_inherited_dupes <- function(lin) {
  inherited <- is.na(lin$line_color_mapped)
  if (!any(inherited) || all(inherited)) {
    return(lin)
  }
  key_cols <- intersect(
    c("i", "j", "line", "line_width", "line_type", "line_trim"),
    colnames(lin)
  )
  key <- do.call(paste, c(lin[key_cols], sep = "\r"))
  lin[!(inherited & key %in% key[!inherited]), , drop = FALSE]
}

typst_hlines <- function(x, lin) {
  if (nrow(lin) == 0) {
    return(x)
  }

  # Normalize colors once before splitting
  lin$line_color_mapped <- normalize_colors(lin$line_color, "typst", default = NA_character_)

  lin$line_type <- typst_line_type(lin)

  lin <- typst_drop_inherited_dupes(lin)

  tmp <- split(
    lin,
    list(lin$i, lin$line, typst_color_key(lin$line_color_mapped), lin$line_width, lin$line_type),
    drop = TRUE
  )
  tmp <- lapply(tmp, function(k) {
    # Drop duplicate cell-level entries. These arise routinely whenever
    # theme_tinytable() and a user style_tt() declare the same line (same
    # i, j, line, color, width, trim). Without deduplication,
    # typst_split_chunks() sees phantom breaks in diff(x) != 1 and emits a
    # chain of fragmented table.hline()/table.vline() segments that cover
    # the same span as a single line. See test-typst-line-dedup.R.
    k <- k[!duplicated(k), , drop = FALSE]

    # Split chunks based on consecutive columns AND line_trim boundaries
    # line_trim marks group boundaries - split chunks at these points
    chunks <- typst_split_chunks(k$j)

    # Further split chunks based on line_trim markers
    # line_trim="l" or "lr" means start of a new group
    # line_trim="r" or "lr" means end of current group
    final_chunks <- list()
    for (chunk_idx in seq_len(nrow(chunks))) {
      chunk_start <- chunks$min[chunk_idx]
      chunk_end <- chunks$max[chunk_idx]
      chunk_cols <- chunk_start:(chunk_end - 1)  # -1 because max is exclusive

      # Find trim boundaries within this chunk
      chunk_rows <- k[k$j %in% chunk_cols, , drop = FALSE]
      chunk_rows <- chunk_rows[order(chunk_rows$j), ]

      # Split on line_trim boundaries
      current_start <- chunk_start
      for (i in seq_len(nrow(chunk_rows))) {
        trim_val <- chunk_rows$line_trim[i]
        col_val <- chunk_rows$j[i]

        # Check if this is a boundary
        if (!is.na(trim_val)) {
          if (grepl("r", trim_val)) {
            # End of current segment; the next one starts after this column
            final_chunks <- c(final_chunks, list(c(min = current_start, max = col_val + 1)))
            current_start <- col_val + 1
          } else if (grepl("l", trim_val) && col_val > current_start) {
            # Start of new segment (close previous if any)
            final_chunks <- c(final_chunks, list(c(min = current_start, max = col_val)))
            current_start <- col_val
          }
        }

        # If last column in chunk, close the segment
        if (i == nrow(chunk_rows) && col_val + 1 >= chunk_end) {
          if (current_start < chunk_end) {
            final_chunks <- c(final_chunks, list(c(min = current_start, max = chunk_end)))
          }
        }
      }
    }

    # If no trim markers found, use original chunks
    if (length(final_chunks) == 0) {
      final_chunks <- lapply(seq_len(nrow(chunks)), function(i) c(min = chunks$min[i], max = chunks$max[i]))
    }

    # Drop zero-width segments and duplicates so each span is drawn once
    final_chunks <- Filter(function(chunk) chunk["min"] < chunk["max"], final_chunks)
    final_chunks <- unique(final_chunks)

    ymin <- k$i[1]
    ymax <- k$i[1] + 1
    line <- k$line[1]
    color <- k$line_color_mapped[1]
    width <- if (is.na(k$line_width[1])) 0.1 else k$line_width[1]
    width <- format_markup_unit(width, "em")
    stroke <- typst_stroke(width, color, k$line_type[1])
    out <- ""

    # Generate hlines for each chunk
    for (chunk in final_chunks) {
      if (grepl("t", line)) {
        tmp <- "table.hline(y: %s, start: %s, end: %s, stroke: %s),"
        tmp <- sprintf(tmp, ymin, chunk["min"], chunk["max"], stroke)
        out <- paste(out, tmp)
      }
      if (grepl("b", line)) {
        tmp <- "table.hline(y: %s, start: %s, end: %s, stroke: %s),"
        tmp <- sprintf(tmp, ymax, chunk["min"], chunk["max"], stroke)
        out <- paste(out, tmp)
      }
    }
    return(out)
  })
  if (length(tmp) > 0) {
    x@table_string <- lines_insert(
      x@table_string,
      paste(unlist(tmp, use.names = FALSE), collapse = "\n"),
      "tinytable lines before",
      "before"
    )
  }
  return(x)
}

# =============================================================================
# NEW SIMPLIFIED FUNCTIONS USING expand_style()
# =============================================================================

#' Process Typst line styles using expanded data
#' @keywords internal
#' @noRd
process_typst_lines <- function(x, lines) {
  if (is.null(lines) || nrow(lines) == 0) {
    return(x)
  }

  lines <- typst_shift_indices(lines, x@nhead)

  # Process horizontal and vertical lines
  x <- typst_hlines(x, lines)
  x <- typst_vlines(x, lines)

  return(x)
}

#' Process Typst other styles using expanded data
#' @keywords internal
#' @noRd
process_typst_other_styles <- function(x, other) {
  if (is.null(other) || nrow(other) == 0) {
    return(x)
  }

  # Map alignments to Typst format
  other$align <- map_align(other$align, "typst")
  other$alignv <- map_alignv(other$alignv, "typst")

  # Normalize colors once
  color_map <- build_color_map(c(other$color, other$background), "typst")

  # Generate CSS for each cell. The reference implementation builds one cell
  # at a time by repeatedly calling typst_insert_field(), which uses regex
  # sub() on a growing string. The vectorized version below produces an
  # identical result but is much faster on tables with many styled cells
  # (heat-map backgrounds, per-cell alignment, etc.). The output strings are
  # joined with ", " and given a trailing comma to match the original
  # typst_insert_field() format exactly (typst_clean_css() collapses runs
  # of commas and trims whitespace, so byte-identical output is achievable).
  css <- .typst_build_css_vectorized(other, color_map)

  # Clean CSS and add to data frame
  other$css <- typst_clean_css(css)

  other <- typst_shift_indices(other, x@nhead)

  # Generate style-dict and style-array for optimized lookup
  x <- typst_apply_styles(x, other)

  return(x)
}

typst_vlines <- function(x, lin) {
  lin <- lin[grepl("l|r", lin$line), , drop = FALSE]
  if (nrow(lin) == 0) {
    return(x)
  }

  # Normalize colors once before splitting
  lin$line_color_mapped <- normalize_colors(lin$line_color, "typst", default = NA_character_)

  lin$line_type <- typst_line_type(lin)

  lin <- typst_drop_inherited_dupes(lin)

  lin <- split(
    lin,
    list(lin$j, lin$line, typst_color_key(lin$line_color_mapped), lin$line_width, lin$line_type),
    drop = TRUE
  )
  lin <- lapply(lin, function(k) {
    # Drop duplicate cell-level entries; see typst_hlines() for rationale.
    k <- k[!duplicated(k), , drop = FALSE]
    chunks <- typst_split_chunks(k$i)
    ymin <- chunks$min
    ymax <- chunks$max
    xmin <- k$j[1]
    xmax <- xmin + 1
    line <- k$line[1]
    color <- k$line_color_mapped[1]
    width <- if (is.na(k$line_width[1])) 0.1 else k$line_width[1]
    width <- format_markup_unit(width, "em")
    stroke <- typst_stroke(width, color, k$line_type[1])
    out <- ""
    if (grepl("l", line)) {
      tmp <- "table.vline(x: %s, start: %s, end: %s, stroke: %s),"
      tmp <- sprintf(tmp, xmin, ymin, ymax, stroke)
      out <- paste(out, tmp)
    }
    if (grepl("r", line)) {
      tmp <- "table.vline(x: %s, start: %s, end: %s, stroke: %s),"
      tmp <- sprintf(tmp, xmax, ymin, ymax, stroke)
      out <- paste(out, tmp)
    }
    return(out)
  })
  if (length(lin) > 0) {
    x@table_string <- lines_insert(
      x@table_string,
      paste(unlist(lin, use.names = FALSE), collapse = "\n"),
      "tinytable lines before",
      "before"
    )
  }
  return(x)
}


#' Internal styling function
#'
#' @inheritParams style_tt
#' @keywords internal
#' @noRd
setMethod(
  f = "style_eval",
  signature = "tinytable_typst",
  definition = function(
    x,
    i = NULL,
    j = NULL,
    bold = FALSE,
    italic = FALSE,
    monospace = FALSE,
    underline = FALSE,
    strikeout = FALSE,
    color = NULL,
    background = NULL,
    fontsize = NULL,
    align = NULL,
    alignv = NULL,
    line = NULL,
    line_color = NA_character_,
    line_width = 0.1,
    indent = 0,
    midrule = FALSE, # undocumented, only used by `group_tt()`
    ...
  ) {
    # Use populated @style_other / @style_lines from build_tt()
    inputs <- style_backend_inputs(x, STYLE_PROPS_TYPST)
    other <- inputs$other
    lines <- inputs$lines

    # Column-group underlines are drawn inside the header cells by
    # typst_build_group_header(); skip them here to avoid drawing them twice.
    consumed <- typst_group_line_rules(x)$consumed
    if (!is.null(lines) && length(consumed) == nrow(lines) && any(consumed)) {
      lines <- lines[!consumed, , drop = FALSE]
      if (nrow(lines) == 0) {
        lines <- NULL
      }
    }

    # Process lines using the expanded data
    x <- process_typst_lines(x, lines)

    # Process other styles using the expanded data
    x <- process_typst_other_styles(x, other)

    return(x)
  }
)

#' Alias for `style_tt()`
#'
#' @rdname style_tt
#' @export
tt_style <- style_tt

#' Alias for `group_tt()`
#'
#' @rdname group_tt
#' @export
tt_group <- group_tt

#' Alias for `format_tt()`
#'
#' @rdname format_tt
#' @export
tt_format <- format_tt

#' Alias for `plot_tt()`
#'
#' @rdname plot_tt
#' @export
tt_plot <- plot_tt

#' Alias for `save_tt()`
#'
#' @rdname save_tt
#' @export
tt_save <- save_tt

# internal function
# style_tt() stores style calls and we only want to evaluate them at the end because
# some rows may be added, which changes how the style is applied
#
# THE ORDER MATTERS A LOT!

#' Build group header and body parts with position calculations
#' @keywords internal
#' @noRd
rbind_body_groupi <- function(x) {
  # Reconstruct the final table by combining formatted data_body and group_data_i parts
  if (nrow(x@group_data_i) == 0) {
    # No groups - @data_body already contains the final formatted data
    return(x)
  }

  # Calculate total final table size
  total_rows <- nrow(x@data_body) + nrow(x@group_data_i)
  final_ncol <- ncol(x@data_body)

  # Create final data frame with proper structure
  final_df <- data.frame(matrix(
    NA_character_,
    nrow = total_rows,
    ncol = final_ncol
  ))
  colnames(final_df) <- colnames(x@data_body)

  # Insert body data at index_body positions
  if (nrow(x@data_body) > 0 && length(x@index_body) > 0) {
    ok <- !is.na(x@index_body) & x@index_body >= 1 & x@index_body <= total_rows
    if (any(ok)) {
      final_df[x@index_body[ok], ] <- x@data_body[which(ok), , drop = FALSE]
    }
  }

  # Insert group i data at group_index_i positions
  if (nrow(x@group_data_i) > 0 && length(x@group_index_i) > 0) {
    idx <- x@group_index_i[seq_len(nrow(x@group_data_i))]
    ok <- !is.na(idx) & idx >= 1 & idx <= total_rows
    if (any(ok)) {
      final_df[idx[ok], ] <- x@group_data_i[which(ok), , drop = FALSE]
    }
  }

  x@data_body <- final_df

  # before format_tt() because we need the indices
  x@nrow <- nrow(x@data) + nrow(x@group_data_i)

  return(x)
}


build_tt <- function(x, output = NULL) {
  x <- sanitize_output(x, output)
  output <- infer_output(x)

  x <- switch(output,
    html = {
      if (identical(x@html_engine, "tabulator")) {
        swap_class(x, "tinytable_tabulator")
      } else {
        swap_class(x, "tinytable_html")
      }
    },
    latex = swap_class(x, "tinytable_tabularray"),
    markdown = swap_class(x, "tinytable_grid"),
    typst = swap_class(x, "tinytable_typst"),
    dataframe = swap_class(x, "tinytable_dataframe"),
    stop("Unsupported output format: '", output, "'. Supported formats are: html, latex, markdown, typst, dataframe", call. = FALSE)
  )

  x@output <- output


  # Calculate which positions are body vs group
  if (nrow(x@group_data_i) == 0) {
    x@index_body <- seq_len(nrow(x))
  } else {
    x@index_body <- setdiff(seq_len(nrow(x)), x@group_index_i)
  }

  # format each component individually, including groups before inserting them into the body
  for (l in x@lazy_format) {
    o <- attr(l, "output")
    if (is.null(o) || x@output %in% o) {
      l[["x"]] <- x
      x <- eval(l)
    }
  }

  # insert group rows into body
  x <- rbind_body_groupi(x)

  # pre-process: theme_*() calls that need formatting conditional on @output
  # this is useful after rbind() because we now have the final indices and headers
  for (p in x@lazy_prepare) {
    o <- attr(p, "output")
    if (is.null(o) || x@output %in% o) {
      x <- p(x)
    }
  }

  # plots and images
  for (l in x@lazy_plot) {
    o <- attr(l, "output")
    if (is.null(o) || x@output %in% o) {
      l[["x"]] <- x
      x <- eval(l)
    }
  }

  # add footnote markers just after formatting, otherwise appending converts to string
  x <- footnote_markers(x)

  # Create rectangular style dataframe with one row per cell
  # This happens after row group operations so we have the final row structure
  iseq <- seq_len(nrow(x))
  # include headers: 0, -1, ... for each header row, but only when there are
  # header rows. The old expression `-1 * 0:(x@nhead - 1)` produced c(0, 1)
  # when nhead == 0, adding a phantom header row and duplicating body row 1.
  if (x@nhead > 0) {
    iseq <- c(-(seq_len(x@nhead) - 1L), iseq)
  }
  jseq <- seq_len(ncol(x))
  rect <- expand.grid(i = iseq, j = jseq)

  # Initialize all style columns with NA
  for (col in STYLE_PROPS) {
    rect[[col]] <- NA
  }

  x@style_other <- rect

  # apply style_tt() and theme_*() after all group operations.
  # style_tt_lazy() exposes only the current call's settings frame in @style;
  # collect those frames here and rbind() them once. This replaces the previous
  # pattern where every style_tt() call grew @style with an incremental rbind(),
  # an O(N^2) cost that dominated on tables with many style_tt() calls.
  lazy_style <- x@lazy_style
  x@lazy_style <- list()

  style_frames <- vector("list", length(lazy_style))
  for (k in seq_along(lazy_style)) {
    p <- lazy_style[[k]]
    o <- attr(p, "output")
    if (is.null(o) || x@output %in% o) {
      p[["x"]] <- x
      x <- eval(p)
      if (nrow(x@style) > 0L) style_frames[[k]] <- x@style
    }
  }
  style_frames <- style_frames[!vapply(style_frames, is.null, logical(1))]
  x@style <- if (length(style_frames)) do.call(rbind, style_frames) else data.frame()

  # Fix colspan that exceeds column count after lazy styles are evaluated
  x@style <- clamp_colspan(x@style, x@ncol)

  # apply styling AFTER formatting/escaping to avoid escaping the style brackets
  x <- style_notes(x)
  x <- style_caption(x)

  # Populate @style_other and @style_lines by resolving @style in a single
  # batched pass. This must happen before build_eval() for backends (like
  # Grid) that apply styles during build.
  # The original per-row loop with apply_style_to_rect()/append_lines_to_rect()
  # is preserved here as a fallback; for non-trivial style counts the batched
  # resolver in resolve_styles_batch() is dramatically faster because it does
  # not rescan the entire rectangular grid for every style_tt() entry.
  resolved <- resolve_styles_batch(x@style_other, x@style)
  has_style <- resolved$has_style
  x@style_other <- resolved$other
  x@style_lines <- resolved$lines

  # Sort style_other for consistent ordering (j first, then i within each j)
  # This ensures test snapshots are deterministic
  ord <- order(x@style_other$j, x@style_other$i)
  x@style_other <- x@style_other[ord, ]
  attr(x@style_other, "has_style") <- has_style[ord]

  # inline plots must be re-boxed for `alignv` to have an effect in LaTeX (#673)
  x <- plot_alignv_latex(x)

  # draw the table
  x <- build_eval(x)

  # groups require the table to be drawn first, especially group_tabularray_col() and friends
  # Handle column groups from @group_data_j
  if (nrow(x@group_data_j) > 0) {
    # Calculate ihead for the group headers - start from -1 for the top header row
    ihead <- -1
    # Apply group_eval_j once with all groups
    x <- group_eval_j(x, j = seq_len(ncol(x)), ihead = ihead)
  }

  x <- style_eval(x)

  # no-op: only modifies markdown
  x <- grid_colspan(x)

  x <- finalize(x)

  # post-process: theme_*() calls that need formatting conditional on @output and table drawn
  for (p in x@lazy_finalize) {
    o <- attr(p, "output")
    if (is.null(o) || x@output %in% o) {
      x <- p(x)
    }
  }

  x@table_string <- lines_drop_consecutive_empty(x@table_string)
  if (x@markdown_style == "gfm") {
    assert_dependency("pandoc")
    x@table_string <- paste(
      pandoc::pandoc_convert(text = x@table_string, to = "gfm"),
      collapse = "\n"
    )
  }

  return(x)
}


build_prepare <- function(x, fn, output = NULL) {
  attr(fn, "output") <- output
  x@lazy_prepare <- c(x@lazy_prepare, list(fn))
  return(x)
}


build_finalize <- function(x, fn, output = NULL) {
  attr(fn, "output") <- output
  x@lazy_finalize <- c(x@lazy_finalize, list(fn))
  return(x)
}

source("helpers.R")

# Issue #414
df1 = df2 = data.frame(a = c(1, 2, 3), b = c(4, 5, 6))
colnames(df1) = NULL
expect_equal(length(colnames(tt(df1, rownames = TRUE))), 0)
expect_equal(length(colnames(tt(df1))), 0)

tab1 = save_tt(tt(df1), "markdown")
tab2 = save_tt(tt(df2, colnames = FALSE), "markdown")
expect_equal(tab1, tab2)


# colnames="label"
dat <- mtcars[1:5, c("cyl", "mpg", "hp")]
attr(dat$cyl, "label") <- "Cylinders"
attr(dat$mpg, "label") <- "Miles per Gallon"
attr(dat$hp, "label") <- "Horse Power"
tab <- tt(dat, colnames = "label") |> save_tt("dataframe")
expect_equal(unlist(tab[1, ]), c("Cylinders", "Miles per Gallon", "Horse Power"))


# Regression: colnames = FALSE must not create a phantom header row (i = 0)
# or duplicate body row 1 in @style_other
x <- build_tt(tt(mtcars[1:3, 1:3], colnames = FALSE), "html")
expect_equal(nrow(x@style_other), 9)
expect_equal(sort(unique(x@style_other$i)), c(1, 2, 3))
expect_false(any(duplicated(x@style_other[, c("i", "j")])))

# header index sequences unchanged for nhead = 1 and nhead = 2
x1 <- build_tt(tt(mtcars[1:3, 1:3]), "html")
expect_equal(sort(unique(x1@style_other$i)), c(0, 1, 2, 3))
x2 <- build_tt(tt(mtcars[1:3, 1:3]) |> group_tt(j = list("G" = 1:2)), "html")
expect_equal(sort(unique(x2@style_other$i)), c(-1, 0, 1, 2, 3))

# Regression: as.character() errored on fresh tables because @output was
# still "tinytable"
out <- as.character(tt(mtcars[1:2, 1:2]))
expect_inherits(out, "character")
expect_true(grepl("mpg", out))

# Regression: colnames<- errored on tables created with colnames = FALSE,
# and did not restore the header row
y <- tt(mtcars[1:3, 1:3], colnames = FALSE)
colnames(y) <- c("a", "b", "c")
expect_equal(colnames(y), c("a", "b", "c"))
expect_equal(y@nhead, 1)
expect_true(grepl("a", save_tt(y, "markdown")))
y <- tt(mtcars[1:3, 1:3], colnames = FALSE)
names(y) <- c("a", "b", "c")
expect_equal(names(y), c("a", "b", "c"))
expect_equal(y@nhead, 1)
# wrong length still errors
y <- tt(mtcars[1:3, 1:3], colnames = FALSE)
expect_error(colnames(y) <- c("a", "b"))

# issue #95: plot_tt() can select j with a regex
tab <- data.frame(a = 1, b = 2) |>
  tt() |>
  plot_tt(1, j = "a", fun = "density", data = list(rnorm(100)))
expect_inherits(tab, "tinytable")

# An erroring plot function must not leak an open graphics device
bad_fun <- function(d, ...) function() stop("boom in user plot function")
tab <- tt(data.frame(x = 1, plot = "")) |>
  plot_tt(j = 2, fun = bad_fun, data = list(1))
n_dev <- length(grDevices::dev.list())
expect_error(save_tt(tab, "markdown"), pattern = "boom in user plot function")
expect_equal(length(grDevices::dev.list()), n_dev)

# Issue #673: `alignv` moves inline plot images in LaTeX via \raisebox
dat673 <- data.frame(Metric = c("Sales", "Growth"), Value = c(75, 92))
mk673 <- function(...) {
  out <- tt(dat673) |>
    plot_tt(j = 2, fun = "bar", data = as.list(dat673$Value), height = 5)
  out <- style_tt(out, ...)
  as.character(build_tt(out, "latex")@table_string)
}
# alignv on the text column applies to the whole row's image (as filed in #673)
expect_true(grepl("\\raisebox{-0.5\\height}{\\includegraphics", mk673(j = 1, alignv = "m"), fixed = TRUE))
# alignv on the plot column itself
expect_true(grepl("\\raisebox{\\dimexpr-\\height+\\ht\\strutbox\\relax}{\\includegraphics", mk673(j = 2, alignv = "t"), fixed = TRUE))
# bottom alignment and the default are unchanged
expect_false(grepl("raisebox", mk673(j = 1, alignv = "b"), fixed = TRUE))
expect_false(grepl("raisebox", mk673(j = 1, bold = TRUE), fixed = TRUE))
# cell-level style wraps only the targeted cell
one <- mk673(i = 1, j = 2, alignv = "m")
expect_equal(lengths(regmatches(one, gregexpr("raisebox", one))), 1L)

source("helpers.R")
using("tinysnapshot")

# Issue: NSE predicates must evaluate against the original typed data, not the
# character-coerced @data_body ("5" > "30" is TRUE lexicographically).
dat_num <- data.frame(x = c(5, 20, 100))
tab <- tt(dat_num) |> style_tt(i = x > 30, bold = TRUE)
tab <- build_tt(tab, "markdown")
expect_equal(tab@style$i[which(tab@style$bold)], 3)
expect_true(grepl("**100**", tab@table_string, fixed = TRUE))
expect_false(grepl("**5**", tab@table_string, fixed = TRUE))

# Same predicate via format_tt()
tab <- tt(dat_num) |> format_tt(i = x > 30, sprintf = "[%s]")
tab <- build_tt(tab, "markdown")
expect_true(grepl("[100]", tab@table_string, fixed = TRUE))
expect_false(grepl("[5]", tab@table_string, fixed = TRUE))

# Issue: with a plain data frame input, NSE expressions must be evaluated in
# the user's calling frame, so function-local variables resolve.
fn <- function() {
  myrows <- 2
  format_tt(data.frame(x = c(1.111, 2.222, 3.333)), i = myrows, digits = 2)
}
out <- fn()
expect_equivalent(out$x, c("1.111", "2.2", "3.333"))

# Issue: NSE row predicates are evaluated against pre-group data rows, but
# numeric `i` is resolved in final-table coordinates after group_tt() inserts
# label rows. Predicate-derived indices must be remapped through @index_body.
dat <- data.frame(x = 1:6, y = letters[1:6])

get_bg_rows <- function(tab) {
  so <- tab@style_other
  sort(unique(so$i[!is.na(so$background)]))
}

# group_tt() before style_tt()
tab <- tt(dat) |>
  group_tt(i = list("G" = 4)) |>
  style_tt(i = x == 5, background = "pink")
tab <- build_tt(tab, "markdown")
# group label at final row 4; data row x == 5 sits at final row 6
expect_equal(get_bg_rows(tab), 6)

# style_tt() before group_tt(): same result
tab <- tt(dat) |>
  style_tt(i = x == 5, background = "pink") |>
  group_tt(i = list("G" = 4))
tab <- build_tt(tab, "markdown")
expect_equal(get_bg_rows(tab), 6)

# Unchanged behavior: explicit numeric `i` refers to final-table positions
tab <- tt(dat) |>
  group_tt(i = list("G" = 4)) |>
  style_tt(i = 5, background = "pink")
tab <- build_tt(tab, "markdown")
expect_equal(get_bg_rows(tab), 5)

# No groups: remap is a no-op
tab <- tt(dat) |> style_tt(i = x == 5, background = "pink")
tab <- build_tt(tab, "markdown")
expect_equal(get_bg_rows(tab), 5)

# format_tt() shares the remap: predicate targets the x == 5 row after grouping
tab <- tt(dat) |>
  group_tt(i = list("G" = 4)) |>
  format_tt(i = x == 5, sprintf = "[%s]")
tab <- build_tt(tab, "markdown")
expect_true(grepl("[5]", tab@table_string, fixed = TRUE))
expect_false(grepl("[4]", tab@table_string, fixed = TRUE))

# plot_tt() shares the remap: image lands on the x == 5 row after grouping
tab <- tt(dat) |>
  group_tt(i = list("G" = 4)) |>
  plot_tt(i = x == 5, j = 2, images = "fake.png", height = 1)
tab <- build_tt(tab, "markdown")
lines <- strsplit(tab@table_string, "\n")[[1]]
img_line <- grep("fake.png", lines, fixed = TRUE, value = TRUE)
expect_equal(length(img_line), 1L)
# the image must land in the row whose first column is 5, not 4
expect_true(grepl("^\\| 5 \\|", img_line))

# NSE predicate with logical variable from the calling frame is also remapped
sel <- dat$x == 5
tab <- tt(dat) |>
  group_tt(i = list("G" = 4)) |>
  style_tt(i = sel, background = "pink")
tab <- build_tt(tab, "markdown")
expect_equal(get_bg_rows(tab), 6)

source("helpers.R")
using("tinysnapshot")

# `j` refers to a position in the rendered table, so subsetting does not shift
# stored indices; entries that fall outside the narrower table are dropped and
# must not error at build.
w <- style_tt(tt(mtcars[1:3, 1:4]), j = 4, color = "red")
s <- subset(w, select = 1:2)
expect_equal(length(s@lazy_style), 0L)
out <- build_tt(s, "html")
expect_inherits(out, "tinytable")
expect_false(grepl("red", out@table_string, fixed = TRUE))

# A style on a kept column survives and targets the correct column.
w <- style_tt(tt(mtcars[1:3, 1:4]), j = 1, color = "red")
s <- subset(w, select = 1:2)
out <- build_tt(s, "html")
sty <- out@style[!is.na(out@style$color) & out@style$color == "red", ]
expect_true(nrow(sty) > 0)
expect_true(all(sty$j == 1))

# A style keeps its position in the rendered table: it does not follow the
# column it originally pointed to.
w <- style_tt(tt(mtcars[1:3, 1:4]), j = 2, color = "blue")
s <- subset(w, select = 2:4)
expect_equal(s@lazy_style[[1]][["j"]], 2)
out <- build_tt(s, "html")
sty <- out@style[!is.na(out@style$color) & out@style$color == "blue", ]
expect_true(all(sty$j == 2))

# Issue: drop = TRUE broke S4 slot assignment when selecting a single column.
# It now warns and is ignored.
expect_warning(
  subset(tt(mtcars[1:3, 1:3]), select = 1, drop = TRUE),
  pattern = "not supported"
)
s <- suppressWarnings(subset(tt(mtcars[1:3, 1:3]), select = 1, drop = TRUE))
expect_inherits(s, "tinytable")
expect_equal(s@ncol, 1L)
expect_inherits(build_tt(s, "markdown"), "tinytable")

# Issue: character select produced an all-NA index vector, silently remapping
# stored lazy calls to every column. Names are now matched and validated.
x <- format_tt(tt(mtcars[1:3, 1:3]), j = 1, digits = 0)
y <- subset(x, select = "mpg")
expect_equal(y@lazy_format[[1]][["j"]], 1L)
out <- build_tt(y, "markdown")
expect_true(grepl("21.0", out@table_string, fixed = TRUE))

# character select leaves in-range indices alone
x <- format_tt(tt(mtcars[1:3, 1:3]), j = 2, digits = 0)
y <- subset(x, select = c("cyl", "disp"))
expect_equal(y@lazy_format[[1]][["j"]], 2)

# unknown column names error informatively instead of corrupting state
expect_error(
  subset(tt(mtcars[1:3, 1:3]), select = "nope"),
  pattern = "nope"
)

# Colspan entries (e.g. group_tt(i = ...) row labels, which style i = row,
# j = 1, colspan = ncol) survive and are clipped to the narrower table.
w <- tt(mtcars[1:5, 1:4]) |> group_tt(i = list("grp" = 3))
s <- subset(w, select = -1)
expect_equal(length(s@lazy_style), 1L)
expect_equal(s@lazy_style[[1]][["j"]], 1)
expect_equal(s@lazy_style[[1]][["colspan"]], 3L)
expect_inherits(build_tt(s, "html"), "tinytable")

# Negative j is a selection against the rendered table, so it is left alone.
x <- format_tt(tt(mtcars[1:3, 1:4]), j = -1, digits = 0)
y <- subset(x, select = 2:3)
expect_equal(y@lazy_format[[1]][["j"]], -1)

# Column spans stay where they are and are clipped at the new right edge:
# "AAA" keeps its two columns, "BBB" is reduced to one.
g <- tt(mtcars[1:3, 1:4]) |> group_tt(j = list("AAA" = 1:2, "BBB" = 3:4))
s <- subset(g, select = -2)
expect_equal(unname(unlist(s@group_data_j[1, ])), c("AAA", "", "BBB"))
spans <- s@lazy_style[vapply(
  s@lazy_style,
  function(e) isTRUE(e[["colspan"]] > 1),
  logical(1)
)]
expect_equal(length(spans), 1L)
expect_equal(spans[[1]][["j"]], 1)
expect_equal(spans[[1]][["colspan"]], 2L)

# A span whose anchor falls outside the subsetted table disappears.
s <- subset(g, select = 1:2)
expect_equal(unname(unlist(s@group_data_j[1, ])), c("AAA", ""))
expect_inherits(build_tt(s, "html"), "tinytable")

# =============================================================================
# TABULATOR PLOT FORMATTERS
# =============================================================================

#' Create Tabulator plot formatters for plot_tt()
#' @param plot_type Character string: "histogram", "density", "bar", "line"
#' @param data Data for the plot
#' @param color Color for the plot (single value or length-2 vector)
#' @param xlim X-axis limits
#' @return List with formatter configuration
#' @keywords internal
#' @noRd
tabulator_plot_formatter <- function(
    plot_type,
    data,
    color = "black",
    xlim = NULL) {

  formatter_config <- list()

  # Only the foreground color is used: the sparkline/histogram formatters in
  # inst/tinytable.js do not support a background color, so the second element
  # of `plot_tt(color = c(fg, bg))` is ignored by the tabulator engine.
  bar_color <- color[1]

  if (plot_type == "line") {
    # Use custom SVG sparkline formatter
    if (is.data.frame(data) && "x" %in% names(data) && "y" %in% names(data)) {
      y_values <- data$y
    } else if (is.numeric(data)) {
      y_values <- data
    } else {
      stop("Line data must be a data frame with x and y columns or a numeric vector", call. = FALSE)
    }

    formatter_config$formatter <- "tinytable_sparkline"
    formatter_config$formatterParams <- list(
      color = bar_color,
      width = 120,
      height = 30,
      strokeWidth = 1.5
    )
    return(list(
      config = formatter_config,
      data = y_values,
      requires_custom_js = TRUE
    ))

  } else if (plot_type == "bar") {
    # Use Tabulator's built-in progress formatter
    if (!is.numeric(data) || length(data) != 1) {
      stop("Bar data must be a single numeric value", call. = FALSE)
    }

    max_val <- if (!is.null(xlim)) xlim[2] else max(data, na.rm = TRUE)

    formatter_config$formatter <- "progress"
    formatter_config$formatterParams <- list(
      min = 0,
      max = max_val,
      color = bar_color,
      legendColor = "#000000",
      legendAlign = "left"
    )

    return(list(config = formatter_config, data = data))

  } else if (plot_type == "histogram") {
    # For histogram, use custom formatter with canvas
    # Data will be stored in cell value, not formatterParams
    if (!is.numeric(data)) {
      stop("Histogram data must be numeric", call. = FALSE)
    }

    return(list(
      config = list(
        formatter = "tinytable_histogram",
        formatterParams = list(
          color = bar_color,
          width = 120,
          height = 30
        )
      ),
      data = as.numeric(data),
      requires_custom_js = TRUE
    ))

  } else if (plot_type == "density") {
    # For density, calculate density points and use custom sparkline with fill
    if (!is.numeric(data)) {
      stop("Density data must be numeric", call. = FALSE)
    }

    dens <- stats::density(stats::na.omit(data))
    y_values <- dens$y

    formatter_config$formatter <- "tinytable_sparkline"
    formatter_config$formatterParams <- list(
      color = bar_color,
      fillArea = TRUE,
      width = 120,
      height = 30,
      strokeWidth = 1.5
    )
    return(list(
      config = formatter_config,
      data = y_values,
      requires_custom_js = TRUE
    ))
  }

  stop(paste("Unknown plot type:", plot_type), call. = FALSE)
}

get_css <- function(table_id = NULL) {
  css_path <- system.file("tinytable.css", package = "tinytable")
  css <- paste(readLines(css_path, warn = FALSE), collapse = "\n")

  # Framework CSS remains global - no scoping needed
  # Individual cell CSS rules are scoped in html_style.R instead

  return(css)
}

#' Deprecated: Use format-specific theme functions instead
#'
#' @description
#' **DEPRECATED**: The `theme_tt()` function has been deprecated. Please use the format-specific or style-specific theme functions instead.
#'
#' @param x deprecated
#' @param theme deprecated
#' @param ... Additional arguments
#'
#' @return Throws an informative error message
#' @export
theme_tt <- function(x, theme, ...) {
  format_functions <- paste(c(
    "* theme_html()",
    "* theme_latex()",
    "* theme_typst()"
  ), collapse = "\n")

  style_functions <- paste(c(
    "* theme_grid()",
    "* theme_revealjs()",
    "* theme_rotate()",
    "* theme_striped()",
    "* theme_empty()"
  ), collapse = "\n")

  format_msg <- paste0("\nFormat-specific functions:\n\n", format_functions)
  style_msg <- paste0("\nStyle-specific functions:\n\n", paste(style_functions, collapse = "\n"))

  warning(
    "The theme_tt() function is deprecated. Please use format-specific or style-specific theme functions instead:\n",
    format_msg,
    "\n",
    style_msg,
    call. = FALSE
  )


  # "void" is a legacy alias for "empty"
  if (identical(theme, "void")) {
    theme <- "empty"
  }

  # dispatch through the same dictionary used by tt(theme = ...)
  if (isTRUE(check_string(theme)) && theme %in% names(theme_dictionary)) {
    x <- theme_dictionary[[theme]](x, ...)
  } else {
    legacy <- c(
      "bootstrap" = "theme_html(x, engine = \"bootstrap\", class = ...)",
      "tabular" = "theme_latex(x, environment = \"tabular\")",
      "resize" = "theme_latex(x, resize_width = ..., resize_direction = ...)",
      "placement" = "theme_latex(x, placement = ...)",
      "multipage" = "theme_latex(x, multipage = TRUE)",
      "spacing" = "the `height` argument of tt()",
      "rotating" = "theme_rotate(x, angle = ...)"
    )
    if (is.character(theme) && length(theme) == 1 && theme %in% names(legacy)) {
      msg <- sprintf(
        "The \"%s\" theme was removed from `tinytable`. Use %s instead.",
        theme,
        legacy[[theme]]
      )
    } else {
      msg <- sprintf(
        "Invalid `theme` argument. `theme_tt()` supports these theme names: %s.",
        paste(sprintf('"%s"', c(
          "default", "grid", "revealjs", "striped", "empty", "void", "rotate"
        )), collapse = ", ")
      )
    }
    stop(msg, call. = FALSE)
  }
  return(x)
}

# =============================================================================
# SHARED TABULARRAY HELPER FUNCTIONS
# =============================================================================

#' Insert content into tabularray string at specific locations
#' @keywords internal
#' @noRd
insert_tabularray_content <- function(x, content = NULL, type = "body") {
    out <- x

    out <- strsplit(out, "\n")[[1]]
    comment <- switch(type,
        "body" = "% tabularray inner close",
        "outer" = "% tabularray outer close",
        "inner" = "% tabularray inner close"
    )
    idx <- grep(comment, out)

    if (length(content) > 0) {
        content <- trimws(content)
        if (type == "body") {
            # Preserve the sequential single-insert semantics: each element is
            # placed immediately after the marker, so a vector lands reversed.
            out <- c(out[1:idx], rev(content), out[(idx + 1):length(out)])
        } else {
            content <- paste0(content, ifelse(grepl(",$", content), "", ","))
            out <- c(out[1:(idx - 1)], content, out[idx:length(out)])
        }
    }

    out <- paste(out, collapse = "\n")

    return(out)
}

#' Define color in tabularray preamble
#' @keywords internal
#' @noRd
define_color_preamble <- function(x, col) {
    if (grepl("^#", col)) {
        # hex color need to be defined in LaTeX
        col <- sub("^#", "c", col)
        # Match the definition token itself, not the color name anywhere in the
        # document (cell text containing the token must not suppress it).
        regex <- sprintf("DefineColor\\{%s\\}", col)
        if (!grepl(regex, x@table_string)) {
            b <- sprintf(
                "\\tinytableDefineColor{%s}{HTML}{%s}",
                col,
                sub("^c", "", col)
            )
            x@table_string <- insert_tabularray_content(
                x@table_string,
                content = b,
                type = "body"
            )
        }
    }
    return(x)
}

#' Calculate d-column specification for tabularray
#' @keywords internal
#' @noRd
calculate_dcolumn_spec <- function(j, x) {
    siunitx <- get_option(
        "tinytable_siunitx_table_format",
        # `table-align-text-before` must stay at its default (true), otherwise
        # siunitx reserves no width for comparators and a value like `<0.001`
        # is typeset outside the column. `table-align-text-after=false` is
        # still needed to avoid reserving width for significance stars.
        default = "table-format=-%s.%s,table-align-text-after=false,input-symbols={-,\\*+()}"
    )
    num <- unlist(x@data_body[, j])

    # empty cells
    num <- sapply(num, trimws)
    num <- num[sapply(num, nchar) > 0]

    num <- strsplit(num, "\\.")
    num <- lapply(num, function(k) if (length(k) == 1) c(k, " ") else k)

    left <- sapply(num, function(k) k[[1]])
    right <- sapply(num, function(k) k[[2]])
    left <- max(nchar(gsub("\\D", "", left)))
    right <- max(nchar(gsub("\\D", "", right)))
    out <- sprintf(siunitx, left, right)
    out <- sprintf("si={%s},", out)
    return(out)
}

#' Generate LaTeX range string for tabularray
#' @keywords internal
#' @noRd
latex_range_string <- function(x) {
    if (length(x) == 0) {
        return("")
    }
    x <- sort(unique(x))
    start <- x[c(TRUE, diff(x) != 1)]
    end <- x[c(diff(x) != 1, TRUE)]
    parts <- ifelse(start == end, start, paste0(start, "-", end))
    paste(parts, collapse = ",")
}

#' Build tabularray header row from group data
#' @keywords internal
#' @noRd
build_tabularray_header <- function(group_row, ncols) {
    header <- rep("", ncols)

    spans <- parse_group_spans(group_row)

    for (span_idx in seq_len(nrow(spans))) {
        header[spans$start[span_idx]] <- spans$label[span_idx]
    }

    header_line <- paste(header, collapse = " & ")
    header_line <- paste(header_line, "\\\\", "")

    return(header_line)
}

#' Insert header row into tabularray string
#' @keywords internal
#' @noRd
insert_tabularray_header <- function(x, header_line) {
    out <- strsplit(x@table_string, split = "\\n")[[1]]

    # Insert the header line
    idx_candidates <- c(
        grep("% tabularray inner close", out),
        grep("\\toprule", out, fixed = TRUE)
    )
    if (length(idx_candidates) == 0) {
        stop(
            "tinytable: cannot locate insertion point for column group header.",
            call. = FALSE
        )
    }
    idx <- max(idx_candidates)

    out <- c(
        out[1:idx],
        trimws(header_line),
        out[(idx + 1):length(out)]
    )

    out <- paste(out, collapse = "\n")
    x@table_string <- out

    return(x)
}

#' tinytable S4 method
#'
#' @keywords internal
setMethod(
  f = "group_eval_j",
  signature = "tinytable_typst",
  definition = function(x, i = NULL, j = NULL, indent = 0, ...) {
    # Column grouping now uses generic system - no special processing needed
    return(x)
  }
)

# All Typst-specific column grouping functions removed
# Column groups now handled by generic system in typst_tt.R

# replaced this with @rawNamespace S3method(knitr::knit_print, tinytable) in print.R

# .onLoad <- function(libname, pkgname) {
#   if (requireNamespace("knitr", quietly = TRUE)) {
#     registerS3method("knit_print", "tinytable", knit_print.tinytable, envir = asNamespace("knitr"))
#   }
# }

#' Deprecated: Use theme_tinytable() instead
#'
#' @description
#' `theme_default()` has been renamed to `theme_tinytable()`. This function is
#' deprecated and will be removed in a future version.
#'
#' @inheritParams theme_tinytable
#' @return A modified `tinytable` object.
#' @export
theme_default <- function(x, ...) {
  warning(
    "`theme_default()` is deprecated. Please use `theme_tinytable()` instead.",
    call. = FALSE
  )
  theme_tinytable(x, ...)
}

#' Format columns of a data frame
#'
#' This function formats the columns of a data frame based on the column type (logical, date, numeric).
#' It allows various formatting options like significant digits, decimal points, and scientific notation.
#' It also includes custom formatting for date and boolean values.
#' If this function is applied several times to the same cell, the last transformation is retained and the previous calls are ignored, except for the `escape` argument which can be applied to previously transformed data.
#' Each call to `format_tt()` is recorded and later replayed sequentially at render time, so later calls operate on the output of earlier ones.
#' Arguments are executed in a fixed order inside each call (see code below), so combining features such as `linebreak`, `escape`, `markdown`, etc. in a single call may produce surprising results.
#' Chain multiple calls when you need finer control (e.g., `format_tt(escape = TRUE) |> format_tt(linebreak = "<br>")`), keeping in mind that the second call will see the already-escaped text.
#'
#' @param x A data frame or a vector to be formatted.
#' @param i Numeric vector or string.
#'   - Numeric vector: Row indices where the styling should be applied. Can be a single value or a vector.
#'   - String: Table components to format "caption", "colnames", "groupi" (row group labels), "~groupi" (non-group rows), "groupj" (column group labels), "notes".
#'   - If both the `i` and `j` are omitted (default: NULL), formatting is applied to all table elements, including caption, notes, and group labels.
#' @param digits Number of significant digits or decimal places, depending on the `num_fmt` argument.
#'   - `num_fmt = "significant"` (default): `digits` is the minimum number of significant digits. Formatting is applied column-wise, so the smallest value in the column determines the decimal representation for the whole column.
#'   - `num_fmt = "significant_cell"`: same as `"significant"`, but formatting is applied cell-by-cell. Each cell displays exactly `digits` significant figures, regardless of other values in the column.
#'   - `num_fmt = "decimal"`: `digits` is the number of decimal places (digits after the decimal point) for all values.
#'   - `num_fmt = "scientific"`: `digits` is the number of decimal places in the mantissa of scientific notation (e.g., `digits = 2` gives `1.23e+04`).
#' @param num_fmt The format for numeric values; one of 'significant', 'significant_cell', 'decimal', or 'scientific'. See the `digits` argument for details on how each format interprets the `digits` value.
#' @param num_zero Logical; if TRUE, trailing zeros are kept in "decimal" format (but not in "significant" format).
#' @param num_mark_big Character to use as a thousands separator.
#' @param num_mark_dec Decimal mark character. Default is the global option 'OutDec'.
#' @param num_suffix Logical; if TRUE display short numbers with `digits` significant digits and K (thousands), M (millions), B (billions), or T (trillions) suffixes.
#' @param date A string passed to the `format()` function, such as "%Y-%m-%d". See the "Details" section in `?strptime`
#' @param bool A function to format logical columns. Defaults to title case.
#' @param math Logical. If TRUE, wrap cell values in math mode `$..$`. This is useful for LaTeX output or with HTML MathJax `options(tinytable_html_mathjax=TRUE)`.
#' @param other A function to format columns of other types. Defaults to `as.character()`.
#' @param replace Logical, String or Named list of vectors
#' - TRUE: Replace `NA` and `NaN` by an empty string.
#' - FALSE: Print `NA` and `NaN` as strings.
#' - String: Replace `NA` and `NaN` entries by the user-supplied string.
#' - Named list: Replace matching elements of the vectors in the list by theirs names. Example:
#'      - `list("-" = c(NA, NaN), "Tiny" = -Inf, "Massive" = Inf)`
#' @param escape Logical or "latex" or "html". If TRUE, escape special characters to display them as text in the format of the output of a `tt()` table.
#' - If `i` and `j` are both `NULL`, escape all cells, column names, caption, notes, and spanning labels created by `group_tt()`.
#' @param markdown Logical; if TRUE, render markdown syntax in cells. Ex: `_italicized text_` is properly italicized in HTML and LaTeX.
#' @param fn Function for custom formatting. Accepts a vector and returns a character vector of the same length.
#' @param quarto Logical. Enable Quarto data processing and wrap cell content in a `data-qmd` span (HTML) or `\QuartoMarkdownBase64{}` macro (LaTeX). See warnings in the Global Options section below.
#' @param sprintf String passed to the `?sprintf` function to format numbers or interpolate strings with a user-defined pattern (similar to the `glue` package, but using Base R).
#' @param linebreak NULL or a single string. If it is a string, replaces that string with appropriate line break sequences depending on the output format (HTML: `<br>`, LaTeX: `\\\\`, Typst: `\\ `). Markdown output is excluded from line break replacement.
#' @param output Apply formatting only if the `tt()` object is rendered in the specified format. A character vector of one or more of "latex", "html", "typst", or "markdown". If `NULL` (default), apply formatting regardless of the output format.
#' @inheritParams tt
#' @inheritParams style_tt
#' @template global_options
#'
#' @return A data frame with formatted columns.
#' @export
#' @examples
#' dat <- data.frame(
#'   a = rnorm(3, mean = 10000),
#'   b = rnorm(3, 10000)
#' )
#' tab <- tt(dat)
#' format_tt(tab,
#'   digits = 2,
#'   num_mark_dec = ",",
#'   num_mark_big = " "
#' )
#'
#' k <- tt(data.frame(x = c(0.000123456789, 12.4356789)))
#' format_tt(k, digits = 2, num_fmt = "significant_cell")
#'
#' dat <- data.frame(
#'   a = c("Burger", "Halloumi", "Tofu", "Beans"),
#'   b = c(1.43202, 201.399, 0.146188, 0.0031),
#'   c = c(98938272783457, 7288839482, 29111727, 93945)
#' )
#' tt(dat) |>
#'   format_tt(j = "a", sprintf = "Food: %s") |>
#'   format_tt(j = 2, digits = 1, num_fmt = "decimal", num_zero = TRUE) |>
#'   format_tt(j = "c", digits = 2, num_suffix = TRUE)
#'
#' y <- tt(data.frame(x = c(123456789.678, 12435.6789)))
#' format_tt(y, digits = 3, num_mark_big = " ")
#'
#' x <- tt(data.frame(Text = c("_italicized text_", "__bold text__")))
#' format_tt(x, markdown = TRUE)
#'
#' # Line breaks using linebreak argument
#' d <- data.frame(Text = "First line<br>Second line")
#' tt(d) |> format_tt(linebreak = "<br>")
#'
#' # Non-standard evaluation (NSE)
#' dat <- data.frame(
#'   w = c(143002.2092, 201399.181, 100188.3883),
#'   x = c(1.43402, 201.399, 0.134588),
#'   y = as.Date(c(897, 232, 198), origin = "1970-01-01"),
#'   z = c(TRUE, TRUE, FALSE)
#' )
#' tt(dat) |>
#'   format_tt(i = w > 150000, j = w, digits = 0, num_mark_big = ",")
#'
#' tab <- data.frame(a = c(NA, 1, 2), b = c(3, NA, 5))
#' tt(tab) |> format_tt(replace = "-")
#'
#' dat <- data.frame(
#'   "LaTeX" = c("Dollars $", "Percent %", "Underscore _"),
#'   "HTML" = c("<br>", "<sup>4</sup>", "<emph>blah</emph>")
#' )
#' tt(dat) |> format_tt(escape = TRUE)
#'
format_tt <- function(
    x,
    i = NULL,
    j = NULL,
    digits = get_option("tinytable_format_digits", default = NULL),
    num_fmt = get_option("tinytable_format_num_fmt", default = "significant"),
    num_zero = get_option("tinytable_format_num_zero", default = FALSE),
    num_suffix = get_option("tinytable_format_num_suffix", default = FALSE),
    num_mark_big = get_option("tinytable_format_num_mark_big", default = ""),
    num_mark_dec = get_option(
      "tinytable_format_num_mark_dec",
      default = getOption("OutDec", default = ".")
    ),
    date = get_option("tinytable_format_date", default = NULL),
    bool = get_option("tinytable_format_bool", default = NULL),
    math = get_option("tinytable_format_math", default = FALSE),
    other = get_option("tinytable_format_other", default = NULL),
    replace = get_option("tinytable_format_replace", default = FALSE),
    escape = get_option("tinytable_format_escape", default = FALSE),
    markdown = get_option("tinytable_format_markdown", default = FALSE),
    quarto = get_option("tinytable_format_quarto", default = FALSE),
    fn = get_option("tinytable_format_fn", default = NULL),
    sprintf = get_option("tinytable_format_sprintf", default = NULL),
    linebreak = get_option("tinytable_format_linebreak", default = NULL),
    output = get_option("tinytable_format_output", default = NULL)) {
  assert_integerish(digits, len = 1, null.ok = TRUE)
  assert_choice(
    num_fmt,
    c("significant", "significant_cell", "decimal", "scientific")
  )
  assert_flag(num_zero)
  assert_flag(num_suffix)
  assert_string(num_mark_big)
  assert_string(num_mark_dec)
  assert_string(date, null.ok = TRUE)
  assert_function(bool, null.ok = TRUE)
  assert_flag(math)
  assert_function(other, null.ok = TRUE)
  assert_flag(markdown)
  assert_flag(quarto)
  assert_function(fn, null.ok = TRUE)
  assert_subset(output, c("latex", "html", "typst", "markdown"), null.ok = TRUE)
  assert_string(sprintf, null.ok = TRUE)
  assert_string(linebreak, null.ok = TRUE)
  replace <- sanitize_replace(replace)
  sanity_num_mark(digits, num_mark_big, num_mark_dec)

  out <- x

  # non-standard evaluation before anything else
  tmp <- nse_i_j(x, i_expr = substitute(i), j_expr = substitute(j), pf = parent.frame())
  list2env(tmp, environment())

  # single argument list shared by the lazy call and the direct invocation
  args <- list(
    i = i,
    j = j,
    digits = digits,
    num_fmt = num_fmt,
    num_zero = num_zero,
    num_suffix = num_suffix,
    num_mark_big = num_mark_big,
    num_mark_dec = num_mark_dec,
    replace = replace,
    fn = fn,
    sprintf = sprintf,
    date_format = date,
    bool = bool,
    math = math,
    escape = escape,
    markdown = markdown,
    quarto = quarto,
    other = other,
    linebreak = linebreak,
    output = output
  )

  if (inherits(out, "tinytable")) {
    cal <- as.call(c(list(quote(format_tt_lazy)), args))
    attr(cal, "output") <- output
    out@lazy_format <- c(out@lazy_format, list(cal))
  } else {
    out <- do.call(format_tt_lazy, c(list(x = out), args))
  }

  return(out)
}


format_tt_lazy <- function(
    x,
    i,
    j,
    digits,
    num_fmt,
    num_zero,
    num_suffix,
    num_mark_big,
    num_mark_dec,
    replace,
    fn,
    sprintf,
    date_format,
    bool,
    math,
    escape,
    markdown,
    quarto,
    other,
    linebreak,
    output) {
  # Row indices derived from an NSE predicate were computed eagerly in
  # format_tt() against the pre-group data rows. Although the @lazy_format
  # loop runs before rbind_body_groupi(), apply_format() interprets numeric
  # `i` in final-table coordinates (it matches `x@index_body %in% i`), so
  # predicate-derived indices must be remapped here as well. @index_body is
  # already available: build_tt() computes it before the lazy_format loop.
  if (inherits(x, "tinytable")) {
    i <- nse_remap_i(i, x)
  }

  if (inherits(x, "tbl_df")) {
    assert_dependency("tibble")
    x_is_tibble <- TRUE
    x <- as.data.frame(x, check.names = FALSE)
  } else {
    x_is_tibble <- FALSE
  }

  # important for tabulator
  if (!is.null(bool) && inherits(x, "tinytable")) {
    x@tabulator_format_bool <- TRUE
  }

  # Check if i contains component names (do this before processing tinytable objects)
  tmp <- resolve_i_components(x, i, j, default = "all")
  i <- tmp$i
  components <- tmp$components

  # cell-targeting components only: the type-based formatters (logical, date,
  # numeric, other) must not reformat body cells when `i` targets a component
  # such as "caption" or "colnames"
  components_cells <- intersect(
    if ("all" %in% components) c("cells", "groupi", "~groupi") else components,
    c("cells", "groupi", "~groupi")
  )

  # format_tt() supports vectors
  if (isTRUE(check_atomic_vector(x))) {
    atomic_vector <- TRUE
    if (is.factor(x)) {
      x <- as.character(x)
    }
    x <- data.frame(tinytable = x, stringsAsFactors = FALSE)
    j <- 1
  } else if (is.data.frame(x)) {
    atomic_vector <- FALSE
  } else if (inherits(x, "tinytable")) {
    atomic_vector <- FALSE
  } else {
    stop(
      "`x` must be a `tinytable` object, a data frame, or an atomic vector.",
      call. = FALSE
    )
  }

  # data frames and vectors do not carry their raw data the way tinytable
  # objects do in @data; keep a copy so `replace` can match typed original
  # values even after digits/date formatting stringified the working copy
  original_input <- if (!inherits(x, "tinytable")) x else NULL
  output_format <- if (inherits(x, "tinytable")) x@output else NULL

  # In sanity_tt(), we fill in missing NULL `j` in the format-specific versions,
  # because tabularray can do whole column styling. Here, we need to fill in
  # NULL for all formats since this is applied before creating the table.
  # nrow(out) because nrow(x) sometimes includes rows that will be added **in the lazy future** by group_tt()
  i <- sanitize_i(i, x, lazy = FALSE, calling_function = "format_tt")
  j <- sanitize_j(j, x, skip_tabulator_types = TRUE)

  x <- apply_format(
    x = x,
    i = i,
    j = j,
    components = components_cells,
    format_fn = format_vector_logical,
    inherit_class = "logical",
    bool_fn = bool
  )

  x <- apply_format(
    x = x,
    i = i,
    j = j,
    components = components_cells,
    format_fn = format_vector_date,
    inherit_class = "Date",
    date_format = date_format
  )

  x <- apply_format(
    x = x,
    i = i,
    j = j,
    components = components_cells,
    format_fn = format_vector_numeric,
    num_suffix = num_suffix,
    digits = digits,
    num_mark_big = num_mark_big,
    num_mark_dec = num_mark_dec,
    num_zero = num_zero,
    num_fmt = num_fmt,
    inherit_class = is.numeric
  )

  is_other <- function(x) {
    !is.numeric(x) && !inherits(x, "Date") && !is.logical(x)
  }
  x <- apply_format(
    x = x,
    i = i,
    j = j,
    components = components_cells,
    format_fn = format_vector_other,
    inherit_class = is_other,
    other_fn = other
  )

  # after other formatting
  if (!is.null(sprintf)) {
    x <- apply_format(
      x = x,
      i = i,
      j = j,
      components = components,
      format_fn = format_vector_sprintf,
      original_data = FALSE,
      sprintf_pattern = sprintf
    )
  }

  # Custom functions overwrite all the other formatting, but is before markdown
  # before escaping
  if (is.function(fn)) {
    x <- apply_format(
      x = x,
      i = i,
      j = j,
      format_fn = format_vector_custom,
      components = components,
      fn = fn
    )
  }

  # close to last
  if (isTRUE(math)) {
    x <- apply_format(
      x = x,
      i = i,
      j = j,
      format_fn = format_vector_math,
      components = components,
      original_data = FALSE,
      math = math
    )
  }

  # linebreak before replace and escape
  if (!is.null(linebreak)) {
    x <- apply_format(
      x = x,
      i = i,
      j = j,
      format_fn = format_vector_linebreak,
      components = components,
      original_data = FALSE,
      linebreak = linebreak,
      output = output_format
    )
  }

  # replace before escape, otherwise overaggressive removal
  x <- apply_format(
    x = x,
    i = i,
    j = j,
    format_fn = format_vector_replace,
    components = components,
    replace = replace,
    original_data = FALSE,
    original = original_input
  )

  # escape latex characters
  if (!isFALSE(escape)) {
    if (isTRUE(escape == "latex")) {
      o <- "latex"
    } else if (isTRUE(escape == "html")) {
      o <- "html"
    } else if (isTRUE(escape == "typst")) {
      o <- "typst"
    } else if (inherits(x, "tinytable")) {
      o <- x@output
    } else {
      o <- FALSE
    }

    x <- apply_format(
      x = x,
      i = i,
      j = j,
      format_fn = format_vector_escape,
      components = components,
      original_data = FALSE,
      output = o
    )
  }

  # markdown and quarto at the very end
  if (isTRUE(markdown)) {
    assert_dependency("litedown")
    x <- apply_format(
      x = x,
      i = i,
      j = j,
      format_fn = format_vector_markdown,
      components = components,
      original_data = FALSE,
      output_format = output_format
    )
  }

  # quarto processing needs a rendered table string; no-op on plain data
  # frames and vectors
  if (isTRUE(quarto) && inherits(x, "tinytable")) {
    for (col in j) {
      x <- format_vector_quarto(i = i, col = col, x = x)
    }
  }

  # output
  if (isTRUE(atomic_vector)) {
    x <- x[[1]]
  }

  if (!inherits(x, "tinytable") && x_is_tibble) {
    x <- tibble::as_tibble(x)
  }

  return(x)
}


# Resolve the `i` argument into row indices and a vector of component names.
# `i` may be a numeric vector of rows, "groupi"/"~groupi" (row group labels or
# non-group rows), or a character vector of component names such as "caption",
# "colnames", "notes", or "groupj". Shared by format_tt_lazy() and
# rotate_cells_setup(); `default` is used when neither `i` nor `j` is supplied.
resolve_i_components <- function(x, i, j, default = "all") {
  if (identical(i, "groupi")) {
    components <- "cells"
    i <- x@group_index_i
  } else if (identical(i, "~groupi")) {
    components <- "cells"
    i <- setdiff(seq_len(nrow(x)), x@group_index_i)
  } else if (is.character(i)) {
    components <- i # before wiping i
    i <- NULL
  } else if (!is.null(i) || !is.null(j)) {
    components <- "cells"
  } else {
    components <- default
  }
  list(i = i, components = components)
}

source("helpers.R")
options(tinytable_html_engine = "tabulator")
options(tinytable_tt_theme = theme_html)

# Basic tabulator table
dat <- head(iris)
tab <- tt(dat)
out <- save_tt(tab, "html")
expect_true(is.character(out))
expect_true(grepl("tabulator", out))
expect_true(grepl("Sepal.Length", out))

# Preserve special characters in column names
special_dat <- data.frame(`hello\\?` = 1, check.names = FALSE)
tab <- tt(special_dat)
out <- save_tt(tab, "html")
expect_true(grepl('"hello_\\?"', out))
expect_false(grepl('"hello__"', out))

# Data with multiple types
dat <- data.frame(
  city = c("Montréal", "Toronto", "Vancouver"),
  salary = c(14002.22, 201399.11, 80188.38),
  random = c(1.43402, 201.399, 0.134588),
  date = as.Date(sample(1:1000, 3), origin = "1970-01-01"),
  best = c(TRUE, FALSE, FALSE)
)

tab <- tt(dat)
out <- save_tt(tab, "html")
expect_true(is.character(out))
expect_true(grepl("Montréal", out))
expect_true(grepl("true", out, ignore.case = TRUE))

# Pagination and filtering
tab <- tt(iris) |> theme_html(
  tabulator_pagination = c(5, 10, 50),
  tabulator_search = "top"
)
out <- save_tt(tab, "html")
expect_true(grepl("pagination", out))
expect_true(grepl("search", out))
expect_true(grepl("paginationSize.*5", out))

# Format numeric columns
tab <- tt(dat) |>
  format_tt(j = "salary", digits = 2, num_mark_big = ",") |>
  format_tt(j = "random", digits = 4)
out <- save_tt(tab, "html")
expect_true(grepl("decimal", out))
expect_true(grepl("salary.*decimal", out))

# Format date columns
tab <- tt(dat) |> format_tt(j = "date", date = "M/d/yyyy")
out <- save_tt(tab, "html")
expect_true(grepl("M/d/yyyy", out))
expect_true(grepl("datetime", out))

# Simple data table for remaining tests
simple_dat <- data.frame(
  name = c("Alice", "Bob", "Charlie"),
  value = c(100, 200, 300)
)

tab <- tt(simple_dat)
out <- save_tt(tab, "html")
expect_true(grepl("Alice", out))

# Style alignment
tab <- tt(simple_dat) |> style_tt(align = "r")
out <- save_tt(tab, "html")
expect_true(grepl("right", out))

# Custom CSS rules
css_rule <- "
$TINYTABLE_ID .tabulator-header .tabulator-col {
  background-color: black;
  color: white;
}
"
tab <- tt(simple_dat) |> theme_html(tabulator_css_rule = css_rule)
out <- save_tt(tab, "html")
expect_true(grepl("background-color.*black", out))
expect_true(grepl("color.*white", out))

# Custom columns
dat_custom <- data.frame(
  city = c("Toronto", "Montreal", "Vancouver"),
  salary = c(75000, 68000, 82000),
  best = c(FALSE, TRUE, FALSE)
)

custom_columns <- '[
  {
    "title": "City",
    "field": "city"
  },
  {
    "title": "Best city",
    "field": "best",
    "formatter": "tickCross"
  },
  {
    "title": "Salary",
    "field": "salary",
    "formatter": "money",
    "formatterParams": {"precision": 0, "symbol": "$"}
  }
]'

tab <- tt(dat_custom) |> theme_html(tabulator_columns = custom_columns)
out <- save_tt(tab, "html")
expect_true(grepl("tickCross", out))
expect_true(grepl("money", out))
expect_true(grepl("Best city", out))

# Custom options
opts <- "
  layout: 'fitColumns',
  height: '200px'
"

tab <- tt(dat_custom) |> theme_html(tabulator_options = opts)
out <- save_tt(tab, "html")
expect_true(grepl("fitColumns", out))
expect_true(grepl("200px", out))

# Reset global option
options(tinytable_html_engine = NULL)

source("helpers.R")
using("tinysnapshot")
options(tinytable_print_output = "latex")

if (!is_local) {
  exit_file("Run on Vincent's machine")
}

x <- mtcars[1:4, 1:5]

expect_snapshot_print(
  tt(x),
  label = "latex-default.tex"
)

k <- x
colnames(k) <- NULL
expect_snapshot_print(
  tt(k),
  label = "latex-nohead.tex"
)

# Issue #613: caption with tabular latex theme
expect_snapshot_print(
  tt(x, caption = "Tabular caption example.") |>
    theme_latex(environment = "tabular"),
  label = "latex-tabular_caption.tex"
)

# Align
expect_snapshot_print(
  tt(x) |> style_tt(j = 1:5, align = "ccllr"),
  label = "latex-align.tex"
)

# Themes
expect_snapshot_print(
  tt(x, theme = "striped"),
  label = "latex-theme_striped.tex"
)

expect_snapshot_print(
  tt(x, theme = "grid"),
  label = "latex-theme_grid.tex"
)

expect_snapshot_print(
  tt(x, theme = "empty"),
  label = "latex-theme_empty.tex"
)

# Styles
expect_snapshot_print(
  tt(x) |> style_tt(i = 1:4, color = "orange"),
  label = "latex-row_color.tex"
)

expect_snapshot_print(
  tt(x) |> style_tt(j = 1:4, color = "orange"),
  label = "latex-col_color.tex"
)

expect_snapshot_print(
  tt(x) |> style_tt(i = 1:2, j = 1:4, color = "orange"),
  label = "latex-cell_color.tex"
)

# Combined font styling in LaTeX
font_dat <- data.frame(
  col_a = c("a", "b", "c"),
  col_b = c("d", "e", "f")
)
expect_snapshot_print(
  tt(font_dat) |>
    style_tt(i = 0, bold = TRUE, fontsize = 0.85),
  label = "latex-font_combined.tex"
)

# Lazy style: group after style is respected
a <- tt(mtcars[1:4, 1:4]) |>
  style_tt(color = "orange", background = "black") |>
  group_tt(j = list("blah" = 1:2, "bar" = 3:4))
b <- tt(mtcars[1:4, 1:4]) |>
  group_tt(j = list("blah" = 1:2, "bar" = 3:4)) |>
  style_tt(color = "orange", background = "black")
expect_snapshot_print(a, label = "latex-group_style_order.tex")
expect_equal(as.character(a@table_string), as.character(b@table_string))

# align d
x <- data.frame(pi = c(pi * 100, pi * 1000, pi * 10000, pi * 100000))
tab <- tt(x) |>
  format_tt(j = 1, digits = 8, num_fmt = "significant_cell") |>
  style_tt(j = 1, align = "d")
expect_snapshot_print(tab, label = "latex-align_d.tex")

dat = data.frame(a = c("(03.1)", "(3.14)**", "(003.1416)+"))
tt(dat) |> style_tt(align = "d")
expect_snapshot_print(tab, label = "latex-align_d_02.tex")

dat = data.frame(a = c("(03.1)", "(3.14)**", "(003.1416)+"))
tt(dat) |>
  group_tt(j = list("blah" = 1)) |>
  style_tt(align = "d")
expect_snapshot_print(tab, label = "latex-align_d_03.tex")

# bug discovered with vignette
x <- tt(mtcars[1:9, 1:8]) |>
  group_tt(
    i = list(
      "I like (fake) hamburgers" = 3,
      "She prefers halloumi" = 4,
      "They love tofu" = 7
    )
  ) |>
  style_tt(
    i = c(3, 5, 9),
    align = "c",
    color = "white",
    background = "gray",
    bold = TRUE
  ) |>
  save_tt("latex")
expect_inherits(x, "character")

# Footnotes
expect_snapshot_print(
  (tt(mtcars[1:4, 1:5], notes = list(a = "Blah.", b = "Blah blah."))),
  label = "latex-footnotes.tex"
)

# Maths
math <- tt(
  data.frame(
    Math = c(
      "$x^2 + y^2 = z^2$",
      "$\\frac{1}{2}$"
    )
  )
) |>
  style_tt(j = 1, align = "c")
expect_snapshot_print(
  math,
  label = "latex-maths.tex"
)

# Line breaks
expect_snapshot_print(
  (tt(
    data.frame(
      "{Sed ut \\\\ perspiciatis unde}",
      "dicta sunt<br> explicabo. Nemo"
    ) |>
      setNames(c("LaTeX line break", "HTML line break")),
    width = 1
  )),
  label = "latex-breaks.tex"
)

# Formatting
dat <- data.frame(
  a = c("Burger", "Halloumi", "Tofu", "Beans"),
  b = c(1.43202, 201.399, 0.146188, 0.0031),
  c = c(98938272783457, 7288839482, 29111727, 93945)
)
expect_snapshot_print(
  (tt(dat) |>
    format_tt(j = "a", sprintf = "Food: %s") |>
    format_tt(j = 2, digits = 1) |>
    format_tt(j = "c", digits = 2, num_suffix = TRUE)),
  label = "latex-formatting.tex"
)

# placement
tab <- mtcars[1:3, 1:3]
tab <- tt(tab) |> theme_latex(placement = "H")
tab@output <- "latex"
expect_snapshot_print(tab, label = "latex-placement.tex")

# Missing value replacement
tab <- data.frame(a = c(NA, 1, 2), b = c(3, NA, 5))
expect_snapshot_print(
  tt(tab) |> format_tt(replace = "-"),
  label = "latex-missing_value_replacement.tex"
)

# Escape special characters
dat <- data.frame(
  "LaTeX" = c("Dollars $", "Percent %", "Underscore _"),
  "HTML" = c("<br>", "<sup>4</sup>", "<emph>blah</emph>")
)
expect_snapshot_print(
  tt(dat) |> format_tt(escape = TRUE),
  label = "latex-escape_special_caracters.tex"
)

# Formatting URLs
dat <- data.frame(
  `Package (link)` = c(
    "[`marginaleffects`](https://www.marginaleffects.com/)",
    "[`modelsummary`](https://www.modelsummary.com/)"
  ),
  Purpose = c(
    "Interpreting statistical models",
    "Data and model summaries"
  ),
  check.names = FALSE
)
dat <- tt(dat) |> format_tt(j = 1, markdown = TRUE)
expect_snapshot_print(
  dat,
  label = "latex-formatting_url.tex"
)

# Style
x <- mtcars[1:4, 1:5]
expect_snapshot_print(
  (tt(x) |>
    style_tt(
      i = 2:3,
      j = c(1, 3, 4),
      italic = TRUE,
      background = "green",
      color = "orange"
    )),
  label = "latex-style.tex"
)

# Font size
expect_snapshot_print(
  (tt(x) |> style_tt(j = "mpg|hp|qsec", fontsize = 1.5)),
  label = "latex-font_size.tex"
)

# Merging cells
expect_snapshot_print(
  (tt(x) |>
    style_tt(
      i = 2,
      j = 2,
      colspan = 3,
      rowspan = 2,
      align = "c",
      alignv = "m",
      color = "white",
      background = "black",
      bold = TRUE
    )),
  label = "latex-merging_cells.tex"
)

# Spanning cells
tab <- aggregate(mpg ~ cyl + am, FUN = mean, data = mtcars)
tab <- tab[order(tab$cyl, tab$am), ]
expect_snapshot_print(
  (tt(tab, digits = 2) |>
    style_tt(i = c(1, 3, 5), j = 1, rowspan = 2, alignv = "t")),
  label = "latex-spanning_cells.tex"
)

# Conditional styling
k <- mtcars[1:10, c("mpg", "am", "vs")]
expect_snapshot_print(
  (tt(k) |>
    style_tt(
      i = which(k$am == k$vs),
      background = "teal",
      color = "white"
    )),
  label = "latex-conditional_styling.tex"
)

# Heatmaps
fs <- seq(.1, 2, length.out = 20)
k <- data.frame(matrix(fs, ncol = 5))
colnames(k) <- NULL
bg <- hcl.colors(20, "Inferno")
fg <- ifelse(as.matrix(k) < 1.7, tail(bg, 1), head(bg, 1))
expect_snapshot_print(
  (tt(k, width = .7, theme = "empty") |>
    style_tt(j = 1:5, align = "ccccc") |>
    style_tt(
      i = 1:4,
      j = 1:5,
      color = fg,
      background = bg,
      fontsize = fs
    )),
  label = "latex-heatmaps.tex"
)

# NA values in background styling
if (Sys.info()["sysname"] == "Darwin") {
  set.seed(48103)
  bg <- sample(c(NA, "green", "orange"), 6 * 5, replace = TRUE)
  expect_snapshot_print(
    tt(head(iris)) |>
      style_tt(i = 1:6, j = 1:5, background = bg),
    label = "latex-background_na_values.tex"
  )
}

# Borders
expect_snapshot_print(
  tt(x, theme = "empty") |>
    style_tt(
      i = 0:3,
      j = 1:3,
      line = "tblr",
      line_width = 0.4,
      line_color = "orange"
    ),
  label = "latex-borders.tex"
)

# Issue #242: multiple notes in tabularray are fiddly
tab <- tt(
  mtcars[1:4, 1:4],
  notes = list(a = "blah", b = "hello world", "oh yeah", "works?")
)
tab@output <- "latex"
expect_snapshot_print(tab, label = "latex-issue242.tex")

# Issue #306
x <- data.frame(x = 1:5)
x <- data.frame(x = 1:5)
colnames(x) <- NULL
tab <- tt(x) |>
  format_tt() |>
  save_tt("latex")
expect_inherits(tab, "character")

# Issue #307
tab <- tt(head(iris)) |>
  group_tt(j = list("blah" = 1:2)) |>
  theme_latex(environment = "tabular", environment_table = FALSE)
expect_snapshot_print(tab, label = "latex-issue307.tex")

# Issue #419
tab <- tt(head(iris)) |>
  theme_latex(
    outer = "label={tbl-species}",
    inner = "rowsep=0pt"
  )
expect_snapshot_print(tab, label = "latex-issue419.tex")

# format_tt math
dat <- data.frame("y^2 = e^x" = c(-2, -pi), check.names = FALSE)
tab <- tt(dat, digits = 3) |> format_tt(math = TRUE)
expect_snapshot_print(tab, label = "latex-format_tt_math.tex")


# Issue #493: theme_multipage does not convert to longtblr
dat <- data.frame(a = rnorm(100), b = rnorm(100), c = rnorm(100))
tab <- tt(dat) |>
  theme_latex(multipage = TRUE, rowhead = 1) |>
  save_tt("latex")
matches <- lengths(gregexpr("longtblr", tab, fixed = TRUE))
expect_equal(matches, 2)


# Issue #535: environment_table = FALSE without an environment argument
tab1 <- head(mtcars, 1) |>
  tt() |>
  theme_latex(environment_table = FALSE) |>
  save_tt("latex")
table_matches <- gregexpr("\\\\begin\\{table\\}", tab1)[[1]]
expect_equal(ifelse(table_matches[1] == -1, 0, length(table_matches)), 0)
expect_equal(lengths(gregexpr("\\{tblr", tab1)), 2)

tab2 <- head(mtcars, 1) |>
  tt() |>
  theme_latex(environment = "longtblr") |>
  save_tt("latex")
table_matches2 <- gregexpr("\\\\begin\\{table\\}", tab2)[[1]]
expect_equal(ifelse(table_matches2[1] == -1, 0, length(table_matches2)), 0)
expect_equal(lengths(gregexpr("longtblr", tab2)), 2)

tab3 <- head(mtcars, 1) |>
  tt() |>
  theme_latex(environment = "talltblr") |>
  save_tt("latex")
expect_equal(lengths(gregexpr("\\{table", tab3)), 2)
expect_equal(lengths(gregexpr("\\{talltblr", tab3)), 2)



options(tinytable_print_output = NULL)

source("helpers.R")
using("tinysnapshot")

options(tinytable_print_output = "markdown")
x <- mtcars[1:4, 1:6]
tab <- tt(x) |>
  format_tt(j = "drat|wt", num_fmt = "decimal", digits = 4, num_zero = TRUE)
expect_snapshot_print(tab, "i_j-sanitize_j_equivalence")

tab <- tt(x) |>
  format_tt(j = c("drat", "wt"), num_fmt = "decimal", num_zero = TRUE, digits = 4)
expect_snapshot_print(tab, "i_j-sanitize_j_equivalence")

tab <- tt(x) |>
  format_tt(j = c(5:6), num_fmt = "decimal", num_zero = TRUE, digits = 4)
expect_snapshot_print(tab, "i_j-sanitize_j_equivalence")

tab <- tt(x) |>
  format_tt(i = 1:2, j = c(5:6), num_fmt = "decimal", digits = 4) |>
  format_tt(i = 3:4, j = c(5:6), num_fmt = "decimal", digits = 2)
expect_snapshot_print(tab, "i_j-format_tt_i_01")

cormat <- data.frame(cor(mtcars[1:5]))
tab <- tt(cormat, digits = 2) |>
  style_tt(i = abs(cormat) > .8, background = "black", color = "white")
expect_equal(dim(tab), c(5, 5))


expect_error(tt(head(iris)) |> style_tt(j = "BadColumn") |> print(),
  pattern = "No columns match the pattern")

options(tinytable_print_output = NULL)

source("helpers.R")
using("tinysnapshot")
options(tinytable_print_output = "latex")

if (!is_local) {
  exit_file("Run on Vincent's machine")
}

x <- mtcars[1:4, 1:5]

expect_snapshot_print(
  tt(x),
  label = "latex-default.tex"
)

k <- x
colnames(k) <- NULL
expect_snapshot_print(
  tt(k),
  label = "latex-nohead.tex"
)

# Issue #613: caption with tabular latex theme
expect_snapshot_print(
  tt(x, caption = "Tabular caption example.") |>
    theme_latex(environment = "tabular"),
  label = "latex-tabular_caption.tex"
)

# Align
expect_snapshot_print(
  tt(x) |> style_tt(j = 1:5, align = "ccllr"),
  label = "latex-align.tex"
)

# Themes
expect_snapshot_print(
  tt(x, theme = "striped"),
  label = "latex-theme_striped.tex"
)

expect_snapshot_print(
  tt(x, theme = "grid"),
  label = "latex-theme_grid.tex"
)

expect_snapshot_print(
  tt(x, theme = "empty"),
  label = "latex-theme_empty.tex"
)

# Styles
expect_snapshot_print(
  tt(x) |> style_tt(i = 1:4, color = "orange"),
  label = "latex-row_color.tex"
)

expect_snapshot_print(
  tt(x) |> style_tt(j = 1:4, color = "orange"),
  label = "latex-col_color.tex"
)

expect_snapshot_print(
  tt(x) |> style_tt(i = 1:2, j = 1:4, color = "orange"),
  label = "latex-cell_color.tex"
)

# Combined font styling in LaTeX
font_dat <- data.frame(
  col_a = c("a", "b", "c"),
  col_b = c("d", "e", "f")
)
expect_snapshot_print(
  tt(font_dat) |>
    style_tt(i = 0, bold = TRUE, fontsize = 0.85),
  label = "latex-font_combined.tex"
)

# Lazy style: group after style is respected
a <- tt(mtcars[1:4, 1:4]) |>
  style_tt(color = "orange", background = "black") |>
  group_tt(j = list("blah" = 1:2, "bar" = 3:4))
b <- tt(mtcars[1:4, 1:4]) |>
  group_tt(j = list("blah" = 1:2, "bar" = 3:4)) |>
  style_tt(color = "orange", background = "black")
expect_snapshot_print(a, label = "latex-group_style_order.tex")
expect_equal(as.character(a@table_string), as.character(b@table_string))

# align d
x <- data.frame(pi = c(pi * 100, pi * 1000, pi * 10000, pi * 100000))
tab <- tt(x) |>
  format_tt(j = 1, digits = 8, num_fmt = "significant_cell") |>
  style_tt(j = 1, align = "d")
expect_snapshot_print(tab, label = "latex-align_d.tex")

dat = data.frame(a = c("(03.1)", "(3.14)**", "(003.1416)+"))
tt(dat) |> style_tt(align = "d")
expect_snapshot_print(tab, label = "latex-align_d_02.tex")

dat = data.frame(a = c("(03.1)", "(3.14)**", "(003.1416)+"))
tt(dat) |>
  group_tt(j = list("blah" = 1)) |>
  style_tt(align = "d")
expect_snapshot_print(tab, label = "latex-align_d_03.tex")

# align d: comparators like `<0.001` need reserved width, otherwise siunitx
# typesets them outside the column
dat <- data.frame(p = c("<0.001", "0.116", "-0.506"))
tab <- tt(dat) |> style_tt(j = 1, align = "d")
expect_false(grepl("table-align-text-before", save_tt(tab, "latex"), fixed = TRUE))
expect_snapshot_print(tab, label = "latex-align_d_comparator.tex")

# bug discovered with vignette
x <- tt(mtcars[1:9, 1:8]) |>
  group_tt(
    i = list(
      "I like (fake) hamburgers" = 3,
      "She prefers halloumi" = 4,
      "They love tofu" = 7
    )
  ) |>
  style_tt(
    i = c(3, 5, 9),
    align = "c",
    color = "white",
    background = "gray",
    bold = TRUE
  ) |>
  save_tt("latex")
expect_inherits(x, "character")

# Footnotes
expect_snapshot_print(
  (tt(mtcars[1:4, 1:5], notes = list(a = "Blah.", b = "Blah blah."))),
  label = "latex-footnotes.tex"
)

# Maths
math <- tt(
  data.frame(
    Math = c(
      "$x^2 + y^2 = z^2$",
      "$\\frac{1}{2}$"
    )
  )
) |>
  style_tt(j = 1, align = "c")
expect_snapshot_print(
  math,
  label = "latex-maths.tex"
)

# Line breaks
expect_snapshot_print(
  (tt(
    data.frame(
      "{Sed ut \\\\ perspiciatis unde}",
      "dicta sunt<br> explicabo. Nemo"
    ) |>
      setNames(c("LaTeX line break", "HTML line break")),
    width = 1
  )),
  label = "latex-breaks.tex"
)

# Formatting
dat <- data.frame(
  a = c("Burger", "Halloumi", "Tofu", "Beans"),
  b = c(1.43202, 201.399, 0.146188, 0.0031),
  c = c(98938272783457, 7288839482, 29111727, 93945)
)
expect_snapshot_print(
  (tt(dat) |>
    format_tt(j = "a", sprintf = "Food: %s") |>
    format_tt(j = 2, digits = 1) |>
    format_tt(j = "c", digits = 2, num_suffix = TRUE)),
  label = "latex-formatting.tex"
)

# placement
tab <- mtcars[1:3, 1:3]
tab <- tt(tab) |> theme_latex(placement = "H")
tab@output <- "latex"
expect_snapshot_print(tab, label = "latex-placement.tex")

# Missing value replacement
tab <- data.frame(a = c(NA, 1, 2), b = c(3, NA, 5))
expect_snapshot_print(
  tt(tab) |> format_tt(replace = "-"),
  label = "latex-missing_value_replacement.tex"
)

# Escape special characters
dat <- data.frame(
  "LaTeX" = c("Dollars $", "Percent %", "Underscore _"),
  "HTML" = c("<br>", "<sup>4</sup>", "<emph>blah</emph>")
)
expect_snapshot_print(
  tt(dat) |> format_tt(escape = TRUE),
  label = "latex-escape_special_caracters.tex"
)

# Formatting URLs
dat <- data.frame(
  `Package (link)` = c(
    "[`marginaleffects`](https://www.marginaleffects.com/)",
    "[`modelsummary`](https://www.modelsummary.com/)"
  ),
  Purpose = c(
    "Interpreting statistical models",
    "Data and model summaries"
  ),
  check.names = FALSE
)
dat <- tt(dat) |> format_tt(j = 1, markdown = TRUE)
expect_snapshot_print(
  dat,
  label = "latex-formatting_url.tex"
)

# Style
x <- mtcars[1:4, 1:5]
expect_snapshot_print(
  (tt(x) |>
    style_tt(
      i = 2:3,
      j = c(1, 3, 4),
      italic = TRUE,
      background = "green",
      color = "orange"
    )),
  label = "latex-style.tex"
)

# Font size
expect_snapshot_print(
  (tt(x) |> style_tt(j = "mpg|hp|qsec", fontsize = 1.5)),
  label = "latex-font_size.tex"
)

# Merging cells
expect_snapshot_print(
  (tt(x) |>
    style_tt(
      i = 2,
      j = 2,
      colspan = 3,
      rowspan = 2,
      align = "c",
      alignv = "m",
      color = "white",
      background = "black",
      bold = TRUE
    )),
  label = "latex-merging_cells.tex"
)

# Spanning cells
tab <- aggregate(mpg ~ cyl + am, FUN = mean, data = mtcars)
tab <- tab[order(tab$cyl, tab$am), ]
expect_snapshot_print(
  (tt(tab, digits = 2) |>
    style_tt(i = c(1, 3, 5), j = 1, rowspan = 2, alignv = "t")),
  label = "latex-spanning_cells.tex"
)

# Conditional styling
k <- mtcars[1:10, c("mpg", "am", "vs")]
expect_snapshot_print(
  (tt(k) |>
    style_tt(
      i = which(k$am == k$vs),
      background = "teal",
      color = "white"
    )),
  label = "latex-conditional_styling.tex"
)

# Heatmaps
fs <- seq(.1, 2, length.out = 20)
k <- data.frame(matrix(fs, ncol = 5))
colnames(k) <- NULL
bg <- hcl.colors(20, "Inferno")
fg <- ifelse(as.matrix(k) < 1.7, tail(bg, 1), head(bg, 1))
expect_snapshot_print(
  (tt(k, width = .7, theme = "empty") |>
    style_tt(j = 1:5, align = "ccccc") |>
    style_tt(
      i = 1:4,
      j = 1:5,
      color = fg,
      background = bg,
      fontsize = fs
    )),
  label = "latex-heatmaps.tex"
)

# NA values in background styling
set.seed(48103)
bg <- sample(c(NA, "green", "orange"), 6 * 5, replace = TRUE)
expect_snapshot_print(
  tt(head(iris)) |>
    style_tt(i = 1:6, j = 1:5, background = bg),
  label = "latex-background_na_values.tex"
)

# Borders
expect_snapshot_print(
  tt(x, theme = "empty") |>
    style_tt(
      i = 0:3,
      j = 1:3,
      line = "tblr",
      line_width = 0.4,
      line_color = "orange"
    ),
  label = "latex-borders.tex"
)

# Issue #242: multiple notes in tabularray are fiddly
tab <- tt(
  mtcars[1:4, 1:4],
  notes = list(a = "blah", b = "hello world", "oh yeah", "works?")
)
tab@output <- "latex"
expect_snapshot_print(tab, label = "latex-issue242.tex")

# Issue #306
x <- data.frame(x = 1:5)
x <- data.frame(x = 1:5)
colnames(x) <- NULL
tab <- tt(x) |>
  format_tt() |>
  save_tt("latex")
expect_inherits(tab, "character")

# Issue #307
tab <- tt(head(iris)) |>
  group_tt(j = list("blah" = 1:2)) |>
  theme_latex(environment = "tabular", environment_table = FALSE)
expect_snapshot_print(tab, label = "latex-issue307.tex")

# Issue #419
tab <- tt(head(iris)) |>
  theme_latex(
    outer = "label={tbl-species}",
    inner = "rowsep=0pt"
  )
expect_snapshot_print(tab, label = "latex-issue419.tex")

# format_tt math
dat <- data.frame("y^2 = e^x" = c(-2, -pi), check.names = FALSE)
tab <- tt(dat, digits = 3) |> format_tt(math = TRUE)
expect_snapshot_print(tab, label = "latex-format_tt_math.tex")


# Issue #493: theme_multipage does not convert to longtblr
dat <- data.frame(a = rnorm(100), b = rnorm(100), c = rnorm(100))
tab <- tt(dat) |>
  theme_latex(multipage = TRUE, rowhead = 1) |>
  save_tt("latex")
matches <- lengths(gregexpr("longtblr", tab, fixed = TRUE))
expect_equal(matches, 2)


# Issue #535: environment_table = FALSE without an environment argument
tab1 <- head(mtcars, 1) |>
  tt() |>
  theme_latex(environment_table = FALSE) |>
  save_tt("latex")
table_matches <- gregexpr("\\\\begin\\{table\\}", tab1)[[1]]
expect_equal(ifelse(table_matches[1] == -1, 0, length(table_matches)), 0)
expect_equal(lengths(gregexpr("\\{tblr", tab1)), 2)

tab2 <- head(mtcars, 1) |>
  tt() |>
  theme_latex(environment = "longtblr") |>
  save_tt("latex")
table_matches2 <- gregexpr("\\\\begin\\{table\\}", tab2)[[1]]
expect_equal(ifelse(table_matches2[1] == -1, 0, length(table_matches2)), 0)
expect_equal(lengths(gregexpr("longtblr", tab2)), 2)

tab3 <- head(mtcars, 1) |>
  tt() |>
  theme_latex(environment = "talltblr") |>
  save_tt("latex")
expect_equal(lengths(gregexpr("\\{table", tab3)), 2)
expect_equal(lengths(gregexpr("\\{talltblr", tab3)), 2)

# rowhead combined with environment = "longtblr" (previously errored)
tab4 <- head(mtcars, 5) |>
  tt() |>
  theme_latex(environment = "longtblr", rowhead = 2) |>
  save_tt("latex")
expect_true(grepl("rowhead=2", tab4, fixed = TRUE))
expect_true(grepl("longtblr", tab4, fixed = TRUE))



options(tinytable_print_output = NULL)

source("helpers.R")
using("tinysnapshot")

# Examples from @examples section in rbind2.R
x <- tt(mtcars[1:3, 1:2], caption = "Combine two tiny tables.")
y <- tt(mtcars[4:5, 8:10])

# basic combination using rbind()
tab1 <- rbind(x, y)
expect_snapshot_print(tab1, label = "rbind2-basic_rbind.md")

# basic combination with format_tt
tab2 <- rbind(x, y) |> format_tt(replace = "")
expect_snapshot_print(tab2, label = "rbind2-basic_rbind_format.md")

# rbind2() with headers = FALSE (omit y header)
tab3 <- rbind2(x, y, headers = FALSE)
expect_snapshot_print(tab3, label = "rbind2-no_headers.md")

# rbind2() with use_names = FALSE (bind by position rather than column names)
tab4 <- rbind2(x, y, use_names = FALSE)
expect_snapshot_print(tab4, label = "rbind2-position_bind.md")

# Examples from vignettes (using a, b variable names like in vignette)
a <- tt(mtcars[1:3, 1:2], caption = "Combine two tiny tables.")
b <- tt(mtcars[4:5, 8:10])

# vignette rbind() example
tab5 <- rbind(a, b)
expect_snapshot_print(tab5, label = "rbind2-vignette_basic.md")

# vignette rbind() with format_tt
tab6 <- rbind(a, b) |> format_tt(replace = "")
expect_snapshot_print(tab6, label = "rbind2-vignette_format.md")

# vignette rbind2() with headers = FALSE
tab7 <- rbind2(a, b, headers = FALSE)
expect_snapshot_print(tab7, label = "rbind2-vignette_no_headers.md")

# vignette rbind2() with use_names = FALSE
tab8 <- rbind2(a, b, use_names = FALSE)
expect_snapshot_print(tab8, label = "rbind2-vignette_position.md")

# Test with identical column names
c <- tt(mtcars[1:2, 1:3])
d <- tt(mtcars[6:7, 1:3])
tab9 <- rbind2(c, d)
expect_snapshot_print(tab9, label = "rbind2-identical_columns.md")

# Test with notes carrying over
e <- tt(mtcars[1:2, 1:2], notes = "First table note")
f <- tt(mtcars[3:4, 1:2], notes = "Second table note")
tab10 <- rbind2(e, f)
expect_snapshot_print(tab10, label = "rbind2-notes_carryover.md")

# Regression: rbind2() inserted a header row built from internal data names
# when `y` was created with colnames = FALSE
tab <- rbind2(tt(mtcars[1:2, 1:2]), tt(mtcars[3:4, 1:2], colnames = FALSE))
expect_equal(nrow(tab@data_body), 4)
expect_false(any(tab@data_body[[1]] == "mpg"))
# header row still inserted when y has visible column names
tab <- rbind2(tt(mtcars[1:2, 1:2]), tt(mtcars[3:4, 1:2]))
expect_equal(nrow(tab@data_body), 5)
expect_true(any(tab@data_body[[1]] == "mpg"))

source("helpers.R")
using("tinysnapshot")
options(tinytable_print_output = "markdown")

# markdown without labels
k = iris[1:10, ]
colnames(k) <- NULL
tt(k)
expect_snapshot_print(
  tt(k),
  label = "markdown-nocolnames"
)

# no colnames
tab <- tt(head(iris))
colnames(tab) <- NULL
tab
expect_snapshot_print(tab, label = "markdown-nocolnames_02")

# group rows and columns
tab <- tt(mtcars[1:10, 1:5]) |>
  group_tt(
    i = list(
      "Hello" = 3,
      "World" = 8
    ),
    j = list(
      "Foo" = 2:3,
      "Bar" = 4:5
    )
  )
expect_snapshot_print(tab, label = "markdown-group_tt")

# simple caption
tab <- tt(mtcars[1:3, 1:3], caption = "Blah blah blah")
expect_snapshot_print(tab, label = "markdown-caption")

# Issue #105
x <- mtcars[1:15, ]
i <- list("blah" = 1, "blah" = 3, "blah" = 10)
tab <- tt(x) |> group_tt(i)
expect_snapshot_print(tab, label = "markdown-issue105")

# Issue 127: Large column
test <- data.frame(a = 1, b = 2, c = 3)
tab <- tt(test) |>
  group_tt(j = list("foobar" = 1:2, "hello world" = 3)) |>
  group_tt(j = list("foobar and hello world" = 2:3))
expect_snapshot_print(tab, label = "markdown-group_j_wider_1")

test <- data.frame(a = 1, b = 2, c = 3)
tab <- tt(test) |>
  group_tt(j = list("foobar" = 1:2, "hello world" = 3))
expect_snapshot_print(tab, label = "markdown-group_j_wider_2")

# Test bold columns
tab <- tt(mtcars[1:8, 1:5]) |>
  style_tt(j = c(2, 4), bold = TRUE)
expect_snapshot_print(tab, label = "markdown-bold_columns")

# Test replace missing value
tab <- mtcars[1:4, 1:3]
tab[2, 3] = NA
tab <- tt(tab) |>
  format_tt(replace = "-")
expect_snapshot_print(tab, label = "markdown-missing_value")

# Test footnote
x <- mtcars[1:3, 1:3]
n <- list("Blah blah", "*" = list(i = 0:1, j = 2, text = "foo bar"))
tab <- tt(x, notes = n)

expect_snapshot_print(tab, label = "markdown-footnote")

# GFM
if (!is_local) {
  exit_file("Run on Vincent's machine")
}
options(tinytable_print_output = "markdown")
expect_snapshot_print(tt(head(iris)) |> theme_markdown(style = "gfm"), label = "markdown-gfm")

# Long group labels - row groups
options(tinytable_print_output = "markdown")
x <- mtcars[1:5, 1:2]
tab <- tt(x) |> group_tt(i = list("AAAAAAAAAAAAAAAAAAAAAAAAAAAA" = 1))
expect_snapshot_print(tab, label = "markdown-long_row_group")

# Long group labels - column groups
x <- mtcars[1:3, 1:4]
tab <- tt(x) |>
  group_tt(j = list("Very Long Column Group Name Here" = 1:2, "Short" = 3:4))
expect_snapshot_print(tab, label = "markdown-long_column_group")

# Long group labels - mixed (both row and column)
x <- mtcars[1:3, 1:4]
tab <- tt(x) |>
  group_tt(
    i = list("SUPER LONG ROW GROUP NAME THAT SHOULD EXPAND TABLE" = 1)
  ) |>
  group_tt(
    j = list("Long Column Group" = 1:2, "Another Long Column Group" = 3:4)
  )
expect_snapshot_print(tab, label = "markdown-long_mixed_groups")

# Multiple long row groups
x <- mtcars[1:6, 1:3]
tab <- tt(x) |>
  group_tt(
    i = list(
      "First Very Long Row Group Name Here" = 1,
      "Second Even Longer Row Group Name That Should Also Expand" = 3,
      "Short" = 5
    )
  )
expect_snapshot_print(tab, label = "markdown-multiple_long_row_groups")

# colspan
x <- head(iris)[1:4, ]
tab <- tt(x) |>
  style_tt(1, 1, italic = TRUE) |>
  style_tt(2, 1, colspan = 3)
expect_snapshot_print(tab, label = "markdown-colspan_basic")

x <- head(iris)[1:4, ]
x[2, 1] <- "Very long content that spans multiple columns"
tab <- tt(x) |>
  style_tt(2, 1, colspan = 3)
expect_snapshot_print(tab, label = "markdown-colspan_long_content")

x <- head(iris)[1:5, ]
tab <- tt(x) |>
  style_tt(2, 1, colspan = 3) |>
  style_tt(4, 2, colspan = 2)
expect_snapshot_print(tab, label = "markdown-colspan_multiple")

# Test theme_empty() followed by theme_markdown() - should restore grid lines
options(tinytable_print_output = "markdown")
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_empty() |>
  theme_markdown(vline = TRUE, hline = TRUE)
expect_snapshot_print(tab, label = "markdown-theme_empty_restore_grid")

# Test theme_empty() alone - should have no grid lines
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_empty()
expect_snapshot_print(tab, label = "markdown-theme_empty_no_grid")

# Test theme_markdown grid options - hline only
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_markdown(hline = TRUE, vline = FALSE)
expect_snapshot_print(tab, label = "markdown-hline_only")

# Test theme_markdown grid options - vline only
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_markdown(hline = FALSE, vline = TRUE)
expect_snapshot_print(tab, label = "markdown-vline_only")

# Test theme_markdown with header line disabled
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_markdown(hline_header = FALSE)
expect_snapshot_print(tab, label = "markdown-no_header_line")

# Test theme_empty() then selective restoration
tab <- tt(head(iris)[1:3, 1:3]) |>
  theme_empty() |>
  theme_markdown(hline = TRUE, vline = FALSE)
expect_snapshot_print(tab, label = "markdown-theme_empty_hline_only")


# Issue #605
cap <- "Issue #605 lm1 and lm2 span 2 and p is in twice in colnames."
out <- data.frame(
  Parameter = c(
    "grp (2)", "grp (3)", "Days × grp (2)", "Days × grp (3)",
    "Days"
  ),
  `Coefficient (CI)` = c(
    "-4.31 (-15.95,  7.32)", "-1.31 (-13.47, 10.84)", "",
    "", "10.44 (  8.84, 12.03)"
  ),
  p = c(" 0.465", " 0.831", "", "", "<0.001"),
  `Coefficient (CI)` = c(
    " 0.32 (-22.56, 23.20)", " 3.77 (-19.72, 27.26)",
    "-1.01 ( -5.35,  3.32)", "-1.11 ( -5.53,  3.31)",
    "11.23 (  7.87, 14.60)"
  ),
  p = c(" 0.978", " 0.752", " 0.645", " 0.621", "<0.001"),
  check.names = FALSE
)
final <- tt(out, caption = cap)
final <- group_tt(
  final,
  i = list(Groups = 1L, Interactions = 3L, Controls = 5L),
  j = list(lm1 = 2:3, lm2 = 4:5)
)
final <- style_tt(final, i = "~groupi", j = 1, indent = 2)
expect_snapshot_print(tab, label = "markdown-issue605")



# restore
options(tinytable_print_output = NULL)

# colspan test from test-colrowspan.R
options(tinytable_print_output = "markdown")
tab <- aggregate(mpg ~ cyl + am, FUN = mean, data = mtcars)
tab <- tab[order(tab$cyl, tab$am), ]
tab <- tt(tab, digits = 2) |>
  style_tt(i = c(1, 3, 5), j = 1, colspan = 2)
expect_snapshot_print(tab, "colrowspan-markdown_multiple")
options(tinytable_print_output = NULL)



# Regression: markdown styling was dropped entirely for tables created with
# colnames = FALSE because the column guard compared against length(names(x))
tab <- tt(data.frame(a = 1:2, b = 3:4), colnames = FALSE) |>
  style_tt(i = 1, bold = TRUE)
out <- save_tt(tab, "markdown")
expect_true(grepl("**1**", out, fixed = TRUE))
expect_true(grepl("**3**", out, fixed = TRUE))

source("helpers.R")
options(tinytable_html_engine = "tabulator")
options(tinytable_tt_theme = theme_html)

# Basic tabulator table
dat <- head(iris)
tab <- tt(dat)
out <- save_tt(tab, "html")
expect_true(is.character(out))
expect_true(grepl("tabulator", out))
expect_true(grepl("Sepal.Length", out))

# Special characters in column names: field is sanitized to [A-Za-z0-9_],
# title keeps the original name (JSON-escaped)
special_dat <- data.frame(`hello\\?` = 1, check.names = FALSE)
tab <- tt(special_dat)
out <- save_tt(tab, "html")
expect_true(grepl('"field":"hello__"', out, fixed = TRUE))
expect_true(grepl('"title":"hello\\\\?"', out, fixed = TRUE))

# Data with multiple types
dat <- data.frame(
  city = c("Montréal", "Toronto", "Vancouver"),
  salary = c(14002.22, 201399.11, 80188.38),
  random = c(1.43402, 201.399, 0.134588),
  date = as.Date(sample(1:1000, 3), origin = "1970-01-01"),
  best = c(TRUE, FALSE, FALSE)
)

tab <- tt(dat)
out <- save_tt(tab, "html")
expect_true(is.character(out))
expect_true(grepl("Montréal", out))
expect_true(grepl("true", out, ignore.case = TRUE))

# Pagination and filtering
tab <- tt(iris) |> theme_html(
  tabulator_pagination = c(5, 10, 50),
  tabulator_search = "top"
)
out <- save_tt(tab, "html")
expect_true(grepl("pagination", out))
expect_true(grepl("search", out))
expect_true(grepl("paginationSize.*5", out))

# Format numeric columns
tab <- tt(dat) |>
  format_tt(j = "salary", digits = 2, num_mark_big = ",") |>
  format_tt(j = "random", digits = 4)
out <- save_tt(tab, "html")
expect_true(grepl("decimal", out))
expect_true(grepl("salary.*decimal", out))

# Format date columns
tab <- tt(dat) |> format_tt(j = "date", date = "M/d/yyyy")
out <- save_tt(tab, "html")
expect_true(grepl("M/d/yyyy", out))
expect_true(grepl("datetime", out))

# Simple data table for remaining tests
simple_dat <- data.frame(
  name = c("Alice", "Bob", "Charlie"),
  value = c(100, 200, 300)
)

tab <- tt(simple_dat)
out <- save_tt(tab, "html")
expect_true(grepl("Alice", out))

# Style alignment
tab <- tt(simple_dat) |> style_tt(align = "r")
out <- save_tt(tab, "html")
expect_true(grepl("right", out))

# Custom CSS rules
css_rule <- "
$TINYTABLE_ID .tabulator-header .tabulator-col {
  background-color: black;
  color: white;
}
"
tab <- tt(simple_dat) |> theme_html(tabulator_css_rule = css_rule)
out <- save_tt(tab, "html")
expect_true(grepl("background-color.*black", out))
expect_true(grepl("color.*white", out))

# Custom columns
dat_custom <- data.frame(
  city = c("Toronto", "Montreal", "Vancouver"),
  salary = c(75000, 68000, 82000),
  best = c(FALSE, TRUE, FALSE)
)

custom_columns <- '[
  {
    "title": "City",
    "field": "city"
  },
  {
    "title": "Best city",
    "field": "best",
    "formatter": "tickCross"
  },
  {
    "title": "Salary",
    "field": "salary",
    "formatter": "money",
    "formatterParams": {"precision": 0, "symbol": "$"}
  }
]'

tab <- tt(dat_custom) |> theme_html(tabulator_columns = custom_columns)
out <- save_tt(tab, "html")
expect_true(grepl("tickCross", out))
expect_true(grepl("money", out))
expect_true(grepl("Best city", out))

# Custom options
opts <- "
  layout: 'fitColumns',
  height: '200px'
"

tab <- tt(dat_custom) |> theme_html(tabulator_options = opts)
out <- save_tt(tab, "html")
expect_true(grepl("fitColumns", out))
expect_true(grepl("200px", out))

# Single styled cell builds without error and emits the color
tab <- tt(mtcars[1:3, 1:2]) |>
  theme_html(engine = "tabulator") |>
  style_tt(i = 2, j = 1, color = "red")
out <- save_tt(tab, "html")
expect_true(grepl("color: #FF0000", out, fixed = TRUE))

# Reset global option
options(tinytable_html_engine = NULL)

grid_notes_caption <- function(x) {
  out <- x@table_string

  # notes
  if (length(x@notes) > 0) {
    lines <- strsplit(out, split = "\\n")[[1]]
    table_width <- max(ansi_nchar(lines))
    target <- table_width - 4

    # Wrap all notes first so the box can widen when a note cannot wrap
    # down to the table width (e.g. very narrow tables or unbreakable words).
    note_lines <- vector("list", length(x@notes))
    for (i in seq_along(x@notes)) {
      no <- x@notes[[i]]
      txt <- if (is.list(no)) no$text else no
      if (isTRUE(names(x@notes)[i] != "")) {
        txt <- sprintf("^%s^ %s", names(x@notes)[i], txt)
      }
      note_lines[[i]] <- ansi_strwrap(txt, width = target)
    }
    box_width <- max(table_width, ansi_nchar(unlist(note_lines)) + 4)

    plus_lines <- grep("^\\+", lines)
    if (length(plus_lines) >= 2) {
      # Bottom border present: convert it into the top of the notes box,
      # widening it when the notes are wider than the table.
      idx <- utils::tail(plus_lines, 1)
      bot <- gsub("-", "=", lines[idx])
      if (box_width > table_width) {
        bot <- paste0(
          sub("\\+$", "", bot),
          strrep("=", box_width - table_width),
          "+"
        )
      }
      lines[idx] <- bot
    } else {
      # No bottom border (e.g. hline = FALSE): open a new box after the table
      bot <- paste0("+", strrep("=", box_width - 2), "+")
      while (length(lines) > 0 && lines[length(lines)] == "") {
        lines <- lines[-length(lines)]
      }
      lines <- c(lines, bot)
    }

    # Each note gets its own box segment, closed by a border line
    for (i in seq_along(note_lines)) {
      txt <- sprintf("| %s |", ansi_pad(note_lines[[i]], box_width - 4))
      lines <- c(lines, txt, bot)
    }
    out <- paste(lines, collapse = "\n")
  }

  # caption
  cap <- x@caption
  if (is.character(cap) && length(cap) == 1 && ansi_nchar(cap) > 0) {
    out <- paste0(out, "\n", "Table: ", cap, "\n")
  }

  x@table_string <- out
  return(x)
}


setMethod(
  f = "finalize",
  signature = "tinytable_grid",
  definition = function(x, ...) {
    x <- style_notes(x)
    x <- style_caption(x)
    x <- grid_hlines(x)
    x <- grid_notes_caption(x)
    return(x)
  }
)

# Process css_rule: detect file paths, URLs, or inline CSS
process_css_rule <- function(css_rule) {
  if (is.null(css_rule)) {
    return(list(type = "none", content = NULL))
  }

  # Check if it's a URL (starts with http/https and ends with .css)
  if (grepl("^https?://.*\\.css$", css_rule, ignore.case = TRUE)) {
    return(list(type = "url", content = css_rule))
  }

  # Check if it's a local file path
  if (file.exists(css_rule)) {
    css_content <- readLines(css_rule, warn = FALSE)
    css_content <- paste(css_content, collapse = "\n")
    return(list(type = "file", content = css_content))
  }

  # Otherwise, treat as inline CSS content
  return(list(type = "inline", content = css_rule))
}

setMethod(
  f = "finalize",
  signature = "tinytable_html",
  definition = function(x, ...) {
    # class
    cl <- x@html_class
    if (!grepl("\\btinytable\\b", cl)) {
      cl <- paste("tinytable", cl)
    }

    out <- sub(
      "$tinytable_HTML_CLASS",
      cl,
      x@table_string,
      fixed = TRUE
    )

    # Handle CSS inclusion - portable vs external
    if (isTRUE(x@html_portable)) {
      # If custom css_rule is provided, use it exclusively (even in portable mode)
      if (!is.null(x@html_css_rule)) {
        css_processed <- process_css_rule(x@html_css_rule)
        if (css_processed$type == "url") {
          # For URLs, use link tag
          css_content <- NULL
          css_include <- sprintf('<link rel="stylesheet" href="%s">', css_processed$content)
        } else {
          # For files or inline CSS, use the content
          css_content <- css_processed$content
        }
      } else {
        # For portable HTML, inline the CSS content from the file
        css_file_path <- system.file("tinytable.css", package = "tinytable")
        if (file.exists(css_file_path)) {
          css_content <- readLines(css_file_path, warn = FALSE)
          css_content <- paste(css_content, collapse = "\n")
        } else {
          # Fallback to external if file not found
          css_include <- sprintf(
            '<link rel="stylesheet" href="%s">',
            "https://cdn.jsdelivr.net/gh/vincentarelbundock/tinytable@main/inst/tinytable.css"
          )
          css_content <- NULL
        }
      }

      if (!is.null(css_content)) {
        css_include <- paste0("<style>\n", css_content, "\n</style>")

        # Also include Bootstrap CSS via external link if using bootstrap engine
        if (identical(x@html_engine, "bootstrap")) {
          bootstrap_css_link <- '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.8/dist/css/bootstrap.min.css">'
          css_include <- paste(bootstrap_css_link, css_include, sep = "\n")
        }
      }
    } else if (is.null(x@html_css_rule)) {
      # Use external CSS file
      css_include <- sprintf(
        '<link rel="stylesheet" href="%s">',
        "https://cdn.jsdelivr.net/gh/vincentarelbundock/tinytable@main/inst/tinytable.css"
      )
    } else {
      # Process css_rule (could be file path, URL, or inline CSS)
      css_processed <- process_css_rule(x@html_css_rule)
      if (css_processed$type == "url") {
        # For URLs, use link tag
        css_include <- sprintf('<link rel="stylesheet" href="%s">', css_processed$content)
      } else {
        # For files or inline CSS, use style tag with content
        css_include <- paste0("<style>\n", css_processed$content, "\n</style>")
      }
    }

    out <- sub(
      "$tinytable_CSS_INCLUDE",
      css_include,
      out,
      fixed = TRUE
    )

    # Handle JavaScript inclusion - portable vs external
    if (isTRUE(x@html_portable)) {
      # For portable HTML, inline the JS content from the file
      js_file_path <- system.file("tinytable.js", package = "tinytable")
      if (file.exists(js_file_path)) {
        js_content <- readLines(js_file_path, warn = FALSE)
        js_content <- paste(js_content, collapse = "\n")
        js_include <- paste0("<script>\n", js_content, "\n</script>")
      } else {
        # Fallback to external if file not found
        js_include <- '<script src="https://cdn.jsdelivr.net/gh/vincentarelbundock/tinytable@main/inst/tinytable.js"></script>'
      }

      # Note: For portable mode, we still use external Bootstrap as it's large
      # Users can override with custom css_rule if they need full portability
      if (identical(x@html_engine, "bootstrap")) {
        bootstrap_include <- '   <script src="https://cdn.jsdelivr.net/npm/bootstrap@5.3.8/dist/js/bootstrap.bundle.min.js"></script>'
        js_include <- paste(js_include, bootstrap_include, sep = "\n")
      }
    } else {
      # Use external files
      js_include <- '<script src="https://cdn.jsdelivr.net/gh/vincentarelbundock/tinytable@main/inst/tinytable.js"></script>'
      bootstrap_include <- '   <script src="https://cdn.jsdelivr.net/npm/bootstrap@5.3.8/dist/js/bootstrap.bundle.min.js"></script>'

      if (identical(x@html_engine, "bootstrap")) {
        js_include <- paste(js_include, bootstrap_include, sep = "\n")
      }
    }

    out <- sub(
      "$tinytable_JS_INCLUDE",
      js_include,
      out,
      fixed = TRUE
    )

    # Handle custom script inclusion
    if (!is.null(x@html_script)) {
      out <- sub(
        "$tinytable_SCRIPT",
        x@html_script,
        out,
        fixed = TRUE
      )
    } else {
      # Remove placeholder if no script is provided
      marker <- "    $tinytable_SCRIPT\n"
      if (grepl(marker, out, fixed = TRUE)) {
        out <- sub(marker, "", out, fixed = TRUE)
      } else {
        out <- lines_drop(
          out,
          "$tinytable_SCRIPT",
          fixed = TRUE,
          unique = FALSE
        )
      }
    }

    if (isTRUE(getOption("knitr.in.progress"))) {
      # Rmarkdown and Quarto load their own html, which we probably don't want to override
      out <- lines_drop(
        out,
        "jsdelivr.*bootstrap",
        fixed = FALSE,
        unique = FALSE
      )
      # avoid nesting full HTML page inside an HTML page
      out <- lines_drop_between(
        out,
        regex_start = "<!-- preamble start -->",
        regex_end = "<!-- preamble end -->",
        fixed = TRUE
      )
      out <- lines_drop_between(
        out,
        regex_start = "<!-- postamble start -->",
        regex_end = "<!-- postamble end -->",
        fixed = TRUE
      )
    }

    x@table_string <- out

    return(x)
  })

#' Rotate table or cell content
#'
#' In HTML, rotation is always applied to cell content (not the whole table) to
#' keep borders intact. In LaTeX and Typst, when `i` and `j` are both `NULL`,
#' the entire table is rotated; otherwise cell content is rotated.
#'
#' @param x A `tinytable` object.
#' @param angle Numeric. Rotation angle in degrees (0-360).
#' @param i Row indices for cell rotation. When `NULL` (and `j` is also `NULL`),
#'   rotates the entire table in LaTeX/Typst or all cells in HTML.
#' @param j Column indices for cell rotation. When `NULL` (and `i` is also `NULL`),
#'   rotates the entire table in LaTeX/Typst or all cells in HTML.
#' @param ... Additional arguments (ignored).
#' @return A modified `tinytable` object.
#' @export
theme_rotate <- function(
  x,
  angle = get_option("tinytable_rotate_angle", default = 90),
  i = NULL,
  j = NULL,
  ...
) {
  assert_numeric(angle, len = 1, lower = 0, upper = 360)

  # LaTeX/Typst: whole-table rotation when no i/j specified
  if (is.null(i) && is.null(j)) {
    x <- rotate_latex_table(x, angle)
    x <- rotate_typst_table(x, angle)
  } else {
    x <- rotate_latex_cells(x, angle, i, j)
    x <- rotate_typst_cells(x, angle, i, j)
  }

  # HTML: always cell-level rotation
  cal <- call(
    "rotate_cells_lazy_html",
    i = i,
    j = j,
    angle = angle
  )
  x@lazy_format <- c(x@lazy_format, list(cal))

  return(x)
}

rotate_latex_table <- function(x, angle) {
  fn <- function(table) {
    rot <- sprintf("\\begin{table}\n\\rotatebox{%s}{", format_markup_num(angle))
    table@table_string <- sub(
      "\\begin{table}",
      rot,
      table@table_string,
      fixed = TRUE
    )
    table@table_string <- sub(
      "\\end{table}",
      "}\n\\end{table}",
      table@table_string,
      fixed = TRUE
    )
    return(table)
  }
  build_finalize(x, fn, output = "latex")
}

rotate_typst_table <- function(x, angle) {
  fn <- function(table) {
    rot <- sprintf("#rotate(%sdeg, reflow: true, [\n  #figure(", format_markup_num(-angle))
    table@table_string <- sub(
      "#figure(",
      rot,
      table@table_string,
      fixed = TRUE
    )
    table@table_string <- sub(
      ") // end figure",
      ") ]) // end figure",
      table@table_string,
      fixed = TRUE
    )
    return(table)
  }
  build_finalize(x, fn, output = "typst")
}

rotate_latex_cells <- function(x, angle, i, j) {
  cal <- call(
    "rotate_cells_lazy_latex",
    i = i,
    j = j,
    angle = angle
  )
  x@lazy_format <- c(x@lazy_format, list(cal))
  return(x)
}

rotate_typst_cells <- function(x, angle, i, j) {
  cal <- call(
    "rotate_cells_lazy_typst",
    i = i,
    j = j,
    angle = angle
  )
  x@lazy_format <- c(x@lazy_format, list(cal))
  return(x)
}

rotate_cells_setup <- function(x, i, j) {
  s <- resolve_i_components(
    x,
    i,
    j,
    default = c("colnames", "cells", "groupi", "~groupi", "groupj")
  )
  i <- sanitize_i(s$i, x, lazy = FALSE, calling_function = "theme_rotate")
  j <- sanitize_j(j, x)
  list(i = i, j = j, components = s$components)
}

rotate_cells_lazy_html <- function(x, i, j, angle) {
  if (!identical(x@output, "html")) return(x)
  s <- rotate_cells_setup(x, i, j)
  sin_a <- abs(sin(angle * pi / 180))
  format_fn <- if (angle == 90 || angle == 270) {
    transform <- if (angle == 90) "writing-mode: vertical-rl; transform: rotate(180deg);" else "writing-mode: vertical-rl;"
    function(vec, ...) {
      ifelse(
        is.na(vec),
        vec,
        sprintf(
          '<div style="display: inline-block; %s white-space: nowrap;">%s</div>',
          transform,
          vec
        )
      )
    }
  } else {
    function(vec, ...) {
      pad <- sprintf("%.1fem", nchar(vec) * 0.15 * sin_a)
      ifelse(
        is.na(vec),
        vec,
        sprintf(
          '<div style="display: inline-block; transform: rotate(%sdeg); transform-origin: center center; white-space: nowrap; margin: %s 0;">%s</div>',
          format_markup_num(angle),
          pad,
          vec
        )
      )
    }
  }
  apply_format(x, i = s$i, j = s$j, format_fn = format_fn, components = s$components, original_data = FALSE)
}

rotate_cells_lazy_latex <- function(x, i, j, angle) {
  if (!identical(x@output, "latex")) return(x)
  s <- rotate_cells_setup(x, i, j)
  format_fn <- function(vec, ...) {
    ifelse(is.na(vec), vec, sprintf("\\rotatebox{%s}{%s}", format_markup_num(angle), vec))
  }
  apply_format(x, i = s$i, j = s$j, format_fn = format_fn, components = s$components, original_data = FALSE)
}

rotate_cells_lazy_typst <- function(x, i, j, angle) {
  if (!identical(x@output, "typst")) return(x)
  s <- rotate_cells_setup(x, i, j)
  format_fn <- function(vec, ...) {
    ifelse(
      is.na(vec),
      vec,
      sprintf("#rotate(%sdeg, reflow: true, [%s])", format_markup_num(-angle), vec)
    )
  }
  apply_format(x, i = s$i, j = s$j, format_fn = format_fn, components = s$components, original_data = FALSE)
}

#' Theme for a void table
#'
#' This function cancels all styles and formatting applied to a `tinytable`
#' object up to that point in the pipeline. Images and inline plots inserted
#' with `plot_tt()` are deliberately preserved: they are table content rather
#' than styling.
#' @inheritParams theme_tinytable
#' @export
theme_empty <- function(x, ...) {
  x@lazy_format <- list()
  x@lazy_style <- list()
  x@lazy_prepare <- list()
  x@lazy_finalize <- list()
  # @lazy_plot is intentionally NOT cleared: plot_tt() insertions are content
  # (images, inline plots), not styling, and must survive theme_empty().
  return(x)
}

options("tinysnapshot_device" = "svglite")
options("tinysnapshot_tol" = 300)
options("tinysnapshot_mode" = "sidebyside")
options("width" = 200) # for diff printout readability
options("tinytable_print_output" = NULL)
options("tinytable_html_engine" = NULL)

# Stabilize collation-dependent ordering in snapshot tests.
try(Sys.setlocale("LC_COLLATE", "C"), silent = TRUE)

# Several test files call unexported functions by their bare name.
# `pkgload::load_all()` puts them on the search path, but `R CMD check` runs
# against the installed namespace, where they are not visible. Bind them here
# so the tests behave the same under both entry points.
for (.fun in c(
  "build_tt",
  "df_to_json",
  "value_to_json",
  "json_raw",
  "filter_style_other",
  "lines_drop",
  "lines_drop_between"
)) {
  assign(.fun, get(.fun, envir = asNamespace("tinytable")))
}
rm(.fun)

# common formatting options
options(tinytable_format_bool = function(x) tools::toTitleCase(tolower(x)))
options(tinytable_format_replace = "")

strip_random_sequential <- function(x, stem = "tinytable_css_") {
  pattern <- paste0(stem, "\\w+\\b")
  match_list <- lapply(x, function(s) {
    regmatches(s, gregexpr(pattern, s, perl = TRUE))[[1]]
  })
  all_matches <- unlist(match_list)
  unique_matches <- unique(all_matches)
  new_ids <- sprintf("tinytable_css_%02d", seq_along(unique_matches))
  replaced <- x
  for (i in seq_along(unique_matches)) {
    replaced <- gsub(unique_matches[i], new_ids[i], replaced, perl = TRUE)
  }
  replaced
}
strip_random <- function(x) {
  for (stem in c(
    "tableFns_",
    "tinytable_(?!css)",
    "tinytable_css_",
    "tinytable/"
  )) {
    x <- strip_random_sequential(x, stem)
  }
  x
}
options(tinysnapshot_fn_current = strip_random)
options(tinysnapshot_fn_target = strip_random)

is_local <- isTRUE(grepl("(?i)vincent|mbp|mandelbrot", Sys.info()["user"]))

# libraries
requiet <- function(package) {
  void <- capture.output(
    pkg_available <- tryCatch(
      suppressPackageStartupMessages(
        suppressWarnings(
          suppressMessages(
            tryCatch(
              isTRUE(
                require(package, warn.conflicts = FALSE, character.only = TRUE)
              ),
              error = function(e) FALSE
            )
          )
        )
      )
    )
  )
  return(pkg_available)
}

requiet("tinytable")
requiet("tinytest")
requiet("tinysnapshot")

print.custom_html_string <- function(x, ...) {
  cat(x, "\n", sep = "", ...)
  invisible(x)
}

print_html <- function(x, output = c("html")) {
  output <- match.arg(output)
  x <- save_tt(x, output)
  class(x) <- c("custom_html_string", class(x))
  x
}

format_extensions <- list(
  markdown = "md",
  latex = "tex",
  typst = "typ",
  html = "html"
)

# Helper function to prepare tables for testing across multiple formats
expect_table <- function(tab, formats = names(format_extensions)) {
  result <- list()

  for (format in formats) {
    # Create a copy and set the output format
    table_copy <- tab
    table_copy@output <- format

    # Create the table output
    if (format == "html") {
      table_output <- print_html(table_copy)
    } else {
      table_output <- table_copy
    }

    # Store the table output
    result[[format]] <- table_output
  }

  return(result)
}

#' tinytable S4 method
#'
#' @keywords internal
setMethod(
  f = "group_eval_j",
  signature = "tinytable_tabulator",
  definition = function(x, i = NULL, j = NULL, ...) {
    # Column group headers are not implemented for the tabulator engine
    warning(
      "group_tt(j = ...) column groups are not supported by the tabulator engine and will be ignored.",
      call. = FALSE
    )
    return(x)
  }
)

# knitr/raster screenshot output shared by knit_print.tinytable and
# print.tinytable. Returns NULL when `output` is neither "knitr" nor "raster".
print_png <- function(x, output) {
  if (identical(output, "knitr")) {
    assert_dependency("webshot2")
    assert_dependency("knitr")
    tmp <- tempfile(fileext = ".png")
    save_tt(x, tmp, overwrite = TRUE)
    return(knitr::include_graphics(tmp))
  }
  if (identical(output, "raster")) {
    assert_dependency("webshot2")
    assert_dependency("png")
    tmp <- tempfile(fileext = ".png")
    save_tt(x, tmp, overwrite = TRUE)
    img <- png::readPNG(tmp)
    grid::grid.newpage()
    grid::grid.raster(img)
    return(invisible(x))
  }
  NULL
}

# wrap raw output in a fenced block (or inline backticks for single-line
# strings) so pandoc passes it through untouched
# adapted from htmltools:::html_preserve, GPL3
knitr_fence <- function(out, format) {
  inline <- !grepl("\n", out, fixed = TRUE)
  if (inline) {
    sprintf("`%s`{=%s}", out, format)
  } else {
    sprintf("\n```{=%s}\n%s\n```\n", format, out)
  }
}

#' @exportS3Method xfun::record_print
record_print.tinytable <- function(x, ...) {
  if (!isTRUE(check_dependency("litedown"))) {
    return(x)
  }

  # litedown
  fmt <- tryCatch(litedown::get_context("format"), error = function(e) NULL)
  if (is.null(fmt)) {
    return(x)
  }

  if (!isTRUE(fmt %in% c("latex", "markdown", "html"))) {
    stop(
      "tinytable in litedown only supports latex, markdown, or html output",
      call. = FALSE
    )
  }

  x <- build_tt(x, output = fmt)
  out <- x@table_string

  out <- litedown::raw_text(out, fmt)
  return(out)
}

#' Print a tinytable object in knitr
#'
#' @keywords internal
#' @return A string with class `knit_asis` to be printed in Rmarkdown or Quarto documents.
#' @rawNamespace S3method(knitr::knit_print, tinytable)
#' @export
knit_print.tinytable <- function(
    x,
    output = get_option("tinytable_print_output", default = NULL),
    ...) {
  if (identical(output, "knitr") || identical(output, "raster")) {
    return(print_png(x, output))
  }

  # lazy styles get evaluated here, at the very end
  x <- build_tt(x, output = output)
  out <- x@table_string

  if (isTRUE(x@output %in% c("html", "typst"))) {
    out <- knitr_fence(out, x@output)
  }

  class(out) <- "knit_asis"
  return(out)
}

#' Print a tinytable object in pkgdown
#'
#' @param x A tinytable object
#' @param visible Whether the object is visible
#' @keywords internal
#' @return HTML output for pkgdown examples
#' @export
pkgdown_print.tinytable <- function(x, visible = TRUE) {
  if (!visible) {
    return(invisible())
  }

  # Build table with HTML output for pkgdown
  x <- build_tt(x, output = "html")
  out <- x@table_string

  # Return as browsable HTML (similar to knit_print pattern)
  htmlText <- paste(enc2utf8(out), collapse = " ")
  attr(htmlText, "html") <- TRUE
  attr(htmlText, "browsable_html") <- TRUE
  class(htmlText) <- c("html", "character")
  htmlText
}

#' Print, display, or convert a tinytable object
#'
#' This function is called automatically by `R` whenever a `tinytable` object is anprinted to the console or in an HTML viewer pane.
#'
#' @inheritParams tt
#' @param output Format in which a Tiny Table is printed. One of:
#' + `NULL`: automatically chosen based on context (see below).
#' + `"html"`: HTML table. In interactive mode, displayed in the viewer pane.
#' + `"latex"`: LaTeX table using tabularray.
#' + `"markdown"`: plain text markdown table.
#' + `"typst"`: Typst table.
#' + `"dataframe"`: formatted data frame.
#' + `"knitr"`: PNG screenshot via `webshot2`, returned through `knitr::include_graphics()`. Useful for embedding styled tables in Rmarkdown or Quarto documents. Requires the `webshot2` package.
#' + `"raster"`: PNG screenshot via `webshot2`, drawn on the graphics device via `grid::grid.raster()`. Useful for pkgdown `@examples` and README files, where the plot is captured as an image. Requires the `webshot2` and `png` packages. For README.Rmd files rendered with `rmarkdown::render()`, set `knitr::opts_chunk$set(fig.path = "man/figures/README-")` in a setup chunk so that pkgdown copies the images correctly.
#'
#' When `output` is `NULL`, the format is chosen automatically:
#' + When called from a script in non-interactive mode, the default is "markdown" (`interactive() == FALSE`).
#' + When called interactively in RStudio, the default is to display an HTML table in the viewer pane.
#' + When called interactively in another development environment, the default is "markdown".
#' + The default print output can be changed for an entire R session by calling: `options(tinytable_print_output = "html")`
#' + The default print output can be changed for a single `tinytable` object by modifying the `output` S4 slot.
#' @details
#' When printing to HTML in `interactive()` mode, a temporary file is created and `viewer()` is called to preview the file with the local browser (ex: Firefox or Chrome). The temporary file is then automatically cleaned up. On some operating systems, like some Linux distributions, browser do not have read access to the `/tmp/` directory. In such cases, users can specify a custom location to store temporary HTML files. Note that this prevents `tinytable` from automatically cleaning up temporary files automatically.
#'
#' `options(tinytable_tempdir = "/home/username/temp_directory")`
#' @param ... Other arguments are ignored.
#' @return launch a browser window or cat() the table to console.
#' @export
print.tinytable <- function(
    x,
    output = get_option("tinytable_print_output", default = NULL),
    ...) {
  x <- sanitize_output(x, output)

  if (identical(output, "knitr") || identical(output, "raster")) {
    return(print_png(x, output))
  }

  output <- infer_output(x)

  dir <- getOption("tinytable_tempdir", default = tempdir())
  dir <- sub("\\/$", "", dir)

  if (identical(output, "html")) {
    x@output_dir <- dir
    x@html_portable <- TRUE
  }


  x <- build_tt(x, output = output)

  tab <- x@table_string

  # lazy styles get evaluated here by build_tt(), at the very end
  if (identical(output, "dataframe")) {
    out <- x@data_body
    colnames(out) <- NULL
    if (x@ansi) {
      rows <- apply(out, 1, paste, collapse = " ")
      tabl <- paste(rows, collapse = "\n")
      cat("\n", tabl, "\n")
      return(invisible(out))
    } else {
      return(out)
    }
  } else if (output %in% c("latex", "typst", "markdown")) {
    cat(tab, "\n")
  } else if (output == "html") {
    if (is_rstudio_notebook()) {
      html_tinytable <- htmltools_browsable(tab)
      print(html_tinytable)
      return(invisible(NULL))
    }

    # need to change the output directory to a temporary directory
    # for plot_tt() inline plots to show up in RStudio
    random_name <- paste0(get_id("tinytable_"), ".html")
    htmlFile <- path.expand(normalizePath(file.path(dir, random_name), mustWork = FALSE))
    # on.exit(unlink(htmlFile))
    cat(tab, file = htmlFile)

    if (interactive()) {
      msg <- "Please choose a default browser with:

      options(browser = 'firefox')
      "
      if (identical(getOption("browser"), "")) {
        stop(msg, call. = FALSE)
      }

      viewer <- getOption("viewer", utils::browseURL)
      viewer(htmlFile)
    } else {
      cat(tab, "\n")
    }
  }

  return(invisible(x))
}

setMethod("show", "tinytable", function(object) {
  print.tinytable(
    object,
    output = get_option("tinytable_print_output", default = NULL)
  )
})

# adapted from {htmltools} under GPL3
htmltools_browsable <- function(x) {
  htmlText <- paste(enc2utf8(x), collapse = " ")
  attr(htmlText, "html") <- TRUE
  attr(htmlText, "browsable_html") <- TRUE
  class(htmlText) <- c("html", "character")
  class(htmlText) <- "shiny.tag.list"
  htmlText
}

format_vector_sprintf <- function(vec, sprintf_pattern = NULL, ...) {
  if (is.null(sprintf_pattern)) {
    return(NULL)
  }
  base::sprintf(sprintf_pattern, vec)
}

format_vector_logical <- function(vec, bool_fn = NULL, ...) {
  if (!is.logical(vec) || is.null(bool_fn)) {
    return(NULL)
  }
  bool_fn(vec)
}

format_vector_date <- function(vec, date_format = NULL, ...) {
  if (!inherits(vec, "Date") || is.null(date_format)) {
    return(NULL)
  }
  format(vec, date_format)
}

format_vector_other <- function(vec, other_fn = NULL, ...) {
  if (!is.function(other_fn)) {
    return(NULL)
  }
  other_fn(vec)
}

format_vector_custom <- function(vec, fn = NULL, ...) {
  if (!is.function(fn)) {
    return(NULL)
  }
  fn(vec)
}

format_vector_math <- function(vec, math = FALSE, ...) {
  if (!isTRUE(math)) {
    return(NULL)
  }
  sprintf("$%s$", vec)
}

format_vector_linebreak <- function(vec, linebreak = NULL, output = NULL, ...) {
  if (is.null(linebreak)) {
    return(NULL)
  }

  # Determine the appropriate line break sequence based on output format
  if (is.null(output) || output == "markdown") {
    return(NULL) # No line break replacement for markdown
  }

  if (output %in% c("html", "bootstrap", "tabulator")) {
    lb <- "<br>"
  } else if (output %in% c("latex", "pdf")) {
    lb <- "\\\\"
    # tabularray wrapper for line breaks
    if (any(grepl(linebreak, vec, fixed = TRUE))) {
      vec <- sprintf("{%s}", vec)
    }
  } else if (output == "typst") {
    # needs a space in typst
    lb <- " \\ "
  } else {
    return(NULL) # Unknown output format
  }

  gsub(linebreak, lb, vec, fixed = TRUE)
}

format_vector_quarto <- function(i, col, x, ...) {
  out <- x@data_body
  if (isTRUE(x@output %in% c("html", "bootstrap", "tabulator"))) {
    fun <- function(z) {
      z@table_string <- sub(
        "data-quarto-disable-processing='true'",
        "data-quarto-disable-processing='false'",
        z@table_string,
        fixed = TRUE
      )
      return(z)
    }
    x <- style_tt(x, finalize = fun)
    out[i, col] <- sprintf(
      '<span data-qmd="%s"></span>',
      out[i, col, drop = TRUE]
    )
  } else if (isTRUE(x@output == "latex")) {
    assert_dependency("base64enc")
    tmp <- sapply(
      out[i, col, drop = TRUE],
      function(z) base64enc::base64encode(charToRaw(z))
    )
    out[i, col] <- sprintf("\\QuartoMarkdownBase64{%s}", tmp)
  }

  x@data_body <- out
  return(x)
}

#' Internal styling function
#'
#' @inheritParams style_tt
#' @keywords internal
#' @noRd
setMethod(
  f = "style_eval",
  signature = "tinytable_tabulator",
  definition = function(x,
                        i = NULL,
                        j = NULL,
                        bold = FALSE,
                        italic = FALSE,
                        monospace = FALSE,
                        underline = FALSE,
                        strikeout = FALSE,
                        color = NULL,
                        background = NULL,
                        fontsize = NULL,
                        align = NULL,
                        alignv = NULL,
                        line = NULL,
                        line_color = "black",
                        line_width = 0.1,
                        colspan = NULL,
                        rowspan = NULL,
                        indent = 0,
                        ...) {

    x <- tabulator_apply_styles(x)
    return(x)
  })


#' Apply styles to Tabulator table
#' @keywords internal
#' @noRd
tabulator_apply_styles <- function(x) {
  if (nrow(x@style) == 0) return(x)

  # Use populated @style_other from build_tt() (Tabulator ignores line styles)
  other <- filter_style_other(x@style_other, STYLE_PROPS_TABULATOR)

  # Precompute field names once (dots/spaces -> underscores)
  field_names <- tabulator_clean_column_name(x@names)

  # First pass: handle column-wide alignment via column definitions
  for (idx in seq_len(nrow(x@style))) {
    orig_style <- x@style[idx, ]
    if (tabulator_is_column_alignment(orig_style)) {
      x <- tabulator_apply_column_alignment(x, orig_style)
    }
  }

  # Group cells by their style properties (before CSS conversion)
  # This avoids calling style_to_css() thousands of times
  style_groups <- list()  # style_signature -> list of cells

  if (!is.null(other) && nrow(other) > 0) {
    # Create a signature for each unique style combination
    style_cols <- c("bold", "italic", "monospace", "underline", "strikeout",
                    "color", "background", "fontsize", "align", "alignv", "indent")

    # Vectorized signature creation: build signatures for all rows at once
    # (do.call(paste) keeps this robust when nrow(other) == 1, where
    # sapply() would collapse to a vector and apply() would fail)
    sig_cols <- lapply(style_cols, function(col) {
      if (col %in% names(other)) {
        vals <- other[[col]]
        ifelse(is.na(vals), "NA", as.character(vals))
      } else {
        rep("NA", nrow(other))
      }
    })
    signatures <- do.call(paste, c(sig_cols, list(sep = "|")))

    # Group by signature
    for (row_idx in seq_len(nrow(other))) {
      signature <- signatures[row_idx]
      i <- other$i[row_idx]
      j <- other$j[row_idx]

      if (is.null(style_groups[[signature]])) {
        style_groups[[signature]] <- list(
          css = NULL,  # Will compute once
          row = other[row_idx, , drop = FALSE],
          cells = list()
        )
      }

      style_groups[[signature]]$cells[[length(style_groups[[signature]]$cells) + 1]] <- list(i = i, j = j)
    }

    # Now convert each unique style to CSS once
    for (sig in names(style_groups)) {
      css_content <- style_to_css(style_groups[[sig]]$row)
      style_groups[[sig]]$css <- css_content
    }
  }

  # Build column_styles from style_groups
  column_styles <- list()  # j -> list(cell_key -> css_content)

  for (sig in names(style_groups)) {
    group <- style_groups[[sig]]
    css_content <- group$css

    if (nchar(css_content) > 0) {
      for (cell in group$cells) {
        j_str <- as.character(cell$j)
        cell_key <- paste(cell$i, cell$j, sep = "_")

        if (is.null(column_styles[[j_str]])) {
          column_styles[[j_str]] <- list(cells = list(), css = list())
        }

        column_styles[[j_str]]$cells[[cell_key]] <- cell
        column_styles[[j_str]]$css[[cell_key]] <- css_content
      }
    }
  }

  # Check which columns have uniform styling across all rows
  css_rules <- character()
  targets <- list()

  for (j_str in names(column_styles)) {
    j <- as.integer(j_str)
    col_data <- column_styles[[j_str]]
    col_css_values <- col_data$css

    # Check if all cells in this column have identical CSS
    unique_css <- unique(unlist(col_css_values))

    # Column-wide if: single unique CSS and covers all data rows (nrow) plus header (nhead)
    expected_cells <- nrow(x) + x@nhead

    if (length(unique_css) == 1 && length(col_css_values) == expected_cells) {
      # Column-wide: all rows in this column have same CSS
      field_name <- field_names[j]
      css_rule <- sprintf("$TINYTABLE_ID .tabulator-cell[tabulator-field='%s'] { %s }",
                        field_name, unique_css)
      css_rules <- c(css_rules, css_rule)
    } else {
      # Cell-specific: need JavaScript for this column
      # Group cells by CSS to reuse classes
      css_to_cells <- list()
      for (cell_key in names(col_css_values)) {
        css_content <- col_css_values[[cell_key]]
        if (is.null(css_to_cells[[css_content]])) {
          css_to_cells[[css_content]] <- list()
        }
        css_to_cells[[css_content]][[length(css_to_cells[[css_content]]) + 1]] <- cell_key
      }

      # Generate CSS class and targets for each unique CSS
      for (css_content in names(css_to_cells)) {
        class_name <- get_id(stem = "tinytable_style_")
        css_rule <- sprintf("$TINYTABLE_ID .%s { %s }", class_name, css_content)
        css_rules <- c(css_rules, css_rule)

        # Add targets for all cells with this CSS
        for (cell_key in css_to_cells[[css_content]]) {
          cell_data <- col_data$cells[[cell_key]]
          targets[[length(targets) + 1]] <- list(
            pos = cell_data$i,
            field = field_names[cell_data$j],
            class = class_name
          )
        }
      }
    }
  }

  # Append CSS rules once
  if (length(css_rules) > 0) {
    css_block <- paste(css_rules, collapse = "\n")
    if (nchar(x@tabulator_css_rule) > 0) {
      x@tabulator_css_rule <- paste(x@tabulator_css_rule, css_block, sep = "\n")
    } else {
      x@tabulator_css_rule <- css_block
    }
  }

  # Build cellStyles map keyed by row index (stable across sorts)
  if (length(targets) > 0) {
    # Group targets by row index
    row_styles <- list()
    for (t in targets) {
      row_key <- as.character(t$pos)
      if (is.null(row_styles[[row_key]])) {
        row_styles[[row_key]] <- list()
      }
      row_styles[[row_key]][[t$field]] <- t$class
    }

    # Generate Map entries: [rowIdx, { field: class, ... }]
    map_entries <- sapply(names(row_styles), function(row_key) {
      fields <- row_styles[[row_key]]
      field_entries <- sapply(names(fields), function(field) {
        sprintf("'%s':'%s'", field, fields[[field]])
      })
      sprintf("[%s,{%s}]", row_key, paste(field_entries, collapse = ","))
    })
    map_str <- paste0("[", paste(map_entries, collapse = ","), "]")

    js_block <- sprintf(
      "const cellStyles_tinytable_%s = new Map(%s);",
      x@id,
      map_str
    )

    # Append to post_init
    if (is.null(x@tabulator_post_init) || nchar(x@tabulator_post_init) == 0) {
      x@tabulator_post_init <- js_block
    } else {
      x@tabulator_post_init <- paste(x@tabulator_post_init, js_block, sep = "\n")
    }

    # Add rowFormatter to options
    row_formatter <- sprintf(
      "index: '_tinytable_row_index',\n    rowFormatter: function(row) {\n      const d = row.getData();\n      const rowIdx = d._tinytable_row_index;\n      const spec = cellStyles_tinytable_%s.get(rowIdx);\n      if (!spec) return;\n      for (const [field, cls] of Object.entries(spec)) {\n        const cell = row.getCell(field);\n        if (cell) cell.getElement().classList.add(cls);\n      }\n    }",
      x@id
    )

    # Prepend to tabulator_options
    if (nchar(x@tabulator_options) > 0) {
      x@tabulator_options <- paste0(row_formatter, ",\n    ", x@tabulator_options)
    } else {
      x@tabulator_options <- row_formatter
    }
  }

  return(x)
}


#' Check if style is column-wide alignment only
#' @keywords internal
#' @noRd
tabulator_is_column_alignment <- function(style_row) {
  # Column alignment if: has j, no i (or i is NA), and only align/alignv set
  has_align <- !is.na(style_row$align) || !is.na(style_row$alignv)
  has_j <- length(style_row$j[[1]]) > 0 && !all(is.na(style_row$j[[1]]))
  i_vals <- style_row$i[[1]]
  has_no_i <- length(i_vals) == 0 || all(is.na(i_vals))

  if (!has_align || !has_j) return(FALSE)
  if (!has_no_i) return(FALSE)

  # Check if only align/alignv are set (nothing else)
  other_fields <- c("bold", "italic", "monospace", "underline", "strikeout",
                    "color", "background", "fontsize", "indent")
  has_other_styles <- any(sapply(other_fields, function(f) {
    if (f %in% names(style_row)) !is.na(style_row[[f]]) && style_row[[f]]
    else FALSE
  }))

  return(!has_other_styles)
}


#' Apply column-wide alignment
#' @keywords internal
#' @noRd
tabulator_apply_column_alignment <- function(x, style_row) {
  j_indices <- style_row$j[[1]]

  for (j in j_indices) {
    col_name <- x@names[j]

    # Initialize column styling if needed
    if (is.null(x@tabulator_column_styles)) {
      x@tabulator_column_styles <- list()
    }
    if (is.null(x@tabulator_column_styles[[col_name]])) {
      x@tabulator_column_styles[[col_name]] <- list()
    }

    # Map align values
    if (!is.na(style_row$align)) {
      x@tabulator_column_styles[[col_name]]$hozAlign <-
        map_align(style_row$align, "tabulator")
    }

    # Map alignv values (Tabulator vertAlign uses the CSS top/middle/bottom values)
    if (!is.na(style_row$alignv)) {
      x@tabulator_column_styles[[col_name]]$vertAlign <-
        map_alignv(style_row$alignv, "css")
    }
  }

  return(x)
}

#' HTML-specific styles and options
#' @param x A `tinytable` object.
#' @param engine Character string specifying the HTML engine: "tinytable", "bootstrap", or "tabulator".
#' @param i Row indices.
#' @param j Column indices.
#' @param class String. HTML table class.
#' @param css Character vector. CSS style declarations.
#' @param css_rule String. CSS rules to customize table styling. Can be one of:
#'   - Shortcut strings: "tinytable" (default) or "dark" for dark theme
#'   - A file path to a local CSS file (e.g., "custom.css" or "path/to/styles.css")
#'   - A URL to an external stylesheet (must start with http/https and end with .css)
#'   - Inline CSS content as a string
#' @param script String. Custom JavaScript code or script tags to inject into the HTML output. Useful for loading external JavaScript libraries like Plotly, D3, etc.
#' @param portable Logical. Sets whether to create portable HTML output with embedded Javascript, CSS, and base64-encoded images.
#' @param tabulator_stylesheet Character string. CSS stylesheet theme for Tabulator.js tables.
#'   Default is "bootstrap5". Available options: "default", "simple", "midnight", "modern",
#'   "site", "site_dark", "bootstrap3", "bootstrap4", "bootstrap5", "semanticui", "bulma",
#'   "materialize", or a custom HTTP URL starting with "http".
#' @param tabulator_layout Character string. Table layout algorithm for column sizing.
#'   Default is "fitDataTable". Available options: "fitDataTable", "fitData", "fitDataFill",
#'   "fitDataStretch", "fitColumns".
#' @param tabulator_pagination Logical or numeric vector. Pagination settings for large tables.
#'   - NULL (default): Preserves existing pagination settings, does not change previous configuration
#'   - FALSE: Explicitly disable pagination
#'   - TRUE: Enable pagination with automatic page sizes (10, 25, 50, 100, 250) filtered by row count
#'   - Numeric vector: First element is default page size, full vector provides page size options
#' @param tabulator_search Character or NULL. Search functionality position.
#'   - NULL (default): Preserves existing search settings, does not change previous configuration
#'   - "top": Adds search box above the table
#'   - "bottom": Adds search box below the table
#'   - "column": Adds header filters to each column for per-column searching
#' @param tabulator_options Custom Tabulator.js configuration options.
#' @param tabulator_css_rule Complete CSS rules.
#' @param tabulator_columns Custom column definitions.
#' @param ... Additional arguments are ignored.
#'
#' @examples
#' \dontrun{
#' # Load an external JavaScript library
#' library(tinytable)
#' x <- tt(mtcars[1:5, 1:5])
#' x <- theme_html(x, script = '<script src="https://cdn.plot.ly/plotly-2.27.0.min.js"></script>')
#'
#' # Add custom inline JavaScript
#' x <- tt(mtcars[1:5, 1:5])
#' x <- theme_html(x, script = '<script>console.log("Table loaded");</script>')
#' }
#'
#' @export
theme_html <- function(
    x,
    engine = get_option("tinytable_html_engine", default = NULL),
    i = NULL,
    j = NULL,
    class = get_option("tinytable_html_class", default = NULL),
    css = get_option("tinytable_html_css", default = NULL),
    css_rule = get_option("tinytable_html_css_rule", default = NULL),
    script = get_option("tinytable_html_script", default = NULL),
    portable = get_option("tinytable_html_portable"),
    tabulator_columns = get_option("tinytable_html_tabulator_columns"),
    tabulator_css_rule = get_option("tinytable_html_tabulator_css_rule"),
    tabulator_layout = get_option("tinytable_html_tabulator_layout", default = "fitDataTable"),
    tabulator_options = get_option("tinytable_html_tabulator_options"),
    tabulator_pagination = get_option("tinytable_html_tabulator_pagination"),
    tabulator_search = get_option("tinytable_html_tabulator_search"),
    tabulator_stylesheet = get_option("tinytable_html_tabulator_stylesheet"),
    ...) {
  assert_string(class, null.ok = TRUE)
  assert_choice(engine, c("tinytable", "bootstrap", "tabulator"), null.ok = TRUE)
  assert_choice(tabulator_search, c("top", "bottom", "column"), null.ok = TRUE)
  sanity_tabulator_css_rule(tabulator_css_rule)
  sanity_tabulator_columns(tabulator_columns)

  if (!is.null(engine)) {
    x@html_engine <- engine
    if (engine == "bootstrap" && is.null(class) && identical(x@html_class, "tinytable")) {
      x@html_class <- "table"
    }
  }

  if (!is.null(css_rule)) {
    assert_string(css_rule)
    # Handle shortcut strings
    if (css_rule == "tinytable") {
      css_rule <- "https://cdn.jsdelivr.net/gh/vincentarelbundock/tinytable@main/inst/tinytable.css"
    } else if (css_rule == "dark") {
      css_rule <- "https://cdn.jsdelivr.net/gh/vincentarelbundock/tinytable@main/inst/tinytable_dark.css"
    }
    x@html_css_rule <- css_rule
  }

  if (!is.null(script)) {
    assert_string(script)
    x@html_script <- script
  }

  if (!is.null(class)) {
    x@html_class <- class
  }

  if (!is.null(portable)) {
    assert_flag(portable)
    x@html_portable <- portable
  }

  # Handle css parameter for styling
  if (!is.null(css)) {
    # Use style_tt with html_css parameter - let style_tt handle the expansion logic
    x <- style_tt(x, i = i, j = j, html_css = css)
  }

  x <- theme_html_tabulator(
    x,
    tabulator_stylesheet = tabulator_stylesheet,
    tabulator_layout = tabulator_layout,
    tabulator_pagination = tabulator_pagination,
    tabulator_search = tabulator_search,
    tabulator_options = tabulator_options,
    tabulator_css_rule = tabulator_css_rule,
    tabulator_columns = tabulator_columns,
    ...
  )

  return(x)
}

#' Mark a string as raw JSON/JavaScript
#'
#' Values wrapped with `json_raw()` are emitted verbatim (unquoted, unescaped)
#' by `df_to_json()`. Use this for pre-serialized JSON fragments (e.g. sparkline
#' data arrays) or JavaScript identifiers that must not be quoted.
#'
#' @param x Character string containing raw JSON or JavaScript
#' @return `x` with class `tt_json_raw`
#' @keywords internal
#' @noRd
json_raw <- function(x) structure(x, class = "tt_json_raw")


#' Escape a character vector for use inside a JSON string literal
#'
#' @param s Character vector
#' @return Escaped character vector
#' @keywords internal
#' @noRd
json_escape_string <- function(s) {
  s <- gsub("\\", "\\\\", s, fixed = TRUE)
  s <- gsub('"', '\\"', s, fixed = TRUE)
  s <- gsub("\n", "\\n", s, fixed = TRUE)
  s <- gsub("\r", "\\r", s, fixed = TRUE)
  s <- gsub("\t", "\\t", s, fixed = TRUE)
  # JS-breaking line separators
  s <- gsub("\u2028", "\\u2028", s, fixed = TRUE)
  s <- gsub("\u2029", "\\u2029", s, fixed = TRUE)
  s
}


# Convert a single value to JSON
value_to_json <- function(value) {
  if (inherits(value, "tt_json_raw")) {
    return(unclass(value))
  }
  if (is.null(value)) {
    return("null")
  }
  if (length(value) == 1 && !is.list(value) && is.na(value)) {
    return("null")
  }
  if (is.numeric(value) && length(value) == 1) {
    if (!is.finite(value)) {
      return("null") # bare Inf/-Inf is a JS ReferenceError
    }
    return(as.character(value)) # Return as unquoted number for JSON
  }
  if (is.logical(value) && length(value) == 1) {
    return(if (value) "true" else "false")
  }
  if (is.list(value)) {
    # Recursively handle nested lists
    return(list_to_json(value))
  }
  if (length(value) == 1) {
    # Single string value: always quoted and escaped
    return(paste0('"', json_escape_string(as.character(value)), '"'))
  }
  # Vector - convert to array
  json_values <- vapply(value, function(v) {
    if (is.na(v)) {
      "null"
    } else if (is.numeric(v)) {
      if (is.finite(v)) as.character(v) else "null"
    } else if (is.logical(v)) {
      if (v) "true" else "false"
    } else {
      paste0('"', json_escape_string(as.character(v)), '"')
    }
  }, character(1))
  paste0("[", paste(json_values, collapse = ","), "]")
}


# Convert a named list to a JSON object
list_to_json <- function(lst) {
  if (length(lst) == 0) {
    return("{}")
  }

  pairs <- sapply(names(lst), function(name) {
    # Special handling for formatter and sorter fields - don't quote custom function names
    if ((name == "formatter" || name == "sorter") && is.character(lst[[name]]) &&
        grepl("^tinytable_", lst[[name]])) {
      # Custom formatter/sorter - output as unquoted function reference
      paste0('"', json_escape_string(name), '":', lst[[name]])
    } else {
      value_json <- value_to_json(lst[[name]])
      paste0('"', json_escape_string(name), '":', value_json)
    }
  })

  paste0("{", paste(pairs, collapse = ","), "}")
}


# Serialize one data frame column to a character vector of JSON values
json_serialize_column <- function(col) {
  if (is.list(col)) {
    # List-columns (e.g. raw JSON sparkline cells) go through value_to_json
    return(vapply(col, value_to_json, character(1)))
  }
  if (is.numeric(col)) {
    out <- as.character(col)
    out[!is.finite(col)] <- "null" # NA/NaN/Inf/-Inf -> null
    return(out)
  }
  if (is.logical(col)) {
    out <- ifelse(col, "true", "false")
    out[is.na(col)] <- "null"
    return(out)
  }
  # character, factor, Date, etc.
  s <- as.character(col)
  out <- paste0('"', json_escape_string(s), '"')
  out[is.na(s)] <- "null"
  out
}


#' Convert data frame or list to JSON
#'
#' Simple helper function to convert a data frame or list to JSON format
#'
#' @param data A data frame or list to convert to JSON
#'
#' @return JSON string
#' @keywords internal
df_to_json <- function(data) {
  # Handle list input (for column definitions)
  if (is.list(data) && !is.data.frame(data)) {
    # Check if this is a list of lists (array of objects)
    if (length(data) > 0 && all(sapply(data, is.list))) {
      # Convert list of lists to JSON array
      json_objects <- sapply(data, list_to_json)
      return(paste0("[", paste(json_objects, collapse = ","), "]"))
    } else {
      # Single object
      return(list_to_json(data))
    }
  }

  # Handle data frame input - convert to array of objects (column-wise, vectorized)
  if (nrow(data) == 0 || ncol(data) == 0) {
    return(paste0("[", paste(rep("{}", nrow(data)), collapse = ","), "]"))
  }

  keys <- paste0('"', json_escape_string(names(data)), '":')
  cols <- lapply(data, json_serialize_column)
  pieces <- Map(function(k, v) paste0(k, v), keys, cols)
  rows <- do.call(paste, c(unname(pieces), list(sep = ",")))
  paste0("[", paste0("{", rows, "}", collapse = ","), "]")
}

# Unit tests for the internal JSON serializer (R/json.R)
df_to_json <- tinytable:::df_to_json
json_raw <- tinytable:::json_raw
value_to_json <- tinytable:::value_to_json

# quotes, backslashes, and newlines are escaped
out <- df_to_json(data.frame(
  x = 'a"b',
  y = "back\\slash",
  z = "new\nline",
  stringsAsFactors = FALSE
))
expect_equal(out, '[{"x":"a\\"b","y":"back\\\\slash","z":"new\\nline"}]')

# leading-zero strings stay quoted strings (no numeric coercion)
expect_equal(
  df_to_json(data.frame(id = "007", stringsAsFactors = FALSE)),
  '[{"id":"007"}]'
)

# non-finite numerics -> null
expect_equal(
  df_to_json(data.frame(v = c(Inf, -Inf, NaN, NA, 1.5))),
  '[{"v":null},{"v":null},{"v":null},{"v":null},{"v":1.5}]'
)

# missing character and logical values -> null
expect_equal(
  df_to_json(data.frame(s = NA_character_, b = NA, stringsAsFactors = FALSE)),
  '[{"s":null,"b":null}]'
)

# json_raw() passes through verbatim: unquoted, unescaped
expect_equal(
  value_to_json(json_raw("function(cell){return 1;}")),
  "function(cell){return 1;}"
)
raw_df <- data.frame(s = 1)
raw_df$s <- list(json_raw("[1,2,3]"))
expect_equal(df_to_json(raw_df), '[{"s":[1,2,3]}]')

# =============================================================================
# CSS Utility Layer
# =============================================================================

#' Get default line color for HTML output
#'
#' Returns CSS variable reference for HTML tinytable engine,
#' otherwise returns "black". When using custom CSS with the tinytable
#' engine, users should define --tt-line-color in their CSS.
#'
#' @param x A tinytable object
#' @return Character string: "var(--tt-line-color)" or "black"
#' @keywords internal
#' @noRd
get_default_line_color <- function(x) {
  use_css_var <- identical(x@output, "html") &&
                 identical(x@html_engine, "tinytable")
  if (use_css_var) {
    "var(--tt-line-color)"
  } else if (identical(x@output, "typst")) {
    # NULL leaves the paint out of the Typst stroke, so the document's own
    # styling supplies it. See Issue #679.
    NULL
  } else {
    "black"
  }
}

style_to_css <- function(row) {
  out <- character()

  row$align <- map_align(row$align, "css")
  row$alignv <- map_alignv(row$alignv, "css")

  # simple properties
  if (isTRUE(row$bold))      out <- c(out, "font-weight: bold")
  if (isTRUE(row$italic))    out <- c(out, "font-style: italic")
  if (isTRUE(row$monospace)) out <- c(out, "font-family: monospace")
  if (isTRUE(row$smallcap))  out <- c(out, "font-variant: small-caps")
  if (!is.na(row$align))     {
    out <- c(out, paste0("text-align: ", row$align))
  }
  if (!is.na(row$alignv))    out <- c(out, paste0("vertical-align: ", row$alignv))
  if (!is.na(row$fontsize))  out <- c(out, paste0("font-size: ", format_markup_unit(row$fontsize, "em")))
  if (!is.na(row$indent))    out <- c(out, paste0("padding-left: ", format_markup_unit(row$indent, "em")))
  if (!is.na(row$color)) {
    out <- c(out, paste0("color: ", standardize_colors(row$color, "hex")))
  }
  if (!is.na(row$background)) {
    out <- c(out, paste0("background-color: ", standardize_colors(row$background, "hex")))
  }
  if (!is.na(row$html_css)) {
    out <- c(out, html_css = row$html_css)
  }

  # text decorations can be multiple
  td <- c()
  if (isTRUE(row$underline)) td <- c(td, "underline")
  if (isTRUE(row$strikeout)) td <- c(td, "line-through")
  if (length(td)) out <- c(out, paste0("text-decoration: ", paste(td, collapse = " ")))

  # combine
  out <- paste(out, collapse = "; ")
  return(out)
}

# Vectorized counterpart used by the HTML backend. Keeping the property order
# identical to style_to_css() preserves byte-for-byte CSS output while avoiding
# one-row data-frame extraction and S3 dispatch for every styled cell.
styles_to_css <- function(x) {
  n <- nrow(x)
  if (n == 0L) {
    return(character())
  }

  css <- matrix("", nrow = n, ncol = 12L)
  css[which(x$bold %in% TRUE), 1L] <- "font-weight: bold"
  css[which(x$italic %in% TRUE), 2L] <- "font-style: italic"
  css[which(x$monospace %in% TRUE), 3L] <- "font-family: monospace"
  css[which(x$smallcap %in% TRUE), 4L] <- "font-variant: small-caps"

  align <- map_align(x$align, "css")
  idx <- which(!is.na(align))
  css[idx, 5L] <- paste0("text-align: ", align[idx])

  alignv <- map_alignv(x$alignv, "css")
  idx <- which(!is.na(alignv))
  css[idx, 6L] <- paste0("vertical-align: ", alignv[idx])

  idx <- which(!is.na(x$fontsize))
  css[idx, 7L] <- paste0("font-size: ", format_markup_unit(x$fontsize[idx], "em"))
  idx <- which(!is.na(x$indent))
  css[idx, 8L] <- paste0("padding-left: ", format_markup_unit(x$indent[idx], "em"))

  color <- normalize_colors(x$color, "hex", default = NA_character_)
  idx <- which(!is.na(color))
  css[idx, 9L] <- paste0("color: ", color[idx])
  background <- normalize_colors(x$background, "hex", default = NA_character_)
  idx <- which(!is.na(background))
  css[idx, 10L] <- paste0("background-color: ", background[idx])

  idx <- which(!is.na(x$html_css))
  css[idx, 11L] <- x$html_css[idx]
  decorations <- paste(
    ifelse(x$underline %in% TRUE, "underline", ""),
    ifelse(x$strikeout %in% TRUE, "line-through", "")
  )
  decorations <- trimws(decorations)
  idx <- which(nzchar(decorations))
  css[idx, 12L] <- paste0("text-decoration: ", decorations[idx])

  apply(css, 1L, function(z) paste(z[nzchar(z)], collapse = "; "))
}


# Borders are drawn on the ::before and ::after pseudo-elements by the
# stylesheet in inst/tinytable.css. The default HTML output loads that
# stylesheet from a CDN, where an older version may still draw solid rules
# only. Emitting these rules alongside the table's own CSS makes dashed and
# dotted rules work with any version of the stylesheet.
line_type_css <- function(table_id) {
  sprintf(
    '    #%s :is(td, th)[class*="tinytable_css_"]::before { border-top: calc(var(--border-top) * var(--line-width-top)) var(--line-type-top, solid) var(--line-color-top); border-right: calc(var(--border-right) * var(--line-width-right)) var(--line-type-right, solid) var(--line-color-right); border-left: calc(var(--border-left) * var(--line-width-left)) var(--line-type-left, solid) var(--line-color-left); }
    #%s :is(td, th)[class*="tinytable_css_"]::after { height: 0; background: none; border-bottom: calc(var(--border-bottom) * var(--line-width-bottom)) var(--line-type-bottom, solid) var(--line-color-bottom); }',
    table_id,
    table_id
  )
}


line_to_css <- function(
  border_top = 0,
  border_right = 0,
  border_bottom = 0,
  border_left = 0,
  color_top = "black",
  color_right = "black",
  color_bottom = "black",
  color_left = "black",
  width_top = .1,
  width_right = .1,
  width_bottom = .1,
  width_left = .1,
  type_top = "solid",
  type_right = "solid",
  type_bottom = "solid",
  type_left = "solid",
  trim_top_left = 0,
  trim_top_right = 0,
  trim_bottom_left = 0,
  trim_bottom_right = 0,
  trim_left_top = 0,
  trim_left_bottom = 0,
  trim_right_top = 0,
  trim_right_bottom = 0
) {
  width_bottom <- format_markup_num(width_bottom)
  width_left <- format_markup_num(width_left)
  width_right <- format_markup_num(width_right)
  width_top <- format_markup_num(width_top)
  trim_bottom_left <- format_markup_num(trim_bottom_left)
  trim_bottom_right <- format_markup_num(trim_bottom_right)
  trim_left_bottom <- format_markup_num(trim_left_bottom)
  trim_left_top <- format_markup_num(trim_left_top)
  trim_right_bottom <- format_markup_num(trim_right_bottom)
  trim_right_top <- format_markup_num(trim_right_top)
  trim_top_left <- format_markup_num(trim_top_left)
  trim_top_right <- format_markup_num(trim_top_right)

  out <- sprintf('
    position: relative;
    --border-bottom: %s;
    --border-left: %s;
    --border-right: %s;
    --border-top: %s;
    --line-color-bottom: %s;
    --line-color-left: %s;
    --line-color-right: %s;
    --line-color-top: %s;
    --line-width-bottom: %sem;
    --line-width-left: %sem;
    --line-width-right: %sem;
    --line-width-top: %sem;
    --line-type-bottom: %s;
    --line-type-left: %s;
    --line-type-right: %s;
    --line-type-top: %s;
    --trim-bottom-left: %s%%;
    --trim-bottom-right: %s%%;
    --trim-left-bottom: %s%%;
    --trim-left-top: %s%%;
    --trim-right-bottom: %s%%;
    --trim-right-top: %s%%;
    --trim-top-left: %s%%;
    --trim-top-right: %s%%;
    ',
    border_bottom, border_left, border_right, border_top,
    color_bottom, color_left, color_right, color_top,
    width_bottom, width_left, width_right, width_top,
    type_bottom, type_left, type_right, type_top,
    trim_bottom_left, trim_bottom_right, trim_left_bottom, trim_left_top,
    trim_right_bottom, trim_right_top, trim_top_left, trim_top_right
  )
  gsub("\\s+", " ", out)
}




#' Internal styling function
#'
#' @inheritParams style_tt
#' @keywords internal
#' @noRd
setMethod(
  f = "style_eval",
  signature = "tinytable_html",
  definition = function(
    x,
    i = NULL,
    j = NULL,
    bold = FALSE,
    italic = FALSE,
    monospace = FALSE,
    underline = FALSE,
    strikeout = FALSE,
    color = NULL,
    background = NULL,
    fontsize = NULL,
    align = NULL,
    alignv = NULL,
    line = NULL,
    line_color = NULL,
    line_width = 0.1,
    colspan = NULL,
    rowspan = NULL,
    indent = 0,
    ...
  ) {
    # Set default line color based on context
    if (is.null(line_color)) {
      line_color <- get_default_line_color(x)
    }
    # CSS rule will be handled by finalize() via template substitution
    # Removed duplicate html_setting call that was causing CSS duplication

    # Use populated @style_other / @style_lines from build_tt()
    inputs <- style_backend_inputs(x, STYLE_PROPS_HTML)
    other <- inputs$other
    lines <- inputs$lines


    # rowspan/colspan spans first
    if (!is.null(other) && nrow(other) > 0) {
      rowspan <- if ("rowspan" %in% names(other)) {
        ifelse(is.na(other$rowspan), 1, other$rowspan)
      } else {
        rep(1, nrow(other))
      }
      colspan <- if ("colspan" %in% names(other)) {
        ifelse(is.na(other$colspan), 1, other$colspan)
      } else {
        rep(1, nrow(other))
      }
      # Skip JavaScript spans for column group headers (negative i) since HTML already handles them via colspan
      idx <- which((rowspan > 1 | colspan > 1) & other$i >= 0)
      if (length(idx) > 0) {
        # Use the factory function approach instead of individual function names
        listeners <- sprintf(
          "      window.addEventListener('load', function () { tableFns_%s.spanCell(%s, %s, %s, %s) })",
          x@id,
          other$i[idx],
          other$j[idx],
          rowspan[idx],
          colspan[idx]
        )
        x@table_string <- lines_insert(
          x@table_string,
          paste(rev(listeners), collapse = "\n"),
          "tinytable span after",
          "after"
        )
      }
    }


    # Collect all styles per cell first, then consolidate
    cell_styles <- list()

    # Dashed and dotted rules need CSS rules which may be missing from an older
    # stylesheet loaded from the CDN. Inject a scoped override when used.
    needs_line_type_css <- FALSE

    # Process line styles - vectorized approach
    if (!is.null(lines) && nrow(lines) > 0) {
      # Use interaction() for compact cell keys
      lines$cell_key <- interaction(lines$i, lines$j, drop = TRUE)

      # Precompute direction booleans once (vectorized regex)
      has_t <- grepl("t", lines$line)
      has_r <- grepl("r", lines$line)
      has_b <- grepl("b", lines$line)
      has_l <- grepl("l", lines$line)

      # Precompute trim booleans once
      lines$line_trim <- ifelse(is.na(lines$line_trim), "", lines$line_trim)
      trim_l <- grepl("l", lines$line_trim)
      trim_r <- grepl("r", lines$line_trim)

      # Map colors to hex (preserve CSS variables, fallback to "black")
      lines$color_hex <- normalize_colors(
        lines$line_color, "hex", preserve_css_vars = TRUE
      )

      # Default widths
      lines$line_width <- ifelse(is.na(lines$line_width), 0.1, lines$line_width)

      # Default line type
      if (!"line_type" %in% names(lines)) {
        lines$line_type <- NA_character_
      }
      needs_line_type_css <- any(
        !is.na(lines$line_type) & lines$line_type != "solid"
      )

      # Aggregate per direction using tapply per cell_key
      cells_unique <- levels(lines$cell_key)

      # Helper to aggregate border flags (any TRUE)
      agg_any <- function(x, cell_key, has_dir) {
        tapply(has_dir, cell_key, any, default = FALSE)[cells_unique]
      }

      # Helper to aggregate widths (max)
      agg_max_width <- function(x, cell_key, has_dir, width) {
        tapply(
          ifelse(has_dir, width, NA),
          cell_key,
          function(v) if (all(is.na(v))) 0.1 else max(v, na.rm = TRUE),
          default = 0.1
        )[cells_unique]
      }

      # Helper to aggregate colors (first non-default)
      agg_first_color <- function(x_obj, cell_key, has_dir, color) {
        default_color <- get_default_line_color(x)
        tapply(
          ifelse(has_dir, color, NA_character_),
          cell_key,
          function(v) {
            v <- v[!is.na(v)]
            if (length(v) > 0) v[1] else default_color
          },
          default = default_color
        )[cells_unique]
      }

      # Helper to aggregate line types (first non-default, mirrors colors)
      agg_first_type <- function(cell_key, has_dir, type) {
        tapply(
          ifelse(has_dir, type, NA_character_),
          cell_key,
          function(v) {
            v <- v[!is.na(v)]
            if (length(v) > 0) v[1] else "solid"
          },
          default = "solid"
        )[cells_unique]
      }

      # Aggregate border flags
      border_top <- agg_any(lines, lines$cell_key, has_t)
      border_right <- agg_any(lines, lines$cell_key, has_r)
      border_bottom <- agg_any(lines, lines$cell_key, has_b)
      border_left <- agg_any(lines, lines$cell_key, has_l)

      # Aggregate widths
      width_top <- agg_max_width(lines, lines$cell_key, has_t, lines$line_width)
      width_right <- agg_max_width(lines, lines$cell_key, has_r, lines$line_width)
      width_bottom <- agg_max_width(lines, lines$cell_key, has_b, lines$line_width)
      width_left <- agg_max_width(lines, lines$cell_key, has_l, lines$line_width)

      # Aggregate colors
      color_top <- agg_first_color(lines, lines$cell_key, has_t, lines$color_hex)
      color_right <- agg_first_color(lines, lines$cell_key, has_r, lines$color_hex)
      color_bottom <- agg_first_color(lines, lines$cell_key, has_b, lines$color_hex)
      color_left <- agg_first_color(lines, lines$cell_key, has_l, lines$color_hex)

      # Aggregate line types
      type_top <- agg_first_type(lines$cell_key, has_t, lines$line_type)
      type_right <- agg_first_type(lines$cell_key, has_r, lines$line_type)
      type_bottom <- agg_first_type(lines$cell_key, has_b, lines$line_type)
      type_left <- agg_first_type(lines$cell_key, has_l, lines$line_type)

      # Aggregate trim flags (any TRUE, multiply by 3 at end)
      trim_top_left <- tapply(trim_l & has_t, lines$cell_key, any, default = FALSE)[cells_unique] * 3
      trim_top_right <- tapply(trim_r & has_t, lines$cell_key, any, default = FALSE)[cells_unique] * 3
      trim_bottom_left <- tapply(trim_l & has_b, lines$cell_key, any, default = FALSE)[cells_unique] * 3
      trim_bottom_right <- tapply(trim_r & has_b, lines$cell_key, any, default = FALSE)[cells_unique] * 3
      # line_trim only encodes "l"/"r", so both ends of a vertical border share
      # the same trim value: compute once per side and reuse.
      trim_left <- tapply(trim_l & has_l, lines$cell_key, any, default = FALSE)[cells_unique] * 3
      trim_left_top <- trim_left_bottom <- trim_left
      trim_right <- tapply(trim_r & has_r, lines$cell_key, any, default = FALSE)[cells_unique] * 3
      trim_right_top <- trim_right_bottom <- trim_right

      # Get first (i,j) per cell_key
      first_idx <- match(cells_unique, lines$cell_key)
      cell_i <- lines$i[first_idx]
      cell_j <- lines$j[first_idx]

      # Filter to cells with at least one border
      has_border_mask <- border_top | border_right | border_bottom | border_left

      if (any(has_border_mask)) {
        # Build consolidated data.frame
        line_params <- data.frame(
          i = cell_i[has_border_mask],
          j = cell_j[has_border_mask],
          border_top = as.integer(border_top[has_border_mask]),
          border_right = as.integer(border_right[has_border_mask]),
          border_bottom = as.integer(border_bottom[has_border_mask]),
          border_left = as.integer(border_left[has_border_mask]),
          color_top = color_top[has_border_mask],
          color_right = color_right[has_border_mask],
          color_bottom = color_bottom[has_border_mask],
          color_left = color_left[has_border_mask],
          width_top = width_top[has_border_mask],
          width_right = width_right[has_border_mask],
          width_bottom = width_bottom[has_border_mask],
          width_left = width_left[has_border_mask],
          type_top = type_top[has_border_mask],
          type_right = type_right[has_border_mask],
          type_bottom = type_bottom[has_border_mask],
          type_left = type_left[has_border_mask],
          trim_top_left = trim_top_left[has_border_mask],
          trim_top_right = trim_top_right[has_border_mask],
          trim_bottom_left = trim_bottom_left[has_border_mask],
          trim_bottom_right = trim_bottom_right[has_border_mask],
          trim_left_top = trim_left_top[has_border_mask],
          trim_left_bottom = trim_left_bottom[has_border_mask],
          trim_right_top = trim_right_top[has_border_mask],
          trim_right_bottom = trim_right_bottom[has_border_mask],
          stringsAsFactors = FALSE
        )

        # Call line_to_css() once per cell via mapply (vectorized)
        css_results <- mapply(
          line_to_css,
          border_top = line_params$border_top,
          border_right = line_params$border_right,
          border_bottom = line_params$border_bottom,
          border_left = line_params$border_left,
          color_top = line_params$color_top,
          color_right = line_params$color_right,
          color_bottom = line_params$color_bottom,
          color_left = line_params$color_left,
          width_top = line_params$width_top,
          width_right = line_params$width_right,
          width_bottom = line_params$width_bottom,
          width_left = line_params$width_left,
          type_top = line_params$type_top,
          type_right = line_params$type_right,
          type_bottom = line_params$type_bottom,
          type_left = line_params$type_left,
          trim_top_left = line_params$trim_top_left,
          trim_top_right = line_params$trim_top_right,
          trim_bottom_left = line_params$trim_bottom_left,
          trim_bottom_right = line_params$trim_bottom_right,
          trim_left_top = line_params$trim_left_top,
          trim_left_bottom = line_params$trim_left_bottom,
          trim_right_top = line_params$trim_right_top,
          trim_right_bottom = line_params$trim_right_bottom,
          SIMPLIFY = TRUE,
          USE.NAMES = FALSE
        )

        # Filter non-empty CSS and build cell_styles list
        has_css <- nzchar(css_results)
        if (any(has_css)) {
          for (idx in which(has_css)) {
            cell_key <- paste(line_params$i[idx], line_params$j[idx], sep = "_")
            cell_styles[[cell_key]] <- list(
              i = line_params$i[idx],
              j = line_params$j[idx],
              css = "",
              css_before = "",
              css_after = "",
              has_border = TRUE,
              consolidated_line_css = css_results[idx]
            )
          }
        }
      }
    }

    # Process other styles - vectorized
    if (!is.null(other) && nrow(other) > 0) {
      css_rules <- styles_to_css(other)

      # Filter to non-empty CSS
      has_css <- nzchar(css_rules)
      if (any(has_css)) {
        other_final <- other[has_css, , drop = FALSE]
        css_final <- css_rules[has_css]
        # (i,j) keys are unique: @style_other has one row per cell
        cell_keys_final <- paste(other_final$i, other_final$j, sep = "_")

        # Build cell_styles entries
        for (idx in seq_along(css_final)) {
          cell_key <- cell_keys_final[idx]

          if (is.null(cell_styles[[cell_key]])) {
            cell_styles[[cell_key]] <- list(
              i = other_final$i[idx],
              j = other_final$j[idx],
              css = "",
              css_before = "",
              css_after = "",
              has_border = FALSE
            )
          }

          cell_styles[[cell_key]]$css <- css_final[idx]
        }
      }
    }

    # Early exit if no styles
    if (length(cell_styles) == 0) {
      return(x)
    }

    # Convert cell_styles list to data.frame for vectorized operations
    n_cells <- length(cell_styles)
    css_df <- data.frame(
      i = integer(n_cells),
      j = integer(n_cells),
      line_css = character(n_cells),
      other_css = character(n_cells),
      has_border = logical(n_cells),
      stringsAsFactors = FALSE
    )

    css_df$i <- vapply(cell_styles, `[[`, numeric(1), "i")
    css_df$j <- vapply(cell_styles, `[[`, numeric(1), "j")
    css_df$line_css <- vapply(
      cell_styles,
      function(z) z$consolidated_line_css %||% "",
      character(1)
    )
    css_df$other_css <- vapply(cell_styles, `[[`, character(1), "css")
    css_df$has_border <- vapply(cell_styles, `[[`, logical(1), "has_border")

    # Merge line_css and other_css
    has_line <- nzchar(css_df$line_css)
    has_other <- nzchar(css_df$other_css)

    css_df$merged_css <- ifelse(
      has_line & has_other,
      paste(css_df$line_css, css_df$other_css, sep = "; "),
      ifelse(has_line, css_df$line_css, css_df$other_css)
    )

    # Filter to non-empty CSS
    css_df <- css_df[nzchar(css_df$merged_css), , drop = FALSE]

    if (nrow(css_df) == 0) {
      return(x)
    }

    # Group by CSS rule and border status using interaction
    css_df$group_key <- interaction(css_df$merged_css, css_df$has_border, drop = TRUE)

    # For each group, collect cells (split() preserves the factor level order,
    # matching the previous per-level scan)
    css_groups <- lapply(split(css_df, css_df$group_key), function(group_rows) {
      list(
        css_rule = group_rows$merged_css[1],
        has_border = group_rows$has_border[1],
        i = group_rows$i,
        j = group_rows$j,
        first_j = group_rows$j[1],
        first_i = group_rows$i[1]
      )
    })

    # Generate CSS rules for each unique group
    # Sort groups by the first cell's coordinates to ensure deterministic ordering
    group_keys_sorted <- names(css_groups)
    if (length(group_keys_sorted) > 0) {
      # Two-key sort: column first, then row (a combined j * 10000 + i key
      # collides for tables with 10000+ rows)
      first_j <- sapply(css_groups, function(g) g$first_j, USE.NAMES = FALSE)
      first_i <- sapply(css_groups, function(g) g$first_i, USE.NAMES = FALSE)
      group_keys_sorted <- group_keys_sorted[order(first_j, first_i)]
    }

    style_arrays <- css_entries <- character(length(group_keys_sorted))
    for (group_idx in seq_along(group_keys_sorted)) {
      group_key <- group_keys_sorted[group_idx]
      group_data <- css_groups[[group_key]]
      css_rule <- group_data$css_rule
      has_border <- group_data$has_border

      # Generate unique ID for this CSS rule
      id_css <- get_id(stem = "tinytable_css_")

      # Generate position array for all cells with this styling
      cell_positions <- sprintf("{ i: '%s', j: %s }", group_data$i, group_data$j)
      arr <- c(
        "          {",
        " positions: [ ",
        paste(cell_positions, collapse = ", "),
        " ],",
        " css_id: '",
        id_css,
        "',",
        "}, "
      )
      arr <- paste(arr, collapse = "")
      style_arrays[group_idx] <- arr

      # Generate CSS entry - scoped to table ID to prevent CSS cascade conflicts
      table_id <- paste0("tinytable_", x@id)
      entry <- sprintf(
        "    #%s td.%s, #%s th.%s { %s }",
        table_id, id_css, table_id, id_css, css_rule
      )
      css_entries[group_idx] <- entry
    }

    if (needs_line_type_css) {
      css_entries <- c(line_type_css(paste0("tinytable_", x@id)), css_entries)
    }

    if (length(style_arrays) > 0) {
      x@table_string <- lines_insert_after_fixed(
        x@table_string,
        c("tinytable style arrays after", "tinytable css entries after"),
        c(
          paste(rev(style_arrays), collapse = "\n"),
          paste(rev(css_entries), collapse = "\n")
        )
      )
    }


    return(x)
  }
)

#' Create Plot Vector
#'
#' @details
#' This function creates plots by creating a temporary `tt()` object and applying `plot_tt()`. It returns a character vector containing the image paths or HTML tags for the plots.
#'
#' @param output Output format. One of "html", "latex", "typst", "markdown", etc.
#' @inheritParams plot_tt
#' @return A character vector with plot file paths or HTML tags.
#' @export
#' @examples
#' \dontrun{
#' # Create histogram plots
#' plot_data <- list(rnorm(100), rnorm(50))
#' plot_vector(fun = "histogram", data = plot_data, output = "html")
#'
#' # Create density plots
#' plot_vector(fun = "density", data = plot_data, output = "latex")
#'
#' # Create bar plots from single values
#' bar_data <- list(0.5, 0.8, 0.3)
#' plot_vector(fun = "barpct", data = bar_data, output = "html")
#' }
plot_vector <- function(
    output = "html",
    fun = NULL,
    data = NULL,
    color = "black",
    xlim = NULL,
    height = 1,
    height_plot = 400,
    width_plot = 1200,
    images = NULL,
    sprintf = "%s",
    assets = "tinytable_assets",
    ...) {
  # Determine the length based on data or images
  if (!is.null(data)) {
    len <- length(data)
  } else if (!is.null(images)) {
    len <- length(images)
  } else {
    stop("Either `data` or `images` must be provided.", call. = FALSE)
  }

  # Create a temporary data frame with empty strings
  df <- data.frame(x = rep("", len), stringsAsFactors = FALSE)
  tt_obj <- tt(df, output = output)

  # Apply plotting (specify j = 1 to target the single column)
  result <- plot_tt(
    tt_obj,
    j = 1,
    fun = fun,
    data = data,
    color = color,
    xlim = xlim,
    height = height,
    height_plot = height_plot,
    width_plot = width_plot,
    images = images,
    sprintf = sprintf,
    assets = assets,
    ...
  )

  # Build the table to trigger plot generation
  built_result <- build_tt(result, output = output)

  # Extract the plot paths/HTML from the built table
  plot_data <- built_result@data_body$x
  return(plot_data)
}

apply_typst_spans <- function(body, sty) {
  # spans must be replaced before concatenating strings
  # Only process spans for positive row indices (data body), not headers (negative indices)
  spans <- sty[
    which(
      ((!is.na(sty$colspan) & sty$colspan > 1) |
        (!is.na(sty$rowspan) & sty$rowspan > 1)) &
        (!is.na(sty$i) & sty$i > 0)
    ),
    ,
    drop = FALSE
  ]
  if (nrow(spans) > 0) {
    # Deduplicate spans by i, j, colspan, and rowspan to avoid applying the same span multiple times
    # This can happen when line styles are expanded (e.g., "lt" becomes "l" and "t" entries)
    spans <- spans[!duplicated(spans[, c("i", "j", "colspan", "rowspan")]), , drop = FALSE]
    table_nrows <- nrow(body)
    table_ncols <- ncol(body)

    for (idx in seq_len(nrow(spans))) {
      rowspan <- spans[idx, "rowspan"]
      colspan <- spans[idx, "colspan"]
      row_idx <- spans[idx, "i"]
      col_idx <- spans[idx, "j"]

      # Sanity checks for span dimensions
      if (!is.na(colspan) && (col_idx + colspan - 1) > table_ncols) {
        stop(sprintf(
          "colspan of %d at column %d exceeds table width of %d columns",
          colspan,
          col_idx,
          table_ncols
        ))
      }
      if (!is.na(rowspan) && (row_idx + rowspan - 1) > table_nrows) {
        stop(sprintf(
          "rowspan of %d at row %d exceeds table height of %d rows",
          rowspan,
          row_idx,
          table_nrows
        ))
      }

      # Build table.cell() arguments
      cell_args <- character(0)
      if (!is.na(colspan) && colspan > 1) {
        cell_args <- c(cell_args, sprintf("colspan: %s", colspan))
      }
      if (!is.na(rowspan) && rowspan > 1) {
        cell_args <- c(cell_args, sprintf("rowspan: %s", rowspan))
      }

      # spanning cell
      body[row_idx, col_idx] <- sprintf(
        "table.cell(%s)%s",
        paste(cell_args, collapse = ", "),
        body[row_idx, col_idx]
      )

      # empty cells
      row_span <- if (!is.na(rowspan)) rowspan else 1
      col_span <- if (!is.na(colspan)) colspan else 1
      row_idx_empty <- seq(row_idx, row_idx + row_span - 1)
      col_idx_empty <- seq(col_idx, col_idx + col_span - 1)
      empty <- expand.grid(i = row_idx_empty, j = col_idx_empty)
      empty <- empty[empty$i != row_idx | empty$j != col_idx, , drop = FALSE]
      for (k in seq_len(nrow(empty))) {
        body[empty[k, "i"], empty[k, "j"]] <- NA
      }
    }
  }
  return(body)
}

setMethod(
  f = "build_eval",
  signature = "tinytable_typst",
  definition = function(x, ...) {
    out <- typst_template()
    out <- typst_body(x, out)
    out <- typst_header(x, out)
    out <- typst_widths(x, out)
    out <- typst_notes(x, out)
    x@table_string <- out
    return(x)
  }
)

# Helper function to load the Typst template
typst_template <- function() {
  out <- readLines(system.file("templates/typst.typ", package = "tinytable"))
  paste(out, collapse = "\n")
}

# Helper function to process table body
typst_body <- function(x, out) {
  # Prepare body data
  body <- apply(x@data_body, 2, function(k) paste0("[", k, "]"), simplify = FALSE)
  body <- do.call(cbind, body)

  # Apply colspan and rowspan transformations
  body <- apply_typst_spans(body, x@style)

  if (nrow(x@data_body) && is.null(dim(body))) {
    body <- matrix(body)
  }

  # Convert body to Typst format
  body <- apply(body, 1, function(k) {
    clean_k <- stats::na.omit(k)
    if (length(clean_k) > 0) {
      paste(clean_k, collapse = ", ")
    } else {
      NA_character_ # Mark completely empty rows for removal
    }
  })

  # Keep only non-empty rows
  body <- stats::na.omit(body)
  body <- paste(body, collapse = ",\n")
  body <- paste0(body, ",\n")

  lines_insert(out, body, "tinytable cell content after", "after")
}

# Identify the column-group underlines created by `group_tt(j = ...)`
#
# Those rules carry a `line_trim` marker, which every other backend uses to
# leave a gap between the rules of two adjacent groups. Typst's
# `table.hline()` snaps to column boundaries and has no trim option, so two
# adjacent group rules merge into a single long line and the reader cannot
# tell where one group ends and the next begins. We therefore draw them
# inside the group header cell with `place()`, which is bounded by the cell's
# inner width and so leaves the gap.
#
# Returns the rules to draw (one row per group span) and a logical vector
# flagging the rows of `x@style_lines` they replace, so `style_eval()` can
# skip them instead of drawing the same rule twice.
#' @keywords internal
#' @noRd
typst_group_line_rules <- function(x) {
  lines <- x@style_lines
  n_lines <- if (is.null(lines)) 0L else nrow(lines)
  out <- list(rules = NULL, consumed = rep(FALSE, n_lines))
  if (n_lines == 0 || nrow(x@group_data_j) == 0) {
    return(out)
  }

  # Two unspecified values match: a NA `line_color` means "inherit the paint
  # from the document", which is as much a shared property as an explicit color.
  eq <- function(a, b) (is.na(a) & is.na(b)) | (!is.na(a) & !is.na(b) & a == b)

  rules <- list()
  n_head <- nrow(x@group_data_j)
  for (row_idx in seq_len(n_head)) {
    # The bottom group row is at i = -1, the one above it at i = -2, etc.
    i_style <- -(n_head - row_idx + 1)
    spans <- parse_group_spans(as.character(x@group_data_j[row_idx, ]))
    for (span_idx in seq_len(nrow(spans))) {
      cols <- spans$start[span_idx]:spans$end[span_idx]
      bottom <- lines$i == i_style & lines$line == "b" & lines$j %in% cols
      trimmed <- which(bottom & !is.na(lines$line_trim))
      if (length(trimmed) == 0) {
        next
      }
      width <- lines$line_width[trimmed[1]]
      color <- lines$line_color[trimmed[1]]
      sel <- which(
        bottom & eq(lines$line_width, width) & eq(lines$line_color, color)
      )
      # Only replace a rule that covers the whole span: a partial rule would
      # be drawn in the wrong place by a cell-wide `place()`.
      if (!all(cols %in% lines$j[sel])) {
        next
      }
      out$consumed[sel] <- TRUE
      rules[[length(rules) + 1L]] <- data.frame(
        header_row = row_idx,
        start = spans$start[span_idx],
        end = spans$end[span_idx],
        line_width = width,
        line_color = color,
        stringsAsFactors = FALSE
      )
    }
  }

  if (length(rules) > 0) {
    out$rules <- do.call(rbind, rules)
  }
  out
}

# Helper function to process header
typst_header <- function(x, out) {
  # Collect all header lines in correct visual order (top to bottom)
  all_headers <- character(0)

  # Add group headers (first call first = top of table)
  if (nrow(x@group_data_j) > 0) {
    rules <- typst_group_line_rules(x)$rules
    for (row_idx in 1:nrow(x@group_data_j)) {
      group_row <- as.character(x@group_data_j[row_idx, ])
      row_rules <- if (is.null(rules)) {
        NULL
      } else {
        rules[rules$header_row == row_idx, , drop = FALSE]
      }
      header_line <- typst_build_group_header(group_row, rules = row_rules)
      if (!is.null(header_line)) {
        all_headers <- c(all_headers, header_line)
      }
    }
  }

  # Add regular column headers (closest to data = bottom of header)
  header <- !is.null(colnames(x)) && length(colnames(x)) > 0
  if (header) {
    header <- paste(paste0("[", colnames(x), "]"), collapse = ", ")
    header <- paste0(header, ",")
    all_headers <- c(all_headers, header)
  }

  # Insert all headers at once
  if (length(all_headers) > 0) {
    all_headers_text <- paste(all_headers, collapse = "\n")
    out <- lines_insert(out, all_headers_text, "repeat: true", "after")
  }

  out
}

# Helper function to process column widths
#
# Three regimes:
# 1. Explicit `width` (scalar or vector): fixed percentage columns, unchanged.
#    A scalar is split equally across columns; a vector gives each column its
#    own share with a total of `sum(width)`.
# 2. Auto width with notes: proportional columns. Typst sizes `auto` columns
#    to fit every cell, including the `table.footer` note cell that spans the
#    full table, so a long note stretches the table to the page width and
#    distorts the column proportions (#669). Instead we measure each column's
#    natural width and convert the measurements to `fr` units, with the table
#    total capped at the natural/page width. The measurement ignores the
#    footer, so notes wrap at the table width instead of dictating it.
# 3. Auto width, no notes: plain `auto` columns, unchanged.
typst_widths <- function(x, out) {
  if (length(x@width) == 1) {
    width <- rep(sprintf("%.2f%%", x@width / ncol(x) * 100), ncol(x))
    width <- sprintf("    columns: (%s),", paste(width, collapse = ", "))
    return(lines_insert(out, width, "tinytable table start", "after"))
  }
  if (length(x@width) > 1) {
    width <- sprintf("%.2f%%", x@width * 100)
    width <- sprintf("    columns: (%s),", paste(width, collapse = ", "))
    return(lines_insert(out, width, "tinytable table start", "after"))
  }

  if (length(x@notes) == 0) {
    width <- paste(rep("auto", ncol(x)), collapse = ", ")
    width <- sprintf("    columns: (%s),", width)
    return(lines_insert(out, width, "tinytable table start", "after"))
  }

  # Proportional columns: measure natural column widths in Typst itself.
  # Group headers and spanning cells are excluded from the measurement; they
  # span several columns, so they do not pin down any single column's width.
  coldata <- apply(x@data_body, 2, function(k) {
    paste0("[", k, "]", collapse = ", ")
  })
  if (!is.null(colnames(x)) && length(colnames(x)) > 0) {
    coldata <- paste0("[", colnames(x), "], ", coldata)
  }
  coldata <- sprintf("    (%s),", coldata)
  coldata <- paste(
    c("  #let tinytable-coldata = (", coldata, "  )", ""),
    collapse = "\n"
  )

  # `grid` with the same inset as `table` measures the same natural width,
  # but is immune to the `show table.cell` styling rule, whose styles are
  # keyed by cell position and would misapply in a single-column layout.
  total <- "calc.min(tinytable-naturals.sum(), size.width)"
  wrapper_open <- paste(
    c(
      coldata,
      "  #context layout(size => {",
      "    let tinytable-naturals = tinytable-coldata.map(col => measure(grid(columns: 1, inset: 5pt, ..col)).width)",
      sprintf("    let tinytable-total = %s", total),
      "    block(width: tinytable-total)[",
      ""
    ),
    collapse = "\n"
  )

  out <- lines_insert(out, wrapper_open, "tinytable align-figure before", "after")
  out <- lines_insert(
    out,
    "    columns: tinytable-naturals.map(w => w.pt() * 1fr),",
    "tinytable table start",
    "after"
  )
  out <- lines_insert(out, "\n  ]\n  })", "end table", "after")
  out
}

# Helper function to process notes
typst_notes <- function(x, out) {
  if (length(x@notes) == 0) {
    return(out)
  }

  # Add footer structure
  ft <- "
    table.footer(
      repeat: false,
      // tinytable notes after
    ),
    "
  out <- lines_insert(out, ft, "tinytable footer after", "after")

  # Process each note
  notes <- rev(x@notes)
  if (is.null(names(notes))) {
    lab <- rep("", length(notes))
  } else {
    lab <- names(notes)
  }

  notes <- sapply(notes, function(n) if (is.list(n)) n$text else n)

  note_text <- vapply(
    seq_along(notes),
    function(k) typst_note(notes[k], lab[k], ncol(x)),
    character(1)
  )
  if (length(note_text) > 0) {
    # Repeated insertion after one marker reverses the input, so reverse the
    # batch to preserve the existing byte-for-byte output order.
    out <- lines_insert(
      out,
      paste(rev(note_text), collapse = "\n"),
      "tinytable notes after",
      "after"
    )
  }

  out
}

# Helper function to format a single note
typst_note <- function(note, label, ncols) {
  if (label == "") {
    sprintf(
      "    table.cell(align: left, colspan: %s, %s),",
      ncols,
      note
    )
  } else {
    l <- sprintf("[#super[%s] ", label)
    n <- sub("[", l, note, fixed = TRUE)
    tmp <- sprintf(
      "    table.cell(align: left, colspan: %s, %s),",
      ncols,
      n
    )
    sub("text(, ", "text(", tmp, fixed = TRUE)
  }
}

# Helper function to build Typst group header from group row data
#
# `rules` (see typst_group_line_rules()) holds the underlines to draw inside
# the span cells rather than as table-wide `table.hline()` calls.
typst_build_group_header <- function(group_row, rules = NULL) {
  spans <- parse_group_spans(group_row)

  header_parts <- character(0)
  pos <- 1

  emit_empty_until <- function(parts, from, to) {
    if (to >= from) {
      parts <- c(parts, rep("[ ]", to - from + 1))
    }
    parts
  }

  # `place()` is bounded by the cell's inner width, so the rule stops short of
  # the column edge on both sides: adjacent group rules stay separated.
  span_rule <- function(start, end) {
    if (is.null(rules) || nrow(rules) == 0) {
      return("")
    }
    hit <- which(rules$start == start & rules$end == end)
    if (length(hit) == 0) {
      return("")
    }
    width <- rules$line_width[hit[1]]
    width <- format_markup_unit(if (is.na(width)) 0.1 else width, "em")
    color <- normalize_colors(rules$line_color[hit[1]], "typst", default = NA_character_)
    stroke <- if (is.na(color)) width else sprintf("%s + %s", width, color)
    sprintf(
      " #place(bottom, dy: 0.4em, line(length: 100%%, stroke: %s))",
      stroke
    )
  }

  for (span_idx in seq_len(nrow(spans))) {
    # Empty cells for uncovered columns before this span
    header_parts <- emit_empty_until(header_parts, pos, spans$start[span_idx] - 1)

    span_length <- spans$end[span_idx] - spans$start[span_idx] + 1
    rule <- span_rule(spans$start[span_idx], spans$end[span_idx])
    if (span_length > 1) {
      # Multi-column span - use table.cell with colspan
      header_parts <- c(header_parts, sprintf(
        "table.cell(colspan: %s, align: center)[%s%s]",
        span_length,
        spans$label[span_idx],
        rule
      ))
    } else {
      # Single column - just centered content
      header_parts <- c(
        header_parts,
        sprintf("[%s%s]", spans$label[span_idx], rule)
      )
    }
    pos <- spans$end[span_idx] + 1
  }

  # Trailing uncovered columns
  header_parts <- emit_empty_until(header_parts, pos, length(group_row))

  if (length(header_parts) > 0) {
    paste0(paste(header_parts, collapse = ", "), ",")
  } else {
    NULL
  }
}

#' Save a Tiny Table to File
#'
#' This function saves an object of class tinytable to a specified file and format, with an option to overwrite existing files.
#'
#' @param x The tinytable object to be saved.
#' @param output String or file path.
#' + If `output` is "markdown", "latex", "html", or "typst", the table is returned in a string as an `R` object.
#' + If `output` is "dataframe", the formatted table is returned as a data frame.
#' + If `output` is a valid file path, the table is saved to file. The supported extensions are: .docx, .html, .png, .pdf, .tex, .typ, and .md (with aliases .txt, .Rmd and .qmd).
#' + If `theme_html(portable = TRUE)` or `theme_typst(portable = TRUE)` is used,
#' images are included as base64 encoded strings instead of links to local files.
#' @param overwrite A logical value indicating whether to overwrite an existing file.
#' @return A string with the table when `output` is a format, and the file path when `output` is a valid path.
#' @template dependencies
#' @template latex_preamble
#' @template global_options
#' @export
#' @examples
#' library(tinytable)
#' x <- mtcars[1:4, 1:5]
#'
#' fn <- file.path(tempdir(), "test.html")
#' tt(x) |> save_tt(fn, overwrite = TRUE)
#'
#' library(tinytable)
#' filename <- file.path(tempdir(), "table.tex")
#' tt(mtcars[1:4, 1:4]) |> save_tt(filename)
#'
save_tt <- function(
    x,
    output = get_option("tinytable_save_output", default = NULL),
    overwrite = get_option("tinytable_save_overwrite", default = FALSE)) {
  assert_class(x, "tinytable")
  assert_string(output)
  assert_flag(overwrite)


  if (identical(output, "markdown")) {
    out <- build_tt(x, output = "markdown")@table_string
    return(as.character(out))
  } else if (identical(output, "html")) {
    out <- build_tt(x, output = "html")@table_string
    return(as.character(out))
  } else if (identical(output, "latex")) {
    out <- build_tt(x, output = "latex")@table_string
    return(as.character(out))
  } else if (identical(output, "typst")) {
    out <- build_tt(x, output = "typst")@table_string
    return(as.character(out))
  } else if (identical(output, "dataframe")) {
    out <- build_tt(x, output = "dataframe")@data_body
    out <- as.data.frame(lapply(out, trimws))
    colnames(out) <- NULL
    return(out)
  }

  # normalize to an absolute path immediately: the PDF branch below changes
  # the working directory, which would otherwise re-resolve a relative path
  # (e.g. "sub/table.pdf" ending up in "sub/sub/table.pdf")
  output <- file.path(
    normalizePath(dirname(output), mustWork = TRUE),
    basename(output)
  )

  if (file.exists(output) && !overwrite) {
    stop("File already exists and overwrite is set to FALSE.", call. = FALSE)
  }

  x@output_dir <- dirname(output)

  file_ext <- tolower(tools::file_ext(output))

  output_format <- switch(file_ext,
    "png" = "html",
    "html" = "html",
    "pdf" = "latex",
    "tex" = "latex",
    "md" = "markdown",
    "rmd" = "markdown",
    "qmd" = "markdown",
    "txt" = "markdown",
    "docx" = "markdown",
    "typ" = "typst",
    stop(
      "The supported file extensions are: .png, .html, .pdf, .tex, .typ, .qmd, .Rmd, .txt, .docx, and .md. Supported output formats are: markdown, latex, typst, html, and dataframe.",
      call. = FALSE
    )
  )

  # evaluate styles at the very end of the pipeline, just before writing
  x <- build_tt(x, output = output_format)

  if (file_ext %in% c("html", "tex", "md", "rmd", "qmd", "txt", "typ")) {
    write(x@table_string, file = output)
  } else if (file_ext == "png") {
    assert_dependency("webshot2")
    # this doesn't work in tempdir() for some reason.
    # probably webshot2's fault. we need to build in `output`
    tmp <- file.path(dirname(output), paste0(get_id(), ".html"))
    write(x@table_string, file = tmp)
    webshot2::webshot(
      tmp,
      file = output,
      selector = "body > div > table",
      zoom = 4,
      quiet = TRUE
    )
    unlink(tmp)
  } else if (file_ext == "pdf") {
    assert_dependency("tinytex")
    # \documentclass{standalone} does not support \begin{table}
    tmp <- strsplit(x@table_string, "\\n")[[1]]
    # tmp <- tmp[!grepl("\\begin{table}", tmp, fixed = TRUE)]
    # tmp <- tmp[!grepl("\\end{table}", tmp, fixed = TRUE)]
    tmp <- paste(tmp, collapse = "\n")
    tmp <- sprintf(latex_standalone, tmp)
    # tinytex is fiddly with file paths, so we need to hack
    # it by changing the working directory
    wd <- getwd()
    on.exit(setwd(wd))
    temp_wd <- dirname(output)
    setwd(temp_wd)
    f <- paste0(get_id(), ".tex")
    write(tmp, f)
    # tinytex sometimes generates logfiles when there are warnings, e.g because
    # table is too wide. See #260.
    # We delete additional logfiles generated by tinytex
    existing_log_files <- list.files(
      temp_wd,
      pattern = "\\.log$",
      full.names = TRUE
    )

    # render
    engine <- x@latex_engine
    assert_choice(
      engine,
      c("xelatex", "pdflatex", "lualatex"),
      name = "latex_engine"
    )
    tinytex::latexmk(f, pdf_file = output, engine = engine)

    # clean
    flag <- get_option("tinytable_pdf_clean", default = TRUE)
    assert_flag(flag, name = "tinytable_pdf_clean")
    if (flag) {
      new_log_files <- setdiff(
        list.files(temp_wd, pattern = "\\.log$", full.names = TRUE),
        existing_log_files
      )
      invisible(file.remove(new_log_files))
    }
    unlink(f)
  } else if (file_ext == "docx") {
    assert_dependency("pandoc")
    pandoc::pandoc_convert(text = x@table_string, to = "docx", output = output)
  }

  return(invisible(path.expand(output)))
}

latex_standalone <- "
\\documentclass{standalone}
\\usepackage{tabularray}
\\usepackage{graphicx}
\\usepackage{rotating}
\\usepackage{float}
\\usepackage[normalem]{ulem}
\\usepackage[x11names, svgnames]{xcolor}
\\UseTblrLibrary{siunitx}
\\newcommand{\\tinytableTabularrayUnderline}[1]{\\underline{#1}}
\\newcommand{\\tinytableTabularrayStrikeout}[1]{\\sout{#1}}
\\NewTableCommand{\\tinytableDefineColor}[3]{\\definecolor{#1}{#2}{#3}}
\\begin{document}
\\minipage{\\textwidth}
%s
\\endminipage
\\end{document}
"

source("helpers.R")
using("tinysnapshot")
options(tinytable_print_output = "dataframe")

x <- head(iris)
tab <- tt(x) |>
    group_tt(i = list("Hello" = 2, "World" = 2)) |>
    save_tt("dataframe")
expect_true("Hello" %in% tab[[1]])
expect_true("World" %in% tab[[1]])

# data frame print uses markdown
expect_snapshot_print(tab, "dataframe-print.md")

# Issue #218
options(tinytable_save_output = "dataframe")
tab <- data.frame(x = 1, y = Inf) |>
    tt() |>
    save_tt()
expect_inherits(tab, "data.frame")


# Issue #513: uninformative error
tab <- tt(head(iris)) |>
    group_tt(j = list("a" = 1:2, "b" = 3:5)) |>
    style_tt("colnames", italic = TRUE)
tab <- tab |> save_tt("dataframe")
expect_inherits(tab, "data.frame")

source("helpers.R")

if (!is_local) {
  exit_file("Run on Vincent's machine")
}

x <- cbind(mtcars[1, ], mtcars[1, ])
dest_dir <- file.path(tempdir(), "tt-pdf-test")
dir.create(dest_dir, showWarnings = FALSE)
dest_pdf <- file.path(dest_dir, "out.pdf")
dest_log <- file.path(dest_dir, "somelog.log")

suppressWarnings(
  tt(x) |>
    theme_latex(resize_direction = "down") |>
    save_tt(dest_pdf, overwrite = TRUE)
)

# logfiles are automatically deleted
expect_equal(
  length(grep("\\.log$", list.files(dest_dir))),
  0
)

# logfiles that exist before call to tinytex are left untouched
cat("some content", file = dest_log)

suppressWarnings(
  tt(x) |>
    theme_latex(resize_direction = "down") |>
    save_tt(dest_pdf, overwrite = TRUE)
)

expect_true("somelog.log" %in% list.files(dest_dir))

# Issue #395
x <- mtcars[1:4, 1:5]
fn <- tempfile(fileext = ".pdf")
cap <- "A simple \\texttt{tinytable} example."
not <- "Nullam odio est, ullamcorper scelerisque lectus a, eleifend luctus nisl. Etiam ullamcorper, nibh vel interdum auctor, odio nulla mollis tortor, vel fringilla ante quam quis est."
tt(x, caption = cap, notes = not, width = .5) |>
  style_tt(i = 1:3, j = 1:2, background = "#1ecebf", bold = TRUE) |>
  group_tt(j = list("Halloumi" = 1:2, "Tofu" = 4:5)) |>
  save_tt(fn, overwrite = TRUE)

# Clean slate
suppressWarnings(invisible(file.remove(list.files(dest_dir))))

source("helpers.R")

# Issue #414
df1 = df2 = data.frame(a = c(1, 2, 3), b = c(4, 5, 6))
colnames(df1) = NULL
expect_equal(length(colnames(tt(df1, rownames = TRUE))), 0)
expect_equal(length(colnames(tt(df1))), 0)

tab1 = save_tt(tt(df1), "markdown")
tab2 = save_tt(tt(df2, colnames = FALSE), "markdown")
expect_equal(tab1, tab2)


# colnames="label"
dat <- mtcars[1:5, c("cyl", "mpg", "hp")]
attr(dat$cyl, "label") <- "Cylinders"
attr(dat$mpg, "label") <- "Miles per Gallon"
attr(dat$hp, "label") <- "Horse Power"
tab <- tt(dat, colnames = "label") |> save_tt("dataframe")
expect_equal(unlist(tab[1, ]), c("Cylinders", "Miles per Gallon", "Horse Power"))

source("helpers.R")

if (!is_local) {
  exit_file("Run on Vincent's machine")
}

x <- cbind(mtcars[1, ], mtcars[1, ])
dest_dir <- file.path(tempdir(), "tt-pdf-test")
dir.create(dest_dir, showWarnings = FALSE)
dest_pdf <- file.path(dest_dir, "out.pdf")
dest_log <- file.path(dest_dir, "somelog.log")

suppressWarnings(
  tt(x) |>
    theme_latex(resize_direction = "down") |>
    save_tt(dest_pdf, overwrite = TRUE)
)

# logfiles are automatically deleted
expect_equal(
  length(grep("\\.log$", list.files(dest_dir))),
  0
)

# logfiles that exist before call to tinytex are left untouched
cat("some content", file = dest_log)

suppressWarnings(
  tt(x) |>
    theme_latex(resize_direction = "down") |>
    save_tt(dest_pdf, overwrite = TRUE)
)

expect_true("somelog.log" %in% list.files(dest_dir))

# Issue #395
x <- mtcars[1:4, 1:5]
fn <- tempfile(fileext = ".pdf")
cap <- "A simple \\texttt{tinytable} example."
not <- "Nullam odio est, ullamcorper scelerisque lectus a, eleifend luctus nisl. Etiam ullamcorper, nibh vel interdum auctor, odio nulla mollis tortor, vel fringilla ante quam quis est."
tt(x, caption = cap, notes = not, width = .5) |>
  style_tt(i = 1:3, j = 1:2, background = "#1ecebf", bold = TRUE) |>
  group_tt(j = list("Halloumi" = 1:2, "Tofu" = 4:5)) |>
  save_tt(fn, overwrite = TRUE)

# Clean slate
suppressWarnings(invisible(file.remove(list.files(dest_dir))))

# save_tt() with a relative sub-directory path
wd <- getwd()
rel_root <- tempfile("tt-pdf-rel")
dir.create(file.path(rel_root, "sub"), recursive = TRUE)
setwd(rel_root)
res <- tryCatch(
  suppressWarnings(
    save_tt(tt(mtcars[1:2, 1:3]), "sub/out.pdf", overwrite = TRUE)
  ),
  error = function(e) e,
  finally = setwd(wd)
)
expect_false(inherits(res, "error"))
expect_true(file.exists(file.path(rel_root, "sub", "out.pdf")))
unlink(rel_root, recursive = TRUE)

# =============================================================================
# Canonical style properties and per-backend capability matrix
# =============================================================================

# Properties shared by every backend keep-list.
STYLE_PROPS_BASE <- c(
  "bold", "italic", "underline", "strikeout", "monospace", "smallcap",
  "align", "alignv", "color", "background", "fontsize", "indent"
)

# Canonical cell-style properties handled by the style pipeline
# (rectangular @style_other grid construction and resolution).
STYLE_PROPS <- c(STYLE_PROPS_BASE, "html_css", "colspan", "rowspan")

# Per-backend keep-lists passed to filter_style_other(). This is the single
# place where the backend capability matrix is declared. Do not extend a
# backend's list without implementing the corresponding rendering support.
STYLE_PROPS_HTML <- STYLE_PROPS
STYLE_PROPS_TABULARRAY <- c(STYLE_PROPS_BASE, "colspan", "rowspan")
STYLE_PROPS_TYPST <- STYLE_PROPS_BASE
STYLE_PROPS_TABULATOR <- STYLE_PROPS_BASE
# Grid (markdown) handles a smaller subset: no monospace, align, alignv, or
# fontsize. Order matters here: prepare_grid_style() also uses this vector to
# select and order the columns of the filtered style data frame.
STYLE_PROPS_GRID <- c(
  "bold", "italic", "strikeout", "underline", "smallcap", "indent",
  "color", "background", "colspan", "rowspan"
)

#' Common style_eval entry: filtered cell styles + line styles
#'
#' @param x A built tinytable object with populated @style_other/@style_lines
#' @param keep Character vector of style properties the backend renders
#' @return list(other = filtered @style_other, lines = @style_lines or NULL)
#' @keywords internal
#' @noRd
style_backend_inputs <- function(x, keep) {
  other <- filter_style_other(x@style_other, keep)
  lines <- x@style_lines
  if (nrow(lines) == 0) {
    lines <- NULL
  }
  list(other = other, lines = lines)
}


# =============================================================================
# Centralized alignment code maps shared across backends
# =============================================================================

#' Map tinytable horizontal alignment codes to backend-specific values
#'
#' Vectorized; unmatched codes (including NA) pass through unchanged.
#' All targets currently share the same map: "d" (decimal) renders as
#' "center". The LaTeX backend is intentionally absent: it keeps the raw
#' l/c/r codes and routes "d" to siunitx d-columns separately.
#'
#' @param code Character vector of alignment codes (l, c, d, r)
#' @param target Backend name: "css", "typst", or "tabulator"
#' @return Character vector of backend-specific alignment values
#' @keywords internal
#' @noRd
map_align <- function(code, target = c("css", "typst", "tabulator")) {
  target <- match.arg(target)
  map <- c(l = "left", c = "center", d = "center", r = "right")
  idx <- which(code %in% names(map))
  code[idx] <- map[code[idx]]
  code
}

#' Map tinytable vertical alignment codes to backend-specific values
#'
#' Vectorized; unmatched codes (including NA) pass through unchanged.
#' The Tabulator engine uses the "css" target (top/middle/bottom).
#'
#' @param code Character vector of vertical alignment codes (t, m, b)
#' @param target Backend name: "css", "typst", or "tabularray"
#' @return Character vector of backend-specific vertical alignment values
#' @keywords internal
#' @noRd
map_alignv <- function(code, target = c("css", "typst", "tabularray")) {
  target <- match.arg(target)
  map <- switch(target,
    css = c(t = "top", m = "middle", b = "bottom"),
    typst = c(t = "top", m = "horizon", b = "bottom"),
    tabularray = c(t = "h", m = "m", b = "f")
  )
  idx <- which(code %in% names(map))
  code[idx] <- map[code[idx]]
  code
}

#' Style a Vector with Text Formatting
#'
#' @details
#' This function applies styling to a vector. It allows customization of text style (bold, italic, monospace), text color, font size, and text decorations (underline, strikeout). The styling is applied element-wise to the vector. Vectors are coerced with `as.character()` before styling.
#'
#' @param x A vector to be styled.
#' @param output Output format for styling. One of "html", "latex", "typst", "markdown", "ansi". Defaults to "html".
#' @param bold Logical; if `TRUE`, text is styled in bold. Must be of length 1 or `length(x)`.
#' @param italic Logical; if `TRUE`, text is styled in italic. Must be of length 1 or `length(x)`.
#' @param monospace Logical; if `TRUE`, text is styled in monospace font. Must be of length 1 or `length(x)`.
#' @param smallcap Logical; if `TRUE`, text is styled in small caps. In Markdown output format, text is converted to uppercase. Must be of length 1 or `length(x)`.
#' @param underline Logical; if `TRUE`, text is underlined. Must be of length 1 or `length(x)`.
#' @param strikeout Logical; if `TRUE`, text has a strike through line. Must be of length 1 or `length(x)`.
#' @param color Text color. Must be of length 1 or `length(x)`. There are several ways to specify colors:
#' + HTML:
#'   - Hex code composed of # and 6 characters, ex: #CC79A7.
#'   - Keywords: black, silver, gray, white, maroon, red, purple, fuchsia, green, lime, olive, yellow, navy, blue, teal, aqua
#' + LaTeX:
#'   - Hex code composed of # and 6 characters, ex: "#CC79A7".
#'   - Keywords: black, blue, brown, cyan, darkgray, gray, green, lightgray, lime, magenta, olive, orange, pink, purple, red, teal, violet, white, yellow.
#'   - Color blending using xcolor, ex: `white!80!blue`, `green!20!red`.
#'   - Color names with luminance levels from the `ninecolors` package.
#' @param fontsize Font size in em units. Must be of length 1 or `length(x)`. Can be `NULL` for default size.
#' @param indent Text indentation in em units. Must be of length 1 or `length(x)`. Positive values only.
#' @return A character vector with applied styling.
#' @export
#' @examples
#' # Basic styling
#' style_vector(c("Hello", "World"), bold = TRUE, color = "red")
#'
#' # Different styles per element
#' style_vector(
#'   c("Bold text", "Italic text", "Monospace"),
#'   bold = c(TRUE, FALSE, FALSE),
#'   italic = c(FALSE, TRUE, FALSE),
#'   monospace = c(FALSE, FALSE, TRUE)
#' )
#'
#' # Single style applied to all elements
#' style_vector(c("A", "B", "C"), color = "blue", fontsize = 1.2)
style_vector <- function(
    x,
    output = "html",
    bold = FALSE,
    italic = FALSE,
    monospace = FALSE,
    smallcap = FALSE,
    underline = FALSE,
    strikeout = FALSE,
    color = NULL,
    fontsize = NULL,
    indent = NULL) {
  assert_atomic_vector(x)

  # Convert input to character if needed
  if (!is.character(x)) {
    x <- as.character(x)
  }

  # Input validation
  assert_choice(output, choice = c("html", "latex", "typst", "markdown", "ansi"))

  # Length validation - all styling arguments must be length 1 or length(x)
  n <- length(x)
  assert_length(bold, len = c(1, n), name = "bold")
  assert_length(italic, len = c(1, n), name = "italic")
  assert_length(monospace, len = c(1, n), name = "monospace")
  assert_length(smallcap, len = c(1, n), name = "smallcap")
  assert_length(underline, len = c(1, n), name = "underline")
  assert_length(strikeout, len = c(1, n), name = "strikeout")
  assert_length(color, len = c(1, n), null.ok = TRUE, name = "color")
  assert_length(fontsize, len = c(1, n), null.ok = TRUE, name = "fontsize")
  assert_length(indent, len = c(1, n), null.ok = TRUE, name = "indent")

  # Type validation
  assert_logical(bold, name = "bold")
  assert_logical(italic, name = "italic")
  assert_logical(monospace, name = "monospace")
  assert_logical(smallcap, name = "smallcap")
  assert_logical(underline, name = "underline")
  assert_logical(strikeout, name = "strikeout")
  assert_character(color, null.ok = TRUE, name = "color")
  assert_numeric(fontsize, null.ok = TRUE, name = "fontsize")
  assert_numeric(indent, lower = 0, null.ok = TRUE, name = "indent")

  # Return early if empty vector
  if (n == 0) {
    return(character(0))
  }

  # Recycle arguments to match vector length
  bold <- rep_len(bold, n)
  italic <- rep_len(italic, n)
  monospace <- rep_len(monospace, n)
  smallcap <- rep_len(smallcap, n)
  underline <- rep_len(underline, n)
  strikeout <- rep_len(strikeout, n)

  if (!is.null(color)) {
    color <- rep_len(standardize_colors(color), n)
  }
  if (!is.null(fontsize)) {
    fontsize <- rep_len(fontsize, n)
  }
  if (!is.null(indent)) {
    indent <- rep_len(indent, n)
  }

  # Build styles data frame
  styles <- data.frame(
    bold = bold,
    italic = italic,
    monospace = monospace,
    smallcap = smallcap,
    underline = underline,
    strikeout = strikeout,
    color = if (is.null(color)) rep(NA_character_, n) else color,
    fontsize = if (is.null(fontsize)) rep(NA_real_, n) else fontsize,
    indent = if (is.null(indent)) rep(NA_real_, n) else indent,
    stringsAsFactors = FALSE
  )

  # Select appropriate styling function based on output format
  style_fun <- switch(output,
    "html" = style_string_html,
    "latex" = style_string_latex,
    "typst" = style_string_typst,
    "markdown" = style_string_markdown,
    "ansi" = style_string_ansi
  )

  # Apply styling to each element
  styled_x <- character(n)
  for (i in seq_len(n)) {
    # Extract style for current element
    current_style <- lapply(styles[i, ], function(col) {
      if (is.na(col)) NULL else col
    })
    # Convert logical NA to FALSE for style functions
    current_style$bold <- isTRUE(current_style$bold)
    current_style$italic <- isTRUE(current_style$italic)
    current_style$monospace <- isTRUE(current_style$monospace)
    current_style$smallcap <- isTRUE(current_style$smallcap)
    current_style$underline <- isTRUE(current_style$underline)
    current_style$strikeout <- isTRUE(current_style$strikeout)

    styled_x[i] <- style_fun(x[i], current_style)
  }

  return(styled_x)
}

style_eval_grid <- function(x) {
  # For grid formats, styling is handled inside build_eval_grid
  # This allows proper ordering of text styles before padding and background after padding
  return(x)
}


prepare_grid_style <- function(x) {
  # Use populated @style_other from build_tt()
  sty <- x@style_other

  # Return early if no styles
  if (nrow(sty) == 0) {
    return(sty)
  }

  sty <- filter_style_other(sty, STYLE_PROPS_GRID)

  # Select only the columns needed for grid styling
  sty <- sty[, c("i", "j", STYLE_PROPS_GRID), drop = FALSE]

  return(sty)
}


#' Apply text styling to table data before padding
#' @keywords internal
#' @noRd
style_grid_body <- function(x) {
  sty <- prepare_grid_style(x)

  if (nrow(sty) == 0) {
    return(x)
  }

  # Determine the styling function to use based on output type and ANSI setting
  style_string_grid <- if (isTRUE(x@ansi)) {
    style_string_ansi
  } else {
    style_string_markdown
  }

  # Apply text styling to each cell (excluding background)
  for (idx in seq_len(nrow(sty))) {
    row <- sty[idx, "i"]
    col <- sty[idx, "j"]

    # when calling subset(select -x) we can have too many columns in @style
    if (col > ncol(x)) next

    # Prepare styles list for the current cell
    styles <- list(
      bold = isTRUE(sty[idx, "bold"]),
      italic = isTRUE(sty[idx, "italic"]),
      strikeout = isTRUE(sty[idx, "strikeout"]),
      underline = isTRUE(sty[idx, "underline"]),
      smallcap = isTRUE(sty[idx, "smallcap"]),
      color = if (!is.na(sty[idx, "color"])) sty[idx, "color"] else NULL,
      indent = if (!is.na(sty[idx, "indent"])) sty[idx, "indent"] else NULL
    )

    # Handle column names (i = 0)
    if (row == 0) {
      current_name <- colnames(x)[col]
      if (!identical(trimws(current_name), "")) {
        tmp <- style_string_grid(current_name, styles)
        colnames(x)[col] <- tmp
      }
    } else if (row < 0) {
      # Handle group headers (negative i)
      if (nrow(x@group_data_j) > 0) {
        # Convert negative row index to positive index in group_data_j
        group_row <- abs(row)
        if (group_row <= nrow(x@group_data_j) && col <= ncol(x@group_data_j)) {
          current_value <- x@group_data_j[group_row, col]
          if (!is.na(current_value) && !identical(trimws(current_value), "")) {
            x@group_data_j[group_row, col] <- style_string_grid(
              current_value,
              styles
            )
          }
        }
      }
    } else {
      # Handle main table body (positive i)
      current_value <- x@data_body[row, col]
      if (!identical(trimws(current_value), "")) {
        x@data_body[row, col] <- style_string_grid(current_value, styles)
      }
    }

    # wipe adjacent cells for colspan/rowspan
    rowspan <- sty[idx, "rowspan"]
    colspan <- sty[idx, "colspan"]
    rowspan <- if (is.na(rowspan)) 1 else rowspan
    colspan <- if (is.na(colspan)) 1 else colspan
    wipe <- expand.grid(
      i = row:(row + rowspan - 1),
      j = col:(col + colspan - 1)
    )
    wipe <- wipe[
      !(wipe$i == row & wipe$j == col) &
        wipe$i >= 1 &
        wipe$i <= nrow(x@data_body) &
        wipe$j >= 1 &
        wipe$j <= ncol(x@data_body),
    ]
    if (nrow(wipe) > 0) {
      x@data_body[cbind(wipe$i, wipe$j)] <- ""
    }
  }

  return(x)
}


#' Apply background styling to padded table matrix
#' @keywords internal
#' @noRd
style_grid_body_background <- function(tab, x, header) {
  sty <- prepare_grid_style(x)

  if (nrow(sty) == 0) {
    return(tab)
  }

  # Determine the styling function to use based on output type and ANSI setting
  style_string_grid <- if (isTRUE(x@ansi)) {
    style_string_ansi
  } else {
    style_string_markdown
  }

  # Apply only background styling to each cell
  for (idx in seq_len(nrow(sty))) {
    # Skip if no background styling
    if (is.na(sty[idx, "background"])) {
      next
    }

    row <- sty[idx, "i"]
    col <- sty[idx, "j"]

    # Prepare styles list with only background
    styles <- list(background = sty[idx, "background"])

    # Handle column names (i = 0)
    if (row == 0 && header) {
      tab_row <- 1 # Header is first row in tab matrix
      if (tab_row <= nrow(tab) && col <= ncol(tab)) {
        current_content <- tab[tab_row, col]
        tab[tab_row, col] <- style_string_grid(current_content, styles)
      }
    } else if (row > 0) {
      # Handle main table body (positive i)
      tab_row <- if (header) row + 1 else row # Adjust for header row
      if (tab_row <= nrow(tab) && col <= ncol(tab)) {
        current_content <- tab[tab_row, col]
        tab[tab_row, col] <- style_string_grid(current_content, styles)
      }
    }
    # Handle group headers would go here if needed (negative i)
  }

  return(tab)
}


###### MUST BE PLACED AFTER style_eval_grid definition
#' tinytable S4 method
#'
#' @keywords internal
setMethod(
  f = "style_eval",
  signature = "tinytable_grid",
  definition = style_eval_grid
)


#' tinytable S4 method
#'
#' @keywords internal
setMethod(
  f = "style_eval",
  signature = "tinytable_dataframe",
  definition = style_eval_grid
)


#' tinytable S4 method
#'
#' @keywords internal
setMethod(
  f = "finalize",
  signature = "tinytable_dataframe",
  definition = identity
)

#' Helper functions for row insertion in group_tt
#'
#' @keywords internal
#' @noRd
group_tt_ij_k <- function(x, i, j) {
  # Convert list input to matrix format for unified processing
  converted_from_list <- FALSE
  if (is.list(i) && is.null(j)) {
    # Convert list like list("Blah blah" = 2) to matrix insertion format
    positions <- unlist(i)
    labels <- names(i)

    if (is.null(labels) || anyNA(labels) || any(labels == "")) {
      stop(
        "All list entries must have (non-missing) names for group labels.",
        call. = FALSE
      )
    }

    # Create matrix with labels repeated in all columns (recycled by column)
    matrix_data <- matrix(labels, nrow = length(labels), ncol = ncol(x))

    # Convert to matrix insertion format
    i <- positions
    j <- matrix_data
    converted_from_list <- TRUE
  }

  # Ensure j is a character matrix (coerce if needed)
  if (!is.character(j)) {
    j <- as.character(j)
  }

  # If x has more than 1 column and j is a 1-column matrix, try to reshape j
  if (is.matrix(j) && ncol(x) > 1 && ncol(j) == 1) {
    total_elements <- nrow(j) * ncol(j)
    if (total_elements %% ncol(x) == 0) {
      # Reshape j to have the same number of columns as x
      j <- matrix(
        j,
        nrow = total_elements / ncol(x),
        ncol = ncol(x),
        byrow = TRUE
      )
    }
  }

  # Check that matrix width matches table width
  if (is.matrix(j) && ncol(j) != ncol(x)) {
    stop(
      sprintf(
        "Matrix must have the same number of columns as the table (%d columns)",
        ncol(x)
      ),
      call. = FALSE
    )
  }

  # Validate row insertion positions against table size
  # For vector-based grouping (consecutive groups), the indices represent group boundaries
  # and can be larger than the table size since they represent insertion points
  if (any(i > nrow(x) + 1)) {
    # Check if this is a vector-based grouping case (consecutive groups)
    # In this case, the indices represent where groups should be inserted
    # and can be larger than the current table size
    if (length(i) > 1 && all(diff(i) >= 0)) {
      # This looks like consecutive grouping, allow it
    } else {
      stop(sprintf("`i` should be smaller than %s", nrow(x) + 2), call. = FALSE)
    }
  }

  # If single position but multiple matrix rows, replicate the position
  matrix_rows <- nrow(j)
  if (length(i) == 1 && matrix_rows > 1) {
    i <- rep(i, matrix_rows)
  }

  # Position 0 is a documented synonym for the first row after the header
  i[i == 0] <- 1

  # The insertion-offset arithmetic in process_group_i() assumes ascending
  # positions: stable-sort positions (and their matrix rows) so unsorted
  # input, e.g. list("B" = 5, "A" = 2), is placed correctly
  if (length(i) > 1) {
    ord <- order(i)
    i <- i[ord]
    if (is.matrix(j) && nrow(j) == length(i)) {
      j <- j[ord, , drop = FALSE]
    }
  }

  list(i, j, converted_from_list)
}

# Make colors more uniform: R definition -> LaTeX definition -> default
standardize_colors <- function(col, format = "hex") {
  if (length(col) == 1 && is.na(col)) {
    return(NA)
  }

  normalize <- isTRUE(getOption("tinytable_color_name_normalization", default = TRUE))
  if (!normalize) {
    return(col)
  }

  single_color <- function(k) {
    # NA means "no color"; do not let it reach col2rgb() (which returns white)
    if (is.na(k)) {
      return(NA_character_)
    }

    # Early return for hex colors
    # do this here to allow typst rgb() wrap later
    if (isTRUE(grepl("^#", k))) {
      return(k)
    }

    # Skip processing for "black" and "white"
    if (k %in% c("black", "white")) {
      return(k)
    }

    # R colors
    out <- try(grDevices::col2rgb(k), silent = TRUE)

    if (!inherits(out, "try-error")) {
      # Convert RGB values to hex format
      out <- grDevices::rgb(out[1], out[2], out[3], maxColorValue = 255)
    } else if (format != "tabularray" && k %in% latex_colors$name) {
      # LaTeX colors (skip for tabularray format)
      hex_val <- latex_colors$rgb[latex_colors$name == k]
      out <- paste0("#", hex_val)
    } else {
      # If still no match, return the original value
      out <- k
    }
    return(out)
  }
  result <- sapply(col, single_color)

  # Format for Typst if needed
  if (format == "typst") {
    result <- ifelse(
      is.character(result) & grepl("^#", result),
      sprintf('rgb("%s")', result),
      result
    )
  }

  # Format for ANSI if needed
  if (format == "ansi") {
    format_ansi_color <- function(color_val) {
      if (is.na(color_val)) {
        return(NA)
      }

      # If it's a hex color, convert to RGB ANSI
      if (grepl("^#[0-9A-Fa-f]{6}$", color_val)) {
        hex_color <- substr(color_val, 2, 7) # Remove #
        r <- as.integer(paste0("0x", substr(hex_color, 1, 2)))
        g <- as.integer(paste0("0x", substr(hex_color, 3, 4)))
        b <- as.integer(paste0("0x", substr(hex_color, 5, 6)))
        return(sprintf("38;2;%d;%d;%d", r, g, b))
      }

      # Check if it's a named ANSI color
      ansi_colors <- list(
        "black" = "30", "red" = "31", "green" = "32", "yellow" = "33",
        "blue" = "34", "magenta" = "35", "cyan" = "36", "white" = "37",
        "gray" = "90", "grey" = "90", "bright_red" = "91", "bright_green" = "92",
        "bright_yellow" = "93", "bright_blue" = "94", "bright_magenta" = "95",
        "bright_cyan" = "96", "bright_white" = "97"
      )

      color_code <- ansi_colors[[tolower(color_val)]]
      if (!is.null(color_code)) {
        return(color_code)
      }

      # If no match, return original
      return(color_val)
    }

    result <- sapply(result, format_ansi_color, USE.NAMES = FALSE)
  }

  return(result)
}


#' Build a named map from unique user-supplied colors to normalized values
#'
#' Calls standardize_colors() once per unique non-NA value, in order of first
#' appearance. Callers that need the raw map (e.g. for later name lookups or
#' to register LaTeX preamble entries) use this directly; others go through
#' normalize_colors().
#'
#' @param values Character vector of user-supplied colors (may contain NA)
#' @param format Passed to standardize_colors ("hex", "tabularray", "typst")
#' @return Named character vector: original color -> normalized color
#' @keywords internal
#' @noRd
build_color_map <- function(values, format) {
  vals <- unique(values[!is.na(values)])
  if (length(vals) == 0) {
    return(character(0))
  }
  stats::setNames(
    sapply(vals, standardize_colors, format = format, USE.NAMES = FALSE),
    vals
  )
}

#' Apply a prebuilt color map to a vector of colors
#'
#' @param values Character vector of user-supplied colors (may contain NA)
#' @param map Named character vector from build_color_map()
#' @param default Value used for NA and unmatched entries
#' @param preserve_css_vars Keep unmatched var(--...) values untouched
#'   (used by the HTML lines path)
#' @return Character vector of normalized colors
#' @keywords internal
#' @noRd
apply_color_map <- function(values, map, default = "black", preserve_css_vars = FALSE) {
  out <- rep(default, length(values))
  idx <- which(!is.na(values) & values %in% names(map))
  out[idx] <- unname(map[values[idx]])
  if (preserve_css_vars) {
    idx <- which(!is.na(values) & !(values %in% names(map)) & grepl("^var\\(", values))
    out[idx] <- values[idx]
  }
  out
}

#' Normalize a vector of user-supplied colors via a unique-value map
#'
#' Convenience wrapper: build_color_map() + apply_color_map().
#'
#' @inheritParams apply_color_map
#' @param format Passed to standardize_colors ("hex", "tabularray", "typst")
#' @keywords internal
#' @noRd
normalize_colors <- function(values, format, default = "black", preserve_css_vars = FALSE) {
  apply_color_map(
    values,
    build_color_map(values, format),
    default = default,
    preserve_css_vars = preserve_css_vars
  )
}

# (pseudo-)unique IDs
# Restore .Random.seed on exit so table creation does not consume the user's
# global RNG stream. Reseed with a per-call unique value (time + counter) so
# consecutive ids still differ even though the user's seed is restored.
get_id_counter <- local({
  i <- 0L
  function() {
    i <<- i + 1L
    i
  }
})

get_id <- function(stem = "id") {
  if (exists(".Random.seed", globalenv(), inherits = FALSE)) {
    old <- get(".Random.seed", globalenv())
    on.exit(assign(".Random.seed", old, globalenv()), add = TRUE)
  } else {
    on.exit(
      if (exists(".Random.seed", globalenv(), inherits = FALSE)) {
        rm(".Random.seed", envir = globalenv())
      },
      add = TRUE
    )
  }
  set.seed(as.integer(Sys.time()) %% 1000000L + get_id_counter())
  id <- sample(c(0:9, letters), 20, replace = TRUE)
  paste0(stem, paste(id, collapse = ""))
}

# Clamp colspan values so spans never extend past the last column
clamp_colspan <- function(style, ncol) {
  if (nrow(style) > 0 && "colspan" %in% colnames(style)) {
    end <- style$j + style$colspan - 1
    style$colspan <- ifelse(!is.na(end) & end > ncol, style$colspan - (end - ncol), style$colspan)
  }
  style
}

# getOption with deprecation warnings
get_option <- function(x, default = NULL) {
  deprecated <- c(
    # old = new
    "tinytable_grid_hlines" = "tinytable_markdown_hlines",
    "tinytable_save_pdf_clean" = "tinytable_pdf_clean",
    "tinytable_save_pdf_engine" = "tinytable_pdf_engine"
  )
  if (x %in% names(deprecated)) {
    x_new <- deprecated[[x]]
    warning(
      sprintf("Option `%s` is deprecated. Use `%s` instead.", x, x_new),
      call. = FALSE
    )
    x <- x_new
  }
  getOption(x, default = default)
}

format_markup_num <- function(x) {
  out <- format(
    x,
    scientific = FALSE,
    trim = TRUE,
    digits = 15,
    decimal.mark = ".",
    drop0trailing = TRUE
  )
  sub("^-0$", "0", out)
}

format_markup_unit <- function(x, unit) {
  paste0(format_markup_num(x), unit)
}

ttempdir <- function() {
  d <- tempdir()
  d <- file.path(d, "tinytable")
  # start fresh
  if (dir.exists(d)) {
    unlink(d, recursive = TRUE)
  }
  dir.create(d)
  return(d)
}

lines_drop_consecutive_empty <- function(x) {
  if (!nzchar(x)) {
    return(x)
  }

  # strsplit() drops one terminal delimiter; mirror that behavior before
  # collapsing runs. The regex matches repeated *identical* whitespace-only
  # lines, which is the exact behavior of the former rle(lines) implementation,
  # without allocating and rejoining every line in large rendered tables.
  if (endsWith(x, "\n")) {
    x <- substr(x, 1L, nchar(x) - 1L)
  }
  gsub(
    "(?m)^([ \\t\\r]*)(?:\\n\\1)+(?=\\n|$)",
    "\\1",
    x,
    perl = TRUE
  )
}

lines_drop <- function(
    old,
    regex,
    position = "equal",
    fixed = FALSE,
    unique = TRUE,
    perl = FALSE) {
  assert_choice(position, c("equal", "before", "after", "all"))
  lines <- strsplit(old, "\n")[[1]]
  idx <- grep(regex, lines, fixed = fixed, perl = perl)
  if (isTRUE(unique) && length(idx) > 1 && position != "all") {
    stop(
      "The `regex` supplied `lines_drop()` did not match a unique line.",
      call. = FALSE
    )
  }
  if (length(idx) > 0) {
    if (position == "equal") {
      lines <- lines[!seq_along(lines) %in% idx]
    } else if (position == "before") {
      lines <- lines[idx:length(lines)]
    } else if (position == "after") {
      lines <- lines[1:idx]
    } else if (position == "all") {
      lines <- lines[!seq_along(lines) %in% idx]
    }
  }
  out <- paste(lines, collapse = "\n")
  return(out)
}

lines_drop_between <- function(text, regex_start, regex_end, fixed = FALSE, perl = FALSE) {
  lines <- strsplit(text, "\n")[[1]]
  idx_start <- grep(regex_start, lines, fixed = fixed, perl = perl)
  idx_end <- grep(regex_end, lines, fixed = fixed, perl = perl)
  if (length(idx_start) != 1) {
    stop("The `regex_start` did not match a unique line.", call. = FALSE)
  }
  if (length(idx_end) != 1) {
    stop("The `regex_end` did not match a unique line.", call. = FALSE)
  }
  if (idx_start >= idx_end) {
    stop("`regex_start` matches a line after `regex_end`.", call. = FALSE)
  }
  keep <- setdiff(seq_along(lines), idx_start:idx_end)
  out <- paste(lines[keep], collapse = "\n")
  return(out)
}

lines_insert <- function(old, new, regex, position = c("before", "after"),
                         fixed = FALSE, perl = FALSE, occurrence = 1L,
                         require_unique = FALSE) {
  position <- match.arg(position)
  lines <- strsplit(old, "\n", fixed = TRUE)[[1]]
  hits <- grep(regex, lines, fixed = fixed, perl = perl)

  if (length(hits) == 0L || anyNA(hits)) stop("`regex` did not match.", call. = FALSE)
  if (require_unique && length(hits) != 1L) stop("`regex` did not match a unique line.", call. = FALSE)
  if (occurrence > length(hits)) stop("`occurrence` out of range.", call. = FALSE)

  idx <- hits[occurrence]
  after <- if (position == "before") idx - 1L else idx
  out <- append(lines, values = new, after = after)
  paste(out, collapse = "\n")
}

# Fast path for inserting content after several unique fixed-text marker lines.
# Character offsets are resolved first and the output is assembled once.
lines_insert_after_fixed <- function(old, markers, values) {
  stopifnot(length(markers) == length(values))
  marker_pos <- vapply(
    markers,
    function(marker) regexpr(marker, old, fixed = TRUE)[1L],
    integer(1)
  )
  if (any(marker_pos < 1L)) {
    stop("`marker` did not match.", call. = FALSE)
  }

  line_end <- vapply(seq_along(marker_pos), function(k) {
    suffix <- substring(old, marker_pos[k])
    rel <- regexpr("\n", suffix, fixed = TRUE)[1L]
    if (rel < 0L) nchar(old) else marker_pos[k] + rel - 1L
  }, integer(1))
  ord <- order(line_end)
  line_end <- line_end[ord]
  values <- values[ord]

  pieces <- character(length(line_end) * 2L + 1L)
  start <- 1L
  p <- 1L
  for (k in seq_along(line_end)) {
    pieces[p] <- substr(old, start, line_end[k])
    pieces[p + 1L] <- paste0(values[k], "\n")
    start <- line_end[k] + 1L
    p <- p + 2L
  }
  pieces[p] <- substr(old, start, nchar(old))
  paste0(pieces, collapse = "")
}


rbind_nocol <- function(...) {
  dflist <- list(...)
  out <- lapply(dflist, function(k) stats::setNames(k, seq_len(ncol(k))))
  do.call(rbind, out)
}


...get <- function(x, ifnotfound = NULL) {
  eval(
    quote(
      if (!anyNA(.m1 <- match(.x, ...names())) && !is.null(.m2 <- ...elt(.m1))) {
        .m2
      } else {
        .ifnotfound
      }),
    pairlist(.x = x[1L], .ifnotfound = ifnotfound),
    parent.frame(1L)
  )
}


#' Helper function to get value with fallback
#' @param value Primary value
#' @param fallback Fallback value
#' @return Value or fallback
#' @keywords internal
#' @noRd
`%||%` <- function(value, fallback) {
  if (is.null(value)) fallback else value
}


#' Parse a column-group header row into spans.
#'
#' Canonical span parser shared by all backends. `group_row` is a character
#' vector where `NA` means "no group" and `""` (or whitespace-only) means
#' "continuation of the preceding label". A non-empty label starts a span;
#' following empty strings extend it; `NA` or a new non-empty label closes it.
#' Empty strings with no open label (e.g. at the start of the row or after an
#' `NA`) belong to no span and are skipped.
#'
#' @param group_row Character vector (one element per table column).
#' @return A `data.frame` with columns `label` (raw, untrimmed), `start`, and
#'   `end`, one row per span, in column order. Duplicate labels are preserved.
#' @keywords internal
#' @noRd
parse_group_spans <- function(group_row) {
  labels <- character(0)
  starts <- integer(0)
  ends <- integer(0)

  pos <- 1L
  n <- length(group_row)
  while (pos <= n) {
    current_label <- group_row[pos]

    # NA = ungrouped column: no span
    if (is.na(current_label)) {
      pos <- pos + 1L
      next
    }

    span_start <- pos
    pos <- pos + 1L

    # Consume continuation cells (empty strings) following this position
    while (pos <= n && !is.na(group_row[pos]) && trimws(group_row[pos]) == "") {
      pos <- pos + 1L
    }

    # Only record spans that start with a non-empty label; orphan empty
    # strings (no preceding label) are skipped along with their run.
    if (trimws(current_label) != "") {
      labels <- c(labels, current_label)
      starts <- c(starts, span_start)
      ends <- c(ends, pos - 1L)
    }
  }

  data.frame(
    label = labels,
    start = starts,
    end = ends,
    stringsAsFactors = FALSE
  )
}


percent_sum_100 <- function(x, digits = 0) {
  stopifnot(is.numeric(x), all(is.finite(x)), all(x >= 0))
  s <- sum(x)
  if (s <= 0) stop("sum(x) must be > 0")
  p <- x / s

  units <- 100L * 10^digits
  scaled <- p * units
  base <- floor(scaled)
  rem <- scaled - base

  leftover <- units - sum(base) # how many 1/10^digits to distribute
  if (leftover != 0) {
    ix <- order(rem, decreasing = TRUE, method = "radix")
    base[ix[seq_len(leftover)]] <- base[ix[seq_len(leftover)]] + 1L
  }

  out <- base / 10^digits
  return(out)
}

# Unit tests for internal helpers in R/utils.R
lines_drop <- tinytable:::lines_drop
lines_drop_between <- tinytable:::lines_drop_between

txt <- "alpha\nbravo\ncharlie\ndelta"

# lines_drop_between: start marker on the first line
expect_equal(lines_drop_between(txt, "alpha", "bravo"), "charlie\ndelta")

# lines_drop_between: end marker on the last line
expect_equal(lines_drop_between(txt, "charlie", "delta"), "alpha\nbravo")

# lines_drop: no match is a no-op for all four positions
for (pos in c("equal", "before", "after", "all")) {
  expect_equal(lines_drop(txt, "zzz", position = pos), txt)
}

# get_id() must not disturb the user's RNG state (building an HTML table
# calls get_id() internally)
set.seed(1)
r1 <- rnorm(1)
set.seed(1)
void <- tinytable::save_tt(tinytable::tt(data.frame(x = 1)), "html")
r2 <- rnorm(1)
expect_equal(r1, r2)

setMethod(
  f = "finalize",
  signature = "tinytable_typst",
  definition = function(x, ...) {
    out <- x@table_string

    cap <- x@caption
    if (length(cap) == 1) {
      out <- sub(
        "$TINYTABLE_TYPST_CAPTION",
        sprintf("caption: %s,", cap),
        out,
        fixed = TRUE
      )
    } else {
      out <- sub("$TINYTABLE_TYPST_CAPTION", "", out, fixed = TRUE)
    }

    # drop the full header block if there are no colnames or group headers
    if (length(x@names) == 0 && nrow(x@group_data_j) == 0) {
      out <- lines_drop_between(
        out,
        regex_start = "// tinytable header start",
        regex_end = "// tinytable header end",
        fixed = TRUE
      )
    }

    if (isTRUE(check_dependency("knitr")) && isTRUE(knitr::pandoc_to("typst"))) {
      lab <- knitr::opts_current$get()[["label"]]
      cap <- knitr::opts_current$get()[["tbl-cap"]]
      # Remove figure environment from template and let Quarto use its own
      if (!is.null(lab) || !is.null(cap)) {
        out <- lines_drop_between(
          out,
          regex_start = "// start preamble figure",
          regex_end = "// end preamble figure",
          fixed = TRUE
        )
        out <- lines_drop(out, regex = "// start preamble figure", fixed = TRUE)
        out <- lines_drop(out, regex = "// end figure", fixed = TRUE)
        out <- sub(" table(", " #table(", out, fixed = TRUE)
      }
    } else {
      # here we kept tinytable's #figure[] so we need to use block[]
      out <- sub("#block", "block", out, fixed = TRUE)
    }

    x@table_string <- out

    return(x)
  })

#' Format a Vector
#'
#' @details
#' This function formats a vector by passing it to `format_tt()`. All formatting arguments must be of length 1 or `length(x)`.
#'
#' @param x A vector to be formatted.
#' @param output Output format. One of "html", "latex", "typst", "markdown", etc.
#' @inheritParams format_tt
#' @return A character vector with formatted values.
#' @export
#' @examples
#' # Format numeric vector
#' format_vector(c(1234.567, 9876.543), digits = 2, num_mark_big = ",")
#'
#' # Format dates
#' dates <- as.Date(c("2023-01-01", "2023-12-31"))
#' format_vector(dates, date = "%B %d, %Y")
#'
#' # Format logical values
#' format_vector(c(TRUE, FALSE, TRUE), bool = function(x) ifelse(x, "Yes", "No"))
format_vector <- function(
    x,
    output = "html",
    digits = NULL,
    num_fmt = "significant",
    num_zero = FALSE,
    num_suffix = FALSE,
    num_mark_big = "",
    num_mark_dec = getOption("OutDec", default = "."),
    date = NULL,
    bool = NULL,
    math = FALSE,
    other = NULL,
    replace = FALSE,
    escape = FALSE,
    markdown = FALSE,
    quarto = FALSE,
    fn = NULL,
    sprintf = NULL,
    linebreak = NULL) {
  assert_atomic_vector(x)

  # Convert vector to data frame and then to tinytable
  df <- data.frame(x = x, stringsAsFactors = FALSE)
  tt_obj <- tt(df, output = output)

  # Apply formatting (specify j = 1 to target the single column)
  result <- format_tt(
    tt_obj,
    j = 1,
    digits = digits,
    num_fmt = num_fmt,
    num_zero = num_zero,
    num_suffix = num_suffix,
    num_mark_big = num_mark_big,
    num_mark_dec = num_mark_dec,
    date = date,
    bool = bool,
    math = math,
    other = other,
    replace = replace,
    escape = escape,
    markdown = markdown,
    quarto = quarto,
    fn = fn,
    sprintf = sprintf,
    linebreak = linebreak
  )

  # Build the table to trigger rendering (including markdown)
  built_result <- build_tt(result, output = output)

  # Extract the formatted vector values from the built table
  formatted_data <- built_result@data_body$x
  return(formatted_data)
}

# =============================================================================
# HELPER FUNCTIONS
# =============================================================================

# Style arguments evaluated eagerly in style_tt() and forwarded to the lazy
# call. Keep in sync with the formals of style_tt() and style_tt_lazy().
# (i and j are excluded: they go through non-standard evaluation.)
STYLE_TT_ARGS <- c(
  "bold", "italic", "monospace", "smallcap", "underline", "strikeout",
  "color", "background", "fontsize", "align", "alignv", "colspan", "rowspan",
  "indent", "line", "line_color", "line_width", "line_type", "line_trim",
  "finalize"
)

# Line types supported by every backend that draws rules. This is the
# intersection of CSS `border-style`, tabularray hline/vline specs, and Typst
# `stroke(dash:)`. Do not extend without checking all three.
LINE_TYPES <- c("solid", "dashed", "dotted")

#' Apply styling to notes or caption
#' @keywords internal
#' @noRd
apply_notes_caption_styling <- function(
  x,
  i,
  color = NULL,
  fontsize = NULL,
  bold = NULL,
  italic = NULL,
  monospace = NULL,
  smallcap = NULL,
  strikeout = NULL,
  underline = NULL
) {
  style_params <- list(
    color = color,
    fontsize = fontsize,
    bold = bold,
    italic = italic,
    monospace = monospace,
    smallcap = smallcap,
    strikeout = strikeout,
    underline = underline
  )

  if (identical(i, "notes")) {
    x@style_notes <- style_params
  } else if (identical(i, "caption")) {
    x@style_caption <- style_params
  }

  return(x)
}

#' Process logical matrix input for styling
#' @keywords internal
#' @noRd
process_logical_matrix_input <- function(x, i, j) {
  if (!is.null(j)) {
    stop("When `i` is a logical matrix, `j` must be NULL.", call. = FALSE)
  }

  settings <- which(i, arr.ind = TRUE)
  settings <- stats::setNames(data.frame(settings), c("i", "j"))
  return(settings)
}

#' Process regular input for styling
#' @keywords internal
#' @noRd
process_regular_input <- function(x, i, j) {
  ival <- sanitize_i(i, x, calling_function = "style_tt")
  jval <- sanitize_j(j, x)

  # Style values recycle over cells in USER-specified order, matching the
  # behavior of `i`: style_tt(j = c(3, 1), background = c("red", "blue"))
  # puts red on column 3 and blue on column 1. sanitize_j() returns positions
  # in column order for character `j`, so restore the user's order here.
  if (is.character(j) && length(j) > 1) {
    jval <- match(j, colnames(x))
  }

  # Handle empty index case - return empty settings dataframe with proper structure
  if (length(ival) == 0) {
    return(data.frame(i = integer(0), j = integer(0)))
  }

  # Create settings grid. expand.grid() varies `i` fastest, so rows follow the
  # user-specified `j` order (important for recycling style value vectors).
  settings <- expand.grid(i = ival, j = jval)

  return(settings)
}


#' Process align argument and add to settings
#' @keywords internal
#' @noRd
process_align_argument <- function(x, settings, align) {
  # Empty selection (e.g., i = integer(0)): nothing to style, and assigning
  # any column to a 0-row frame would fail with a replacement-length error.
  if (nrow(settings) == 0) {
    return(settings)
  }

  if (is.null(align)) {
    settings[["align"]] <- NA_character_
    return(settings)
  }

  # Get the number of columns being styled
  n_cols_styled <- length(unique(settings$j))

  if (nchar(align) == n_cols_styled) {
    align <- strsplit(align, "")[[1]]
  } else if (nchar(align) == 1) {
    align <- rep(align, n_cols_styled)
  } else {
    msg <- sprintf("`align` must be a single character or a string of length %s.", n_cols_styled)
    stop(msg, call. = FALSE)
  }

  if (!all(align %in% c("c", "l", "r", "d"))) {
    stop("`align` must be characters c, l, r, or d.", call. = FALSE)
  }

  # Initialize align column if it doesn't exist
  if (!"align" %in% names(settings)) {
    settings$align <- NA_character_
  }

  unique_j <- unique(settings$j)
  for (j in seq_along(align)) {
    idx <- which(settings$j == unique_j[j])
    settings$align[idx] <- align[j]
  }

  return(settings)
}



style_tt_lazy <- function(
  x,
  i = NULL,
  j = NULL,
  bold = NULL,
  italic = NULL,
  monospace = NULL,
  smallcap = NULL,
  underline = NULL,
  strikeout = NULL,
  color = NULL,
  background = NULL,
  fontsize = NULL,
  align = NULL,
  alignv = NULL,
  colspan = NULL,
  rowspan = NULL,
  indent = NULL,
  line = NULL,
  line_color = NULL,
  line_width = 0.1,
  line_type = NULL,
  line_trim = NULL,
  finalize = NULL,
  ...) {

  out <- x
  # Each lazy call exposes only its own settings frame for build_tt() to
  # collect. In particular, notes and caption styling return early below and
  # must not expose the preceding call's frame again.
  out@style <- data.frame()

  # Row indices derived from an NSE predicate (e.g., i = x > 30) were computed
  # eagerly in style_tt() against the pre-group data rows. Here, at lazy-eval
  # time, group_tt() label rows have been inserted and numeric `i` refers to
  # final-table positions, so remap through @index_body.
  i <- nse_remap_i(i, x)

  # Set default line_color if NULL. HTML and Typst both have a native mechanism
  # to inherit the color from the surrounding document (a CSS variable, and
  # Typst stroke folding respectively), so we leave the paint unspecified there
  # rather than hard-coding black and overriding the user's global styling.
  if (is.null(line_color) && !is.null(line)) {
    if (identical(x@output, "html") && identical(x@html_engine, "tinytable")) {
      line_color <- "var(--tt-line-color)"
    } else if (!identical(x@output, "typst")) {
      line_color <- "black"
    }
  }

  # Deprecated argument shims. These must mutate `out`, which is the object
  # returned to build_tt(); changes made to `x` after `out <- x` would be lost.
  if ("tabularray_inner" %in% ...names()) {
    out <- theme_latex(out, inner = ...get("tabularray_inner"))
    warning("The `tabularray_inner` argument is deprecated. Use `theme_latex(x, inner = ...)` instead.",
      call. = FALSE
    )
  }
  if ("tabularray_outer" %in% ...names()) {
    out <- theme_latex(out, outer = ...get("tabularray_outer"))
    warning("The `tabularray_outer` argument is deprecated. Use `theme_latex(x, outer = ...)` instead.",
      call. = FALSE
    )
  }
  if ("html_class" %in% ...names()) {
    out <- theme_html(out, class = ...get("html_class"))
    warning(
      "The `html_class` argument is deprecated. Use `theme_html(x, class = ...)` instead.",
      call. = FALSE
    )
  }
  if ("html_css_rule" %in% ...names()) {
    out <- theme_html(out, css_rule = ...get("html_css_rule"))
    warning("The `html_css_rule` argument is deprecated. Use `theme_html(x, css_rule = ...)` instead.",
      call. = FALSE
    )
  }

  # this must be handled here rather than theme_html() because it is a cell-level issue
  html_css <- ...get("html_css")

  # Handle special cases first (before validation)
  if (isTRUE(i %in% c("notes", "caption"))) {
    return(apply_notes_caption_styling(
      out,
      i,
      color,
      fontsize,
      bold,
      italic,
      monospace,
      smallcap,
      strikeout,
      underline
    ))
  }

  # Validate inputs (after special cases)
  assert_style_tt(
    x = out,
    i = i,
    j = j,
    bold = bold,
    italic = italic,
    monospace = monospace,
    smallcap = smallcap,
    underline = underline,
    strikeout = strikeout,
    color = color,
    background = background,
    fontsize = fontsize,
    align = align,
    alignv = alignv,
    colspan = colspan,
    rowspan = rowspan,
    indent = indent,
    line = line,
    line_color = line_color,
    line_width = line_width,
    line_type = line_type,
    line_trim = line_trim,
    html_css = html_css,
    finalize = finalize,
    ...
  )

  # Add Tabulator-specific validation via lazy_prepare
  # This checks alignment restrictions when output is determined
  if (!is.null(align) || !is.null(alignv)) {
    validate_fn <- function(x) {
      if (x@html_engine == "tabulator") {
        if (!is.null(i)) {
          stop(
            "Tabulator does not support row-specific alignment. ",
            "When using `align` or `alignv` with Tabulator output, `i` must be NULL to apply alignment to entire columns.",
            call. = FALSE
          )
        }
      }
      return(x)
    }
    out <- build_prepare(out, validate_fn, output = "html")
  }

  sanity_align(align, i)

  # Process inputs and create settings.
  # Logical-matrix `i` dimensions were validated in assert_style_tt().
  if (is.matrix(i) && is.logical(i)) {
    settings <- process_logical_matrix_input(x, i, j)
  } else {
    settings <- process_regular_input(x, i, j)
  }

  # Build complete settings - skip if no rows to style
  if (nrow(settings) > 0) {
    # Cell-style columns: NA when unset, otherwise the user value recycled over
    # the settings rows. STYLE_PROPS is the canonical property vector defined
    # in style_maps.R; align, alignv, and html_css need special handling below.
    for (prop in setdiff(STYLE_PROPS, c("align", "alignv", "html_css"))) {
      value <- get(prop, inherits = FALSE)
      settings[[prop]] <- if (is.null(value)) NA else as.vector(value)
    }
    settings[["alignv"]] <- if (is.null(alignv)) NA else alignv
    settings[["line_color"]] <- if (is.null(line) || is.null(line_color)) {
      NA_character_
    } else {
      line_color
    }
    settings[["line_width"]] <- if (is.null(line)) NA else line_width
    settings[["line_type"]] <- if (is.null(line) || is.null(line_type)) {
      NA_character_
    } else {
      line_type
    }
    settings[["html_css"]] <- if (!is.null(html_css)) {
      html_css
    } else {
      NA
    }
    # Always create tabularray column for consistency
    settings[["tabularray"]] <- ""

    # Expand compound line directions like "tblr" into separate entries.
    # Only the first copy keeps the non-line properties (background, color,
    # etc.); the extra copies carry line-relevant columns only, so other
    # properties are not resolved multiple times per cell.
    if (!is.null(line) && nchar(line) > 1) {
      line_chars <- strsplit(line, "")[[1]]
      line_only <- settings
      for (prop in setdiff(colnames(line_only), c("i", "j", "line_color", "line_width", "line_type", "tabularray"))) {
        line_only[[prop]] <- NA
      }
      expanded_settings <- vector("list", length(line_chars))
      for (k in seq_along(line_chars)) {
        new_settings <- if (k == 1) settings else line_only
        new_settings[["line"]] <- line_chars[k]
        expanded_settings[[k]] <- new_settings
      }
      settings <- do.call(rbind, expanded_settings)
    } else {
      settings[["line"]] <- if (is.null(line)) NA else line
    }

    if (!is.null(line_trim)) {
      split_consecutive <- function(x) {
        x <- sort(unique(x))  # optional: ensure sorted & unique
        groups <- cumsum(c(1, diff(x) != 1))
        split(x, groups)
      }
      idx <- split_consecutive(settings$j)
      left <- grepl("l", line_trim)
      right <- grepl("r", line_trim)
      for (d in idx) {
        if (length(d) == 1) {
          settings[["line_trim"]][settings$j == d] <- line_trim
        } else {
          if (left) {
            settings[["line_trim"]][settings$j == min(d)] <- "l"
          }
          if (right) {
            settings[["line_trim"]][settings$j == max(d)] <- "r"
          }
        }
      }
      ends <- unlist(lapply(idx, range))
      settings[["line_trim"]][!settings$j %in% ends] <- NA
    } else {
      settings[["line_trim"]] <- NA
    }

  }

  if (!is.matrix(i) || !is.logical(i)) {
    settings <- process_align_argument(x, settings, align)
  } else {
    if (!is.null(align)) {
      warning("`align` is not supported when `i` is a logical matrix; it is ignored.", call. = FALSE)
    }
    if (nrow(settings) > 0) {
      settings$align <- NA_character_
    }
  }

  # sort column: important for bind
  cols <- unique(c("i", "j", sort(colnames(settings))))
  settings <- settings[, cols, drop = FALSE]

  # Expose this call's settings frame in @style. build_tt() collects the
  # per-call frames and rbind()s them once at the end of the lazy loop, instead
  # of growing @style with an incremental rbind() on every style_tt() call.
  # That per-call rbind() was O(N^2) and dominated the lazy-style evaluation
  # phase on tables with many style_tt() calls (e.g. per-cell heat-maps).
  out@style <- settings

  if (is.function(finalize)) {
    out@lazy_finalize <- c(out@lazy_finalize, list(finalize))
  }

  return(out)
}

assert_style_tt <- function(
  x,
  i,
  j,
  bold,
  italic,
  monospace,
  smallcap,
  underline,
  strikeout,
  color,
  background,
  fontsize,
  align,
  alignv,
  colspan,
  rowspan,
  indent,
  line,
  line_color,
  line_width,
  line_type,
  line_trim,
  finalize = NULL,
  ...
) {
  # Validate alignv choice
  assert_choice(alignv, c("t", "m", "b"), null.ok = TRUE)


  # Validate finalize function
  assert_function(finalize, null.ok = TRUE)

  # Check for deprecated width argument
  if ("width" %in% names(list(...))) {
    stop("The `width` argument is now in the `tt()` function.", call. = FALSE)
  }

  assert_integerish(colspan, len = 1, lower = 2, null.ok = TRUE)
  assert_integerish(rowspan, len = 1, lower = 2, null.ok = TRUE)
  assert_numeric(indent, len = 1, lower = 0, null.ok = TRUE)
  assert_character(background, null.ok = TRUE)
  assert_character(color, null.ok = TRUE)
  assert_numeric(fontsize, null.ok = TRUE)
  assert_logical(bold, null.ok = TRUE)
  assert_logical(italic, null.ok = TRUE)
  assert_logical(monospace, null.ok = TRUE)
  assert_logical(smallcap, null.ok = TRUE)
  assert_logical(underline, null.ok = TRUE)
  assert_logical(strikeout, null.ok = TRUE)
  assert_string(line, null.ok = TRUE)
  assert_string(line_color, null.ok = TRUE) # default determined by output format
  assert_numeric(line_width, len = 1, lower = 0, null.ok = FALSE) # 0.1 default
  assert_choice(line_type, LINE_TYPES, null.ok = TRUE)
  assert_choice(line_trim, c("l", "r", "lr"), null.ok = TRUE)

  # Validate that line_trim is only used with bottom lines
  if (!is.null(line_trim) && !is.null(line)) {
    if (!identical("b", line)) {
      stop("line_trim can only be used with bottom lines (line must contain 'b').", call. = FALSE)
    }
  }

  # must be handled here rather than theme_html() because it is a cell-level issue
  html_css <- ...get("html_css")
  assert_character(html_css, null.ok = TRUE)

  if (is.character(line)) {
    line <- strsplit(line, split = "")[[1]]
    if (!all(line %in% c("t", "b", "l", "r"))) {
      msg <- "`line` must be a string of characters t, b, l, or r."
      stop(msg, call. = FALSE)
    }
  }

  ival <- sanitize_i(i, x, calling_function = "style_tt")
  jval <- sanitize_j(j, x)
  inull <- isTRUE(attr(ival, "null"))
  jnull <- isTRUE(attr(jval, "null"))

  # Valid recycling lengths for style value vectors
  if (is.matrix(ival) && is.logical(ival)) {
    # Logical-matrix selection: dimensions must match the table exactly,
    # otherwise TRUE cells would be silently coerced to garbage indices.
    if (nrow(ival) != nrow(x) || ncol(ival) != ncol(x)) {
      msg <- sprintf(
        "When `i` is a logical matrix, its dimensions must match the table: %s rows and %s columns (received %s by %s).",
        nrow(x), ncol(x), nrow(ival), ncol(ival)
      )
      stop(msg, call. = FALSE)
    }
    # One settings row per TRUE cell, filled in column-major order
    len <- c(1, sum(ival))
  } else if (inull && jnull) {
    # 1
    len <- 1
  } else if (!inull && jnull) {
    # 1 or #rows
    len <- c(1, length(ival))
  } else if (inull && !jnull) {
    # 1 or #cols
    len <- c(1, length(jval))
  } else {
    # 1 or #cells
    len <- c(1, length(ival) * length(jval))
  }

  recycled_props <- list(
    color = color,
    background = background,
    fontsize = fontsize,
    bold = bold,
    italic = italic,
    monospace = monospace,
    smallcap = smallcap,
    underline = underline,
    strikeout = strikeout
  )
  for (nm in names(recycled_props)) {
    assert_length(recycled_props[[nm]], len = len, null.ok = TRUE, name = nm)
  }
}
# =============================================================================
# MAIN FUNCTION
# =============================================================================

#' Style a Tiny Table
#'
#' @details
#' This function applies styling to a table created by `tt()`. It allows customization of text style (bold, italic, monospace), text and background colors, font size, cell width, text alignment, column span, and indentation. The function also supports passing native instructions to LaTeX (tabularray) and HTML (bootstrap) formats.
#'
#' Vector values for style arguments (e.g., `color`, `background`, `fontsize`) are recycled over the selected cells in the order specified by the user in `i` and `j`, with `i` varying fastest. For example, `style_tt(x, j = c(3, 1), background = c("red", "blue"))` colors column 3 red and column 1 blue. When `i` is a logical matrix, values are recycled over the `TRUE` cells in column-major order.
#'
#' @param x A table object created by `tt()`.
#' @param i Numeric vector, logical matrix, string, or unquoted expression.
#'   - Numeric vector: Row indices where the styling should be applied. Can be a single value or a vector.
#'   - Logical matrix: A matrix with the same number of rows and columns as `x`. `i=0` is the header, and negative values are higher level headers. Row indices refer to rows *after* the insertion of row labels by `group_tt()`, when applicable.
#'   - String: Table components "caption", "colnames", "groupi" (row group labels), "~groupi" (non-group rows), "groupj" (column group labels), "notes".
#'   - Unquoted expression: When supplying an unquoted expression, it is first evaluated in the calling environment, then in the data frame passed to `tt()`.
#' @param j Column indices where the styling should be applied. Can be:
#' + Integer vectors indicating column positions.
#' + Character vector indicating column names.
#' + A single string specifying a Perl-style regular expression used to match column names.
#' + Unquoted expression: Non-standard evaluation is supported. When supplying an unquoted expression, it is first evaluated in the calling environment, then in an environment that includes the columns of the original data passed to `tt()`, and `groupi` indices. See examples below.
#' @param bold Logical; if `TRUE`, text is styled in bold.
#' @param italic Logical; if `TRUE`, text is styled in italic.
#' @param monospace Logical; if `TRUE`, text is styled in monospace font.
#' @param smallcap Logical; if `TRUE`, text is styled in small caps. In Markdown output format, text is converted to uppercase.
#' @param underline Logical; if `TRUE`, text is underlined.
#' @param strikeout Logical; if `TRUE`, text has a strike through line.
#' @param color Text color. Colors are standardized across output formats and can be specified as:
#'   - Hex codes: "#CC79A7", "#FF0000", "#123ABC"
#'   - R color names: Any color recognized by R, such as "red", "blue", "forestgreen", "lightblue"
#'   - Extended color names: 749+ named colors from the LaTeX xcolor package (see `tinytable:::latex_colors` for the full list)
#'   - LaTeX color blending (LaTeX output only): "white!80!blue", "red!50", "green!20!red"
#' @param background Background color. Same color specification options as the `color` parameter. Can be `NULL` for default color.
#' @param fontsize Font size in em units. Can be `NULL` for default size.
#' @param align A single character or a string with a number of characters equal to the number of columns in `j`. Valid characters include 'c' (center), 'l' (left), 'r' (right), 'd' (decimal). Decimal alignment is only available in LaTeX via the `siunitx` package. The width of columns is determined by the maximum number of digits to the left and to the right in all cells specified by `i` and `j`.
#' @param alignv A single character specifying vertical alignment. Valid characters include 't' (top), 'm' (middle), 'b' (bottom).
#' @param colspan Number of columns a cell should span. `i` and `j` must be of length 1.
#' @param rowspan Number of rows a cell should span. `i` and `j` must be of length 1.
#' @param indent Text indentation in em units. Positive values only.
#' @param line String determines if lines (rules or borders) should be drawn around the cell, row, or column. See `line_type` to draw dashed or dotted lines.
#' + "t": top
#' + "b": bottom
#' + "l": left
#' + "r": right
#' + Can be combined such as: "lbt" to draw borders at the left, bottom, and top.
#' @param line_color Color of the line. See the `color` argument for details.
#' @param line_width Width of the line in em units (default: 0.1).
#' @param line_type Style of the line: "solid" (default), "dashed", or "dotted". Ignored in Markdown and Word output, and by the Tabulator HTML engine, which always draw solid rules.
#' @param line_trim String specifying line trimming. Acceptable values: "l" (left), "r" (right), "lr" (both sides). When specified, shortens the lines by 0.8pt on the specified side(s). Default: NULL (no trimming).
#' @param finalize A function applied to the table object at the very end of table-building, for post-processing. For example, the function could use regular expressions to add LaTeX commands to the text version of the table hosted in `x@table_string`, or it could programmatically change the caption in `x@caption`.
#' @param output Apply styling only when the table is rendered in the specified format. A character vector of one or more of "latex", "html", "typst", or "markdown". If `NULL` (default), styling is applied regardless of the output format.
#' @param ... extra arguments are ignored
#' @return An object of class `tt` representing the table.
#' @template limitations_word_markdown
#' @export
#' @examplesIf knitr::is_html_output()
#' if (knitr::is_html_output()) options(tinytable_print_output = "html")
#'
#' library(tinytable)
#'
#' tt(mtcars[1:5, 1:6])
#'
#' # Alignment
#' tt(mtcars[1:5, 1:6]) |>
#'   style_tt(j = 1:5, align = "lcccr")
#'
#' # Colors and styles
#' tt(mtcars[1:5, 1:6]) |>
#'   style_tt(i = 2:3, background = "black", color = "orange", bold = TRUE)
#'
#' # column selection with `j``
#' tt(mtcars[1:5, 1:6]) |>
#'   style_tt(j = 5:6, background = "pink")
#'
#' tt(mtcars[1:5, 1:6]) |>
#'   style_tt(j = "drat|wt", background = "pink")
#'
#' tt(mtcars[1:5, 1:6]) |>
#'   style_tt(j = c("drat", "wt"), background = "pink")
#'
#' tt(mtcars[1:5, 1:6], theme = "empty") |>
#'   style_tt(
#'     i = 2, j = 2,
#'     colspan = 3,
#'     rowspan = 2,
#'     align = "c",
#'     alignv = "m",
#'     color = "white",
#'     background = "black",
#'     bold = TRUE)
#'
#' tt(mtcars[1:5, 1:6], theme = "empty") |>
#'   style_tt(
#'     i = 0:3,
#'     j = 1:3,
#'     line = "tblr",
#'     line_width = 0.4,
#'     line_color = "teal")
#'
#' tt(mtcars[1:5, 1:6], theme = "striped") |>
#'   style_tt(
#'     i = c(2, 5),
#'     j = 3,
#'     strikeout = TRUE,
#'     fontsize = 0.7)
#'
#' # Non-standard evaluation (NSE)
#' dat <- data.frame(
#'   w = c(143002.2092, 201399.181, 100188.3883),
#'   x = c(1.43402, 201.399, 0.134588),
#'   y = as.Date(c(897, 232, 198), origin = "1970-01-01"),
#'   z = c(TRUE, TRUE, FALSE)
#' )
#' tt(dat) |>
#'   style_tt(i = w > 150000, j = c("w", "x"), 
#'            color = "white", background = "black")
#'
#' tt(mtcars[1:5, 1:6]) |>
#'   theme_html(class = "table table-dark table-hover")
#'
#'
#' inner <- "
#' column{1-4}={halign=c},
#' hlines = {fg=white},
#' vlines = {fg=white},
#' cell{1,6}{odd} = {bg=teal7},
#' cell{1,6}{even} = {bg=green7},
#' cell{2,4}{1,4} = {bg=red7},
#' cell{3,5}{1,4} = {bg=purple7},
#' cell{2}{2} = {r=4,c=2}{bg=azure7},
#' "
#' tt(mtcars[1:5, 1:4], theme = "empty") |>
#'   theme_latex(inner = inner)
#'
#' # Style group rows and non-group rows
#' dat <- data.frame(x = 1:6, y = letters[1:6])
#' dat |>
#'   tt() |>
#'   group_tt(i = list("Group A" = 3)) |>
#'   style_tt(i = "groupi", background = "lightblue") |>
#'   style_tt(i = "~groupi", background = "lightgray")
#'
#' # unquote expressions
#' dat <- mtcars[1:10,]
#' dat <- dat[order(dat$am),]
#' tt(dat) |>
#'    subset(mpg > 20) |>
#'    group_tt(am)
#'
#' # style elements: captions and colnames
#' notes <- list(
#'   "*" = "Hello world", 
#'   "a" = "Bacon ipsum dolor amet kevin t-bone porchetta.")
#' tt(head(iris), 
#'    width = .8,
#'    caption = "This is a Caption Example.", 
#'    notes = notes) |>
#'   style_tt(2, 2, background = "pink", rowspan = 2, colspan = 2, 
#'     alignv = "m", align = "c", line = "tblr") |>
#'   style_tt("colnames", italic = TRUE) |>
#'   style_tt("caption", smallcap = TRUE)
#'
style_tt <- function(
  x,
  i = NULL,
  j = NULL,
  bold = NULL,
  italic = NULL,
  monospace = NULL,
  smallcap = NULL,
  underline = NULL,
  strikeout = NULL,
  color = NULL,
  background = NULL,
  fontsize = NULL,
  align = NULL,
  alignv = NULL,
  colspan = NULL,
  rowspan = NULL,
  indent = NULL,
  line = NULL,
  line_color = NULL,
  line_width = 0.1,
  line_type = NULL,
  line_trim = NULL,
  finalize = NULL,
  output = NULL,
  ...
) {

  
  assert_subset(output, c("latex", "html", "typst", "markdown"), null.ok = TRUE)

  # non-standard evaluation before anything else
  tmp <- nse_i_j(x, i_expr = substitute(i), j_expr = substitute(j), pf = parent.frame())
  list2env(tmp, environment())

  # evaluate arguments immediately, except i and j, to avoid scoping issues
  obj <- c(
    list(style_tt_lazy, x = quote(x), i = i, j = j),
    mget(STYLE_TT_ARGS, envir = environment()),
    list(...)
  )
  obj <- as.call(obj)
  attr(obj, "output") <- output

  x@lazy_style <- c(x@lazy_style, list(obj))
  return(x)
}

usepackage_latex <- function(name, options = NULL, extra_lines = NULL) {
  assert_dependency("rmarkdown")
  invisible(
    knitr::knit_meta_add(
      list(rmarkdown::latex_dependency(name, options, extra_lines))
    )
  )
}

sanity_align <- function(align, i) {
  if (any(grepl("d", align)) && !is.null(i)) {
    msg <- "d column alignment can only be applied to entire columns. `i` must be `NULL`."
    stop(msg, call. = FALSE)
  }
}

sanitize_i <- function(
  i,
  x,
  pre_group_i = FALSE,
  lazy = TRUE,
  calling_function = "other"
) {
  if (is.character(i)) {
    # format_tt() accepts certain character inputs
    # if (calling_function %in% c("format_tt", "style_tt")) {
    assert_choice(
      i,
      choice = c("caption", "colnames", "groupi", "~groupi", "groupj", "notes"),
      null.ok = TRUE
    )
  }
  if (identical(i, "groupi")) {
    i <- x@group_index_i
    if (length(i) == 0) {
      msg <- "No grouping index found. Please use `group_tt()` first."
      stop(msg, call. = FALSE)
    }
  } else if (identical(i, "~groupi")) {
    i <- setdiff(seq_len(nrow(x)), x@group_index_i)
  } else if (identical(i, "groupj")) {
    has_colnames <- length(names(x)) > 0
    has_groupj <- nrow(x@group_data_j) > 0
    if (!has_groupj) {
      msg <- "No column grouping found. Please use `group_tt(j = ...)` first."
      stop(msg, call. = FALSE)
    }
    if (has_colnames) {
      i <- -(1:x@nhead)
    } else {
      i <- -(0:x@nhead)
    }
  } else if (identical(i, "colnames")) {
    if (isTRUE(length(colnames(x)) == 0)) {
      msg <- "No column names found. Please set column names first."
      stop(msg, call. = FALSE)
    }
    i <- 0
  }

  if (is.matrix(i) && is.logical(i)) {
    return(i)
  }

  assert_numeric(i, null.ok = TRUE, name = "i")
  if (anyNA(i)) {
    stop("`i` must not contain missing values.", call. = FALSE)
  }

  if (is.null(i) && isTRUE(lazy)) {
    out <- NA
    attr(out, "null") <- TRUE
    attr(out, "body") <- seq_len(nrow(x))
    attr(out, "head") <- integer()
  } else {
    if (!is.null(i)) {
      out <- i
    } else {
      out <- seq_len(nrow(x))
    }
    attr(out, "null") <- FALSE
    attr(out, "body") <- out[out > 0]
    attr(out, "head") <- out[out < 1]
  }

  return(out)
}

sanitize_j <- function(j, x, skip_tabulator_types = FALSE) {
  # regex
  if (is.character(j) && length(j) == 1 && !is.null(colnames(x))) {
    out <- grep(j, colnames(x), perl = TRUE)
    # Check if regex pattern matched any columns
    if (length(out) == 0) {
      msg <- sprintf("No columns match the pattern: %s", j)
      stop(msg, call. = FALSE)
    }
    # full names
  } else if (is.character(j) && length(j) > 1 && !is.null(colnames(x))) {
    bad <- setdiff(j, colnames(x))
    if (length(bad) > 0) {
      msg <- sprintf("Missing columns: %s", paste(bad, collapse = ", "))
      stop(msg, call. = FALSE)
    }
    out <- which(colnames(x) %in% j)
  } else {
    assert_integerish(j, lower = 1, upper = ncol(x), null.ok = TRUE)
    if (anyNA(j)) {
      stop("`j` must not contain missing values.", call. = FALSE)
    }
    if (is.null(j)) {
      out <- seq_len(ncol(x))
    } else {
      out <- j
    }
  }

  # Filter out columns that should be skipped for tabulator formatting:
  # Tabulator formats numeric/date/logical columns client-side via JS
  # formatters, so the underlying data must be sent raw.
  is_tabulator <- inherits(x, "tinytable") &&
    identical(x@output, "html") &&
    identical(x@html_engine, "tabulator")
  if (skip_tabulator_types && is_tabulator) {
    skip_classes <- c("integer", "numeric", "logical", "Date", "POSIXct", "POSIXlt")
    # when `bool` formatting is requested, logical columns must still be
    # formatted in the data because tabulator_clean_data() reads them from
    # @data_body
    if (isTRUE(x@tabulator_format_bool)) {
      skip_classes <- setdiff(skip_classes, "logical")
    }
    skip <- vapply(
      out,
      function(col_idx) inherits(x@data[[col_idx]], skip_classes),
      logical(1)
    )
    out <- out[!skip]
  }

  attr(out, "null") <- is.null(j)
  return(out)
}

sanitize_output <- function(x, output) {
  if (is.null(x)) {
    stop("'x' cannot be NULL", call. = FALSE)
  }
  if (is.null(output)) {
    return(x)
  }
  
  assert_choice(
    output,
    choice = c(
      "latex",
      "html",
      "typst",
      "dataframe",
      "markdown",
      "knitr",
      "raster"
    )
  )
  
  x@output <- output
  return(x)
}


insert_latex_preamble <- function(x) {
  if (isTRUE(check_dependency("knitr"))) {
    pandoc_to <- knitr::pandoc_to()
    if (isTRUE(pandoc_to %in% c("latex", "beamer"))) {
      if (isTRUE(x@latex_preamble)) {
        usepackage_latex("float")
        usepackage_latex(
          "tabularray",
          extra_lines = c(
            "\\usepackage[normalem]{ulem}",
            "\\usepackage{graphicx}",
            "\\usepackage{rotating}",
            "\\UseTblrLibrary{siunitx}",
            "\\NewTableCommand{\\tinytableDefineColor}[3]{\\definecolor{#1}{#2}{#3}}",
            "\\newcommand{\\tinytableTabularrayUnderline}[1]{\\underline{#1}}",
            "\\newcommand{\\tinytableTabularrayStrikeout}[1]{\\sout{#1}}"
          )
        )
      }
    }
  }
}


infer_output <- function(x) {
  output <- x@output

  insert_latex_preamble(x)
  
  # If output is already set to a specific format, respect it
  if (!is.null(output) && !identical(output, "tinytable")) {
    return(output)
  } else {
    # Only do inference when output is NULL or "tinytable"
    has_viewer <- interactive() && !is.null(getOption("viewer"))
    out <- if (has_viewer) "html" else "markdown"
  }

  # Environmental overrides only apply when we're doing inference (not explicit output)
  if (!is.null(getOption("calepin", default = NULL))) {
    return("typst")
  }

  if (isTRUE(check_dependency("litedown"))) {
    fmt <- tryCatch(litedown::get_context("format"), error = function(e) NULL)
    if (identical(fmt, "latex")) {
      return("latex")
    } else if (identical(fmt, "markdown")) {
      return("markdown")
    } else if (identical(fmt, "commonmark")) {
      return("markdown")
    } else if (identical(fmt, "html")) {
      return("html")
    }
  }

  if (isTRUE(check_dependency("knitr"))) {

    pandoc_to <- knitr::pandoc_to()
    if (isTRUE(pandoc_to %in% c("latex", "beamer"))) {
      out <- "latex"
    } else if (isTRUE(pandoc_to %in% c("html", "revealjs"))) {
      out <- "html"
    } else if (isTRUE(pandoc_to == "typst")) {
      out <- "typst"
      if (isTRUE(check_dependency("quarto"))) {
        if (isTRUE(quarto::quarto_version() < "1.5.29")) {
          msg <- "Typst tables require version 1.5.29 or later of Quarto and version 0.11.0 or later of Typst. This software may (or may not) only be available in pre-release builds: https://quarto.org/docs/download"
          stop(msg, call. = FALSE)
        }
      }
    # docx, markdown variants, and unknown formats
    } else if (!is.null(pandoc_to)) {
      out <- "markdown"
    }
  }
  
  return(out)
}

#' check if dependency is installed
#'
#' @noRd
check_dependency <- function(library_name) {
  flag <- requireNamespace(library_name, quietly = TRUE)
  if (isFALSE(flag)) {
    msg <- sprintf("Please install the `%s` package.", library_name)
    return(msg)
  } else {
    return(TRUE)
  }
}

assert_dependency <- function(library_name) {
  flag <- check_dependency(library_name)
  if (!isTRUE(flag)) {
    stop(flag, call. = FALSE)
  }
  return(invisible())
}

# shared core for assert_choice() and assert_subset()
assert_membership <- function(x, choice, all, null.ok, name, template) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(TRUE)
  }
  ok <- if (isTRUE(all)) {
    is.character(x) && length(x) >= 1 && all(x %in% choice)
  } else {
    is.character(x) && length(x) == 1 && x %in% choice
  }
  if (ok) {
    return(TRUE)
  }
  msg <- sprintf(template, name, paste(choice, collapse = ", "))
  stop(msg, call. = FALSE)
}

assert_choice <- function(
  x,
  choice,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  assert_membership(
    x,
    choice,
    all = FALSE,
    null.ok = null.ok,
    name = name,
    template = "`%s` must be one of: %s"
  )
}


assert_subset <- function(
  x,
  choice,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  assert_membership(
    x,
    choice,
    all = TRUE,
    null.ok = null.ok,
    name = name,
    template = "`%s` must be a subset of: %s"
  )
}

check_string <- function(x, null.ok = FALSE) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(invisible(TRUE))
  }
  if (is.character(x) && length(x) == 1 && !is.na(x)) {
    return(invisible(TRUE))
  }
  return(FALSE)
}

assert_string <- function(
  x,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")) {
  if (!isTRUE(check_string(x, null.ok = null.ok))) {
    msg <- sprintf("`%s` must be a string.", name)
    stop(msg, call. = FALSE)
  }
}

check_flag <- function(x, null.ok = FALSE) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(TRUE)
  }
  if (is.logical(x) && length(x) == 1 && !is.na(x)) {
    return(TRUE)
  }
  return(FALSE)
}

assert_flag <- function(
  x,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  msg <- sprintf("`%s` must be a logical flag.", name)
  if (!isTRUE(check_flag(x, null.ok = null.ok))) {
    stop(msg, call. = FALSE)
  }
}

check_function <- function(x, null.ok = FALSE) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(TRUE)
  }
  if (is.function(x)) {
    return(TRUE)
  }
  return(FALSE)
}

assert_length <- function(
  x,
  len = 1,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(invisible(TRUE))
  }
  msg <- sprintf(
    "`%s` must be one of these lengths: %s",
    name,
    paste(unique(len), collapse = ", ")
  )
  if (!length(x) %in% len) {
    stop(msg, call. = FALSE)
  }
}

assert_logical <- function(
  x,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(invisible(TRUE))
  }
  msg <- sprintf("`%s` must be a logical vector", name)
  if (!is.logical(x)) stop(msg, call. = FALSE)
}

# shared core for check_numeric() and check_integerish()
# NA policy: NA values are tolerated; bounds and integer checks ignore them
# (na.rm = TRUE) instead of crashing with "missing value where TRUE/FALSE
# needed". Indices are rejected explicitly in sanitize_i()/sanitize_j().
check_numeric_core <- function(x, len, lower, upper, null.ok, integerish) {
  if (is.null(x)) {
    return(isTRUE(null.ok))
  }
  if (!is.numeric(x)) {
    return(FALSE)
  }
  if (integerish) {
    x <- stats::na.omit(x)
  }
  if (!is.null(len) && length(x) != len) {
    return(FALSE)
  }
  if (!is.null(lower) && isTRUE(any(x < lower, na.rm = TRUE))) {
    return(FALSE)
  }
  if (!is.null(upper) && isTRUE(any(x > upper, na.rm = TRUE))) {
    return(FALSE)
  }
  if (
    integerish &&
      isTRUE(any(abs(x - round(x)) > (.Machine$double.eps)^0.5, na.rm = TRUE))
  ) {
    return(FALSE)
  }
  return(TRUE)
}

check_integerish <- function(
  x,
  len = NULL,
  lower = NULL,
  upper = NULL,
  null.ok = TRUE
) {
  check_numeric_core(
    x,
    len = len,
    lower = lower,
    upper = upper,
    null.ok = null.ok,
    integerish = TRUE
  )
}

check_numeric <- function(
  x,
  len = NULL,
  lower = NULL,
  upper = NULL,
  null.ok = TRUE
) {
  check_numeric_core(
    x,
    len = len,
    lower = lower,
    upper = upper,
    null.ok = null.ok,
    integerish = FALSE
  )
}

# shared core for assert_numeric() and assert_integerish()
# builds the failure message in a single pass (each condition evaluated once)
assert_numeric_core <- function(x, len, lower, upper, null.ok, integerish, name) {
  what <- if (integerish) "integer-ish" else "numeric"
  if (is.null(x)) {
    if (isTRUE(null.ok)) {
      return(invisible())
    }
    if (integerish) {
      stop(sprintf("%s should not be NULL.", name), call. = FALSE)
    }
    stop(sprintf("`%s` must be %s", name, what), call. = FALSE)
  }
  msg <- sprintf("`%s` must be %s", name, what)
  if (!is.numeric(x)) {
    if (integerish) {
      msg <- paste0(msg, "; it is not numeric")
    }
    stop(msg, call. = FALSE)
  }
  xcheck <- if (integerish) stats::na.omit(x) else x
  problems <- character()
  if (!is.null(len) && length(xcheck) != len) {
    problems <- c(problems, sprintf("; its length must be %s", len))
  }
  if (!is.null(lower) && isTRUE(any(x < lower, na.rm = TRUE))) {
    problems <- c(
      problems,
      sprintf("; all values must be greater than or equal to %s", lower)
    )
  }
  if (!is.null(upper) && isTRUE(any(x > upper, na.rm = TRUE))) {
    problems <- c(
      problems,
      sprintf("; all values must be less than or equal to %s", upper)
    )
  }
  if (
    integerish &&
      isTRUE(any(abs(x - round(x)) > (.Machine$double.eps)^0.5, na.rm = TRUE))
  ) {
    problems <- c(problems, "; all values must be close to integers")
  }
  if (length(problems) > 0) {
    stop(paste0(msg, paste(problems, collapse = "")), call. = FALSE)
  }
  return(invisible())
}

assert_integerish <- function(
  x,
  len = NULL,
  lower = NULL,
  upper = NULL,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  assert_numeric_core(
    x,
    len = len,
    lower = lower,
    upper = upper,
    null.ok = null.ok,
    integerish = TRUE,
    name = name
  )
}

assert_numeric <- function(
  x,
  len = NULL,
  lower = NULL,
  upper = NULL,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  assert_numeric_core(
    x,
    len = len,
    lower = lower,
    upper = upper,
    null.ok = null.ok,
    integerish = FALSE,
    name = name
  )
}

assert_data_frame <- function(
  x,
  min_rows = 0,
  min_cols = 0,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  msg <- sprintf("`%s` must be a data.frame.", name)
  if (!is.data.frame(x)) {
    stop(msg, call. = FALSE)
  }
  msg <- sprintf("Number of rows in `%s` must be at least `%s`", name, min_rows)
  if (nrow(x) < min_rows) {
    stop(msg, call. = FALSE)
  }
  msg <- sprintf(
    "Number of columns in `%s` must be at least `%s`",
    name,
    min_cols
  )
  if (ncol(x) < min_cols) stop(msg, call. = FALSE)
}

check_matrix <- function(
  x,
  min_rows = 0,
  min_cols = 0,
  null.ok = FALSE
) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(TRUE)
  }
  if (!is.matrix(x)) {
    return(FALSE)
  }
  if (nrow(x) < min_rows) {
    return(FALSE)
  }
  if (ncol(x) < min_cols) {
    return(FALSE)
  }
  return(TRUE)
}

check_character <- function(
  x,
  len = NULL,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  if (isTRUE(null.ok) && is.null(x)) {
    return(TRUE)
  } else if (!is.character(x)) {
    msg <- sprintf("`%s` must be character.", name)
    return(msg)
  } else if (!is.null(len) && length(x) != len) {
    msg <- sprintf("`%s` must have length %s.", name, len)
    return(msg)
  }
  return(TRUE)
}

assert_character <- function(
  x,
  len = NULL,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  flag <- check_character(x, len = len, null.ok = null.ok, name = name)
  if (!isTRUE(flag)) {
    stop(flag, call. = FALSE)
  } else {
    return(invisible(TRUE))
  }
}

assert_list <- function(
  x,
  named = FALSE,
  len = NULL,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  if (isTRUE(null.ok) && is.null(x)) {
    return(invisible(TRUE))
  }
  if (!is.list(x)) {
    stop("Input is not a list.", call. = FALSE)
  }
  if (isTRUE(named)) {
    if (is.null(names(x))) {
      stop(sprintf("`%s` should be named list.", name), call. = FALSE)
    }
  }
  if (!is.null(len)) {
    if (length(x) != len) {
      stop(sprintf("`%s` must be of length %s.", name, len), call. = FALSE)
    }
  }
}

assert_function <- function(
  x,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  if (isTRUE(null.ok) && is.null(x)) {
    return(invisible(TRUE))
  }
  if (!is.function(x)) {
    msg <- sprintf("`%s` must be a function.", name)
    stop(msg, call. = FALSE)
  }
}

check_atomic_vector <- function(
  x,
  null.ok = FALSE,
  name = paste(deparse(substitute(x)), collapse = " ")
) {
  if (isTRUE(null.ok) && is.null(x)) {
    return(invisible(TRUE))
  }
  # doesn't work on glue::glue() output
  # flag <- is.atomic(x) && is.vector(x) && !is.list(x)
  flag <- is.atomic(x) && is.null(dim(x)) && length(x) > 0 && !is.list(x)
  if (flag) {
    out <- TRUE
  } else if (is.factor(x) && is.null(dim(x))) {
    out <- TRUE
  } else {
    out <- sprintf("`%s` must be an atomic vector.", name)
  }
  return(out)
}

assert_atomic_vector <- function(x, 
                                 null.ok = FALSE, 
                                 name = paste(deparse(substitute(x)), collapse = " ")) {
  flag <- check_atomic_vector(x, null.ok = null.ok, name = name)
  if (!isTRUE(flag)) {
    stop(flag, call. = FALSE)
  } else {
    return(invisible(TRUE))
  }
}

assert_class <- function(x, classname) {
  if (!inherits(x, classname)) {
    msg <- sprintf("`x` must be of class `%s`.", classname)
    stop(msg, call. = FALSE)
  }
}

sanitize_notes <- function(notes) {
  if (is.character(notes) && length(notes) > 0) {
    notes <- as.list(notes)
  }
  assert_list(notes, null.ok = TRUE)
  for (idx in seq_along(notes)) {
    n <- notes[[idx]]
    bad <- FALSE
    if (is.list(n)) {
      if (is.null(names(notes)[idx])) {
        bad <- TRUE
      }
      if (!all(c("i", "j", "text") %in% names(n))) {
        bad <- TRUE
      }
    } else if (!is.character(n) || length(n) != 1) {
      bad <- TRUE
    }
    if (isTRUE(bad)) {
      stop(
        "`notes` includes invalid elements. Please refer to the documentation for details.",
        call. = FALSE
      )
    }
  }
  return(notes)
}

sanitize_replace <- function(replace) {
  if (isTRUE(replace)) {
    # only target real NA/NaN values: format_vector_replace() also matches
    # against the original (pre-formatting) data, so formatted missing values
    # are caught without blanking literal "NA"/"NaN" strings (e.g. the
    # country code "NA" for Namibia)
    replace <- list(" " = c(NA, NaN))
  } else if (isFALSE(replace)) {
    replace <- list(NULL)
  } else if (isTRUE(check_string(replace))) {
    if (identical(replace, "")) {
      replace <- " "
    }
    replace <- stats::setNames(list(NA, NaN), c(replace, replace))
  } else if (!is.list(replace) || is.null(names(replace))) {
    stop(
      "`replace` should be TRUE/FALSE, a single string, or a named list.",
      call. = FALSE
    )
  }
  return(replace)
}

sanity_num_mark <- function(digits, num_mark_big, num_mark_dec) {
  # commented out because doesn't work with french decimals
  if (identical(getOption("OutDec"), ".")) {
    if (is.null(digits)) {
      if (num_mark_big != "") {
        stop("`num_mark_big` requires a `digits` value.", call. = FALSE)
      }
      if (num_mark_dec != ".") {
        stop("`num_mark_dec` requires a `digits` value.", call. = FALSE)
      }
    }
  }
}

footnote_markers <- function(x) {
  notes <- x@notes
  tab <- x@data_body

  # Define superscript syntax for each output format
  superscript_syntax <- list(
    latex = "\\textsuperscript{%s}",
    html = "<sup>%s</sup>",
    bootstrap = "<sup>%s</sup>",
    tabulator = "<sup>%s</sup>",
    typst = "#super[%s]",
    markdown = "^%s^"
  )

  # Get the appropriate superscript format
  sup_format <- superscript_syntax[[x@output]] %||%
    superscript_syntax[["markdown"]]

  # Helper function to add superscript to a value
  add_superscript <- function(value, sup) {
    paste0(value, sprintf(sup_format, sup))
  }

  # Helper function to update column names and x@names
  update_column_names <- function(x, n, sup) {
    if (0 %in% n$i) {
      colnames(tab)[n$j] <- add_superscript(colnames(tab)[n$j], sup)
      x@names[n$j] <- add_superscript(x@names[n$j], sup)
    }
    return(x)
  }

  for (idx in seq_along(notes)) {
    n <- notes[[idx]]
    sup <- names(notes)[idx]

    if (is.list(n)) {
      # Add superscript to table cells
      tab[n$i, n$j] <- add_superscript(tab[n$i, n$j], sup)

      # Update column names and x@names if needed
      x <- update_column_names(x, n, sup)
    }
  }

  x@data_body <- tab
  return(x)
}

#' tinytable S4 method
#'
#' @keywords internal
setMethod(
  f = "group_eval_j",
  signature = "tinytable_tabularray",
  definition = function(x, i = NULL, j = NULL, ...) {
    # Only handle column grouping - row insertions now use matrix insertion
    x <- group_tabularray_col(x, j, ...)
    return(x)
  })

group_tabularray_col <- function(x, j, ...) {
  # Process column groups from @group_data_j
  if (nrow(x@group_data_j) > 0) {
    # Process each header row from bottom to top (reverse order) to match expected header order
    for (row_idx in nrow(x@group_data_j):1) {
      group_row <- as.character(x@group_data_j[row_idx, ])

      # Build and insert header row
      header_line <- build_tabularray_header(group_row, ncol(x))
      x <- insert_tabularray_header(x, header_line)
    }
  }

  return(x)
}

#' Apply a single style entry to the rectangular style dataframe
#' @param style_other Rectangular dataframe with all cells
#' @param style_row Single row from @style dataframe
#' @return Modified style_other dataframe
#' @keywords internal
#' @noRd
apply_style_to_rect <- function(style_other, style_row) {
  # Determine which cells this entry applies to
  if (is.na(style_row$i)) {
    # Apply to all rows
    i_vals <- unique(style_other$i)
  } else {
    i_vals <- style_row$i
  }

  if (is.na(style_row$j)) {
    # Apply to all columns
    j_vals <- unique(style_other$j)
  } else {
    j_vals <- style_row$j
  }

  # Find matching rows in style_other
  mask <- style_other$i %in% i_vals & style_other$j %in% j_vals

  # Overwrite non-NA values from style_row
  for (prop in STYLE_PROPS) {
    if (prop %in% names(style_row) && !is.na(style_row[[prop]])) {
      style_other[[prop]][mask] <- style_row[[prop]]
    }
  }

  return(style_other)
}


#' Append line style entry to the style_lines dataframe
#' @param style_lines Dataframe accumulating line entries
#' @param style_row Single row from @style dataframe
#' @param rect Full rectangular grid of (i,j) cells
#' @return Modified style_lines dataframe with new entries appended
#' @keywords internal
#' @noRd
append_lines_to_rect <- function(style_lines, style_row, rect) {
  # Skip if no line property
  if (is.na(style_row$line)) {
    return(style_lines)
  }

  # Determine which cells this line entry applies to
  if (is.na(style_row$i) && is.na(style_row$j)) {
    # Apply to all cells
    i_vals <- unique(rect$i)
    j_vals <- unique(rect$j)
  } else if (is.na(style_row$i)) {
    # Apply to all rows in these columns
    i_vals <- unique(rect$i)
    j_vals <- style_row$j
  } else if (is.na(style_row$j)) {
    # Apply to all columns in these rows
    i_vals <- style_row$i
    j_vals <- unique(rect$j)
  } else {
    # Specific cells
    i_vals <- style_row$i
    j_vals <- style_row$j
  }

  # Create new entries for all matching (i,j) combinations
  new_entries <- expand.grid(i = i_vals, j = j_vals, stringsAsFactors = FALSE)

  # Add line properties
  new_entries$line <- style_row$line
  new_entries$line_color <- if ("line_color" %in% names(style_row)) style_row$line_color else NA
  new_entries$line_width <- if ("line_width" %in% names(style_row)) style_row$line_width else NA
  new_entries$line_type <- if ("line_type" %in% names(style_row)) style_row$line_type else NA_character_
  new_entries$line_trim <- if ("line_trim" %in% names(style_row)) style_row$line_trim else NA

  # Append to style_lines
  if (nrow(style_lines) == 0) {
    return(new_entries)
  } else {
    return(rbind(style_lines, new_entries))
  }
}


#' Resolve all entries of @style into the rectangular @style_other grid and
#' the line list @style_lines in a single batched pass.
#'
#' This is functionally equivalent to (and produces value-identical output as)
#' the per-row loop:
#' ```r
#'   for (idx in seq_len(nrow(x@style))) {
#'     style_other <- apply_style_to_rect(style_other, x@style[idx,])
#'     style_lines <- append_lines_to_rect(style_lines, x@style[idx,], rect)
#'   }
#' ```
#' but is O(N + cells) instead of O(N * cells * props) by avoiding the per-row
#' full-rect mask scan. The naive loop becomes the dominant bottleneck once a
#' table accumulates more than a few hundred style_tt() entries (which is
#' common when styling individual cells for heat-map-like effects, alternating
#' row backgrounds, or per-column alignment).
#'
#' @param style_other Rectangular dataframe of (i, j, <props>) initialized with NAs
#' @param style_df @style data frame (one row per style_tt setting entry)
#' @return list(other=, lines=) with the populated data frames
#' @keywords internal
#' @noRd
resolve_styles_batch <- function(style_other, style_df) {
  n <- nrow(style_df)
  empty_lines <- data.frame(
    i = integer(0), j = integer(0),
    line = character(0),
    line_color = character(0),
    line_width = numeric(0),
    line_type = character(0),
    line_trim = character(0),
    stringsAsFactors = FALSE
  )
  if (n == 0) {
    return(list(
      other = style_other,
      lines = empty_lines,
      has_style = rep(FALSE, nrow(style_other))
    ))
  }

  # Unique i and j present in style_other (used when i or j is NA meaning "all")
  so_i <- style_other$i
  so_j <- style_other$j
  uniq_i <- unique(so_i)
  uniq_j <- unique(so_j)

  idx_matrix <- matrix(
    NA_integer_,
    nrow = length(uniq_i),
    ncol = length(uniq_j),
    dimnames = list(as.character(uniq_i), as.character(uniq_j))
  )
  idx_matrix[cbind(match(so_i, uniq_i), match(so_j, uniq_j))] <- seq_along(so_i)

  sd_i <- style_df$i
  sd_j <- style_df$j
  i_na <- is.na(sd_i)
  j_na <- is.na(sd_j)

  target_i <- vector("list", n)
  target_j <- vector("list", n)
  target_idx <- vector("list", n)
  for (k in seq_len(n)) {
    target_i[[k]] <- if (i_na[k]) uniq_i else sd_i[k]
    target_j[[k]] <- if (j_na[k]) uniq_j else sd_j[k]

    i_pos <- if (i_na[k]) seq_along(uniq_i) else match(sd_i[k], uniq_i)
    j_pos <- if (j_na[k]) seq_along(uniq_j) else match(sd_j[k], uniq_j)
    if (anyNA(i_pos) || anyNA(j_pos)) {
      target_idx[[k]] <- integer(0)
    } else {
      idx <- as.vector(idx_matrix[i_pos, j_pos, drop = FALSE])
      target_idx[[k]] <- idx[!is.na(idx)]
    }
  }

  has_style <- rep(FALSE, nrow(style_other))

  # For each property, walk style rows that have a non-NA value, in order,
  # and write directly into style_other at the matching linear indices.
  # "Last write wins" semantics is preserved by iterating in source order.
  for (prop in STYLE_PROPS) {
    if (!prop %in% names(style_df)) next
    vals <- style_df[[prop]]
    has <- which(!is.na(vals))
    if (length(has) == 0L) next
    for (k in has) {
      idx <- target_idx[[k]]
      if (length(idx)) {
        style_other[[prop]][idx] <- vals[[k]]
        has_style[idx] <- TRUE
      }
    }
  }

  # Lines are additive: expand every style_df row that has a non-NA `line`
  # into one entry per (i, j) cell, then rbind() once at the end.
  has_line <- which(!is.na(style_df$line))
  if (length(has_line)) {
    line_color_col <- if ("line_color" %in% names(style_df)) style_df$line_color else rep(NA_character_, n)
    line_width_col <- if ("line_width" %in% names(style_df)) style_df$line_width else rep(NA_real_, n)
    line_type_col  <- if ("line_type"  %in% names(style_df)) style_df$line_type  else rep(NA_character_, n)
    line_trim_col  <- if ("line_trim"  %in% names(style_df)) style_df$line_trim  else rep(NA_character_, n)

    line_entries <- vector("list", length(has_line))
    for (kp in seq_along(has_line)) {
      k <- has_line[kp]
      # Issue #664: unname targets before creating data.frame() to avoid
      # "row names contain missing values" when grouping order leaves NA names.
      ri <- as.integer(unname(target_i[[k]]))
      rj <- as.integer(unname(target_j[[k]]))
      # cartesian product (preserves append_lines_to_rect ordering)
      line_entries[[kp]] <- data.frame(
        i = rep(ri, times = length(rj)),
        j = rep(rj, each  = length(ri)),
        line = style_df$line[k],
        line_color = line_color_col[k],
        line_width = line_width_col[k],
        line_type  = line_type_col[k],
        line_trim  = line_trim_col[k],
        stringsAsFactors = FALSE,
        row.names = NULL
      )
    }
    style_lines <- do.call(rbind, line_entries)
  } else {
    style_lines <- empty_lines
  }

  list(other = style_other, lines = style_lines, has_style = has_style)
}


#' Filter a rectangular @style_other grid to rows relevant for a backend
#' @keywords internal
#' @noRd
filter_style_other <- function(style_other, style_cols) {
  if (is.null(style_other) || nrow(style_other) == 0) {
    return(style_other)
  }

  style_cols <- intersect(style_cols, names(style_other))
  if (length(style_cols) == 0L) {
    return(style_other[0, , drop = FALSE])
  }

  has_style <- attr(style_other, "has_style", exact = TRUE)
  if (!is.null(has_style) && length(has_style) == nrow(style_other)) {
    style_other <- style_other[has_style, , drop = FALSE]
  }

  if (nrow(style_other) == 0) {
    return(style_other)
  }

  has_backend_style <- rowSums(!is.na(style_other[, style_cols, drop = FALSE])) > 0
  style_other[has_backend_style, , drop = FALSE]
}


# NOTE: expand_lines() and expand_style() have been removed.
# Line styles are now handled via @style_lines in build_tt()
# Other styles are handled via @style_other in build_tt()

source("helpers.R")

style_resolution_props <- c(
  "bold", "italic", "underline", "strikeout", "monospace", "smallcap",
  "align", "alignv", "color", "background", "fontsize", "indent",
  "html_css", "colspan", "rowspan"
)

make_style_rect <- function() {
  rect <- expand.grid(i = c(0L, 1L, 2L), j = 1:3)
  for (col in style_resolution_props) {
    rect[[col]] <- NA
  }
  rect
}

make_style_settings <- function() {
  data.frame(
    i = c(NA, 1, 2, NA, 1, 2),
    j = c(1, NA, 2, NA, 3, 3),
    bold = c(TRUE, NA, FALSE, NA, TRUE, NA),
    italic = c(NA, TRUE, NA, NA, FALSE, NA),
    underline = NA,
    strikeout = NA,
    monospace = NA,
    smallcap = NA,
    align = c(NA, "l", NA, NA, NA, NA),
    alignv = NA,
    color = c("red", NA, "blue", NA, "green", NA),
    background = c(NA, "yellow", NA, NA, NA, NA),
    fontsize = c(NA, 1.2, NA, NA, 0.9, NA),
    indent = c(NA, NA, 2, NA, NA, NA),
    html_css = c(NA, NA, NA, NA, NA, "font-feature-settings: 'tnum'"),
    colspan = NA,
    rowspan = NA,
    line = c(NA, "b", "t", "b", "l", NA),
    line_color = c(NA, "red", "blue", "gray", "black", NA),
    line_width = c(NA, 0.1, 0.2, 0.4, 0.3, NA),
    line_type = c(NA, "dashed", NA, "dotted", NA, NA),
    line_trim = c(NA, "lr", NA, NA, NA, NA),
    stringsAsFactors = FALSE
  )
}

resolve_styles_reference <- function(rect, settings) {
  other <- rect
  lines <- data.frame()
  for (idx in seq_len(nrow(settings))) {
    row <- settings[idx, , drop = FALSE]
    other <- tinytable:::apply_style_to_rect(other, row)
    lines <- tinytable:::append_lines_to_rect(lines, row, rect)
  }
  list(other = other, lines = lines)
}

rect <- make_style_rect()
settings <- make_style_settings()
reference <- resolve_styles_reference(rect, settings)
resolved <- tinytable:::resolve_styles_batch(rect, settings)

expect_true(isTRUE(all.equal(reference$other, resolved$other, check.attributes = FALSE)))
expect_true(isTRUE(all.equal(reference$lines, resolved$lines, check.attributes = FALSE)))

expected_has_style <- rowSums(!is.na(reference$other[, style_resolution_props, drop = FALSE])) > 0
expect_identical(resolved$has_style, expected_has_style)

expect_true(exists("filter_style_other", envir = asNamespace("tinytable"), inherits = FALSE))
if (exists("filter_style_other", envir = asNamespace("tinytable"), inherits = FALSE)) {
  attr(resolved$other, "has_style") <- resolved$has_style
  compact <- tinytable:::filter_style_other(resolved$other, style_resolution_props)
  expected <- reference$other[expected_has_style, , drop = FALSE]

  expect_true(nrow(compact) < nrow(resolved$other))
  expect_true(isTRUE(all.equal(expected, compact, check.attributes = FALSE)))
}

empty_resolved <- tinytable:::resolve_styles_batch(rect, data.frame())
expect_identical(empty_resolved$has_style, rep(FALSE, nrow(rect)))
attr(empty_resolved$other, "has_style") <- empty_resolved$has_style
expect_identical(
  tinytable:::filter_style_other(empty_resolved$other, style_resolution_props),
  empty_resolved$other[0, , drop = FALSE]
)

#' tinytable S4 method
#'
#' @keywords internal
setMethod(
  f = "group_eval_j",
  signature = "tinytable_html",
  definition = function(x, i = NULL, j = NULL, ihead = NULL, ...) {
    # Only handle column grouping - row insertions now use matrix insertion
    if (!is.null(j)) {
      x <- html_groupj(x, j = j, ihead = ihead, ...)
    }
    return(x)
  })

html_groupj <- function(x, j, ihead, ...) {
  # Check if there are any column groups to process
  if (nrow(x@group_data_j) == 0) {
    return(x)
  }

  all_groupj_rows <- list()

  # Process each row in @group_data_j separately (from last to first to maintain proper order)
  for (groupj_idx in nrow(x@group_data_j):1) {
    groupj <- as.character(x@group_data_j[groupj_idx, ])

    # Calculate the appropriate ihead for this group row
    current_ihead <- ihead - (nrow(x@group_data_j) - groupj_idx)

    # Convert group row to column spans
    spans <- parse_group_spans(groupj)

    if (nrow(spans) > 0) {
      # Create HTML for this group row
      group_html <- html_groupj_html(x, spans, current_ihead)
      all_groupj_rows[[groupj_idx]] <- group_html
    }
  }

  if (length(all_groupj_rows) > 0) {
    x <- html_groupj_insert(x, all_groupj_rows)
  }

  return(x)
}


# Helper function to create HTML for a group row
# `spans` is the data.frame returned by parse_group_spans(): label/start/end,
# positional so duplicate labels are preserved.
html_groupj_html <- function(x, spans, ihead) {
  # Add missing columns as empty single-column groups
  covered <- unlist(Map(seq.int, spans$start, spans$end))
  miss <- setdiff(seq_len(ncol(x)), covered)
  if (length(miss) > 0) {
    spans <- rbind(
      spans,
      data.frame(label = " ", start = miss, end = miss, stringsAsFactors = FALSE)
    )
  }

  # Sort by column position
  spans <- spans[order(spans$end), , drop = FALSE]

  # Generate HTML for each group
  jstring <- lapply(seq_len(nrow(spans)), function(k) {
    colspan_val <- spans$end[k] - spans$start[k] + 1

    # Calculate width style if x@width has multiple values (individual column widths)
    width_style <- ""
    if (length(x@width) > 1 && colspan_val > 1) {
      # Sum the widths of the columns this header spans
      spanned_cols <- spans$start[k]:spans$end[k]
      total_width <- sum(x@width[spanned_cols]) / sum(x@width) * 100
      width_style <- sprintf(' style="width: %s;"', format_markup_unit(round(total_width, 2), "%"))
    }

    sprintf(
      '<th scope="col" align="center" colspan=%s data-row="%d" data-col="%d"%s>%s</th>',
      colspan_val,
      ihead,
      spans$start[k], # Use the first column of the span, not the loop index
      width_style,
      spans$label[k]
    )
  })

  jstring <- paste(unlist(jstring), collapse = "\n")
  sprintf("<tr>\n%s\n</tr>", jstring)
}

# Helper function to insert group rows into the table
html_groupj_insert <- function(x, groupj_rows) {
  out <- strsplit(x@table_string, "\\n")[[1]]

  all_jstrings <- paste(groupj_rows, collapse = "\n")
  # Anchor on the template's own indented <thead> line so injected content
  # (e.g. a caption containing the literal string "<thead>") cannot match.
  idx <- grep("^        <thead>$", out)[1]
  out <- c(out[seq_len(idx)], all_jstrings, out[(idx + 1):length(out)])
  out <- paste(out, collapse = "\n")

  x@table_string <- out
  x
}

# -----------------------------------------------------------------------------
# constants
# -----------------------------------------------------------------------------
TAB_NUM <- c("integer", "numeric", "double")
TAB_DATE <- c("Date", "POSIXct", "POSIXlt")
TAB_FORM <- c(TAB_NUM, "logical", TAB_DATE)

# small helpers
merge_lists <- function(x, y) { # shallow merge; rhs wins
    x[names(y)] <- y
    x
}

# Translate an R strptime format string to Luxon tokens for Tabulator's
# datetime formatter (outputFormat). Known %-tokens are mapped; literal ASCII
# letters are wrapped in single quotes so Luxon does not interpret them as
# tokens; other characters (separators, spaces, ...) pass through as-is.
# Strings without any % token (e.g. an already-Luxon "M/d/yyyy") are returned
# unchanged.
strptime_to_luxon <- function(fmt) {
    if (!is.character(fmt) || length(fmt) != 1 || !grepl("%", fmt, fixed = TRUE)) {
        return(fmt)
    }
    map <- c(
        "Y" = "yyyy", "y" = "yy", "m" = "MM", "d" = "dd", "e" = "d",
        "H" = "HH", "I" = "hh", "M" = "mm", "S" = "ss", "p" = "a",
        "B" = "MMMM", "b" = "MMM", "A" = "cccc", "a" = "ccc", "j" = "ooo"
    )
    chars <- strsplit(fmt, "", fixed = TRUE)[[1]]
    out <- character(0)
    i <- 1
    n <- length(chars)
    while (i <= n) {
        if (chars[i] == "%" && i < n) {
            token <- chars[i + 1]
            if (token == "%") {
                out <- c(out, "%")
            } else if (token %in% names(map)) {
                out <- c(out, map[[token]])
            } else {
                # unknown token: pass through untranslated
                out <- c(out, "%", token)
            }
            i <- i + 2
        } else if (grepl("[A-Za-z]", chars[i])) {
            # quote runs of literal letters so Luxon treats them as text
            run <- i
            while (run < n && grepl("[A-Za-z]", chars[run + 1])) {
                run <- run + 1
            }
            out <- c(out, "'", chars[i:run], "'")
            i <- run + 1
        } else {
            out <- c(out, chars[i])
            i <- i + 1
        }
    }
    paste(out, collapse = "")
}

# -----------------------------------------------------------------------------
# single formatter registry (data-driven)
# Each entry returns a list(title/field/formatter/params/...),
# given (col_def, args).
# -----------------------------------------------------------------------------
tabulator_tickcross_params <- function() {
    list(
        allowEmpty = TRUE,
        allowTruthy = TRUE,
        tickElement = "<i class='fa-solid fa-check'></i>",
        crossElement = "<i class='fa-solid fa-xmark'></i>"
    )
}

tabulator_column_registry <- list(
    numeric = function(col_def, args) {
        digits <- args$digits %||% get_option("tinytable_format_digits")
        thousand <- args$num_mark_big %||% get_option("tinytable_format_num_mark_big", "")
        decimal <- args$num_mark_dec %||% get_option("tinytable_format_num_mark_dec", get_option("OutDec", "."))
        # Note: Tabulator's `money` formatter only supports a fixed number of
        # decimal places (`precision`), so `num_fmt` ("significant",
        # "scientific", ...) and `num_zero` are not honored client-side.

        merge_lists(col_def, list(
            formatter = "money",
            formatterParams = list(
                decimal = decimal,
                thousand = thousand,
                precision = digits %||% 2,
                symbol = "",
                symbolAfter = FALSE
            )
        ))
    },
    logical = function(col_def, args) {
        bool_fun <- args$bool %||% get_option("tinytable_format_bool")
        if (!is.null(bool_fun) && is.function(bool_fun)) {
            merge_lists(col_def, list(
                formatter = "plaintext"
            ))
        } else {
            merge_lists(col_def, list(
                formatter = "tickCross",
                formatterParams = tabulator_tickcross_params()
            ))
        }
    },
    Date = function(col_def, args) {
        fmt_out <- strptime_to_luxon(args$date %||% get_option("tinytable_format_date") %||% "M/d/yyyy")
        merge_lists(col_def, list(
            formatter = "datetime",
            sorter = "datetime",
            formatterParams = list(
                inputFormat = "yyyy-MM-dd",
                outputFormat = fmt_out,
                invalidPlaceholder = ""
            ),
            sorterParams = list(
                format = "yyyy-MM-dd",
                alignEmptyValues = "bottom"
            )
        ))
    },
    POSIXct = function(col_def, args) {
        fmt_out <- strptime_to_luxon(args$date %||% get_option("tinytable_format_date") %||% "M/d/yyyy HH:mm:ss")
        merge_lists(col_def, list(
            formatter = "datetime",
            sorter = "datetime",
            formatterParams = list(
                inputFormat = "yyyy-MM-dd HH:mm:ss",
                outputFormat = fmt_out,
                invalidPlaceholder = ""
            ),
            sorterParams = list(
                format = "yyyy-MM-dd HH:mm:ss",
                alignEmptyValues = "bottom"
            )
        ))
    },
    POSIXlt = function(col_def, args) {
        tabulator_column_registry$POSIXct(col_def, args)
    })

# -----------------------------------------------------------------------------
# inject the columns JSON into the table string exactly once
# - replaces the template placeholder when still present, otherwise rewrites the
#   previously injected `columns: [...]` line
# - the replacement is inserted literally (gsub replacement semantics would
#   otherwise mangle backslashes in properly escaped JSON)
# -----------------------------------------------------------------------------
tabulator_replace_columns_json <- function(table_string, columns_json) {
    if (grepl("$tinytable_TABULATOR_COLUMNS", table_string, fixed = TRUE)) {
        return(gsub(
            "$tinytable_TABULATOR_COLUMNS",
            columns_json,
            table_string,
            fixed = TRUE
        ))
    }
    # columns JSON is emitted on a single line: match to the last `]` on it
    literal <- gsub("\\", "\\\\", columns_json, fixed = TRUE)
    sub(
        "columns: \\[[^\n]*\\]",
        paste0("columns: ", literal),
        table_string
    )
}

# -----------------------------------------------------------------------------
# one function to build the columns, applying
# 1) base spec, 2) lazy-format overrides, 3) style overrides, 4) write JSON
# -----------------------------------------------------------------------------
tabulator_apply_columns <- function(x) {
    stopifnot(!is.null(x@names), length(x@names) > 0)

    # 1) base column specs
    # Clean the full vector at once so de-duplication is consistent
    fields <- tabulator_clean_column_name(x@names)
    columns <- lapply(seq_along(x@data), function(j) {
        col_name <- x@names[j]
        field <- fields[j]
        ctype <- class(x@data[[j]])[1]

        col_def <- list(title = col_name, field = field)

        if (ctype %in% TAB_FORM) {
            # pick the registry key
            key <- if (ctype %in% TAB_NUM) "numeric" else ctype
            formatter_fun <- tabulator_column_registry[[key]]
            if (!is.null(formatter_fun)) {
                col_def <- formatter_fun(col_def, args = list())
            }
        }

        col_def
    })
    names(columns) <- vapply(columns, `[[`, character(1), "title")

    # 2) lazy-format to per-column overrides (digits/date/marks/etc.)
    if (length(x@lazy_format) > 0) {
        # build a map: title -> merged formatter col_def
        for (l in x@lazy_format) {
            js <- if (is.null(l$j)) seq_along(x@data) else sanitize_j(l$j, x)
            for (j in js) {
                col_name <- x@names[j]
                ctype <- class(x@data[[j]])[1]
                if (!(ctype %in% TAB_FORM)) next

                key <- if (ctype %in% TAB_NUM) "numeric" else ctype
                f <- tabulator_column_registry[[key]]
                if (is.null(f)) next

                # args from lazy_format
                args <- list(
                    digits       = l$digits,
                    num_fmt      = l$num_fmt,
                    num_mark_big = l$num_mark_big,
                    num_mark_dec = l$num_mark_dec,
                    num_zero     = l$num_zero,
                    date         = l$date_format,
                    bool         = l$bool
                )
                columns[[col_name]] <- f(columns[[col_name]], args)
            }
        }
    }

    # 3) column-level styles (hozAlign, vertAlign, ...)
    if (length(x@tabulator_column_styles) > 0) {
        for (nm in names(x@tabulator_column_styles)) {
            if (!nm %in% names(columns)) next
            st <- x@tabulator_column_styles[[nm]]
            if (!is.null(st$hozAlign)) columns[[nm]]$hozAlign <- st$hozAlign
            if (!is.null(st$vertAlign)) columns[[nm]]$vertAlign <- st$vertAlign
        }
    }

    # 3.5) apply column formatters from plot_tt and other sources
    if (length(x@tabulator_column_formatters) > 0) {
        for (nm in names(x@tabulator_column_formatters)) {
            if (!nm %in% names(columns)) next
            fmt <- x@tabulator_column_formatters[[nm]]
            if (!is.null(fmt$formatter)) columns[[nm]]$formatter <- fmt$formatter
            if (!is.null(fmt$formatterParams)) columns[[nm]]$formatterParams <- fmt$formatterParams
            if (!is.null(fmt$sorter)) columns[[nm]]$sorter <- fmt$sorter
            if (!is.null(fmt$sorterParams)) columns[[nm]]$sorterParams <- fmt$sorterParams
        }
    }

    # 4) persist back and inject JSON
    x@tabulator_columns <- unname(columns)

    columns_json <- df_to_json(x@tabulator_columns)
    x@table_string <- tabulator_replace_columns_json(x@table_string, columns_json)

    x
}

#' Markdown theme with optional ANSI color support and grid customization
#'
#' @param x A `tinytable` object.
#' @param ansi Logical. If TRUE, enables ANSI color codes for grid styling; if FALSE, disables them. Default is NULL (no change; ANSI is disabled unless previously enabled).
#' @param style Character. Markdown style format. Can be "grid" or "gfm". Default is NULL.
#' @param hline Logical. Enable/disable horizontal lines. Default is TRUE.
#' @param hline_header Logical. Enable/disable the special header separator line below column names. Default is TRUE.
#' @param vline Logical. Enable/disable vertical lines. Default is TRUE.
#' @param ... Additional arguments (ignored).
#' @return A modified `tinytable` object.
#' @details
#' When `ansi = TRUE`, colors and text styling (bold, italic, strikeout, underline)
#' are applied using ANSI escape sequences for terminal display. ANSI colors require
#' a terminal or application that supports ANSI escape sequences. Common supported
#' terminals include: Terminal.app (macOS), iTerm2 (macOS), Windows Terminal, most
#' Linux terminals, RStudio Console, and VS Code terminal. Colors may not display
#' correctly in basic text editors or older terminals.
#' @export
theme_markdown <- function(x,
                           ansi = NULL,
                           style = NULL,
                           vline = NULL,
                           hline = NULL,
                           hline_header = NULL,
                           ...) {
  # Validate logical arguments
  assert_flag(ansi, null.ok = TRUE)
  assert_choice(style, c("grid", "gfm"), null.ok = TRUE)
  assert_flag(vline, null.ok = TRUE)
  assert_flag(hline, null.ok = TRUE)
  assert_flag(hline_header, null.ok = TRUE)

  # Set the ansi slot based on the argument
  if (is.logical(ansi)) x@ansi <- ansi

  # Set the markdown style slot if provided
  if (!is.null(style)) {
    x@markdown_style <- style
  }

  # Set the grid formatting slots
  if (is.logical(hline)) x@grid_hline <- hline
  if (is.logical(hline_header)) x@grid_hline_header <- hline_header
  if (is.logical(vline)) x@grid_vline <- vline

  return(x)
}

# =============================================================================
# TABULATOR CSS FINALIZATION
# =============================================================================

#' Apply custom CSS rules
#' @param x tinytable object
#' @return Modified tinytable object
#' @keywords internal
#' @noRd
tabulator_apply_css <- function(x) {
    if (nchar(x@tabulator_css_rule) > 0) {
        # Replace $TINYTABLE_ID with actual table ID
        table_id <- paste0("tinytable_", x@id)
        custom_css <- gsub(
            "\\$TINYTABLE_ID",
            paste0("#", table_id),
            x@tabulator_css_rule
        )
        css_block <- sprintf("<style>\n%s\n</style>", custom_css)

        # Replace CSS placeholder
        x@table_string <- gsub(
            "$tinytable_TABULATOR_CSS",
            css_block,
            x@table_string,
            fixed = TRUE
        )
    } else {
        # Clean up CSS placeholder if no custom CSS
        x@table_string <- gsub(
            "$tinytable_TABULATOR_CSS",
            "",
            x@table_string,
            fixed = TRUE
        )
    }

    return(x)
}

#'
#' https://stackoverflow.com/questions/9439256/how-can-i-handle-r-cmd-check-no-visible-binding-for-global-variable-notes-when#comment20826625_12429344
#' 2012 hadley says "globalVariables is a hideous hack and I will never use it"
#' 2014 hadley updates his own answer with globalVariables as one of "two solutions"
#' @noRd
utils::globalVariables(
  c(
    "bootstrap",
    "i",
    "j"
  )
)

ansi_nchar <- function(x, keepNA = TRUE, z = FALSE) {
  x <- as.character(x)
  is_na <- is.na(x)

  # Replace NA if keepNA = FALSE
  if (!keepNA) x[is_na] <- "NA"

  # Strip ANSI/CSI sequences
  x_stripped <- gsub("\u001B\\[[0-?]*[ -/]*[@-~]", "", x, perl = TRUE)

  if (z) {
    nzchar(x_stripped, keepNA = keepNA)
  } else {
    nchar(x_stripped, type = "width", keepNA = keepNA)
  }
}


ansi_pad <- function(txt, width) {
    w <- ansi_nchar(txt)
    pad <- pmax(width - w, 0)
    paste0(txt, strrep(" ", pad))
}


ansi_strwrap <- function(text, width) {
    # For ANSI text, only wrap if visual width exceeds target width
    visual_width <- ansi_nchar(text)
    if (visual_width <= width) {
        return(text)
    } else {
        # If it's too long, fall back to regular strwrap
        # This is a simplified implementation - could be enhanced for better ANSI handling
        return(strwrap(text, width = width))
    }
}

is_rstudio <- function() {
  identical(Sys.getenv("RSTUDIO"), "1")
}

is_positron <- function() {
  identical(Sys.getenv("POSITRON"), "1")
}

is_rstudio_notebook_context <- function(path, contents = NULL) {
  if (isTRUE(grepl("\\.qmd$|\\.Rmd$", path))) {
    return(TRUE)
  }

  contents <- paste(contents, collapse = "\n")
  if (!nzchar(contents)) {
    return(FALSE)
  }

  has_yaml_header <- grepl("^---\\s*\n", contents)
  has_notebook_format <- grepl("(?m)^\\s*(output|format)\\s*:", contents, perl = TRUE)
  has_chunk <- grepl("(?m)^```\\{[a-zA-Z]", contents, perl = TRUE)

  isTRUE(has_chunk || (has_yaml_header && has_notebook_format))
}

is_rstudio_notebook <- function() {
  flag <- FALSE
  # inline doesn't work in Positron
  if (is_rstudio()) {
    assert_dependency("rstudioapi")
    con <- rstudioapi::getActiveDocumentContext()
    flag <- is_rstudio_notebook_context(con[["path"]], con[["contents"]])
  }
  return(flag)
}

source("helpers.R")

expect_true(tinytable:::is_rstudio_notebook_context("/tmp/report.Rmd"))
expect_true(tinytable:::is_rstudio_notebook_context("/tmp/report.qmd"))
expect_false(tinytable:::is_rstudio_notebook_context("/tmp/script.R"))

unsaved_rmd <- c(
  "---",
  "title: Untitled",
  "output: html_document",
  "---",
  "",
  "```{r}",
  "tt(head(mtcars))",
  "```"
)

unsaved_qmd <- c(
  "---",
  "title: Untitled",
  "format: html",
  "---",
  "",
  "```{r}",
  "tt(head(mtcars))",
  "```"
)

unsaved_chunk <- c(
  "```{r}",
  "tt(head(mtcars))",
  "```"
)

expect_true(tinytable:::is_rstudio_notebook_context("", unsaved_rmd))
expect_true(tinytable:::is_rstudio_notebook_context("Untitled3", unsaved_qmd))
expect_true(tinytable:::is_rstudio_notebook_context("Untitled3", unsaved_chunk))
expect_false(tinytable:::is_rstudio_notebook_context("Untitled3", "tt(head(mtcars))"))

setMethod(
  f = "finalize",
  signature = "tinytable_tabulator",
  definition = function(x, ...) {
    # Apply stylesheet
    if (nchar(x@tabulator_stylesheet) > 0) {
      x <- tabulator_stylesheet(x, x@tabulator_stylesheet)
    }

    # Inject column definitions
    if (is.list(x@tabulator_columns) && !is.null(x@tabulator_columns$json_string)) {
      # User-supplied JSON string: inject verbatim
      x@table_string <- tabulator_replace_columns_json(x@table_string, x@tabulator_columns$json_string)
    } else if (isTRUE(attr(x@tabulator_columns, "user_defined"))) {
      # User-supplied list of column definitions: serialize and inject as-is
      x@table_string <- tabulator_replace_columns_json(x@table_string, df_to_json(x@tabulator_columns))
    } else {
      # Process columns (formatting, styling, conversion) only for basic columns
      x <- tabulator_apply_columns(x)
    }

    # Apply options
    x <- tabulator_apply_options(x)

    # Apply search functionality
    x <- tabulator_apply_search(x)

    # Apply custom CSS
    x <- tabulator_apply_css(x)

    # Apply post-initialization JavaScript (placeholder is removed when empty)
    x@table_string <- gsub(
      "$tinytable_TABULATOR_POST_INIT",
      x@tabulator_post_init,
      x@table_string,
      fixed = TRUE
    )

    # Clean up JS markers (formatters are now in inst/tinytable.js)
    x@tabulator_options <- gsub("\n// NEEDS_HISTOGRAM_JS", "", x@tabulator_options, fixed = TRUE)
    x@tabulator_options <- gsub("\n// NEEDS_SPARKLINE_JS", "", x@tabulator_options, fixed = TRUE)

    tabulator_cdn_tag <- sprintf('<script src="%s"></script>', tabulator_js_cdn)

    # Inline tinytable Tabulator theme stylesheet when in portable mode
    if (isTRUE(x@html_portable)) {
      x@table_string <- tabulator_inline_theme_css(x@table_string)
    }

    # Inject tinytable.js (inline if portable, external link otherwise)
    if (isTRUE(x@html_portable)) {
      # Inline the JS for portable HTML
      js_file <- system.file("tinytable.js", package = "tinytable")
      if (file.exists(js_file)) {
        js_content <- paste(readLines(js_file, warn = FALSE), collapse = "\n")
        js_tag <- sprintf("<script>\n%s\n</script>", js_content)
      } else {
        js_tag <- ""
      }

      tabulator_inline <- tabulator_inline_tabulator_js()
      if (!is.null(tabulator_inline)) {
        inline_tag <- paste0("    <script>\n", tabulator_inline, "\n    </script>")
        x@table_string <- sub(
          paste0("    ", tabulator_cdn_tag),
          inline_tag,
          x@table_string,
          fixed = TRUE
        )
      } else {
        warning(
          "Failed to inline Tabulator JavaScript in portable mode; falling back to CDN reference.",
          call. = FALSE
        )
      }
    } else {
      # External link (will work when package is installed)
      js_tag <- sprintf('<script src="%s"></script>', tinytable_js_cdn)
    }

    x@table_string <- gsub(
      "$tinytable_TINYTABLE_JS",
      js_tag,
      x@table_string,
      fixed = TRUE
    )

    # Handle custom script inclusion
    if (!is.null(x@html_script)) {
      x@table_string <- gsub(
        "$tinytable_SCRIPT",
        x@html_script,
        x@table_string,
        fixed = TRUE
      )
    } else {
      # Remove placeholder if no script is provided
      x@table_string <- lines_drop(
        x@table_string,
        "\\$tinytable_SCRIPT",
        fixed = FALSE,
        unique = FALSE
      )
    }

    return(x)
  })

tabulator_inline_tabulator_js <- function() {
  custom_path <- get_option("tinytable_tabulator_js_path", default = NULL)
  if (!is.null(custom_path) && file.exists(custom_path)) {
    return(paste(readLines(custom_path, warn = FALSE), collapse = "\n"))
  }

  package_path <- system.file("tabulator.min.js", package = "tinytable")
  if (nzchar(package_path) && file.exists(package_path)) {
    return(paste(readLines(package_path, warn = FALSE), collapse = "\n"))
  }

  inline_url <- get_option("tinytable_tabulator_js_url", default = tabulator_js_cdn)
  if (is.null(inline_url) || !nzchar(inline_url)) {
    return(NULL)
  }

  # Attempt to fetch from remote CDN as a last resort
  js_source <- tryCatch({
    con <- url(inline_url)
    on.exit(close(con), add = TRUE)
    suppressWarnings(paste(readLines(con, warn = FALSE), collapse = "\n"))
  }, error = function(e) NULL)

  return(js_source)
}

tabulator_inline_theme_css <- function(table_string) {
  css_path <- system.file("tabulator_tinytable.min.css", package = "tinytable")
  if (!file.exists(css_path)) {
    return(table_string)
  }

  css_content <- paste(readLines(css_path, warn = FALSE), collapse = "\n")
  pattern <- sprintf('([ \t]*)<link href="%s" rel="stylesheet">', tinytable_tabulator_theme_cdn)

  match <- regexpr(pattern, table_string, perl = TRUE)
  if (match[1] > 0) {
    match_len <- attr(match, "match.length")
    match_str <- substr(table_string, match[1], match[1] + match_len - 1)
    indent <- sub(pattern, "\\1", match_str, perl = TRUE)
    replacement <- paste0(indent, "<style>\n", css_content, "\n", indent, "</style>")
    table_string <- paste0(
      substr(table_string, 1, match[1] - 1),
      replacement,
      substr(table_string, match[1] + match_len, nchar(table_string))
    )
  }

  return(table_string)
}

# Shared CSS template for the revealjs theme. Placeholders (in order):
# body font size, header/last-row border color, caption color, caption font size.
css_revealjs_template <- "
/* tables */

.reveal table {
  /* height: auto; Adjust table width to fit content up to the available slide space */
  margin: auto;
  border-collapse: collapse;
  border-spacing: 0;
}

.reveal table th,
.reveal table td {
  border: none;
  padding: .23em;
  font-size: %sem;
}

/* Adds a bottom border to the table header row for distinction */
.reveal table thead th,
.reveal .slides table tr:last-child td,
.reveal .slides table {
  border-bottom: 2px solid %s;
}

/* Make column headers bold */
.reveal table thead th {
}

/* Styling table captions */
.reveal table caption {
  color: %s;
  font-size: %sem;
}
"

# Extra rules appended for the dark variant only.
css_revealjs_dark_addendum <- "
.reveal table {
  background-color: #2d2d2d; /* Dark background for tables */
  color: white; /* White text color */
}

.reveal table th,
.reveal table td {
  color: white; /* White text color */
}
"

#' RevealJS presentation theme
#'
#' @param x A `tinytable` object.
#' @param css String. CSS theme: "light" (default) or "dark".
#' @param fontsize Numeric. Font size multiplier for table content.
#' @param fontsize_caption Numeric. Font size multiplier for table captions.
#' @return A modified `tinytable` object.
#' @export
theme_revealjs <- function(
  x,
  css = get_option("tinytable_revealjs_css", default = "light"),
  fontsize = get_option("tinytable_revealjs_fontsize", default = 0.8),
  fontsize_caption = get_option("tinytable_revealjs_fontsize_caption", default = 1)) {
  if (css == "light") {
    css <- sprintf(
      css_revealjs_template,
      format_markup_num(fontsize),
      "#D3D3D3",
      "#666666", # Dark grey color for the caption
      format_markup_num(fontsize_caption)
    )
  } else if (css == "dark") {
    css <- paste0(
      sprintf(
        css_revealjs_template,
        format_markup_num(fontsize),
        "white", # White border color for dark theme
        "white", # White color for the caption
        format_markup_num(fontsize_caption)
      ),
      css_revealjs_dark_addendum
    )
  }
  x <- theme_html(x, engine = "tinytable", css_rule = css)
  return(x)
}

format_vector_markdown <- function(vec, output_format, ...) {
  if (is.null(output_format)) {
    return(NULL)
  }

  if (output_format %in% c("html", "bootstrap", "tabulator")) {
    vapply(
      vec,
      function(k) {
        k <- litedown::mark(I(k), "html")
        k <- sub("<p>", "", k, fixed = TRUE)
        k <- sub("</p>", "", k, fixed = TRUE)
        return(k)
      },
      character(1)
    )
  } else if (output_format == "latex") {
    vapply(
      vec,
      function(k) {
        litedown::mark(I(k), "latex")
      },
      character(1)
    )
  } else {
    vec
  }
}

#' Convert delimiter-based column names to named list
#'
#' @keywords internal
#' @noRd
j_delim_to_named_list <- function(x, j) {
  nm <- x@names

  # Preprocess column names: add space at start/end if they start/end with delimiter
  for (i in seq_along(nm)) {
    # Check if ends with delimiter
    # Do not check if starts with delimiter, because we sometimes want empty groups in the top header
    if (endsWith(nm[i], j)) {
      nm[i] <- paste0(nm[i], " ")
    }
  }

  # Find which columns contain the delimiter
  indices <- grepl(j, nm, fixed = TRUE)

  if (!any(indices)) {
    return(list(colnames = nm, groupnames = NULL))
  }

  # For columns with delimiters, parse the hierarchical structure
  grouped_cols <- nm[indices]

  # Split each column name by the delimiter to get all levels
  split_names <- strsplit(grouped_cols, j, fixed = TRUE)
  split_lengths <- lengths(split_names)

  if (length(unique(split_lengths)) > 1) {
    delim_counts <- split_lengths - 1
    details <- paste(
      sprintf("%s (%d)", grouped_cols, delim_counts),
      collapse = ", "
    )
    stop(
      sprintf(
        "Each column name must have the same number of delimiters. Got: %s.",
        details
      ),
      call. = FALSE
    )
  }
  max_levels <- max(lengths(split_names))

  # Build the nested grouping structure from the outermost level down
  groupings <- list()
  current_colnames <- nm

  for (level in 1:max_levels) {
    # Extract the prefix for this level (first 'level' parts)
    prefixes <- character(length(grouped_cols))
    for (i in seq_along(split_names)) {
      if (length(split_names[[i]]) >= level) {
        prefixes[i] <- paste(split_names[[i]][1:level], collapse = j)
      } else {
        prefixes[i] <- paste(split_names[[i]], collapse = j)
      }
    }

    # Extract just the group name at this level
    level_names <- character(length(grouped_cols))
    for (i in seq_along(split_names)) {
      if (length(split_names[[i]]) >= level) {
        level_names[i] <- split_names[[i]][level]
      } else {
        level_names[i] <- ""
      }
    }

    unique_level_names <- unique(level_names[level_names != ""])
    level_groups <- list()
    
    for (level_name in unique_level_names) {
      unique_prefixes <- unique(prefixes[level_names == level_name])
      for (prefix in unique_prefixes) {
        matching_indices <- which(prefixes == prefix)
        col_indices <- (which(indices))[matching_indices]

        # Only create groupings with more than one column or if it's not the final level
        if (length(col_indices) > 1 || level < max_levels) {
          # Use the level name as the group name
          group_name <- level_name
          
          # Handle empty group names by replacing with a space (as per tinytable convention)
          if (is.na(group_name) || group_name == "") {
            group_name <- " "
          }

          # If this group name already exists, append to it
          if (group_name %in% names(level_groups)) {
            level_groups[[group_name]] <- c(level_groups[[group_name]], col_indices)
          } else {
            level_groups[[group_name]] <- col_indices
          }
        }
      }
    }

    # Only add level if it has meaningful groups
    if (length(level_groups) > 0) {
      groupings <- c(groupings, list(level_groups))
    }
  }

  # Set final column names (the last part after all delimiters)
  for (i in which(indices)) {
    parts <- split_names[[which(which(indices) == i)]]
    final_name <- parts[length(parts)]

    # Handle empty final names
    if (is.na(final_name) || final_name == "") {
      final_name <- " "
    }

    current_colnames[i] <- final_name
  }

  out <- list(colnames = current_colnames, groupnames = groupings)
  return(out)
}

#' Validate and process group indices
#'
#' @keywords internal
#' @noRd
sanitize_group_index <- function(idx, hi, orientation, warn = TRUE) {
  if (is.null(idx)) {
    return(idx)
  }
  assert_list(idx, named = TRUE)
  for (n in names(idx)) {
    if (orientation == "row") {
      assert_integerish(idx[[n]], len = 1, lower = 1, upper = hi, name = n)
    } else {
      assert_integerish(idx[[n]], lower = 1, upper = hi, name = n)
    }
  }
  # allow duplicated indices for consecutive rows
  # if (anyDuplicated(unlist(idx)) > 0) stop("Duplicate group indices.", call. = FALSE)
  out <- lapply(seq_along(idx), function(k) {
    v <- sort(unique(idx[[k]]))
    span <- min(v):max(v)
    # spans must be contiguous; warn instead of silently swallowing interior columns
    if (isTRUE(warn) && length(v) != length(span)) {
      warning(
        sprintf(
          "Group \"%s\": requested %ss (%s) are not contiguous. The label will span %ss %d to %d.",
          names(idx)[k],
          orientation,
          paste(idx[[k]], collapse = ", "),
          orientation,
          min(v),
          max(v)
        ),
        call. = FALSE
      )
    }
    span
  })
  names(out) <- names(idx)
  return(out)
}

#' Convert vector to list format for grouping
#'
#' @keywords internal
#' @noRd
sanitize_group_vec2list <- function(vec) {
  if (is.factor(vec)) {
    vec <- as.character(vec)
  }
  rle_result <- rle(vec)
  idx <- cumsum(c(1, utils::head(rle_result$lengths, -1)))
  idx <- as.list(idx)
  names(idx) <- rle_result$values
  return(idx)
}

#' Default theme for TinyTable
#'
#' @param x A tinytable object.
#' @param ... Additional arguments are ignored.
#' @return A modified `tinytable` object.
#' @export
theme_tinytable <- function(x, ...) {
  # run this after rbind_body_groupi()
  # Placement for LaTeX and Typst
  x <- theme_latex(x, placement = get_option("tinytable_latex_placement", default = NULL))
  x <- theme_typst(x)

  fn <- function(x) {
    if (identical(x@output, "html") && identical(x@html_engine, "tabulator")) {
      return(x)
    }
    # Use CSS variable when default tinytable.css is loaded, otherwise "black"
    col <- get_default_line_color(x)
    # top
    x <- style_tt(
      x,
      i = -x@nhead + 1,
      line = "t",
      line_color = col,
      line_width = 0.08
    )
    # middle
    if (length(x@names) > 0) {
      x <- style_tt(
        x,
        i = 0,
        line = "b",
        line_color = col,
        line_width = 0.05
      )
    }
    # bottom
    x <- style_tt(
      x,
      i = nrow(x),
      line = "b",
      line_color = col,
      line_width = 0.08
    )
  }
  x <- build_prepare(x, fn)

  return(x)
}

# Format numeric values with different formats
# digits check needs to be done here to avoid the other() formatting from ori, which zaps the original setting
format_vector_numeric <- function(
    vec,
    num_suffix,
    digits,
    num_mark_big,
    num_mark_dec,
    num_zero,
    num_fmt,
    ...) {
  if (!is.numeric(vec) || is.null(digits)) {
    return(NULL)
  }

  # numeric suffix
  if (isTRUE(num_suffix)) {
    out <- format_num_suffix(
      vec,
      digits = digits,
      num_mark_big = num_mark_big,
      num_mark_dec = num_mark_dec,
      num_zero = num_zero,
      num_fmt = num_fmt
    )
    # non-integer numeric
  } else if (!isTRUE(check_integerish(vec))) {
    out <- format_non_integer_numeric(
      vec,
      digits = digits,
      num_mark_big = num_mark_big,
      num_mark_dec = num_mark_dec,
      num_zero = num_zero,
      num_fmt = num_fmt
    )
    # integerish columns (including whole-number doubles) take a shortcut that
    # ignores `digits`, `num_zero`, and `num_mark_dec` unless num_fmt = "scientific"
  } else {
    out <- format_integer(
      vec,
      digits = digits,
      num_mark_big = num_mark_big,
      num_mark_dec = num_mark_dec,
      num_zero = num_zero,
      num_fmt = num_fmt
    )
  }
  if (is.character(out)) {
    out <- trimws(out)
  }
  return(out)
}

# `digits` significant digits; column-wise when called on a vector, cell-wise
# when vapply()ed element by element. Shared by the "significant",
# "significant_cell", and num_suffix paths.
format_significant <- function(x, digits, num_zero, num_mark_big, num_mark_dec) {
  format(
    x,
    digits = digits,
    drop0trailing = !num_zero,
    big.mark = num_mark_big,
    decimal.mark = num_mark_dec,
    scientific = FALSE
  )
}

format_scientific <- function(vec, digits, num_zero, num_mark_big, num_mark_dec) {
  formatC(
    vec,
    digits = digits,
    format = "e",
    drop0trailing = !num_zero,
    big.mark = num_mark_big,
    decimal.mark = num_mark_dec
  )
}

format_num_suffix <- function(
    x,
    digits,
    num_mark_big,
    num_mark_dec,
    num_zero,
    num_fmt) {
  # tier 0 = no suffix; thresholds compare abs(x) with `>=` so exact powers
  # get the higher tier (1e6 -> "1M", not "1000K") and negatives abbreviate
  tier <- findInterval(abs(x), c(1e3, 1e6, 1e9, 1e12))
  # NA/NaN/Inf get no suffix and pass through format() untouched
  tier[!is.finite(x)] <- 0L
  scaled <- x / 1000^tier
  number <- vapply(
    scaled,
    format_significant,
    character(1),
    digits = digits,
    num_zero = num_zero,
    num_mark_big = num_mark_big,
    num_mark_dec = num_mark_dec
  )
  paste0(number, c("", "K", "M", "B", "T")[tier + 1])
}

# Format non-integer numeric values
format_non_integer_numeric <- function(
    vec,
    digits,
    num_mark_big,
    num_mark_dec,
    num_zero,
    num_fmt) {
  if (num_fmt == "significant") {
    return(format_significant(
      vec,
      digits = digits,
      num_zero = num_zero,
      num_mark_big = num_mark_big,
      num_mark_dec = num_mark_dec
    ))
  } else if (num_fmt == "significant_cell") {
    return(vapply(
      vec,
      format_significant,
      character(1),
      digits = digits,
      num_zero = num_zero,
      num_mark_big = num_mark_big,
      num_mark_dec = num_mark_dec
    ))
  } else if (num_fmt == "decimal") {
    out <- formatC(
      vec,
      digits = digits,
      format = "f",
      drop0trailing = !num_zero,
      big.mark = num_mark_big,
      decimal.mark = num_mark_dec
    )
    # no signed zero: -0.0001 with digits = 2 must print "0.00"/"0", not "-0"
    dec <- gsub("(\\W)", "\\\\\\1", num_mark_dec)
    neg_zero <- grepl(paste0("^-0(", dec, "0*)?$"), out)
    out[neg_zero] <- sub("-", "", out[neg_zero], fixed = TRUE)
    return(out)
  } else if (num_fmt == "scientific") {
    return(format_scientific(
      vec,
      digits = digits,
      num_zero = num_zero,
      num_mark_big = num_mark_big,
      num_mark_dec = num_mark_dec
    ))
  }
  return(vec)
}

# Format integer values
format_integer <- function(
    vec,
    digits,
    num_mark_big,
    num_mark_dec,
    num_zero,
    num_fmt) {
  if (num_fmt == "scientific") {
    return(format_scientific(
      vec,
      digits = digits,
      num_zero = num_zero,
      num_mark_big = num_mark_big,
      num_mark_dec = num_mark_dec
    ))
  } else {
    return(format(vec, big.mark = num_mark_big, scientific = FALSE))
  }
}

#' Striped theme with alternating row colors
#'
#' @param x A `tinytable` object.
#' @param ... Additional arguments (ignored).
#' @return A modified `tinytable` object.
#' @export
theme_striped <- function(x, ...) {
  # For Tabulator, use CSS rules to create striped rows
  x <- theme_html(
    x,
    tabulator_css_rule = "$TINYTABLE_ID .tabulator-row:nth-child(odd) .tabulator-cell { background-color: #ededed !important; }"
  )

  # Stripe rows are computed at build time, after group rows have been
  # inserted by rbind_body_groupi(), so every output format (LaTeX, HTML,
  # Typst, ...) stripes exactly the same rows even when group_tt() adds rows
  # before or after theme_striped() in the pipeline. LaTeX previously used a
  # `row{even}={bg=black!5!white}` tabularray spec, which counted header and
  # group rows by absolute parity and used a slightly different gray; the
  # single explicit row set with #ededed replaces it.
  #
  # The lazy call is appended to @lazy_style directly (instead of calling
  # style_tt() inside a build_prepare() hook) so the stripe styles keep this
  # call's position in the style queue: styles applied after theme_striped()
  # must still override the stripes (see issue #531).
  cal <- as.call(list(
    function(x, ...) {
      style_tt_lazy(x, i = seq(1, nrow(x), by = 2), background = "#ededed")
    },
    x = quote(x)
  ))
  x@lazy_style <- c(x@lazy_style, list(cal))

  return(x)
}

grid_colspan <- function(x) {
    if (!x@output %in% c("markdown", "dataframe")) {
        return(x)
    }

    sty <- prepare_grid_style(x)

    if (nrow(sty) == 0 || !is.data.frame(sty)) {
        return(x)
    }

    # Check if colspan column exists before accessing it
    if (!"colspan" %in% colnames(sty)) {
        return(x)
    }

    # Find rows with colspan > 1
    colspan_rows <- sty[!is.na(sty$colspan) & sty$colspan > 1, ]

    if (nrow(colspan_rows) == 0) {
        return(x)
    }

    # Get the table string and split into lines
    table_lines <- strsplit(x@table_string, "\\n")[[1]]

    # Find the header separator line (+=====+) as reference point
    header_sep_line <- grep("^\\+={2,}", table_lines)
    if (length(header_sep_line) == 0) {
        # No header separator found, try to find any separator line starting with +
        header_sep_line <- which(grepl("^\\+-", table_lines))[1]
        if (is.na(header_sep_line)) {
            return(x) # Can't find reference point
        }
    } else {
        header_sep_line <- header_sep_line[1]
    }

    for (idx in seq_len(nrow(colspan_rows))) {
        row_idx <- colspan_rows[idx, "i"]
        col_idx <- colspan_rows[idx, "j"]
        colspan <- colspan_rows[idx, "colspan"]

        # Calculate target line: header_sep_line + row_idx (rows are consecutive after header separator)
        if (row_idx > 0) {
            target_line <- header_sep_line + row_idx
        } else if (row_idx == 0) {
            target_line <- header_sep_line + row_idx - 1
        } else {
            # group_tt(j) are always spanned
            next
        }

        if (
            target_line <= length(table_lines) &&
                startsWith(table_lines[target_line], "|")
        ) {
            line <- table_lines[target_line]

            # Split the line by | to get cells
            cells <- strsplit(line, "\\|")[[1]]

            # Remove empty first and last elements
            if (length(cells) > 0 && cells[1] == "") {
                cells <- cells[-1]
            }
            if (length(cells) > 0 && cells[length(cells)] == "") {
                cells <- cells[-length(cells)]
            }

            # Remove | markers for the colspan range
            if (
                col_idx <= length(cells) &&
                    (col_idx + colspan - 1) <= length(cells)
            ) {
                # For colspan, use only the content from the first cell (the spanning cell)
                # and ignore content from subsequent cells that are being spanned over
                # Preserve leading space but trim trailing spaces
                spanned_content <- sub(" *$", "", cells[col_idx])

                # Calculate total width for the spanned cells
                if (
                    !is.null(x@width_cols) &&
                        length(x@width_cols) >= (col_idx + colspan - 1)
                ) {
                    total_width <- sum(x@width_cols[
                        col_idx:(col_idx + colspan - 1)
                    ]) +
                        (colspan - 1)
                    current_width <- ansi_nchar(spanned_content)
                    if (current_width < total_width) {
                        padding <- total_width - current_width
                        spanned_content <- paste0(
                            spanned_content,
                            strrep(" ", padding)
                        )
                    }
                }

                # Replace the cells with the spanned content
                new_cells <- cells
                new_cells[col_idx] <- spanned_content
                # Remove the subsequent cells that are part of the colspan
                if (colspan > 1) {
                    indices_to_remove <- (col_idx + 1):(col_idx + colspan - 1)
                    new_cells <- new_cells[-indices_to_remove]
                }

                # Reconstruct the line, preserving the original format
                table_lines[target_line] <- paste0(
                    "|",
                    paste(new_cells, collapse = "|"),
                    "|"
                )
            }
        }
    }

    # Reconstruct the table string
    x@table_string <- paste(table_lines, collapse = "\n")

    return(x)
}

setMethod(
  f = "build_eval",
  signature = "tinytable_html",
  definition = function(x, ...) {
    template <- readLines(
      system.file("templates/html.html", package = "tinytable")
    )

    mathjax <- get_option("tinytable_html_mathjax", default = FALSE)
    assert_flag(mathjax, name = "tinytable_html_mathjax")
    if (isFALSE(mathjax)) {
      template <- paste(template, collapse = "\n")
      sta <- "    <!-- tinytable mathjax start -->"
      end <- "    <!-- tinytable mathjax end -->"
      template <- lines_drop_between(template, sta, end, fixed = TRUE)
      template <- strsplit(template, "\n")[[1]]
    }

    quartoprocessing <- get_option(
      "tinytable_quarto_disable_processing",
      default = TRUE
    )
    assert_flag(quartoprocessing, name = "tinytable_quarto_disable_processing")
    if (isFALSE(quartoprocessing)) {
      template <- sub(
        "data-quarto-disable-processing='true'",
        "data-quarto-disable-processing='false'",
        template,
        fixed = TRUE
      )
    }

    if (identical(x@html_engine, "bootstrap")) {
      template <- sub(
        "$tinytable_BOOTSTRAP_CDN",
        '<link href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.2/dist/css/bootstrap.min.css" rel="stylesheet">',
        template,
        fixed = TRUE)
    } else {
      template <- sub("$tinytable_BOOTSTRAP_CDN", "", template, fixed = TRUE)
    }

    # caption
    if (length(x@caption) != 1) {
      template <- sub(
        "$tinytable_HTML_CAPTION",
        "",
        template,
        fixed = TRUE
      )
    } else {
      template <- sub(
        "$tinytable_HTML_CAPTION",
        sprintf("<caption>%s</caption>", x@caption),
        template,
        fixed = TRUE
      )
    }

    # note
    if (length(x@notes) == 0) {
      template <- sub(
        "$tinytable_HTML_NOTE",
        "",
        template,
        fixed = TRUE
      )
    } else {
      notes_tmp <- NULL
      for (k in seq_along(x@notes)) {
        if (!is.null(names(x@notes))) {
          if (is.list(x@notes[[k]])) {
            tmp <- sprintf(
              "<tr><td colspan='%s'><sup>%s</sup> %s</td></tr>",
              ncol(x),
              names(x@notes)[k],
              x@notes[[k]]$text
            )
            # note is a string
          } else {
            tmp <- sprintf(
              "<tr><td colspan='%s'><sup>%s</sup> %s</td></tr>",
              ncol(x),
              names(x@notes)[k],
              x@notes[k]
            )
          }
        } else {
          tmp <- sprintf(
            "<tr><td colspan='%s'>%s</td></tr>",
            ncol(x),
            x@notes[[k]]
          )
        }
        notes_tmp <- c(notes_tmp, tmp)
      }
      notes <- paste(notes_tmp, collapse = "\n")
      notes <- paste0("<tfoot>", notes, "</tfoot>")
      template <- sub(
        "$tinytable_HTML_NOTE",
        notes,
        template,
        fixed = TRUE
      )
    }

    # width
    if (length(x@width) == 1) {
      template <- sub(
        "width: auto;",
        sprintf(
          "table-layout: fixed; width: %s !important;",
          format_markup_unit(round(x@width * 100), "%")
        ),
        template,
        fixed = TRUE
      )
    } else if (length(x@width) > 1) {
      template <- sub(
        "width: auto;",
        sprintf(
          "table-layout: fixed; width: %s;",
          format_markup_unit(round(sum(x@width) * 100), "%")
        ),
        template,
        fixed = TRUE
      )
    }

    # (pseudo-)unique table IDs
    id <- get_id("")
    x@id <- id

    # Function factory eliminates need for function name manipulation
    template <- gsub(
      "$tinytable_TABLE_ID",
      paste0("tinytable_", id),
      template,
      fixed = TRUE
    )
    template <- gsub(
      "$tinytable_ID",
      id,
      template,
      fixed = TRUE
    )

    # header
    idx <- grep("$tinytable_HTML_HEADER", template, fixed = TRUE)

    if (length(colnames(x)) > 0) {
      # Generate all header cells at once
      col_indices <- seq_along(colnames(x))
      header_cells <- sprintf(
        '    <th scope="col" data-row="0" data-col="%d">%s</th>',
        col_indices,
        colnames(x)
      )
      header <- c("  <tr>", header_cells, "  </tr>")
      header <- paste(strrep(" ", 11), header)
    } else {
      header <- NULL
    }
    template <- c(
      template[1:(idx - 1)],
      header,
      template[(idx + 1):length(template)]
    )
    # body
    body <- NULL

    # Calculate row indices with vectorized operations
    # All rows in table_dataframe should get consecutive indices starting from 1
    # Group rows have already been inserted by rbind_**()
    # The user's row indices (i parameter) should match the HTML data-row values
    i_idx <- seq_len(nrow(x@data_body))

    # Generate all cells at once using matrix operations
    row_indices <- rep(i_idx, each = ncol(x@data_body))
    col_indices <- rep(
      seq_len(ncol(x@data_body)),
      times = nrow(x@data_body)
    )
    cell_values <- as.vector(t(x@data_body))

    # Create all cells in one operation
    cells <- sprintf(
      '    <td data-row="%d" data-col="%d">%s</td>',
      row_indices,
      col_indices,
      cell_values
    )

    # Reshape into rows
    cells_matrix <- matrix(cells, ncol = ncol(x@data_body), byrow = TRUE)
    rows <- apply(cells_matrix, 1, function(row) {
      c("  <tr>", row, "  </tr>")
    })

    body <- unlist(rows)

    idx <- grep("$tinytable_HTML_BODY", template, fixed = TRUE)
    template <- c(
      template[1:(idx - 1)],
      paste(strrep(" ", 13), body),
      template[(idx + 1):length(template)]
    )

    out <- paste(template, collapse = "\n")

    # before style_eval()
    x@table_string <- out


    if (length(x@html_class) == 0) {
      if (
        length(x@theme) == 0 ||
          is.null(x@theme[[1]]) ||
          is.function(x@theme[[1]]) ||
          isTRUE("default" %in% x@theme[[1]])
      ) {
        x <- theme_tinytable(x)
      }
    }

    return(x)
  })

typst_resize_table <- function(table_string, resize_width, resize_height, resize_direction) {
  if (is.null(resize_direction)) {
    return(table_string)
  }

  if (is.null(resize_height)) {
    target_name <- "target-width"
    target_definition <- sprintf("let target-width = size.width * %s", format_markup_num(resize_width))
    body_dimension <- "body-size.width"
  } else {
    target_name <- "target-height"
    target_definition <- sprintf("let target-height = size.height * %s", format_markup_num(resize_height))
    body_dimension <- "body-size.height"
  }

  condition <- switch(
    resize_direction,
    "down" = sprintf("%s > %s", body_dimension, target_name),
    "up" = sprintf("%s < %s", body_dimension, target_name),
    "both" = "true"
  )

  factor <- sprintf("%s / %s * 100%%", target_name, body_dimension)

  paste0(
    "#layout(size => {\n",
    "  let body = [\n",
    table_string,
    "\n  ]\n",
    "  let body-size = measure(body)\n",
    "  ", target_definition, "\n",
    "  if ", condition, " {\n",
    "    let factor = ", factor, "\n",
    "    scale(x: factor, y: factor, reflow: true, body)\n",
    "  } else {\n",
    "    body\n",
    "  }\n",
    "})"
  )
}

#' Typst-specific styles and options
#'
#' @param x A `tinytable` object.
#' @param multipage Logical. When `TRUE`, emits a Typst show rule that allows figures to break across pages. When `FALSE` (default), tables are kept on a single page. When `NULL`, no show rule is emitted.
#' @param figure Logical, whether to wrap the table in a Typst figure environment and block.
#' @param portable Logical. Sets whether to create portable Typst output with base64-encoded local images embedded directly in the Typst code. Remote image URLs are not downloaded.
#' @param align_figure Character string indicating horizontal alignment: "l", "c", or "r".
#'   Defaults to `get_option("tinytable_typst_align_figure", NULL)`. When NULL, no figure-level alignment is emitted.
#' @param resize_width Numeric value between 0.01 and 1.0 specifying the target width
#'   as a fraction of the available Typst layout width when resize_direction is
#'   specified.
#' @param resize_height Numeric value between 0.01 and 1.0 specifying the target
#'   height as a fraction of the available Typst layout height. When specified,
#'   height controls the scaling and width scales proportionally.
#' @param resize_direction Character string specifying how to resize tables. Options
#'   are: "down" to shrink oversized tables, "up" to expand undersized tables,
#'   and "both" to always scale to the target size.
#' @param ... Additional arguments.
#'
#' @export
theme_typst <- function(x,
                        multipage = get_option("tinytable_typst_multipage", default = FALSE),
                        figure = get_option("tinytable_typst_figure", default = TRUE),
                        portable = get_option("tinytable_typst_portable", default = NULL),
                        align_figure = get_option("tinytable_typst_align_figure", NULL),
                        resize_width = get_option("tinytable_typst_resize_width", 1),
                        resize_height = get_option("tinytable_typst_resize_height", default = NULL),
                        resize_direction = get_option("tinytable_typst_resize_direction", default = NULL), ...) {
  assert_flag(multipage, null.ok = TRUE)
  assert_flag(figure)
  assert_flag(portable, null.ok = TRUE)
  assert_choice(align_figure, c("l", "c", "r"), null.ok = TRUE)
  assert_numeric(resize_width, len = 1, lower = 0.01, upper = 1)
  assert_numeric(resize_height, len = 1, lower = 0.01, upper = 1, null.ok = TRUE)
  assert_choice(resize_direction, c("down", "up", "both"), null.ok = TRUE)

  if (!is.null(portable)) {
    x@typst_portable <- portable
  }

  fn <- function(table) {
    tab <- table@table_string
    if (isTRUE(multipage)) {
      tab <- sub("breakable: false", "breakable: true", tab, fixed = TRUE)
    } else if (isFALSE(multipage)) {
      tab <- sub("breakable: true", "breakable: false", tab, fixed = TRUE)
    } else {
      tab <- lines_drop(tab, regex = "show figure.*breakable")
    }
    table@table_string <- tab
    return(table)
  }
  x <- build_finalize(x, fn, output = "typst")

  if (!figure) {
    fn <- function(table) {
      tab <- table@table_string
      # Drop only the figure wrapper. The #block[] preamble must be kept: it
      # defines style-dict/style-array/get-style() and the table.cell show
      # rule, which the table's align/fill closures call. Same marker-based
      # approach as the Quarto handling in typst_finalize.R. When Quarto has
      # already stripped the figure wrapper, the markers are gone and there
      # is nothing left to drop.
      if (grepl("// start preamble figure", tab, fixed = TRUE)) {
        tab <- lines_drop_between(
          tab,
          regex_start = "// start preamble figure",
          regex_end = "// end preamble figure",
          fixed = TRUE
        )
        tab <- lines_drop(tab, regex = "// end figure", fixed = TRUE)
        # finalize() rewrites "#block" as "block" because the block is an
        # argument of #figure(); at the top level it must be a code call.
        tab <- sub("\nblock[ // start block", "\n#block[ // start block", tab, fixed = TRUE)
      }
      table@table_string <- tab
      return(table)
    }
    x <- build_finalize(x, fn, output = "typst")
  }

  # Handle align_figure functionality
  if (!is.null(align_figure)) {
    fn <- function(table) {
      tab <- table@table_string
      figure_align <- switch(
        align_figure,
        "l" = "left",
        "c" = "center",
        "r" = "right"
      )
      tab <- sub(
        "// tinytable align-figure before",
        sprintf("#align(%s, [", figure_align),
        tab,
        fixed = TRUE
      )
      tab <- sub(
        "// tinytable align-figure after",
        "]) // end align",
        tab,
        fixed = TRUE
      )
      table@table_string <- tab
      return(table)
    }
    x <- build_finalize(x, fn, output = "typst")
  }

  if (!is.null(resize_direction)) {
    fn <- function(table) {
      table@table_string <- typst_resize_table(
        table@table_string,
        resize_width = resize_width,
        resize_height = resize_height,
        resize_direction = resize_direction
      )
      return(table)
    }
    x <- build_finalize(x, fn, output = "typst")
  }

  return(x)
}

nse_i_j <- function(x, i_expr, j_expr, pf) {
  if (!inherits(x, "tinytable")) {
    # `pf` is the frame of the user's call (e.g., format_tt() on a plain data
    # frame). Evaluating in parent.frame() here would look up variables in
    # nse_i_j()'s caller instead, where user-defined objects do not exist.
    return(list(i = eval(i_expr, pf), j = eval(j_expr, pf)))
  }

  i <- i_expr
  j <- j_expr

  tmpenv <- new.env()
  # Evaluate NSE expressions against the *original* typed columns in @data.
  # @data_body is character-coerced by tt(), so a predicate like `x > 30`
  # would otherwise compare strings lexicographically ("5" > "30" is TRUE)
  # and silently select the wrong rows. Columns that only exist in the
  # working copy fall back to their @data_body version.
  val <- as.list(x@data_body)
  if (nrow(x@data) == nrow(x@data_body)) {
    for (nm in intersect(names(val), colnames(x@data))) {
      val[[nm]] <- x@data[[nm]]
    }
  }
  val <- c(val, list(groupi = x@group_index_i))
  val <- val[names(val) != ""]
  list2env(val, tmpenv)
  i <- tryCatch(eval(i, pf), error = function(e) eval(i, tmpenv))

  if (is.logical(i) && length(i) == nrow(x@data_body)) {
    i <- which(i)
    # A logical predicate is evaluated against the data rows of @data_body,
    # which does not (yet) include the label rows inserted by group_tt():
    # those are only merged into @data_body by rbind_body_groupi() at build
    # time. Numeric `i`, however, is interpreted in final-table coordinates
    # when the lazy style/format/plot calls are resolved. Mark
    # predicate-derived indices so lazy consumers can remap them through
    # @index_body (see nse_remap_i()).
    attr(i, "tt_nse_data_rows") <- TRUE
  }

  j <- tryCatch(eval(j_expr, pf), error = function(e) NULL)

  # if j is a symbol matching a column in x@data_body
  if (is.null(j)) {
    j <- j_expr
    if (is.symbol(j) && as.character(j) %in% colnames(x@data_body)) {
      j <- as.character(j)

      # if j is a call to c() of column names
    } else if (is.call(j) && identical(j[[1L]], as.name("c"))) {
      syms <- as.list(j[-1L])
      if (all(vapply(syms, function(s) is.symbol(s) && as.character(s) %in% colnames(x@data_body), logical(1)))) {
        j <- vapply(syms, as.character, character(1))
      } else {
        j <- eval(j, pf)
      }

      # otherwise: evaluate normally
    } else {
      j <- eval(j, pf)
    }
  }

  return(list(i = i, j = j))
}


# Remap row indices produced by an NSE predicate in nse_i_j() from data-row
# coordinates (rows of the pre-group @data_body) to final-table coordinates,
# using @index_body. @index_body is computed by build_tt() before any lazy
# call is resolved, so this must only be called at lazy-evaluation time
# (style_tt_lazy(), format_tt_lazy(), plot_tt_lazy()). This is correct
# regardless of whether group_tt() was called before or after the styling
# call: group_tt() never modifies @data_body eagerly, so predicates always
# see data rows only. When the table has no group rows, @index_body is the
# identity mapping and this is a no-op. Unmarked `i` is returned unchanged.
nse_remap_i <- function(i, x) {
  if (isTRUE(attr(i, "tt_nse_data_rows")) && length(x@index_body) > 0) {
    i <- x@index_body[as.vector(i, mode = "integer")]
  }
  return(i)
}

typst_escape_string <- function(x) {
  x <- gsub("\\", "\\\\", x, fixed = TRUE)
  x <- gsub('"', '\\"', x, fixed = TRUE)
  x <- gsub("\n", "\\n", x, fixed = TRUE)
  x
}

typst_image_mime <- function(images) {
  ext <- tolower(tools::file_ext(images))

  if (any(ext == "")) {
    stop("Empty image extensions are not allowed", call. = FALSE)
  }

  ext <- ifelse(ext == "jpg", "jpeg", ext)
  ext <- ifelse(ext == "svg", "svg+xml", ext)
  paste0("image/", ext)
}

encode_typst <- function(images, height, width_plot, height_plot) {
  assert_dependency("base64enc")

  if (any(grepl("^https?://", trimws(images)))) {
    stop("Portable Typst images require local image files.", call. = FALSE)
  }

  width_plot <- format_markup_num(width_plot)
  height_plot <- format_markup_num(height_plot)
  height <- format_markup_num(height)
  mime <- typst_image_mime(images)
  encoded <- sapply(images, base64enc::base64encode)
  encoded <- gsub("\\s+", "", encoded)

  svg <- sprintf(
    "<svg xmlns='http://www.w3.org/2000/svg' width='%s' height='%s' viewBox='0 0 %s %s'><image href='data:%s;base64,%s' width='%s' height='%s' preserveAspectRatio='xMidYMid meet'/></svg>",
    width_plot,
    height_plot,
    width_plot,
    height_plot,
    mime,
    encoded,
    width_plot,
    height_plot
  )
  svg <- typst_escape_string(svg)

  base::sprintf('#image(bytes("%s"), format: "svg", height: %sem)', svg, height)
}

#' Grid theme with borders around all cells
#'
#' @param x A `tinytable` object.
#' @param ... Additional arguments (ignored).
#' @return A modified `tinytable` object.
#' @export
theme_grid <- function(x, ...) {
  # prepare: before table is drawn
  x <- theme_latex(x, inner = "hlines, vlines,")
  x <- theme_html(x, engine = "bootstrap", class = "table table-bordered")

  # finalize: after table is drawn
  fn <- function(x) {
    x@table_string <- sub(
      "stroke: none,",
      "stroke: (paint: black),",
      x@table_string
    )
    return(x)
  }
  x <- build_finalize(x, fn, output = "typst")

  return(x)
}

#' Draw a Tiny Table
#'
#' @description
#' The `tt` function renders a table in different formats with various styling options: HTML, Markdown, LaTeX, Word, PDF, PNG, or Typst. The table can be customized with additional functions:
#'
#' * `style_tt()`: style fonts, colors, alignment, etc.
#' * `format_tt()`: format numbers, dates, strings, etc.
#' * `group_tt()`: row or column group labels.
#' * `save_tt()`: save the table to a file or return the table as a string.
#' * `print()`: print to a specific format, ex: `print(x, "latex")`
#' * `theme_*()` functions apply a collection of format-specific or visual transformations to a `tinytable.`
#'
#' `tinytable` attempts to determine the appropriate way to print the table based on interactive use, RStudio availability, and output format in RMarkdown or Quarto documents. Users can call `print(x, output="markdown")` to print the table in a specific format. Alternatively, they can set a global option: `options("tinytable_print_output"="markdown")`
#'
#' @param x A data frame, data table, or tibble to be rendered as a table.
#' @param digits Number of significant digits to keep for numeric variables. When `digits` is an integer, `tt()` calls `format_tt(x, digits = digits)` before proceeding to draw the table. Note that only the `digits` argument is passed explicitly; all other `format_tt()` arguments use their default values, which can be influenced by global options (e.g., `options(tinytable_format_num_fmt = "decimal")`). When `digits` is `NULL` (the default), `format_tt()` is not called at all, so global `tinytable_format_*` options have no effect unless the user calls `format_tt()` explicitly. Users who need more control can use the `format_tt()` function instead.
#' @param caption A string that will be used as the caption of the table. This argument should *not* be used in Quarto or Rmarkdown documents. In that context, please use the appropriate chunk options.
#' @param width Table or column width.
#' - Single numeric value smaller than or equal to 1 determines the full table width, in proportion of line width.
#' - Numeric vector of length equal to the number of columns in `x` determines the width of each column, in proportion of line width. If the sum of `width` exceeds 1, each element is divided by `sum(width)`. This makes the table full-width with relative column sizes.
#' @param height Row height in em units. Single numeric value greater than zero that determines the row height spacing.
#' @param theme Function or string.
#' - String: `r paste(setdiff(names(theme_dictionary), "default"), collapse = ", ")`
#' - Function: Applied to the `tinytable` object.
#' @param notes Notes to append to the bottom of the table. This argument accepts several different inputs:
#' * Single string insert a single note: `"blah blah"`
#' * Multiple strings insert multiple notes sequentially: `list("Hello world", "Foo bar")`
#' * A named list inserts a list with the name as superscript: `list("a" = list("Hello World"))`
#' * A named list with positions inserts markers as superscripts inside table cells: `list("a" = list(i = 0:1, j = 2, text = "Hello World"))`
#' @param colnames `TRUE`, `FALSE`, or "label". If "label", use the `attr(x$col,"label")` attribute if available and fall back on column names otherwise.
#' @param rownames Logical. If `TRUE`, rownames are included as the first column
#' @param escape Logical. If `TRUE`, escape special characters in the table. Equivalent to `format_tt(tt(x), escape = TRUE)`.
#' @param engine Character string specifying the HTML engine: "tinytable", "bootstrap", or "tabulator". Equivalent to `theme_html(tt(x), engine = engine)`. When `NULL` (the default), the `tinytable_html_engine` global option is honored by `theme_html()`.
#' @param ... Additional arguments are ignored
#' @return An object of class `tt` representing the table.
#'
#' The table object has S4 slots which hold information about the structure of the table. For example, the `table@group_index_i` slot includes the row indices for grouping labels added by `group_tt()`.
#'
#' Warning: Relying on or modifying the contents of these slots is strongly discouraged. Their names and contents could change at any time, and the `tinytable` developers do not consider changes to the internal structure of the output object to be a "breaking  change" for versioning or changelog purposes.
#'
#' @template dependencies
#' @template latex_preamble
#' @template limitations_word_markdown
#' @template tabulator
#' @template global_options
#' @template order_of_operations
#'
#' @examples
#' library(tinytable)
#' x <- mtcars[1:4, 1:5]
#'
#' tt(x)
#'
#' tt(x,
#'   theme = "striped",
#'   width = 0.5,
#'   caption = "Data about cars."
#' )
#'
#' tt(x, notes = "Hello World!")
#'
#' fn <- list(i = 0:1, j = 2, text = "Hello World!")
#' tab <- tt(x, notes = list("*" = fn))
#' print(tab, "latex")
#'
#' k <- data.frame(x = c(0.000123456789, 12.4356789))
#' tt(k, digits = 2)
#'
#' # use variable labels stored in attributes as column names
#' dat = mtcars[1:5, c("cyl", "mpg", "hp")]
#' attr(dat$cyl, "label") <- "Cylinders"
#' attr(dat$mpg, "label") <- "Miles per Gallon"
#' attr(dat$hp, "label") <- "Horse Power"
#' tt(dat, colnames = "label")
#'
#' @export
# Generic function
tt <- function(x, ...) {
  UseMethod("tt")
}

#' @rdname tt
#' @export

# Default method (for data.frame, data.table, tbl_df, and other data frame-like objects)
tt.default <- function(
    x,
    digits = get_option("tinytable_tt_digits", default = NULL),
    caption = get_option("tinytable_tt_caption", default = NULL),
    notes = get_option("tinytable_tt_notes", default = NULL),
    width = get_option("tinytable_tt_width", default = NULL),
    height = get_option("tinytable_tt_height", default = NULL),
    theme = get_option("tinytable_tt_theme", default = "default"),
    colnames = get_option("tinytable_tt_colnames", default = TRUE),
    rownames = get_option("tinytable_tt_rownames", default = FALSE),
    escape = get_option("tinytable_tt_escape", default = FALSE),
    engine = NULL,
    ...) {
  # sanity checks
  assert_string(caption, null.ok = TRUE)
  assert_integerish(digits, len = 1, null.ok = TRUE)
  notes <- sanitize_notes(notes)

  if (identical(theme, "void")) {
    theme <- "empty"
    warning("`theme = 'void'` is deprecated. Please use `theme = 'empty'` instead.", call. = FALSE)
  }
  if (!isTRUE(check_function(theme)) && !isTRUE(check_string(theme))) {
    stop("The `theme` argument must be a function or a string.", call. = FALSE)
  }
  if (isTRUE(check_string(theme))) {
    assert_choice(theme, names(theme_dictionary), null.ok = TRUE)
  }

  # x should be a data frame, not a tibble or slopes, for indexing convenience
  assert_data_frame(x, min_rows = 1, min_cols = 1)
  if (!isTRUE(identical(class(x), "data.frame"))) {
    cn <- colnames(x)
    # weird bug on as.data.frame.data.table() when there is no columns. in general, we don't want to modify in-place.
    if (is.null(cn) && inherits(x, "data.table")) {
      assert_dependency("data.table")
      data.table::setDF(x)
    } else {
      x <- as.data.frame(x, check.names = FALSE)
      colnames(x) <- cn
    }
  }

  if (!(is.logical(colnames) && length(colnames) == 1) && !identical(colnames, "label")) {
    stop("The `colnames` argument must be TRUE, FALSE, or 'label'.", call. = FALSE)
  }

  # factors should all be characters (for replace, etc.)
  # it might be dangerous to leave non-numerics, but what about dates and other character-coercibles?
  for (i in seq_along(x)) {
    if (is.factor(x[[i]])) {
      x[[i]] <- as.character(x[[i]])
    }
  }

  assert_numeric(width, lower = 0, null.ok = TRUE)
  if (!length(width) %in% c(0, 1, ncol(x))) {
    msg <- sprintf("The `width` argument must have length 1 or %s.", ncol(x))
    stop(msg, call. = FALSE)
  }
  if (sum(width) > 1) {
    # handle remainders gracefully to avoid sum > 100% after rounding
    width <- percent_sum_100(width) / 100
  }

  assert_numeric(height, lower = 0, len = 1, null.ok = TRUE)

  # bind the row names if the user explicitly asks for it in global option.
  # Same name as tibble::rownames_to_column()
  assert_flag(rownames)
  if (isTRUE(rownames) && !is.null(row.names(x))) {
    nocol <- is.null(colnames(x))
    rn <- data.frame(format(row.names(x)))
    rn <- stats::setNames(rn, "rowname")
    x <- cbind(rn, x)
    if (isTRUE(nocol)) {
      colnames(x) <- NULL
    }
  }

  # formatting options are limited here
  # after creating the table since the new lazy system
  tab <- x
  if (!is.null(digits)) {
    tab <- format_tt(tab, digits = digits)
  }

  # baseline character format
  # twice because format() leaves Date type, which cannot be partially reasigned
  # with indexed format_tt(i)
  tab <- data.frame(lapply(tab, format))
  tab <- data.frame(lapply(tab, trimws))
  colnames(tab) <- colnames(x)

  out <- methods::new(
    "tinytable",
    data = x,
    data_body = tab,
    caption = caption,
    notes = notes,
    theme = list(theme),
    width = width,
    height = height,
    colnames = colnames
  )

  if (is.null(theme)) {
    out <- theme_tinytable(out)
  } else if (is.function(theme)) {
    out <- theme(out)
  } else if (is.character(theme)) {
    for (th in theme) {
      if (!th %in% names(theme_dictionary)) {
        stop(sprintf("Unknown theme: '%s'. Available themes: %s", th, paste(names(theme_dictionary), collapse = ", ")), call. = FALSE)
      }
    }
    out <- theme_dictionary[[theme]](out)
  }

  # explicit engine choice delegates to theme_html(), which validates the
  # value and sets @html_engine (plus the bootstrap class swap)
  if (!is.null(engine)) {
    out <- theme_html(out, engine = engine)
  }

  if (isTRUE(escape)) {
    out <- format_tt(out, escape = TRUE)
  }

  if (!is.null(height)) {
    # LaTeX tabularray: use height/2 as rowsep
    out <- theme_latex(
      out,
      inner = sprintf("rowsep={%s}", format_markup_unit(height / 2, "em"))
    )

    # HTML: use CSS padding for row height
    out <- theme_html(
      out,
      css = sprintf(
        "padding-top: %s; padding-bottom: %s;",
        format_markup_unit(height / 2, "em"),
        format_markup_unit(height / 2, "em")
      )
    )

    # Typst: use rows parameter in table settings
    fun_typst_height <- function(table) {
      if (!is.null(table@height)) {
        table@table_string <- gsub(
          "rows: auto,",
          sprintf("rows: %s,", format_markup_unit(table@height, "em")),
          table@table_string
        )
      }
      return(table)
    }
    out <- style_tt(out, finalize = fun_typst_height)
  }

  # HTML: add column width styles if multiple widths specified
  if (!is.null(width) && length(width) > 1) {
    for (j in seq_len(ncol(x))) {
      css <- sprintf("width: %s;", format_markup_unit(width[j] / sum(width) * 100, "%"))
      out <- theme_html(out, j = j, css = css)
    }
  }

  return(out)
}

#' @export
tt.data.frame <- function(x, ...) {
  tt.default(x, ...)
}

#' @export
tt.data.table <- function(x, ...) {
  tt.default(x, ...)
}

#' @export
tt.tbl_df <- function(x, ...) {
  tt.default(x, ...)
}

format_vector_replace <- function(
  vec,
  vec_original = NULL,
  replace = NULL,
  ...
) {
  if (is.null(replace) || isFALSE(replace) || length(replace) == 0) {
    return(vec)
  }

  result <- vec

  # match against the original (pre-formatting) values only: a cell that
  # merely *formats* to the same string as a replacement target (e.g. 1.4
  # printed as "1" with digits = 0) must not be clobbered
  ori <- if (is.null(vec_original)) vec else vec_original
  ori_chr <- as.character(ori)
  ori_numeric <- is.numeric(ori)

  for (z in seq_along(replace)) {
    new <- names(replace)[z]

    for (old in replace[[z]]) {
      if (is.nan(old)) {
        match_idx <- is.nan(ori)
      } else if (is.na(old)) {
        match_idx <- is.na(ori)
      } else if (ori_numeric && is.numeric(old) && is.infinite(old)) {
        match_idx <- is.infinite(ori) & sign(ori) == sign(old)
      } else {
        match_idx <- !is.na(ori_chr) & ori_chr == as.character(old)
      }

      result[match_idx] <- new
    }
  }

  return(result)
}

format_vector_escape <- function(vec, output = "latex", ...) {
  # Early returns for edge cases
  if (length(vec) < 1 || all(is.na(vec)) || !is.character(vec)) {
    return(vec)
  }
  if (isFALSE(output)) {
    return(vec)
  }

  # Define escape patterns for each output format
  escape_patterns <- list(
    latex = list(
      # LaTeX escaping code adapted from the `gt` package, published under MIT
      # https://github.com/rstudio/gt/
      # YEAR: 2018-2024
      # COPYRIGHT HOLDER: gt authors
      chars = c(
        "\\" = "\\textbackslash{}",
        "~" = "\\textasciitilde{}",
        "^" = "\\textasciicircum{}",
        "&" = "\\&",
        "%" = "\\%",
        "$" = "\\$",
        "#" = "\\#",
        "_" = "\\_",
        "{" = "\\{",
        "}" = "\\}"
      ),
      pattern = "[\\\\&%$#_{}~^]"
    ),
    html = list(
      chars = c(
        "&" = "&amp;",
        "<" = "&lt;",
        ">" = "&gt;"
      ),
      pattern = "[&<>]"
    ),
    typst = list(
      chars = c(
        "\\" = "\\\\",
        "<" = "\\<",
        ">" = "\\>",
        "*" = "\\*",
        "_" = "\\_",
        "@" = "\\@",
        "=" = "\\=",
        "-" = "\\-",
        "+" = "\\+",
        "/" = "\\/",
        "$" = "\\$",
        "#" = "\\#",
        "[" = "\\[",
        "]" = "\\]",
        "`" = "\\`",
        "~" = "\\~",
        "\"" = "\\\""
      ),
      pattern = "[\\\\<>*_@=+/\\$#\\[\\]\\-`~\"]"
    )
  )

  out <- vec

  if (isTRUE(output == "latex")) {
    out <- apply_escape_pattern(out, escape_patterns$latex)
  } else if (isTRUE(output %in% c("html", "bootstrap", "tabulator"))) {
    out <- apply_escape_pattern(out, escape_patterns$html)
  } else if (isTRUE(output == "typst")) {
    out <- apply_escape_pattern(out, escape_patterns$typst)
  }

  return(out)
}



apply_escape_pattern <- function(vec, pattern_info) {
  na_out <- is.na(vec)
  if (all(na_out)) {
    return(vec)
  }

  vec_clean <- vec[!na_out]

  # Short circuit if no special characters found (performance optimization)
  if (!any(grepl(pattern_info$pattern, vec_clean, perl = TRUE))) {
    return(vec)
  }

  m <- gregexpr(pattern_info$pattern, vec_clean, perl = TRUE)
  special_chars <- regmatches(vec_clean, m)

  # Safer lookup
  escaped_chars <- lapply(special_chars, function(x) {
    vapply(x, function(ch) pattern_info$chars[[ch]], character(1))
  })

  regmatches(vec_clean, m) <- escaped_chars
  vec[!na_out] <- vec_clean

  return(vec)
}

handle_latex_environment <- function(x, environment) {
  if (!is.null(environment)) {
    fn <- function(table) {
      table_string <- table@table_string

      assert_choice(environment, c("tblr", "talltblr", "longtblr", "tabular"))

      # Switch the environment
      table_string <- gsub("\\{tblr\\}(\\[.*?\\])?", sprintf("{%s}\\1", environment), table_string)
      table_string <- gsub("\\{talltblr\\}(\\[.*?\\])?", sprintf("{%s}\\1", environment), table_string)
      table_string <- gsub("\\{longtblr\\}(\\[.*?\\])?", sprintf("{%s}\\1", environment), table_string)

      # Special handling for tabular environment
      if (environment == "tabular") {
        # insert a duplicate \begin{tabular} without the tblr baggage
        table_string <- lines_insert(table_string,
          new = "\\begin{tabular}",
          regex = "\\\\begin\\{tabular\\}",
          position = "before")
        # tabularray options
        table_string <- lines_drop_between(
          table_string,
          regex_start = "tabularray outer open",
          regex_end = "tabularray inner close"
        )

        # reinsert caption for plain tabular tables
        table_env <- "\\\\begin\\{table\\}"
        if (length(table@caption) > 0 && grepl(table_env, table_string, perl = TRUE)) {
          cap <- sprintf("\\caption{%s}", table@caption[1])
          table_string <- lines_insert(
            table_string,
            new = cap,
            regex = table_env,
            position = "after",
            perl = TRUE
          )
        }

        # Convert tabularray syntax to tabular
        table_string <- gsub("cmidrule\\[(.*?)\\]", "cmidrule(\\1)", table_string)
        table_string <- gsub("\\\\toprule|\\\\midrule|\\\\bottomrule", "\\\\hline", table_string)
        table_string <- sub("\\s*%% tabularray outer open", "", table_string)
        table_string <- sub("\\s*%% TinyTableHeader", "", table_string)
      }

      # Set proper column alignment for tabular
      if (environment == "tabular" && !is.null(ncol(table))) {
        a <- sprintf("begin{tabular}{%s}", strrep("l", ncol(table)))
        table_string <- sub("begin{tabular}", a, table_string, fixed = TRUE)
      }

      table@table_string <- table_string
      return(table)
    }
    x <- build_finalize(x, fn, output = "latex")
  }
  return(x)
}

handle_latex_environment_table <- function(x, environment_table) {
  if (!environment_table) {
    fn <- function(table) {
      # Remove table wrapper elements if requested
      table_string <- table@table_string
      regex <- "\\\\begin\\s*\\{(?:tabular|(?:long|tall)?tblr)\\}"
      table_string <- lines_drop(table_string, regex = regex, position = "before", perl = TRUE)
      table_string <- lines_drop(table_string, regex = "\\\\end\\{table\\}", position = "after")
      table_string <- lines_drop(table_string, regex = "\\\\end\\{table\\}", position = "equal")
      table@table_string <- table_string
      return(table)
    }
    x <- build_finalize(x, fn, output = "latex")
  }
  return(x)
}


#' LaTeX-Specific Theme for `tinytable`
#'
#' This function provides comprehensive LaTeX-specific theming and configuration options
#' for `tinytable` objects. It allows customization of LaTeX environments, table layout,
#' multipage behavior, resizing, and placement within LaTeX documents.
#'
#' @param x A `tinytable` object to apply LaTeX theming to.
#' @param inner Character string specifying inner tabularray options. These options
#'   control the internal formatting of the table (e.g., column alignment, spacing).
#'   Will be added to any existing inner options. Default is `NULL`.
#' @param outer Character string specifying outer tabularray options. These options
#'   control the external formatting around the table. Will be added to any existing
#'   outer options. Default is `NULL`.
#' @param environment Character string specifying the LaTeX table environment to use.
#'   Options are:
#'   - `"tblr"` - Standard tabularray table (default)
#'   - `"talltblr"` - Tall tabularray table for tables that may break across pages
#'   - `"longtblr"` - Long tabularray table for multi-page tables
#'   - `"tabular"` - Basic LaTeX tabular environment without tabularray features
#'
#'   Default is controlled by `tinytable_latex_environment` option.
#' @param environment_table Logical indicating whether to wrap the table in a `table`
#'   environment. When `FALSE`, only the core table structure is output without the
#'   surrounding table wrapper. Automatically set to `FALSE` when `environment = "longtblr"`.
#'   Default is controlled by `tinytable_latex_environment_table` option.
#' @param multipage Logical indicating whether to enable multipage table functionality.
#'   When `TRUE`, automatically switches to `longtblr` environment and sets appropriate
#'   options for tables that span multiple pages. Default is controlled by
#'   `tinytable_latex_multipage` option.
#' @param rowhead Integer specifying the number of header rows to repeat on each page
#'   in multipage tables. Only valid with `longtblr` environment. Default is controlled
#'   by `tinytable_latex_rowhead` option.
#' @param rowfoot Integer specifying the number of footer rows to repeat on each page
#'   in multipage tables. Only valid with `longtblr` environment. Default is controlled
#'   by `tinytable_latex_rowfoot` option.
#' @param resize_width Numeric value between 0.01 and 1.0 specifying the target width
#'   as a fraction of `\\linewidth` when resizing tables. Only used when `resize_direction`
#'   is specified. Default is controlled by `tinytable_latex_resize_width` option.
#' @param resize_height Numeric value between 0.01 and 1.0 specifying the target height
#'   as a fraction of `\\textheight` when resizing tables. When specified (and `resize_direction`
#'   is set), the table height is controlled instead of its width. If both `resize_width`
#'   and `resize_height` are specified, `resize_height` takes precedence (width scales
#'   proportionally). Default is controlled by `tinytable_latex_resize_height` option.
#' @param resize_direction Character string specifying how to resize tables that are
#'   too wide or too narrow. Options are:
#'   - `"down"` - Only shrink tables wider than `\\linewidth` (or taller than `\\textheight` when `resize_height` is set)
#'   - `"up"` - Only expand tables narrower than `\\linewidth` (or shorter than `\\textheight` when `resize_height` is set)
#'   - `"both"` - Resize all tables to exactly `resize_width * \\linewidth` (or `resize_height * \\textheight`)
#'
#'   Default is controlled by `tinytable_latex_resize_direction` option.
#' @param placement Character string specifying LaTeX float placement options for the
#'   table environment (e.g., "h", "t", "b", "p", "H"). Only used when `environment_table = TRUE`.
#'   Default is controlled by `tinytable_latex_placement` option.
#' @param preamble Logical value specifying whether to include LaTeX preamble packages.
#'   If not NULL, overrides the table's preamble setting.
#' @param engine Character string specifying the LaTeX engine to use for PDF compilation.
#'   Options are "xelatex", "pdflatex", or "lualatex". If not NULL, overrides the table's engine setting.
#' @param ... Additional arguments (currently unused).
#'
#' @return A modified `tinytable` object with LaTeX-specific theming applied.
#'
#' @details
#' The function provides fine-grained control over LaTeX table output through several mechanisms:
#'
#' **Environment Selection:**
#' Different LaTeX environments offer different capabilities:
#' - `tblr`: Modern tabularray syntax with full styling support
#' - `talltblr`: Like `tblr` but optimized for tall tables
#' - `longtblr`: Supports page breaks and repeated headers/footers
#' - `tabular`: Basic LaTeX syntax, limited styling but maximum compatibility
#'
#' **Multipage Tables:**
#' When `multipage = TRUE` or when `rowhead`/`rowfoot` are specified, the function
#' automatically switches to `longtblr` environment and disables the table wrapper.
#' This allows tables to break across pages while maintaining headers and footers.
#'
#' **Resizing:**
#' The resize functionality uses LaTeX's `\\resizebox` command to automatically
#' adjust table dimensions based on content and page constraints. Use `resize_width`
#' to control width (fraction of `\\linewidth`) or `resize_height` to control height
#' (fraction of `\\textheight`). When `resize_height` is specified, the table is scaled
#' to fit the target height with width adjusting proportionally.
#'
#' **Tabularray Options:**
#' Inner and outer options directly control tabularray formatting. Inner options
#' affect cell content and spacing, while outer options control the table's
#' relationship with surrounding text.
#'
#' @seealso
#' [tt()], [style_tt()], [save_tt()]
#'
#' @export
theme_latex <- function(x,
                        inner = NULL,
                        outer = NULL,
                        environment = get_option("tinytable_latex_environment", default = NULL),
                        environment_table = get_option("tinytable_latex_environment_table", default = TRUE),
                        multipage = get_option("tinytable_latex_multipage", default = FALSE),
                        rowhead = get_option("tinytable_latex_rowhead", 0L),
                        rowfoot = get_option("tinytable_latex_rowfoot", 0L),
                        resize_width = get_option("tinytable_latex_resize_width", 1),
                        resize_height = get_option("tinytable_latex_resize_height", default = NULL),
                        resize_direction = get_option("tinytable_latex_resize_direction", default = NULL),
                        placement = get_option("tinytable_latex_placement", NULL),
                        preamble = NULL,
                        engine = NULL,
                        ...) {
  assert_flag(environment_table)
  assert_string(inner, null.ok = TRUE)
  assert_string(outer, null.ok = TRUE)
  assert_integerish(rowhead, lower = 0, len = 1)
  assert_integerish(rowfoot, lower = 0, len = 1)
  assert_choice(environment, c("tblr", "talltblr", "longtblr", "tabular"), null.ok = TRUE)
  assert_numeric(resize_width, len = 1, lower = 0.01, upper = 1)
  assert_numeric(resize_height, len = 1, lower = 0.01, upper = 1, null.ok = TRUE)
  assert_choice(resize_direction, c("down", "up", "both"), null.ok = TRUE)
  assert_string(placement, null.ok = TRUE)
  assert_flag(preamble, null.ok = TRUE)
  assert_choice(engine, c("xelatex", "pdflatex", "lualatex"), null.ok = TRUE)

  if (!is.null(preamble)) {
    x@latex_preamble <- preamble
  }

  if (!is.null(engine)) {
    x@latex_engine <- engine
  }

  # Set environment_table = FALSE when environment is longtblr
  if (!is.null(environment) && environment == "longtblr") {
    environment_table <- FALSE
  }

  if (!is.null(inner)) x@tabularray_inner <- c(x@tabularray_inner, inner)
  if (!is.null(outer)) x@tabularray_outer <- c(x@tabularray_outer, outer)

  # environment switch
  if (!is.null(environment) && environment != "longtblr" && (rowhead >= 1 || rowfoot >= 1)) {
    stop("When using multipage functionality (rowhead or rowfoot >= 1), the environment must be 'longtblr'.", call. = FALSE)
  }

  # rowhead/rowfoot imply a multipage (longtblr) table
  if (rowhead >= 1 || rowfoot >= 1) {
    multipage <- TRUE
  }

  # a longtblr can span multiple pages, so it cannot be resized as a single box
  if (!is.null(resize_direction) && (isTRUE(multipage) || identical(environment, "longtblr"))) {
    stop("`resize_direction` cannot be combined with multipage tables (`multipage = TRUE`, `rowhead`/`rowfoot` >= 1, or `environment = \"longtblr\"`), because a `longtblr` can span multiple pages and cannot be resized as a single box.", call. = FALSE)
  }

  # Handle environment using separate helper function
  x <- handle_latex_environment(x, environment)

  # multipage
  if (isTRUE(multipage)) {
    if (rowhead > 0) {
      x@tabularray_inner <- c(x@tabularray_inner, sprintf("rowhead=%s", rowhead))
    }
    if (rowfoot > 0) {
      x@tabularray_inner <- c(x@tabularray_inner, sprintf("rowfoot=%s", rowfoot))
    }
    # Apply both environment and environment_table changes for multipage
    x <- handle_latex_environment(x, "longtblr")
    environment_table <- FALSE
  }

  # Handle environment_table using separate helper function
  x <- handle_latex_environment_table(x, environment_table)

  # Handle resize functionality
  if (!is.null(resize_direction)) {
    fn <- function(table) {
      tab <- table@table_string

      if (!is.null(resize_height)) {
        # Height-based resizing: scale height, width proportional
        if (resize_direction == "both") {
          new <- sprintf("\\resizebox{!}{%s\\textheight}{", format_markup_num(resize_height))
        } else if (resize_direction == "down") {
          new <- sprintf(
            "\\resizebox{!}{\\ifdim\\totalheight>\\textheight %s\\textheight\\else\\totalheight\\fi}{",
            format_markup_num(resize_height)
          )
        } else if (resize_direction == "up") {
          new <- sprintf(
            "\\resizebox{!}{\\ifdim\\totalheight<\\textheight %s\\textheight\\else\\totalheight\\fi}{",
            format_markup_num(resize_height)
          )
        }
      } else {
        # Width-based resizing: scale width, height proportional
        if (resize_direction == "both") {
          new <- sprintf("\\resizebox{%s\\linewidth}{!}{", format_markup_num(resize_width))
        } else if (resize_direction == "down") {
          new <- sprintf(
            "\\resizebox{\\ifdim\\width>\\linewidth %s\\linewidth\\else\\width\\fi}{!}{",
            format_markup_num(resize_width)
          )
        } else if (resize_direction == "up") {
          new <- sprintf(
            "\\resizebox{\\ifdim\\width<\\linewidth %s\\linewidth\\else\\width\\fi}{!}{",
            format_markup_num(resize_width)
          )
        }
      }

      # `longtblr` is excluded: resize + multipage raises an error above
      reg <- "\\\\begin\\{(tblr|talltblr|tabular)\\}"
      tab <- lines_insert(tab, regex = reg, new = new, position = "before")

      new <- "}"
      reg <- "\\\\end\\{(tblr|talltblr|tabular)\\}"
      tab <- lines_insert(tab, regex = reg, new = new, position = "after")

      table@table_string <- tab
      return(table)
    }
    x <- build_finalize(x, fn, output = "latex")
  }

  # Handle placement functionality
  if (!is.null(placement)) {
    fn <- function(table) {
      tab <- table@table_string
      tab <- sub(
        "\\\\begin\\{table\\}([^\\[])",
        sprintf("\\\\begin{table}[%s]\\1", placement),
        tab
      )
      table@table_string <- tab
      return(table)
    }
    x <- build_finalize(x, fn, output = "latex")
  }

  return(x)
}

#' Insert images and inline plots into tinytable objects
#'
#' The `plot_tt()` function allows for the insertion of images and inline plots into
#' tinytable objects. This function can handle both local and web-based images.
#'
#' @param x A tinytable object.
#' @param i Integer vector, the row indices where images are to be inserted. If `NULL`,
#'    images will be inserted in all rows.
#' @param j Integer vector, the column indices where images are to be inserted. If `NULL`,
#'    images will be inserted in all columns.
#' @param height Numeric, the height of the images in the table in em units.
#' @param height_plot Numeric, the height of generated plot images in pixels (default: 400).
#' @param width_plot Numeric, the width of generated plot images in pixels (default: 1200).
#' @param color string Name of color to use for inline plots (passed to the `col` argument base `graphics` plots in `R`). For bar plots in static output formats (PNG, PDF, etc.), can be a vector of length 2: c(bar_color, background_color) to show progress against a maximum. Note: Tabulator format only uses the first color.
#' @param xlim Numeric vector of length 2. Controls the range of bar plots.
#' @param fun  String or function to generate inline plots.
#' - Built-in plot types (strings):
#'   - `"histogram"`: Creates histograms from numeric vectors. Accepts `color` argument.
#'   - `"density"`: Creates density plots from numeric vectors. Accepts `color` argument.
#'   - `"bar"`: Creates horizontal bar charts from single numeric values. Accepts `color` (single value, or length-2 vector for bar and background colors in static formats) and `xlim` arguments.
#'   - `"line"`: Creates line plots from data frames with `x` and `y` columns. Accepts `color` and `xlim` arguments.
#' - Custom functions:
#'   - Functions that return `ggplot2` objects.
#'   - Functions that return another function which generates a base `R` plot, ex: `function(x) {function() hist(x)}`
#'   - Note: When using custom ggplot2 functions that return plots with text elements, the text size will normally need to be adjusted because the plot is inserted as a very small image in the table. Text sizes of 1 or smaller often work well (e.g., `theme(text = element_text(size = 1))`).
#' - See the tutorial on the `tinytable` website for more information.
#' @param data a list of data frames or vectors to be used by the plotting functions in `fun`.
#' @param images Character vector, the paths to the images to be inserted. Paths are relative to the main table file or Quarto (Rmarkdown) document.
#' @param sprintf Character string, a sprintf format string to format the generated cell content. Default is "%s" which displays the content as-is. Use this to wrap images or plots in custom markup.
#' @param assets Path to the directory where generated assets are stored. This path is relative to the location where a table is saved.
#' @param output Apply plot insertion only when the table is rendered in the specified format. A character vector of one or more of "latex", "html", "typst", or "markdown". If `NULL` (default), plots are inserted regardless of the output format.
#' @param ... Extra arguments are passed to the function in `fun`. Important: Custom plotting functions must always have `...` as an argument.
#'
#' @return A modified tinytable object with images or plots inserted.
#'
#' @details The `plot_tt()` can insert images and inline plots into tables.
#'
#' @examples
#' \dontrun{
#' # Bar plots with single and dual colors
#' dat <- data.frame(
#'   Metric = c("Sales", "Conversion", "Growth", "Efficiency"),
#'   Value = c(75, 45, 92, 38),
#'   Percentage = c(0.75, 0.45, 0.92, 0.38)
#' )
#'
#' tt(dat) |>
#'   plot_tt(j = 2, fun = "bar", data = as.list(dat$Value), color = "darkorange") |>
#'   plot_tt(j = 3, fun = "bar", data = as.list(dat$Percentage),
#'           color = c("steelblue", "lightgrey"), xlim = c(0, 1))
#'
#' # Built-in plot types
#' plot_data <- list(mtcars$mpg, mtcars$hp, mtcars$qsec)
#'
#' dat <- data.frame(
#'   Variables = c("mpg", "hp", "qsec"),
#'   Histogram = "",
#'   Density = "",
#'   Line = ""
#' )
#'
#' # Random data for sparklines
#' lines <- lapply(1:3, \(x) data.frame(x = 1:10, y = rnorm(10)))
#'
#' tt(dat) |>
#'   plot_tt(j = 2, fun = "histogram", data = plot_data) |>
#'   plot_tt(j = 3, fun = "density", data = plot_data, color = "darkgreen") |>
#'   plot_tt(j = 4, fun = "line", data = lines, color = "blue") |>
#'   style_tt(j = 2:4, align = "c")
#'
#' # Custom function example (must have ... argument)
#' custom_hist <- function(d, ...) {
#'   function() hist(d, axes = FALSE, ann = FALSE, col = "lightblue")
#' }
#'
#' tt(data.frame(Variables = "mpg", Histogram = "")) |>
#'   plot_tt(j = 2, fun = custom_hist, data = list(mtcars$mpg))
#' }
#'
#' @export
plot_tt <- function(
    x,
    i = NULL,
    j = NULL,
    fun = NULL,
    data = NULL,
    color = "black",
    xlim = NULL,
    height = 1,
    height_plot = 400,
    width_plot = 1200,
    images = NULL,
    sprintf = "%s",
    assets = "tinytable_assets",
    output = NULL,
    ...) {
  assert_subset(output, c("latex", "html", "typst", "markdown"), null.ok = TRUE)

  # non-standard evaluation before anything else
  tmp <- nse_i_j(x, i_expr = substitute(i), j_expr = substitute(j), pf = parent.frame())
  list2env(tmp, environment())

  jval <- sanitize_j(j, x)
  ival <- sanitize_i(i, x, calling_function = "plot_tt")
  assert_numeric(height, len = 1, lower = 0)
  assert_class(x, "tinytable")

  # Calculate actual length considering NULL i values
  ival_length <- if (isTRUE(attr(ival, "null"))) {
    length(attr(ival, "body"))
  } else {
    length(ival)
  }
  len <- ival_length * length(jval)

  assert_list(data, len = len, null.ok = TRUE)
  assert_character(images, len = len, null.ok = TRUE)

  if (!is.null(fun) && !is.null(images)) {
    stop("`fun` and `images` cannot be used together.", call. = FALSE)
  }

  if (!is.null(fun) && is.null(data)) {
    stop("Please specify `data`.", call. = FALSE)
  }

  if (is.character(fun)) {
    assert_choice(fun, c("histogram", "density", "bar", "barpct", "line"))
  } else {
    assert_function(fun, null.ok = TRUE)
  }

  # built-in plots
  if (identical(fun, "histogram")) {
    fun <- rep(list(tiny_histogram), length(data))
  } else if (identical(fun, "density")) {
    fun <- rep(list(tiny_density), length(data))
  } else if (identical(fun, "line")) {
    fun <- rep(list(tiny_line), length(data))
  } else if (identical(fun, "bar")) {
    for (idx in seq_along(data)) {
      assert_numeric(data[[idx]], len = 1, name = base::sprintf("data[[%d]]", idx))
    }
    if (is.null(xlim)) {
      xlim <- c(0, max(unlist(data)))
    }
    fun <- rep(list(tiny_bar), length(data))
  } else if (identical(fun, "barpct")) {
    for (idx in seq_along(data)) {
      assert_numeric(data[[idx]], len = 1, name = base::sprintf("data[[%d]]", idx))
      if (!all(data[[idx]] >= 0 & data[[idx]] <= 1, na.rm = TRUE)) {
        stop("Data for 'barpct' must be between 0 and 1 (percentages).", call. = FALSE)
      }
    }
    # Always set xlim to c(0, 1) for barpct
    xlim <- c(0, 1)
    # Default to lightgrey background if color is single value
    if (length(color) == 1) {
      color <- c(color, "lightgrey")
    }
    fun <- rep(list(tiny_bar), length(data))
  } else {
    fun <- rep(list(fun), length(data))
  }

  # needed when rendering in tempdir()
  cal <- list(
    "plot_tt_lazy",
    x = x,
    i = ival,
    j = jval,
    data = data,
    fun = fun,
    color = color,
    xlim = xlim,
    height = height,
    height_plot = height_plot,
    width_plot = width_plot,
    images = images,
    sprintf = sprintf,
    assets = assets
  )
  cal <- c(cal, list(...))
  cal <- do.call(call, cal)
  attr(cal, "output") <- output

  x@lazy_plot <- c(x@lazy_plot, list(cal))

  return(x)
}

plot_tt_lazy <- function(
    x,
    i = NULL,
    j = NULL,
    height = 1,
    height_plot = 400,
    width_plot = 1200,
    fun = NULL,
    color = NULL,
    data = NULL,
    xlim = NULL,
    images = NULL,
    sprintf = "%s",
    assets = "tinytable_assets",
    ...) {
  # Row indices derived from an NSE predicate were computed eagerly in
  # plot_tt() against the pre-group data rows (the marker set by nse_i_j()
  # survives sanitize_i()). The @lazy_plot loop runs after
  # rbind_body_groupi(), so @data_body below is the final table and numeric
  # `i` must be in final-table coordinates: remap through @index_body.
  i <- nse_remap_i(i, x)

  out <- x@data_body

  # Handle Tabulator plots with JavaScript formatters
  # Note: images use the standard HTML path below, which works for tabulator too
  is_tabulator <- isTRUE(x@output == "html" && x@html_engine == "tabulator")
  if (is_tabulator && !is.null(data)) {
    result <- plot_tt_tabulator(
      x, i = i, j = j, fun = fun, data = data,
      color = color, xlim = xlim, ...
    )
    # If plot_tt_tabulator returns NULL, it means we should fall back to PNG rendering
    # (e.g., for custom functions). Otherwise, return the result.
    if (!is.null(result)) {
      return(result)
    }
    # Fall through to PNG rendering below
  }

  is_html <- isTRUE(x@output %in% c("html", "bootstrap", "tabulator"))
  is_typst_portable <- isTRUE(x@output == "typst" && x@typst_portable)
  is_quarto <- isTRUE(check_dependency("knitr")) && !is.null(knitr::pandoc_to())

  # paths are tricky in Quarto HTML (website vs single file)
  is_portable <- is_html && (isTRUE(x@html_portable) || is_quarto)
  if (is_portable || is_typst_portable) assert_dependency("base64enc")

  # Normalize user-provided image paths to full paths
  if (!is.null(images)) {
    # quarto requires relative links or url
    # print("html") must be run from a tempdir on linux, so we need absolute paths
    # LaTeX requires relative paths with forward slashes
    if (isTRUE(x@output == "latex")) {
      # For LaTeX: convert backslashes to forward slashes for cross-platform compatibility
      images <- gsub("\\\\", "/", images)
    } else if (!is_quarto) {
      # URLs are not file paths: normalizePath() would prepend the working
      # directory and flip the slashes on Windows.
      url <- grepl("^http", trimws(images))
      images[!url] <- normalizePath(images[!url], mustWork = FALSE)
    }
  }

  if (!is.null(data)) {
    assert_dependency("ggplot2")
    images <- NULL

    # path_assets directory stores dynamically generated plots
    if (is_portable || is_typst_portable) {
      path_assets <- tempdir()
      # quarto and LaTeX require relative paths from the project folder
    } else if (is_quarto || isTRUE(x@output == "latex")) {
      path_assets <- assets
    } else {
      path_assets <- file.path(x@output_dir, assets)
    }
    if (!dir.exists(path_assets)) {
      dir.create(path_assets, recursive = TRUE, showWarnings = FALSE)
      if (!dir.exists(path_assets)) {
        stop(
          base::sprintf("Unable to create the assets directory: %s", path_assets),
          call. = FALSE
        )
      }
    }

    # Rank hack: prepend zero-padded rank to filename to allow sorting based on
    # file names in interactive tables like tabulator
    last_values <- sapply(data, plot_data_rank)
    ranks <- rank(last_values, ties.method = "first")
    n_digits <- nchar(as.character(length(data)))
    zero_padded_ranks <- sprintf(paste0("%0", n_digits, "d"), ranks)

    for (idx in seq_along(data)) {
      fn <- paste0("tinytable_", zero_padded_ranks[idx], "_", get_id(), ".png")
      fn_full <- file.path(path_assets, fn)

      # For LaTeX and Typst output: use relative paths with forward slashes (portable across all platforms)
      # For other outputs: use normalized absolute paths
      if (isTRUE(x@output == "latex" | x@output == "typst")) {
        # Keep relative path and convert backslashes to forward slashes for LaTeX
        # fn_full is used for saving the file (OS-native path)
        # images[idx] is used in \includegraphics (needs forward slashes)
        images[idx] <- gsub("\\\\", "/", fn_full)
      } else {
        # For HTML/other formats, normalize to absolute paths for proper file access
        fn_full <- normalizePath(fn_full, mustWork = FALSE)
        images[idx] <- fn_full
      }

      plot_fun <- fun[[idx]]
      if (!"..." %in% names(formals(plot_fun))) {
        stop(
          "Inline plotting function must have `...` as argument. See tutorial on the `tinytable` website for examples.",
          call. = FALSE
        )
      }

      p <- plot_fun(data[[idx]], xlim = xlim, color = color, ...)

      # ggplot2
      if (inherits(p, "ggplot")) {
        assert_dependency("ggplot2")
        suppressMessages(
          ggplot2::ggsave(
            p,
            filename = fn_full,
            width = width_plot,
            height = height_plot,
            units = "px"
          )
        )

        # base R
      } else if (is.function(p)) {
        render_base_plot(p, fn_full, width_plot, height_plot)

        # sanity check
      } else {
        msg <- "The functions in the `fun` list must return a function or a `ggplot2` object. See the tutorial online for examples: https://vincentarelbundock.github.io/tinytable"
        stop(msg, call. = FALSE)
      }
    }
  }

  if (isTRUE(x@output == "latex")) {
    cell <- "\\includegraphics[height=%sem]{%s}"
    cell <- base::sprintf(cell, format_markup_num(height), images)
  } else if (is_portable) {
    http <- grepl("^http", trimws(images))
    images[!http] <- encode(images[!http])
    cell <- base::sprintf('<img src="%s" style="height: %sem;">', images, format_markup_num(height))
  } else if (is_html) {
    # Convert relative paths to absolute paths for save_tt/print
    http <- grepl("^http", trimws(images))
    for (img_idx in seq_along(images)) {
      if (!http[img_idx]) {
        # Convert relative paths to absolute paths
        if (!grepl("^/", trimws(images[img_idx])) && !grepl("^[A-Za-z]:", trimws(images[img_idx]))) {
          images[img_idx] <- file.path(x@output_dir, images[img_idx])
        }
      }
    }
    cell <- base::sprintf('<img src="%s" style="height: %sem;">', images, format_markup_num(height))
  } else if (isTRUE(x@output == "markdown")) {
    cell <- "![](%s){ height=%s }"
    cell <- base::sprintf(cell, images, format_markup_num(height * 16))
  } else if (isTRUE(x@output == "typst")) {
    if (is_typst_portable) {
      cell <- encode_typst(images, height, width_plot, height_plot)
    } else {
      cell <- '#image("%s", height: %sem)'
      cell <- base::sprintf(cell, images, format_markup_num(height))
    }
  } else if (isTRUE(x@output == "dataframe")) {
    cell <- "%s"
    cell <- base::sprintf(cell, images)
  } else {
    stop("here be dragons")
  }

  cell <- base::sprintf(sprintf, cell)

  # Handle column header insertions (i=0)
  if (0 %in% i) {
    # Insert into header (column names)
    if (is.null(x@names) || length(x@names) == 0) {
      stop("Cannot insert images into header: table has no column names.", call. = FALSE)
    }

    # Fill column-major, matching the body-only path (`out[i, j] <- cell`):
    # the same `data`/`images` vector maps to the same cells whether or not
    # `i` includes the header row (0).
    cell_idx <- 1
    for (j_val in j) {
      for (i_val in i) {
        if (i_val == 0) {
          if (j_val <= length(x@names)) {
            x@names[j_val] <- cell[cell_idx]
          }
        } else {
          out[i_val, j_val] <- cell[cell_idx]
        }
        cell_idx <- cell_idx + 1
      }
    }
  } else {
    # Original behavior: insert into data body
    out[resolve_i_body(i), j] <- cell
  }

  x@data_body <- out

  # Mark columns with HTML content for HTML formatter in Tabulator
  # For custom functions with PNG images, also add rank fields for sorting
  if (isTRUE(x@html_engine == "tabulator")) {
    i_body <- resolve_i_body(i)

    for (col_idx in j) {
      col_name <- x@names[col_idx]
      if (!is.null(col_name) && !(col_name %in% names(x@tabulator_column_formatters))) {
        x@tabulator_column_formatters[[col_name]] <- list(formatter = "html")

        # Add rank fields for sorting (custom functions use PNG, but still need sorting)
        if (!is.null(data)) {
          x <- tabulator_add_rank_fields(x, col_name, col_idx, j, i_body, data)
        }
      }
    }
  }

  return(x)
}

#' Resolve body row indices from a sanitized `i`
#'
#' `sanitize_i()` returns `NA` with a "null" attribute when `i` was `NULL`;
#' in that case the body rows are stored in the "body" attribute.
#' @keywords internal
#' @noRd
resolve_i_body <- function(i) {
  if (all(is.na(i)) && isTRUE(attr(i, "null"))) {
    attr(i, "body")
  } else {
    i
  }
}

#' Write hidden rank fields and sorter configuration for a Tabulator plot column
#'
#' The `data` list is ordered column-major over the `i_body` x `j` grid;
#' `plot_data_rank()` extracts a sortable scalar per cell. The rank values are
#' stored in a hidden `rank_<column>` field (no leading underscore: R converts
#' it to `X_rank_`) used by the custom `tinytable_rank_sorter`.
#' @keywords internal
#' @noRd
tabulator_add_rank_fields <- function(x, col_name, col_idx, j, i_body, data) {
  rank_col_name <- paste0("rank_", col_name)

  # Get data for this column
  col_data_idx <- 1
  if (length(j) > 1) {
    col_data_idx <- which(j == col_idx)
  }

  for (row_idx in seq_along(i_body)) {
    data_idx <- (col_data_idx - 1) * length(i_body) + row_idx
    sort_value <- plot_data_rank(data[[data_idx]])
    x@data_body[i_body[row_idx], rank_col_name] <- sort_value
  }

  # Configure sorter to use rank field
  x@tabulator_column_formatters[[col_name]]$sorter <- "tinytable_rank_sorter"
  x@tabulator_column_formatters[[col_name]]$sorterParams <- list(rankField = rank_col_name)

  return(x)
}

#' Render a base R plot function to a PNG file
#'
#' The device is closed via `on.exit()` so an error in the user's plot
#' function does not leak an open graphics device.
#' @keywords internal
#' @noRd
render_base_plot <- function(p, path, width, height) {
  grDevices::png(path, width = width, height = height)
  on.exit(grDevices::dev.off(), add = TRUE)
  graphics::par(mar = c(0, 0, 0, 0))
  p()
  invisible(NULL)
}

plot_data_rank <- function(x) {
  if (is.list(x) || is.data.frame(x)) {
    if (is.data.frame(x) && "y" %in% names(x)) {
      utils::tail(x$y, n = 1)
    } else if (is.list(x)) {
      utils::tail(unlist(x), n = 1)
    } else {
      utils::tail(x, n = 1)
    }
  } else {
    utils::tail(x, n = 1)
  }
}

tiny_histogram <- function(d, color = "black", ...) {
  function() graphics::hist(d, col = color, axes = FALSE, ann = FALSE)
}

tiny_density <- function(d, color = "black", ...) {
  function() {
    d <- stats::density(stats::na.omit(d))
    graphics::plot(d, axes = FALSE, ann = FALSE, col = color)
    graphics::polygon(d, col = color)
  }
}

tiny_bar <- function(d, color = "black", xlim = 0:1, ...) {
  function() {
    if (length(color) == 2) {
      # Two colors: stacked bar with background
      bar_col <- standardize_colors(color[1])
      bg_col <- standardize_colors(color[2])

      # Calculate the remaining portion based on xlim
      max_val <- xlim[2]
      comp <- max_val - d
      mat <- rbind(d, comp)

      graphics::barplot(
        mat,
        horiz  = TRUE,
        col    = c(bar_col, bg_col),
        xlim   = xlim,
        space  = 0,
        beside = FALSE,
        axes   = FALSE,
        ...
      )
    } else {
      # Single color: simple bar without background
      graphics::barplot(d, horiz = TRUE, col = color, xlim = xlim)
    }
  }
}

tiny_line <- function(d, xlim = 0:1, color = "black", ...) {
  function() {
    if (
      !inherits(d, "data.frame") || !"x" %in% names(d) || !"y" %in% names(d)
    ) {
      stop(
        "The data to plot a `line` must be a data frame with columns `x` and `y`.",
        call. = FALSE
      )
    }
    plot(d$x, d$y, type = "l", col = color, axes = FALSE, ann = FALSE, lwd = 50)
  }
}

encode <- function(images) {
  assert_dependency("base64enc")
  ext <- tools::file_ext(images)

  if (any(ext == "")) {
    stop("Empty image extensions are not allowed", call. = FALSE)
  }

  encoded <- sapply(images, base64enc::base64encode)
  base::sprintf("data:image/%s;base64, %s", ext, encoded)
}


#' Handle plot_tt for Tabulator tables
#' @keywords internal
#' @noRd
plot_tt_tabulator <- function(
    x,
    i = NULL,
    j = NULL,
    fun = NULL,
    data = NULL,
    color = "black",
    xlim = NULL,
    ...) {

  # Determine plot type from fun
  plot_type <- NULL
  if (is.list(fun) && length(fun) > 0) {
    # Extract plot type from function name
    fun_obj <- fun[[1]]
    if (identical(fun_obj, tiny_histogram)) {
      plot_type <- "histogram"
    } else if (identical(fun_obj, tiny_density)) {
      plot_type <- "density"
    } else if (identical(fun_obj, tiny_bar)) {
      plot_type <- "bar"
    } else if (identical(fun_obj, tiny_line)) {
      plot_type <- "line"
    }
  }

  if (is.null(plot_type)) {
    # For custom functions, we cannot use JavaScript formatters
    # Signal to continue with standard PNG rendering by returning NULL
    # This will cause plot_tt_lazy to skip the tabulator path and use PNG
    return(NULL)
  }

  # Handle the case where i is NA (from sanitize_i when i was NULL)
  i_body <- resolve_i_body(i)

  # Track which custom JS has been marked as needed
  needs_histogram <- FALSE
  needs_sparkline <- FALSE

  # Process each column
  for (col_idx in j) {
    col_name <- x@names[col_idx]

    # Get data for this column
    col_data_idx <- 1
    if (length(j) > 1) {
      col_data_idx <- which(j == col_idx)
    }

    for (row_idx in seq_along(i_body)) {
      data_idx <- (col_data_idx - 1) * length(i_body) + row_idx
      plot_data <- data[[data_idx]]

      # Create formatter configuration
      formatter_info <- tabulator_plot_formatter(
        plot_type = plot_type,
        data = plot_data,
        color = color,
        xlim = xlim
      )

      # Store the formatted data in the cell
      if (plot_type %in% c("line", "density", "histogram")) {
        # For sparkline and histogram, store as JSON array string
        # (non-finite values are not valid JS literals -> null)
        vals <- ifelse(is.finite(formatter_info$data), formatter_info$data, "null")
        json_array <- paste0("[", paste(vals, collapse = ","), "]")
        x@data_body[i_body[row_idx], col_idx] <- json_array
      } else {
        # For progress/bar, store the numeric value
        x@data_body[i_body[row_idx], col_idx] <- formatter_info$data
      }

      # Store the formatter configuration in tabulator_column_formatters (once per column)
      if (is.null(x@tabulator_column_formatters[[col_name]])) {
        x@tabulator_column_formatters[[col_name]] <- formatter_info$config
      }

      # Track which custom JS is needed (once per plot type)
      if (isTRUE(formatter_info$requires_custom_js)) {
        if (plot_type == "histogram") {
          needs_histogram <- TRUE
        } else if (plot_type %in% c("line", "density")) {
          needs_sparkline <- TRUE
        }
      }
    }

    # Hidden rank fields + sorter configuration for this column
    x <- tabulator_add_rank_fields(x, col_name, col_idx, j, i_body, data)
  }

  # Add markers for needed custom JS (once total)
  if (needs_histogram && !grepl("NEEDS_HISTOGRAM_JS", x@tabulator_options, fixed = TRUE)) {
    x@tabulator_options <- paste0(x@tabulator_options, "\n// NEEDS_HISTOGRAM_JS")
  }
  if (needs_sparkline && !grepl("NEEDS_SPARKLINE_JS", x@tabulator_options, fixed = TRUE)) {
    x@tabulator_options <- paste0(x@tabulator_options, "\n// NEEDS_SPARKLINE_JS")
  }

  return(x)
}


#' Wrap inline plot images so `alignv` works in LaTeX
#'
#' In tabularray, `\includegraphics` sits on the row baseline and defines the
#' row height, so `valign=` on the neighboring cells is a no-op (#673). To
#' honor `alignv`, the image itself must be lowered with `\raisebox`. The
#' effective `alignv` for an image cell is the one styled on that cell, or
#' failing that, on any other cell of the same row.
#'
#' Called in `build_tt()` after `@style_other` is resolved but before
#' `build_eval()` draws the body.
#'
#' @keywords internal
#' @noRd
plot_alignv_latex <- function(x) {
  if (!isTRUE(x@output == "latex")) {
    return(x)
  }
  so <- x@style_other
  if (is.null(so) || nrow(so) == 0L || !"alignv" %in% names(so)) {
    return(x)
  }
  so <- so[!is.na(so$alignv) & so$i >= 1, , drop = FALSE]
  if (nrow(so) == 0L) {
    return(x)
  }
  # baseline shift: middle of image (m) or top of image at text ascender (t)
  lift <- c(
    m = "-0.5\\height",
    t = "\\dimexpr-\\height+\\ht\\strutbox\\relax"
  )
  for (col in seq_len(ncol(x@data_body))) {
    cells <- x@data_body[[col]]
    for (row in which(startsWith(cells, "\\includegraphics"))) {
      a <- so$alignv[so$i == row & so$j == col] # cell-level style wins
      if (length(a) == 0L) a <- so$alignv[so$i == row] # else row-level
      if (length(a) == 0L || !a[1] %in% names(lift)) next # b: keep as-is
      cells[row] <- base::sprintf("\\raisebox{%s}{%s}", lift[[a[1]]], cells[row])
    }
    x@data_body[[col]] <- cells
  }
  return(x)
}

apply_caption <- function(x, format_fn, source = "original", ...) {
  if (!inherits(x, "tinytable")) {
    return(x)
  }
  x@caption <- format_fn(x@caption, ...)
  return(x)
}

apply_notes <- function(x, format_fn, ...) {
  if (!inherits(x, "tinytable")) {
    return(x)
  }

  for (idx in seq_along(x@notes)) {
    n <- x@notes[[idx]]
    if (is.character(n) && length(n) == 1) {
      x@notes[[idx]] <- format_fn(n, ...)
    } else if (is.list(n) && "text" %in% names(n)) {
      n$text <- format_fn(n$text, ...)
      x@notes[[idx]] <- n
    }
  }
  return(x)
}

update_group_data_cells <- function(data_slot, rows, cols, format_fn, ...) {
  for (row in rows) {
    for (col in cols) {
      current_value <- data_slot[row, col]
      if (!is.na(current_value) && trimws(current_value) != "") {
        formatted_value <- format_fn(current_value, ...)
        if (!is.null(formatted_value)) {
          data_slot[row, col] <- formatted_value
        }
      }
    }
  }
  return(data_slot)
}

apply_group_j <- function(x, i, j, format_fn, components, ...) {
  if (!inherits(x, "tinytable")) {
    return(x)
  }

  if (!"groupj" %in% components && !any(i < 0, na.rm = TRUE)) {
    return(x)
  }

  data_slot <- x@group_data_j

  if ("groupj" %in% components) {
    data_slot <- update_group_data_cells(
      data_slot,
      rows = seq_len(nrow(data_slot)),
      cols = seq_len(ncol(data_slot)),
      format_fn = format_fn,
      ...
    )
  } else {
    i_idx <- i[i < 0 & !is.na(i)]
    i_idx <- i_idx - min(i_idx) + 1
    data_slot <- update_group_data_cells(
      data_slot,
      rows = i_idx,
      cols = j,
      format_fn = format_fn,
      ...
    )
  }

  x@group_data_j <- data_slot
  return(x)
}

apply_colnames <- function(x, i, j, format_fn, components, ...) {
  if ("colnames" %in% components) {
    colnames(x) <- format_fn(colnames(x), ...)
  } else if (0 %in% i) {
    formatted <- format_fn(colnames(x)[j], ...)
    if (length(formatted) == length(j)) {
      colnames(x)[j] <- formatted
    }
  }
  return(x)
}

# Global dispatcher function for applying formatting functions
apply_format <- function(
  x,
  i,
  j,
  format_fn,
  components = NULL,
  inherit_class = NULL,
  original_data = TRUE,
  original = NULL,
  ...
) {
  if (is.character(components)) {
    if ("all" %in% components) {
      components <- c(
        "colnames",
        "caption",
        "notes",
        "~groupi",
        "groupi",
        "groupj",
        "cells"
      )
    }
  }

  if (identical(components, "groupi")) {
    i <- x@group_index_i
  }

  if (inherits(x, "tinytable")) {
    out <- x@data_body
    ori <- x@data
  } else {
    out <- x
    # `original` preserves the raw input for data frames / vectors, which do
    # not carry their pre-formatting data the way tinytable objects do
    ori <- if (is.null(original)) x else original
  }

  # Apply formatting to specified components only
  x <- apply_colnames(x, i, j, format_fn, components = components, ...)
  x <- apply_group_j(x, i, j, format_fn, components = components, ...)

  if ("caption" %in% components) {
    x <- apply_caption(x, format_fn, ...)
  }
  if ("notes" %in% components) {
    x <- apply_notes(x, format_fn, ...)
  }

  if (
    is.null(components) || any(c("cells", "groupi", "~groupi") %in% components)
  ) {
    # Apply to specific cells
    # Filter columns based on inherits argument and original data
    j_filtered <- j

    if (inherits(x, "tinytable")) {
      classref <- x@data
    } else {
      classref <- out
    }

    if (length(j) == 0) {
      j_filtered <- integer(0)
    } else if (is.character(inherit_class)) {
      j_filtered <- j[sapply(j, function(col) {
        inherits(classref[[col]], inherit_class)
      })]
    } else if (is.function(inherit_class)) {
      j_filtered <- j[sapply(j, function(col) inherit_class(classref[[col]]))]
    }

    i <- sort(unique(i))

    # group i
    if (inherits(x, "tinytable") && length(x@group_index_i) > 0) {
      idx_group <- x@group_index_i %in% i
      for (col in j_filtered) {
        tmp <- tryCatch(
          format_fn(x@group_data_i[idx_group, col], ...),
          error = function(e) NULL
        )
        if (length(tmp) > 0) {
          x@group_data_i[idx_group, col] <- tmp
        }
      }
    }

    # body
    # index: we are only formatting the body rows
    # ori & out currently have the same dimensions
    if (inherits(x, "tinytable")) {
      idx <- x@index_body %in% i
    } else {
      idx <- seq_len(nrow(out)) %in% i
    }
    for (col in j_filtered) {
      vec_original <- ori[idx, col, drop = TRUE]
      vec <- if (original_data) vec_original else out[idx, col, drop = TRUE]
      formatted <- tryCatch(
        format_fn(
          vec = vec,
          vec_original = vec_original,
          ...
        ),
        error = function(e) NULL
      )
      if (length(formatted) > 0) {
        out[idx, col] <- formatted
      }
    }

    if (inherits(x, "tinytable")) {
      x@data_body <- out
    } else {
      x <- out
    }
  }

  return(x)
}

#' Clean column names for JavaScript field identifiers
#'
#' Sanitizes every character outside `[A-Za-z0-9_]` to `_` and de-duplicates
#' the results with `make.unique()`. Always call this on the full vector of
#' column names so that de-duplication is consistent across call sites.
#'
#' @param name Character vector of column names to clean
#' @return Cleaned column names
#' @keywords internal
#' @noRd
tabulator_clean_column_name <- function(name) {
    make.unique(gsub("[^A-Za-z0-9_]", "_", name), sep = "_")
}


#' Clean data for Tabulator
#' @param x tinytable object
#' @return Cleaned data frame
#' @keywords internal
#' @noRd
tabulator_clean_data <- function(x) {
    # Process data based on column types
    data_clean <- list()

    for (i in seq_along(x@data)) {
        col_name <- names(x@data)[i]
        original_col <- x@data[[i]]
        col_type <- class(original_col)[1]

        if (col_type %in% TAB_NUM) {
            # Use raw data for numeric columns (Tabulator formatters will handle display)
            data_clean[[col_name]] <- original_col
        } else if (col_type == "Date") {
            # Convert Date to ISO string for Tabulator datetime parsing
            data_clean[[col_name]] <- format(original_col, "%Y-%m-%d")
        } else if (col_type %in% c("POSIXct", "POSIXlt")) {
            # Convert POSIXct/POSIXlt to ISO string for Tabulator datetime parsing
            data_clean[[col_name]] <- format(original_col, "%Y-%m-%dT%H:%M:%S")
        } else if (col_type == "logical") {
            # For logical columns, use formatted data if bool formatting is applied
            if (isTRUE(x@tabulator_format_bool)) {
                if (nrow(x@data_body) == nrow(x@data)) {
                    data_clean[[col_name]] <- as.character(x@data_body[[i]])
                } else {
                    # Handle case where rows were modified (groups, etc.)
                    data_clean[[col_name]] <- as.character(original_col)
                }
            } else {
                # Use raw logical data for Tabulator's default boolean handling
                data_clean[[col_name]] <- original_col
            }
        } else {
            # Use formatted data for other column types (character, factor, etc.)
            if (nrow(x@data_body) == nrow(x@data)) {
                data_clean[[col_name]] <- as.character(x@data_body[[i]])
            } else {
                # Handle case where rows were modified (groups, etc.)
                # For now, fall back to original data as character
                data_clean[[col_name]] <- as.character(original_col)
            }
        }
    }

    # Add any additional columns from data_body that aren't in data (e.g., rank columns)
    if (nrow(x@data_body) == nrow(x@data)) {
        body_cols <- names(x@data_body)
        data_cols <- names(x@data)
        extra_cols <- setdiff(body_cols, data_cols)
        for (col_name in extra_cols) {
            data_clean[[col_name]] <- x@data_body[[col_name]]
        }
    }

    # Convert to data frame
    data_clean <- as.data.frame(
        data_clean,
        stringsAsFactors = FALSE,
        check.names = FALSE
    )

    # Add stable row index for styling (1-based, matching R's row numbering)
    data_clean$`_tinytable_row_index` <- seq_len(nrow(data_clean))

    # Mark plot_tt() cells holding pre-serialized JSON arrays (sparkline,
    # histogram) as raw JSON so df_to_json() emits them unquoted
    raw_formatters <- c("tinytable_sparkline", "tinytable_histogram")
    for (col_name in names(x@tabulator_column_formatters)) {
        fmt <- x@tabulator_column_formatters[[col_name]]
        if (
            !is.null(fmt$formatter) &&
                fmt$formatter %in% raw_formatters &&
                col_name %in% names(data_clean)
        ) {
            data_clean[[col_name]] <- I(lapply(
                data_clean[[col_name]],
                function(v) {
                    if (is.character(v) && grepl("^\\[.*\\]$", v)) {
                        json_raw(v)
                    } else {
                        v
                    }
                }
            ))
        }
    }

    # Clean column names (sanitize to [A-Za-z0-9_] and de-duplicate)
    names(data_clean) <- tabulator_clean_column_name(names(data_clean))

    return(data_clean)
}

# =============================================================================
# UTILITY FUNCTIONS
# =============================================================================

#' Create a grid line with specified character
#' @keywords internal
#' @noRd
grid_create_line <- function(width_cols, char = "-") {
  paste0("+", paste(strrep(char, width_cols), collapse = "+"), "+")
}


# =============================================================================
# DATA EXTRACTION FUNCTIONS
# =============================================================================

#' Extract colspan information from table styling
#' @keywords internal
#' @noRd
extract_colspan_info <- function(x) {
  if (!inherits(x, "tinytable") || nrow(x@style) == 0) {
    return(NULL)
  }

  sty <- x@style
  colspan_info <- sty[
    !is.na(sty$colspan) & sty$colspan > 1,
    c("i", "j", "colspan")
  ]

  if (nrow(colspan_info) == 0) {
    return(NULL)
  }

  return(colspan_info)
}

#' Check if a cell is part of a colspan
#' @keywords internal
#' @noRd
is_colspan_cell <- function(i, j, colspan_info, header) {
  if (is.null(colspan_info)) {
    return(FALSE)
  }

  data_row_idx <- if (header) i - 1 else i
  if (data_row_idx <= 0) {
    return(FALSE)
  }

  any(colspan_info$i == data_row_idx & colspan_info$j == j)
}

# =============================================================================
# WIDTH CALCULATION FUNCTIONS
# =============================================================================

#' Adjust column widths for colspan cells
#' @keywords internal
#' @noRd
adjust_colspan_widths <- function(tab, width_cols, colspan_info, header) {
  if (is.null(colspan_info)) {
    return(width_cols)
  }

  for (idx in seq_len(nrow(colspan_info))) {
    i_row <- colspan_info[idx, "i"]
    j_col <- colspan_info[idx, "j"]
    colspan <- colspan_info[idx, "colspan"]

    # Get the actual row index in the tab matrix
    tab_row_idx <- if (header) i_row + 1 else i_row

    # Skip if row index is invalid (negative indices are used for group headers)
    if (tab_row_idx <= 0 || tab_row_idx > nrow(tab) || j_col <= 0 || j_col > ncol(tab)) {
      next
    }
      # Get content width of the colspan cell
      cell_content <- tab[tab_row_idx, j_col]
      content_width <- ansi_nchar(cell_content)

      # Calculate current total width of spanned columns
      spanned_cols <- j_col:(j_col + colspan - 1)
      spanned_cols <- spanned_cols[spanned_cols <= length(width_cols)]
      current_total_width <- sum(width_cols[spanned_cols]) +
        length(spanned_cols) -
        1

      # Expand columns if content is longer than current width
      if (content_width > current_total_width) {
        extra_width <- content_width - current_total_width
        width_per_col <- extra_width / length(spanned_cols)

        for (col in spanned_cols) {
          width_cols[col] <- width_cols[col] + ceiling(width_per_col)
        }
      }
  }

  return(width_cols)
}

#' Adjust column widths for group headers
#' @keywords internal
#' @noRd
adjust_group_widths <- function(x, width_cols) {
  # Handle column groups using @group_data_j
  if (nrow(x@group_data_j) > 0) {
    # Process each header row to extract column spans
    for (row_idx in seq_len(nrow(x@group_data_j))) {
      group_row <- as.character(x@group_data_j[row_idx, ])
      spans <- parse_group_spans(group_row)

      # Adjust widths for each span
      for (span_idx in seq_len(nrow(spans))) {
        group_cols <- spans$start[span_idx]:spans$end[span_idx]
        g_len <- ansi_nchar(spans$label[span_idx]) + 2
        # Space available to the spanning cell includes absorbed separators
        c_len <- sum(width_cols[group_cols]) + length(group_cols) - 1

        if (g_len > c_len) {
          width_cols[group_cols] <- width_cols[group_cols] +
            ceiling((g_len - c_len) / length(group_cols))
        }
      }
    }
  }

  # Handle row groups (they span entire table width)
  if (nrow(x@group_data_i) > 0) {
    # Check labels in first column of group data
    if (ncol(x@group_data_i) >= 1) {
      labels <- x@group_data_i[, 1]
      for (label in labels) {
        if (!is.na(label) && ansi_nchar(label) > 0) {
          g_len <- ansi_nchar(label) + 2
          # Total table width including separators
          c_len <- sum(width_cols) + length(width_cols) - 1
          if (g_len > c_len) {
            # Distribute extra width across all columns
            extra_width <- ceiling((g_len - c_len) / length(width_cols))
            width_cols <- width_cols + extra_width
          }
        }
      }
    }
  }

  return(width_cols)
}

#' Calculate all column widths including colspan and group adjustments
#' @keywords internal
#' @noRd
calculate_column_widths <- function(tab, x, header, width_cols) {
  colspan_info <- extract_colspan_info(x)

  # Mark colspan cells once instead of rescanning colspan_info for every cell.
  # Rows in colspan_info use body indices; tab includes the header when present.
  excluded <- matrix(FALSE, nrow = nrow(tab), ncol = ncol(tab))
  if (!is.null(colspan_info)) {
    tab_i <- colspan_info$i + as.integer(header)
    tab_j <- colspan_info$j
    valid <- !is.na(tab_i) & !is.na(tab_j) &
      tab_i > 0L & tab_i <= nrow(tab) & tab_j > 0L & tab_j <= ncol(tab)
    if (any(valid)) {
      excluded[cbind(tab_i[valid], tab_j[valid])] <- TRUE
    }
  }

  # Calculate basic widths excluding colspan cells. ansi_nchar() is vectorized,
  # so this avoids growing cell_widths in the former nested row loop.
  for (j in seq_len(ncol(tab))) {
    keep <- !excluded[, j]
    if (any(keep)) {
      width_cols[j] <- max(ansi_nchar(tab[keep, j]))
    } else {
      width_cols[j] <- max(ansi_nchar(tab[, j]))
    }
  }

  # Adjust for colspan cells
  width_cols <- adjust_colspan_widths(tab, width_cols, colspan_info, header)

  # Adjust for group headers
  if (inherits(x, "tinytable")) {
    width_cols <- adjust_group_widths(x, width_cols)
  }

  return(width_cols)
}


# =============================================================================
# TABLE BUILDING FUNCTIONS
# =============================================================================

#' Pad table cells to match column widths
#' @keywords internal
#' @noRd
pad_table_cells <- function(tab, width_cols) {
  for (j in seq_len(ncol(tab))) {
    nc <- ansi_nchar(tab[, j])
    pad <- strrep(" ", pmax.int(0L, width_cols[j] - nc))
    tab[, j] <- paste0(tab[, j], pad)
  }
  return(tab)
}

#' Format table rows with proper separators
#' @keywords internal
#' @noRd
format_table_rows <- function(tab, x, header, width_cols) {
  body <- character(nrow(tab))

  # Get vertical line character - use "|" if enabled, space if disabled
  vline_char <- if (inherits(x, "tinytable") && isTRUE(x@grid_vline)) {
    "|"
  } else if (inherits(x, "tinytable") && !isTRUE(x@grid_vline)) {
    " " # Use space if vertical lines are disabled
  } else {
    "|" # Default fallback
  }

  # Use vertical line character for borders
  left_border <- vline_char
  right_border <- vline_char

  # Get group row indices if available
  group_rows <- if (inherits(x, "tinytable")) x@group_index_i else integer(0)
  if (header) {
    group_rows <- group_rows + 1 # Header adds one row at the top
  }

  spanning_rows <- logical(nrow(tab))
  candidates <- intersect(group_rows, seq_len(nrow(tab)))
  for (row_idx in candidates) {
    row_data <- tab[row_idx, ]
    spanning_rows[row_idx] <-
      ansi_nchar(trimws(row_data[1])) > 0 &&
      all(trimws(row_data[-1]) == "")
  }

  for (row_idx in seq_len(nrow(tab))) {
    row_data <- tab[row_idx, ]

    if (spanning_rows[row_idx]) {
      # This is a colspan row - create a single spanning cell
      total_width <- sum(width_cols) + length(width_cols) - 1
      content_width <- ansi_nchar(row_data[1])
      padding_needed <- total_width - content_width
      body[row_idx] <- paste0(
        left_border,
        row_data[1],
        strrep(" ", padding_needed),
        right_border
      )
    } else {
      # Regular row - use column separators
      body[row_idx] <- paste0(
        left_border,
        paste(row_data, collapse = vline_char),
        right_border
      )
    }
  }

  return(body)
}

#' Assemble final table string with borders
#' @keywords internal
#' @noRd
assemble_table_string <- function(body, header, width_cols, x = NULL) {
  # Get line characters - use hardcoded values
  if (!is.null(x) && inherits(x, "tinytable")) {
    hline_char <- if (isTRUE(x@grid_hline)) "-" else NULL
    hline_header_char <- if (isTRUE(x@grid_hline_header)) "=" else NULL
  } else {
    hline_char <- "-"
    hline_header_char <- "="
  }

  # Create rule lines only if enabled; NULL parts drop out of c() below
  rule_head <- if (!is.null(hline_header_char)) {
    grid_create_line(width_cols, hline_header_char)
  } else {
    NULL
  }

  rule_line <- if (!is.null(hline_char)) {
    grid_create_line(width_cols, hline_char)
  } else {
    NULL
  }

  tab <- c(
    "\n",
    rule_line,
    if (header) c(body[1], rule_head, body[-1]) else body,
    rule_line,
    "\n"
  )

  return(paste(tab, collapse = "\n"))
}

# =============================================================================
# MAIN FUNCTION
# =============================================================================

build_eval_grid <- function(x, width_cols = NULL, ...) {
  is_matrix <- is.matrix(x)
  if (is_matrix) {
    tab <- x
  } else {
    tab <- x@data_body
  }

  if (is.null(width_cols)) {
    width_cols <- x@width_cols
  }

  # Apply text styling before padding (for non-matrix formats)
  if (!is_matrix && inherits(x, "tinytable") && nrow(x@style) > 0) {
    x <- style_grid_body(x)
    tab <- x@data_body
  }

  tthead <- inherits(x, "tinytable") && isTRUE(x@nhead > 0)
  if (length(colnames(x)) != 0 || tthead) {
    tab <- as.matrix(tab)
    tab <- base::rbind(colnames(x), tab)
    header <- TRUE
  } else {
    tab <- as.matrix(tab)
    header <- FALSE
  }

  # pad for readability
  padded <- sprintf(" %s ", tab)
  tab <- matrix(padded, ncol = ncol(tab))

  if (is.null(width_cols) || length(width_cols) == 0) {
    width_cols <- calculate_column_widths(tab, x, header, width_cols)
  }

  # Pad cells to match column widths
  tab <- pad_table_cells(tab, width_cols)

  # Apply background styling to padded cells (for grid formats only)
  if (!is_matrix && inherits(x, "tinytable") && nrow(x@style) > 0) {
    tab <- style_grid_body_background(tab, x, header)
  }

  if (inherits(x, "tinytable")) {
    x@data_body <- stats::setNames(as.data.frame(tab), names(x@data_body))
  }

  # Format rows with proper separators
  body <- format_table_rows(tab, x, header, width_cols)

  # Assemble final table string
  out <- assemble_table_string(body, header, width_cols, x)

  if (is_matrix) {
    return(out)
  }

  # rebuild output
  x@width_cols <- width_cols
  x@table_string <- out

  # output
  return(x)
}

# =============================================================================
# AUXILIARY FUNCTIONS
# =============================================================================

# insert horizontal rules everywhere (important for word)
grid_hlines <- function(x) {
  # Only add lines if horizontal lines are enabled
  if (!isTRUE(x@grid_hline)) {
    return(x)
  }

  rule_line <- grid_create_line(x@width_cols, "-")
  out <- x@table_string
  lines <- strsplit(out, split = "\\n")[[1]]
  if (length(lines) > 1) {
    corner_char <- "+"
    add_before <- c(
      FALSE,
      !startsWith(lines[-length(lines)], corner_char) &
        !startsWith(lines[-1L], corner_char) &
        lines[-1L] != ""
    )
    if (any(add_before)) {
      end_pos <- cumsum(1L + add_before)
      expanded <- character(length(lines) + sum(add_before))
      expanded[end_pos] <- lines
      expanded[end_pos[add_before] - 1L] <- rule_line
      lines <- expanded
    }
  }
  x@table_string <- paste(lines, collapse = "\n")
  return(x)
}

setMethod(
  f = "build_eval",
  signature = "tinytable_grid",
  definition = build_eval_grid
)

setMethod(
  f = "build_eval",
  signature = "matrix",
  definition = build_eval_grid
)

setMethod(
  f = "build_eval",
  signature = "tinytable_dataframe",
  definition = build_eval_grid
)

source("helpers.R")
using("tinysnapshot")
options(tinytable_print_output = "html")

x <- mtcars[1:4, 1:5]

tab <- tt(x, theme = "striped")
expect_snapshot_print(print_html(tab), "html-striped.html")

tab <- tt(x, theme = "striped") |>
  style_tt(color = "orange")
expect_snapshot_print(print_html(tab), "html-striped_orange.html")

# Issue #92: header alignment
k <- structure(
  list(
    Column1 = c("Some text", "123"),
    Column2 = c(
      "Some text",
      "456"
    )
  ),
  row.names = c(NA, -2L),
  class = "data.frame"
)
tab <- tt(k, width = 1) |> style_tt(j = 2, align = "r")
expect_snapshot_print(print_html(tab), "html-issue92.html")

# tutorial.qmd: vectorized settings
tab <- tt(x) |>
  style_tt(
    j = 2:3,
    color = c("orange", "green"),
    background = "black"
  )
expect_snapshot_print(print_html(tab), "html-vectorized_color_j.html")

# Issue #58
tab <- tt(iris[1:10, ]) |>
  style_tt(align = "c") |>
  group_tt(j = list("Sepal" = 1:2, "Petal" = 3:4))
expect_snapshot_print(print_html(tab), "html-issue58.html")

# Issue #88: indent
tab <- tt(iris[1:10, ]) |>
  style_tt(i = 2:4, indent = 3)
expect_snapshot_print(print_html(tab), "html-issue88.html")

# tutorial.qmd: heatmap
k <- data.frame(matrix(1:20, ncol = 5))
colnames(k) <- NULL
bg <- hcl.colors(20, "Inferno")
fg <- ifelse(as.matrix(k) < 17, tail(bg, 1), head(bg, 1))
fs <- 1:20
tab <- tt(k, width = .5, theme = "empty") |>
  style_tt(
    i = 1:4,
    j = 1:5,
    color = fg,
    background = bg,
    fontsize = fs
  )
expect_snapshot_print(print_html(tab), "html-heatmap.html")

# Caption
tab <- tt(mtcars[1:3, 1:3], caption = "Blah blah")
expect_snapshot_print(print_html(tab), "html-caption.html")

# Footnote
tab <- tt(mtcars[1:3, 1:3], notes = list(a = "Blah.", b = "Blah blah."))
expect_snapshot_print(print_html(tab), "html-footnote.html")

# Style individual cells
tab <- tt(mtcars[1:4, 1:4]) |>
  style_tt(
    i = 2:3,
    j = c(1, 3, 4),
    italic = TRUE,
    background = "pink",
    color = "orange"
  )
expect_snapshot_print(print_html(tab), "html-individual_cells.html")

# Issue #432: S4 refactor broke bootstrap_css argument
x <- mtcars[1:4, 1:5]
tab <- tt(x) |>
  style_tt(j = 1, bootstrap_css = "font-weight: bold; color: red;")
expect_snapshot_print(print_html(tab), "html-bootstrap_css.html")

# Line break
d <- data.frame(
  "{Sed ut \\\\ perspiciatis unde}",
  "dicta sunt<br> explicabo. Nemo"
) |>
  setNames(c("LaTeX line break", "HTML line break"))
d <- tt(d)
expect_snapshot_print(print_html(d), "html-line_break.html")

# Formatting
dat <- data.frame(
  w = c(143002.2092, 201399.181, 100188.3883),
  x = c(1.43402, 201.399, 0.134588),
  y = as.Date(c(999, 675, 3), origin = "1970-01-01"),
  z = c(TRUE, TRUE, FALSE)
)
dat <- tt(dat, digits = 2)
expect_snapshot_print(print_html(dat), "html-formatting.html")

# Missing value replacement
tab <- data.frame(a = c(NA, 1, 2), b = c(3, NA, 5))
tab <- tt(tab)
tab <- format_tt(tab, replace = "-")
expect_snapshot_print(print_html(tab), "html-missing_value.html")

# Alignment
dat <- data.frame(
  a = c("a", "aa", "aaa"),
  b = c("b", "bb", "bbb"),
  c = c("c", "cc", "ccc")
)
dat <- tt(dat) |> style_tt(j = 1:3, align = "lcr")
expect_snapshot_print(print_html(dat), "html-alignment.html")

# Font size
x <- mtcars[1:4, 1:5]
x <- tt(x) |> style_tt(j = "mpg|hp|qsec", fontsize = 1.5)
expect_snapshot_print(print_html(x), "html-font_size.html")

# Merge cells
x <- (mtcars[1:4, 1:5])
x <- tt(x) |>
  style_tt(
    i = 2,
    j = 2,
    colspan = 3,
    rowspan = 2,
    align = "c",
    alignv = "m",
    color = "white",
    background = "black",
    bold = TRUE
  )
expect_snapshot_print(print_html(x), "html-merge_cells.html")

# Spanning cells
tab <- aggregate(mpg ~ cyl + am, FUN = mean, data = mtcars)
tab <- tab[order(tab$cyl, tab$am), ]
tab <- tt(tab, digits = 2) |>
  style_tt(i = c(1, 3, 5), j = 1, rowspan = 2, alignv = "t")
expect_snapshot_print(print_html(tab), "html-spanning_cells.html")

# Omit headers
k <- (mtcars[1:4, 1:5])
colnames(k) <- NULL
k <- tt(k)
expect_snapshot_print(print_html(k), "html-omit_headers.html")

# Conditional styling
k <- mtcars[1:10, c("mpg", "am", "vs")]
k <- tt(k) |>
  style_tt(
    i = which(k$am == k$vs),
    background = "teal",
    color = "white"
  )
expect_snapshot_print(print_html(k), "html-conditional_styling.html")

# Borders
x <- tt(mtcars[1:4, 1:5], theme = "empty") |>
  style_tt(
    i = 0:3,
    j = 1:3,
    line = "tblr",
    line_width = 0.4,
    line_color = "orange"
  )
expect_snapshot_print(print_html(x), "html-borders.html")

# Images
dat <- data.frame(
  Species = c("Spider", "Squirrel"),
  Image = ""
)
img <- c(
  "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcSeLPSPrPtVgPg6BLCiN6lBYy8l1xNy0T5yttVjkIk0L3Rva8Zl",
  "https://encrypted-tbn2.gstatic.com/images?q=tbn:ANd9GcQdBlFVajljNz5qMbO622ihkIU2r6yA5whM9b8MbRGKOfJ8_UmZ"
)
dat <- tt(dat) |>
  plot_tt(j = 2, images = img, height = 3)
expect_snapshot_print(print_html(dat), "html-images.html")

# Issue #297: group_tt() breaks alignment
tab <- data.frame(
  Person = c("Alice", "Bob", "Charlemagne"),
  Fruit = c("Apple", "Banana", "Cantaloupe"),
  Count = c(4, 238432, 32)
) |>
  tt() |>
  group_tt(i = list("Thing" = 1, "Thing again" = 2)) |>
  style_tt(i = c(1, 3), align = "l") |>
  style_tt(j = 1:3, align = "l")
expect_snapshot_print(print_html(tab), "html-issue297.html")


# # Issue #355a: rowspan breaks indexing
tab <- tt(mtcars[1:6, 1:5]) |>
  group_tt(j = list("Hello" = 1:2, "World" = 3:5)) |>
  group_tt(j = list("Foo" = 2:3, "Bar" = 4:5)) |>
  style_tt(i = c(1, 3, 5), j = 1:2, bold = TRUE, color = "red", rowspan = 2) |>
  style_tt(i = 1 + c(1, 3, 5), j = 3, bold = TRUE, color = "green") |>
  style_tt(i = c(1, 3, 5), j = 3, bold = TRUE, color = "orange") |>
  style_tt(i = 0, background = "black", color = "white") |>
  style_tt(i = -1, color = "pink") |>
  style_tt(i = -2, color = "blue") |>
  style_tt(i = 6, background = "pink")
expect_snapshot_print(print_html(tab), "html-issue355a.html")

tab <- tt(mtcars[1:9, 1:8]) |>
  group_tt(
    j = list("Foo" = 1:3, "Bar" = 4:8),
    i = list(
      "I like (fake) hamburgers" = 3,
      "She prefers halloumi" = 4,
      "They love tofu" = 7
    )
  ) |>
  style_tt(
    i = c(3, 5, 9),
    align = "c",
    color = "white",
    background = "gray",
    bold = TRUE
  ) |>
  style_tt(4, background = "pink")
expect_snapshot_print(print_html(tab), "html-issue355b.html")



# Issue #575
tab <- head(iris) |>
  tt() |>
  theme_empty() |>
  group_tt(j = list("Hello" = 1:2, "Cruel" = 3, "World" = 4:5)) |>
  group_tt(j = list("Foo" = 1, "Bar" = 2:5)) |>
  style_tt(align = "c") |>
  style_tt(i = 2, j = 2, colspan = 3) |>
  style_tt(i = 2, j = 2, line = "b", line_color = "green") |>
  style_tt(i = 2, j = 5, line = "b", line_color = "orange") |>
  style_tt(i = -1, j = 1:2, line = "b", line_color = "orange") |>
  style_tt(i = -2, j = 3, line = "b", line_color = "green") |>
  style_tt(i = -1, j = 4:5, line = "b", line_color = "blue") |>
  style_tt(i = 0, line = "b", line_color = "pink") |>
  style_tt(i = 6, line = "b", line_color = "pink")
expect_snapshot_print(print_html(tab), "html-issue575.html")




## TODO: reinstate portable test, but there's a snapshot challenge
# # Issue #340: plot_tt should be able to create self-contained HTML
# if (Sys.info()["user"] == "vincent") {
#     dat <- data.frame(
#         Name = c("bar", "line"),
#         Image = "") |>
#         tt() |>
#         plot_tt(i = 1, j = 2, fun = "bar", data = list(2)) |>
#         plot_tt(i = 2, j = 2, fun = "line", data = list(data.frame(x = 1:3, y = 3:1)))
#     expect_snapshot_print(print_html(dat, "html_portable"), "html-images-portable.html")
#     op = options("tinytable_html_portable" = TRUE)
#     expect_snapshot_print(print_html(dat, "html"), "html-images-portable.html")
#     options(op)
# }

# # Built-in plots
# # cannot be tested because the names of plots are random and set seed doesn't work
# set.seed(1024)
# plot_data <- list(mtcars$mpg, mtcars$hp, mtcars$qsec)
# dat <- data.frame(
#   Variables = c("mpg", "hp", "qsec"),
#   Histogram = "",
#   Density = "",
#   Bar = "",
#   Line = ""
# )
# lines <- lapply(1:3, \(x) data.frame(x = 1:10, y = rnorm(10)))
# dat<-tt(dat) |>
#   plot_tt(j = 2, fun = "histogram", data = plot_data) |>
#   plot_tt(j = 3, fun = "density", data = plot_data, color = "darkgreen") |>
#   plot_tt(j = 4, fun = "bar", data = list(2, 3, 6), color = "orange") |>
#   plot_tt(j = 5, fun = "line", data = lines, color = "blue") |>
#   style_tt(j = 2:5, align = "c")
# expect_snapshot_print(print_html(dat), "html-built_in_plots")

options(tinytable_print_output = NULL)

# =============================================================================
# HELPER FUNCTIONS
# =============================================================================

#' Clean style strings
#' @keywords internal
#' @noRd
clean_style_strings <- function(k) {
  k <- gsub("\\s*", "", k)
  k <- gsub(",+", ",", k)
  k <- gsub("^,", "", k, perl = TRUE)

  # Remove duplicates from each style string
  k <- sapply(k, function(style_string) {
    if (is.na(style_string) || trimws(style_string) == "") {
      return(style_string)
    }

    # Split by comma and remove duplicates
    parts <- trimws(strsplit(style_string, ",")[[1]])
    parts <- parts[parts != ""] # Remove empty parts
    unique_parts <- unique(parts)

    # Rejoin with proper spacing
    paste(unique_parts, collapse = ", ")
  }, USE.NAMES = FALSE)

  k <- trimws(k)
  return(k)
}

#' Prepare d-column styling
#' @keywords internal
#' @noRd
prepare_dcolumn <- function(x, sty) {
  if (nrow(sty) > 0) {
    dcol_j <- sty[grepl("^d$", sty[["align"]]), "j"]
    dcol_j <- if (length(dcol_j) == 0) NULL else unique(dcol_j)
    for (idx_j in dcol_j) {
      spec <- calculate_dcolumn_spec(idx_j, x)
      spec <- sprintf("column{%s}={%s}\n", idx_j, spec)
      x@table_string <- insert_tabularray_content(
        x@table_string,
        content = spec,
        type = "inner"
      )
      spec <- sprintf("cell{%s}{%s}={guard,halign=c,},", seq_len(x@nhead), idx_j)
      x@table_string <- insert_tabularray_content(
        x@table_string,
        content = spec,
        type = "inner"
      )
    }
  }
  return(x)
}

#' Generate tabularray column specifications
#' @keywords internal
#' @noRd
tabularray_columns <- function(x, rec) {
  # Complete columns (first because of d-column)
  cols <- unique(
    rec[
      (rec$span != "" | rec$set != "") & rec$complete_column,
      c("j", "set", "span"),
      drop = FALSE
    ]
  )

  spec <- by(cols, list(cols$set, cols$span), function(k) {
    sprintf("column{%s}={%s}{%s}", latex_range_string(k$j), k$span, k$set)
  })
  spec <- unique(as.vector(unlist(spec)))

  x@table_string <- insert_tabularray_content(
    x@table_string,
    content = spec,
    type = "inner"
  )

  return(x)
}

#' Generate tabularray row specifications
#' @keywords internal
#' @noRd
tabularray_rows <- function(x, rec) {
  # Complete rows
  rows <- unique(
    rec[
      (rec$span != "" | rec$set != "") &
        rec$complete_row &
        !rec$complete_column,
      c("i", "set", "span"),
      drop = FALSE
    ]
  )

  spec <- by(rows, list(rows$set, rows$span), function(k) {
    sprintf("row{%s}={%s}{%s}", latex_range_string(k$i), k$span, k$set)
  })
  spec <- unique(as.vector(unlist(spec)))

  x@table_string <- insert_tabularray_content(
    x@table_string,
    content = spec,
    type = "inner"
  )

  return(x)
}

#' Generate tabularray cell specifications
#' @keywords internal
#' @noRd
tabularray_cells <- function(x, rec) {
  # Individual cells
  cells <- unique(
    rec[
      (rec$span != "" | rec$set != "") &
        !rec$complete_row &
        !rec$complete_column, ,
      drop = FALSE
    ]
  )

  if (nrow(cells) > 0) {
    # need to split by j otherwise we can end up with rectangular index that
    # cover cells that should not be styled
    cellsplit <- split(cells, list(cells$j, cells$set, cells$span))
    cellsplit <- Filter(function(k) nrow(k) > 0, cellsplit)
    spec <- sapply(cellsplit, function(cells) {
      sprintf(
        "cell{%s}{%s}={%s}{%s}",
        latex_range_string(cells$i),
        cells$j[1],
        cells$span[1],
        cells$set[1]
      )
    })
    spec <- sort(unique(as.vector(unlist(spec))))

    x@table_string <- insert_tabularray_content(
      x@table_string,
      content = spec,
      type = "inner"
    )
  }

  return(x)
}



#' Apply tabularray specifications
#' @keywords internal
#' @noRd
apply_tabularray_specs <- function(x) {
  x@table_string <- insert_tabularray_content(
    x@table_string,
    content = unique(as.vector(stats::na.omit(x@tabularray_inner))),
    type = "inner"
  )

  x@table_string <- insert_tabularray_content(
    x@table_string,
    content = unique(as.vector(stats::na.omit(x@tabularray_outer))),
    type = "outer"
  )

  return(x)
}

#' Process tabularray lines using expanded style data
#' @keywords internal
#' @noRd
process_tabularray_lines <- function(x, lines) {
  if (is.null(lines) || nrow(lines) == 0) {
    return(x)
  }

  # Adjust i values for header offset
  lines$i <- lines$i + x@nhead

  # Normalize colors once and define preambles
  line_color_map <- build_color_map(lines$line_color, "tabularray")
  for (col in line_color_map) {
    x <- define_color_preamble(x, col)
  }
  lines$line_color_mapped <- apply_color_map(lines$line_color, line_color_map)

  # Process horizontal lines
  hlines <- lines[!is.na(lines$line) & grepl("b|t", lines$line), ]
  if (nrow(hlines) > 0) {
    x <- process_tabularray_axis_lines(x, hlines, axis = "i")
  }

  # Process vertical lines
  vlines <- lines[!is.na(lines$line) & grepl("l|r", lines$line), ]
  if (nrow(vlines) > 0) {
    x <- process_tabularray_axis_lines(x, vlines, axis = "j")
  }

  return(x)
}

#' Process tabularray other styles using expanded style data
#' @keywords internal
#' @noRd
process_tabularray_other_styles <- function(x, other) {
  if (is.null(other) || nrow(other) == 0) {
    # Apply tabularray specifications even if no other styles
    x <- apply_tabularray_specs(x)
    return(x)
  }

  # Adjust i values for header offset and map alignv values
  other$i <- other$i + x@nhead
  other$alignv <- map_alignv(other$alignv, "tabularray")

  # Create record grid
  rec <- expand.grid(
    i = c(seq_len(x@nrow + x@nhead)),
    j = seq_len(x@ncol)
  )

  set <- span <- rep("", nrow(rec))

  # Prepare d-columns (special case)
  x <- prepare_dcolumn(x, other)

  # Normalize color map once and define preambles
  color_map <- build_color_map(c(other$color, other$background), "tabularray")
  for (col in color_map) {
    x <- define_color_preamble(x, col)
  }

  # Build style strings directly from columns. `other` is the filtered
  # rectangular style grid, so each row already represents one concrete cell;
  # extracting one-row data frames and rescanning `rec` is unnecessary.
  yes <- function(z) !is.na(z) & z
  font_cmds <- paste0(
    ifelse(yes(other$bold), "\\bfseries", ""),
    ifelse(yes(other$italic), "\\itshape", ""),
    ifelse(yes(other$monospace), "\\ttfamily", ""),
    ifelse(yes(other$smallcap), "\\scshape", "")
  )
  cmd_strs <- paste0(
    ifelse(yes(other$underline), "\\tinytableTabularrayUnderline", ""),
    ifelse(yes(other$strikeout), "\\tinytableTabularrayStrikeout", "")
  )

  col_idx <- which(!is.na(other$color) & other$color %in% names(color_map))
  if (length(col_idx)) {
    mapped <- unname(color_map[other$color[col_idx]])
    mapped <- sub("^#", "c", mapped)
    cmd_strs[col_idx] <- sprintf("%s, fg=%s", cmd_strs[col_idx], mapped)
  }
  bg_idx <- which(!is.na(other$background) & other$background %in% names(color_map))
  if (length(bg_idx)) {
    mapped <- unname(color_map[other$background[bg_idx]])
    mapped <- sub("^#", "c", mapped)
    cmd_strs[bg_idx] <- sprintf("%s, bg=%s", cmd_strs[bg_idx], mapped)
  }

  font_sets <- rep("", nrow(other))
  fontsize <- suppressWarnings(as.numeric(other$fontsize))
  idx <- which(!is.na(fontsize))
  if (length(idx)) {
    font_cmds[idx] <- sprintf(
      "%s\\fontsize{%sem}{%sem}\\selectfont",
      font_cmds[idx],
      format_markup_num(fontsize[idx]),
      format_markup_num(fontsize[idx] + 0.3)
    )
  }
  idx <- which(!is.na(other$align) & !grepl("d", other$align))
  font_sets[idx] <- sprintf("%s, halign=%s,", font_sets[idx], other$align[idx])
  idx <- which(!is.na(other$alignv))
  font_sets[idx] <- sprintf("%s, valign=%s,", font_sets[idx], other$alignv[idx])
  idx <- which(!is.na(other$indent) & other$indent > 0)
  if (length(idx)) {
    font_sets[idx] <- sprintf(
      "%s preto={\\hspace{%sem}},",
      font_sets[idx],
      format_markup_num(other$indent[idx])
    )
  }

  span_strs <- rep("", nrow(other))
  idx <- which(!is.na(other$colspan))
  span_strs[idx] <- paste0(span_strs[idx], "c=", other$colspan[idx], ",")
  idx <- which(!is.na(other$rowspan))
  span_strs[idx] <- paste0(span_strs[idx], "r=", other$rowspan[idx], ",")
  if (length(x@width) == ncol(x)) {
    for (row in which(!is.na(other$colspan))) {
      cols <- other$j[row]:(other$j[row] + other$colspan[row] - 1)
      w <- sum(x@width[cols])
      font_sets[row] <- paste(font_sets[row], sprintf("wd=%s\\linewidth,", format_markup_num(w)))
    }
  }

  n_i <- x@nrow + x@nhead
  target_idx <- other$i + (other$j - 1L) * n_i
  valid_target <- !is.na(target_idx) & target_idx >= 1L & target_idx <= nrow(rec)

  # Preserve append semantics defensively if duplicate coordinates ever reach
  # this function, while avoiding a full-grid logical scan for every row.
  for (row in which(valid_target)) {
    idx <- target_idx[row]

    # Add font styling if present
    font_cmd <- font_cmds[row]
    if (trimws(font_cmd) != "") {
      set[idx] <- sprintf("%s font=%s, ", set[idx], font_cmd)
    }

    # Format command string for remaining cmd styles
    cmd <- cmd_strs[row]
    if (grepl("^,", cmd)) {
      tmp <- "%s, %s, "
    } else {
      tmp <- "%s, cmd=%s, "
    }
    if (trimws(cmd) != "") {
      set[idx] <- sprintf(tmp, set[idx], cmd)
    }

    # Add font settings
    set[idx] <- paste0(set[idx], font_sets[row])

    # Add spans
    span[idx] <- paste0(span[idx], span_strs[row])
  }

  # Only styled cells need the comparatively expensive split/dedup cleanup.
  used <- unique(target_idx[valid_target])
  set[used] <- clean_style_strings(set[used])
  span[used] <- clean_style_strings(span[used])
  rec$set <- set
  rec$span <- span

  # Mark complete rows and columns
  all_i <- seq_len(x@nrow + x@nhead)
  all_j <- seq_len(x@ncol)

  # rec contains one row per (i, j), so a style group covers a complete column
  # or row exactly when its group size equals the corresponding dimension.
  # This replaces two by()/transform()/rbind() cycles over the full grid.
  col_group <- interaction(rec$j, rec$set, rec$span, drop = TRUE)
  col_size <- tabulate(as.integer(col_group))
  rec$complete_column <- col_size[as.integer(col_group)] == length(all_i)

  row_group <- interaction(rec$i, rec$set, rec$span, drop = TRUE)
  row_size <- tabulate(as.integer(row_group))
  rec$complete_row <- row_size[as.integer(row_group)] == length(all_j)

  # Generate tabularray specifications
  x <- tabularray_columns(x, rec)
  x <- tabularray_rows(x, rec)
  x <- tabularray_cells(x, rec)

  # Apply tabularray specifications
  x <- apply_tabularray_specs(x)

  return(x)
}

#' Process horizontal or vertical lines from expanded style data
#'
#' Shared implementation for hlines (`axis = "i"`) and vlines (`axis = "j"`).
#' Lines are grouped by (position, line spec) so that disjoint requests
#' sharing the same style never merge into a rectangular cross-product.
#' @keywords internal
#' @noRd
process_tabularray_axis_lines <- function(x, lines, axis) {
  # Vectorize line specification building
  # Use precomputed line_color_mapped
  line_colors <- ifelse(
    grepl("^#", lines$line_color_mapped),
    sub("^#", "c", lines$line_color_mapped),
    lines$line_color_mapped
  )

  line_widths <- ifelse(is.na(lines$line_width), 0.1, lines$line_width)
  line_types <- if ("line_type" %in% names(lines)) lines$line_type else NA_character_
  line_types <- ifelse(is.na(line_types), "solid", line_types)
  line_specs <- sprintf(
    "%s, %s, %sem", line_types, line_colors, format_markup_num(line_widths)
  )

  # Add trimming vectorized
  has_trim <- !is.na(lines$line_trim) & nzchar(lines$line_trim)
  trim_l <- has_trim & grepl("l", lines$line_trim)
  trim_r <- has_trim & grepl("r", lines$line_trim)

  line_specs <- ifelse(trim_l, paste0(line_specs, ", l=-0.5"), line_specs)
  line_specs <- ifelse(trim_r, paste0(line_specs, ", r=-0.5"), line_specs)

  if (axis == "i") {
    # "b" draws below the row: shift to the next hline slot (vectorized)
    pos <- ifelse(grepl("b", lines$line), lines$i + 1, lines$i)
    extent <- lines$j
    template <- "hline{%s}={%s}{%s}"
  } else {
    # "r" draws right of the column: shift to the next vline slot (vectorized)
    pos <- ifelse(grepl("r", lines$line), lines$j + 1, lines$j)
    extent <- lines$i
    template <- "vline{%s}={%s}{%s}"
  }

  dat <- data.frame(
    pos = pos,
    extent = extent,
    lin = line_specs,
    stringsAsFactors = FALSE
  )

  spec <- by(dat, list(dat$pos, dat$lin), function(k) {
    pos_val <- latex_range_string(k$pos)
    extent_val <- latex_range_string(k$extent)
    lin_val <- k$lin[1]
    # Skip invalid line specifications
    if (is.na(lin_val) || lin_val == "" || pos_val == "" || extent_val == "") {
      return(NULL)
    }
    sprintf(template, pos_val, extent_val, lin_val)
  })
  spec <- unique(as.vector(unlist(spec)))
  # Remove any NULL or NA entries
  spec <- spec[!is.na(spec) & spec != "NULL"]

  x@table_string <- insert_tabularray_content(
    x@table_string,
    content = spec,
    type = "inner"
  )

  return(x)
}

# =============================================================================
# MAIN METHOD
# =============================================================================

setMethod(
  f = "style_eval",
  signature = "tinytable_tabularray",
  definition = function(
      x,
      i = NULL,
      j = NULL,
      bold = FALSE,
      italic = FALSE,
      monospace = FALSE,
      underline = FALSE,
      strikeout = FALSE,
      color = NULL,
      background = NULL,
      fontsize = NULL,
      align = NULL,
      alignv = NULL,
      line = NULL,
      line_color = "black",
      line_width = 0.1,
      colspan = NULL,
      rowspan = NULL,
      indent = 0,
      ...) {
    # Use populated @style_other / @style_lines from build_tt()
    inputs <- style_backend_inputs(x, STYLE_PROPS_TABULARRAY)
    other <- inputs$other
    lines <- inputs$lines

    # Process lines using the expanded data
    x <- process_tabularray_lines(x, lines)

    # Process other styles using the expanded data
    x <- process_tabularray_other_styles(x, other)

    return(x)
  })

source("helpers.R")
using("tinysnapshot")
requiet("tibble")
requiet("pillar")
# exit_file("fail on github actions")

# Issue #426
x <- tibble(a = c(pi, 2.3, 1), b = letters[1:3])
x <- format_tt(x, digits = 1)
expect_inherits(x, "tbl_df")
expect_equivalent(x[[1]], c("3", "2", "1"))
expect_equivalent(x[[2]], c("a", "b", "c"))

# Issue #141
k2 <- tibble(
  x3 = 9:11 * 100 + 0.5,
  x4 = 9:11 * 100 + 0.5,
  x5 = 9:11 * 100 + 0.5,
)
tab <- tt(k2, digits = 1)

options("tinytable_print_output" = "markdown")
expect_snapshot_print(tab, "tibble-tibble_issue141")


options("tinytable_print_output" = NULL)
x <- tibble(
  x3 = pillar::num(9:11 * 100 + 0.5, sigfig = 3),
  x4 = pillar::num(9:11 * 100 + 0.5, sigfig = 4),
  x5 = pillar::num(9:11 * 100 + 0.5, sigfig = 5),
)
tab <- tt(x)
t <- expect_table(tab)
expect_snapshot_print(t[["typst"]], "tibble-pillar_num.typ")
expect_snapshot_print(t[["markdown"]], "tibble-pillar_num.md")
expect_snapshot_print(t[["latex"]], "tibble-pillar_num.tex")

exit_file("fail on github actions")
expect_snapshot_print(t[["html"]], "tibble-pillar_num.html")


# markdown
exit_file("ANSI tinysnapshot issue")
options("tinytable_print_output" = "markdown")
expect_snapshot_print(tab, "tibble-pillar_num")
options("tinytable_print_output" = NULL)

group_grid_col <- function(x, j, ...) {
  # Check if there are any column groups to process
  if (nrow(x@group_data_j) == 0) {
    return(x)
  }

  tab <- x@table_string
  cw <- x@width_cols

  # Process @group_data_j to create column group headers
  # Process all rows of column groups (not just the first)
  if (nrow(x@group_data_j) > 0) {
    all_header_lines <- list()

    # Process each row in @group_data_j (from last to first to maintain proper order)
    for (group_row_idx in nrow(x@group_data_j):1) {
      group_row <- as.character(x@group_data_j[group_row_idx, ])

      spans <- parse_group_spans(group_row)

      if (nrow(spans) > 0) {
        # Positional cells: labelled spans plus " "-labelled runs filling the
        # holes up to the last spanned column (duplicate labels preserved)
        runs <- Map(seq.int, spans$start, spans$end)
        labels <- spans$label
        missing_nums <- setdiff(seq_len(max(spans$end)), unlist(runs))
        if (length(missing_nums) > 0) {
          holes <- split(missing_nums, cumsum(c(1, diff(missing_nums) != 1)))
          runs <- c(runs, unname(holes))
          labels <- c(labels, rep(" ", length(holes)))
        }
        idx <- order(vapply(runs, min, numeric(1)))
        runs <- runs[idx]
        labels <- labels[idx]

        cw_grouped <- sapply(runs, function(k) sum(cw[k]) + length(cw[k]) - 1)
        header_matrix <- t(matrix(labels))
        header_formatted <- build_eval(header_matrix, cw_grouped)
        header_lines <- strsplit(header_formatted, split = "\\n")[[1]]
        header_lines <- header_lines[header_lines != ""]
        header_line <- header_lines[2]

        # Store header line for this group row
        all_header_lines[[group_row_idx]] <- header_line
      }
    }

    # Insert all header lines into the table
    if (length(all_header_lines) > 0) {
      z <- strsplit(tab, split = "\\n")[[1]]
      z <- z[z != ""]

      # Determine insertion position based on grid_hline setting
      header_lines_to_insert <- unlist(all_header_lines)

      if (isTRUE(x@grid_hline)) {
        # With hlines: insert after the first line (top border)
        z <- c(z[1], header_lines_to_insert, z[2:length(z)])
      } else {
        # Without hlines: insert at the very beginning (before column names)
        z <- c(header_lines_to_insert, z)
      }

      # missing cell at the end
      nc <- ansi_nchar(z)
      idx <- ansi_nchar(z) < max(nc)
      z[idx] <- paste0(
        z[idx],
        strrep(" ", max(nc) - ansi_nchar(z[idx]) - 1),
        "|"
      )

      tab <- paste(z, collapse = "\n")
      x@table_string <- tab
    }
  }

  return(x)
}


#' tinytable S4 method
#'
#' @keywords internal
setMethod(
  f = "group_eval_j",
  signature = "tinytable_dataframe",
  definition = function(x, i = NULL, j = NULL, ...) {
    # Only handle column grouping - row insertions now use matrix insertion
    x <- group_grid_col(x, j)
    return(x)
  }
)

#' tinytable S4 method
#'
#' @keywords internal
setMethod(
  f = "group_eval_j",
  signature = "tinytable_grid",
  definition = function(x, i = NULL, j = NULL, ...) {
    # Only handle column grouping - row insertions now use matrix insertion
    x <- group_grid_col(x, j)
    return(x)
  }
)

setMethod(
  f = "build_eval",
  signature = "tinytable_tabularray",
  definition = function(x, ...) {
    template <- readLines(
      system.file("templates/tabularray.tex", package = "tinytable")
    )

    ncols <- ncol(x)
    nrows <- nrow(x)

    tall <- FALSE
    if (length(x@caption) > 0) {
      tall <- TRUE
    }
    if (length(x@notes) > 0) {
      tall <- TRUE
    }

    # placement
    if (length(x@placement) == 1) {
      assert_string(x@placement)
      template <- sub(
        "\\\\begin\\{table\\}",
        sprintf("\\\\begin{table}[%s]\n", x@placement),
        template
      )
    }

    # body: main
    if (length(colnames(x)) > 0) {
      header <- paste(colnames(x), collapse = " & ")
      header <- paste(header, "\\\\")
    } else {
      header <- NULL
    }
    body <- apply(as.data.frame(x@data_body), 1, paste, collapse = " & ")
    body <- paste(body, "\\\\")

    # body: finish
    idx <- grep("\\$tinytable_BODY", template)
    out <- c(
      template[1:(idx - 1)],
      header,
      body,
      template[(idx + 1):length(template)]
    )

    out <- trimws(out)
    out <- paste(out, collapse = "\n")

    if (length(x@caption) > 0) {
      spec <- sprintf("caption={%s},", x@caption[1])
      out <- insert_tabularray_content(out, content = spec, type = "outer")
    }

    if (length(x@width) == 0) {
      tabularray_cols <- rep("Q[]", ncol(x))
    } else if (length(x@width) == 1) {
      tabularray_cols <- rep("X[]", ncol(x))
      spec <- sprintf("width={%s\\linewidth},", format_markup_num(round(x@width, 4)))
      out <- insert_tabularray_content(out, content = spec, type = "inner")
    } else if (length(x@width) > 1) {
      tabularray_cols <- sprintf("X[%s]", format_markup_num(x@width))
      spec <- sprintf("width={%s\\linewidth},", format_markup_num(round(sum(x@width), 4)))
      out <- insert_tabularray_content(out, content = spec, type = "inner")
    }

    # colspec (we don't need rowspec)
    colspec <- sprintf("colspec={%s},", paste(tabularray_cols, collapse = ""))
    out <- insert_tabularray_content(out, content = colspec, type = "inner")

    # notes
    if (length(x@notes) > 0) {
      if (length(x@caption) == 0) {
        # otherwise an empty caption is created automatically
        out <- insert_tabularray_content(
          out,
          content = "entry=none,label=none",
          type = "outer"
        )
      }
      if (is.null(names(x@notes))) {
        lab <- sapply(seq_along(x@notes), function(k) strrep(" ", k - 1))
      } else {
        lab <- NULL
        pad <- 0
        for (i in seq_along(x@notes)) {
          # tabularray requires unique labels, but multiple blanks work
          if (names(x@notes)[i] == "") {
            lab[i] <- strrep(" ", pad) # not sure why -1 is necessary in tabularray
            pad <- pad + 1
          } else {
            lab[i] <- names(x@notes)[i]
          }
        }
      }
      notes <- sapply(x@notes, function(n) if (is.list(n)) n$text else n)
      for (k in seq_along(notes)) {
        spec <- sprintf("note{%s}={%s}", lab[k], notes[k])
        out <- insert_tabularray_content(out, content = spec, type = "outer")
      }
    }

    if (isTRUE(tall)) {
      out <- sub("\\begin{tblr}", "\\begin{talltblr}", out, fixed = TRUE)
      out <- sub("\\end{tblr}", "\\end{talltblr}", out, fixed = TRUE)
    }

    x@table_string <- out
    x@body <- body

    return(x)
  })

# =============================================================================
# HELPER FUNCTIONS
# =============================================================================

#' Handle deprecated arguments
#' @keywords internal
#' @noRd
handle_deprecated_args <- function(...) {
  if ("indent" %in% ...names()) {
    warning(
      'Argument `indent` is deprecated and will be removed in a future version. Use `style_tt("~group", indent = 1)` instead.',
      call. = FALSE
    )
  }
}

#' Process row-group insertion (named list `i`, or integer `i` with matrix `j`)
#'
#' `j = NULL` triggers list-to-matrix conversion in `group_tt_ij_k()`; a
#' matrix `j` is inserted as-is at the positions in `i`. Repeated calls
#' append to `@group_data_i` so row groups can be stacked.
#' @keywords internal
#' @noRd
process_group_i <- function(x, i, j = NULL) {
  k <- group_tt_ij_k(x, i, j)
  converted_from_list <- k[[3]]

  # Calculate indices and update table. Positions are sorted (ascending) by
  # group_tt_ij_k(), so each insertion shifts subsequent positions by one row.
  positions <- k[[1]]
  idx <- positions + seq_along(positions) - 1
  x@nrow <- x@nrow + length(positions)

  # Create group data frame
  group_matrix <- k[[2]]
  group_df <- as.data.frame(group_matrix, stringsAsFactors = FALSE)

  # Handle row duplication for multiple positions with single matrix row
  if (length(positions) > 1 && nrow(group_df) == 1) {
    group_df <- group_df[rep(1, length(positions)), , drop = FALSE]
  }

  # Set column names to match the table
  if (ncol(group_df) == ncol(x@data)) {
    colnames(group_df) <- colnames(x@data)
  }

  # Add to existing group data or create new
  if (nrow(x@group_data_i) == 0) {
    x@group_data_i <- group_df
  } else {
    x@group_data_i <- rbind_nocol(x@group_data_i, group_df)
  }

  # Add indices to track final positions
  x@group_index_i <- c(x@group_index_i, idx)

  # Apply styling for list-converted group headers
  if (converted_from_list) {
    x <- style_tt(x, i = idx, j = 1, colspan = ncol(x))
  }

  return(x)
}

#' Process column grouping
#' @keywords internal
#' @noRd
process_group_j <- function(x, j) {
  j <- sanitize_group_index(j, hi = ncol(x), orientation = "column")
  x@nhead <- x@nhead + 1

  # Create group row
  new_row <- rep(NA_character_, ncol(x))

  # Set group labels in appropriate columns
  for (i in seq_along(j)) {
    group_name <- names(j)[i]
    group_cols <- j[[i]]
    # Set the label in the first column of the span
    new_row[group_cols[1]] <- group_name
    # For spans > 1: set continuation columns based on output format
    # HTML needs empty strings, Typst needs repeated text
    if (length(group_cols) > 1) {
      new_row[group_cols[-1]] <- ""
    }
  }

  # Add to existing group data or create new
  if (nrow(x@group_data_j) == 0) {
    x@group_data_j <- data.frame(matrix(
      NA_character_,
      nrow = 1,
      ncol = ncol(x@data)
    ))
    colnames(x@group_data_j) <- colnames(x@data)
    x@group_data_j[1, ] <- new_row
  } else {
    new_header_row <- data.frame(matrix(
      NA_character_,
      nrow = 1,
      ncol = ncol(x@data)
    ))
    colnames(new_header_row) <- colnames(x@data)
    new_header_row[1, ] <- new_row
    x@group_data_j <- rbind_nocol(new_header_row, x@group_data_j)
  }

  return(x)
}

#' Regenerate the alignment and underline styles of every column group
#'
#' The styles are derived entirely from `@group_data_j`, and the entries this
#' function appends are tagged with a `"tt_group_j"` attribute so that a later
#' call can discard and rebuild them. `subset()` relies on this: when columns
#' are removed, the spans are re-laid out and their styles must follow, and
#' patching the stored anchors and colspans one by one is not possible because
#' a span's position depends on the spans to its left.
#' @keywords internal
#' @noRd
style_group_j <- function(x) {
  is_tagged <- vapply(
    x@lazy_style,
    function(e) isTRUE(attr(e, "tt_group_j")),
    logical(1)
  )
  x@lazy_style <- x@lazy_style[!is_tagged]

  if (nrow(x@group_data_j) == 0) {
    return(x)
  }

  n_before <- length(x@lazy_style)
  n_groups <- nrow(x@group_data_j)

  # The last group row in @group_data_j is closest to the table body, so it
  # sits at i = -1; earlier rows go to -2, -3, etc.
  for (group_idx in n_groups:1) {
    table_row_i <- -(n_groups - group_idx + 1)

    # Center the whole header row
    x <- style_tt(x, i = table_row_i, align = "c")

    spans <- parse_group_spans(as.character(x@group_data_j[group_idx, ]))

    # Duplicate labels each get their own span, alignment, and underline
    for (idx in seq_len(nrow(spans))) {
      group_cols <- spans$start[idx]:spans$end[idx]

      if (length(group_cols) > 1) {
        x <- style_tt(
          x,
          i = table_row_i,
          j = group_cols[1],
          align = "c",
          colspan = length(group_cols)
        )
      } else {
        x <- style_tt(x, i = table_row_i, j = group_cols[1], align = "c")
      }

      # Determine trimming based on column positions
      trim_left <- min(group_cols) > 1
      trim_right <- max(group_cols) < ncol(x)

      # Build trim specification
      line_trim_spec <- ""
      if (trim_left && trim_right) {
        line_trim_spec <- "lr"
      } else if (trim_left) {
        line_trim_spec <- "l"
      } else if (trim_right) {
        line_trim_spec <- "r"
      } else {
        line_trim_spec <- NULL # No trimming
      }

      x <- style_tt(
        x,
        i = table_row_i,
        j = group_cols,
        line = "b",
        line_width = 0.03,
        line_trim = line_trim_spec
      )
    }
  }

  if (length(x@lazy_style) > n_before) {
    for (k in seq.int(n_before + 1L, length(x@lazy_style))) {
      attr(x@lazy_style[[k]], "tt_group_j") <- TRUE
    }
  }

  return(x)
}

#' Process delimiter-based column grouping
#' @keywords internal
#' @noRd
process_delimiter_grouping <- function(x, j) {
  if (any(grepl(j, x@names, fixed = TRUE))) {
    j_delim <- j_delim_to_named_list(x = x, j = j)
    x@names <- j_delim$colnames

    # Apply multiple levels of grouping if they exist (in reverse order)
    if (length(j_delim$groupnames) > 0) {
      for (level_groups in rev(j_delim$groupnames)) {
        x <- process_group_j(x, level_groups)
      }
    }
    j <- NULL # Set to NULL since we've already applied the groupings
  } else {
    j <- NULL
  }
  return(list(x = x, j = j))
}

# =============================================================================
# MAIN FUNCTION
# =============================================================================

#' Spanning labels to identify groups of rows or columns
#'
#' @export
#' @inheritParams tt
#' @inheritParams style_tt
#' @param i Character vector, named list, or integer vector
#' + A character vector of labels with length equal to the number of rows in `x`
#' + A named list of row indices to group. The names of the list will be used as labels. The indices represent the position where labels should be inserted in the original table. For example,
#'   - `i=list("Hello"=5)`: insert the "Hello" label after the 4th row in the original table.
#'   - `i=list("Hello"=2, "World"=2)`: insert the two labels consecutively after the 1st row in the original table.
#'   - `i=list("Foo Bar"=0)`: insert the label in the first row after the header.
#' + Vector of positive integers: For matrix insertion: `i` specifies row positions and `j` must be a character matrix to insert in the table (see below for details).
#' @param j String, named list, or character matrix
#' - Named list of column indices to group, ex: `j=list("A"=1:2,"B"=3:6)`. The names of the list will be used as labels. See below for more examples. Note: empty labels must be a space: " ".
#' - A single string when column names include the group name as a prefix, ex: group1_column1, group1_column2, etc.
#' - Character matrix for inserting rows at positions specified by `i`. The matrix must have the same number of columns as the table, or be a single column with a number of elements that is a multiple of the table's column count (which will be automatically reshaped). Each row of the matrix matches an element
#' - Unquoted expression: Non-standard evaluation is supported. When supplying an unquoted expression, it is first evaluated in the calling environment, then in the data frame passed to `tt()`.
#' @param ... Other arguments are ignored.
#' @return An object of class `tt` representing the table.
#' @template limitations_word_markdown
#' @details
#' Warning: The `style_tt()` can normally be used to style the group headers, as expected, but that feature is not available for Markdown and Word tables.
#' @examples
#'
#' # vector of row labels
#' dat <- data.frame(
#'   label = c("a", "a", "a", "b", "b", "c", "a", "a"),
#'   x1 = rnorm(8),
#'   x2 = rnorm(8)
#' )
#' tt(dat[, 2:3]) |> group_tt(i = dat$label)
#'
#' # named lists of labels
#' tt(mtcars[1:10, 1:5]) |>
#'   group_tt(
#'     i = list(
#'       "Hello" = 3,
#'       "World" = 8
#'     ),
#'     j = list(
#'       "Foo" = 2:3,
#'       "Bar" = 4:5
#'     )
#'   )
#'
#' dat <- mtcars[1:9, 1:8]
#' tt(dat) |>
#'   group_tt(i = list(
#'     "I like (fake) hamburgers" = 3,
#'     "She prefers halloumi" = 4,
#'     "They love tofu" = 7
#'   ))
#'
#' tt(dat) |>
#'   group_tt(
#'     j = list(
#'       "Hamburgers" = 1:3,
#'       "Halloumi" = 4:5,
#'       "Tofu" = 7
#'     )
#'   )
#'
#' x <- mtcars[1:5, 1:6]
#' tt(x) |>
#'   group_tt(j = list("Hello" = 1:2, "World" = 3:4, "Hello" = 5:6)) |>
#'   group_tt(j = list("Foo" = 1:3, "Bar" = 4:6))
#'
#' # column names with delimiters
#' dat <- data.frame(
#'   A_id = 1,
#'   A_a1 = 2,
#'   A_a2 = "3",
#'   B_b1 = 4,
#'   B_b2 = 5,
#'   B_C = 6
#' )
#' tt(dat) |> group_tt(j = "_")
#'
#' # matrix insertion
#' rowmat <- matrix(colnames(iris))
#' tt(head(iris, 7)) |>
#'   group_tt(i = c(2, 5), j = rowmat)
#'
#' rowmat <- matrix(c(
#'   "a", "b", "c", "d", "e",
#'   1, 2, 3, 4, 5))
#' tt(head(iris, 7)) |>
#'   group_tt(i = 2, j = rowmat) |>
#'   style_tt(i = "groupi", background = "pink")
group_tt <- function(
    x,
    i = getOption("tinytable_group_i", default = NULL),
    j = getOption("tinytable_group_j", default = NULL),
    ...) {
  handle_deprecated_args(...)
  if (!inherits(x, "tinytable")) {
    stop("`x` must be generated by `tinytable::tt()`.", call. = FALSE)
  }

  # non-standard evaluation before anything else
  tmp <- nse_i_j(x,
    i_expr = substitute(i),
    j_expr = substitute(j),
    pf = parent.frame())
  i <- tmp$i

  if (is.null(i) && is.null(j)) {
    stop("At least one of `i` or `j` must be specified.", call. = FALSE)
  }

  # Atomic vector `i` supplies one label per row: convert to named-list form.
  # This works regardless of `j`, so row labels can be combined with column
  # spans in a single call. Character and factor vectors are always labels;
  # other vectors (e.g., numeric columns selected via NSE) are treated as
  # labels only when they supply one label per row.
  if (!is.null(i) && !is.list(i) && isTRUE(check_atomic_vector(i)) && !isTRUE(check_matrix(j))) {
    if (is.character(i) || is.factor(i) || length(i) == nrow(x)) {
      if (anyNA(i)) {
        stop("`i` must not contain missing values (`NA` group labels).", call. = FALSE)
      }
      i <- sanitize_group_vec2list(i)
    } else {
      stop(
        "`i` must be a vector of row labels (one per row), a named list of positions, or a vector of row positions combined with a character matrix `j`.",
        call. = FALSE
      )
    }
  }

  # matrix insertion case
  if (isTRUE(check_integerish(i)) && isTRUE(check_matrix(j))) {
    return(process_group_i(x, i, j))
  }

  # row grouping when i is (or has been converted to) a named list
  if (is.list(i)) {
    x <- process_group_i(x, i, NULL)
  }

  # delimiter-based column grouping
  grouped_j <- FALSE
  if (isTRUE(check_string(j))) {
    result <- process_delimiter_grouping(x, j)
    grouped_j <- !identical(result$x@group_data_j, x@group_data_j)
    x <- result$x
    j <- result$j
  }

  # column grouping
  if (!is.null(j)) {
    x <- process_group_j(x, j)
    grouped_j <- TRUE
  }

  if (grouped_j) {
    x <- style_group_j(x)
  }

  return(x)
}

#' Subsetting a `tinytable` object
#'
#' Return subsets `tinytable` which meet conditions.
#' @inheritParams base::subset.data.frame
#' @export
subset.tinytable <- function(x, subset, select, drop = FALSE, ...) {
  chkDots(...)

  # `drop = TRUE` would return vectors, which cannot be assigned to the
  # data.frame S4 slots. Warn and ignore.
  if (!isFALSE(drop)) {
    warning(
      "`drop = TRUE` is not supported for tinytable objects; ignoring.",
      call. = FALSE
    )
  }

  # Evaluate subset condition
  r <- rep_len(TRUE, nrow(x@data_body))
  if (!missing(subset)) {
    warning("The `subset` argument of the `subset()` function is not supported for `tinytable` objects. Filter the rows before supplying the data frame to `tt()`.", call. = FALSE)
  }

  # Evaluate select condition
  old_ncol <- ncol(x@data_body)
  vars <- if (missing(select)) {
    rep_len(TRUE, old_ncol)
  } else {
    nl <- as.list(seq_along(x@data_body))
    names(nl) <- names(x@data_body)
    eval(substitute(select), nl, parent.frame())
  }

  # Resolve `vars` to positive column positions. Character selections must be
  # matched by name: `seq_len(old_ncol)[vars]` would silently yield NA for
  # character input, corrupting the column selection below.
  kept <- if (is.logical(vars)) {
    which(vars)
  } else if (is.character(vars)) {
    match(vars, names(x@data_body))
  } else {
    seq_len(old_ncol)[vars]
  }
  if (anyNA(kept)) {
    bad <- if (is.character(vars)) {
      vars[!vars %in% names(x@data_body)]
    } else {
      vars[is.na(seq_len(old_ncol)[vars])]
    }
    stop(
      sprintf(
        "Invalid `select`: cannot match column(s): %s",
        paste(bad, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  new_ncol <- length(kept)

  # Apply subsetting to data
  x@data_body <- x@data_body[r, kept, drop = FALSE]
  x@data <- x@data[r, kept, drop = FALSE]

  # Subset row group data
  if (nrow(x@group_data_i) > 0) {
    x@group_data_i <- x@group_data_i[, kept, drop = FALSE]
  }

  # Truncate the column group labels. Like `i` in `style_tt()`, span positions
  # refer to the rendered table rather than to the input data, so a span stays
  # where it was and is only clipped against the new right edge. Spans that
  # start beyond the last surviving column disappear.
  if (nrow(x@group_data_j) > 0) {
    gdj <- x@group_data_j
    new_gdj <- data.frame(
      matrix(NA_character_, nrow = nrow(gdj), ncol = new_ncol),
      stringsAsFactors = FALSE
    )
    for (row_idx in seq_len(nrow(gdj))) {
      spans <- parse_group_spans(as.character(gdj[row_idx, ]))
      new_row <- rep(NA_character_, new_ncol)
      for (span_idx in seq_len(nrow(spans))) {
        start <- spans$start[span_idx]
        if (start > new_ncol) {
          next
        }
        end <- min(spans$end[span_idx], new_ncol)
        new_row[start] <- spans$label[span_idx]
        if (end > start) {
          new_row[seq.int(start + 1L, end)] <- ""
        }
      }
      new_gdj[row_idx, ] <- new_row
    }
    colnames(new_gdj) <- colnames(x@data_body)
    x@group_data_j <- new_gdj
  }

  # Update dimensions and names
  x@nrow <- nrow(x@data_body) + nrow(x@group_data_i)
  x@ncol <- new_ncol
  x@names <- names(x@data_body)

  # Update width_cols to match new column count
  if (length(x@width_cols) > 0) {
    x@width_cols <- x@width_cols[kept]
  }

  # Numeric `j` in a lazy call refers to a position in the rendered table, the
  # same convention `i` follows, so removing a column does not shift the stored
  # indices. Entries only need to be clipped to the narrower table.
  x@lazy_format <- clip_lazy_list(x@lazy_format, new_ncol)
  x@lazy_plot <- clip_lazy_list(x@lazy_plot, new_ncol)
  x@lazy_style <- clip_lazy_list(x@lazy_style, new_ncol)

  # Column group styles are derived from @group_data_j, which was just
  # truncated: rebuild them rather than clip them.
  x <- style_group_j(x)

  # Avoid colspan that exceeds the new number of columns
  x@style <- clamp_colspan(x@style, x@ncol)

  return(x)
}


#' Clip j-indices in a list of lazy calls after column subsetting
#'
#' Entries that no longer target any column are dropped entirely: setting
#' their `j` to NULL instead would make them apply to every remaining column.
#' @keywords internal
#' @noRd
clip_lazy_list <- function(lst, new_ncol) {
  out <- lapply(lst, clip_j, new_ncol = new_ncol)
  out[!vapply(out, is.null, logical(1))]
}


#' Clip the j-indices of a lazy call to the width of a subsetted table
#'
#' Returns the (possibly modified) lazy entry, or NULL if every column the
#' entry targeted now falls outside the table. Only plain positive numeric `j`
#' values are clipped; anything else (NULL, negative selections, character
#' column names, NSE markers, ...) is deliberately left untouched so we never
#' mangle entries we cannot interpret.
#' @keywords internal
#' @noRd
clip_j <- function(call_args, new_ncol) {
  if (is.call(call_args)) {
    j <- call_args[["j"]]
  } else if (is.list(call_args)) {
    j <- call_args$j
  } else {
    return(call_args)
  }

  if (!is.numeric(j) || length(j) == 0L || anyNA(j) || any(j <= 0)) {
    return(call_args)
  }

  new_j <- j[j <= new_ncol]
  if (length(new_j) == 0L) {
    return(NULL)
  }

  colspan <- if (is.call(call_args)) {
    call_args[["colspan"]]
  } else {
    call_args$colspan
  }

  if (is.numeric(colspan) && length(colspan) == 1L && !is.na(colspan)) {
    # Shrink the span so it stops at the last surviving column. `colspan` must
    # be >= 2: a span reduced to a single column is a plain cell, so drop the
    # argument instead of setting it to 1.
    new_colspan <- min(colspan, new_ncol - min(new_j) + 1L)
    if (new_colspan < 2L) {
      new_colspan <- NULL
    }
    if (is.call(call_args)) {
      call_args[["colspan"]] <- new_colspan
    } else {
      call_args$colspan <- new_colspan
    }
  }

  if (is.call(call_args)) {
    call_args[["j"]] <- new_j
  } else {
    call_args$j <- new_j
  }
  call_args
}

# =============================================================================
# TABULATOR CDN CONSTANTS
#
# Single source of truth for every CDN URL and the default CSS <link> block
# used by the tabulator engine. `tabulator_stylesheet()` swaps stylesheets by
# exact-match replacement of `tabulator_default_css_block`, so these must not
# be redefined elsewhere: drift between copies would silently break that
# replacement. The Tabulator version is also referenced in
# inst/templates/tabulator.html.
# =============================================================================
tabulator_js_cdn <- "https://cdn.jsdelivr.net/npm/tabulator-tables@6.3/dist/js/tabulator.min.js"
tabulator_cdn_base <- "https://cdn.jsdelivr.net/npm/tabulator-tables@6.3/dist/css/"
tabulator_css_cdn <- paste0(tabulator_cdn_base, "tabulator.min.css")
tinytable_tabulator_theme_cdn <- "https://cdn.jsdelivr.net/gh/vincentarelbundock/tinytable@main/inst/tabulator_tinytable.min.css"
tinytable_js_cdn <- "https://cdn.jsdelivr.net/gh/vincentarelbundock/tinytable@main/inst/tinytable.js"
fontawesome_css_cdn <- "https://cdn.jsdelivr.net/npm/@fortawesome/fontawesome-free@6.5.2/css/all.min.css"

tabulator_default_css_block <- sprintf(
  '<link href="%s" rel="stylesheet">\n    <link href="%s" rel="stylesheet">\n    <link href="%s" rel="stylesheet">',
  tabulator_css_cdn,
  tinytable_tabulator_theme_cdn,
  fontawesome_css_cdn
)

setMethod(
  f = "build_eval",
  signature = "tinytable_tabulator",
  definition = function(x, ...) {
    # Check that column names exist
    if (is.null(x@names) || length(x@names) == 0) {
      stop(
        "Column names are required for tabulator tables. Use `colnames(x) <- ...` to set column names.",
        call. = FALSE
      )
    }

    # Check that no format_tt() calls use the i argument (row-based formatting)
    # Tabulator only supports column-based or full-table formatting
    for (l in x@lazy_format) {
      if (!is.null(l$i)) {
        stop(
          "Row-based formatting (i argument) is not supported with tabulator output. ",
          "Use column-based formatting (j argument) or full-table formatting instead.",
          call. = FALSE
        )
      }
    }

    template <- readLines(
      system.file("templates/tabulator.html", package = "tinytable")
    )

    # (pseudo-)unique table IDs
    id <- get_id("")
    x@id <- id

    # Replace table ID in template
    template <- gsub(
      "$tinytable_TABLE_ID",
      paste0("tinytable_", id),
      template,
      fixed = TRUE
    )

    # Convert data to JSON for Tabulator
    data_clean <- tabulator_clean_data(x)
    js_data <- df_to_json(data_clean)

    # Build basic column definitions only if not already set by theme
    if (length(x@tabulator_columns) == 0) {
      # Clean the full vector at once so de-duplication is consistent
      fields <- tabulator_clean_column_name(x@names)
      columns <- lapply(seq_along(x@names), function(i) {
        list(
          title = x@names[i],
          field = fields[i]
        )
      })

      # Store columns list for later processing
      x@tabulator_columns <- columns
    }

    # Replace data in template
    template <- gsub(
      "$tinytable_TABULATOR_DATA",
      js_data,
      template,
      fixed = TRUE
    )

    # Leave columns placeholder for finalize_tabulator.R to handle
    # (Custom columns and formatting are handled there)

    # Add default CDN (base Tabulator CSS + tinytable theme)
    template <- gsub(
      "$tinytable_TABULATOR_CDN",
      tabulator_default_css_block,
      template,
      fixed = TRUE
    )

    # Leave CSS placeholder for finalize to handle

    # Leave options placeholder for potential theme replacement
    # (it will be cleaned up in finalize if not replaced)

    out <- paste(template, collapse = "\n")

    # Store table string for potential styling later
    x@table_string <- out

    return(x)
  })

# =============================================================================
# TABULATOR SEARCH FUNCTIONALITY
# =============================================================================

clean_search_placeholders <- function(x, top = "", bottom = "", listener = "") {
  x@table_string <- sub("$tinytable_TABULATOR_SEARCH_TOP", top, x@table_string, fixed = TRUE)
  x@table_string <- sub("$tinytable_TABULATOR_SEARCH_BOTTOM", bottom, x@table_string, fixed = TRUE)
  x@table_string <- sub("$tinytable_TABULATOR_SEARCH_LISTENER", listener, x@table_string, fixed = TRUE)
  x
}

#' Create search listener JavaScript
#' @param search_id Search ID
#' @param columns_json Columns JSON string
#' @return Search listener JavaScript
#' @keywords internal
#' @noRd
tabulator_search_listener <- function(search_id, columns_json) {
  # Simple string replacement approach
  js <- '
        // Expose table reference globally for search functionality
        window.table_tinytable_SEARCH_ID = table_tinytable_SEARCH_ID;

        // Search functionality (runs within the table IIFE scope)
        const columns_SEARCH_ID = COLUMNS_JSON;
        const searchFields_SEARCH_ID = columns_SEARCH_ID.map(col => col.field);

        // Attach search listener using the globally exposed table reference
        const searchElement = document.getElementById("search_SEARCH_ID");
        if (searchElement) {
          searchElement.addEventListener("input", function () {
            const term = this.value.trim();
            if (!term) {
              window.table_tinytable_SEARCH_ID.clearFilter();
            } else {
              window.table_tinytable_SEARCH_ID.setFilter(function(data) {
                return searchFields_SEARCH_ID.some(field => {
                  const value = data[field];
                  if (value === null || value === undefined) return false;
                  return String(value).toLowerCase().includes(term.toLowerCase());
                });
              });
            }
          });
        }
    '

  # Replace placeholders (escape backslashes: gsub replacements are not literal)
  js <- gsub("SEARCH_ID", search_id, js, fixed = TRUE)
  columns_json <- gsub("\\", "\\\\", columns_json, fixed = TRUE)
  js <- gsub("COLUMNS_JSON", columns_json, js, fixed = TRUE)

  return(js)
}

#' Apply column header filters
#' @param x tinytable object
#' @return Modified tinytable object
#' @keywords internal
#' @noRd
tabulator_apply_column_search <- function(x) {
  x <- clean_search_placeholders(x)

  # Column definitions supplied as a JSON string cannot be augmented with
  # header filters (we would have to parse arbitrary JavaScript)
  if (is.list(x@tabulator_columns) && !is.null(x@tabulator_columns$json_string)) {
    stop(
      "`tabulator_search = 'column'` is not supported when `tabulator_columns` is supplied as a JSON string. ",
      "Provide `tabulator_columns` as a list of column definitions instead.",
      call. = FALSE
    )
  }

  # Add headerFilter to each column definition based on data type
  if (length(x@tabulator_columns) > 0) {
    for (i in seq_along(x@tabulator_columns)) {
      col_title <- x@tabulator_columns[[i]][["title"]]

      # Find the column in the original data to check its type
      col_idx <- which(x@names == col_title)

      if (length(col_idx) == 1 && col_idx <= ncol(x@data)) {
        col_data <- x@data[[col_idx]]

        # Determine appropriate filter type based on column data type
        if (is.numeric(col_data)) {
          # Use number filter for numeric columns with min/max operators
          x@tabulator_columns[[i]][["headerFilter"]] <- "number"
          x@tabulator_columns[[i]][["headerFilterPlaceholder"]] <- ""
          x@tabulator_columns[[i]][["headerFilterFunc"]] <- ">="
        } else if (inherits(col_data, c("Date", "POSIXct", "POSIXlt"))) {
          # Use input filter for dates (could be enhanced with date picker)
          x@tabulator_columns[[i]][["headerFilter"]] <- "input"
          x@tabulator_columns[[i]][["headerFilterPlaceholder"]] <- ""
        } else if (is.logical(col_data)) {
          # Use select filter for logical columns
          x@tabulator_columns[[i]][["headerFilter"]] <- "tickCross"
          x@tabulator_columns[[i]][["headerFilterParams"]] <- c(
            tabulator_tickcross_params(),
            list(tristate = TRUE)
          )
          x@tabulator_columns[[i]][["headerFilterPlaceholder"]] <- "All"
        } else {
          # Use input filter for text columns
          x@tabulator_columns[[i]][["headerFilter"]] <- "input"
          x@tabulator_columns[[i]][["headerFilterPlaceholder"]] <- ""
        }
      } else {
        # Default to input filter if we can't determine the type
        x@tabulator_columns[[i]][["headerFilter"]] <- "input"
        x@tabulator_columns[[i]][["headerFilterPlaceholder"]] <- ""
      }
    }
  }

  # Re-serialize columns to JSON after adding search filters
  columns_json <- df_to_json(x@tabulator_columns)
  x@table_string <- tabulator_replace_columns_json(x@table_string, columns_json)

  return(x)
}

#' Apply search functionality to table
#' @param x tinytable object
#' @return Modified tinytable object
#' @keywords internal
#' @noRd
tabulator_apply_search <- function(x) {
  if (is.null(x@tabulator_search)) {
    x <- clean_search_placeholders(x)
    return(x)
  }

  # Handle column header filters
  if (x@tabulator_search == "column") {
    x <- tabulator_apply_column_search(x)
    return(x)
  }

  # Create unique search ID
  search_id <- gsub(".*tinytable_", "", x@id)

  # Get columns JSON directly from the S4 object
  if (length(x@tabulator_columns) == 0) {
    x <- clean_search_placeholders(x)
    return(x)
  }

  # Column definitions supplied as a JSON string are a JavaScript array
  # already: use them verbatim so the listener can map over `col.field`
  if (!is.null(x@tabulator_columns$json_string)) {
    columns_json <- x@tabulator_columns$json_string
  } else {
    columns_json <- df_to_json(x@tabulator_columns)
  }

  # Determine search position
  search_position <- x@tabulator_search

  # Create search bar HTML
  search_bar_template <- '
        <div class="mb-3"><input type="text" id="search_%s" class="form-control" placeholder="Search table..." style="margin-bottom: 10px;"></div>
    '
  search_bar_html <- sprintf(search_bar_template, search_id)

  # Create search listener JS
  search_listener_js <- tabulator_search_listener(search_id, columns_json)

  # Replace search bar placeholder based on position
  if (search_position == "top") {
    x <- clean_search_placeholders(x, top = search_bar_html, listener = search_listener_js)
  } else if (search_position == "bottom") {
    x <- clean_search_placeholders(x, bottom = search_bar_html, listener = search_listener_js)
  }

  return(x)
}

swap_class <- function(x, new_class) {
  class(x) <- new_class
  return(x)
}

setClassUnion("NULLorCharacter", c("NULL", "character"))
setClassUnion("NULLorNumeric", c("NULL", "numeric"))
setClassUnion("NULLorList", c("NULL", "list"))

#' tinytable S4 class
#'
#' @keywords internal
#' @export
setClass(
  Class = "tinytable",
  slots = representation(
    ansi = "logical",
    body = "character",
    html_class = "character",
    html_css_rule = "NULLorCharacter",
    html_script = "NULLorCharacter",
    caption = "NULLorCharacter",
    data = "data.frame",
    data_body = "data.frame",
    grid_hline = "logical",
    grid_hline_header = "logical",
    grid_vline = "logical",
    group_data_i = "data.frame",
    group_data_j = "data.frame",
    group_index_i = "numeric",
    height = "NULLorNumeric",
    id = "character",
    index_body = "numeric",
    lazy_finalize = "list",
    lazy_format = "list",
    lazy_plot = "list",
    lazy_prepare = "list",
    lazy_style = "list",
    markdown_style = "character",
    names = "NULLorCharacter",
    ncol = "numeric",
    nhead = "numeric",
    notes = "NULLorList",
    nrow = "numeric",
    output = "character",
    output_dir = "character",
    placement = "NULLorCharacter",
    html_portable = "logical",
    typst_portable = "logical",
    html_engine = "character",
    latex_preamble = "logical",
    latex_engine = "character",
    style = "data.frame",
    style_other = "data.frame",
    style_lines = "data.frame",
    style_caption = "list",
    style_notes = "list",
    table_string = "character",
    tabularray_inner = "character",
    tabularray_outer = "character",
    tabulator_format_bool = "logical",
    tabulator_column_formatters = "list",
    tabulator_column_styles = "list",
    tabulator_columns = "list",
    tabulator_css_rule = "character",
    tabulator_options = "character",
    tabulator_post_init = "character",
    tabulator_search = "NULLorCharacter",
    tabulator_stylesheet = "character",
    theme = "list",
    width = "NULLorNumeric",
    width_cols = "numeric"
  )
)

#' Method for a tinytable S4 object
#'
#' @inheritParams tt
#' @keywords internal
setMethod(
  "initialize",
  "tinytable",
  function(.Object,
           data = data.frame(),
           caption = NULL,
           notes = NULL,
           theme = list("default"),
           data_body = data.frame(),
           placement = NULL,
           width = NULL,
           height = NULL,
           colnames = TRUE) {
    # explicit
    .Object@data <- data
    .Object@data_body <- data_body
    .Object@theme <- theme
    .Object@placement <- placement
    .Object@caption <- caption
    .Object@width <- width
    .Object@notes <- notes
    .Object@height <- height
    # Generate unique ID first
    .Object@id <- get_id("tinytable_")

    # Default to NULL - framework CSS will be loaded externally
    .Object@html_css_rule <- NULL
    .Object@html_script <- NULL

    # dynamic
    .Object@nrow <- nrow(.Object@data)
    .Object@ncol <- ncol(.Object@data)
    .Object@nhead <- if (is.null(colnames(data)) || isFALSE(colnames)) 0 else 1

    # colnames
    if (isTRUE(colnames)) {
      .Object@names <- colnames(data)
    } else if (isFALSE(colnames) || is.null(colnames(data))) {
      .Object@names <- character()
    } else if (identical(colnames, "label")) {
      cols <- character()
      for (cn in colnames(data)) {
        lab <- attr(data[[cn]], "label")
        if (!is.null(lab)) {
          cols <- c(cols, lab)
        } else {
          cols <- c(cols, cn)
        }
      }
      .Object@names <- cols
    }

    # empty
    .Object@ansi <- FALSE
    .Object@grid_hline <- TRUE
    .Object@grid_hline_header <- TRUE
    .Object@grid_vline <- TRUE
    .Object@group_data_i <- data.frame()
    .Object@group_data_j <- data.frame()
    .Object@group_index_i <- numeric(0)
    .Object@html_class <- "tinytable"
    .Object@html_engine <- "tinytable"
    .Object@html_portable <- FALSE
    .Object@typst_portable <- FALSE
    .Object@index_body <- numeric(0)
    .Object@latex_engine <- "xelatex"
    .Object@latex_preamble <- TRUE
    .Object@lazy_finalize <- list()
    .Object@lazy_format <- list()
    .Object@lazy_plot <- list()
    .Object@lazy_prepare <- list()
    .Object@markdown_style <- "grid"
    .Object@output <- "tinytable"
    .Object@output_dir <- getwd()
    .Object@style <- data.frame()
    .Object@style_other <- data.frame()
    .Object@style_lines <- data.frame()
    .Object@tabularray_inner <- character()
    .Object@tabularray_outer <- character()
    .Object@tabulator_column_formatters <- list()
    .Object@tabulator_column_styles <- list()
    .Object@tabulator_columns <- list()
    .Object@tabulator_css_rule <- ""
    .Object@tabulator_format_bool <- FALSE
    .Object@tabulator_options <- ""
    .Object@tabulator_post_init <- ""
    .Object@tabulator_search <- NULL
    .Object@tabulator_stylesheet <- ""

    return(.Object)
  })

#' Method for a tinytable S4 object
#'
#' @inheritParams tt
#' @keywords internal
setMethod("nrow", "tinytable", function(x) {
  return(x@nrow)
})

#' Method for a tinytable S4 object
#'
#' @inheritParams tt
#' @keywords internal
setMethod("ncol", "tinytable", function(x) {
  return(x@ncol)
})

# shared implementation for the colnames() and names() getters
ttnames <- function(x) {
  return(x@names)
}

# shared implementation for the colnames<-() and names<-() setters
# @nhead counts the column name row (0 or 1) plus one row per group_tt(j=...)
# call, so we adjust it by 1 rather than overwrite it.
`ttnames<-` <- function(x, value) {
  # Issue #306
  if (length(value) == 0) {
    value <- NULL
  }
  if (!is.null(value)) {
    assert_character(value, len = x@ncol)
    # restore the header row when the table previously had no column names
    if (length(x@names) == 0) {
      x@nhead <- x@nhead + 1
    }
  } else {
    # drop the header row when removing existing column names
    if (length(x@names) > 0 && x@nhead > 0) {
      x@nhead <- x@nhead - 1
    }
  }
  x@names <- value
  return(x)
}

#' Method for a tinytable S4 object
#'
#' @inheritParams tt
#' @keywords internal
#' @export
setMethod("colnames", "tinytable", ttnames)

#' Method for a tinytable S4 object
#'
#' @inheritParams tt
#' @keywords internal
#' @export
setMethod("names", "tinytable", ttnames)

#' Method for a tinytable S4 object
#'
#' @inheritParams tt
#' @keywords internal
#' @export
setReplaceMethod(
  "colnames",
  signature = "tinytable",
  definition = `ttnames<-`)

#' Method for a tinytable S4 object
#'
#' @inheritParams tt
#' @keywords internal
#' @export
setReplaceMethod(
  "names",
  signature = "tinytable",
  definition = `ttnames<-`)

#' Dimensions a tinytable S4 object
#'
#' @inheritParams tt
#' @keywords internal
setMethod("dim", "tinytable", function(x) {
  return(c(x@nrow, x@ncol))
})

#' Convert a tinytable S4 object to a string
#'
#' @inheritParams tt
#' @keywords internal
setMethod("as.character", "tinytable", function(x) {
  # fresh tables have @output set to "tinytable", which save_tt() rejects, so
  # we infer the output format the same way print() does
  save_tt(x, infer_output(x))
})

setClass("tinytable_tabularray", contains = "tinytable")
setClass("tinytable_html", contains = "tinytable")
setClass("tinytable_typst", contains = "tinytable")
setClass("tinytable_grid", contains = "tinytable")
setClass("tinytable_dataframe", contains = "tinytable")
setClass("tinytable_tabulator", contains = "tinytable")

#' Apply style settings to a tinytable
#'
#' @inheritParams tt
#' @keywords internal
setGeneric(
  name = "style_eval",
  def = function(x, ...) standardGeneric("style_eval")
)

#' Apply group settings to a tinytable
#'
#' @inheritParams tt
#' @keywords internal
setGeneric(
  name = "build_eval",
  def = function(x, ...) standardGeneric("build_eval")
)

#' Apply group settings to a tinytable
#'
#' @inheritParams tt
#' @keywords internal
setGeneric(
  name = "group_eval_j",
  def = function(x, ...) standardGeneric("group_eval_j")
)


#' Apply final settings to a tinytable
#'
#' @inheritParams tt
#' @keywords internal
setGeneric(
  name = "finalize",
  def = function(x, ...) standardGeneric("finalize")
)

sanity_tabulator_css_rule <- function(tabulator_css_rule) {
  if (!is.null(tabulator_css_rule)) {
    if (!is.character(tabulator_css_rule) || length(tabulator_css_rule) != 1) {
      stop(
        "tabulator_css_rule must be a single character string",
        call. = FALSE
      )
    }

    if (!grepl("\\$TINYTABLE_ID", tabulator_css_rule)) {
      stop(
        "tabulator_css_rule must contain '$TINYTABLE_ID' placeholder for table scoping.\n",
        "Example: tabulator_css_rule = '$TINYTABLE_ID .tabulator-col { background: red; }'",
        call. = FALSE
      )
    }
  }
}

sanity_tabulator_columns <- function(tabulator_columns) {
  if (!is.null(tabulator_columns)) {
    if (!is.character(tabulator_columns) && !is.list(tabulator_columns)) {
      stop(
        "tabulator_columns must be a character string containing valid JavaScript array or a list of column definitions",
        call. = FALSE
      )
    }
    if (is.character(tabulator_columns) && length(tabulator_columns) != 1) {
      stop(
        "tabulator_columns character string must be a single element",
        call. = FALSE
      )
    }
  }
}

# Returns a list with:
# - options: the pagination options string for the Tabulator constructor
# - sizes: the resolved numeric page-size vector (first element is the active
#   page size), or NULL when pagination is disabled/unset. The layout helper
#   uses sizes[1] to compute the table height from *visible* rows.
tabulator_pagination_options <- function(tabulator_pagination, nrow) {
  # NULL
  pagination_opts <- ""

  if (isFALSE(tabulator_pagination)) {
    pagination_opts <- "pagination: false,"
  }

  if (isTRUE(tabulator_pagination)) {
    # Create pagination options: 10, 25, 50, 100, 250 (filtered by available rows)
    tabulator_pagination <- c(10, 25, 50, 100, 250)
    tabulator_pagination <- tabulator_pagination[
      tabulator_pagination <= nrow
    ]

    # If there are no options above nrow, add nrow as an option
    if (length(tabulator_pagination) > 0 && max(tabulator_pagination) < nrow) {
      tabulator_pagination <- c(tabulator_pagination, nrow)
    }

    # If no pagination options remain, use the actual number of rows
    if (length(tabulator_pagination) == 0) {
      tabulator_pagination <- nrow
    }
  }

  if (is.numeric(tabulator_pagination)) {
    # Vector of integers: first is size, sorted vector is selector
    paginationSize <- tabulator_pagination[1]

    # If the number of rows is smaller than the pagination size, adjust pagination size
    if (nrow <= paginationSize) {
      # Use the actual number of rows as pagination size
      paginationSize <- nrow
      tabulator_pagination <- paginationSize
    }

    # Now set up pagination options
    if (length(tabulator_pagination) > 1) {
      # Multiple pagination options: include selector
      paginationSizeSelector <- sort(tabulator_pagination)
      selector_str <- paste0(
        "[",
        paste(paginationSizeSelector, collapse = ", "),
        "]"
      )
      pagination_opts <- sprintf(
        "
        pagination: 'local',
        paginationSizeSelector: %s,
        paginationSize: %s,",
        selector_str,
        paginationSize
      )
    } else {
      # Single pagination option: no selector
      pagination_opts <- sprintf(
        "
        pagination: 'local',
        paginationSize: %s,",
        paginationSize
      )
    }
  }

  return(list(
    options = pagination_opts,
    sizes = if (is.numeric(tabulator_pagination)) tabulator_pagination else NULL
  ))
}

tabulator_layout_options <- function(
    tabulator_layout,
    x,
    page_sizes) {
  # Build layout options
  if (!is.null(x@height)) {
    # Calculate total height: height per row * number of visible rows + header space
    if (is.numeric(page_sizes) && length(page_sizes) > 0) {
      # Use pagination size for number of visible rows
      visible_rows <- page_sizes[1]
    } else {
      # Use actual number of rows when no pagination
      visible_rows <- nrow(x)
    }
    total_height <- (x@height * visible_rows) + 2.5 # 2.5em for header
    layout_opts <- sprintf(
      "    layout: '%s',\n    height: '%sem'",
      tabulator_layout,
      format_markup_num(total_height)
    )
  } else {
    layout_opts <- sprintf("  layout: '%s'", tabulator_layout)
  }

  return(layout_opts)
}

theme_html_tabulator <- function(
    x,
    tabulator_stylesheet,
    tabulator_layout,
    tabulator_pagination,
    tabulator_search,
    tabulator_options,
    tabulator_css_rule,
    tabulator_columns,
    ...) {
  assert_choice(
    tabulator_layout,
    choice = c(
      "fitDataTable",
      "fitData",
      "fitDataFill",
      "fitDataStretch",
      "fitColumns"
    )
  )

  if (identical(tabulator_search, "column") && is.character(tabulator_columns)) {
    stop(
      "`tabulator_search = 'column'` is not supported when `tabulator_columns` is supplied as a JSON string. ",
      "Provide `tabulator_columns` as a list of column definitions instead.",
      call. = FALSE
    )
  }

  if (!is.null(tabulator_search)) {
    x@tabulator_search <- tabulator_search
  }

  pagination <- tabulator_pagination_options(tabulator_pagination, nrow(x))
  pagination_opts <- pagination$options
  layout_opts <- tabulator_layout_options(
    tabulator_layout,
    x,
    pagination$sizes
  )

  # Build options string based on whether pagination options exist
  if (nchar(pagination_opts) > 0) {
    opts <- sprintf(
      "    %s
      %s
      ",
      pagination_opts,
      layout_opts
    )
  } else {
    opts <- sprintf(
      "
      %s
      ",
      layout_opts
    )
  }

  # If tabulator_options is provided, use it instead of the individual arguments
  if (!is.null(tabulator_options) && nchar(tabulator_options) > 0) {
    opts <- tabulator_options
  }

  # Only apply to tabulator engine with html output
  tabulator_theme_fn <- function(table) {
    if (!isTRUE(table@output == "html" && table@html_engine == "tabulator")) {
      return(table)
    }

    # Store stylesheet, options, search, css_rule, and columns in S4 slots
    if (!is.null(tabulator_stylesheet)) {
      table@tabulator_stylesheet <- tabulator_stylesheet
    }
    table@tabulator_options <- opts
    if (!is.null(tabulator_css_rule)) {
      # Append CSS rules instead of overwriting
      if (nchar(table@tabulator_css_rule) > 0) {
        table@tabulator_css_rule <- paste(
          table@tabulator_css_rule,
          tabulator_css_rule,
          sep = "\n"
        )
      } else {
        table@tabulator_css_rule <- tabulator_css_rule
      }
    }
    if (!is.null(tabulator_columns)) {
      # Handle both character strings (backward compatibility) and lists
      if (is.character(tabulator_columns)) {
        # Keep as character for now - will be handled in finalize_tabulator.R
        # But we need to store it in a way that doesn't break S4 validation
        # Convert to a temporary list structure that indicates it's a JSON string
        table@tabulator_columns <- list(json_string = tabulator_columns)
      } else {
        # Mark list-form definitions as user-supplied so that finalize
        # serializes them verbatim instead of regenerating column
        # definitions from the data
        attr(tabulator_columns, "user_defined") <- TRUE
        table@tabulator_columns <- tabulator_columns
      }
    }

    return(table)
  }
  x <- build_prepare(x, tabulator_theme_fn, output = "html")

  return(x)
}

# =============================================================================
# TABULATOR OPTIONS FINALIZATION
# =============================================================================

#' Apply tabulator options
#' @param x tinytable object
#' @return Modified tinytable object
#' @keywords internal
#' @noRd
tabulator_apply_options <- function(x) {
    options_string <- if (nchar(x@tabulator_options) > 0) {
        paste0(x@tabulator_options, ",")
    } else {
        ""
    }
    x@table_string <- gsub(
        "$tinytable_TABULATOR_OPTIONS",
        options_string,
        x@table_string,
        fixed = TRUE
    )
    return(x)
}

#' Combine `tinytable` objects by rows (vertically)
#'
#' @details
#' Transformations recorded via `format_tt()` and `style_tt()` are evaluated at the very end of the rendering pipeline, after `rbind2()` has combined the tables. When headers are inserted or columns differ in type, the combined data is first coerced to character, so subsequent formatting/styling works on strings. Apply `format_tt()` directly to raw data frames before calling `tt()`, or re-run the formatting/styling steps on the combined table to preserve rounding and other rules.
#'
#' Calls to other `tinytable` functions such as `style_tt()` or `group_tt()` are ignored when applied to `x` or `y`. These functions should be applied to the final table instead.
#'
#' Information in these S4 slots is carried over from `x` to the combined table:
#'
#' + `x@output`
#' + `x@caption`
#' + `x@width`
#'
#' Information in these S4 slots is concatenated and carried over to the combined table:
#'
#' + `c(x@notes, y@notes)`
#'
#' This function relies on the `rbindlist()` function from the `data.table` package.
#'
#' @param x `tinytable` object
#' @param y `tinytable` object
#' @param use_names `TRUE` binds by matching column name, `FALSE` by position
#' @param headers Logical. TRUE inserts the colnames of `y` as an extra row between the two tables.
#' @param ... Additional arguments are ignored.
#' @aliases rbind2
#' @examples
#' library(tinytable)
#' x <- tt(mtcars[1:3, 1:2], caption = "Combine two tiny tables.")
#' y <- tt(mtcars[4:5, 8:10])
#'
#' # rbind() does not support additional aarguments
#' # rbind2() supports additional arguments
#'
#' # basic combination
#' rbind(x, y)
#'
#' rbind(x, y) |> format_tt(replace = "")
#'
#' # omit y header
#' rbind2(x, y, headers = FALSE)
#'
#' # bind by position rather than column names
#' rbind2(x, y, use_names = FALSE)
#'
#' # `iris` example with pre-tt() formatting
#' dat <- iris[1:3, 1:4]
#' a <- format_tt(dat, i = 1:3, digits = 1) |> tt()
#' b <- format_tt(dat, i = 1:3, digits = 2) |> tt()
#' rbind2(a, b)
#'
#' @importFrom methods rbind2
#' @export
setMethod(
  "rbind2",
  signature = signature(x = "tinytable", y = "tinytable"),
  definition = function(x, y, use_names = TRUE, headers = TRUE, ...) {
    assert_class(x, "tinytable")
    assert_class(y, "tinytable")
    assert_dependency("data.table")
    assert_flag(use_names)
    assert_flag(headers)

    x_df <- x@data_body
    y_df <- y@data_body

    # colnames() on a tinytable returns character(0) (never NULL) when the
    # table was created with `colnames = FALSE`, so check the length
    if (isTRUE(headers) && length(colnames(y)) > 0) {
      y_df <- base::rbind(colnames(y_df), y_df)
    }

    out <- data.table::rbindlist(
      list(x_df, y_df),
      fill = TRUE,
      use.names = use_names
    )

    out <- tt(out)

    out@output <- x@output
    out@notes <- c(x@notes, y@notes)
    out@width <- x@width
    out@caption <- x@caption

    return(out)
  }
)

source("helpers.R")

tt.lm <- function(x, ...) {
  results <- broom::tidy(x)
  tt(results, ...)
}

tab <- lm(mpg ~ hp, data = mtcars) |> tt()
expect_inherits(tab, "tinytable")
expect_equal(nrow(tab), 2)
expect_equal(ncol(tab), 5)

tab <- lm(mpg ~ hp + factor(cyl), data = mtcars) |> tt()
expect_equal(nrow(tab), 4)
expect_equal(ncol(tab), 5)

rm("tt.lm")

                                                           
1 Sepal.Length Sepal.Width Petal.Length Petal.Width Species
2          5.1         3.5          1.4         0.2  setosa
3        Hello                                             
4        World                                             
5          4.9         3.0          1.4         0.2  setosa
6          4.7         3.2          1.3         0.2  setosa
7          4.6         3.1          1.5         0.2  setosa
8          5.0         3.6          1.4         0.2  setosa
9          5.4         3.9          1.7         0.4  setosa

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa_class.R
\name{initialize,tinytable-method}
\alias{initialize,tinytable-method}
\title{Method for a tinytable S4 object}
\usage{
\S4method{initialize}{tinytable}(
  .Object,
  data = data.frame(),
  caption = NULL,
  notes = NULL,
  theme = list("default"),
  data_body = data.frame(),
  placement = NULL,
  width = NULL,
  height = NULL,
  colnames = TRUE
)
}
\arguments{
\item{caption}{A string that will be used as the caption of the table. This argument should \emph{not} be used in Quarto or Rmarkdown documents. In that context, please use the appropriate chunk options.}

\item{notes}{Notes to append to the bottom of the table. This argument accepts several different inputs:
\itemize{
\item Single string insert a single note: \code{"blah blah"}
\item Multiple strings insert multiple notes sequentially: \code{list("Hello world", "Foo bar")}
\item A named list inserts a list with the name as superscript: \code{list("a" = list("Hello World"))}
\item A named list with positions inserts markers as superscripts inside table cells: \code{list("a" = list(i = 0:1, j = 2, text = "Hello World"))}
}}

\item{theme}{Function or string.
\itemize{
\item String: grid, revealjs, rotate, striped, empty
\item Function: Applied to the \code{tinytable} object.
}}

\item{width}{Table or column width.
\itemize{
\item Single numeric value smaller than or equal to 1 determines the full table width, in proportion of line width.
\item Numeric vector of length equal to the number of columns in \code{x} determines the width of each column, in proportion of line width. If the sum of \code{width} exceeds 1, each element is divided by \code{sum(width)}. This makes the table full-width with relative column sizes.
}}

\item{height}{Row height in em units. Single numeric value greater than zero that determines the row height spacing.}

\item{colnames}{\code{TRUE}, \code{FALSE}, or "label". If "label", use the \code{attr(x$col,"label")} attribute if available and fall back on column names otherwise.}
}
\description{
Method for a tinytable S4 object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{pkgdown_print.tinytable}
\alias{pkgdown_print.tinytable}
\title{Print a tinytable object in pkgdown}
\usage{
pkgdown_print.tinytable(x, visible = TRUE)
}
\arguments{
\item{x}{A tinytable object}

\item{visible}{Whether the object is visible}
}
\value{
HTML output for pkgdown examples
}
\description{
Print a tinytable object in pkgdown
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa_class.R
\docType{class}
\name{tinytable-class}
\alias{tinytable-class}
\title{tinytable S4 class}
\description{
tinytable S4 class
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_style.R
\name{style_eval,tinytable_dataframe-method}
\alias{style_eval,tinytable_dataframe-method}
\title{tinytable S4 method}
\usage{
\S4method{style_eval}{tinytable_dataframe}(x)
}
\description{
tinytable S4 method
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_html.R
\name{theme_html}
\alias{theme_html}
\title{HTML-specific styles and options}
\usage{
theme_html(
  x,
  engine = get_option("tinytable_html_engine", default = NULL),
  i = NULL,
  j = NULL,
  class = get_option("tinytable_html_class", default = NULL),
  css = get_option("tinytable_html_css", default = NULL),
  css_rule = get_option("tinytable_html_css_rule", default = NULL),
  script = get_option("tinytable_html_script", default = NULL),
  portable = get_option("tinytable_html_portable"),
  tabulator_columns = get_option("tinytable_html_tabulator_columns"),
  tabulator_css_rule = get_option("tinytable_html_tabulator_css_rule"),
  tabulator_layout = get_option("tinytable_html_tabulator_layout", default =
    "fitDataTable"),
  tabulator_options = get_option("tinytable_html_tabulator_options"),
  tabulator_pagination = get_option("tinytable_html_tabulator_pagination"),
  tabulator_search = get_option("tinytable_html_tabulator_search"),
  tabulator_stylesheet = get_option("tinytable_html_tabulator_stylesheet"),
  ...
)
}
\arguments{
\item{x}{A \code{tinytable} object.}

\item{engine}{Character string specifying the HTML engine: "tinytable", "bootstrap", or "tabulator".}

\item{i}{Row indices.}

\item{j}{Column indices.}

\item{class}{String. HTML table class.}

\item{css}{Character vector. CSS style declarations.}

\item{css_rule}{String. CSS rules to customize table styling. Can be one of:
\itemize{
\item Shortcut strings: "tinytable" (default) or "dark" for dark theme
\item A file path to a local CSS file (e.g., "custom.css" or "path/to/styles.css")
\item A URL to an external stylesheet (must start with http/https and end with .css)
\item Inline CSS content as a string
}}

\item{script}{String. Custom JavaScript code or script tags to inject into the HTML output. Useful for loading external JavaScript libraries like Plotly, D3, etc.}

\item{portable}{Logical. Sets whether to create portable HTML output with embedded Javascript, CSS, and base64-encoded images.}

\item{tabulator_columns}{Custom column definitions.}

\item{tabulator_css_rule}{Complete CSS rules.}

\item{tabulator_layout}{Character string. Table layout algorithm for column sizing.
Default is "fitDataTable". Available options: "fitDataTable", "fitData", "fitDataFill",
"fitDataStretch", "fitColumns".}

\item{tabulator_options}{Custom Tabulator.js configuration options.}

\item{tabulator_pagination}{Logical or numeric vector. Pagination settings for large tables.
\itemize{
\item NULL (default): Preserves existing pagination settings, does not change previous configuration
\item FALSE: Explicitly disable pagination
\item TRUE: Enable pagination with automatic page sizes (10, 25, 50, 100, 250) filtered by row count
\item Numeric vector: First element is default page size, full vector provides page size options
}}

\item{tabulator_search}{Character or NULL. Search functionality position.
\itemize{
\item NULL (default): Preserves existing search settings, does not change previous configuration
\item "top": Adds search box above the table
\item "bottom": Adds search box below the table
\item "column": Adds header filters to each column for per-column searching
}}

\item{tabulator_stylesheet}{Character string. CSS stylesheet theme for Tabulator.js tables.
Default is "bootstrap5". Available options: "default", "simple", "midnight", "modern",
"site", "site_dark", "bootstrap3", "bootstrap4", "bootstrap5", "semanticui", "bulma",
"materialize", or a custom HTTP URL starting with "http".}

\item{...}{Additional arguments are ignored.}
}
\description{
HTML-specific styles and options
}
\examples{
\dontrun{
# Load an external JavaScript library
library(tinytable)
x <- tt(mtcars[1:5, 1:5])
x <- theme_html(x, script = '<script src="https://cdn.plot.ly/plotly-2.27.0.min.js"></script>')

# Add custom inline JavaScript
x <- tt(mtcars[1:5, 1:5])
x <- theme_html(x, script = '<script>console.log("Table loaded");</script>')
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_vector.R
\name{plot_vector}
\alias{plot_vector}
\title{Create Plot Vector}
\usage{
plot_vector(
  output = "html",
  fun = NULL,
  data = NULL,
  color = "black",
  xlim = NULL,
  height = 1,
  height_plot = 400,
  width_plot = 1200,
  images = NULL,
  sprintf = "\%s",
  assets = "tinytable_assets",
  ...
)
}
\arguments{
\item{output}{Output format. One of "html", "latex", "typst", "markdown", etc.}

\item{fun}{String or function to generate inline plots.
\itemize{
\item Built-in plot types (strings):
\itemize{
\item \code{"histogram"}: Creates histograms from numeric vectors. Accepts \code{color} argument.
\item \code{"density"}: Creates density plots from numeric vectors. Accepts \code{color} argument.
\item \code{"bar"}: Creates horizontal bar charts from single numeric values. Accepts \code{color} (single value, or length-2 vector for bar and background colors in static formats) and \code{xlim} arguments.
\item \code{"line"}: Creates line plots from data frames with \code{x} and \code{y} columns. Accepts \code{color} and \code{xlim} arguments.
}
\item Custom functions:
\itemize{
\item Functions that return \code{ggplot2} objects.
\item Functions that return another function which generates a base \code{R} plot, ex: \code{function(x) {function() hist(x)}}
\item Note: When using custom ggplot2 functions that return plots with text elements, the text size will normally need to be adjusted because the plot is inserted as a very small image in the table. Text sizes of 1 or smaller often work well (e.g., \code{theme(text = element_text(size = 1))}).
}
\item See the tutorial on the \code{tinytable} website for more information.
}}

\item{data}{a list of data frames or vectors to be used by the plotting functions in \code{fun}.}

\item{color}{string Name of color to use for inline plots (passed to the \code{col} argument base \code{graphics} plots in \code{R}). For bar plots in static output formats (PNG, PDF, etc.), can be a vector of length 2: c(bar_color, background_color) to show progress against a maximum. Note: Tabulator format only uses the first color.}

\item{xlim}{Numeric vector of length 2. Controls the range of bar plots.}

\item{height}{Numeric, the height of the images in the table in em units.}

\item{height_plot}{Numeric, the height of generated plot images in pixels (default: 400).}

\item{width_plot}{Numeric, the width of generated plot images in pixels (default: 1200).}

\item{images}{Character vector, the paths to the images to be inserted. Paths are relative to the main table file or Quarto (Rmarkdown) document.}

\item{sprintf}{Character string, a sprintf format string to format the generated cell content. Default is "\%s" which displays the content as-is. Use this to wrap images or plots in custom markup.}

\item{assets}{Path to the directory where generated assets are stored. This path is relative to the location where a table is saved.}

\item{...}{Extra arguments are passed to the function in \code{fun}. Important: Custom plotting functions must always have \code{...} as an argument.}
}
\value{
A character vector with plot file paths or HTML tags.
}
\description{
Create Plot Vector
}
\details{
This function creates plots by creating a temporary \code{tt()} object and applying \code{plot_tt()}. It returns a character vector containing the image paths or HTML tags for the plots.
}
\examples{
\dontrun{
# Create histogram plots
plot_data <- list(rnorm(100), rnorm(50))
plot_vector(fun = "histogram", data = plot_data, output = "html")

# Create density plots
plot_vector(fun = "density", data = plot_data, output = "latex")

# Create bar plots from single values
bar_data <- list(0.5, 0.8, 0.3)
plot_vector(fun = "barpct", data = bar_data, output = "html")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_empty.R
\name{theme_empty}
\alias{theme_empty}
\title{Theme for a void table}
\usage{
theme_empty(x, ...)
}
\arguments{
\item{x}{A tinytable object.}

\item{...}{Additional arguments are ignored.}
}
\description{
This function cancels all styles and formatting applied to a \code{tinytable}
object up to that point in the pipeline. Images and inline plots inserted
with \code{plot_tt()} are deliberately preserved: they are table content rather
than styling.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_tt.R, R/zzz-aliases.R
\name{plot_tt}
\alias{plot_tt}
\alias{tt_plot}
\title{Insert images and inline plots into tinytable objects}
\usage{
plot_tt(
  x,
  i = NULL,
  j = NULL,
  fun = NULL,
  data = NULL,
  color = "black",
  xlim = NULL,
  height = 1,
  height_plot = 400,
  width_plot = 1200,
  images = NULL,
  sprintf = "\%s",
  assets = "tinytable_assets",
  output = NULL,
  ...
)

tt_plot(
  x,
  i = NULL,
  j = NULL,
  fun = NULL,
  data = NULL,
  color = "black",
  xlim = NULL,
  height = 1,
  height_plot = 400,
  width_plot = 1200,
  images = NULL,
  sprintf = "\%s",
  assets = "tinytable_assets",
  output = NULL,
  ...
)
}
\arguments{
\item{x}{A tinytable object.}

\item{i}{Integer vector, the row indices where images are to be inserted. If \code{NULL},
images will be inserted in all rows.}

\item{j}{Integer vector, the column indices where images are to be inserted. If \code{NULL},
images will be inserted in all columns.}

\item{fun}{String or function to generate inline plots.
\itemize{
\item Built-in plot types (strings):
\itemize{
\item \code{"histogram"}: Creates histograms from numeric vectors. Accepts \code{color} argument.
\item \code{"density"}: Creates density plots from numeric vectors. Accepts \code{color} argument.
\item \code{"bar"}: Creates horizontal bar charts from single numeric values. Accepts \code{color} (single value, or length-2 vector for bar and background colors in static formats) and \code{xlim} arguments.
\item \code{"line"}: Creates line plots from data frames with \code{x} and \code{y} columns. Accepts \code{color} and \code{xlim} arguments.
}
\item Custom functions:
\itemize{
\item Functions that return \code{ggplot2} objects.
\item Functions that return another function which generates a base \code{R} plot, ex: \code{function(x) {function() hist(x)}}
\item Note: When using custom ggplot2 functions that return plots with text elements, the text size will normally need to be adjusted because the plot is inserted as a very small image in the table. Text sizes of 1 or smaller often work well (e.g., \code{theme(text = element_text(size = 1))}).
}
\item See the tutorial on the \code{tinytable} website for more information.
}}

\item{data}{a list of data frames or vectors to be used by the plotting functions in \code{fun}.}

\item{color}{string Name of color to use for inline plots (passed to the \code{col} argument base \code{graphics} plots in \code{R}). For bar plots in static output formats (PNG, PDF, etc.), can be a vector of length 2: c(bar_color, background_color) to show progress against a maximum. Note: Tabulator format only uses the first color.}

\item{xlim}{Numeric vector of length 2. Controls the range of bar plots.}

\item{height}{Numeric, the height of the images in the table in em units.}

\item{height_plot}{Numeric, the height of generated plot images in pixels (default: 400).}

\item{width_plot}{Numeric, the width of generated plot images in pixels (default: 1200).}

\item{images}{Character vector, the paths to the images to be inserted. Paths are relative to the main table file or Quarto (Rmarkdown) document.}

\item{sprintf}{Character string, a sprintf format string to format the generated cell content. Default is "\%s" which displays the content as-is. Use this to wrap images or plots in custom markup.}

\item{assets}{Path to the directory where generated assets are stored. This path is relative to the location where a table is saved.}

\item{output}{Apply plot insertion only when the table is rendered in the specified format. A character vector of one or more of "latex", "html", "typst", or "markdown". If \code{NULL} (default), plots are inserted regardless of the output format.}

\item{...}{Extra arguments are passed to the function in \code{fun}. Important: Custom plotting functions must always have \code{...} as an argument.}
}
\value{
A modified tinytable object with images or plots inserted.
}
\description{
The \code{plot_tt()} function allows for the insertion of images and inline plots into
tinytable objects. This function can handle both local and web-based images.
}
\details{
The \code{plot_tt()} can insert images and inline plots into tables.
}
\examples{
\dontrun{
# Bar plots with single and dual colors
dat <- data.frame(
  Metric = c("Sales", "Conversion", "Growth", "Efficiency"),
  Value = c(75, 45, 92, 38),
  Percentage = c(0.75, 0.45, 0.92, 0.38)
)

tt(dat) |>
  plot_tt(j = 2, fun = "bar", data = as.list(dat$Value), color = "darkorange") |>
  plot_tt(j = 3, fun = "bar", data = as.list(dat$Percentage),
          color = c("steelblue", "lightgrey"), xlim = c(0, 1))

# Built-in plot types
plot_data <- list(mtcars$mpg, mtcars$hp, mtcars$qsec)

dat <- data.frame(
  Variables = c("mpg", "hp", "qsec"),
  Histogram = "",
  Density = "",
  Line = ""
)

# Random data for sparklines
lines <- lapply(1:3, \(x) data.frame(x = 1:10, y = rnorm(10)))

tt(dat) |>
  plot_tt(j = 2, fun = "histogram", data = plot_data) |>
  plot_tt(j = 3, fun = "density", data = plot_data, color = "darkgreen") |>
  plot_tt(j = 4, fun = "line", data = lines, color = "blue") |>
  style_tt(j = 2:4, align = "c")

# Custom function example (must have ... argument)
custom_hist <- function(d, ...) {
  function() hist(d, axes = FALSE, ann = FALSE, col = "lightblue")
}

tt(data.frame(Variables = "mpg", Histogram = "")) |>
  plot_tt(j = 2, fun = custom_hist, data = list(mtcars$mpg))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa_class.R
\name{finalize}
\alias{finalize}
\title{Apply final settings to a tinytable}
\usage{
finalize(x, ...)
}
\arguments{
\item{x}{A data frame, data table, or tibble to be rendered as a table.}

\item{...}{Additional arguments are ignored}
}
\description{
Apply final settings to a tinytable
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_typst.R
\name{theme_typst}
\alias{theme_typst}
\title{Typst-specific styles and options}
\usage{
theme_typst(
  x,
  multipage = get_option("tinytable_typst_multipage", default = FALSE),
  figure = get_option("tinytable_typst_figure", default = TRUE),
  portable = get_option("tinytable_typst_portable", default = NULL),
  align_figure = get_option("tinytable_typst_align_figure", NULL),
  resize_width = get_option("tinytable_typst_resize_width", 1),
  resize_height = get_option("tinytable_typst_resize_height", default = NULL),
  resize_direction = get_option("tinytable_typst_resize_direction", default = NULL),
  ...
)
}
\arguments{
\item{x}{A \code{tinytable} object.}

\item{multipage}{Logical. When \code{TRUE}, emits a Typst show rule that allows figures to break across pages. When \code{FALSE} (default), tables are kept on a single page. When \code{NULL}, no show rule is emitted.}

\item{figure}{Logical, whether to wrap the table in a Typst figure environment and block.}

\item{portable}{Logical. Sets whether to create portable Typst output with base64-encoded local images embedded directly in the Typst code. Remote image URLs are not downloaded.}

\item{align_figure}{Character string indicating horizontal alignment: "l", "c", or "r".
Defaults to \code{get_option("tinytable_typst_align_figure", NULL)}. When NULL, no figure-level alignment is emitted.}

\item{resize_width}{Numeric value between 0.01 and 1.0 specifying the target width
as a fraction of the available Typst layout width when resize_direction is
specified.}

\item{resize_height}{Numeric value between 0.01 and 1.0 specifying the target
height as a fraction of the available Typst layout height. When specified,
height controls the scaling and width scales proportionally.}

\item{resize_direction}{Character string specifying how to resize tables. Options
are: "down" to shrink oversized tables, "up" to expand undersized tables,
and "both" to always scale to the target size.}

\item{...}{Additional arguments.}
}
\description{
Typst-specific styles and options
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa_class.R
\name{colnames,tinytable-method}
\alias{colnames,tinytable-method}
\title{Method for a tinytable S4 object}
\usage{
\S4method{colnames}{tinytable}(x)
}
\arguments{
\item{x}{A data frame, data table, or tibble to be rendered as a table.}
}
\description{
Method for a tinytable S4 object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_latex.R
\name{theme_latex}
\alias{theme_latex}
\title{LaTeX-Specific Theme for \code{tinytable}}
\usage{
theme_latex(
  x,
  inner = NULL,
  outer = NULL,
  environment = get_option("tinytable_latex_environment", default = NULL),
  environment_table = get_option("tinytable_latex_environment_table", default = TRUE),
  multipage = get_option("tinytable_latex_multipage", default = FALSE),
  rowhead = get_option("tinytable_latex_rowhead", 0L),
  rowfoot = get_option("tinytable_latex_rowfoot", 0L),
  resize_width = get_option("tinytable_latex_resize_width", 1),
  resize_height = get_option("tinytable_latex_resize_height", default = NULL),
  resize_direction = get_option("tinytable_latex_resize_direction", default = NULL),
  placement = get_option("tinytable_latex_placement", NULL),
  preamble = NULL,
  engine = NULL,
  ...
)
}
\arguments{
\item{x}{A \code{tinytable} object to apply LaTeX theming to.}

\item{inner}{Character string specifying inner tabularray options. These options
control the internal formatting of the table (e.g., column alignment, spacing).
Will be added to any existing inner options. Default is \code{NULL}.}

\item{outer}{Character string specifying outer tabularray options. These options
control the external formatting around the table. Will be added to any existing
outer options. Default is \code{NULL}.}

\item{environment}{Character string specifying the LaTeX table environment to use.
Options are:
\itemize{
\item \code{"tblr"} - Standard tabularray table (default)
\item \code{"talltblr"} - Tall tabularray table for tables that may break across pages
\item \code{"longtblr"} - Long tabularray table for multi-page tables
\item \code{"tabular"} - Basic LaTeX tabular environment without tabularray features
}

Default is controlled by \code{tinytable_latex_environment} option.}

\item{environment_table}{Logical indicating whether to wrap the table in a \code{table}
environment. When \code{FALSE}, only the core table structure is output without the
surrounding table wrapper. Automatically set to \code{FALSE} when \code{environment = "longtblr"}.
Default is controlled by \code{tinytable_latex_environment_table} option.}

\item{multipage}{Logical indicating whether to enable multipage table functionality.
When \code{TRUE}, automatically switches to \code{longtblr} environment and sets appropriate
options for tables that span multiple pages. Default is controlled by
\code{tinytable_latex_multipage} option.}

\item{rowhead}{Integer specifying the number of header rows to repeat on each page
in multipage tables. Only valid with \code{longtblr} environment. Default is controlled
by \code{tinytable_latex_rowhead} option.}

\item{rowfoot}{Integer specifying the number of footer rows to repeat on each page
in multipage tables. Only valid with \code{longtblr} environment. Default is controlled
by \code{tinytable_latex_rowfoot} option.}

\item{resize_width}{Numeric value between 0.01 and 1.0 specifying the target width
as a fraction of \verb{\\\\linewidth} when resizing tables. Only used when \code{resize_direction}
is specified. Default is controlled by \code{tinytable_latex_resize_width} option.}

\item{resize_height}{Numeric value between 0.01 and 1.0 specifying the target height
as a fraction of \verb{\\\\textheight} when resizing tables. When specified (and \code{resize_direction}
is set), the table height is controlled instead of its width. If both \code{resize_width}
and \code{resize_height} are specified, \code{resize_height} takes precedence (width scales
proportionally). Default is controlled by \code{tinytable_latex_resize_height} option.}

\item{resize_direction}{Character string specifying how to resize tables that are
too wide or too narrow. Options are:
\itemize{
\item \code{"down"} - Only shrink tables wider than \verb{\\\\linewidth} (or taller than \verb{\\\\textheight} when \code{resize_height} is set)
\item \code{"up"} - Only expand tables narrower than \verb{\\\\linewidth} (or shorter than \verb{\\\\textheight} when \code{resize_height} is set)
\item \code{"both"} - Resize all tables to exactly \verb{resize_width * \\\\linewidth} (or \verb{resize_height * \\\\textheight})
}

Default is controlled by \code{tinytable_latex_resize_direction} option.}

\item{placement}{Character string specifying LaTeX float placement options for the
table environment (e.g., "h", "t", "b", "p", "H"). Only used when \code{environment_table = TRUE}.
Default is controlled by \code{tinytable_latex_placement} option.}

\item{preamble}{Logical value specifying whether to include LaTeX preamble packages.
If not NULL, overrides the table's preamble setting.}

\item{engine}{Character string specifying the LaTeX engine to use for PDF compilation.
Options are "xelatex", "pdflatex", or "lualatex". If not NULL, overrides the table's engine setting.}

\item{...}{Additional arguments (currently unused).}
}
\value{
A modified \code{tinytable} object with LaTeX-specific theming applied.
}
\description{
This function provides comprehensive LaTeX-specific theming and configuration options
for \code{tinytable} objects. It allows customization of LaTeX environments, table layout,
multipage behavior, resizing, and placement within LaTeX documents.
}
\details{
The function provides fine-grained control over LaTeX table output through several mechanisms:

\strong{Environment Selection:}
Different LaTeX environments offer different capabilities:
\itemize{
\item \code{tblr}: Modern tabularray syntax with full styling support
\item \code{talltblr}: Like \code{tblr} but optimized for tall tables
\item \code{longtblr}: Supports page breaks and repeated headers/footers
\item \code{tabular}: Basic LaTeX syntax, limited styling but maximum compatibility
}

\strong{Multipage Tables:}
When \code{multipage = TRUE} or when \code{rowhead}/\code{rowfoot} are specified, the function
automatically switches to \code{longtblr} environment and disables the table wrapper.
This allows tables to break across pages while maintaining headers and footers.

\strong{Resizing:}
The resize functionality uses LaTeX's \verb{\\\\resizebox} command to automatically
adjust table dimensions based on content and page constraints. Use \code{resize_width}
to control width (fraction of \verb{\\\\linewidth}) or \code{resize_height} to control height
(fraction of \verb{\\\\textheight}). When \code{resize_height} is specified, the table is scaled
to fit the target height with width adjusting proportionally.

\strong{Tabularray Options:}
Inner and outer options directly control tabularray formatting. Inner options
affect cell content and spacing, while outer options control the table's
relationship with surrounding text.
}
\seealso{
\code{\link[=tt]{tt()}}, \code{\link[=style_tt]{style_tt()}}, \code{\link[=save_tt]{save_tt()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_grid.R
\name{theme_grid}
\alias{theme_grid}
\title{Grid theme with borders around all cells}
\usage{
theme_grid(x, ...)
}
\arguments{
\item{x}{A \code{tinytable} object.}

\item{...}{Additional arguments (ignored).}
}
\value{
A modified \code{tinytable} object.
}
\description{
Grid theme with borders around all cells
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_markdown.R
\name{theme_markdown}
\alias{theme_markdown}
\title{Markdown theme with optional ANSI color support and grid customization}
\usage{
theme_markdown(
  x,
  ansi = NULL,
  style = NULL,
  vline = NULL,
  hline = NULL,
  hline_header = NULL,
  ...
)
}
\arguments{
\item{x}{A \code{tinytable} object.}

\item{ansi}{Logical. If TRUE, enables ANSI color codes for grid styling; if FALSE, disables them. Default is NULL (no change; ANSI is disabled unless previously enabled).}

\item{style}{Character. Markdown style format. Can be "grid" or "gfm". Default is NULL.}

\item{vline}{Logical. Enable/disable vertical lines. Default is TRUE.}

\item{hline}{Logical. Enable/disable horizontal lines. Default is TRUE.}

\item{hline_header}{Logical. Enable/disable the special header separator line below column names. Default is TRUE.}

\item{...}{Additional arguments (ignored).}
}
\value{
A modified \code{tinytable} object.
}
\description{
Markdown theme with optional ANSI color support and grid customization
}
\details{
When \code{ansi = TRUE}, colors and text styling (bold, italic, strikeout, underline)
are applied using ANSI escape sequences for terminal display. ANSI colors require
a terminal or application that supports ANSI escape sequences. Common supported
terminals include: Terminal.app (macOS), iTerm2 (macOS), Windows Terminal, most
Linux terminals, RStudio Console, and VS Code terminal. Colors may not display
correctly in basic text editors or older terminals.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_tinytable.R
\name{theme_tinytable}
\alias{theme_tinytable}
\title{Default theme for TinyTable}
\usage{
theme_tinytable(x, ...)
}
\arguments{
\item{x}{A tinytable object.}

\item{...}{Additional arguments are ignored.}
}
\value{
A modified \code{tinytable} object.
}
\description{
Default theme for TinyTable
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.tinytable}
\alias{print.tinytable}
\title{Print, display, or convert a tinytable object}
\usage{
\method{print}{tinytable}(x, output = get_option("tinytable_print_output", default = NULL), ...)
}
\arguments{
\item{x}{A data frame, data table, or tibble to be rendered as a table.}

\item{output}{Format in which a Tiny Table is printed. One of:
\itemize{
\item \code{NULL}: automatically chosen based on context (see below).
\item \code{"html"}: HTML table. In interactive mode, displayed in the viewer pane.
\item \code{"latex"}: LaTeX table using tabularray.
\item \code{"markdown"}: plain text markdown table.
\item \code{"typst"}: Typst table.
\item \code{"dataframe"}: formatted data frame.
\item \code{"knitr"}: PNG screenshot via \code{webshot2}, returned through \code{knitr::include_graphics()}. Useful for embedding styled tables in Rmarkdown or Quarto documents. Requires the \code{webshot2} package.
\item \code{"raster"}: PNG screenshot via \code{webshot2}, drawn on the graphics device via \code{grid::grid.raster()}. Useful for pkgdown \verb{@examples} and README files, where the plot is captured as an image. Requires the \code{webshot2} and \code{png} packages. For README.Rmd files rendered with \code{rmarkdown::render()}, set \code{knitr::opts_chunk$set(fig.path = "man/figures/README-")} in a setup chunk so that pkgdown copies the images correctly.
}

When \code{output} is \code{NULL}, the format is chosen automatically:
\itemize{
\item When called from a script in non-interactive mode, the default is "markdown" (\code{interactive() == FALSE}).
\item When called interactively in RStudio, the default is to display an HTML table in the viewer pane.
\item When called interactively in another development environment, the default is "markdown".
\item The default print output can be changed for an entire R session by calling: \code{options(tinytable_print_output = "html")}
\item The default print output can be changed for a single \code{tinytable} object by modifying the \code{output} S4 slot.
}}

\item{...}{Other arguments are ignored.}
}
\value{
launch a browser window or cat() the table to console.
}
\description{
This function is called automatically by \code{R} whenever a \code{tinytable} object is anprinted to the console or in an HTML viewer pane.
}
\details{
When printing to HTML in \code{interactive()} mode, a temporary file is created and \code{viewer()} is called to preview the file with the local browser (ex: Firefox or Chrome). The temporary file is then automatically cleaned up. On some operating systems, like some Linux distributions, browser do not have read access to the \verb{/tmp/} directory. In such cases, users can specify a custom location to store temporary HTML files. Note that this prevents \code{tinytable} from automatically cleaning up temporary files automatically.

\code{options(tinytable_tempdir = "/home/username/temp_directory")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json.R
\name{df_to_json}
\alias{df_to_json}
\title{Convert data frame or list to JSON}
\usage{
df_to_json(data)
}
\arguments{
\item{data}{A data frame or list to convert to JSON}
}
\value{
JSON string
}
\description{
Simple helper function to convert a data frame or list to JSON format
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/typst_group.R
\name{group_eval_j,tinytable_typst-method}
\alias{group_eval_j,tinytable_typst-method}
\title{tinytable S4 method}
\usage{
\S4method{group_eval_j}{tinytable_typst}(x, i = NULL, j = NULL, indent = 0, ...)
}
\description{
tinytable S4 method
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa_class.R
\name{as.character,tinytable-method}
\alias{as.character,tinytable-method}
\title{Convert a tinytable S4 object to a string}
\usage{
\S4method{as.character}{tinytable}(x)
}
\arguments{
\item{x}{A data frame, data table, or tibble to be rendered as a table.}
}
\description{
Convert a tinytable S4 object to a string
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_tt.R
\name{theme_tt}
\alias{theme_tt}
\title{Deprecated: Use format-specific theme functions instead}
\usage{
theme_tt(x, theme, ...)
}
\arguments{
\item{x}{deprecated}

\item{theme}{deprecated}

\item{...}{Additional arguments}
}
\value{
Throws an informative error message
}
\description{
\strong{DEPRECATED}: The \code{theme_tt()} function has been deprecated. Please use the format-specific or style-specific theme functions instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset.R
\name{subset.tinytable}
\alias{subset.tinytable}
\title{Subsetting a \code{tinytable} object}
\usage{
\method{subset}{tinytable}(x, subset, select, drop = FALSE, ...)
}
\arguments{
\item{x}{object to be subsetted.}

\item{subset}{logical expression indicating elements or rows to keep:
    missing values are taken as false.}

\item{select}{expression, indicating columns to select from a
    data frame.}

\item{drop}{passed on to \code{[} indexing operator.}

\item{...}{further arguments to be passed to or from other methods.}
}
\description{
Return subsets \code{tinytable} which meet conditions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/style_tt.R, R/zzz-aliases.R
\name{style_tt}
\alias{style_tt}
\alias{tt_style}
\title{Style a Tiny Table}
\usage{
style_tt(
  x,
  i = NULL,
  j = NULL,
  bold = NULL,
  italic = NULL,
  monospace = NULL,
  smallcap = NULL,
  underline = NULL,
  strikeout = NULL,
  color = NULL,
  background = NULL,
  fontsize = NULL,
  align = NULL,
  alignv = NULL,
  colspan = NULL,
  rowspan = NULL,
  indent = NULL,
  line = NULL,
  line_color = NULL,
  line_width = 0.1,
  line_type = NULL,
  line_trim = NULL,
  finalize = NULL,
  output = NULL,
  ...
)

tt_style(
  x,
  i = NULL,
  j = NULL,
  bold = NULL,
  italic = NULL,
  monospace = NULL,
  smallcap = NULL,
  underline = NULL,
  strikeout = NULL,
  color = NULL,
  background = NULL,
  fontsize = NULL,
  align = NULL,
  alignv = NULL,
  colspan = NULL,
  rowspan = NULL,
  indent = NULL,
  line = NULL,
  line_color = NULL,
  line_width = 0.1,
  line_type = NULL,
  line_trim = NULL,
  finalize = NULL,
  output = NULL,
  ...
)
}
\arguments{
\item{x}{A table object created by \code{tt()}.}

\item{i}{Numeric vector, logical matrix, string, or unquoted expression.
\itemize{
\item Numeric vector: Row indices where the styling should be applied. Can be a single value or a vector.
\item Logical matrix: A matrix with the same number of rows and columns as \code{x}. \code{i=0} is the header, and negative values are higher level headers. Row indices refer to rows \emph{after} the insertion of row labels by \code{group_tt()}, when applicable.
\item String: Table components "caption", "colnames", "groupi" (row group labels), "~groupi" (non-group rows), "groupj" (column group labels), "notes".
\item Unquoted expression: When supplying an unquoted expression, it is first evaluated in the calling environment, then in the data frame passed to \code{tt()}.
}}

\item{j}{Column indices where the styling should be applied. Can be:
\itemize{
\item Integer vectors indicating column positions.
\item Character vector indicating column names.
\item A single string specifying a Perl-style regular expression used to match column names.
\item Unquoted expression: Non-standard evaluation is supported. When supplying an unquoted expression, it is first evaluated in the calling environment, then in an environment that includes the columns of the original data passed to \code{tt()}, and \code{groupi} indices. See examples below.
}}

\item{bold}{Logical; if \code{TRUE}, text is styled in bold.}

\item{italic}{Logical; if \code{TRUE}, text is styled in italic.}

\item{monospace}{Logical; if \code{TRUE}, text is styled in monospace font.}

\item{smallcap}{Logical; if \code{TRUE}, text is styled in small caps. In Markdown output format, text is converted to uppercase.}

\item{underline}{Logical; if \code{TRUE}, text is underlined.}

\item{strikeout}{Logical; if \code{TRUE}, text has a strike through line.}

\item{color}{Text color. Colors are standardized across output formats and can be specified as:
\itemize{
\item Hex codes: "#CC79A7", "#FF0000", "#123ABC"
\item R color names: Any color recognized by R, such as "red", "blue", "forestgreen", "lightblue"
\item Extended color names: 749+ named colors from the LaTeX xcolor package (see \code{tinytable:::latex_colors} for the full list)
\item LaTeX color blending (LaTeX output only): "white!80!blue", "red!50", "green!20!red"
}}

\item{background}{Background color. Same color specification options as the \code{color} parameter. Can be \code{NULL} for default color.}

\item{fontsize}{Font size in em units. Can be \code{NULL} for default size.}

\item{align}{A single character or a string with a number of characters equal to the number of columns in \code{j}. Valid characters include 'c' (center), 'l' (left), 'r' (right), 'd' (decimal). Decimal alignment is only available in LaTeX via the \code{siunitx} package. The width of columns is determined by the maximum number of digits to the left and to the right in all cells specified by \code{i} and \code{j}.}

\item{alignv}{A single character specifying vertical alignment. Valid characters include 't' (top), 'm' (middle), 'b' (bottom).}

\item{colspan}{Number of columns a cell should span. \code{i} and \code{j} must be of length 1.}

\item{rowspan}{Number of rows a cell should span. \code{i} and \code{j} must be of length 1.}

\item{indent}{Text indentation in em units. Positive values only.}

\item{line}{String determines if lines (rules or borders) should be drawn around the cell, row, or column. See \code{line_type} to draw dashed or dotted lines.
\itemize{
\item "t": top
\item "b": bottom
\item "l": left
\item "r": right
\item Can be combined such as: "lbt" to draw borders at the left, bottom, and top.
}}

\item{line_color}{Color of the line. See the \code{color} argument for details.}

\item{line_width}{Width of the line in em units (default: 0.1).}

\item{line_type}{Style of the line: "solid" (default), "dashed", or "dotted". Ignored in Markdown and Word output, and by the Tabulator HTML engine, which always draw solid rules.}

\item{line_trim}{String specifying line trimming. Acceptable values: "l" (left), "r" (right), "lr" (both sides). When specified, shortens the lines by 0.8pt on the specified side(s). Default: NULL (no trimming).}

\item{finalize}{A function applied to the table object at the very end of table-building, for post-processing. For example, the function could use regular expressions to add LaTeX commands to the text version of the table hosted in \code{x@table_string}, or it could programmatically change the caption in \code{x@caption}.}

\item{output}{Apply styling only when the table is rendered in the specified format. A character vector of one or more of "latex", "html", "typst", or "markdown". If \code{NULL} (default), styling is applied regardless of the output format.}

\item{...}{extra arguments are ignored}
}
\value{
An object of class \code{tt} representing the table.
}
\description{
Style a Tiny Table

Alias for \code{style_tt()}
}
\details{
This function applies styling to a table created by \code{tt()}. It allows customization of text style (bold, italic, monospace), text and background colors, font size, cell width, text alignment, column span, and indentation. The function also supports passing native instructions to LaTeX (tabularray) and HTML (bootstrap) formats.

Vector values for style arguments (e.g., \code{color}, \code{background}, \code{fontsize}) are recycled over the selected cells in the order specified by the user in \code{i} and \code{j}, with \code{i} varying fastest. For example, \code{style_tt(x, j = c(3, 1), background = c("red", "blue"))} colors column 3 red and column 1 blue. When \code{i} is a logical matrix, values are recycled over the \code{TRUE} cells in column-major order.
}
\section{Markdown limitations}{


Markdown is a text-only format that only supports these styles: italic, bold, strikeout. The \code{width} argument is also unavailable.
These limitations exist because there is no standard markdown syntax for the other styling options.

However, in terminals (consoles) that support it, \code{tinytable} can display colors and text styles using
ANSI escape codes by setting \code{theme_markdown(ansi = TRUE)}. This allows for rich formatting in
compatible terminal environments.
}

\section{Word limitations}{


Word tables only support these styles: italic, bold, strikeout. The \code{width} argument is also unavailable.
Moreover, the \code{style_tt()} function cannot be used to style headers inserted by the \code{group_tt()} function;
instead, you should style the headers directly in the header definition using markdown syntax:
\code{group_tt(i = list("*italic header*" = 2))}. These limitations are due to the fact that we create
Word documents by converting a markdown table to .docx via the Pandoc software, which requires
going through a text-only intermediate format.
}

\examples{
\dontshow{if (knitr::is_html_output()) withAutoprint(\{ # examplesIf}
if (knitr::is_html_output()) options(tinytable_print_output = "html")

library(tinytable)

tt(mtcars[1:5, 1:6])

# Alignment
tt(mtcars[1:5, 1:6]) |>
  style_tt(j = 1:5, align = "lcccr")

# Colors and styles
tt(mtcars[1:5, 1:6]) |>
  style_tt(i = 2:3, background = "black", color = "orange", bold = TRUE)

# column selection with `j``
tt(mtcars[1:5, 1:6]) |>
  style_tt(j = 5:6, background = "pink")

tt(mtcars[1:5, 1:6]) |>
  style_tt(j = "drat|wt", background = "pink")

tt(mtcars[1:5, 1:6]) |>
  style_tt(j = c("drat", "wt"), background = "pink")

tt(mtcars[1:5, 1:6], theme = "empty") |>
  style_tt(
    i = 2, j = 2,
    colspan = 3,
    rowspan = 2,
    align = "c",
    alignv = "m",
    color = "white",
    background = "black",
    bold = TRUE)

tt(mtcars[1:5, 1:6], theme = "empty") |>
  style_tt(
    i = 0:3,
    j = 1:3,
    line = "tblr",
    line_width = 0.4,
    line_color = "teal")

tt(mtcars[1:5, 1:6], theme = "striped") |>
  style_tt(
    i = c(2, 5),
    j = 3,
    strikeout = TRUE,
    fontsize = 0.7)

# Non-standard evaluation (NSE)
dat <- data.frame(
  w = c(143002.2092, 201399.181, 100188.3883),
  x = c(1.43402, 201.399, 0.134588),
  y = as.Date(c(897, 232, 198), origin = "1970-01-01"),
  z = c(TRUE, TRUE, FALSE)
)
tt(dat) |>
  style_tt(i = w > 150000, j = c("w", "x"), 
           color = "white", background = "black")

tt(mtcars[1:5, 1:6]) |>
  theme_html(class = "table table-dark table-hover")


inner <- "
column{1-4}={halign=c},
hlines = {fg=white},
vlines = {fg=white},
cell{1,6}{odd} = {bg=teal7},
cell{1,6}{even} = {bg=green7},
cell{2,4}{1,4} = {bg=red7},
cell{3,5}{1,4} = {bg=purple7},
cell{2}{2} = {r=4,c=2}{bg=azure7},
"
tt(mtcars[1:5, 1:4], theme = "empty") |>
  theme_latex(inner = inner)

# Style group rows and non-group rows
dat <- data.frame(x = 1:6, y = letters[1:6])
dat |>
  tt() |>
  group_tt(i = list("Group A" = 3)) |>
  style_tt(i = "groupi", background = "lightblue") |>
  style_tt(i = "~groupi", background = "lightgray")

# unquote expressions
dat <- mtcars[1:10,]
dat <- dat[order(dat$am),]
tt(dat) |>
   subset(mpg > 20) |>
   group_tt(am)

# style elements: captions and colnames
notes <- list(
  "*" = "Hello world", 
  "a" = "Bacon ipsum dolor amet kevin t-bone porchetta.")
tt(head(iris), 
   width = .8,
   caption = "This is a Caption Example.", 
   notes = notes) |>
  style_tt(2, 2, background = "pink", rowspan = 2, colspan = 2, 
    alignv = "m", align = "c", line = "tblr") |>
  style_tt("colnames", italic = TRUE) |>
  style_tt("caption", smallcap = TRUE)
\dontshow{\}) # examplesIf}
}

\name{tinytable-package}
\alias{tinytable-package}
\alias{tinytable}
\docType{package}
\title{\packageTitle{tinytable}}
\description{\packageDescription{tinytable}}
\section{Package Content}{\packageIndices{tinytable}}
\author{\packageAuthor{tinytable}}
\section{Maintainer}{\packageMaintainer{tinytable}}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_default.R
\name{theme_default}
\alias{theme_default}
\title{Deprecated: Use theme_tinytable() instead}
\usage{
theme_default(x, ...)
}
\arguments{
\item{x}{A tinytable object.}

\item{...}{Additional arguments are ignored.}
}
\value{
A modified \code{tinytable} object.
}
\description{
\code{theme_default()} has been renamed to \code{theme_tinytable()}. This function is
deprecated and will be removed in a future version.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa_class.R
\name{group_eval_j}
\alias{group_eval_j}
\title{Apply group settings to a tinytable}
\usage{
group_eval_j(x, ...)
}
\arguments{
\item{x}{A data frame, data table, or tibble to be rendered as a table.}

\item{...}{Additional arguments are ignored}
}
\description{
Apply group settings to a tinytable
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbind2.R
\name{rbind2,tinytable,tinytable-method}
\alias{rbind2,tinytable,tinytable-method}
\alias{rbind2}
\title{Combine \code{tinytable} objects by rows (vertically)}
\usage{
\S4method{rbind2}{tinytable,tinytable}(x, y, use_names = TRUE, headers = TRUE, ...)
}
\arguments{
\item{x}{\code{tinytable} object}

\item{y}{\code{tinytable} object}

\item{use_names}{\code{TRUE} binds by matching column name, \code{FALSE} by position}

\item{headers}{Logical. TRUE inserts the colnames of \code{y} as an extra row between the two tables.}

\item{...}{Additional arguments are ignored.}
}
\description{
Combine \code{tinytable} objects by rows (vertically)
}
\details{
Transformations recorded via \code{format_tt()} and \code{style_tt()} are evaluated at the very end of the rendering pipeline, after \code{rbind2()} has combined the tables. When headers are inserted or columns differ in type, the combined data is first coerced to character, so subsequent formatting/styling works on strings. Apply \code{format_tt()} directly to raw data frames before calling \code{tt()}, or re-run the formatting/styling steps on the combined table to preserve rounding and other rules.

Calls to other \code{tinytable} functions such as \code{style_tt()} or \code{group_tt()} are ignored when applied to \code{x} or \code{y}. These functions should be applied to the final table instead.

Information in these S4 slots is carried over from \code{x} to the combined table:
\itemize{
\item \code{x@output}
\item \code{x@caption}
\item \code{x@width}
}

Information in these S4 slots is concatenated and carried over to the combined table:
\itemize{
\item \code{c(x@notes, y@notes)}
}

This function relies on the \code{rbindlist()} function from the \code{data.table} package.
}
\examples{
library(tinytable)
x <- tt(mtcars[1:3, 1:2], caption = "Combine two tiny tables.")
y <- tt(mtcars[4:5, 8:10])

# rbind() does not support additional aarguments
# rbind2() supports additional arguments

# basic combination
rbind(x, y)

rbind(x, y) |> format_tt(replace = "")

# omit y header
rbind2(x, y, headers = FALSE)

# bind by position rather than column names
rbind2(x, y, use_names = FALSE)

# `iris` example with pre-tt() formatting
dat <- iris[1:3, 1:4]
a <- format_tt(dat, i = 1:3, digits = 1) |> tt()
b <- format_tt(dat, i = 1:3, digits = 2) |> tt()
rbind2(a, b)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/save_tt.R, R/zzz-aliases.R
\name{save_tt}
\alias{save_tt}
\alias{tt_save}
\title{Save a Tiny Table to File}
\usage{
save_tt(
  x,
  output = get_option("tinytable_save_output", default = NULL),
  overwrite = get_option("tinytable_save_overwrite", default = FALSE)
)

tt_save(
  x,
  output = get_option("tinytable_save_output", default = NULL),
  overwrite = get_option("tinytable_save_overwrite", default = FALSE)
)
}
\arguments{
\item{x}{The tinytable object to be saved.}

\item{output}{String or file path.
\itemize{
\item If \code{output} is "markdown", "latex", "html", or "typst", the table is returned in a string as an \code{R} object.
\item If \code{output} is "dataframe", the formatted table is returned as a data frame.
\item If \code{output} is a valid file path, the table is saved to file. The supported extensions are: .docx, .html, .png, .pdf, .tex, .typ, and .md (with aliases .txt, .Rmd and .qmd).
\item If \code{theme_html(portable = TRUE)} or \code{theme_typst(portable = TRUE)} is used,
images are included as base64 encoded strings instead of links to local files.
}}

\item{overwrite}{A logical value indicating whether to overwrite an existing file.}
}
\value{
A string with the table when \code{output} is a format, and the file path when \code{output} is a valid path.
}
\description{
This function saves an object of class tinytable to a specified file and format, with an option to overwrite existing files.
}
\section{Dependencies}{

\itemize{
\item \code{.pdf} output requires a full LaTeX installation on the local computer.
\item \code{.png} output requires the \code{webshot2} package.
\item \code{.html} self-contained files and \code{theme_typst(portable = TRUE)} require the \code{base64enc} package.
}
}

\section{LaTeX preamble}{


\code{tinytable} uses the \code{tabularray} package from your LaTeX distribution to draw tables. \code{tabularray}, in turn, uses the special \code{tblr}, \code{talltblr}, and \code{longtblr} environments.

When rendering a document from Quarto or Rmarkdown directly to PDF, \code{tinytable} will populate the LaTeX preamble automatically with all the required packages. For standalone LaTeX documents, these commands should be inserted in the preamble manually:

Note: Your document will fail to compile to PDF in Quarto if you enable caching and you use tinytable due to missing LaTeX headers. To avoid this problem, set the option \verb{#| cache: false} for the chunk(s) where you use tinytable.

\if{html}{\out{<div class="sourceCode latex">}}\preformatted{\\usepackage\{tabularray\}
\\usepackage\{float\}
\\usepackage\{graphicx\}
\\usepackage\{rotating\}
\\usepackage[normalem]\{ulem\}
\\UseTblrLibrary\{siunitx\}
\newcommand{\tinytableTabularrayUnderline}[1]\{\\underline\{#1\}\}
\newcommand{\tinytableTabularrayStrikeout}[1]\{\\sout\{#1\}\}
\\NewTableCommand\{\\tinytableDefineColor\}[3]\{\\definecolor\{#1\}\{#2\}\{#3\}\}
}\if{html}{\out{</div>}}
}

\section{Global options}{


Options can be set with \code{options()} and change the default behavior of tinytable. For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(tinytable_tt_digits = 4)
tt(head(iris))
}\if{html}{\out{</div>}}

You can set options in a script or via \code{.Rprofile}. Note: be cautious with \code{.Rprofile} settings as they may affect reproducibility.
\subsection{Default values for function arguments}{

Nearly all of the package's functions retrieve their default values from global options. This allows you to set defaults once and apply them to all tables without needing to specify them each time. For example, to fix the the \code{digits} argument of the \code{tt()} function globally, call:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(tinytable_tt_digits = 4)
}\if{html}{\out{</div>}}

In addition, some more specific options are available to control the behavior of the package in specific contexts.
\itemize{
\item \code{tinytable_html_mathjax}: Insert MathJax scripts (warning: may conflict if MathJax is loaded elsewhere)
\item \code{tinytable_pdf_clean}: Delete temporary and log files for pdf output in \code{save_tt()}
\item \code{tinytable_color_name_normalization}: Enable/disable automatic color name processing (default: TRUE). When enabled, R color names recognized by \code{col2rgb()} are converted to hex format for consistent rendering across HTML, LaTeX, and Typst formats. If R color conversion fails, LaTeX color names are used as fallback. Colors explicitly supplied as hex values with "#" prefix are passed through unchanged. Set to FALSE to disable processing and pass color names unchanged.
}
\subsection{Quarto}{

The \code{format_tt(quarto=TRUE)} argument enables Quarto data processing with some limitations:
\enumerate{
\item The \verb{\\QuartoMarkdownBase64\{\}} LaTeX macro may not process references and markdown as expected
\item Quarto processing may conflict with \code{tinytable} styling/formatting
}

Options:
\itemize{
\item \code{tinytable_quarto_disable_processing}: Disable Quarto cell processing
}

Example of Quarto-specific code in cells:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{x <- data.frame(Math = "x^2^", Citation = "@Lovelace1842")
fn <- function(z) sprintf("<span data-qmd='\%s'></span>", z)
tt(x) |> format_tt(i = 1, fn = fn)
}\if{html}{\out{</div>}}

For more details on Quarto table processing: https://quarto.org/docs/authoring/tables.html#disabling-quarto-table-processing
}

}
}

\examples{
library(tinytable)
x <- mtcars[1:4, 1:5]

fn <- file.path(tempdir(), "test.html")
tt(x) |> save_tt(fn, overwrite = TRUE)

library(tinytable)
filename <- file.path(tempdir(), "table.tex")
tt(mtcars[1:4, 1:4]) |> save_tt(filename)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/style_vector.R
\name{style_vector}
\alias{style_vector}
\title{Style a Vector with Text Formatting}
\usage{
style_vector(
  x,
  output = "html",
  bold = FALSE,
  italic = FALSE,
  monospace = FALSE,
  smallcap = FALSE,
  underline = FALSE,
  strikeout = FALSE,
  color = NULL,
  fontsize = NULL,
  indent = NULL
)
}
\arguments{
\item{x}{A vector to be styled.}

\item{output}{Output format for styling. One of "html", "latex", "typst", "markdown", "ansi". Defaults to "html".}

\item{bold}{Logical; if \code{TRUE}, text is styled in bold. Must be of length 1 or \code{length(x)}.}

\item{italic}{Logical; if \code{TRUE}, text is styled in italic. Must be of length 1 or \code{length(x)}.}

\item{monospace}{Logical; if \code{TRUE}, text is styled in monospace font. Must be of length 1 or \code{length(x)}.}

\item{smallcap}{Logical; if \code{TRUE}, text is styled in small caps. In Markdown output format, text is converted to uppercase. Must be of length 1 or \code{length(x)}.}

\item{underline}{Logical; if \code{TRUE}, text is underlined. Must be of length 1 or \code{length(x)}.}

\item{strikeout}{Logical; if \code{TRUE}, text has a strike through line. Must be of length 1 or \code{length(x)}.}

\item{color}{Text color. Must be of length 1 or \code{length(x)}. There are several ways to specify colors:
\itemize{
\item HTML:
\itemize{
\item Hex code composed of # and 6 characters, ex: #CC79A7.
\item Keywords: black, silver, gray, white, maroon, red, purple, fuchsia, green, lime, olive, yellow, navy, blue, teal, aqua
}
\item LaTeX:
\itemize{
\item Hex code composed of # and 6 characters, ex: "#CC79A7".
\item Keywords: black, blue, brown, cyan, darkgray, gray, green, lightgray, lime, magenta, olive, orange, pink, purple, red, teal, violet, white, yellow.
\item Color blending using xcolor, ex: \verb{white!80!blue}, \verb{green!20!red}.
\item Color names with luminance levels from the \code{ninecolors} package.
}
}}

\item{fontsize}{Font size in em units. Must be of length 1 or \code{length(x)}. Can be \code{NULL} for default size.}

\item{indent}{Text indentation in em units. Must be of length 1 or \code{length(x)}. Positive values only.}
}
\value{
A character vector with applied styling.
}
\description{
Style a Vector with Text Formatting
}
\details{
This function applies styling to a vector. It allows customization of text style (bold, italic, monospace), text color, font size, and text decorations (underline, strikeout). The styling is applied element-wise to the vector. Vectors are coerced with \code{as.character()} before styling.
}
\examples{
# Basic styling
style_vector(c("Hello", "World"), bold = TRUE, color = "red")

# Different styles per element
style_vector(
  c("Bold text", "Italic text", "Monospace"),
  bold = c(TRUE, FALSE, FALSE),
  italic = c(FALSE, TRUE, FALSE),
  monospace = c(FALSE, FALSE, TRUE)
)

# Single style applied to all elements
style_vector(c("A", "B", "C"), color = "blue", fontsize = 1.2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_striped.R
\name{theme_striped}
\alias{theme_striped}
\title{Striped theme with alternating row colors}
\usage{
theme_striped(x, ...)
}
\arguments{
\item{x}{A \code{tinytable} object.}

\item{...}{Additional arguments (ignored).}
}
\value{
A modified \code{tinytable} object.
}
\description{
Striped theme with alternating row colors
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format_tt.R, R/zzz-aliases.R
\name{format_tt}
\alias{format_tt}
\alias{tt_format}
\title{Format columns of a data frame}
\usage{
format_tt(
  x,
  i = NULL,
  j = NULL,
  digits = get_option("tinytable_format_digits", default = NULL),
  num_fmt = get_option("tinytable_format_num_fmt", default = "significant"),
  num_zero = get_option("tinytable_format_num_zero", default = FALSE),
  num_suffix = get_option("tinytable_format_num_suffix", default = FALSE),
  num_mark_big = get_option("tinytable_format_num_mark_big", default = ""),
  num_mark_dec = get_option("tinytable_format_num_mark_dec", default =
    getOption("OutDec", default = ".")),
  date = get_option("tinytable_format_date", default = NULL),
  bool = get_option("tinytable_format_bool", default = NULL),
  math = get_option("tinytable_format_math", default = FALSE),
  other = get_option("tinytable_format_other", default = NULL),
  replace = get_option("tinytable_format_replace", default = FALSE),
  escape = get_option("tinytable_format_escape", default = FALSE),
  markdown = get_option("tinytable_format_markdown", default = FALSE),
  quarto = get_option("tinytable_format_quarto", default = FALSE),
  fn = get_option("tinytable_format_fn", default = NULL),
  sprintf = get_option("tinytable_format_sprintf", default = NULL),
  linebreak = get_option("tinytable_format_linebreak", default = NULL),
  output = get_option("tinytable_format_output", default = NULL)
)

tt_format(
  x,
  i = NULL,
  j = NULL,
  digits = get_option("tinytable_format_digits", default = NULL),
  num_fmt = get_option("tinytable_format_num_fmt", default = "significant"),
  num_zero = get_option("tinytable_format_num_zero", default = FALSE),
  num_suffix = get_option("tinytable_format_num_suffix", default = FALSE),
  num_mark_big = get_option("tinytable_format_num_mark_big", default = ""),
  num_mark_dec = get_option("tinytable_format_num_mark_dec", default =
    getOption("OutDec", default = ".")),
  date = get_option("tinytable_format_date", default = NULL),
  bool = get_option("tinytable_format_bool", default = NULL),
  math = get_option("tinytable_format_math", default = FALSE),
  other = get_option("tinytable_format_other", default = NULL),
  replace = get_option("tinytable_format_replace", default = FALSE),
  escape = get_option("tinytable_format_escape", default = FALSE),
  markdown = get_option("tinytable_format_markdown", default = FALSE),
  quarto = get_option("tinytable_format_quarto", default = FALSE),
  fn = get_option("tinytable_format_fn", default = NULL),
  sprintf = get_option("tinytable_format_sprintf", default = NULL),
  linebreak = get_option("tinytable_format_linebreak", default = NULL),
  output = get_option("tinytable_format_output", default = NULL)
)
}
\arguments{
\item{x}{A data frame or a vector to be formatted.}

\item{i}{Numeric vector or string.
\itemize{
\item Numeric vector: Row indices where the styling should be applied. Can be a single value or a vector.
\item String: Table components to format "caption", "colnames", "groupi" (row group labels), "~groupi" (non-group rows), "groupj" (column group labels), "notes".
\item If both the \code{i} and \code{j} are omitted (default: NULL), formatting is applied to all table elements, including caption, notes, and group labels.
}}

\item{j}{Column indices where the styling should be applied. Can be:
\itemize{
\item Integer vectors indicating column positions.
\item Character vector indicating column names.
\item A single string specifying a Perl-style regular expression used to match column names.
\item Unquoted expression: Non-standard evaluation is supported. When supplying an unquoted expression, it is first evaluated in the calling environment, then in an environment that includes the columns of the original data passed to \code{tt()}, and \code{groupi} indices. See examples below.
}}

\item{digits}{Number of significant digits or decimal places, depending on the \code{num_fmt} argument.
\itemize{
\item \code{num_fmt = "significant"} (default): \code{digits} is the minimum number of significant digits. Formatting is applied column-wise, so the smallest value in the column determines the decimal representation for the whole column.
\item \code{num_fmt = "significant_cell"}: same as \code{"significant"}, but formatting is applied cell-by-cell. Each cell displays exactly \code{digits} significant figures, regardless of other values in the column.
\item \code{num_fmt = "decimal"}: \code{digits} is the number of decimal places (digits after the decimal point) for all values.
\item \code{num_fmt = "scientific"}: \code{digits} is the number of decimal places in the mantissa of scientific notation (e.g., \code{digits = 2} gives \code{1.23e+04}).
}}

\item{num_fmt}{The format for numeric values; one of 'significant', 'significant_cell', 'decimal', or 'scientific'. See the \code{digits} argument for details on how each format interprets the \code{digits} value.}

\item{num_zero}{Logical; if TRUE, trailing zeros are kept in "decimal" format (but not in "significant" format).}

\item{num_suffix}{Logical; if TRUE display short numbers with \code{digits} significant digits and K (thousands), M (millions), B (billions), or T (trillions) suffixes.}

\item{num_mark_big}{Character to use as a thousands separator.}

\item{num_mark_dec}{Decimal mark character. Default is the global option 'OutDec'.}

\item{date}{A string passed to the \code{format()} function, such as "\%Y-\%m-\%d". See the "Details" section in \code{?strptime}}

\item{bool}{A function to format logical columns. Defaults to title case.}

\item{math}{Logical. If TRUE, wrap cell values in math mode \verb{$..$}. This is useful for LaTeX output or with HTML MathJax \code{options(tinytable_html_mathjax=TRUE)}.}

\item{other}{A function to format columns of other types. Defaults to \code{as.character()}.}

\item{replace}{Logical, String or Named list of vectors
\itemize{
\item TRUE: Replace \code{NA} and \code{NaN} by an empty string.
\item FALSE: Print \code{NA} and \code{NaN} as strings.
\item String: Replace \code{NA} and \code{NaN} entries by the user-supplied string.
\item Named list: Replace matching elements of the vectors in the list by theirs names. Example:
\itemize{
\item \code{list("-" = c(NA, NaN), "Tiny" = -Inf, "Massive" = Inf)}
}
}}

\item{escape}{Logical or "latex" or "html". If TRUE, escape special characters to display them as text in the format of the output of a \code{tt()} table.
\itemize{
\item If \code{i} and \code{j} are both \code{NULL}, escape all cells, column names, caption, notes, and spanning labels created by \code{group_tt()}.
}}

\item{markdown}{Logical; if TRUE, render markdown syntax in cells. Ex: \verb{_italicized text_} is properly italicized in HTML and LaTeX.}

\item{quarto}{Logical. Enable Quarto data processing and wrap cell content in a \code{data-qmd} span (HTML) or \verb{\\QuartoMarkdownBase64\{\}} macro (LaTeX). See warnings in the Global Options section below.}

\item{fn}{Function for custom formatting. Accepts a vector and returns a character vector of the same length.}

\item{sprintf}{String passed to the \code{?sprintf} function to format numbers or interpolate strings with a user-defined pattern (similar to the \code{glue} package, but using Base R).}

\item{linebreak}{NULL or a single string. If it is a string, replaces that string with appropriate line break sequences depending on the output format (HTML: \verb{<br>}, LaTeX: \verb{\\\\\\\\}, Typst: \verb{\\\\ }). Markdown output is excluded from line break replacement.}

\item{output}{Apply formatting only if the \code{tt()} object is rendered in the specified format. A character vector of one or more of "latex", "html", "typst", or "markdown". If \code{NULL} (default), apply formatting regardless of the output format.}
}
\value{
A data frame with formatted columns.
}
\description{
This function formats the columns of a data frame based on the column type (logical, date, numeric).
It allows various formatting options like significant digits, decimal points, and scientific notation.
It also includes custom formatting for date and boolean values.
If this function is applied several times to the same cell, the last transformation is retained and the previous calls are ignored, except for the \code{escape} argument which can be applied to previously transformed data.
Each call to \code{format_tt()} is recorded and later replayed sequentially at render time, so later calls operate on the output of earlier ones.
Arguments are executed in a fixed order inside each call (see code below), so combining features such as \code{linebreak}, \code{escape}, \code{markdown}, etc. in a single call may produce surprising results.
Chain multiple calls when you need finer control (e.g., \code{format_tt(escape = TRUE) |> format_tt(linebreak = "<br>")}), keeping in mind that the second call will see the already-escaped text.
}
\section{Global options}{


Options can be set with \code{options()} and change the default behavior of tinytable. For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(tinytable_tt_digits = 4)
tt(head(iris))
}\if{html}{\out{</div>}}

You can set options in a script or via \code{.Rprofile}. Note: be cautious with \code{.Rprofile} settings as they may affect reproducibility.
\subsection{Default values for function arguments}{

Nearly all of the package's functions retrieve their default values from global options. This allows you to set defaults once and apply them to all tables without needing to specify them each time. For example, to fix the the \code{digits} argument of the \code{tt()} function globally, call:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(tinytable_tt_digits = 4)
}\if{html}{\out{</div>}}

In addition, some more specific options are available to control the behavior of the package in specific contexts.
\itemize{
\item \code{tinytable_html_mathjax}: Insert MathJax scripts (warning: may conflict if MathJax is loaded elsewhere)
\item \code{tinytable_pdf_clean}: Delete temporary and log files for pdf output in \code{save_tt()}
\item \code{tinytable_color_name_normalization}: Enable/disable automatic color name processing (default: TRUE). When enabled, R color names recognized by \code{col2rgb()} are converted to hex format for consistent rendering across HTML, LaTeX, and Typst formats. If R color conversion fails, LaTeX color names are used as fallback. Colors explicitly supplied as hex values with "#" prefix are passed through unchanged. Set to FALSE to disable processing and pass color names unchanged.
}
\subsection{Quarto}{

The \code{format_tt(quarto=TRUE)} argument enables Quarto data processing with some limitations:
\enumerate{
\item The \verb{\\QuartoMarkdownBase64\{\}} LaTeX macro may not process references and markdown as expected
\item Quarto processing may conflict with \code{tinytable} styling/formatting
}

Options:
\itemize{
\item \code{tinytable_quarto_disable_processing}: Disable Quarto cell processing
}

Example of Quarto-specific code in cells:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{x <- data.frame(Math = "x^2^", Citation = "@Lovelace1842")
fn <- function(z) sprintf("<span data-qmd='\%s'></span>", z)
tt(x) |> format_tt(i = 1, fn = fn)
}\if{html}{\out{</div>}}

For more details on Quarto table processing: https://quarto.org/docs/authoring/tables.html#disabling-quarto-table-processing
}

}
}

\examples{
dat <- data.frame(
  a = rnorm(3, mean = 10000),
  b = rnorm(3, 10000)
)
tab <- tt(dat)
format_tt(tab,
  digits = 2,
  num_mark_dec = ",",
  num_mark_big = " "
)

k <- tt(data.frame(x = c(0.000123456789, 12.4356789)))
format_tt(k, digits = 2, num_fmt = "significant_cell")

dat <- data.frame(
  a = c("Burger", "Halloumi", "Tofu", "Beans"),
  b = c(1.43202, 201.399, 0.146188, 0.0031),
  c = c(98938272783457, 7288839482, 29111727, 93945)
)
tt(dat) |>
  format_tt(j = "a", sprintf = "Food: \%s") |>
  format_tt(j = 2, digits = 1, num_fmt = "decimal", num_zero = TRUE) |>
  format_tt(j = "c", digits = 2, num_suffix = TRUE)

y <- tt(data.frame(x = c(123456789.678, 12435.6789)))
format_tt(y, digits = 3, num_mark_big = " ")

x <- tt(data.frame(Text = c("_italicized text_", "__bold text__")))
format_tt(x, markdown = TRUE)

# Line breaks using linebreak argument
d <- data.frame(Text = "First line<br>Second line")
tt(d) |> format_tt(linebreak = "<br>")

# Non-standard evaluation (NSE)
dat <- data.frame(
  w = c(143002.2092, 201399.181, 100188.3883),
  x = c(1.43402, 201.399, 0.134588),
  y = as.Date(c(897, 232, 198), origin = "1970-01-01"),
  z = c(TRUE, TRUE, FALSE)
)
tt(dat) |>
  format_tt(i = w > 150000, j = w, digits = 0, num_mark_big = ",")

tab <- data.frame(a = c(NA, 1, 2), b = c(3, NA, 5))
tt(tab) |> format_tt(replace = "-")

dat <- data.frame(
  "LaTeX" = c("Dollars $", "Percent \%", "Underscore _"),
  "HTML" = c("<br>", "<sup>4</sup>", "<emph>blah</emph>")
)
tt(dat) |> format_tt(escape = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_revealjs.R
\name{theme_revealjs}
\alias{theme_revealjs}
\title{RevealJS presentation theme}
\usage{
theme_revealjs(
  x,
  css = get_option("tinytable_revealjs_css", default = "light"),
  fontsize = get_option("tinytable_revealjs_fontsize", default = 0.8),
  fontsize_caption = get_option("tinytable_revealjs_fontsize_caption", default = 1)
)
}
\arguments{
\item{x}{A \code{tinytable} object.}

\item{css}{String. CSS theme: "light" (default) or "dark".}

\item{fontsize}{Numeric. Font size multiplier for table content.}

\item{fontsize_caption}{Numeric. Font size multiplier for table captions.}
}
\value{
A modified \code{tinytable} object.
}
\description{
RevealJS presentation theme
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format_vector.R
\name{format_vector}
\alias{format_vector}
\title{Format a Vector}
\usage{
format_vector(
  x,
  output = "html",
  digits = NULL,
  num_fmt = "significant",
  num_zero = FALSE,
  num_suffix = FALSE,
  num_mark_big = "",
  num_mark_dec = getOption("OutDec", default = "."),
  date = NULL,
  bool = NULL,
  math = FALSE,
  other = NULL,
  replace = FALSE,
  escape = FALSE,
  markdown = FALSE,
  quarto = FALSE,
  fn = NULL,
  sprintf = NULL,
  linebreak = NULL
)
}
\arguments{
\item{x}{A vector to be formatted.}

\item{output}{Output format. One of "html", "latex", "typst", "markdown", etc.}

\item{digits}{Number of significant digits or decimal places, depending on the \code{num_fmt} argument.
\itemize{
\item \code{num_fmt = "significant"} (default): \code{digits} is the minimum number of significant digits. Formatting is applied column-wise, so the smallest value in the column determines the decimal representation for the whole column.
\item \code{num_fmt = "significant_cell"}: same as \code{"significant"}, but formatting is applied cell-by-cell. Each cell displays exactly \code{digits} significant figures, regardless of other values in the column.
\item \code{num_fmt = "decimal"}: \code{digits} is the number of decimal places (digits after the decimal point) for all values.
\item \code{num_fmt = "scientific"}: \code{digits} is the number of decimal places in the mantissa of scientific notation (e.g., \code{digits = 2} gives \code{1.23e+04}).
}}

\item{num_fmt}{The format for numeric values; one of 'significant', 'significant_cell', 'decimal', or 'scientific'. See the \code{digits} argument for details on how each format interprets the \code{digits} value.}

\item{num_zero}{Logical; if TRUE, trailing zeros are kept in "decimal" format (but not in "significant" format).}

\item{num_suffix}{Logical; if TRUE display short numbers with \code{digits} significant digits and K (thousands), M (millions), B (billions), or T (trillions) suffixes.}

\item{num_mark_big}{Character to use as a thousands separator.}

\item{num_mark_dec}{Decimal mark character. Default is the global option 'OutDec'.}

\item{date}{A string passed to the \code{format()} function, such as "\%Y-\%m-\%d". See the "Details" section in \code{?strptime}}

\item{bool}{A function to format logical columns. Defaults to title case.}

\item{math}{Logical. If TRUE, wrap cell values in math mode \verb{$..$}. This is useful for LaTeX output or with HTML MathJax \code{options(tinytable_html_mathjax=TRUE)}.}

\item{other}{A function to format columns of other types. Defaults to \code{as.character()}.}

\item{replace}{Logical, String or Named list of vectors
\itemize{
\item TRUE: Replace \code{NA} and \code{NaN} by an empty string.
\item FALSE: Print \code{NA} and \code{NaN} as strings.
\item String: Replace \code{NA} and \code{NaN} entries by the user-supplied string.
\item Named list: Replace matching elements of the vectors in the list by theirs names. Example:
\itemize{
\item \code{list("-" = c(NA, NaN), "Tiny" = -Inf, "Massive" = Inf)}
}
}}

\item{escape}{Logical or "latex" or "html". If TRUE, escape special characters to display them as text in the format of the output of a \code{tt()} table.
\itemize{
\item If \code{i} and \code{j} are both \code{NULL}, escape all cells, column names, caption, notes, and spanning labels created by \code{group_tt()}.
}}

\item{markdown}{Logical; if TRUE, render markdown syntax in cells. Ex: \verb{_italicized text_} is properly italicized in HTML and LaTeX.}

\item{quarto}{Logical. Enable Quarto data processing and wrap cell content in a \code{data-qmd} span (HTML) or \verb{\\QuartoMarkdownBase64\{\}} macro (LaTeX). See warnings in the Global Options section below.}

\item{fn}{Function for custom formatting. Accepts a vector and returns a character vector of the same length.}

\item{sprintf}{String passed to the \code{?sprintf} function to format numbers or interpolate strings with a user-defined pattern (similar to the \code{glue} package, but using Base R).}

\item{linebreak}{NULL or a single string. If it is a string, replaces that string with appropriate line break sequences depending on the output format (HTML: \verb{<br>}, LaTeX: \verb{\\\\\\\\}, Typst: \verb{\\\\ }). Markdown output is excluded from line break replacement.}
}
\value{
A character vector with formatted values.
}
\description{
Format a Vector
}
\details{
This function formats a vector by passing it to \code{format_tt()}. All formatting arguments must be of length 1 or \code{length(x)}.
}
\examples{
# Format numeric vector
format_vector(c(1234.567, 9876.543), digits = 2, num_mark_big = ",")

# Format dates
dates <- as.Date(c("2023-01-01", "2023-12-31"))
format_vector(dates, date = "\%B \%d, \%Y")

# Format logical values
format_vector(c(TRUE, FALSE, TRUE), bool = function(x) ifelse(x, "Yes", "No"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{knit_print.tinytable}
\alias{knit_print.tinytable}
\title{Print a tinytable object in knitr}
\usage{
knit_print.tinytable(
  x,
  output = get_option("tinytable_print_output", default = NULL),
  ...
)
}
\value{
A string with class \code{knit_asis} to be printed in Rmarkdown or Quarto documents.
}
\description{
Print a tinytable object in knitr
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_tt.R, R/zzz-aliases.R
\name{group_tt}
\alias{group_tt}
\alias{tt_group}
\title{Spanning labels to identify groups of rows or columns}
\usage{
group_tt(
  x,
  i = getOption("tinytable_group_i", default = NULL),
  j = getOption("tinytable_group_j", default = NULL),
  ...
)

tt_group(
  x,
  i = getOption("tinytable_group_i", default = NULL),
  j = getOption("tinytable_group_j", default = NULL),
  ...
)
}
\arguments{
\item{x}{A data frame, data table, or tibble to be rendered as a table.}

\item{i}{Character vector, named list, or integer vector
\itemize{
\item A character vector of labels with length equal to the number of rows in \code{x}
\item A named list of row indices to group. The names of the list will be used as labels. The indices represent the position where labels should be inserted in the original table. For example,
\itemize{
\item \code{i=list("Hello"=5)}: insert the "Hello" label after the 4th row in the original table.
\item \code{i=list("Hello"=2, "World"=2)}: insert the two labels consecutively after the 1st row in the original table.
\item \code{i=list("Foo Bar"=0)}: insert the label in the first row after the header.
}
\item Vector of positive integers: For matrix insertion: \code{i} specifies row positions and \code{j} must be a character matrix to insert in the table (see below for details).
}}

\item{j}{String, named list, or character matrix
\itemize{
\item Named list of column indices to group, ex: \code{j=list("A"=1:2,"B"=3:6)}. The names of the list will be used as labels. See below for more examples. Note: empty labels must be a space: " ".
\item A single string when column names include the group name as a prefix, ex: group1_column1, group1_column2, etc.
\item Character matrix for inserting rows at positions specified by \code{i}. The matrix must have the same number of columns as the table, or be a single column with a number of elements that is a multiple of the table's column count (which will be automatically reshaped). Each row of the matrix matches an element
\item Unquoted expression: Non-standard evaluation is supported. When supplying an unquoted expression, it is first evaluated in the calling environment, then in the data frame passed to \code{tt()}.
}}

\item{...}{Other arguments are ignored.}
}
\value{
An object of class \code{tt} representing the table.
}
\description{
Spanning labels to identify groups of rows or columns

Alias for \code{group_tt()}
}
\details{
Warning: The \code{style_tt()} can normally be used to style the group headers, as expected, but that feature is not available for Markdown and Word tables.
}
\section{Markdown limitations}{


Markdown is a text-only format that only supports these styles: italic, bold, strikeout. The \code{width} argument is also unavailable.
These limitations exist because there is no standard markdown syntax for the other styling options.

However, in terminals (consoles) that support it, \code{tinytable} can display colors and text styles using
ANSI escape codes by setting \code{theme_markdown(ansi = TRUE)}. This allows for rich formatting in
compatible terminal environments.
}

\section{Word limitations}{


Word tables only support these styles: italic, bold, strikeout. The \code{width} argument is also unavailable.
Moreover, the \code{style_tt()} function cannot be used to style headers inserted by the \code{group_tt()} function;
instead, you should style the headers directly in the header definition using markdown syntax:
\code{group_tt(i = list("*italic header*" = 2))}. These limitations are due to the fact that we create
Word documents by converting a markdown table to .docx via the Pandoc software, which requires
going through a text-only intermediate format.
}

\examples{

# vector of row labels
dat <- data.frame(
  label = c("a", "a", "a", "b", "b", "c", "a", "a"),
  x1 = rnorm(8),
  x2 = rnorm(8)
)
tt(dat[, 2:3]) |> group_tt(i = dat$label)

# named lists of labels
tt(mtcars[1:10, 1:5]) |>
  group_tt(
    i = list(
      "Hello" = 3,
      "World" = 8
    ),
    j = list(
      "Foo" = 2:3,
      "Bar" = 4:5
    )
  )

dat <- mtcars[1:9, 1:8]
tt(dat) |>
  group_tt(i = list(
    "I like (fake) hamburgers" = 3,
    "She prefers halloumi" = 4,
    "They love tofu" = 7
  ))

tt(dat) |>
  group_tt(
    j = list(
      "Hamburgers" = 1:3,
      "Halloumi" = 4:5,
      "Tofu" = 7
    )
  )

x <- mtcars[1:5, 1:6]
tt(x) |>
  group_tt(j = list("Hello" = 1:2, "World" = 3:4, "Hello" = 5:6)) |>
  group_tt(j = list("Foo" = 1:3, "Bar" = 4:6))

# column names with delimiters
dat <- data.frame(
  A_id = 1,
  A_a1 = 2,
  A_a2 = "3",
  B_b1 = 4,
  B_b2 = 5,
  B_C = 6
)
tt(dat) |> group_tt(j = "_")

# matrix insertion
rowmat <- matrix(colnames(iris))
tt(head(iris, 7)) |>
  group_tt(i = c(2, 5), j = rowmat)

rowmat <- matrix(c(
  "a", "b", "c", "d", "e",
  1, 2, 3, 4, 5))
tt(head(iris, 7)) |>
  group_tt(i = 2, j = rowmat) |>
  style_tt(i = "groupi", background = "pink")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tt.R
\name{tt}
\alias{tt}
\alias{tt.default}
\title{Draw a Tiny Table}
\usage{
tt(x, ...)

\method{tt}{default}(
  x,
  digits = get_option("tinytable_tt_digits", default = NULL),
  caption = get_option("tinytable_tt_caption", default = NULL),
  notes = get_option("tinytable_tt_notes", default = NULL),
  width = get_option("tinytable_tt_width", default = NULL),
  height = get_option("tinytable_tt_height", default = NULL),
  theme = get_option("tinytable_tt_theme", default = "default"),
  colnames = get_option("tinytable_tt_colnames", default = TRUE),
  rownames = get_option("tinytable_tt_rownames", default = FALSE),
  escape = get_option("tinytable_tt_escape", default = FALSE),
  engine = NULL,
  ...
)
}
\arguments{
\item{x}{A data frame, data table, or tibble to be rendered as a table.}

\item{...}{Additional arguments are ignored}

\item{digits}{Number of significant digits to keep for numeric variables. When \code{digits} is an integer, \code{tt()} calls \code{format_tt(x, digits = digits)} before proceeding to draw the table. Note that only the \code{digits} argument is passed explicitly; all other \code{format_tt()} arguments use their default values, which can be influenced by global options (e.g., \code{options(tinytable_format_num_fmt = "decimal")}). When \code{digits} is \code{NULL} (the default), \code{format_tt()} is not called at all, so global \verb{tinytable_format_*} options have no effect unless the user calls \code{format_tt()} explicitly. Users who need more control can use the \code{format_tt()} function instead.}

\item{caption}{A string that will be used as the caption of the table. This argument should \emph{not} be used in Quarto or Rmarkdown documents. In that context, please use the appropriate chunk options.}

\item{notes}{Notes to append to the bottom of the table. This argument accepts several different inputs:
\itemize{
\item Single string insert a single note: \code{"blah blah"}
\item Multiple strings insert multiple notes sequentially: \code{list("Hello world", "Foo bar")}
\item A named list inserts a list with the name as superscript: \code{list("a" = list("Hello World"))}
\item A named list with positions inserts markers as superscripts inside table cells: \code{list("a" = list(i = 0:1, j = 2, text = "Hello World"))}
}}

\item{width}{Table or column width.
\itemize{
\item Single numeric value smaller than or equal to 1 determines the full table width, in proportion of line width.
\item Numeric vector of length equal to the number of columns in \code{x} determines the width of each column, in proportion of line width. If the sum of \code{width} exceeds 1, each element is divided by \code{sum(width)}. This makes the table full-width with relative column sizes.
}}

\item{height}{Row height in em units. Single numeric value greater than zero that determines the row height spacing.}

\item{theme}{Function or string.
\itemize{
\item String: grid, revealjs, rotate, striped, empty
\item Function: Applied to the \code{tinytable} object.
}}

\item{colnames}{\code{TRUE}, \code{FALSE}, or "label". If "label", use the \code{attr(x$col,"label")} attribute if available and fall back on column names otherwise.}

\item{rownames}{Logical. If \code{TRUE}, rownames are included as the first column}

\item{escape}{Logical. If \code{TRUE}, escape special characters in the table. Equivalent to \code{format_tt(tt(x), escape = TRUE)}.}

\item{engine}{Character string specifying the HTML engine: "tinytable", "bootstrap", or "tabulator". Equivalent to \code{theme_html(tt(x), engine = engine)}. When \code{NULL} (the default), the \code{tinytable_html_engine} global option is honored by \code{theme_html()}.}
}
\value{
An object of class \code{tt} representing the table.

The table object has S4 slots which hold information about the structure of the table. For example, the \code{table@group_index_i} slot includes the row indices for grouping labels added by \code{group_tt()}.

Warning: Relying on or modifying the contents of these slots is strongly discouraged. Their names and contents could change at any time, and the \code{tinytable} developers do not consider changes to the internal structure of the output object to be a "breaking  change" for versioning or changelog purposes.
}
\description{
The \code{tt} function renders a table in different formats with various styling options: HTML, Markdown, LaTeX, Word, PDF, PNG, or Typst. The table can be customized with additional functions:
\itemize{
\item \code{style_tt()}: style fonts, colors, alignment, etc.
\item \code{format_tt()}: format numbers, dates, strings, etc.
\item \code{group_tt()}: row or column group labels.
\item \code{save_tt()}: save the table to a file or return the table as a string.
\item \code{print()}: print to a specific format, ex: \code{print(x, "latex")}
\item \verb{theme_*()} functions apply a collection of format-specific or visual transformations to a \code{tinytable.}
}

\code{tinytable} attempts to determine the appropriate way to print the table based on interactive use, RStudio availability, and output format in RMarkdown or Quarto documents. Users can call \code{print(x, output="markdown")} to print the table in a specific format. Alternatively, they can set a global option: \code{options("tinytable_print_output"="markdown")}
}
\section{Dependencies}{

\itemize{
\item \code{.pdf} output requires a full LaTeX installation on the local computer.
\item \code{.png} output requires the \code{webshot2} package.
\item \code{.html} self-contained files and \code{theme_typst(portable = TRUE)} require the \code{base64enc} package.
}
}

\section{LaTeX preamble}{


\code{tinytable} uses the \code{tabularray} package from your LaTeX distribution to draw tables. \code{tabularray}, in turn, uses the special \code{tblr}, \code{talltblr}, and \code{longtblr} environments.

When rendering a document from Quarto or Rmarkdown directly to PDF, \code{tinytable} will populate the LaTeX preamble automatically with all the required packages. For standalone LaTeX documents, these commands should be inserted in the preamble manually:

Note: Your document will fail to compile to PDF in Quarto if you enable caching and you use tinytable due to missing LaTeX headers. To avoid this problem, set the option \verb{#| cache: false} for the chunk(s) where you use tinytable.

\if{html}{\out{<div class="sourceCode latex">}}\preformatted{\\usepackage\{tabularray\}
\\usepackage\{float\}
\\usepackage\{graphicx\}
\\usepackage\{rotating\}
\\usepackage[normalem]\{ulem\}
\\UseTblrLibrary\{siunitx\}
\newcommand{\tinytableTabularrayUnderline}[1]\{\\underline\{#1\}\}
\newcommand{\tinytableTabularrayStrikeout}[1]\{\\sout\{#1\}\}
\\NewTableCommand\{\\tinytableDefineColor\}[3]\{\\definecolor\{#1\}\{#2\}\{#3\}\}
}\if{html}{\out{</div>}}
}

\section{Markdown limitations}{


Markdown is a text-only format that only supports these styles: italic, bold, strikeout. The \code{width} argument is also unavailable.
These limitations exist because there is no standard markdown syntax for the other styling options.

However, in terminals (consoles) that support it, \code{tinytable} can display colors and text styles using
ANSI escape codes by setting \code{theme_markdown(ansi = TRUE)}. This allows for rich formatting in
compatible terminal environments.
}

\section{Word limitations}{


Word tables only support these styles: italic, bold, strikeout. The \code{width} argument is also unavailable.
Moreover, the \code{style_tt()} function cannot be used to style headers inserted by the \code{group_tt()} function;
instead, you should style the headers directly in the header definition using markdown syntax:
\code{group_tt(i = list("*italic header*" = 2))}. These limitations are due to the fact that we create
Word documents by converting a markdown table to .docx via the Pandoc software, which requires
going through a text-only intermediate format.
}

\section{Tabulator (interactive tables)}{


\strong{Experimental Feature:} The Tabulator.js integration is experimental and the API may change in future versions.

The Tabulator.js library provides powerful interactive table features including sorting, filtering, pagination, data export, and real-time editing capabilities. This theme customizes the appearance and behavior of Tabulator tables.

\strong{Features:}
\itemize{
\item Sorting and filtering of all columns
\item Pagination with configurable page sizes
\item Search functionality across all columns
\item Multiple CSS themes and custom styling
\item Real-time data export options
\item Accessibility features (ARIA compliant)
}

\strong{Limitations:}
\itemize{
\item \code{style_tt()} supports most styling options (bold, italic, color, background, fontsize, etc.) with cell-level precision. Styles persist across sorting and pagination. Column-wide alignment only (row-specific \code{align}/\code{alignv} with \code{i} argument not supported)
\item Row-based formatting (\code{format_tt()} with \code{i} argument) not supported
\item Global stylesheets affect all tables in multi-table documents
\item Date formatting uses Luxon tokens, not R's \code{strptime} format
\item Boolean formatting requires \code{format_tt()} with \code{bool} argument for custom display
}
}

\section{Global options}{


Options can be set with \code{options()} and change the default behavior of tinytable. For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(tinytable_tt_digits = 4)
tt(head(iris))
}\if{html}{\out{</div>}}

You can set options in a script or via \code{.Rprofile}. Note: be cautious with \code{.Rprofile} settings as they may affect reproducibility.
\subsection{Default values for function arguments}{

Nearly all of the package's functions retrieve their default values from global options. This allows you to set defaults once and apply them to all tables without needing to specify them each time. For example, to fix the the \code{digits} argument of the \code{tt()} function globally, call:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(tinytable_tt_digits = 4)
}\if{html}{\out{</div>}}

In addition, some more specific options are available to control the behavior of the package in specific contexts.
\itemize{
\item \code{tinytable_html_mathjax}: Insert MathJax scripts (warning: may conflict if MathJax is loaded elsewhere)
\item \code{tinytable_pdf_clean}: Delete temporary and log files for pdf output in \code{save_tt()}
\item \code{tinytable_color_name_normalization}: Enable/disable automatic color name processing (default: TRUE). When enabled, R color names recognized by \code{col2rgb()} are converted to hex format for consistent rendering across HTML, LaTeX, and Typst formats. If R color conversion fails, LaTeX color names are used as fallback. Colors explicitly supplied as hex values with "#" prefix are passed through unchanged. Set to FALSE to disable processing and pass color names unchanged.
}
\subsection{Quarto}{

The \code{format_tt(quarto=TRUE)} argument enables Quarto data processing with some limitations:
\enumerate{
\item The \verb{\\QuartoMarkdownBase64\{\}} LaTeX macro may not process references and markdown as expected
\item Quarto processing may conflict with \code{tinytable} styling/formatting
}

Options:
\itemize{
\item \code{tinytable_quarto_disable_processing}: Disable Quarto cell processing
}

Example of Quarto-specific code in cells:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{x <- data.frame(Math = "x^2^", Citation = "@Lovelace1842")
fn <- function(z) sprintf("<span data-qmd='\%s'></span>", z)
tt(x) |> format_tt(i = 1, fn = fn)
}\if{html}{\out{</div>}}

For more details on Quarto table processing: https://quarto.org/docs/authoring/tables.html#disabling-quarto-table-processing
}

}
}

\section{Order of Operations}{


The specified order of operations used when building tables is defined in the \code{build.R} file, which can be viewed on Github: https://github.com/vincentarelbundock/tinytable/tree/main/R

A few important things to note:
\itemize{
\item The \code{i} and \code{j} indices in \code{format_tt()} and \code{style_tt()} refer to the table structure \emph{after} any grouping operations have been applied.
\item The \code{theme} argument in \code{tt()} is applied before any other theme, and pre-empts the default theme.
\item Theme functions apply transformations immediately when they are called, but they can delay the execution of certain operations using internal helpers: \code{build_prepare()} and \code{build_finalize()}. See \verb{theme_*()} files in the Github repository for examples.
\item \code{rbind()} and \code{rbind2()} combine two \code{tinytable} objects before formatting. In some cases, this has the undesirable consequence of coercing numeric variables to character, which prevents further numeric formatting. To avoid this, users can apply \code{format_tt()} directly to the data frames before calling \code{tt()}.
}
}

\examples{
library(tinytable)
x <- mtcars[1:4, 1:5]

tt(x)

tt(x,
  theme = "striped",
  width = 0.5,
  caption = "Data about cars."
)

tt(x, notes = "Hello World!")

fn <- list(i = 0:1, j = 2, text = "Hello World!")
tab <- tt(x, notes = list("*" = fn))
print(tab, "latex")

k <- data.frame(x = c(0.000123456789, 12.4356789))
tt(k, digits = 2)

# use variable labels stored in attributes as column names
dat = mtcars[1:5, c("cyl", "mpg", "hp")]
attr(dat$cyl, "label") <- "Cylinders"
attr(dat$mpg, "label") <- "Miles per Gallon"
attr(dat$hp, "label") <- "Horse Power"
tt(dat, colnames = "label")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_rotate.R
\name{theme_rotate}
\alias{theme_rotate}
\title{Rotate table or cell content}
\usage{
theme_rotate(
  x,
  angle = get_option("tinytable_rotate_angle", default = 90),
  i = NULL,
  j = NULL,
  ...
)
}
\arguments{
\item{x}{A \code{tinytable} object.}

\item{angle}{Numeric. Rotation angle in degrees (0-360).}

\item{i}{Row indices for cell rotation. When \code{NULL} (and \code{j} is also \code{NULL}),
rotates the entire table in LaTeX/Typst or all cells in HTML.}

\item{j}{Column indices for cell rotation. When \code{NULL} (and \code{i} is also \code{NULL}),
rotates the entire table in LaTeX/Typst or all cells in HTML.}

\item{...}{Additional arguments (ignored).}
}
\value{
A modified \code{tinytable} object.
}
\description{
In HTML, rotation is always applied to cell content (not the whole table) to
keep borders intact. In LaTeX and Typst, when \code{i} and \code{j} are both \code{NULL},
the entire table is rotated; otherwise cell content is rotated.
}
